using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System.Xml;
using System.IO;
using Util;

namespace VolumeViewer
{
    class Section
    {
        /// <summary>
        /// The number of this section in the sequence
        /// </summary>
        public int Number; 

        public string path;
        protected System.Collections.Generic.SortedList<string, Tile> Tiles = new SortedList<string, Tile>();

        public Rectangle Bounds;
        public Volume volume;

        /// <summary>
        /// These transforms are not used directly, but are rendered with the draw method to display how the section has been warped
        /// </summary>
        private System.Collections.Generic.SortedList<string, GridTransform> ReferenceTransforms = new SortedList<string, GridTransform>();

        public System.Collections.Generic.SortedList<string, TileGridTransform[]> Transforms = new SortedList<string, TileGridTransform[]>();

        public string VolumeTransform;  //The name of the transform used to create the volume

        /// <summary>
        /// Name and path of available tile types
        /// </summary>
        public SortedList<string, string> AvailableTileTypes = new SortedList<string, string>();

        public override string ToString()
        {
            return "Section " + Number.ToString(); 
        }

        public override int GetHashCode()
        {
            return Number; 
        }

        protected string _CurrentTransform;
        public string CurrentTransform
        {
            get { return _CurrentTransform; }
            set
            {
                Debug.Assert(value != null); 
                if (value == null)
                    return; 
//                if (_CurrentTransform == value)
//                    return;

                if(Transforms.ContainsKey(value) == false)
                    return;

                _CurrentTransform = value;

                //Initialize all of the tile verticies
                foreach(TileGridTransform T in this.Transforms[_CurrentTransform])
                {
                    Tile tile = this.Tiles[T.filename];

                    tile.CalculateVerticies(T); 
                }
            }
        }

        public Section(Volume vol)
        {
            this.volume = vol; 
        }

        public Section(Volume vol, string path, XmlReader reader) : this(vol)
        {
            this.path = path;

            reader.Read();

            this.Number = System.Convert.ToInt32(reader.GetAttribute("number"));

            while (reader.Read())
            {
                if (reader.NodeType != XmlNodeType.Element)
                    continue; 

                switch (reader.LocalName)
                {
                    case "transform":
                        string mosaicPath = this.path + '/' + reader.GetAttribute("path");
                        System.Net.HttpWebRequest request = (System.Net.HttpWebRequest)System.Net.HttpWebRequest.CreateDefault(new Uri(mosaicPath));
                        request.CachePolicy = new System.Net.Cache.RequestCachePolicy(System.Net.Cache.RequestCacheLevel.Revalidate); 
                        request.AutomaticDecompression = System.Net.DecompressionMethods.Deflate | System.Net.DecompressionMethods.GZip; 
                        System.Net.WebResponse response = request.GetResponse();
                        Debug.WriteLine(mosaicPath);
                        Stream MosaicData = response.GetResponseStream();
                        string[] MosaicLines = Util.Common.StreamToLines(MosaicData);

                        TileGridTransform[] tileTransforms = LoadMosaic(this.path, MosaicLines);

                        string Key = reader.GetAttribute("name");
                        Transforms.Add(Key, tileTransforms);

                        if (Key.ToLower().Contains("grid"))
                        {
                            Transforms.Add("Volume", tileTransforms);
                            VolumeTransform = "Volume";
                            this._CurrentTransform = "Volume";
                        }

                        if (this._CurrentTransform == null)
                        {
                            this._CurrentTransform = Key; 
                        }

                        MosaicData.Close(); 
                        response.Close(); 
                        break;
                    case "Pyramid":
                        string name = reader.GetAttribute("name");
                        string Pyramidpath = reader.GetAttribute("path");

                        AvailableTileTypes.Add(name, Pyramidpath); 
                        break;
                }
            }
        }

        public Section(Volume vol, string path) : this(vol)
        {
            this.path = path;
            string filename = System.IO.Path.GetFileNameWithoutExtension(path);

            //We expect the first part of the filename to be a section #:
            int iSectionNumberEnd = 0;
            for (int i = 0; i < filename.Length; i++)
            {
                if (char.IsDigit(filename[i]) == false)
                    break;
                iSectionNumberEnd = i;
            }

            string sectionNumber = filename;
            if(iSectionNumberEnd + 1 < filename.Length)
                sectionNumber = filename.Remove(iSectionNumberEnd + 1);


            try
            {
                this.Number = System.Convert.ToInt32(sectionNumber);
            }
            catch (Exception e)
            {
                System.Windows.Forms.MessageBox.Show("The name of each directory in a volume must start with a number indicating which section the directory contains.\n" + 
                                                     "This directory did not have a section number: " + path);
                throw e; 
                                                        
            }

            LoadLocal(path);
        }

        protected void LoadLocal(string path)
        {
            //List directories under the section, each directory name is an available tile type
            string[] fullPathModes = System.IO.Directory.GetDirectories(this.path);

            for (int i = 0; i < fullPathModes.Length; i++)
            {
                string mode = System.IO.Path.GetFileNameWithoutExtension(fullPathModes[i]);
                AvailableTileTypes.Add(mode, mode); 
            }



            // Try to find a grid mosaic
            string[] availableMosaics = System.IO.Directory.GetFiles(path, "*.mosaic");

            bool LoadDefault = false;
            foreach(string file in availableMosaics)
            {
                string Key = System.IO.Path.GetFileNameWithoutExtension(file);

                string[] mosaicLines = System.IO.File.ReadAllLines(file);
                TileGridTransform[] tileTransforms = LoadMosaic(System.IO.Path.GetDirectoryName(file), mosaicLines);
                Transforms.Add(Key, tileTransforms);

                if (file.ToLower().Contains("grid"))
                {
                    Transforms.Add("Volume", tileTransforms); 
                    VolumeTransform = "Volume"; 
                    this._CurrentTransform = "Volume";
                    LoadDefault = true;
                }
            }

            if(!LoadDefault)
            {
                this.CurrentTransform = Transforms.Keys[0];
            }
                
        }

        /// <summary>
        /// Load mosaic from specified file and add it to transforms list using specified key
        /// </summary>
        /// <param name="file"></param>
        /// <param name="Key"></param>
        protected TileGridTransform[] LoadMosaic(string path, string[] mosaic)
        {
            string[] parts = mosaic[0].Split(':');
            int numTiles = System.Convert.ToInt32(parts[1]);

            TileGridTransform[] tileTransforms = new TileGridTransform[numTiles];

            int iTileStart = 3; 
            for (int i = 0; i < mosaic.Length; i++)
            {
                string line = mosaic[i];
                if (line.StartsWith("image:"))
                {
                    iTileStart = i;
                    break;
                }
            }

            Debug.WriteLine("Loading " + numTiles.ToString() + " tiles");
            for (int i = iTileStart; i < mosaic.Length; i++)
            {
                string line = mosaic[i];
                //Debug.WriteLine(line);

                tileTransforms[i - iTileStart] = new TileGridTransform(path, line);
                tileTransforms[i - iTileStart].ControlSection = i - iTileStart;
                tileTransforms[i - iTileStart].MappedSection = i - iTileStart;
                Tile t = new Tile(this, tileTransforms[i - iTileStart].filename);
                      
                //The tiles list contains all possible tiles, so add the tile if it doesn't already exist
                if(Tiles.ContainsKey(t.filename) == false)
                    Tiles.Add(t.filename, t);
            }

            Rectangle R = CalculateBounds(tileTransforms); 

            foreach (GridTransform T in tileTransforms)
            {
                //Adjusting to center here breaks vclume tranformation since volume transforms assume mosaic origin of 0,0
                //T.Translate(new Vector2(-R.X - (R.Width / 2), -R.Y - (R.Height / 2))); 
                T.Translate(new Vector2(-R.X , -R.Y )); 
            }

            R = CalculateBounds(tileTransforms); 

            //Add transform list to our collection of transforms
            return tileTransforms;
        }

        protected Rectangle CalculateBounds(GridTransform[] transforms)
        {
            Single minX = Single.MaxValue;
            Single minY = Single.MaxValue;
            Single maxX = Single.MinValue;
            Single maxY = Single.MinValue; 

            foreach(GridTransform T in transforms)
            {
                Rectangle R = T.CalculateControlBounds();

                if (R.X < minX)
                    minX = R.X;
                if (R.X + R.Width > maxX)
                    maxX = R.X + R.Width;
                if (R.Y < minY)
                    minY = R.Y;
                if (R.Y + R.Height > maxY)
                    maxY = R.Y + R.Height;
            }

            return new Rectangle((int)Math.Floor(minX), (int)Math.Floor(minY), (int)Math.Ceiling(maxX - minX), (int)Math.Ceiling(maxY - minY)); 
        }

        /// <summary>
        /// Display modes the section supports such as 8-bit, 16-bit,clahe etc...
        /// </summary>
        public string[] Modes
        {
            get
            {
                IList<string> values = AvailableTileTypes.Values;

                string[] _modes = new string[values.Count]; 

                for(int i = 0; i < _modes.Length; i++)
                {
                    _modes[i] = values[i]; 
                }

                return _modes; 
            }

        }

        /// <summary>
        /// Adds a new transform to the section that maps it to the volume space. Returns 
        /// true if section had a grid-refine mapping that could be mapped. Otherwise false
        /// </summary>
        /// <param name="transform">Tranformation to appy. If null is passed then a copy is made unmodified and added under the volume name</param>
        /// <param name="volumeName">Name of the transform</param>
        public bool AddVolumeTransform(StosGridTransform transform, string volumeName)
        {
            if(this.VolumeTransform == null)
                return false;

            TileGridTransform[] volTransforms = this.Transforms[this.VolumeTransform]; 
            List<TileGridTransform> volTranCopy = new List<TileGridTransform>(volTransforms.Length); 

            foreach(TileGridTransform T in volTransforms)
            {
                TileGridTransform copy = (TileGridTransform)T.Copy();
                
                if(transform != null)
                    copy.Add(transform); 

                volTranCopy.Add(copy); 
            }
            
            this.Transforms.Add(volumeName, volTranCopy.ToArray());
            if(transform != null)
                ReferenceTransforms.Add(volumeName, (GridTransform)transform.Copy()); 

            return true;
        }

        public void Transform(StosGridTransform transform, string transformName)
        {
            foreach (GridTransform T in this.Transforms[transformName])
            {
                T.Add(transform);
            }

            this.Bounds = CalculateBounds(this.Transforms[transformName]);
        }

        public void FreeTextures()
        {
            foreach(Tile T in Tiles.Values)
            {
                T.FreeTexture();
            }

            System.GC.Collect();
        }

        public void Draw(GraphicsDevice graphicsDevice, Quad VisibleBounds, double DownSample, BasicEffect basicEffect)
        {
            
  //          Debug.WriteLine("Visible Bounds: " + VisibleBounds.ToString());

            //Setup a larger boundary outside of which we release textures
            Quad releaseBounds = VisibleBounds;
            Quad loadBounds = VisibleBounds;
            releaseBounds.Scale(3);
            loadBounds.Scale(1.5f);

            foreach(TileGridTransform T in Transforms[_CurrentTransform])
            {
                if (VisibleBounds.Contains(T.CachedControlBounds))
                {
                    Tile tile = Tiles[T.filename];
                    tile.Draw(graphicsDevice, DownSample, basicEffect);
                }
                else if (loadBounds.Contains(T.CachedControlBounds))
                {
                    Tile tile = Tiles[T.filename];
                    //Get ready by loading a smaller texture in case the user scrolls this direction 
                    int predictiveDownsample = DownSample * 2 > 64 ? 64 : (int)DownSample * 2;
                    tile.GetTexture(graphicsDevice, predictiveDownsample); 
                }
                else if (false == releaseBounds.Contains(T.CachedControlBounds))
                {
                    Tile tile = Tiles[T.filename];
                    tile.FreeTexture();
                }
            }

            if (Controller.ShowMesh && ReferenceTransforms.ContainsKey(CurrentTransform))
            {
                GridTransform T = ReferenceTransforms[CurrentTransform];
                T.Draw(graphicsDevice, basicEffect); 
            }
        }        
    }
}
