using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.IO;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System.Threading;

using Util; 

namespace VolumeViewer
{
    class Tile
    {
        public string filename;
        int currentDownsample = int.MaxValue;

        public VertexPositionNormalTexture[] verticies;
        Microsoft.Xna.Framework.Graphics.Texture2D texture;

        TextureReader TexReader; 

        Section section;

        bool ReadRequestInProgress = false;
        ManualResetEvent ReadCompleteEvent = new ManualResetEvent(false); 

        List<int> gridIndicies;

        public Tile(Section section, string filename)
        {
            this.section = section;
            this.filename = filename;
        }

        public void CalculateVerticies(TileGridTransform transform)
        {
            this.verticies = transform.GetVerticies();

            this.gridIndicies = new List<int>();

            for (int i = 0; i < verticies.Length; i++)
            {
                this.gridIndicies.Add(i);
            }
        }

        public override string ToString()
        {
            return filename;
        }

        public void FreeTexture()
        {
            if (this.texture != null)
            {
      //          Monitor.Enter(this);
  //              Debug.WriteLine("Drop Texture" + currentDownsample.ToString("D3") + " " + filename);

                //TODO: Put a dispose call on a seperate thread
  //              this.texture.Dispose();
                this.texture = null;
                this.currentDownsample = int.MaxValue;
      //          Monitor.Exit(this);
            }   
        }

        public Texture2D GetTexture(GraphicsDevice graphicsDevice, int Level)
        {
           //If we have no texture read the smallest texture first synchronously to get moving
           if (texture != null && this.currentDownsample <= Level)
           {
               //Drop any outstanding requests since we seem to be in the right spot
               this.TexReader = null; 

               return this.texture; 
           }

           //We are trying to read a new texture
           if (TexReader != null)
           {
               if (TexReader.FinishedReading)
               {
                   bool Entered = Monitor.TryEnter(TexReader);
                   if (Entered)
                   {
                       if (this.texture != null)
                       {
                           this.texture.Dispose();
                           this.texture = null;
                       }

                       this.texture = TexReader.Result;
                       this.currentDownsample = TexReader.Downsample;
                       Monitor.Exit(TexReader);

                       this.TexReader = null;
                   }
//                   Debug.WriteLine("AsSynch load completed " + Level.ToString("D3") + " " + filename);
               }
           }
           else //No texture and no reader
           {
               string filePath = section.path + '\\' + Controller.CurrentMode + '\\' + Level.ToString("D3") + '\\' + filename;

 //              Debug.WriteLine("ASynch request for " + Level.ToString("D3") + " " + filename);
               this.ReadCompleteEvent.Reset(); 
               this.TexReader = new TextureReader(graphicsDevice, filePath, Level, this.ReadCompleteEvent);

               ThreadPool.QueueUserWorkItem(TexReader.ThreadPoolCallback);

           }

           return this.texture; // Can return null or a texture with a different downsample level if we have one
            /*
               
               System.IO.FileInfo info = new FileInfo(filePath);
               int bufferSize = (int)info.Length;
               int streamBufferSize = (int)info.Length;

               FileStream stream;
               try
               {
                   stream = new FileStream(filePath, FileMode.Open, FileAccess.Read, FileShare.Read, streamBufferSize, FileOptions.Asynchronous | FileOptions.SequentialScan);
               }
               catch (Exception e)
               {
                   FileNotFoundException fileExcept = e as FileNotFoundException;
                   if (fileExcept != null)
                   {
                       Debug.WriteLine("Texture file not found!", "Tile");
                       return null;
                   }
                   else
                   {
                       throw e;
                   }
               }
               
               byte[] fileBuffer = new byte[bufferSize];
               IAsyncResult ar = stream.BeginRead(fileBuffer,
                                0,
                                fileBuffer.Length,
                                new AsyncCallback(this.FileReadComplete),
                                new object[] { stream, graphicsDevice, fileBuffer, Level });

               ReadRequestInProgress = true; 

               Debug.WriteLine("ASynch request for " + Level.ToString("D3") + " " + filename);

               if (ar.CompletedSynchronously)
               {
                   MemoryStream MemStream = new MemoryStream(fileBuffer, false);
                   if (texture != null)
                   {
                       texture.Dispose();
                       texture = null;
                   }

                   Texture2D oldTexture = texture;
                   TextureCreationParameters par = TextureCreationParameters.Default;
                   par.Filter = FilterOptions.Linear; 
                   texture = Texture2D.FromFile(graphicsDevice, MemStream, par);

                   if (oldTexture != null)
                   {
                       oldTexture.Dispose();
                       oldTexture = null;
                   }

                   stream.Close();
                   stream.Dispose();
                   MemStream.Close();
                   MemStream.Dispose();
                   fileBuffer = null; 
                   Debug.WriteLine("Synch load for " + Level.ToString("D3") + " " + filename);
                   ReadRequestInProgress = false;
               }

               //texture = Texture2D.FromFile(graphicsDevice, filePath);
               Monitor.Exit(this);
                */
                              
        }

        public void FileReadComplete(IAsyncResult ar)
        {
            Debug.Assert(ar.IsCompleted, "Wasn't expecting Asynch read not to complete successfully");
            if (!ar.IsCompleted)
            {
                ReadRequestInProgress = false;
                return;
            }

            object[] objs = ar.AsyncState as object[];
            FileStream stream = objs[0] as FileStream;
            GraphicsDevice graphicsDevice = objs[1] as GraphicsDevice;
            byte[] fileBuffer = objs[2] as byte[];
            int Level = (int)objs[3];

            stream.Close();
            stream.Dispose();
            stream = null;

            MemoryStream MemStream = new MemoryStream(fileBuffer, false);

            Monitor.Enter(this);

            Texture oldTexture = texture;
           // TextureCreationParameters par = TextureCreationParameters.Default;
         //   par.Filter = FilterOptions.Point;
            texture = Texture2D.FromFile(graphicsDevice, MemStream);

            if (oldTexture != null)
            {
                oldTexture.Dispose();
                oldTexture = null;
            }

            MemStream.Close();
            MemStream.Dispose();
            MemStream = null; 
            
            this.currentDownsample = Level;

            Debug.WriteLine("AsSynch load completed " + Level.ToString("D3") + " " + filename);

            ReadRequestInProgress = false;

            Monitor.Exit(this);
        }

        public void Draw(GraphicsDevice graphicsDevice, double DownSample, BasicEffect basicEffect)
        {
            if (gridIndicies == null)
            {
                Debug.WriteLine("No Grid Indicies for " + this.filename);
                return;
            }

            Texture2D currentTexture = GetTexture(graphicsDevice, (int)DownSample);

            basicEffect.Texture = currentTexture;
            basicEffect.TextureEnabled = null != currentTexture;    //Draw white if we don't have a texture
            basicEffect.VertexColorEnabled = false; 
                        
            basicEffect.CommitChanges();
            
            basicEffect.Begin();

            foreach (EffectPass pass in basicEffect.CurrentTechnique.Passes)
            {
                pass.Begin();

                if(gridIndicies.Count > 0)

                   graphicsDevice.DrawUserIndexedPrimitives<VertexPositionNormalTexture>(PrimitiveType.TriangleList, this.verticies, 0, verticies.Length, gridIndicies.ToArray(), 0, gridIndicies.Count / 3);
                    //  GraphicsDevice.DrawUserIndexedPrimitives<VertexPositionNormalTexture>(PrimitiveType.TriangleList, GridVerticies, 0, GridVerticies.Length, GridIndicies, 2, 2); 
                   // graphicsDevice.DrawUserIndexedPrimitives<VertexPositionNormalTexture>(PrimitiveType.TriangleList, this.verticies, 0, verticies.Length, gridIndicies, 0, 1);
                    
                pass.End();
            }

            basicEffect.End();

//            Monitor.Exit(this);
        }
    }
}
