using System;
using System.IO;
using System.Net;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System.Xml;
using Util;
using NGVV.Controls; 

namespace VolumeViewer 
{
    /// <summary>
    /// Collection of volumes, sections and tiles. There is only one dataset loaded at a time
    /// </summary>
    class Volume : INGVVDraw
    {
        public string Name = ""; 
        /// <summary>
        /// Keep the textures from this many sections +/- the current section in memory
        /// </summary>
        const int SectionsCached = 3; 
        public string path;
        public SortedList<int, Section> sections = new SortedList<int,Section>();
        int _SectionIndex = 0;

        /// <summary>
        /// Sorted list containing the transforms for each volume transform we find
        /// Key = Downsample level
        /// </summary>
        public SortedList<string, StosGridTransform[]> Transforms = new SortedList<string, StosGridTransform[]>();

        protected string _CurrentTransform;
        public string CurrentTransform
        {
            get { return _CurrentTransform; }
            set
            {
                if (_CurrentTransform == value)
                    return;

                Debug.Assert(Transforms.ContainsKey(value), "Requested transform not available to volume");

                _CurrentTransform = value;

                //Initialize all of the sections verticies
            }
        }

        public int SectionIndex
        {
            get { return _SectionIndex; }
            set
            {
                if (value < 0)
                    _SectionIndex = 0;
                else if (value >= sections.Count)
                    _SectionIndex = sections.Count - 1;
                else
                {
                    _SectionIndex = value;

                    //Free texture memory from non-visible sections
                    if(_SectionIndex - (SectionsCached + 1) > 0)
                        sections.Values[_SectionIndex - (SectionsCached + 1)].FreeTextures();
                    if(_SectionIndex + (SectionsCached + 1) < sections.Count)
                        sections.Values[_SectionIndex + (SectionsCached + 1)].FreeTextures(); 
                }
            }
        }

        public int NumSections
        {
            get { return sections.Count; }
        }

        public Section CurrentSection
        {
            get
            {
                if (sections == null)
                    return null;

                return sections.Values[_SectionIndex];
            }
        }

        public Volume(string path)
        {
            this.path = path; 
            Load(path);
            
        }

        /// <summary>
        /// Loads a path, determines whether path refers to XML file or a local directory
        /// </summary>
        /// <param name="path"></param>
        protected void Load(string path)
        {
            Uri uri = new Uri(path);

            if (uri.Scheme == "http")
                LoadHTTP(path);
            else
                LoadLocal(uri.LocalPath);

            CreateVolumeTransforms();

        }

        protected void LoadHTTP(string path)
        {
            //Remove the .xml file from the path
            int iRemove = path.LastIndexOf('/');
            string VolumePath = path.Remove(iRemove) ;
            XmlTextReader reader = new XmlTextReader(path);

            while (reader.Read())
            {
                if (reader.NodeType == XmlNodeType.Whitespace)
                    continue;

                //Fetch the name if we know it
                switch(reader.LocalName)
                {
                    case "Volume": 
                        Name = reader.GetAttribute("name"); 
                        break;
                    case "stos":
                        string StosPath = VolumePath + '/' + reader.GetAttribute("path");
                        string type = reader.GetAttribute("type");
                        int pixelSpacing = System.Convert.ToInt32(reader.GetAttribute("pixelSpacing"));

                        string key = type + " " + pixelSpacing.ToString();
                        if(this.Transforms.ContainsKey(key))
                        {
                            StosGridTransform[] transforms = this.Transforms[key];
                            List<StosGridTransform> listTransforms = new List<StosGridTransform>(transforms);
                            listTransforms.Add(new StosGridTransform(StosPath, reader.ReadSubtree()));
                            this.Transforms[key] = listTransforms.ToArray();
                        }
                        else
                        {
                            this.Transforms.Add(key, new StosGridTransform[] {new StosGridTransform(StosPath, reader.ReadSubtree() ) } );
                        }
                        break;
                    case "Section":
                        string SectionPath = VolumePath + '/' + reader.GetAttribute("path");
                        Section s = new Section(this, SectionPath, reader.ReadSubtree());
                        this.sections.Add(s.Number, s);
                        break; 
                }
            }
        }

        protected void LoadLocal(string path)
        {
            string[] dirs = System.IO.Directory.GetDirectories(path);
            
            //Load all of the sections we can find
            foreach( string dir in dirs)
            {
                string[] files = System.IO.Directory.GetFiles(dir, "*.mosaic");
                if (files.Length == 0)
                {
                    Debug.WriteLine("No .Mosaic Found: " + dir);
                    continue;
                }

                Section s = new Section(this, dir);
                this.sections.Add(s.Number,s);
            }

            if(this.sections.Count > 0)
                this._SectionIndex = 0;

            //grid transforms have the form ####-####_grid_## where the last two digits indicate the pixel spacing,
            //the first two numbers seperated by the hyphen indicate the mapped and control section #'s.
            //Each pixel spacing group needs to be put in a seperate tranform collection. 
            string[] stosfiles = System.IO.Directory.GetFiles(path, "*.stos");

            SortedList<string, List<string>> SortFilesByDownsample = new SortedList<string, List<string>>();

            for (int i = 0; i < stosfiles.Length; i++)
            {
                string file = stosfiles[i];
                string basefilename = System.IO.Path.GetFileNameWithoutExtension(file);
                string[] parts = basefilename.Split('_');
                string TransformType = ""; 

                //Third part should have downsample level, skip it if we can't figure it out
                int downsample = 0;
                try
                {
                    TransformType = parts[1]; 
                    downsample = System.Convert.ToInt32(parts[2]);
                }
                catch (Exception e)
                {
                    Debug.WriteLine("Skipping improperly named stos file: " + file.ToString()); 
                    continue; 
                }

                string key = TransformType + " " + downsample.ToString();

                if (SortFilesByDownsample.ContainsKey(key))
                    SortFilesByDownsample[key].Add(file);
                else
                {
                    List<string> files = new List<string>();
                    files.Add(file);
                    SortFilesByDownsample[key] = files;
                }
            }

            foreach (string key in SortFilesByDownsample.Keys)
            {
                List<string> files = SortFilesByDownsample[key];
                StosGridTransform[] volTransforms = new StosGridTransform[files.Count];

                for(int i = 0; i < files.Count; i++)
                {
                    string file = files[i];
                    volTransforms[i] = new StosGridTransform(file);
                }

                this.Transforms.Add(key, volTransforms); 
            }
        }

        /// <summary>
        /// Adds a transform to each section mapping it into each of the volume spaces we found
        /// </summary>
        public void CreateVolumeTransforms()
        {
            foreach (string TransformKey in Transforms.Keys)
            {
                StosGridTransform[] volTransforms = Transforms[TransformKey];

                //Make a list sorted by which section the transform maps from.  Then we'll add transfroms so every transform maps from the mapped section to section #1
                SortedList<int, StosGridTransform> TList = new SortedList<int, StosGridTransform>();

                //Make transforms map to the control section
                for (int i = 0; i < volTransforms.Length; i++)
                {
                    TList.Add(volTransforms[i].MappedSection, volTransforms[i]); 
                }

                foreach (int iSection in TList.Keys)
                {
                    StosGridTransform trans = TList[iSection];
                    Section S = sections[trans.MappedSection];
                    S.AddVolumeTransform(trans, "slice-to-slice " + TransformKey.ToString());

                    if (trans.ControlSection == 1)
                        continue;

                    if(false == TList.ContainsKey(trans.ControlSection))
                        continue;

                    //Find the section that can map our transform
                    StosGridTransform addTrans = TList[trans.ControlSection];

                    //Sections should register to section 1, but if a volume hasn't finished registration or was done in parts we may register to a section other than 1
                    //Debug.Assert(addTrans.ControlSection == 1);
                    Debug.WriteLine("Adding transforms: " + trans.ToString() + " to " + addTrans.ToString()); 
                    trans.Add(addTrans);

                    //Let windows process some events
                    System.Windows.Forms.Application.DoEvents();
                }

                int controlSection = 1; 

                //Transform the sections if we can
                for (int i = 0; i < volTransforms.Length; i++)
                {
                    StosGridTransform trans = volTransforms[i];

                    if (i == 0)
                        controlSection = trans.ControlSection; 

                    //Subtract one because sections start at one, not zero
                    Debug.WriteLine(volTransforms[i].ToString());

                    Section S = sections[trans.MappedSection];
                    S.AddVolumeTransform(trans, "Volume " + TransformKey.ToString());

                    //Let windows process some events
                    System.Windows.Forms.Application.DoEvents();
                }
                
                //Add a volume transform to the control section as well
                sections[controlSection].AddVolumeTransform(null, "slice-to-slice " + TransformKey);
                sections[controlSection].AddVolumeTransform(null, "Volume " + TransformKey);
            }
        }

        void INGVVDraw.Draw(GraphicsDevice graphicsDevice, Quad bounds, double DownSample, BasicEffect basicEffect)
        {
            sections.Values[_SectionIndex].Draw(graphicsDevice, bounds, DownSample, basicEffect);
          
            /*
            if (this.Transforms != null)
            {
                //Make a list sorted by which section the transform maps from.  Then we'll add transfroms so every transform maps from the mapped section to section #1
                SortedList<int, StosGridTransform> TList = new SortedList<int, StosGridTransform>();

                StosGridTransform[] volTransforms = Transforms.Values[0]; 

                //Make transforms map to the control section
                for (int i = 0; i < volTransforms.Length; i++)
                {
                    TList.Add(volTransforms[i].MappedSection, volTransforms[i]);
                }

                if(SectionIndex < TList.Count)
                    TList.Values[SectionIndex].Draw(graphicsDevice, basicEffect); 
            }
            */

            
            /*
            VertexPositionColor v1 = new VertexPositionColor(new Vector3(bounds.X, bounds.Y, 0), Color.Yellow);
            VertexPositionColor v2 = new VertexPositionColor(new Vector3(bounds.X + bounds.Width, bounds.Y, 0), Color.Yellow);
            VertexPositionColor v3 = new VertexPositionColor(new Vector3(bounds.X, bounds.Y+bounds.Height, 0), Color.Yellow);
            VertexPositionColor v4 = new VertexPositionColor(new Vector3(bounds.X + bounds.Width, bounds.Y + bounds.Height, 0), Color.Yellow);

            VertexPositionColor[] verts = { v1, v2, v4, v3 }; 
            int[] ind = { 0, 1, 2, 3,0};
            */

            /*
             * 
            basicEffect.Begin();

            foreach (EffectPass pass in basicEffect.CurrentTechnique.Passes)
            {
                pass.Begin();

                bounds.Draw(graphicsDevice); 

                
                pass.End();
            }

            basicEffect.End(); 
            */

            /*
            MapP = Transforms[0].Transform(ControlP); 

            VertexPositionColor v1 = new VertexPositionColor(new Vector3(ControlP, 0), Color.Red);
            VertexPositionColor v2 = new VertexPositionColor(new Vector3(MapP, 0), Color.Blue);
            VertexPositionColor[] verts = {v1,v2};

            graphicsDevice.RenderState.PointSize = 10f; 

            int[] ind = {0,1};

            basicEffect.Begin();

            foreach (EffectPass pass in basicEffect.CurrentTechnique.Passes)
            {
                pass.Begin();

                graphicsDevice.DrawUserIndexedPrimitives<VertexPositionColor>(PrimitiveType.PointList, verts, 0, verts.Length, ind, 0, ind.Length);
                
                pass.End(); 
            }

            basicEffect.End(); 
             */
        }
    }
}
