using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Util
{
    /// <summary>
    /// This is the parent class of all transforms.  All transforms, even continuos, are implemented by placing a grid 
    /// of control points on the mapped and control section.  
    /// </summary>
    public class GridTransform : ICloneable
    {
        /// <summary>
        /// Section the transform maps to
        /// </summary>
        public int ControlSection = 0;
        protected Rectangle ControlBounds = new Rectangle();

        /// <summary>
        /// Section the transform maps from
        /// </summary>
        public int MappedSection = 1;
        protected Rectangle MappedBounds = new Rectangle();

        /// <summary>
        /// Control points that define the transform
        /// </summary>
        protected List<MappingTriangle> mappings = new List<MappingTriangle>();

        public GridTransform() : base() 
        {

        }

        public override string ToString()
        {
           return "GridTransform " + MappedSection.ToString() + " to " + ControlSection.ToString();
        }

        public GridTransform Copy()
        {
            return ((ICloneable)this).Clone() as GridTransform; 
        }

        object ICloneable.Clone()
        {
            GridTransform newObj = new GridTransform();
            newObj = this.MemberwiseClone() as GridTransform;

            newObj.mappings = new List<MappingTriangle>(); 

            foreach (MappingTriangle mapping in mappings)
            {
                newObj.mappings.Add(mapping.Copy()); 
            }

            return newObj; 
        }

        /// <summary>
        /// Return the mapping triangle which can map the point
        /// </summary>
        /// <param name="Point"></param>
        /// <returns></returns>
        public bool CanTransform(Vector2 Point)
        {
            if (GetTransform(Point) == null)
                return false;

            return true;
        }

        /// <summary>
        /// Return the mapping triangle which can map the point
        /// </summary>
        /// <param name="Point"></param>
        /// <returns></returns>
        protected MappingTriangle GetTransform(Vector2 Point)
        {
            //Having a smaller epsilon caused false positives.  
            //We just want to know if we are close enough to check with the more time consuming math
            double epsilon = 5;

            if (Point.X < CachedMappedBounds.Left - epsilon)
                return null;
            if (Point.X > CachedMappedBounds.Right + epsilon)
                return null;
            if (Point.Y < CachedMappedBounds.Top - epsilon)
                return null;
            if (Point.Y > CachedMappedBounds.Bottom + epsilon)
                return null; 
            

            foreach (MappingTriangle t in mappings)
            {
                if (t.IntersectsMapped(Point))
                    return t;
            }

            return null;
        }

        /// <summary>
        /// Transform point from mapped space to control space
        /// </summary>
        /// <param name="Point"></param>
        /// <returns></returns>
        public Vector2 Transform(Vector2 Point)
        {
            //TEMP Code
            /*
            if (Point.X > MappedBounds.Width)
                Point.X = MappedBounds.Width;
            if (Point.Y > MappedBounds.Height)
                Point.Y = MappedBounds.Height;
             */

            MappingTriangle t = GetTransform(Point);
            if (t == null)
            {
                throw new ArgumentException("Point could not be mapped"); 
            }

            return t.Transform(Point);
        }

       

        /// <summary>
        /// Translates all verticies in the tile according to the vector
        /// </summary>
        /// <param name="vector"></param>
        public void Translate(Vector2 vector)
        {
            for (int i = 0; i < mappings.Count; i++)
            {
                mappings[i].Control.p1 += vector;
                mappings[i].Control.p2 += vector;
                mappings[i].Control.p3 += vector; 
            }

            _CachedControlBounds = new Rectangle(int.MinValue, int.MinValue, 0, 0);
//            Bounds.X += (int)Math.Round(vector.X);
//            Bounds.Y += (int)Math.Round(vector.Y);
        }

        protected Rectangle _CachedMappedBounds = new Rectangle(int.MinValue, int.MinValue, 0, 0);
        public Rectangle CachedMappedBounds
        {
            get
            {
                if (_CachedMappedBounds.X == int.MinValue)
                    return CalculateMappedBounds();
                return _CachedMappedBounds;
            }

        }

        public Rectangle CalculateMappedBounds()
        {
            double minX = double.MaxValue;
            double minY = double.MaxValue;
            double maxX = double.MinValue;
            double maxY = double.MinValue;

            //Looking at gridIndicies isn't efficient, but it prevents adding removed verticies to 
            //boundary
            for (int i = 0; i < mappings.Count; i++)
            {
                minX = Math.Min(minX, mappings[i].Mapped.p1.X);
                minX = Math.Min(minX, mappings[i].Mapped.p2.X);
                minX = Math.Min(minX, mappings[i].Mapped.p3.X);

                maxX = Math.Max(maxX, mappings[i].Mapped.p1.X);
                maxX = Math.Max(maxX, mappings[i].Mapped.p2.X);
                maxX = Math.Max(maxX, mappings[i].Mapped.p3.X);

                minY = Math.Min(minY, mappings[i].Mapped.p1.Y);
                minY = Math.Min(minY, mappings[i].Mapped.p2.Y);
                minY = Math.Min(minY, mappings[i].Mapped.p3.Y);

                maxY = Math.Max(maxY, mappings[i].Mapped.p1.Y);
                maxY = Math.Max(maxY, mappings[i].Mapped.p2.Y);
                maxY = Math.Max(maxY, mappings[i].Mapped.p3.Y);
            }

            _CachedMappedBounds = new Rectangle((int)Math.Floor(minX), (int)Math.Floor(minY), (int)Math.Ceiling(maxX - minX), (int)Math.Ceiling(maxY - minY));
            return _CachedMappedBounds;
        }


        protected Rectangle _CachedControlBounds = new Rectangle(int.MinValue, int.MinValue, 0,0);
        public Rectangle CachedControlBounds
        {
            get
            {
                if (_CachedControlBounds.X == int.MinValue)
                    return CalculateControlBounds();
                return _CachedControlBounds;
            }
        }

        public Rectangle CalculateControlBounds()
        {
            double minX = double.MaxValue;
            double minY = double.MaxValue;
            double maxX = double.MinValue;
            double maxY = double.MinValue;

            //Looking at gridIndicies isn't efficient, but it prevents adding removed verticies to 
            //boundary
            for (int i = 0; i < mappings.Count; i++)
            {
                minX = Math.Min(minX, mappings[i].Control.p1.X);
                minX = Math.Min(minX, mappings[i].Control.p2.X);
                minX = Math.Min(minX, mappings[i].Control.p3.X);

                maxX = Math.Max(maxX, mappings[i].Control.p1.X);
                maxX = Math.Max(maxX, mappings[i].Control.p2.X);
                maxX = Math.Max(maxX, mappings[i].Control.p3.X);

                minY = Math.Min(minY, mappings[i].Control.p1.Y);
                minY = Math.Min(minY, mappings[i].Control.p2.Y);
                minY = Math.Min(minY, mappings[i].Control.p3.Y);

                maxY = Math.Max(maxY, mappings[i].Control.p1.Y);
                maxY = Math.Max(maxY, mappings[i].Control.p2.Y);
                maxY = Math.Max(maxY, mappings[i].Control.p3.Y);
            }

            _CachedControlBounds = new Rectangle((int)Math.Floor(minX), (int)Math.Floor(minY), (int)Math.Ceiling(maxX - minX), (int)Math.Ceiling(maxY - minY));
            return _CachedControlBounds;
        }

        /// <summary>
        /// Returns the coordinate on the section to be mapped given a grid coordinate from reading the transform
        /// </summary>
        /// <param name="x"></param>
        /// <param name="y"></param>
        /// <returns></returns>
        protected static Vector2 CoordinateFromGridPos(int x, int y, float gridWidth, float gridHeight, float MappedWidth, float MappedHeight)
        {
            return new Vector2(((x) / (gridWidth - 1)) * MappedWidth, (y / (gridHeight - 1)) * MappedHeight);
        }

        /// <summary>
        /// Returns the coordinate on the section to be mapped given a grid coordinate from reading the transform
        /// </summary>
        /// <param name="x"></param>
        /// <param name="y"></param>
        /// <returns></returns>
        protected Vector2 CoordinateFromGridPos(int x, int y, float gridWidth, float gridHeight)
        {
            return new Vector2(((x) / (gridWidth - 1)) * (float)MappedBounds.Width, (y / (gridHeight - 1)) * (float)MappedBounds.Height);
        }


        /// <summary>
        /// Takes two transforms and transforms the control grid of this section into the control grid space of the passed transfrom. Requires control section
        /// of this transform to match mapped section of adding transform
        /// 
        /// When we add a transform (tAdd) to another transform (this) we follow this algorithm:
        //We pass the control points of this to tAdd and find out if they can be mapped into the space of tAdd.   The triangle of this is ABC. 
        //If all three points of a triangle map inside the space we accept the new points as the control points and proceed.
        //If two points of a triangle map inside the space and we assume point A is outside space we find the intersection of
        //AB & AC with mapped space and create new points, D & E.  We bisect the line between BC to create point F 
        //(Asserting F is able to be mapped).  We then remove triangle ABC from the mapping list and add BDF, FDE, CFE.
        //Optimization:  Before bisecting we check if the lines intersecting AC and AB have a common origin point.  If they 
        //do we use this point as the bisection point.  This prevents erosion in the convex boundary case.
        //Problems: If there is a concave boundary then the line connecting ED may pass outside mapped space, but it should 
        //be mappable because we know all points of the triangle are in mappable space.
        //If two points of a triangle maps outside the space and we assume points BC are these points then we find the
        //intersection of AB and AC with mapped space to define points D & E.  We then redefine C=E and B=D. 
        //Optimization:  Check if lines intersecting AC and AB intersect outside ADE.  If they do define EDF and add it
        //to the triangle list.  This expands the mapped space in a concave boundary case.
        //If three points of a triangle map outside the space we discard the triangle for now.
        //Optimization: Check each line to ensure it doesnt intersect with map space.  If it does
        /// </summary>
        /// <param name="t1"></param>
        /// <param name="t2"></param>
        /// <returns></returns>
        public void Add(StosGridTransform transform)
        {
 //           Debug.WriteLine("Adding transforms: " + this.ToString() + " to " + transform.ToString()); 
 //           Debug.Assert(transform.MappedSection == this.ControlSection, "Can't skip sections when assembling volume transforms");

            //Reset boundaries since they will be changed
            _CachedControlBounds = new Rectangle(int.MinValue, int.MinValue, 0, 0);
            _CachedMappedBounds = new Rectangle(int.MinValue, int.MinValue, 0, 0);

            //Search all mapping triangles and update control points, if they fall outside the grid then discard the triangle
            int iTri = 0;
            while (iTri < mappings.Count)
            {
                MappingTriangle mapTri = mappings[iTri];
                Triangle control = mapTri.Control;

                Vector2 p1 = control.p1;
                Vector2 p2 = control.p2;
                Vector2 p3 = control.p3;

                MappingTriangle m1 = transform.GetTransform(p1);
                MappingTriangle m2 = transform.GetTransform(p2);
                MappingTriangle m3 = transform.GetTransform(p3);

                int numNull = 0;
                if (m1 == null)
                    numNull++;
                if (m2 == null)
                    numNull++;
                if (m3 == null)
                    numNull++;

                MappingTriangle[] newMaps = null;
                switch (numNull)
                {
                    case 0:
                        control.p1 = m1.Transform(p1);
                        control.p2 = m2.Transform(p2);
                        control.p3 = m3.Transform(p3);

                        mapTri.Control = control;
                        mappings[iTri] = mapTri;
                        iTri++;
                        break;

                    case 1:
                        newMaps = transform.FitTriangleOneExternalVertex(mapTri);
                        mappings.RemoveAt(iTri);
                        mappings.InsertRange(iTri,newMaps);
                        iTri += newMaps.Length; 
                        break;
                    case 2:
                        newMaps = transform.FitTriangleTwoExternalVertex(mapTri);
                        mappings.RemoveAt(iTri);
                        mappings.InsertRange(iTri, newMaps);
                        iTri += newMaps.Length;
                        break; 
                    case 3:
                        mappings.RemoveAt(iTri);
                        break; 
                }
            }

            _CachedControlBounds = new Rectangle(int.MinValue, int.MinValue, 0, 0);
            _CachedMappedBounds = new Rectangle(int.MinValue, int.MinValue, 0, 0);

            this.ControlSection = transform.ControlSection;
        }

        /// <summary>
        /// Given a mapping triangle with one vertex outside our mapping space (A) we return a list
        /// of mapping triangles which fill our mapping space and crop the unmappable space
        /// </summary>
        /// <param name="mapTri">Mapping triangle which doesn't fit</param>
        /// <returns></returns>
        protected MappingTriangle[] FitTriangleOneExternalVertex(MappingTriangle mapTri)
        {
            //Figure out which vertex is external and relabel vectors A,B,C in clockwise fashion
            //with A being outside mapped space
            Vector2 A,B,C;

            if(false == CanTransform(mapTri.Control.p1))
            {
                A = mapTri.Control.p1;
                B = mapTri.Control.p2;
                C = mapTri.Control.p3; 
            }
            else if (false == CanTransform(mapTri.Control.p2))
            {
                A = mapTri.Control.p2;
                B = mapTri.Control.p3;
                C = mapTri.Control.p1; 
            }
            else
            {
                Debug.Assert(CanTransform(mapTri.Control.p3) == false, "FitTriangleOneExternalVertex - all verts inside mapped space");
                A = mapTri.Control.p3;
                B = mapTri.Control.p1;
                C = mapTri.Control.p2; 
            }

            LineSegment AB = new LineSegment(A, B); //Leave points outside mapping as first term so I can calculate distance to closest point in mapped space
            LineSegment AC = new LineSegment(A, C);
            LineSegment BC = new LineSegment(B, C); 

            //Find the nearest point inside the grid from the external points
            LineSegment lineThroughAB;
            Vector2 D = new Vector2(); //The intersection point of BA with mapped space
            float nearestABintersect = float.MaxValue;
            LineSegment lineThroughAC;
            Vector2 E = new Vector2(); //The intersection point of CA with mapped space
            float nearestACintersect = float.MaxValue;

            LineSegment lineThroughBC;
            Vector2 F = new Vector2(); //The intersection point of CA with mapped space
            float nearestBCintersect = float.MaxValue;


            nearestABintersect = NearestLine(AB, out lineThroughAB, out D);
            nearestACintersect = NearestLine(AC, out lineThroughAC, out E);
            nearestBCintersect = NearestLine(BC, out lineThroughBC, out F);

            //TODO: Fiddle with epsilon or figure out why this error happens sometimes
            //Debug.Assert(lineThroughAB != null && lineThroughAC != null, "FitTriangleOneExternalVertex: Couldn't find intersection with mapped space");
            if (lineThroughAB == null || lineThroughAC == null)
                return new MappingTriangle[0]; 

            //OK, we know where the lines intersect with mapped space, find midpoint of BC to define F.  Then create
            //mapping triangles BDF, FDE, CFE

            //First we check to see if there is a common point between two intersection lines we can use as a node;
            if (nearestBCintersect >= float.MaxValue)
            {
                //TODO: If this fails to find a valid transform spot I want to remove the triangle from the transform.
                //hopefully this case is only hit when two nodes are in the same triangle.
                F =  BC.Bisect(); 
            }

            //Find the closest line intersection to BC

            //My approach to dealing with failed transforms due to floating point error is to remove those triangles

            Vector2 Bctrl;
            Vector2 Cctrl;
            Vector2 Dctrl;
            Vector2 Ectrl;
            Vector2 Fctrl;

            //Work backwards to find where new control points where in the mapped space
            Vector2 Bmap;
            Vector2 Cmap;
            Vector2 Dmap;
            Vector2 Emap;;
            Vector2 Fmap;

            try
            {
                //Transform all points into the new control space
                Bctrl = Transform(B);
                Cctrl = Transform(C);
                Dctrl = Transform(D);
                Ectrl = Transform(E);
                Fctrl = Transform(F);

                //Work backwards to find where new control points where in the mapped space
                Bmap = mapTri.InverseTransform(B);
                Cmap = mapTri.InverseTransform(C);
                Dmap = mapTri.InverseTransform(D);
                Emap = mapTri.InverseTransform(E);
                Fmap = mapTri.InverseTransform(F);
            }
            catch (ArgumentException e)
            {
                return new MappingTriangle[0]; 
            }

            List<MappingTriangle> newMaps = new List<MappingTriangle>();
            MappingTriangle newTri = null;
            try
            {
                newTri = new MappingTriangle(new Triangle(Cctrl, Fctrl, Ectrl, Color.Yellow), new Triangle(Cmap, Fmap, Emap, Color.Gold));
                newMaps.Add(newTri); 
            }catch(ArgumentException e)
            {
                Debug.WriteLine("Tried to create triangle which is really a line"); 
            }
            
            try
            {
                newTri = new MappingTriangle(new Triangle(Fctrl, Dctrl, Ectrl, Color.Violet), new Triangle(Fmap, Dmap, Emap, Color.Black));
                newMaps.Add(newTri); 
            }
            catch(ArgumentException e)
            {
                Debug.WriteLine("Tried to create triangle which is really a line"); 
            }

            try{
                newTri = new MappingTriangle(new Triangle(Fctrl, Bctrl, Dctrl, Color.Green), new Triangle(Fmap, Bmap, Dmap, Color.White));
                newMaps.Add(newTri); 
            }catch(ArgumentException e)
            {
                Debug.WriteLine("Tried to create triangle which is really a line"); 
            }

            return newMaps.ToArray(); 
        }

        /// <summary>
        /// Given a mapping triangle with two vertex (B&C) outside our mapping space we return a list
        /// of mapping triangles which fill our mapping space and crop the unmappable space
        /// </summary>
        /// <param name="mapTri">Mapping triangle which doesn't fit</param>
        /// <returns></returns>
        protected MappingTriangle[] FitTriangleTwoExternalVertex(MappingTriangle mapTri)
        {
            //Figure out which two vertex are external and relabel vectors A,B,C in clockwise fashion
            //with A being inside mapped space
            Vector2 A, B, C;

            if (CanTransform(mapTri.Control.p1))
            {
                A = mapTri.Control.p1;
                B = mapTri.Control.p2;
                C = mapTri.Control.p3;
            }
            else if (CanTransform(mapTri.Control.p2))
            {
                A = mapTri.Control.p2;
                B = mapTri.Control.p3;
                C = mapTri.Control.p1;
            }
            else
            {
                Debug.Assert(CanTransform(mapTri.Control.p3), "FitTriangleOneExternalVertex - all verts outside mapped space");
                A = mapTri.Control.p3;
                B = mapTri.Control.p1;
                C = mapTri.Control.p2;
            }

            LineSegment BA = new LineSegment(B, A); //Leave points outside mapping as first term so I can calculate distance to closest point in mapped space
            LineSegment CA = new LineSegment(C, A);
            LineSegment BC = new LineSegment(B, C);

            //Find the nearest point inside the grid from the external points
            LineSegment lineThroughBA;
            Vector2 D = new Vector2(); //The intersection point of BA with mapped space
            float nearestBAintersect = float.MaxValue;
            LineSegment lineThroughCA;
            Vector2 E = new Vector2(); //The intersection point of CA with mapped space
            float nearestCAintersect = float.MaxValue;

//            LineSegment lineThroughBC;
//            Vector2 F = new Vector2(); //The intersection point of CA with mapped space
//            float nearestBCintersect = float.MaxValue;


            nearestBAintersect = NearestLine(BA, out lineThroughBA, out D);
            nearestCAintersect = NearestLine(CA, out lineThroughCA, out E);
        //    nearestBCintersect = NearestLine(BC, out lineThroughBC, out F);

            //TODO: Figure out if we can prevent this from occuring
            //Debug.Assert(lineThroughBA != null && lineThroughCA != null, "FitTriangleOneExternalVertex: Couldn't find intersection with mapped space");
            if(lineThroughBA == null || lineThroughCA == null)
                return new MappingTriangle[0]; 

            //Transform all points into the new control space
            MappingTriangle[] newMaps = new MappingTriangle[1];

            try
            {
                Vector2 Actrl = Transform(A);
                Vector2 Dctrl = Transform(D);
                Vector2 Ectrl = Transform(E);

                //Work backwards to find where new control points where in the mapped space
                Vector2 Amap = mapTri.InverseTransform(A);
                Vector2 Dmap = mapTri.InverseTransform(D);
                Vector2 Emap = mapTri.InverseTransform(E);

                

                newMaps[0] = new MappingTriangle(new Triangle(Actrl, Dctrl, Ectrl, Color.Ivory), new Triangle(Amap, Dmap, Emap, Color.Gold));
            }
            catch (ArgumentException e)
            {
                Debug.WriteLine("Tried to create triangle which is really a line");
                return new MappingTriangle[0]; 
            }

            return newMaps;
        }

        protected float NearestLine(LineSegment L, out LineSegment foundLine, out Vector2 intersection)
        {
            float nearestIntersect = float.MaxValue;

            //For debugging only
            float nearestFailedIntersect = float.MaxValue;
            Vector2 nearestFailedPoint = new Vector2(); 
            LineSegment nearestFailedSegment = null;


            foundLine = null;
            intersection = new Vector2(); 

            foreach (MappingTriangle T in mappings)
            {
                Vector2 result;

                LineSegment One = new LineSegment(T.Mapped.p1, T.Mapped.p2);
                LineSegment Two = new LineSegment(T.Mapped.p2, T.Mapped.p3);
                LineSegment Three = new LineSegment(T.Mapped.p3, T.Mapped.p1);

                bool bIntersected = One.Intersects(L, out result);
                float distance = Vector2.Distance(L.A, result);
                if (distance < nearestIntersect && bIntersected)
                {
                    nearestIntersect = distance;
                    intersection = result;
                    foundLine = One;
                }
                if(distance < nearestFailedIntersect && !bIntersected)
                {
                    nearestFailedPoint = result; 
                    nearestFailedSegment = One;
                    nearestFailedIntersect = distance;
                }

                bIntersected = Two.Intersects(L, out result);
                distance = Vector2.Distance(L.A, result);
                if (distance < nearestIntersect &&  bIntersected)
                {
                    nearestIntersect = distance;
                    intersection = result;
                    foundLine = Two;
                }
                if(distance < nearestFailedIntersect && !bIntersected)
                {
                    nearestFailedPoint = result; 
                    nearestFailedSegment = Two;
                    nearestFailedIntersect = distance;
                }

                bIntersected = Three.Intersects(L, out result);
                distance = Vector2.Distance(L.A, result);
                if (distance < nearestIntersect && bIntersected)
                {
                    nearestIntersect = distance;
                    intersection = result;   
                    foundLine = Three;
                }
                if(distance < nearestFailedIntersect && !bIntersected)
                {
                    nearestFailedPoint = result; 
                    nearestFailedSegment = Three;
                    nearestFailedIntersect = distance;
                }
            }

            return nearestIntersect;

        }

        public void Draw(GraphicsDevice graphicsDevice, BasicEffect basicEffect)
        {
            /*
            BasicEffect basicEffect = new BasicEffect(graphicsDevice, null);
            basicEffect.AmbientLightColor = templateEffect.AmbientLightColor;
            basicEffect.Projection = templateEffect.Projection;
            basicEffect.World = templateEffect.World;
            basicEffect.View = templateEffect.View;
            basicEffect.Projection = templateEffect.Projection;
             */
            basicEffect.TextureEnabled = false;
            basicEffect.VertexColorEnabled = true;
            basicEffect.CommitChanges();

            VertexDeclaration originalVertDeclare = graphicsDevice.VertexDeclaration;

            basicEffect.Begin();

            foreach (EffectPass pass in basicEffect.CurrentTechnique.Passes)
            {
                pass.Begin();
                foreach (MappingTriangle tri in mappings)
                {
                    tri.Draw(graphicsDevice);
                }
                pass.End();
            }

            basicEffect.End();

            graphicsDevice.VertexDeclaration = originalVertDeclare;
        }
    }
}
