﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Util
{
    /// <summary>
    /// Two triangles that make a rectangle
    /// </summary>
    public struct Quad
    {
        Triangle T0;
        Triangle T1;

        public Quad(Vector2 pos, float Width, float Height)
            : this(pos, new Vector2(pos.X + Width, pos.Y), new Vector2(pos.X, pos.Y + Height), new Vector2(pos.X + Width, pos.Y + Height), Color.Blue)
        { }

        public Quad(Vector2 pos, float Width, float Height, Color color)
            : this(pos, new Vector2(pos.X + Width, pos.Y), new Vector2(pos.X, pos.Y + Height), new Vector2(pos.X + Width, pos.Y + Height), color)
        {}

        public Quad(Vector3 p1, Vector3 p2, Vector3 p3, Vector3 p4, Color color)
            : this(new Vector2(p1.X,p1.Y), new Vector2(p2.X,p2.Y), new Vector2(p3.X,p3.Y), new Vector2(p4.X, p4.Y), color)
        {}

        public Quad(Vector3 p1, Vector3 p2, Vector3 p3, Vector3 p4)
            : this(new Vector2(p1.X,p1.Y), new Vector2(p2.X,p2.Y), new Vector2(p3.X,p3.Y), new Vector2(p4.X, p4.Y), Color.Blue)
        {}
         
        public Quad(Vector2 p1, Vector2 p2, Vector2 p3, Vector2 p4) 
            : this(p1,p2,p3,p4, Color.Blue)
        {}

        /// <summary>
        /// 
        /// </summary>
        /// <param name="p1">BottomLeft</param>
        /// <param name="p2">BottomRight</param>
        /// <param name="p3">TopLeft</param>
        /// <param name="p4">TopRight</param>
        /// <param name="color"></param>
        public Quad(Vector2 p1, Vector2 p2, Vector2 p3, Vector2 p4, Color color)
        {
           T0 = new Triangle(p1,p2,p3,color); 
           T1 = new Triangle(p2,p4,p3,color); 
        }

        public Vector2 Center
        {
            get
            {
                return new LineSegment(T0.p2, T0.p3).Bisect();
            }
        }

        public float Width
        {
            get
            {
                return Vector2.Distance(T0.p1, T0.p2);
            }
        }

        public float Height
        {
            get
            {
                return Vector2.Distance(T0.p1, T0.p3);
            }
        }

        public Vector2 BottomLeft
        {
            get { return T0.p1; }
            set { 
                T0.p1 = value; 
            }
        }

        public Vector2 BottomRight
        {
            get { return T0.p2; }
            set { 
                T0.p2 = value; 
                T1.p1 = value; 
            }
        }

        public Vector2 TopLeft
        {
            get { return T0.p3; }
            set { 
                T0.p3 = value; 
                T1.p3 = value;
            }
        }

        public Vector2 TopRight
        {
            get { return T1.p2; }
            set { 
                T1.p2 = value; 
            }
        }

        public void Scale(float scalar)
        {
            //Have to cache center because it changes as we update points
            Vector2 center = this.Center;
            Vector2 directionA = this.TopRight - center;
            Vector2 directionB = this.TopLeft - center; 

            directionA = Vector2.Multiply(directionA, scalar);
            directionB = Vector2.Multiply(directionB, scalar);

            this.BottomLeft = Vector2.Subtract(center, directionA);
            this.TopRight = Vector2.Add(center, directionA);

            this.BottomRight = Vector2.Subtract(center, directionB);
            this.TopLeft = Vector2.Add(center, directionB);
        }

        /// <summary>
        /// Draw the triangle on a graphics device
        /// </summary>
        /// <param name="graphicsDevice"></param>
        public void Draw(GraphicsDevice graphicsDevice)
        {
            T0.Draw(graphicsDevice);
            T1.Draw(graphicsDevice); 
        }

        public bool Contains(Rectangle R)
        {
            return Contains(new Quad(new Vector2(R.X, R.Y), R.Width, R.Height)); 
        }

        public bool Contains(Quad R)
        {
            Vector2 v1 = R.BottomLeft;
            Vector2 v2 = R.BottomRight;
            Vector2 v3 = R.TopRight;
            Vector2 v4 = R.TopLeft;

            //If any verticies are in the quad we return true 
            if (T0.Intersects(v1) || T0.Intersects(v2) || T0.Intersects(v3) || T0.Intersects(v4) ||
                T1.Intersects(v1) || T1.Intersects(v2) || T1.Intersects(v3) || T1.Intersects(v4))
                return true;

            if (R.T0.Intersects(TopLeft) || R.T0.Intersects(TopRight) || R.T0.Intersects(BottomLeft) || R.T0.Intersects(BottomRight) ||
                R.T1.Intersects(TopLeft) || R.T1.Intersects(TopRight) || R.T1.Intersects(BottomLeft) || R.T1.Intersects(BottomRight))
                return true; 

            LineSegment RL1 = new LineSegment(v1, v2);
            LineSegment RL2 = new LineSegment(v2, v3);
            LineSegment RL3 = new LineSegment(v3, v4);
            LineSegment RL4 = new LineSegment(v4, v1);

            LineSegment L1 = new LineSegment(this.BottomLeft, this.BottomRight);
            LineSegment L2 = new LineSegment(this.BottomRight, this.TopRight);
            LineSegment L3 = new LineSegment(this.TopRight, this.TopLeft);
            LineSegment L4 = new LineSegment(this.TopLeft, this.BottomLeft);

            LineSegment[] RA = new LineSegment[4] { RL1, RL2, RL3, RL4 };
            LineSegment[] A = new LineSegment[4] { L1, L2, L3, L4 };

            Vector2 outparam; 
            foreach (LineSegment RL in RA)
            {
                foreach (LineSegment L in A)
                {
                    if (RL.Intersects(L, out outparam))
                        return true; 
                }
            }

            return false; 
        }

        public double Area
        {
            get
            {
                return T0.Area + T1.Area; 
            }
        }

    }
}
