#!/bin/bash

# include some useful functions:
. $HOME/bin/string_utils.sh

#----------------------------------------------------------------
# usage
# 
usage()
{
  printf "\nUSAGE:\n%s name width height x_orig y_orig x_step y_step bytes_per_scanline\n" $0
  printf "\nEXAMPLE:\n%s point 8 8 4.0 5.0 0.0 0.0 1\n\n" $0
  printf "This example would produce two output files - the_point_symbols.hxx "
  printf "and the_point_symbols.cxx,\nwhich would implement a C++ class by the"
  printf "name the_point_symbols_t, which \ninherits from the_symbols_t "
  printf "base class.\n\nclass the_point_symbols_t would implement a bitmap "
  printf "symbol list to be used with OpenGL.\nThe symbol lists will be called "
  printf "\"point\", with symbol dimensions 8x8, symbol \norigin "
  printf "located at (4.0, 5.0). During text display,the raster"
  printf "position will \nadvance by 0.0 point horizontally, and 0.0"
  printf "points vertically. The number of \nbytes necessary to represent"
  printf "a single scanline of a character of this symbols is 1.\n\n"
}

#----------------------------------------------------------------
# bytes_per_scanline
# 
bytes_per_scanline()
{
  STR=$1
  POWER=$((${#STR}))
  WHOLE=$(($POWER/8));
  PARTIAL=$(($POWER%8));
  if [ $PARTIAL != 0 ]; then
    WHOLE=$(($WHOLE+1))
  fi
  echo $WHOLE
}

#----------------------------------------------------------------
# str_to_int
# 
str_to_int()
{
  STR=$1
  ZERO=$2
  RESULT=0
  POWER=$(($((`bytes_per_scanline $STR`*8))-1))
  i=0;
  while true; do
    CHAR=${STR:$i:1}
    if [ -z "$CHAR" ]; then
      break
    elif [ "$CHAR" != "$ZERO" ]; then
      RESULT=$(($RESULT+2**$(($POWER-$i))))
    fi
    i=$(($i+1))
  done
  echo $RESULT
}

#----------------------------------------------------------------
# str_accum
# 
# args: number, bytes_per_scanline, string accumulator:
# 
str_accum()
{
  NUMBER=$1
  BYTES_PER_SCANLINE=$2
  ACCUM=$3
  BYTES=""
  i=0;
  while true; do
    BASE=$(($BYTES_PER_SCANLINE-$i-1))
    WHOLE=$(($NUMBER/$((2**$(($BASE*8))))))
    PART=$(($NUMBER%$((2**$(($BASE*8))))))
    NUMBER=$PART
    if [ -z "$BYTES" ]; then
      BYTES=`printf "0x%.2X" $WHOLE`
    else
      BYTES="`printf \"0x%.2X\" $WHOLE`, $BYTES"
    fi
    
    i=$(($i+1))
    if [ $i = $BYTES_PER_SCANLINE ]; then
      break
    fi
  done
  
  if [ -z "$ACCUM" ]; then
    printf "%s" $BYTES
  else
    printf "%s, %s" $BYTES $ACCUM
  fi
}

#----------------------------------------------------------------
# gen_hxx
# 
gen_hxx()
{
  printf "// NOTE: \n"
  printf "//       DO NOT EDIT THIS FILE, IT IS AUTOGENERATED.\n"
  printf "//       ALL CHANGES WILL BE LOST UPON REGENERATION.\n\n"
  printf "#ifndef THE_%s_SYMBOLS_HXX_\n" `str_to_upper ${NAME}`
  printf "#define THE_%s_SYMBOLS_HXX_\n\n" `str_to_upper ${NAME}`
  printf "// local includes:\n"
  printf "#include \"the_symbols.hxx\"\n\n\n"
  
  NUM=0
  printf "//----------------------------------------------------------------\n"
  printf "// the_%s_symbol_id_t\n" $NAME
  printf "// \n"
  printf "typedef enum\n{"
  for i in `/bin/ls *.xpm | sort`; do
    if [ $NUM != 0 ]; then 
    printf ","
    fi
    
    ENUM=`echo $i | cut -d'.' -f1`
    printf "\n  THE_%.32s_SYMBOL_E" `str_to_upper ${ENUM}`
    NUM=$(($NUM+1))
  done
  printf "\n} the_%s_symbol_id_t;\n\n\n" $NAME
  
  printf "//----------------------------------------------------------------\n"
  printf "// the_%s_symbols_t\n" $NAME
  printf "// \n"
  printf "class the_%s_symbols_t : public the_symbols_t\n{\npublic:\n" $NAME
  printf "  the_%s_symbols_t(): the_symbols_t(\"%s\") {}\n" $NAME $NAME
  printf "  virtual ~the_%s_symbols_t() {}\n\n" $NAME
  printf "  // virtual:\n"
  printf "  unsigned int width() const  { return %s; }\n" $WIDTH
  printf "  unsigned int height() const { return %s; }\n\n" $HEIGHT
  printf "  // virtual:\n"
  printf "  real_t x_origin() const { return %s; }\n" $X_ORIG
  printf "  real_t y_origin() const { return %s; }\n\n" $Y_ORIG
  printf "  // virtual:\n"
  printf "  real_t x_step() const { return %s; }\n" $X_STEP
  printf "  real_t y_step() const { return %s; }\n\n" $Y_STEP
  printf "  // virtual:\n"
  printf "  byte_t * bitmap(const unsigned int & id) const "
  printf "{ return symbols_[id]; }\n\n"
  printf "  // virtual: the number of display lists to be compiled:\n"
  printf "  unsigned int size() const { return %s; }\n\n" $NFILES
  printf "private:\n"
  printf "  static byte_t symbols_[%s][%s];\n" $NFILES $(($HEIGHT*$NBYTES))
  printf "};\n\n"
  printf "//----------------------------------------------------------------\n"
  printf "// THE_%s_SYMBOLS\n" `str_to_upper ${NAME}`
  printf "// \n"
  printf "// a single global instance of the symbols font:\n"
  printf "// \n"
  printf "extern the_%s_symbols_t THE_%s_SYMBOLS;\n" $NAME `str_to_upper ${NAME}`
  printf "\n\n#endif // THE_%s_SYMBOLS_HXX_\n" `str_to_upper ${NAME}`
}

#----------------------------------------------------------------
# gen_cxx
# 
gen_cxx()
{
  printf "// NOTE: \n"
  printf "//       DO NOT EDIT THIS FILE, IT IS AUTOGENERATED.\n"
  printf "//       ALL CHANGES WILL BE LOST UPON REGENERATION.\n\n"
  printf "// local includes:\n"
  printf "#include \"the_%s_symbols.hxx\"\n\n\n" $NAME
  printf "//----------------------------------------------------------------\n"
  printf "// the_%s_symbols_t::symbols_\n" $NAME
  printf "// \n"
  printf "byte_t\n"
  printf "the_%s_symbols_t::symbols_[%s][%s] =\n{" $NAME $NFILES $(($HEIGHT*$NBYTES))

  PREV_IFS=${IFS}
  COLUMNS=0;
  NUM=0
  ZERO='.'
  ONE='\#' 
  for i in `/bin/ls *.xpm | sort`; do
    IFS=''
    if [ $NUM != 0 ]; then
      printf ",\n  {"
    else
      printf "\n  {"
    fi
    exec 0<$i
    LINENUM=0
    ACCUM=""
    SCANLINE=0
    while read -r j; do
      if [ 1 = $(($LINENUM>2)) ]; then
        LINE=`echo $j | cut -d'"' -f2`
        HAS_NONE=`echo $LINE | grep ' c None'`
        HAS_HASH=`echo $LINE | grep ' c #'`
        HAS_END=`echo $j | grep '"};'`
        if [ -n "$HAS_NONE" ]; then
          ZERO=`echo $LINE | cut -d' ' -f1`
        elif [ -n "$HAS_HASH" ]; then
          ONE=`echo $LINE | cut -d' ' -f1`
        else
          BYTES=`bytes_per_scanline $LINE`
          SCANLINE=$(($SCANLINE+1))
          if [ 1 = $(($BYTES>$NBYTES)) ]; then
            printf "\nERROR: byte count mismatch, expected %s, got %s\n" $NBYTES $BYTES
            printf "file: %s, line: %s, %s\n\naborting...\n\n" $i $LINENUM $j
            exit 1;
          fi
          
          NUMBER=`str_to_int $LINE $ZERO`
          ACCUM=`str_accum $NUMBER $NBYTES $ACCUM`
          if [ -n "$HAS_END" ]; then
            if [ $SCANLINE != $HEIGHT ]; then
              printf "\nERROR: height mismatch, expected %s, got %s\n" $HEIGHT $SCANLINE
              printf "file: %s, line: %s\n\naborting...\n\n" $i $LINENUM
              exit 1;
            fi
            printf " %s" $ACCUM
          fi
        fi
      fi
      LINENUM=$(($LINENUM+1))
    done
    NUM=$(($NUM+1))
    printf " }"
  done
  printf "\n};\n\n"
  printf "//----------------------------------------------------------------\n"
  printf "// THE_%s_SYMBOLS\n" `str_to_upper ${NAME}`
  printf "// \n"
  printf "// a single global instance of the symbols font:\n"
  printf "// \n"
  printf "the_%s_symbols_t THE_%s_SYMBOLS;\n" $NAME `str_to_upper ${NAME}`
}

#----------------------------------------------------------------
# main
# 
main()
{
  FILES=`/bin/ls *.xpm | sort`
  if [ -z "$FILES" ]; then
    printf "error: no *.xpm files to process, aborting\n"
    exit 1;
  fi
  
  NFILES=`/bin/ls *.xpm | sort | wc | cut -c0-8`
  NAME=$1
  WIDTH=$2
  HEIGHT=$3
  X_ORIG=$4
  Y_ORIG=$5
  X_STEP=$6
  Y_STEP=$7
  NBYTES=$8
  if [[ -z "$NBYTES" || -z "$Y_STEP" || \
      -z "$X_STEP" || -z "$Y_ORIG" || \
      -z "$X_ORIG" || -z "$HEIGHT" || \
      -z "$WIDTH"  || -z "$NAME" ]]; 
  then
    usage $*
    printf "ARGMENTS GIVEN:\n  1: %s\n  2: %s\n  3: %s\n  4: %s\n  5: %s\n  6: %s\n  7: %s\n  8: %s\n\n" "$NAME" "$WIDTH" "$HEIGHT" "$X_ORIG" "$Y_ORIG" "$X_STEP" "$Y_STEP" "$NBYTES"
    exit 1
  fi
  
  printf "generating the_%s_symbols.hxx ..... " $NAME
  gen_hxx > the_${NAME}_symbols.hxx
  printf "done\n"
  printf "generating the_%s_symbols.cxx ..... " $NAME
  gen_cxx > the_${NAME}_symbols.cxx
  printf "done\n"
}

# declare some global variables:
FILES=""
NFILES=0
NAME=""
WIDTH=8
HEIGHT=8
X_ORIG=0.0
Y_ORIG=2.0
X_STEP=9.0
Y_STEP=0.0
NBYTES=1
export NAME WIDTH HEIGHT X_ORIG Y_ORIG X_STEP Y_STEP NBYTES

# start the process:
main $*
