#include "SynapsesViewer.h"
#include <iostream>
#include <QtGui>

#include <math.h>

static const int MouseCount = 7;

SynapsesViewer::SynapsesViewer(int argc, char **argv)
{
  initializeApplication();
  if (argc > 1)
    openSVFFile (argv[1]);
}

SynapsesViewer::SynapsesViewer()
{
  initializeApplication();
}

void SynapsesViewer::initializeApplication()
{
  initializeWidgets ();
  initializeSynapseGroupManager();
  initializeImageGroup();
  this->updateViewer (pictureFilename, synapsesFilename);
}

SynapsesViewer::~SynapsesViewer()
{
  // delete varname;
}

void SynapsesViewer::initializeSynapseGroupManager()
{
  synapseGroupManager = new SynapseGroupManager();
  synapseGroupManager->setCurrentSynapseGroup(-1);
}

void SynapsesViewer::initializeImageGroup()
{
  imageGroup = new ImageGroup();
}

void SynapsesViewer::initializeWidgets()
{
  createActions ();
  createMenus ();
  synapseMode = MOVE_SYNAPSE_MODE;
  imageScale = 100;

  setWindowTitle(tr("Synapses Viewer"));
  resize(500, 400); 

  scene = new QGraphicsScene();
  scene->setItemIndexMethod(QGraphicsScene::NoIndex);
  view = new QGraphicsView(scene);
  view->setRenderHint(QPainter::Antialiasing);
  pictureItem = new QGraphicsPixmapItem();
  scene->addItem(pictureItem);

  QGridLayout *layout = new QGridLayout();
  layout->addWidget(view,1,1);
  initializeViewController();
  layout->addWidget (viewerController ,1,1, Qt::AlignTop);
  layout->setAlignment(Qt::AlignCenter);
  layout->setMargin(5);
  QWidget *widget = new QWidget;
  widget->setLayout(layout);
  setCentralWidget(widget);
  this->show();
}

void SynapsesViewer::initializeViewController()
{
  viewerController = new ViewerController();
  connect(viewerController->getZoomInButton(), SIGNAL(clicked()), this, SLOT(zoomInPicture()));
  connect(viewerController->getZoomOutButton(), SIGNAL(clicked()), this, SLOT(zoomOutPicture()));
  connect(viewerController->getScaleSlider(), SIGNAL(valueChanged(int)), this, SLOT(setScale(int)));
  connect(viewerController->getMoveRightButton(), SIGNAL(clicked()), this, SLOT(moveSceneRight()));
  connect(viewerController->getMoveLeftButton(), SIGNAL(clicked()), this, SLOT(moveSceneLeft()));
  connect(viewerController->getMoveUpButton(), SIGNAL(clicked()), this, SLOT(moveSceneUp()));
  connect(viewerController->getMoveDownButton(), SIGNAL(clicked()), this, SLOT(moveSceneDown()));
}

void SynapsesViewer::moveSceneRight()
{
  printf ("Move Scene Right\n");
  QMatrix oldMatrix = view->matrix();
  printf ("dx = %f , dy = %f\n", oldMatrix.dx(), oldMatrix.dy());
}

void SynapsesViewer::moveSceneLeft()
{
	printf ("Move Scene Left\n");
}

void SynapsesViewer::moveSceneUp()
{
	printf ("Move Scene Up\n");
}

void SynapsesViewer::moveSceneDown()
{
	printf ("Move Scene Down\n");
}

void SynapsesViewer::updateViewer(QString pictureFileName, QString synapsisFileName)
{
  if (!pictureFileName.isEmpty())
    this->setPicture (pictureFileName);
  if (!synapsisFileName.isEmpty())
    this->setSynapses (synapsisFileName);
}

void SynapsesViewer::setSynapses (SynapseGroup* newSynapseGroup)
{
  synapseGroupManager->addNewSynapseGroup (newSynapseGroup);
  synapseGroupVisibilityAct->menu()->addAction (newSynapseGroup->getGroupMenuItem());
  printf ("Synapse No. of Groups = %d\n", synapseGroupManager->getNumberOfGroups());
  printf ("Synapse new Group size = %d\n", newSynapseGroup->getNumberOfSynapses());
  int SynapseMarkerCount = newSynapseGroup->getNumberOfSynapses();
  for (int i = 0; i < SynapseMarkerCount; ++i)
  {
    SynapseMarker *synapseMarker = new SynapseMarker();
	synapseMarker = newSynapseGroup->getSynapseAt(i);
    synapseMarker->setPos(synapseMarker->getPosition().x(), synapseMarker->getPosition().y());
    synapseMarker->rotate(-synapseMarker->getDirection());
    scene->addItem(synapseMarker);
  }
  view->show();
}

void SynapsesViewer::setSynapses (QString synapsesFileName)
{
  SynapseGroup *synapseGroup = new SynapseGroup();
  
  QFile xmlFile (synapsesFileName);
  domSynapses = new QDomDocument ();

  domSynapses->setContent(&xmlFile);
  QDomElement root = domSynapses->documentElement();

  if (root.tagName() != "rois")
     QMessageBox::information(window(), tr("Synapses Viewer"), tr("This file is not a Synapses file."));
  else
    synapseGroup->addSynapsesFromFile (synapsesFileName);

  this->setSynapses (synapseGroup);
}


void SynapsesViewer::setPicture (QString fileName)
{
  imageGroup->addImageFromFile(fileName);
  imageGroupVisibilityAct->menu()->addAction (imageGroup->getCurrentAction());
  imageVisibilityGroup->addAction(imageGroup->getCurrentAction());
  imageGroup->getCurrentAction()->setChecked(TRUE);
  scene->addItem(imageGroup->getCurrentImage());
  scene->setSceneRect(0, 0, imageGroup->getCurrentImage()->pixmap().width(), imageGroup->getCurrentImage()->pixmap().height());
  view->setDragMode(QGraphicsView::ScrollHandDrag);
  view->setWindowTitle(QT_TRANSLATE_NOOP(QGraphicsView, "Synapses Viewer - "+fileName));
  view->show();
}

void SynapsesViewer::openPicture()
{
  //QString fileName = QFileDialog::getOpenFileName(this, tr("Open File"), QDir::currentPath());
  //QString fileName = QFileDialog::getOpenFileName(this, tr("Open File"), "C:\\WorkingFiles\\Synapses\\data", "Tiff Images (*.tif)");
  QString fileName = QFileDialog::getOpenFileName(this, tr("Open Image"), "/usr/sci/crcnsdata/CRCNS/Synapses/data/roiExport3/", "All Image Formats (*.jpg *.png *.tif);; TIFF Images (*.tif);; JPEG Images (*.jpg)");
  if (!fileName.isEmpty()) 
  {
    QImage image(fileName);
    if (image.isNull())
    {
      QMessageBox::information(this, tr("Synapses Viewer"), tr("Cannot load %1.").arg(fileName));
      return;
    }
    this->updateViewer (fileName, NULL);
  }
}

void SynapsesViewer::openSynapses()
{
  //QString fileName = QFileDialog::getOpenFileName(this, tr("Open File"), QDir::currentPath());
  //QString fileName = QFileDialog::getOpenFileName(this, tr("Open File"), "C:\\WorkingFiles\\Synapses\\data", "Synapses XMLs (*.xml)");
  printf ("Number of images = %d\n", imageGroup->getNumberOfImages());
  if (imageGroup->getNumberOfImages() == 0)
    QMessageBox::information(window(), tr("Synapses Viewer"), tr("You can't open a Markup unless you open an image."));
  else
  {
    QString fileName = QFileDialog::getOpenFileName(this, tr("Open Synapse XML"), "/usr/sci/crcnsdata/CRCNS/Synapses/data/roiExport3/", "Synapses XMLs (*.xml)");
	this->updateViewer (NULL, fileName);
  }
}

void SynapsesViewer::openSVF()
{
  //QString fileName = QFileDialog::getOpenFileName(this, tr("Open File"), QDir::currentPath());
  QString svfFileName = QFileDialog::getOpenFileName(this, tr("Open File"), "/usr/sci/crcnsdata/CRCNS/Synapses/data/roiExport3/", "Synapses Viewer Files (*.svf)");
  openSVFFile (svfFileName);
}

void SynapsesViewer::setAddSynapseMode()
{
	synapseMode = ADD_SYNAPSE_MODE;
	printf ("Synapse Mode set to Add\n");
}

void SynapsesViewer::setMoveSynapseMode()
{
	synapseMode = MOVE_SYNAPSE_MODE;
	printf ("Synapse Mode set to Move\n");
}

void SynapsesViewer::setRemoveSynapseMode()
{
	synapseMode = REMOVE_SYNAPSE_MODE;
	printf ("Synapse Mode set to Remove\n");
}

void SynapsesViewer::openSVFFile(QString svfFileName)
{
  QFile xmlFile (svfFileName);
  svfElements = new QDomDocument ();

  svfElements->setContent(&xmlFile);
  QDomElement root = svfElements->documentElement();
  if (root.tagName() != "synapsesViewerFile")
        QMessageBox::information(window(), tr("Synapses Viewer"), tr("This file %1 is not a valid SVF file.").arg(svfFileName));
  QString originalImageFileName = root.firstChildElement("originalImage").text();
  QString thresholdImageFileName = root.firstChildElement("thresholdImage").text();
  QString xmlFileName = root.firstChildElement("xml").text();

  this->updateViewer (originalImageFileName, xmlFileName);
}

void SynapsesViewer::setScale (int newScale)
{
  qreal scalingFactor = qreal (newScale) / qreal (imageScale);
  if (!(newScale < 10) && !(newScale > 100))
  {
    //QMessageBox::information(this, tr("Synapses Viewer"), tr("newScale = %1, value = %2.").arg(newScale).arg(scaleSlider->value()));
    QMatrix oldMatrix = view->matrix();
    view->translate(oldMatrix.dx(), oldMatrix.dy());
    view->scale(scalingFactor, scalingFactor);
	//printf ("slider value = %d , newScale Value = %d, scaling Factor = %f\n", scaleSlider->value(), newScale, scalingFactor);
	imageScale = newScale;
  }
}

void SynapsesViewer::zoomInPicture()
{
  int currentScale = imageScale;
  int newScale = currentScale + 10;
  if (viewerController->getScaleSlider()->value() != newScale)
	  viewerController->getScaleSlider()->setValue(newScale);
}

void SynapsesViewer::zoomOutPicture()
{
  int currentScale = imageScale;
  int newScale = currentScale - 10;
  if (viewerController->getScaleSlider()->value() != newScale)
	  viewerController->getScaleSlider()->setValue(newScale);
}

void SynapsesViewer::editSynapseGroup()
{
  //printf ("Calling SynapsesViewer::editSynapseGroup()\n
  StandardMethodInit();
  if (synapseGroupManager->getNumberOfGroups() == 0)
	QMessageBox::information(this, tr("Synapses Viewer"), tr("No groups to Edit."));
  else
  {
    synapseGroupPropertyEditor = new SynapseGroupPropertyEditor();
	synapseGroupPropertyEditor->setDialogDefaults (synapseGroupManager);
  }
}

void SynapsesViewer::setActivePicture()
{
  StandardMethodInit ();
  view->show();
}

void SynapsesViewer::createActions()
{
  openPictureAct = new QAction(tr("&Open Image"), this);
  openPictureAct->setShortcut(tr("Ctrl+O"));
  connect(openPictureAct, SIGNAL(triggered()), this, SLOT(openPicture()));

  openSynapsesAct = new QAction(tr("Open S&ynapses"), this);
  openSynapsesAct->setShortcut(tr("Ctrl+Y"));
  connect(openSynapsesAct, SIGNAL(triggered()), this, SLOT(openSynapses()));

  openSVFAct = new QAction(tr("Open S&VF"), this);
  openSVFAct->setShortcut(tr("Ctrl+V"));
  connect(openSVFAct, SIGNAL(triggered()), this, SLOT(openSVF()));

  exitAct = new QAction(tr("E&xit"), this);
  exitAct->setShortcut(tr("Ctrl+Q"));
  connect(exitAct, SIGNAL(triggered()), this, SLOT(close()));

  zoomInPictureAct = new QAction(tr("Zoom &In"), this);
  zoomInPictureAct->setShortcut(tr("Ctrl+="));
  connect(zoomInPictureAct, SIGNAL(triggered()), this, SLOT(zoomInPicture()));

  zoomOutPictureAct = new QAction(tr("Zoom &Out"), this);
  zoomOutPictureAct->setShortcut(tr("Ctrl+-"));
  connect(zoomOutPictureAct, SIGNAL(triggered()), this, SLOT(zoomOutPicture()));

  synapseGroupVisibilityAct = new QAction(tr("Synapse Group Visibility"), this);
  imageGroupVisibilityAct = new QAction(tr("Image Visibility"), this);

  imageVisibilityGroup = new QActionGroup(this);
  connect(imageVisibilityGroup, SIGNAL(triggered(QAction*)), this, SLOT(setActivePicture()));

  setSynapseGroupPropertiesAct = new QAction(tr("Edit Synapse Group"), this);
  connect(setSynapseGroupPropertiesAct, SIGNAL(triggered()), this, SLOT(editSynapseGroup()));
  
  addSynapseModeAct = new QAction(tr("Add Synapse"), this);
  connect(addSynapseModeAct, SIGNAL(triggered()), this, SLOT(setAddSynapseMode()));

  moveSynapseModeAct = new QAction(tr("Move Synapse"), this);
  connect(moveSynapseModeAct, SIGNAL(triggered()), this, SLOT(setMoveSynapseMode()));

  removeSynapseModeAct = new QAction(tr("Remove Synapse"), this);
  connect(removeSynapseModeAct, SIGNAL(triggered()), this, SLOT(setRemoveSynapseMode()));
}

void SynapsesViewer::createMenus()
{
  fileMenu = new QMenu(tr("&File"), this);
  fileMenu->addAction(openPictureAct);
  fileMenu->addAction(openSynapsesAct);
  fileMenu->addAction(openSVFAct);
  fileMenu->addSeparator();
  fileMenu->addAction(exitAct);

  viewMenu = new QMenu(tr("&View"), this);
  viewMenu->addAction(zoomInPictureAct);
  viewMenu->addAction(zoomOutPictureAct);
  viewMenu->addSeparator();
  viewMenu->addAction(imageGroupVisibilityAct);
  imageGroupVisibilityMenu = new QMenu(tr("Synapse Group Visibility"), this);
  imageGroupVisibilityAct->setMenu(imageGroupVisibilityMenu);
  viewMenu->addAction(synapseGroupVisibilityAct);
  synapseGroupVisibilityMenu = new QMenu(tr("Image Visibility"), this);
  synapseGroupVisibilityAct->setMenu(synapseGroupVisibilityMenu);

  toolsMenu = new QMenu(tr("&Tools"), this);
  toolsMenu->addAction(setSynapseGroupPropertiesAct);
  fileMenu->addSeparator();
  toolsMenu->addAction(addSynapseModeAct);
  toolsMenu->addAction(removeSynapseModeAct);
  toolsMenu->addAction(moveSynapseModeAct);

  menuBar()->addMenu(fileMenu);
  menuBar()->addMenu(viewMenu);
  menuBar()->addMenu(toolsMenu);
}
