// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : ThreadPoolDialog.cpp
// Author       : Pavel A. Koshevoy
// Created      : Thu Mar 22 09:30:06 MDT 2007
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : A modal dialog for the image processing thread pool.

// local includes:
#include "ThreadPoolDialog.h"

// the includes:
#include <ui/the_trail.hxx>
#include <thread/the_transaction.hxx>
#include <thread/the_mutex_interface.hxx>

// Qt includes:
#include <QPushButton>
#include <QTextEdit>
#include <QCloseEvent>
#include <QScrollBar>
#include <QTimer>
#include <QTime>

//----------------------------------------------------------------
// SUPPRESS_DEBUG
// 
#define SUPPRESS_DEBUG


//----------------------------------------------------------------
// ThreadPoolDialog::ThreadPoolDialog
// 
ThreadPoolDialog::ThreadPoolDialog(unsigned int num_threads,
				   QWidget * parent,
				   Qt::WFlags f):
  QDialog(parent, f),
  Ui::ThreadPoolDialog(),
  pool_(num_threads)
{
  Ui::ThreadPoolDialog::setupUi(this);
  
  connect(cancelButton, SIGNAL(clicked()),
	  this, SLOT(cancel()));
  
  connect(&pool_, SIGNAL(status_update(const QString &)),
	  this, SLOT(updateStatus(const QString &)));
  
  connect(&pool_, SIGNAL(transaction_finished(the_transaction_t *)),
	  this, SLOT(transactionFinished(the_transaction_t *)));
}

//----------------------------------------------------------------
// ThreadPoolDialog::~ThreadPoolDialog
// 
ThreadPoolDialog::~ThreadPoolDialog()
{
  pool_.stop();
  pool_.wait();
}

//----------------------------------------------------------------
// ThreadPoolDialog::exec
// 
int
ThreadPoolDialog::exec()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  QTimer::singleShot(0, this, SLOT(startThread()));
  return QDialog::exec();
}

//----------------------------------------------------------------
// ThreadPoolDialog::next
// 
void
ThreadPoolDialog::startThread()
{
#ifndef SUPPRESS_DEBUG
  cerr << "\n\nTRANSACTION STARTED AT "
       << qPrintable(QTime::currentTime().toString()) << endl << endl;
#endif
  
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  pool_.start();
}

//----------------------------------------------------------------
// ThreadPoolDialog::cancel
// 
void
ThreadPoolDialog::cancel()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
#ifndef SUPPRESS_DEBUG
  cerr << "\n\nTRANSACTION ABORT REQUESTED AT "
       << qPrintable(QTime::currentTime().toString()) << endl << endl;
#endif
  
  // NOTE: this will block the UI:
  pool_.stop();
  pool_.wait();
  
  done(QDialog::Rejected);
}

//----------------------------------------------------------------
// ThreadPoolDialog::transactionFinished
// 
void
ThreadPoolDialog::transactionFinished(the_transaction_t * t)
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
#ifndef SUPPRESS_DEBUG
  cerr << "\n\nTRANSACTION STOPPED AT "
       << qPrintable(QTime::currentTime().toString()) << endl << endl;
#endif
  
  if (t->state() == the_transaction_t::DONE_E)
  {
    if (!pool_.has_work())
    {
      pool_.stop();
      pool_.wait();
      done(QDialog::Accepted);
    }
#ifndef SUPPRESS_DEBUG
    else
    {
      cerr << "POOL STILL HAS WORK" << endl;
    }
#endif
  }
}

//----------------------------------------------------------------
// ThreadPoolDialog::updateStatus
// 
void
ThreadPoolDialog::updateStatus(const QString & status)
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  QScrollBar * vbar = statusTextView->verticalScrollBar();
  int ia = vbar->maximum();
  int ib = vbar->value();
  
  QTextCursor cursor = statusTextView->textCursor();
  cursor.movePosition(QTextCursor::End);
  cursor.insertText(status);
  if (ia != ib) return;
  
  vbar->setValue(vbar->maximum());
}

//----------------------------------------------------------------
// ThreadPoolDialog::closeEvent
// 
void
ThreadPoolDialog::closeEvent(QCloseEvent * e)
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  e->ignore();
  ThreadPoolDialog::cancel();
}
