// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : mosaic_cache.cxx
// Author       : Pavel A. Koshevoy
// Created      : Wed Oct 18 10:37:00 MDT 2005
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : mosaic cache -- supports on-demand mosaic image regeneration.

// local includes:
#include "mosaic_cache.hxx"

// the includes:
#include <thread/the_terminator.hxx>
#include <thread/the_thread_storage.hxx>
#include <utils/the_exception.hxx>


//----------------------------------------------------------------
// DEBUG_CLUMPING
// 
// #define DEBUG_CLUMPING

//----------------------------------------------------------------
// DEBUG_MOSAIC
// 
// #define DEBUG_MOSAIC

//----------------------------------------------------------------
// DEBUG_TINT
// 
// #define DEBUG_TINT

//----------------------------------------------------------------
// get_bbox
// 
inline static void
get_bbox(const the_aa_bbox_t & bbox, pnt2d_t & min, pnt2d_t & max)
{
  min[0] = bbox.min_[0];
  min[1] = bbox.min_[1];
  max[0] = bbox.max_[0];
  max[1] = bbox.max_[1];
}

//----------------------------------------------------------------
// inside_bbox
// 
inline static bool
inside_bbox(const the_aa_bbox_t & bbox, const pnt2d_t & pt)
{
  return
    bbox.min_[0] <= pt[0] &&
    pt[0] <= bbox.max_[0] &&
    bbox.min_[1] <= pt[1] &&
    pt[1] <= bbox.max_[1];
}

//----------------------------------------------------------------
// inside_bbox
// 
inline static bool
inside_bbox(const the_aa_bbox_t & bbox, const p3x1_t & pt)
{
  return
    bbox.min_[0] <= pt[0] &&
    pt[0] <= bbox.max_[0] &&
    bbox.min_[1] <= pt[1] &&
    pt[1] <= bbox.max_[1];
}

//----------------------------------------------------------------
// get_overlap_stats
// 
static void
get_overlap_stats(const the_aa_bbox_t & a,
		  const the_aa_bbox_t & b,
		  double & max_overlap_area,
		  double & overlap_area,
		  double & area_a,
		  double & area_b)
{
  const double w0 = a.max_[0] - a.min_[0];
  const double h0 = a.max_[1] - a.min_[1];
  
  const double w1 = b.max_[0] - b.min_[0];
  const double h1 = b.max_[1] - b.min_[1];
  
  max_overlap_area = std::min(w0, w1) * std::min(h0, h1);
  area_a = w0 * h0;
  area_b = w1 * h1;
  
  double ul_x =
    std::max(0.0, std::min(w1, double(a.min_.x()) - double(b.min_[0])));
  double ul_y =
    std::max(0.0, std::min(h1, double(a.min_.y()) - double(b.min_[1])));
  
  double lr_x =
    std::max(0.0, std::min(w1, double(a.max_.x()) - double(b.min_[0])));
  double lr_y =
    std::max(0.0, std::min(h1, double(a.max_.y()) - double(b.min_[1])));
  
  const double dx = lr_x - ul_x;
  const double dy = lr_y - ul_y;
  overlap_area = (dx * dy);
}

//----------------------------------------------------------------
// get_region_map
// 
static void
get_region_map(const the_aa_bbox_t & region_boundary,
	       std::list<the_aa_bbox_t> & regions,
	       std::vector<double> & x,
	       std::vector<double> & y,
	       array2d(bool) & region_map)
{
  // 1. clamp the regions to the requested boundary:
  for (std::list<the_aa_bbox_t>::iterator i = regions.begin();
       i != regions.end(); ++i)
  {
    the_aa_bbox_t & region = *i;
    region.clamp(region_boundary);
  }
  
  // 2. figure out the scanline layouts:
  std::list<double> x_list;
  std::list<double> y_list;
  
  x_list.push_back(region_boundary.min_.x());
  x_list.push_back(region_boundary.max_.x());
  y_list.push_back(region_boundary.min_.y());
  y_list.push_back(region_boundary.max_.y());
  
  for (std::list<the_aa_bbox_t>::const_iterator i = regions.begin();
       i != regions.end(); ++i)
  {
    const the_aa_bbox_t & region = *i;
    x_list.push_back(region.min_.x());
    x_list.push_back(region.max_.x());
    y_list.push_back(region.min_.y());
    y_list.push_back(region.max_.y());
  }
  
  x_list.sort();
  x_list.unique();
  
  y_list.sort();
  y_list.unique();
  
  x.assign(x_list.begin(), x_list.end());
  y.assign(y_list.begin(), y_list.end());
  
  // 3. build a cell table indicating the regions region cells:
  unsigned int rows = y.size() - 1;
  unsigned int cols = x.size() - 1;
  region_map.resize(rows);
  
  for (unsigned int row = 0; row < rows; row++)
  {
    region_map[row].assign(cols, false);
    for (unsigned int col = 0; col < cols; col++)
    {
      p3x1_t min(x[col], y[row], 0);
      p3x1_t max(x[col + 1], y[row + 1], 0);
      
      // check whether the current cell is contained
      // by any of the regions regions:
      for (std::list<the_aa_bbox_t>::const_iterator i = regions.begin();
	   i != regions.end(); ++i)
      {
	const the_aa_bbox_t & region = *i;
	if (region.contains(min) && region.contains(max))
	{
	  region_map[row][col] = true;
	  break;
	}
      }
    }
  }
  
#ifdef DEBUG_CLUMPING
  cerr << " -- regions: -------------------------------------" << endl;
  for (unsigned int row = 0; row < rows; row++)
  {
    for (unsigned int col = 0; col < cols; col++)
    {
      char c = region_map[row][col] ? 'R' : '.';
      cerr << c;
    }
    cerr << endl;
  }
  
  for (unsigned int col = 0; col <= cols; col++)
  {
    cerr << "x[" << col << "] = " << x[col] << endl;
  }
  
  for (unsigned int row = 0; row <= rows; row++)
  {
    cerr << "y[" << row << "] = " << y[row] << endl;
  }
#endif // DEBUG_CLUMPING
}

//----------------------------------------------------------------
// get_clumps
// 
static void
get_clumps(const the_aa_bbox_t & region_boundary,
	   std::list<the_aa_bbox_t> & regions,
	   std::list<the_aa_bbox_t> & clumps)
{
  std::vector<double> x;
  std::vector<double> y;
  array2d(bool) is_region;
  
  get_region_map(region_boundary, regions, x, y, is_region);
  unsigned int rows = y.size() - 1;
  unsigned int cols = x.size() - 1;
  
  array2d(bool) is_taken;
  assign(is_taken, rows, cols, false);

#ifdef DEBUG_CLUMPING
  array2d(char) clump_map;
  assign(clump_map, rows, cols, '.');
  char clump_id = 'a';
#endif
  
  for (unsigned int row = 0; row < rows; row++)
  {
    for (unsigned int col = 0; col < cols; col++)
    {
      if (is_taken[row][col] || !is_region[row][col]) continue;
      
      the_aa_bbox_t clump;
      clump << p3x1_t(x[col], y[row], 0)
	    << p3x1_t(x[col + 1], y[row + 1], 0);
      is_taken[row][col] = true;

#ifdef DEBUG_CLUMPING
      clump_map[row][col] = clump_id;
#endif // DEBUG_CLUMPING
      
      // find the max square ready region clump:
      unsigned int max_size = std::max(rows - row, cols - col);
      unsigned int size = 1;
      
      // the stop flags:
      //   [0] == true  -->  obstructed on the right
      //   [1] == true  -->  obstructed on the bottom
      bool stop[] = { false, false };
      
      for (; size < max_size; size++)
      {
	stop[0] = !(col + size < cols);
	stop[1] = !(row + size < rows);
	
	for (unsigned int i = 0; i <= size && !stop[0] && !stop[1]; i++)
	{
	  // check for obstruction on the right:
	  stop[0] =
	    is_taken[row + i][col + size] || !is_region[row + i][col + size];
	  
	  // check for obstruction on the bottom:
	  stop[1] =
	    is_taken[row + size][col + i] || !is_region[row + size][col + i];
	}
	
	// check for obstruction:
	if (stop[0] || stop[1]) break;
	
	// take the regions:
	for (unsigned int i = 0; i <= size; i++)
	{
	  is_taken[row + i][col + size] = true;
	  is_taken[row + size][col + i] = true;
	  
#ifdef DEBUG_CLUMPING
	  clump_map[row + i][col + size] = clump_id;
	  clump_map[row + size][col + i] = clump_id;
#endif // DEBUG_CLUMPING
	}
	
	// add the regions to the clump:
	p3x1_t max(x[col + size + 1], y[row + size + 1], 0);
	clump << max;
      }
      
#if 1
      // it may still be possible to expand in the non-obstructed direction:
      unsigned int expand = stop[0] ? 1 : 0;
      unsigned int min_size = size;
      max_size = stop[0] ? rows - row : cols - col;
      for (; size < max_size && !stop[expand]; size++)
      {
	for (unsigned int i = 0; i < min_size && !stop[expand]; i++)
	{
	  if (expand == 0)
	  {
	    // check for obstruction on the right:
	    stop[expand] =
	      is_taken[row + i][col + size] || !is_region[row + i][col + size];
	  }
	  else
	  {
	    // check for obstruction on the left:
	    stop[expand] =
	      is_taken[row + size][col + i] || !is_region[row + size][col + i];
	  }
	}
	
	if (stop[expand]) break;
	
	if (expand == 0)
	{
	  // take the regions:
	  for (unsigned int i = 0; i < min_size; i++)
	  {
	    is_taken[row + i][col + size] = true;
#ifdef DEBUG_CLUMPING
	    clump_map[row + i][col + size] = clump_id;
#endif // DEBUG_CLUMPING
	  }
	  
	  // expand the clump on the right:
	  p3x1_t max(x[col + size + 1], y[row + min_size], 0);
	  clump << max;
	}
	else
	{
	  // take the regions:
	  for (unsigned int i = 0; i < min_size; i++)
	  {
	    is_taken[row + size][col + i] = true;
#ifdef DEBUG_CLUMPING
	    clump_map[row + size][col + i] = clump_id;
#endif // DEBUG_CLUMPING
	  }
	  
	  // expand the clump towards the bottom:
	  p3x1_t max(x[col + min_size], y[row + size + 1], 0);
	  clump << max;
	}
      }
#endif
      
#ifdef DEBUG_CLUMPING
      clump_id++;
#endif // DEBUG_CLUMPING
      
      clumps.push_back(clump);
    }
  }
  
#ifdef DEBUG_CLUMPING
  cerr << " -- clumps -----------------------------------------------" << endl;
  for (unsigned int row = 0; row < rows; row++)
  {
    for (unsigned int col = 0; col < cols; col++)
    {
      cerr << clump_map[row][col];
    }
    cerr << endl;
  }
#endif // DEBUG_CLUMPING
}

//----------------------------------------------------------------
// blend_to_rgbaf
// 
inline static void
blend_to_rgbaf(unsigned int num_tiles,
	       const double * lum,
	       const double * alf,
	       const the_color_t * tint,
	       float * rgba)
{
  float c[] = { 0, 0, 0, 0 };
  float t[] = { 0, 0, 0 };
  
  for (unsigned int j = 0; j < num_tiles; j++)
  {
    float a = alf[j];
    float l = lum[j];
    float wa = a / 255.0f;
    
    float tw[] = {
      wa * tint[j][0],
      wa * tint[j][1],
      wa * tint[j][2]
    };
    
    c[0] += l * tw[0];
    c[1] += l * tw[1];
    c[2] += l * tw[2];
    c[3] = std::max(c[3], a);
    
    t[0] += wa * tw[0];
    t[1] += wa * tw[1];
    t[2] += wa * tw[2];
  }
  
#if 1
  float t_max = std::max(t[0], std::max(t[1], t[2]));
  if (t_max > 1)
  {
    float s = 1 / t_max;
    c[0] *= s;
    c[1] *= s;
    c[2] *= s;
  }
#elif 0
  if (t[0] > 1) c[0] /= t[0];
  if (t[1] > 1) c[1] /= t[1];
  if (t[2] > 1) c[2] /= t[2];
#else
  c[0] /= 1e-8f + t[0];
  c[1] /= 1e-8f + t[1];
  c[2] /= 1e-8f + t[2];
#endif
  
  rgba[0] = std::min(255.0f, std::max(0.0f, c[0]));
  rgba[1] = std::min(255.0f, std::max(0.0f, c[1]));
  rgba[2] = std::min(255.0f, std::max(0.0f, c[2]));
  rgba[3] = c[3];
}

//----------------------------------------------------------------
// blend_to_bgra
// 
inline static void
blend_to_bgra(unsigned int num_tiles,
	      const double * lum,
	      const double * alf,
	      const the_color_t * tint,
	      unsigned char * bgra_bytes)
{
  float c[] = { 0, 0, 0, 0 };
  blend_to_rgbaf(num_tiles, lum, alf, tint, c);
  
#ifdef __BIG_ENDIAN__
  #define B 3
  #define G 2
  #define R 1
  #define A 0
#else
  #define B 0
  #define G 1
  #define R 2
  #define A 3
#endif
  
  bgra_bytes[R] = (unsigned char)(c[0]);
  bgra_bytes[G] = (unsigned char)(c[1]);
  bgra_bytes[B] = (unsigned char)(c[2]);
  bgra_bytes[A] = (unsigned char)(c[3]);
  
#undef R
#undef G
#undef B
#undef A
}


//----------------------------------------------------------------
// mosaic_cache_shared_data_t::mosaic_cache_shared_data_t
// 
mosaic_cache_shared_data_t::mosaic_cache_shared_data_t():
  work_in_progress_(0),
  mutex_(the_mutex_interface_t::create())
{
#ifdef DEBUG_MOSAIC
  cerr << the_thread_storage().thread_id() << ". "
       << "create shared data: " << this << endl;
#endif // DEBUG_MOSAIC
}

//----------------------------------------------------------------
// mosaic_cache_shared_data_t::mosaic_cache_shared_data_t
// 
mosaic_cache_shared_data_t::
mosaic_cache_shared_data_t
(const unsigned int & active,
 const std::vector<base_transform_t::ConstPointer> & xforms,
 const image_tile_generator_t & mosaic,
 const the_aa_bbox_t & mosaic_bbox,
 const unsigned int & bytes_per_pixel):
  active_(active),
  transform_(xforms),
  mosaic_(mosaic),
  mosaic_bbox_(mosaic_bbox),
  work_in_progress_(0),
  mutex_(the_mutex_interface_t::create())
{
#ifdef DEBUG_MOSAIC
  cerr << the_thread_storage().thread_id() << ". "
       << "create shared data: " << this << ", "
       << mosaic.w_pad_ << ' ' << mosaic.h_pad_
       << endl;
#endif // DEBUG_MOSAIC
  
  // sanity test (refuse to allocate enormous textures):
  static const double _1MB = 1024.0 * 1024.0;
  double size_MB =
    double(mosaic_.w_pad_) *
    double(mosaic_.h_pad_) *
    double(bytes_per_pixel) / _1MB;
  
  static const unsigned int texture_ram = 256;
  if (size_MB > double(texture_ram) * 0.75)
  {
    std::ostringstream os;
    os << "will not allocate texture of size " << size_MB << "MB" << endl;
    the_exception_t e(os.str().c_str());
    throw e;
  }
  
  // allocate the mosaic:
  mosaic_.allocate(bytes_per_pixel);
}

//----------------------------------------------------------------
// mosaic_cache_shared_data_t::~mosaic_cache_shared_data_t
// 
mosaic_cache_shared_data_t::~mosaic_cache_shared_data_t()
{
  mutex_->lock();
#ifdef DEBUG_MOSAIC
  cerr << the_thread_storage().thread_id() << ". "
       << "delete shared data: " << this
       << ", work in progress: " << work_in_progress_ << endl;
#endif // DEBUG_MOSAIC
  assert(work_in_progress_ == 0);
  mutex_->delete_this();
  mutex_ = NULL;
}

//----------------------------------------------------------------
// mosaic_cache_shared_data_t::make_tiles
// 
void
mosaic_cache_shared_data_t::make_tiles(const GLenum & data_type,
				       const GLenum & format_internal,
				       const GLenum & format,
				       const unsigned int max_texture)
{
  // cut new tiles:
  mosaic_.make_tiles(data_type,
		     format_internal,
		     format,
		     max_texture,
		     mosaic_.bbox_.min_[0],
		     mosaic_.bbox_.min_[1],
		     mosaic_.bbox_.max_[0],
		     mosaic_.bbox_.max_[1]);
  
  // update the shared image tiles structure:
  data_.buffer_ = mosaic_.buffer_;
  data_.tiles_ = mosaic_.tiles_;
}

//----------------------------------------------------------------
// mosaic_cache_shared_data_t::is_ready
// 
bool
mosaic_cache_shared_data_t::
is_ready(const the_aa_bbox_t & requested_region,
	 const std::vector<base_transform_t::ConstPointer> & xforms,
	 const unsigned int & active) const
{
  the_lock_t<the_mutex_interface_t> locker(mutex_);
  return
    same_mosaic(xforms, active) &&
    region_ready(requested_region);
}

//----------------------------------------------------------------
// mosaic_cache_shared_data_t::same_mosaic
// 
// check whether the mosaic is the same:
// 
bool
mosaic_cache_shared_data_t::
same_mosaic(const std::vector<base_transform_t::ConstPointer> & xforms,
	    const unsigned int & active) const
{
  if (active_ != active)
  {
    return false;
  }
  
  unsigned int num_xforms = xforms.size();
  if (num_xforms != transform_.size()) return false;
  
  for (unsigned int i = 0; i < num_xforms; i++)
  {
    const base_transform_t * t_old = transform_[i].GetPointer();
    const base_transform_t * t_new = xforms[i].GetPointer();
    
    if ((t_old == NULL && t_new != NULL) ||
	(t_old != NULL && t_new == NULL))
    {
      return false;
    }
    
    // this is retarder -- ITK may throw an exception if you access the
    // fixed parameters of the transform, and that transform doesn't
    // have any fixed parameters and it hasn't reimplemented the
    // itk::Transform::GetFixedParameters() method:
    try
    {
      const base_transform_t::ParametersType & new_params_fixed =
	t_new->GetFixedParameters();
      
      const base_transform_t::ParametersType & old_params_fixed =
	t_old->GetFixedParameters();
      
      if (new_params_fixed.size() != old_params_fixed.size())
      {
	return false;
      }
      
      for (unsigned int i = 0; i < new_params_fixed.size(); i++)
      {
	if (new_params_fixed[i] != old_params_fixed[i])
	{
	  return false;
	}
      }
    }
    catch (...)
    {}
    
    const base_transform_t::ParametersType & new_params =
      t_new->GetParameters();
    
    const base_transform_t::ParametersType & old_params =
      t_old->GetParameters();
    
    if (new_params.size() != old_params.size())
    {
      return false;
    }
    
    for (unsigned int i = 0; i < new_params.size(); i++)
    {
      if (new_params[i] != old_params[i])
      {
	return false;
      }
    }
  }
  
  return true;
}

//----------------------------------------------------------------
// mosaic_cache_shared_data_t::region_ready
// 
bool
mosaic_cache_shared_data_t::
region_ready(const the_aa_bbox_t & requested_region) const
{
  // if the requested region is outside of the mosaic,
  // than it is automatiacally ready -- there is nothing to do there:
  if (!requested_region.intersects(mosaic_bbox_))
  {
    return true;
  }
  
  // get the work regions:
  the_aa_bbox_t clamped_region(requested_region);
  clamped_region.clamp(mosaic_.bbox_);
  
  std::list<the_aa_bbox_t> work;
  get_work_regions(clamped_region, work);
  
  return work.empty();
}

//----------------------------------------------------------------
// mosaic_cache_shared_data_t::find_overlapping_regions
// 
// find all ready regions that overlap the requested region
// by at least as much as the given overlap ratio:
// 
void
mosaic_cache_shared_data_t::
find_overlapping_regions(std::list<the_aa_bbox_t> & found,
			 const the_aa_bbox_t & requested,
			 const double & overlap_ratio_threshold) const
{
  assert(mosaic_.bbox_.contains(requested));
  
  double max_overlap_area;
  double overlap_area;
  double area_requested;
  double area_ready;
  
  for (std::list<the_aa_bbox_t>::const_iterator i = ready_.begin();
       i != ready_.end(); ++i)
  {
    const the_aa_bbox_t & ready = *i;
    get_overlap_stats(requested,
		      ready,
		      max_overlap_area,
		      overlap_area,
		      area_requested,
		      area_ready);
    
    double overlap_ratio = overlap_area / max_overlap_area;
    if (overlap_area != 0.0 && overlap_ratio >= overlap_ratio_threshold)
    {
      found.push_back(ready);
    }
  }
}

//----------------------------------------------------------------
// add_work_region
// 
static void
add_work_region(std::list<the_aa_bbox_t> & work,
		const the_aa_bbox_t & work_region,
		const double & target_region_area)
{
  if (work_region.is_empty()) return;
  
  // FIXME:
  // work.push_back(work_region);
  // return;
  
  // split up the region into pieces as necessary:
  if (target_region_area > 0.0)
  {
    double w = work_region.max_[0] - work_region.min_[0];
    double h = work_region.max_[1] - work_region.min_[1];
    double area = w * h;
    unsigned int rows = int(std::max(1.0, area / target_region_area));
    double dh = h / double(rows);
    the_aa_bbox_t work_portion(work_region);
    for (unsigned int r = 0; r < rows; r++)
    {
      work_portion.min_[1] = work_region.min_[1] + r * dh;
      work_portion.max_[1] = std::min(work_portion.min_[1] + dh,
				      double(work_region.max_[1]));
      work.push_back(work_portion);
    }
  }
  else
  {
    work.push_back(work_region);
  }
}

//----------------------------------------------------------------
// mosaic_cache_shared_data_t::get_work_regions
// 
void
mosaic_cache_shared_data_t::
get_work_regions(const the_aa_bbox_t & requested_region,
		 std::list<the_aa_bbox_t> & work,
		 const double & target_region_area) const
{
  assert(mosaic_.bbox_.contains(requested_region));
  
  std::list<the_aa_bbox_t> ready;
  find_overlapping_regions(ready, requested_region, 0.0);
  
  if (ready.empty())
  {
    // the trivial case -- nothing has been cached yet:
    add_work_region(work, requested_region, target_region_area);
    return;
  }
  
  // generate a set of smaller work regions excluding the ready regions:
  
  // 1. clamp the ready regions to the requested region:
  std::vector<double> x;
  std::vector<double> y;
  array2d(bool) cell_is_ready;
  
  get_region_map(requested_region, ready, x, y, cell_is_ready);
  
  unsigned int rows = y.size() - 1;
  unsigned int cols = x.size() - 1;
  
  // make work regions out of contiguous horizontal scanlines
  // within the cell table:
  std::list<unsigned int> scanline;
  
  for (unsigned int row = 0; row < rows; row++)
  {
    for (unsigned int col = 0; col <= cols; col++)
    {
      if (col < cols && !cell_is_ready[row][col])
      {
	scanline.push_back(col);
	continue;
      }
      
      // process the scanline:
      the_aa_bbox_t work_region;
      for (std::list<unsigned int>::const_iterator i = scanline.begin();
	   i != scanline.end(); ++i)
      {
	const unsigned int & col = *i;
	p3x1_t min(x[col], y[row], 0);
	p3x1_t max(x[col + 1], y[row + 1], 0);
	work_region << min << max;
      }
      
      add_work_region(work, work_region, target_region_area);
      
      // clear the scanline:
      scanline.clear();
    }
  }
}

//----------------------------------------------------------------
// mosaic_cache_shared_data_t::update
// 
// update the mosaic for the requested region:
// 
void
mosaic_cache_shared_data_t::
update(const the_aa_bbox_t & requested_region,
       const std::vector<image_tile_generator_t> & tile,
       const std::vector<the_color_t> & tint)
{
  mutex_->lock();
  the_scoped_increment_t<unsigned int> tmp(work_in_progress_);
  mutex_->unlock();
  
  // FIXME: TODO:
  // 1. remove from consideration tiles which do not overlap with
  //    the requested region
  
  // get the work regions:
  {
    the_lock_t<the_mutex_interface_t> locker(mutex_);
    the_aa_bbox_t clamped_region(requested_region);
    clamped_region.clamp(mosaic_.bbox_);
    
#ifdef DEBUG_MOSAIC
    static int counter = 0;
    cerr << counter << ". thread " << the_thread_storage().thread_id()
	 << ", shared data " << this
	 << ", work in progress: " << work_in_progress_
	 << " updating " << clamped_region << endl;
    if (counter >= 49)
    {
      cerr << "counter == " << counter << endl;
    }
    counter++;
#endif // DEBUG_MOSAIC
    
    if (clamped_region != work_region_)
    {
      work_region_ = clamped_region;
      work_.clear();
      get_work_regions(work_region_, work_,
		       // 4096 * mosaic_.spacing_x_ * mosaic_.spacing_y_);
		       32768 * mosaic_.spacing_x_ * mosaic_.spacing_y_);
    }
    if (work_.empty()) return;
  }
  
  // process the work regions:
  itk_terminator_t terminator("mosaic_cache_shared_data_t::update");
  
  while (true)
  {
    terminator.terminate_on_request();
    the_aa_bbox_t next_region;
    {
      the_lock_t<the_mutex_interface_t> locker(mutex_);
      if (work_.empty()) break;
      next_region = remove_head(work_);
    }
    
    try
    {
      the_aa_bbox_t processed;
      image_tile_t::quad_t quad;
      process(next_region,
	      tile,
	      tint,
	      processed,
	      quad);
      
      // salvage as much work as possible, even if the process was terminated:
      if (quad.h_ > 0)
      {
	// avoid simultaneous access to the ready list and upload list:
	the_lock_t<the_mutex_interface_t> locker(mutex_);
	ready_.push_back(processed);
	upload_.push_back(quad);
      }
    }
    catch (...)
    {
      the_lock_t<the_mutex_interface_t> locker(mutex_);
      work_.push_front(next_region);
#ifdef DEBUG_MOSAIC
      cerr << "thread " << the_thread_storage().thread_id()
	   << ". shared data " << this
	   << " caught exception, region " << next_region << endl;
#endif // DEBUG_MOSAIC
      throw;
    }
  }
  
  // the overlapping/adjacent ready regions should be
  // clumped together (defragmented):
  the_lock_t<the_mutex_interface_t> locker(mutex_);
  if (work_in_progress_ == 1)
  {
    std::list<the_aa_bbox_t> ready_copy(ready_);
    std::list<the_aa_bbox_t> ready_more;
    get_clumps(mosaic_.bbox_, ready_copy, ready_more);
    
    ready_.clear();
    ready_.splice(ready_.end(), ready_more);
  }
}

//----------------------------------------------------------------
// calc_image_bboxes
// 
void
calc_image_bboxes(const std::vector<image_tile_generator_t> & image,
		  
		  // image space bounding boxes (for feathering):
		  std::vector<pnt2d_t> & image_min,
		  std::vector<pnt2d_t> & image_max)
{
  const unsigned int num_images = image.size();
  image_min.resize(num_images);
  image_max.resize(num_images);
  
  // calculate the bounding boxes:
  for (unsigned int i = 0; i < num_images; i++)
  {
    // initialize an empty bounding box:
    get_bbox(image[i].bbox_, image_min[i], image_max[i]);
  }
}

//----------------------------------------------------------------
// mosaic_cache_shared_data_t::process
// 
// update the requested region of the mosaic:
// 
bool
mosaic_cache_shared_data_t::
process(const the_aa_bbox_t & clamped_region,
	const std::vector<image_tile_generator_t> & tile,
	const std::vector<the_color_t> & tile_colors,
	the_aa_bbox_t & processed,
	image_tile_t::quad_t & upload)
{
  // itk_terminator_t terminator("mosaic_cache_t::process");
#ifdef DEBUG_TINT
  std::vector<the_color_t> tint_vec(tile.size());
  for (unsigned int i = 0; i < tile.size(); i++)
  {
    tint_vec[i] = hsv_to_rgb(the_color_t(drand(), 1, 1));
  }
  const the_color_t * tint = &tint_vec[0];
#else
  const the_color_t * tint = &tile_colors[0];
#endif
  
  // shortcut:
  const unsigned int num_tiles = tile.size();
  assert(num_tiles == transform_.size());
  
  // sanity check:
  assert(mosaic_.bbox_.contains(clamped_region.min_) &&
	 mosaic_.bbox_.contains(clamped_region.max_));
  
  const double & origin_x = mosaic_.origin_x_;
  const double & origin_y = mosaic_.origin_y_;
  const double & spacing_x = mosaic_.spacing_x_;
  const double & spacing_y = mosaic_.spacing_y_;
  const double & roi_x = clamped_region.min_.x();
  const double & roi_y = clamped_region.min_.y();
  
  int ix = int(floor((roi_x - origin_x) / spacing_x));
  int iy = int(floor((roi_y - origin_y) / spacing_y));
  
  the_aa_bbox_t process_bbox;
  process_bbox << p3x1_t(origin_x + ix * spacing_x,
			 origin_y + iy * spacing_y,
			 0);
  
  const double roi_w = clamped_region.max_.x() - process_bbox.min_.x();
  const double roi_h = clamped_region.max_.y() - process_bbox.min_.y();
  
  int iw = int(ceil(roi_w / spacing_x));
  int ih = int(ceil(roi_h / spacing_y));
  
  iw = std::min(iw, std::max(0, int(mosaic_.w_) - ix));
  ih = std::min(ih, std::max(0, int(mosaic_.h_) - iy));
  
  process_bbox << p3x1_t(origin_x + (ix + iw) * spacing_x,
			 origin_y + (iy + ih) * spacing_y,
			 0);
  
  // sanity check:
  assert(ix >= 0 && iy >= 0 &&
	 (ix + iw <= int(mosaic_.w_)) &&
	 (iy + ih <= int(mosaic_.h_)));
  
  // image space bounding boxes (for accelerated hit/miss detection):
  std::vector<pnt2d_t> bbox_min(num_tiles);
  std::vector<pnt2d_t> bbox_max(num_tiles);
  calc_image_bboxes(tile, bbox_min, bbox_max);
  
  std::vector<pnt2d_t> mosaic_min(num_tiles);
  std::vector<pnt2d_t> mosaic_max(num_tiles);
  calc_mosaic_bboxes<pnt2d_t, base_transform_t::ConstPointer>
    (transform_, bbox_min, bbox_max, mosaic_min, mosaic_max);
  
  // expand the bounding boxes to allow some leeway within the
  // individual image pixel spacing bounds:
  for (unsigned int i = 0; i < num_tiles; i++)
  {
    mosaic_min[i][0] -= tile[i].spacing_x_;
    mosaic_min[i][1] -= tile[i].spacing_y_;
    mosaic_max[i][0] += tile[i].spacing_x_;
    mosaic_max[i][1] += tile[i].spacing_y_;
  }
  
  // reset the progress accumulators:
  processed.clear();
  processed << p3x1_t(origin_x + ix * spacing_x,
		      origin_y + iy * spacing_y,
		      0);
  
  upload.x_ = ix;
  upload.y_ = iy;
  upload.w_ = (iw + 2);
  upload.h_ = 0;
  
  // setup the luminance/alpha interpolator:
  bool la_compressed = tile[0].bytes_per_pixel_ == 1;
  la_interpolator_t interpolator = la_compressed
    ? interpolate_compressed_luminance_alpha
    : interpolate_luminance_alpha;
  
  bool mosaic_compressed = mosaic_.bytes_per_pixel_ == 2;
  if (mosaic_compressed)
  {
    // if using compressed pixels, the mosaic OpenGL texture will
    // be claimed as 1 byte per pixel, but in reality each pixels
    // is represented by 2 bytes, so the texture upload offset
    // and size have to be scaled accordingly:
    upload.x_ *= mosaic_.bytes_per_pixel_;
    upload.w_ *= mosaic_.bytes_per_pixel_;
    
    // on the other hand, if compression is not used, than 4 bytes will
    // be used per pixel, and no scaling is necessary.
  }
  
  // accumulators:
  std::vector<double> lum(num_tiles, 0.0);
  std::vector<double> alf(num_tiles, 0.0);
  
  // shortcuts:
  pnt2d_t itk_pt;
  pnt2d_t tmp_pt;
  const unsigned char * pixel[4];
  double weight[4];
  
  // iterate over the requested mosaic region:
  // for (int j = -1; j <= ih; j++)
  for (int j = 0; j < ih; j++)
  {
    int y = iy + j;
    unsigned char * dst_scanline = mosaic_.scanline(y);
    itk_pt[1] = origin_y + spacing_y * double(y);
    
    // for (int i = -1; i <= iw; i++)
    for (int i = 0; i < iw; i++)
    {
      // make sure there hasn't been an interrupt:
      /*
      if (terminator.termination_requested())
      {
	return true;
      }
      */
      
      int x = ix + i;
      itk_pt[0] = origin_x + spacing_x * double(x);
      
      for (unsigned int k = 0; k < num_tiles; k++)
      {
	lum[k] = 0;
	alf[k] = 0;
	if (inside_bbox(mosaic_min[k], mosaic_max[k], itk_pt))
	{
	  tmp_pt = transform_[k]->TransformPoint(itk_pt);
	  
	  if (tile[k].evaluate(tmp_pt[0], tmp_pt[1], pixel, weight))
	  {
	    // get luminance/alpha from the moving image:
	    interpolator(lum[k], alf[k], pixel, weight);
	  }
	}
      }
      
      unsigned char * dst_pixel = dst_scanline + x * mosaic_.bytes_per_pixel_;
      assert(dst_pixel + mosaic_.bytes_per_pixel_ <=
	     mosaic_.buffer_->data() + mosaic_.buffer_->size() &&
	     dst_pixel >= mosaic_.buffer_->data());
      
      if (mosaic_compressed)
      {
	float rgba[4];
	blend_to_rgbaf(num_tiles, &lum[0], &alf[0], tint, rgba);
	encode_rgba16(rgba, dst_pixel);
	
#if 0
	// sanity test:
	float test[4];
	decode_rgba16(dst_pixel, test);
	float err = std::max(fabs(rgba[0] - test[0]),
			     std::max(fabs(rgba[1] - test[1]),
				      fabs(rgba[2] - test[2])));
	if (err > 9.0)
	{
	  cerr << "ERROR -- the encode is not working: " << err << endl
	       << rgba[0] << endl
	       << rgba[1] << endl
	       << rgba[2] << endl
	       << rgba[3] << endl;
	}
#endif
      }
      else
      {
	blend_to_bgra(num_tiles, &lum[0], &alf[0], tint, dst_pixel);
      }
    }
    
    // a scanline is ready, update the progress accumulators:
    upload.h_ = j + 1 + 2;
    processed << p3x1_t(origin_x + (ix + iw) * spacing_x,
			origin_y + (iy + j + 1) * spacing_y,
			0);
  }
  
  return false;
}


//----------------------------------------------------------------
// mosaic_cache_t::mosaic_cache_t
// 
mosaic_cache_t::mosaic_cache_t(const unsigned int & id,
			       const double & spacing_x,
			       const double & spacing_y):
  id_(id),
  spacing_x_(spacing_x),
  spacing_y_(spacing_y),
  shared_(new mosaic_cache_shared_data_t())
{}

//----------------------------------------------------------------
// mosaic_cache_t::update
// 
// generate a mosaic for the requested region:
// 
void
mosaic_cache_t::
update(const the_aa_bbox_t & requested_region,
       const std::vector<image_tile_generator_t> & tile,
       const std::vector<the_color_t> & tint,
       const std::vector<base_transform_t::ConstPointer> & xforms,
       const unsigned int & active,
       const bool & mosaic_compressed)
{
  // make sure the shared data doesn't get deleted from some other thread:
  boost::shared_ptr<mosaic_cache_shared_data_t> current(shared_);
  
#ifdef DEBUG_MOSAIC
  cerr << the_thread_storage().thread_id()
       << ". current shared data: " << current.get() << endl;
#endif // DEBUG_MOSAIC
  
  // update the mosaic as necessary:
  {
    the_lock_t<the_mutex_interface_t> locker(mutex());
    if (!shared_->same_mosaic(xforms, active))
    {
      // calculate the new mosaic bounding box:
      const unsigned int num_tiles = tile.size();
      std::vector<pnt2d_t> bbox_min(num_tiles);
      std::vector<pnt2d_t> bbox_max(num_tiles);
      calc_image_bboxes(tile, bbox_min, bbox_max);
      
      std::vector<pnt2d_t> mosaic_min(num_tiles);
      std::vector<pnt2d_t> mosaic_max(num_tiles);
      calc_mosaic_bboxes<pnt2d_t, base_transform_t::ConstPointer>
	(xforms, bbox_min, bbox_max, mosaic_min, mosaic_max);
      
      pnt2d_t min;
      pnt2d_t max;
      calc_mosaic_bbox(mosaic_min, mosaic_max, min, max);
      
      the_aa_bbox_t mosaic_bbox;
      mosaic_bbox << p3x1_t(min[0], min[1], 0)
		  << p3x1_t(max[0], max[1], 0);
      
      // mosaic pixel spacing can not be smaller than the smallest
      // pixel spacing of the composing tile images:
      double spacing_x = tile[0].spacing_x_;
      double spacing_y = tile[0].spacing_y_;
      for (unsigned int i = 1; i < num_tiles; i++)
      {
	spacing_x = std::min(spacing_x, tile[i].spacing_x_);
	spacing_y = std::min(spacing_y, tile[i].spacing_y_);
      }
      spacing_x = std::max(spacing_x, spacing_x_);
      spacing_y = std::max(spacing_y, spacing_y_);
      
      // mosaic origin:
      double ox = mosaic_bbox.min_.x();
      double oy = mosaic_bbox.min_.y();
      
      // mosaic width and height in world coordinate system:
      double mw = mosaic_bbox.max_.x() - ox;
      double mh = mosaic_bbox.max_.y() - oy;
      
      // corresponding discrete image dimensions
      // (equal or greater than the mosaic dimensions):
      unsigned int iw = int(ceil(mw / spacing_x));
      unsigned int ih = int(ceil(mh / spacing_y));
      
      image_tile_generator_t mosaic;
      mosaic.layout(iw, ih, ox, oy, spacing_x, spacing_y);
      
      size_t bytes_per_pixel = mosaic_compressed ? 2 : 4;
#ifdef DEBUG_MOSAIC
      cerr << the_thread_storage().thread_id() << ". "
	   << id_ << " -- allocating "
	   << iw << " x " << ih << " x " << bytes_per_pixel
	   << " (" << mosaic.w_pad_ << " x " << mosaic.h_pad_ << " x "
	   << bytes_per_pixel << " = "
	   << mosaic.w_pad_ * mosaic.h_pad_ * bytes_per_pixel / (1 << 20)
	   << " MB) "<< endl;
#endif
      
      // NOTE: this will throw an exception if the mosaic image is too large:
      shared_ =
	boost::shared_ptr<mosaic_cache_shared_data_t>
	(new mosaic_cache_shared_data_t(active,
					xforms,
					mosaic,
					mosaic_bbox,
					bytes_per_pixel));
    }
  }
  
#ifdef DEBUG_MOSAIC
  cerr << the_thread_storage().thread_id() << ". "
       << "update shared data: " << shared_.get() << endl;
#endif
  
  // the shared data may have been updated, so transfer the current here:
  current = shared_;
  
  // update the requested region:
  current->update(requested_region, tile, tint);
  
#ifdef DEBUG_MOSAIC
  {
    the_lock_t<the_mutex_interface_t> locker(mutex());
    cerr << "thread " << the_thread_storage().thread_id()
	 << ". shared data " << current.get()
	 << ", work in progress: " << current->work_in_progress()
	 << " done with " << current->work_region_ << endl;
  }
#endif // DEBUG_MOSAIC
  
#ifdef DEBUG_MOSAIC
  cerr << the_thread_storage().thread_id() << ". "
       << "shared data updated: " << current.get() << endl;
#endif
}
