// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : mosaic_layout_controller.cxx
// Author       : Pavel A. Koshevoy
// Created      : Wed Mar 21 14:59:24 MDT 2007
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Wrapper class for automatic mosaic layout.

// local includes:
#include "mosaic_layout_controller.hxx"

// the includes:
#include <itk/common.hxx>
#include <itk/mosaic_layout_common.hxx>
#include <utils/the_log.hxx>
#include <thread/the_thread_interface.hxx>


//----------------------------------------------------------------
// mosaic_layout_controller_t::mosaic_layout_controller_t
// 
mosaic_layout_controller_t::
mosaic_layout_controller_t(std::list<the_tile_t *> & tile_primitives,
			   const unsigned int & num_tiles):
  num_tiles_(num_tiles),
  max_cascade_len_(num_tiles - 2),
  affine_(num_tiles),
  matched_(num_tiles, false),
  peaks_(num_tiles),
  optimal_level_(UINT_MAX),
  use_std_mask_(false),
  overlap_min_(0.05), // 5 percent
  overlap_max_(1.00), // 100 percent
  max_peaks_(3)
{
  // choose the optimal pyramid level at which the tiles will be matched:
  const the_tile_t * smallest_tile = NULL;
  unsigned int max_pyramid_size = UINT_MAX;
  for (std::list<the_tile_t *>::iterator iter = tile_primitives.begin();
       iter != tile_primitives.end(); ++iter)
  {
    const the_tile_t * t = *iter;
    const the_image_pyramid_t & pyramid = t->pyramid();
    if (pyramid.image_.size() < max_pyramid_size)
    {
      max_pyramid_size = pyramid.image_.size();
      smallest_tile = t;
    }
  }
  
  // once the minimum acceptable resolution level is found, try to
  // select the optimum resolution level -- 128x128 or better:
  unsigned int max_res_level = max_pyramid_size - 1;
  unsigned int min_res_level = max_pyramid_size - 1;
  {
    static const unsigned int min_edge = 128;
    static const unsigned int max_edge = 512;
    
    const the_image_pyramid_t & pyramid = smallest_tile->pyramid();
    while (max_pyramid_size != 1)
    {
      const image_tile_generator_t & img =
	pyramid.image_[max_pyramid_size - 1];
      
      if (img.w_ / min_edge == 0 && img.h_ / min_edge == 0)
      {
	min_res_level--;
      }
      
      if (img.w_ / max_edge == 0 && img.h_ / max_edge == 0)
      {
	max_res_level--;
      }
      
      max_pyramid_size--;
    }
  }
  // cerr << max_res_level << " .. " << min_res_level << endl;
  
  use_std_mask_ = false;
  resize(tile_pyramid_, min_res_level - max_res_level + 1, num_tiles);
  resize(mask_pyramid_, min_res_level - max_res_level + 1, num_tiles);
  unsigned int i = 0;
  for (std::list<the_tile_t *>::iterator iter = tile_primitives.begin();
       iter != tile_primitives.end(); ++iter, ++i)
  {
    the_tile_t * t = *iter;
    
    // the tile pyramid:
    the_image_pyramid_t & pyramid = t->pyramid();
    use_std_mask_ = use_std_mask_ || pyramid.use_std_mask_;
    for (unsigned int j = max_res_level; j <= min_res_level; j++)
    {
      tile_pyramid_[min_res_level - j][i] = pyramid.itk_image_[j];
      mask_pyramid_[min_res_level - j][i] = pyramid.itk_mask_[j];
    }
  }
  
  // initialize the path and cost:
  resize(peaks_, num_tiles_, num_tiles_);
  resize(cost_, max_cascade_len_ + 1, num_tiles_, num_tiles_);
  resize(path_, max_cascade_len_ + 1, num_tiles_, num_tiles_);
  reset(num_tiles_, max_cascade_len_, path_, cost_);
}

//----------------------------------------------------------------
// mosaic_layout_controller_t::~mosaic_layout_controller_t
// 
mosaic_layout_controller_t::~mosaic_layout_controller_t()
{}

//----------------------------------------------------------------
// mosaic_layout_controller_t::layout
// 
void
mosaic_layout_controller_t::layout(the_log_t & log)
{
  std::list<unsigned int> tile_order;
  layout_mosaic<native_image_t, mask_t>(log,
					num_tiles_,
					tile_pyramid_,
					mask_pyramid_,
					max_cascade_len_,
					affine_,
					matched_,
					tile_order,
					cost_,
					path_,
					optimal_level_ != 0,
					use_std_mask_,
					overlap_min_,
					overlap_max_,
					max_peaks_,
					false, // refining
					the_text_t(""));
}


//----------------------------------------------------------------
// tile_match_transaction_t::tile_match_transaction_t
// 
tile_match_transaction_t::
tile_match_transaction_t(mosaic_layout_controller_t * controller):
  controller_(controller)
{}

//----------------------------------------------------------------
// tile_match_transaction_t::execute
// 
void
tile_match_transaction_t::execute(the_thread_interface_t * thread)
{
  the_transaction_log_t log(thread);
  controller_->layout(log);
}
