// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : mosaic_refinement_controller.cxx
// Author       : Pavel A. Koshevoy
// Created      : Mon Mar 26 11:14:21 MDT 2007
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Wrapper class for automatic mosaic refinement.

// local includes:
#include "mosaic_refinement_controller.hxx"

// the includes:
#include <itk/common.hxx>
#include <itk/mosaic_refinement_common.hxx>
#include <itk/itkLegendrePolynomialTransform.h>
#include <utils/the_log.hxx>
#include <thread/the_thread_interface.hxx>


//----------------------------------------------------------------
// mosaic_refinement_controller_t::mosaic_refinement_controller_t
// 
mosaic_refinement_controller_t::
mosaic_refinement_controller_t(std::list<the_tile_t *> & tile_primitives,
			       const unsigned int & num_tiles):
  tile_(num_tiles),
  mask_(num_tiles),
  grid_(num_tiles),
  transform_(num_tiles)
{
  typedef itk::LegendrePolynomialTransform<double, 3> cubic_transform_t;
  
  // choose the optimal pyramid level at which the tiles will be matched:
  const the_tile_t * smallest_tile = NULL;
  unsigned int max_pyramid_size = UINT_MAX;
  for (std::list<the_tile_t *>::iterator iter = tile_primitives.begin();
       iter != tile_primitives.end(); ++iter)
  {
    const the_tile_t * t = *iter;
    const the_image_pyramid_t & pyramid = t->pyramid();
    if (pyramid.image_.size() < max_pyramid_size)
    {
      max_pyramid_size = pyramid.image_.size();
      smallest_tile = t;
    }
  }
  
  // once the minimum acceptable resolution level is found, try to
  // select the optimum resolution level -- 256x256 or better:
  unsigned int max_res_level = max_pyramid_size - 1;
  unsigned int min_res_level = max_pyramid_size - 1;
  {
    static const unsigned int min_edge = 256;
    static const unsigned int max_edge = UINT_MAX;
    
    const the_image_pyramid_t & pyramid = smallest_tile->pyramid();
    while (max_pyramid_size != 1)
    {
      const image_tile_generator_t & img =
	pyramid.image_[max_pyramid_size - 1];
      
      if (img.w_ / min_edge == 0 && img.h_ / min_edge == 0)
      {
	min_res_level--;
      }
      
      if (img.w_ / max_edge == 0 && img.h_ / max_edge == 0)
      {
	max_res_level--;
      }
      
      max_pyramid_size--;
    }
  }
  // cerr << max_res_level << " .. " << min_res_level << endl;
  
  resize(tile_pyramid_, min_res_level - max_res_level + 1, num_tiles);
  resize(mask_pyramid_, min_res_level - max_res_level + 1, num_tiles);
  unsigned int i = 0;
  for (std::list<the_tile_t *>::iterator iter = tile_primitives.begin();
       iter != tile_primitives.end(); ++iter, ++i)
  {
    the_tile_t * t = *iter;
    
    // the tile pyramid:
    the_image_pyramid_t & pyramid = t->pyramid();
    for (unsigned int j = max_res_level; j <= min_res_level; j++)
    {
      tile_pyramid_[min_res_level - j][i] = pyramid.itk_image_[j];
      mask_pyramid_[min_res_level - j][i] = pyramid.itk_mask_[j];
    }
    
    // the tile and its mask:
    tile_[i] = pyramid.itk_image_[0];
    mask_[i] = pyramid.itk_mask_[0];
    
    // the tile transform:
    the_tile_transform_t * s = first_dependent<the_tile_transform_t>(t);
    grid_[i] = s->transform_;
    
    itk::GridTransform::Pointer itk_gt = itk::GridTransform::New();
    itk_gt->setup(s->transform_);
    
#ifdef USE_GRID_REFINEMENT
    transform_[i] = itk_gt;
#else
    cubic_transform_t::Pointer legendre =
      setup_transform<cubic_transform_t, native_image_t>
      (tile_pyramid_[max_pyramid_size - 1][i].GetPointer());
    
    solve_for_transform<native_image_t, cubic_transform_t>
      (tile_pyramid_[max_pyramid_size - 1][i].GetPointer(),
       mask_[i].GetPointer(),
       itk_gt.GetPointer(),
       legendre,
       16,	// samples per edge
       1,	// landmark generator version
       true);	// refine
    transform_[i] = legendre;
#endif
  }
}

//----------------------------------------------------------------
// mosaic_refinement_controller_t::~mosaic_refinement_controller_t
// 
mosaic_refinement_controller_t::~mosaic_refinement_controller_t()
{}

//----------------------------------------------------------------
// mosaic_refinement_controller_t::refine
// 
void
mosaic_refinement_controller_t::refine(the_log_t & log)
{
#ifdef USE_GRID_REFINEMENT
  // use FFT to refine the grid transforms directly:
  typedef itk::GridTransform itk_grid_transform_t;
  
  unsigned int pyramid_levels = tile_pyramid_.size();
  if (pyramid_levels == 0) return;
  
  unsigned int num_tiles = tile_.size();
  if (num_tiles == 0) return;
  
  // select an appropriate pyramid level:
  unsigned int neighborhood = 96;
  int best_level = pyramid_levels - 1;
  for (; best_level > -1; best_level--)
  {
    native_image_t::Pointer & tile = tile_pyramid_[best_level][0];
    const the_grid_transform_t & gt = grid_[0];
    
    native_image_t::SizeType sz = tile->GetLargestPossibleRegion().GetSize();
    unsigned int dx = (sz[0] * 3) / gt.cols_;
    unsigned int dy = (sz[1] * 3) / gt.rows_;
    
    if (dx <= neighborhood && dy <= neighborhood) break;
  }
  
  if (best_level == -1)
  {
    best_level = 0;
    native_image_t::Pointer & tile = tile_pyramid_[best_level][0];
    const the_grid_transform_t & gt = grid_[0];
    
    native_image_t::SizeType sz = tile->GetLargestPossibleRegion().GetSize();
    unsigned int dx = (sz[0] * 3) / gt.cols_;
    unsigned int dy = (sz[1] * 3) / gt.rows_;
    
    neighborhood = std::max(dx, dy);
  }
  
  for (unsigned int i = 0; i < num_tiles; i++)
  {
    transform_[i] = itk_grid_transform_t::New();
    transform_[i]->setup(grid_[i]);
  }
  
  // FIXME:
  std::vector<native_image_t::Pointer> & tile_ = tile_pyramid_[best_level];
  std::vector<mask_t::Pointer> & mask_ = mask_pyramid_[best_level];
  
  refine_mosaic<native_image_t::Pointer, mask_t::Pointer>
    (log,
     transform_,
     tile_,
     mask_,
     neighborhood,
     true,	// prewarp tiles
     0.25,	// minimum overlap
     1,		// median filter radius
     1);	// number of refinement passes
  
#else // USE_GRID_REFINEMENT
  
  refine_mosaic<native_image_t, mask_t>(log,
					tile_pyramid_,
					mask_,
					transform_,
					50);
#endif // USE_GRID_REFINEMENT
}


//----------------------------------------------------------------
// mosaic_refinement_transaction_t::mosaic_refinement_transaction_t
// 
mosaic_refinement_transaction_t::
mosaic_refinement_transaction_t(mosaic_refinement_controller_t * controller):
  controller_(controller)
{}

//----------------------------------------------------------------
// mosaic_refinement_transaction_t::execute
// 
void
mosaic_refinement_transaction_t::execute(the_thread_interface_t * thread)
{
  the_transaction_log_t log(thread);
  controller_->refine(log);
}
