// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : the_main_window.cxx
// Author       : Pavel A. Koshevoy
// Created      : 2006/07/11 13:20
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : main application window widget

// system includes:
#include <vector>

// Qt includes:
#include <QWidget>
#include <QTabWidget>
#include <QApplication>
#include <QMessageBox>
#include <QInputDialog>
#include <QFileDialog>
#include <QFileInfo>
#include <QString>
#include <QAction>
#include <QListView>
#include <QPixmap>
#include <QGridLayout>
#include <QList>
#include <QDateTime>
#include <QImageReader>

// local includes:
#include "the_main_window.hxx"
#include "the_mosaic_proc_ui.hxx"
#include "the_mosaic_proc.hxx"
#include "the_tile_transaction.hxx"
#include "mosaic_layout_controller.hxx"
#include "mosaic_refinement_controller.hxx"
#include "ThreadPoolDialog.h"

// the includes:
#include <math/the_color.hxx>
#include <eh/the_view_mgr_eh.hxx>
#include <io/the_file_io.hxx>
#include <utils/the_indentation.hxx>
#include <utils/the_text.hxx>
#include <utils/the_exception.hxx>
#include <Qt/the_qt_trail.hxx>
#include <Qt/the_qt_view.hxx>
#include <Qt/the_qt_thread_pool.hxx>
#include <Qt/qimage_pixel_converter.hxx>
#include <image/la_pixel_converter.hxx>
#include <opengl/the_appearance.hxx>
#include <itk/grid_common.hxx>
#include <itk/fft_common.hxx>

// Cg includes:
#include <Cg/cgGL.h>

// forward declarations:
extern CGprogram CG_LA_FRAGMENT_PROGRAM;
extern CGprogram CG_RGBA16_FRAGMENT_PROGRAM;
extern CGprogram * CG_LA_FRAGMENT_PROGRAM_PTR;
extern CGprogram * CG_RGBA16_FRAGMENT_PROGRAM_PTR;
extern bool CG_LA_FRAGMENT_PROGRAM_OK;
extern bool CG_RGBA16_FRAGMENT_PROGRAM_OK;


//----------------------------------------------------------------
// TILE_THREAD_POOL
// 
the_qt_thread_pool_t *
TILE_THREAD_POOL()
{
  static the_qt_thread_pool_t * pool_ = NULL;
  if (pool_ == NULL)
  {
    pool_ = new the_qt_thread_pool_t(4);
  }
  
  return pool_;
}


//----------------------------------------------------------------
// get_image_filters
// 
QString
get_image_filters()
{
  QList<QByteArray> supported = QImageReader::supportedImageFormats();
  the_dynamic_array_t<QString> format;
  the_dynamic_array_t<QString> extension;
  for(QList<QByteArray>::const_iterator i = supported.begin();
      i != supported.end(); i++)
  {
    QString f(*i);
    format.append(f);
    extension.append(f.toLower());
  }
  
  QString filter = "Images (";
  for (unsigned int i = 0; i < format.size(); i++)
  {
    filter += QString("*.") + extension[i] + QString(" ");
  }
  filter += ")";
  return filter;
}


//----------------------------------------------------------------
// get_image_filters
// 
void
get_image_filters(QStringList & filters, bool decorated)
{
  QList<QByteArray> supported = QImageReader::supportedImageFormats();
  the_dynamic_array_t<QString> format;
  the_dynamic_array_t<QString> extension;
  for(QList<QByteArray>::const_iterator i = supported.begin();
      i != supported.end(); i++)
  {
    QString f(*i);
    format.append(f);
    extension.append(f.toLower());
  }
  
  for (unsigned int i = 0; i < format.size(); i++)
  {
    QString filter;
    if (decorated)
    {
      filter = format[i] + " images (*." + extension[i] + ")";
    }
    else
    {
      filter = QString("*.") + extension[i];
    }
    
    filters << filter;
  }
}

//----------------------------------------------------------------
// the_tile_thumbnail_t
// 
class the_tile_thumbnail_t : public QIconEngine
{
public:
  the_tile_thumbnail_t(boost::shared_ptr<the_image_pyramid_t> & pyramid):
    pyramid_(pyramid)
  {}
  
  // virtual:
  QSize actualSize(const QSize & size, QIcon::Mode, QIcon::State)
  { return size; }
  
  // virtual:
  void paint(QPainter * painter,
	     const QRect & rect,
	     QIcon::Mode /* mode */,
	     QIcon::State /* state */)
  {
    // shortcut:
    the_image_pyramid_t & pyramid = *(pyramid_.get());
    
    if (pyramid.is_ready())
    {
      QImage & icon = pyramid.thumbnail_;
      QSize size = rect.size();
      
      // generate an thumbnail image as necessary:
      if (icon.isNull() ||
	  std::max(size.width(), size.height()) !=
	  std::max(icon.size().width(), icon.size().height()))
      {
	/*
	cerr << "icon size: " << icon.size().width()
	     << " x " << icon.size().height() << endl
	     << "requested: " << size.width()
	     << " x " << size.height() << endl;
	*/
	
	// look through the pyramid and find the lowest resolution level
	// with resolution higher than the requested resolution:
	
	unsigned int num_levels = pyramid.image_.size();
	unsigned int next_level = 1;
	for (; next_level < num_levels; ++next_level)
	{
	  image_tile_generator_t & image = pyramid.image_[next_level];
	  if (int(image.w_) < size.width() ||
	      int(image.h_) < size.height()) break;
	}
	
	unsigned int best_level = next_level - 1;
	native_image_t::Pointer tile = pyramid.itk_image_[best_level];
	mask_t::Pointer mask = pyramid.itk_mask_[best_level];
	
	native_image_t::SizeType tile_sz =
	  tile->GetLargestPossibleRegion().GetSize();
	native_image_t::SpacingType tile_sp =
	  tile->GetSpacing();
	native_image_t::PointType tile_origin =
	  tile->GetOrigin();
	
	int nx = tile_sz[0];
	int ny = tile_sz[1];
	
	double wr = double(nx) / double(size.width());
	double hr = double(ny) / double(size.height());
	double ar = double(nx) / double(ny);
	
	// scale the image to fit the thumbnail:
	QSize sz(nx, ny);
	if (nx > size.width() || ny > size.height())
	{
	  if (wr > hr)
	  {
	    sz.setWidth(size.width());
	    sz.setHeight(std::max(1, int(double(size.width()) / ar)));
	  }
	  else
	  {
	    sz.setWidth(std::max(1, int(double(size.height()) * ar)));
	    sz.setHeight(size.height());
	  }
	}
	
	QImage qt_image(sz, QImage::Format_ARGB32);
	int icon_w = sz.width();
	int icon_h = sz.height();
	
	// setup the interpolators:
	typedef itk::LinearInterpolateImageFunction<native_image_t, double>
	  tile_interpolator_t;
	tile_interpolator_t::Pointer tile_interpolator =
	  tile_interpolator_t::New();
	tile_interpolator->SetInputImage(tile);
	
	typedef itk::LinearInterpolateImageFunction<mask_t, double>
	  mask_interpolator_t;
	mask_interpolator_t::Pointer mask_interpolator =
	  mask_interpolator_t::New();
	mask_interpolator->SetInputImage(mask);
	
	// resample the image into a QImage:
	native_image_t::PointType point;
	native_image_t::PointType tmp_pt;
	for (int y = 0; y < icon_h; y++)
	{
	  point[1] =
	    tile_origin[1] +
	    (double(y) / double(icon_h)) * tile_sp[1] * double(ny);
	  
	  for (int x = 0; x < icon_w; x++)
	  {
	    point[0] =
	      tile_origin[0] +
	      (double(x) / double(icon_w)) * tile_sp[0] * double(nx);
	    
	    double p = 0;
	    double a = 0;
	    double n = 0;
	    
	    for (unsigned int dy = 0; dy < 2; dy++)
	    {
	      tmp_pt[1] = point[1] + tile_sp[1] * double(dy);
	      for (unsigned int dx = 0; dx < 2; dx++)
	      {
		tmp_pt[0] = point[0] + tile_sp[0] * double(dx);
		
		if (tile_interpolator->IsInsideBuffer(tmp_pt))
		{
		  p += tile_interpolator->Evaluate(tmp_pt);
		  a += mask_interpolator->Evaluate(tmp_pt);
		  n += 1.0;
		}
	      }
	    }
	    
	    if (n != 0)
	    {
	      p = std::min(255.0, std::max(0.0, p / n));
	      a = std::min(255.0, std::max(0.0, a / n));
	    }
	    
	    QRgb pixel = qRgba(int(p), int(p), int(p), int(a));
	    qt_image.setPixel(x, y, pixel);
	  }
	}
	
	icon = qt_image;
      }
      
      // paint the thumbnail:
      int x = rect.x() + (rect.width() - icon.width()) / 2;
      int y = rect.y() + (rect.height() - icon.height()) / 2;
      painter->drawImage(QPoint(x, y), icon);
    }
    else
    {
      // paint the default thumbnail
      painter->fillRect(rect, QBrush(QColor(64, 128, 255, 255)));
    }
  }
  
  // virtual:
  QPixmap pixmap(const QSize & size, QIcon::Mode mode, QIcon::State state)
  {
    QPixmap pm(size);
    QPainter painter(&pm);
    paint(&painter, QRect(0, 0, size.width(), size.height()), mode, state);
    return pm;
  }
  
private:
  boost::shared_ptr<the_image_pyramid_t> pyramid_;
};


//----------------------------------------------------------------
// the_tiles_model_t::the_tiles_model_t
// 
the_tiles_model_t::the_tiles_model_t(the_document_so_t * doc_so,
				     QObject * parent):
  QAbstractListModel(parent),
  doc_so_(doc_so)
{}

//----------------------------------------------------------------
// the_tiles_model_t::rowCount
// 
int
the_tiles_model_t::rowCount(const QModelIndex & /* parent */) const
{
  the_mosaic_proc_t * p = proc();
  if (p == NULL) return 0;
  
  std::list<the_tile_t *> tiles;
  unsigned int num_tiles = direct_dependents<the_tile_t>(p, tiles);
  return num_tiles;
}

//----------------------------------------------------------------
// the_tiles_model_t::data
// 
QVariant
the_tiles_model_t::data(const QModelIndex & index, int role) const
{
  the_tile_t * t = tile(index);
  if (t != NULL)
  {
    const char * fn = t->pyramid().filename_.text();
    const QFileInfo info(fn);
    
    if (role == Qt::DisplayRole)
    {
      return info.fileName();
    }
    else if (role == Qt::DecorationRole)
    {
      return QIcon(new the_tile_thumbnail_t(t->pyramid_));
    }
  }
  
  return QVariant();
}

//----------------------------------------------------------------
// the_tiles_model_t::index
// 
QModelIndex
the_tiles_model_t::index(const QString & tile_absolute_file_path) const
{
  the_mosaic_proc_t * p = proc();
  if (p == NULL) return QModelIndex();
  
  the_text_t filename(qPrintable(tile_absolute_file_path));
  
  std::list<the_tile_t *> tiles;
  direct_dependents<the_tile_t>(p, tiles);
  
  unsigned int row = 0;
  for (std::list<the_tile_t *>::const_iterator i = tiles.begin();
       i != tiles.end(); ++i, ++row)
  {
    const the_tile_t * tile = *i;
    if (tile->pyramid().filename_.match_tail(filename))
    {
      return QAbstractItemModel::createIndex(row, 0);
    }
  }
  
  return QModelIndex();
}

//----------------------------------------------------------------
// the_tiles_model_t::proc
// 
the_mosaic_proc_t *
the_tiles_model_t::proc() const
{
  return doc_so_->document()->active_procedure<the_mosaic_proc_t>();
}

//----------------------------------------------------------------
// the_tiles_model_t::tile
// 
the_tile_t *
the_tiles_model_t::tile(const QModelIndex & index) const
{
  the_mosaic_proc_t * p = proc();
  if (p != NULL)
  {
    unsigned int row = index.row();
    return nth_dependent<the_tile_t>(p, row);
  }
  
  return NULL;
}

//----------------------------------------------------------------
// the_tiles_model_t::trigger_update
// 
void
the_tiles_model_t::trigger_update(const QModelIndex & index)
{
  the_mosaic_proc_t * p = proc();
  if (p == NULL) return;
  
  unsigned int row = index.row();
  the_tile_t * tile = nth_dependent<the_tile_t>(p, row);
  
  if (tile != NULL)
  {
    // cerr << "trigger update: " << tiles[index.row()] << endl;
    emit dataChanged(index, index);
  }
}

//----------------------------------------------------------------
// the_tiles_model_t::trigger_reset
// 
void
the_tiles_model_t::trigger_reset()
{
  reset();
}


//----------------------------------------------------------------
// orientation
// 
static the_view_mgr_orientation_t orientation[] =
{
  THE_BOTTOM_VIEW_E,
};


//----------------------------------------------------------------
// the_main_window_t::the_main_window_t
// 
the_main_window_t::the_main_window_t():
  QMainWindow(NULL, 0),
  the_document_ui_t("ir-mosaic/20070214"),
  tiles_model_(NULL),
  mosaic_proc_ui_(NULL),
  preferences_ui_(this)
{
  the_view_t * the_views[] = {
    new the_qt_view_t(this, "the mosaic", NULL, orientation[0]),
  };
  setup(NULL, the_views, 1);
  
  // shortcut:
  the_qt_view_t * qt_views[] = {
    view<the_qt_view_t>(0),
  };
  
  qt_views[0]->setObjectName("mosaic");
  
  // disable 3D rotation:
  view_mgr_eh_[0]->allow_spin(false);
  
  setupUi(this);
  
  // make certain shortcuts global to the entire application:
  actionOpen->setShortcutContext(Qt::ApplicationShortcut);
  actionSave->setShortcutContext(Qt::ApplicationShortcut);
  actionQuit->setShortcutContext(Qt::ApplicationShortcut);
  
  actionUndo->setShortcutContext(Qt::ApplicationShortcut);
  actionRedo->setShortcutContext(Qt::ApplicationShortcut);
  
  actionRestore->setShortcutContext(Qt::ApplicationShortcut);
  
  actionAdd->setShortcutContext(Qt::ApplicationShortcut);
  actionRemove->setShortcutContext(Qt::ApplicationShortcut);
  actionClear->setShortcutContext(Qt::ApplicationShortcut);
  
  actionLayout->setShortcutContext(Qt::ApplicationShortcut);
  actionRefine->setShortcutContext(Qt::ApplicationShortcut);
  
  actionAbout->setShortcutContext(Qt::ApplicationShortcut);
  
  // connect signals and slots:
  connect(actionOpen, SIGNAL(triggered()), this, SLOT(file_open()));
  connect(actionSave, SIGNAL(triggered()), this, SLOT(file_save()));
  connect(actionQuit, SIGNAL(triggered()), this, SLOT(file_exit()));
  
  connect(actionUndo, SIGNAL(triggered()), this, SLOT(edit_undo()));
  connect(actionRedo, SIGNAL(triggered()), this, SLOT(edit_redo()));
  connect(actionUnwarp, SIGNAL(triggered()), this, SLOT(edit_unwarp()));
  connect(actionPreferences, SIGNAL(triggered()), this, SLOT(edit_prefs()));
  
  connect(actionRestore, SIGNAL(triggered()), this, SLOT(view_restore()));
  
  connect(actionAdd, SIGNAL(triggered()), this, SLOT(tiles_add()));
  connect(actionRemove, SIGNAL(triggered()), this, SLOT(tiles_remove()));
  connect(actionClear, SIGNAL(triggered()), this, SLOT(tiles_clear()));
  
  connect(actionLayout, SIGNAL(triggered()), this, SLOT(mosaic_layout()));
  connect(actionRefine, SIGNAL(triggered()), this, SLOT(mosaic_refine()));
  
  connect(actionAbout, SIGNAL(triggered()), this, SLOT(help_about()));
  
  // start with a new document:
  QString time_stamp =
    QDateTime::currentDateTime().toString("yyyyMMdd-hhmm");
  doc_so().new_document(new the_document_t(qPrintable(time_stamp +
						      ".mosaic")));
  
  // UI objects of various procedures:
  mosaic_proc_ui_ = new the_mosaic_proc_ui_t(this);
  
  // mosaic layout view:
  {
    QGridLayout * layout = new QGridLayout(layoutTab);
    layout->setSpacing(0);
    layout->setMargin(0);
    qt_views[0]->setParent(layoutTab);
    layout->addWidget(qt_views[0], 0, 0);
    qt_views[0]->show();
  }
  
  mosaic_proc_ui_->install();
  
  tiles_model_ = new the_tiles_model_t(doc_so_, this);
  listView->setModel(tiles_model_);
  
  connect(TILE_THREAD_POOL(),
	  SIGNAL(transaction_finished(the_transaction_t *)),
	  this, SLOT(tile_ready(the_transaction_t *)));
  
  connect(listView->selectionModel(),
	  SIGNAL(currentChanged(const QModelIndex &, const QModelIndex &)),
	  this,
	  SLOT(selection_changed(const QModelIndex &, const QModelIndex &)));
}

//----------------------------------------------------------------
// the_main_window_t::~the_main_window_t
// 
the_main_window_t::~the_main_window_t()
{
  assert(mosaic_proc_ui_ == NULL);
}

//----------------------------------------------------------------
// import_transform
// 
static bool
update_transform(the_tile_t * tile,
		 const base_transform_t * t,
		 const unsigned int & rows = 15,
		 const unsigned int & cols = 15,
		 const unsigned int & max_iterations = 100)
{
  base_transform_t::ConstPointer transform = t;
  if (t == NULL)
  {
    transform = identity_transform_t::New();
  }
  
  // shortcut:
  the_tile_transform_t * s = first_dependent<the_tile_transform_t>(tile);
  assert(s != NULL);
  if (s == NULL) return false;
  
  // resample the imported transform onto a grid transform:
  const the_image_pyramid_t & pyramid = tile->pyramid();
  pnt2d_t tile_min = pnt2d(pyramid.image_[0].bbox_.min_[0],
			   pyramid.image_[0].bbox_.min_[1]);
  pnt2d_t tile_max = pnt2d(pyramid.image_[0].bbox_.max_[0],
			   pyramid.image_[0].bbox_.max_[1]);
  
  bool ok = setup_grid_transform(s->transform_,
				 rows,
				 cols,
				 tile_min,
				 tile_max,
				 pyramid.itk_mask_[0].GetPointer(),
				 transform.GetPointer(),
				 max_iterations,
				 1e-12,
				 1e-16,
				 5);
  assert(ok);
  
#if 0
  if (!ok)
  {
    // FIXME: this will die for sure:
    char * blah = NULL;
    blah[0] = 0;
  }
  
  cerr << " - - - - - - - - - - - - - - - - - - - - - - - - - -" << endl;
  for (unsigned int i = 0; i < s->transform_.mesh_.size(); i++)
  {
    vertex_t & vx = s->transform_.mesh_[i];
    cerr << vx.uv_[0] << ", " << vx.uv_[1] << '\t'
	 << vx.xy_[0] << ", " << vx.xy_[1] << endl;
  }
  cerr << endl;
#endif
  
  return ok;
}

//----------------------------------------------------------------
// the_main_window_t::tile_ready
// 
void
the_main_window_t::tile_ready(the_transaction_t * transaction)
{
  the_tile_transaction_t * t =
    dynamic_cast<the_tile_transaction_t *>(transaction);
  
  if (t != NULL && t->done())
  {
    the_mosaic_proc_t * proc =
      document()->active_procedure<the_mosaic_proc_t>();
    the_registry_t * r = proc->registry();
    the_tile_t * tile = r->elem<the_tile_t>(t->tile_id_);
    
    // shortcut:
    the_image_pyramid_t & pyramid = *(tile->pyramid_.get());
    
    if (pyramid.state_ == the_image_pyramid_t::READY_E)
    {
      const char * fn = pyramid.filename_.text();
      const QFileInfo info(fn);
      const QString path(info.absoluteFilePath());
      tiles_model_->trigger_update(tiles_model_->index(path));
      listView->repaint();
      
      the_tile_transform_t * s = first_dependent<the_tile_transform_t>(tile);
      if (!s->transform_.is_ready())
      {
	update_transform(tile, tile->imported_.GetPointer());
	tile->imported_ = NULL;
      }
    }
    else if (pyramid.state_ == the_image_pyramid_t::FAILED_E)
    {
      proc->tiles_remove(tile->id());
      listView->repaint();
    }
    
    proc->request_regeneration();
    sync_ui();
    view(0)->restore_orientation();
  }
  
  delete transaction;
}

//----------------------------------------------------------------
// the_main_window_t::selection_changed
// 
void
the_main_window_t::selection_changed(const QModelIndex & ix_new,
				     const QModelIndex & /* ix_old */)
{
  the_tile_t * tile = tiles_model_->tile(ix_new);
  if (tile == NULL) return;
  
  tiles_model_->proc()->select(tile->id());
  // cerr << "selection changed: " << tile->pyramid_->filename_ << endl;
  sync_ui();
}

//----------------------------------------------------------------
// ok_to_discard_current_document
// 
static bool
ok_to_discard_current_document(the_main_window_t * main_window,
			       the_document_so_t & doc_so)
{
  if (doc_so.changes_saved()) return true;
  
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  QMessageBox mb(main_window);
  mb.setIcon(QMessageBox::Question);	
  mb.setWindowTitle(QString("Save changes?"));
  
  std::ostringstream os;
  os << "The document \"" << doc_so.document()->name()
     <<"\" has unsaved changes." << endl;
  mb.setText(QString(os.str().c_str()));
  
  os.str("");
  os << "Click [Save] to save the changes and proceed." << endl
     << "Click [Discard] to discard the changes and proceed." << endl
     << "Click [Cancel] to keep the changes and continue working." << endl;
  mb.setDetailedText(QString(os.str().c_str()));
  
  QAbstractButton * btn_save =
    mb.addButton(QObject::tr("Save"), QMessageBox::ActionRole);
  
  QAbstractButton * btn_discard =
    mb.addButton(QObject::tr("Discard"), QMessageBox::ActionRole);
  
  // QAbstractButton * btn_cancel =
  mb.addButton(QObject::tr("Cancel"), QMessageBox::ActionRole);
  
  mb.exec();
  if (mb.clickedButton() == btn_save)
  {
    main_window->file_save();
    if (doc_so.changes_saved()) return true;
    
    QMessageBox::critical(main_window,
			  QString("Could not save changes"),
			  QString("Could not save the document, aborting..."),
			  QMessageBox::Ok,
			  QMessageBox::NoButton,
			  QMessageBox::NoButton);
    return false;
  }
  
  return (mb.clickedButton() == btn_discard);
}

//----------------------------------------------------------------
// the_main_window_t::file_open
// 
void
the_main_window_t::file_open()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  QString filename
    (QFileDialog::getOpenFileName(this,
				  "Open a mosaic",
				  QString(),
				  "mosaic files (*.mosaic)"));
  file_open(filename);
}

//----------------------------------------------------------------
// the_main_window_t::file_open
// 
void
the_main_window_t::file_open(const QString & filename)
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  std::fstream fin;
  fin.open(qPrintable(filename), ios::in);
  if (!fin.is_open()) return;
  
  double pixel_spacing = 1.0;
  bool use_std_mask = false;
  std::list<the_text_t> fn_tiles;
  std::vector<base_transform_t::Pointer> transform;
  
  ::load_mosaic<base_transform_t>(fin,
				  pixel_spacing,
				  use_std_mask,
				  fn_tiles,
				  transform);
  fin.close();
  
  unsigned int num_images = transform.size();
  if (num_images == 0) return;
  if (pixel_spacing == 0.0) return;
  
  QStringList tmp;
  for (std::list<the_text_t>::const_iterator iter = fn_tiles.begin();
       iter != fn_tiles.end(); ++iter)
  {
    const the_text_t & fn_image = *iter;
    QFileInfo file_info(fn_image.text());
    if (!file_info.isFile()) continue;
    
    tmp.push_back(file_info.absoluteFilePath());
  }
  
  if (tmp.empty()) return;
  
  // update the preferences:
  prefsSetUseStdMask(use_std_mask);
  
  // FIXME: what about pixel spacing?
  tiles_clear();
  tiles_add(tmp, 1, pixel_spacing, pixel_spacing);
  doc_so().big_changes();
  doc_so().changes_saved(true);
  
  std::list<the_tile_t *> tiles;
  direct_dependents<the_tile_t>(tiles_model_->proc(), tiles);
  unsigned int index = 0;
  for (std::list<the_tile_t *>::iterator iter = tiles.begin();
       iter != tiles.end(); ++iter, ++index)
  {
    the_tile_t * t = *iter;
    t->imported_ = transform[index];
  }
  
  // change the document name:
  QFileInfo file_info(filename);
  document()->name().assign(qPrintable(file_info.fileName()));
}

//----------------------------------------------------------------
// the_main_window_t::file_save
// 
bool
the_main_window_t::file_save()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  std::list<the_tile_t *> tiles;
  unsigned int num_tiles =
    direct_dependents<the_tile_t>(tiles_model_->proc(), tiles);
  if (num_tiles == 0) return false;
  
  // select a filename:
  QString suggest_fn = QString(document()->name().text());
  QString filename
    (QFileDialog::getSaveFileName(this,
				  "Save this mosaic",
				  suggest_fn,
				  "mosaic files (*.mosaic)"));
  if (filename.isEmpty()) return false;
  doc_so().set_filename(filename.toUtf8().data());
  
  // save the mosaic:
  double pixel_spacing = tiles.front()->pyramid().image_[0].spacing_x_;
  bool use_std_mask = prefsUseStdMask();
  
  std::list<the_text_t> fn_tiles;
  std::vector<base_transform_t::Pointer> transform(num_tiles);
  
  unsigned int index = 0;
  for (std::list<the_tile_t *>::iterator iter = tiles.begin();
       iter != tiles.end(); ++iter, ++index)
  {
    the_tile_t * t = *iter;
    the_tile_transform_t * s = first_dependent<the_tile_transform_t>(t);
    
    fn_tiles.push_back(t->pyramid().filename_);
    itk::GridTransform::Pointer itk_gt = itk::GridTransform::New();
    itk_gt->setup(s->transform_);
    transform[index] = itk_gt;
  }
  
  std::fstream sout;
  sout.open(qPrintable(filename), std::ios::out);
  if (!sout.is_open()) return false;
  
  save_mosaic<base_transform_t>
    (sout, pixel_spacing, use_std_mask, fn_tiles, transform);
  sout.close();
  
  doc_so().changes_saved(true);
  return true;
}

//----------------------------------------------------------------
// the_main_window_t::file_exit
// 
void
the_main_window_t::file_exit()
{
  // check whether the existing document needs to be saved:
  if (ok_to_discard_current_document(this, doc_so()) == false) return;
  
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  // stop trail playback:
  THE_TRAIL.stop();
  
  // shut down the procedure UI:
  mosaic_proc_ui_->uninstall();
  delete mosaic_proc_ui_;
  mosaic_proc_ui_ = NULL;
  
  // close the document:
  doc_so().close_document();
  
  // shut dowm the document UI:
  shutdown();
  
  // done:
  qApp->quit();
}

//----------------------------------------------------------------
// the_main_window_t::edit_undo
// 
void
the_main_window_t::edit_undo()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  doc_so().undo();
  document()->active_procedure<the_mosaic_proc_t>()->request_regeneration();
  sync_ui();
}

//----------------------------------------------------------------
// the_main_window_t::edit_redo
// 
void
the_main_window_t::edit_redo()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  doc_so().redo();
  document()->active_procedure<the_mosaic_proc_t>()->request_regeneration();
  sync_ui();
}

//----------------------------------------------------------------
// the_main_window_t::edit_unwarp
// 
void
the_main_window_t::edit_unwarp()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  the_mosaic_proc_t * proc = tiles_model_->proc();
  if (!proc->has_selection()) return;
  
  the_registry_t * r = proc->registry();
  the_tile_t * t = r->elem<the_tile_t>(proc->selection());
  the_tile_transform_t * s = ::first_dependent<the_tile_transform_t>(t);
  
  // compute the current center of mass:
  the_grid_transform_t & gt = s->transform_;
  unsigned int n_vx = gt.mesh_.size();
  if (n_vx == 0) return;
  
  // save an undo record:
  doc_so().save_undo_record();
  
  pnt2d_t cm = pnt2d(0, 0);
  for (unsigned int i = 0; i < n_vx; i++)
  {
    const pnt2d_t & p = gt.mesh_[i].xy_;
    cm[0] += p[0];
    cm[1] += p[1];
  }
  
  cm[0] /= double(n_vx);
  cm[1] /= double(n_vx);
  
  pnt2d_t center = gt.tile_min_ + gt.tile_ext_ * 0.5;
  vec2d_t shift = cm - center;
  
  for (unsigned int i = 0; i < n_vx; i++)
  {
    vertex_t & vx = gt.mesh_[i];
    vx.xy_[0] = gt.tile_min_[0] + gt.tile_ext_[0] * vx.uv_[0] + shift[0];
    vx.xy_[1] = gt.tile_min_[1] + gt.tile_ext_[1] * vx.uv_[1] + shift[1];
  }
  gt.rebuild_grid();
  s->request_regeneration();
  sync_ui();
}

//----------------------------------------------------------------
// the_main_window_t::edit_prefs
// 
void
the_main_window_t::edit_prefs()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;

  the_mosaic_proc_t * proc = tiles_model_->proc();
  bool prev_gpu_pref = prefsUseFragmentShader();
  bool prev_msk_pref = prefsUseStdMask();
  
  int r = preferences_ui_.exec();
  if (r == QDialog::Accepted)
  {
    bool curr_gpu_pref = prefsUseFragmentShader();
    bool curr_msk_pref = prefsUseStdMask();
    
    if (curr_gpu_pref != prev_gpu_pref)
    {
      if (!curr_gpu_pref)
      {
	CG_LA_FRAGMENT_PROGRAM_PTR = NULL;
	CG_RGBA16_FRAGMENT_PROGRAM_PTR = NULL;
      }
      else
      {
	CG_LA_FRAGMENT_PROGRAM_PTR =
	  CG_LA_FRAGMENT_PROGRAM_OK ?
	  &CG_LA_FRAGMENT_PROGRAM :
	  NULL;
	
	CG_RGBA16_FRAGMENT_PROGRAM_PTR =
	  CG_RGBA16_FRAGMENT_PROGRAM_OK ?
	  &CG_RGBA16_FRAGMENT_PROGRAM :
	  NULL;
      }
      
      // reload the fixed/moving images in correct format:
      proc->apply_prefs();
    }
    else if (prev_msk_pref != curr_msk_pref)
    {
      proc->apply_prefs();
    }
    
    document()->active_procedure()->request_regeneration();
    sync_ui();
  }
}

//----------------------------------------------------------------
// the_main_window_t::view_restore
// 
void
the_main_window_t::view_restore()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  std::vector<the_view_t *> views;
  get_visible_views(views);
  unsigned int num_views = views.size();
  for (unsigned int i = 0; i < num_views; i++)
  {
    views[i]->restore_orientation();
  }
}

//----------------------------------------------------------------
// the_main_window_t::tiles_add
// 
void
the_main_window_t::tiles_add()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  QString filters = get_image_filters();
  QStringList filenames(QFileDialog::getOpenFileNames(this,
						      QString(),
						      QString(),
						      filters));
  
  tiles_add(filenames, 1, 1, 1);
}

//----------------------------------------------------------------
// the_main_window_t::tiles_add
// 
void
the_main_window_t::tiles_add(const QStringList & filenames,
			     unsigned int shrink_factor,
			     double spacing_x,
			     double spacing_y)
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  if (filenames.empty()) return;
  
  doc_so().save_undo_record();
  
  for (QStringList::const_iterator i = filenames.begin();
       i != filenames.end(); ++i)
  {
    const QString & filename = *i;
    the_tile_transaction_t * loader = tile_load_transaction(filename,
							    shrink_factor,
							    spacing_x,
							    spacing_y);
    TILE_THREAD_POOL()->push_back(loader);
  }
  TILE_THREAD_POOL()->start();
  
  sync_ui();
}

//----------------------------------------------------------------
// the_main_window_t::tiles_remove
//
void
the_main_window_t::tiles_remove()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  QItemSelectionModel * sel_model = listView->selectionModel();
  QModelIndex index = sel_model->currentIndex();
  the_tile_t * tile = tiles_model_->tile(index);
  if (tile == NULL) return;
  
  doc_so().save_undo_record();
  tiles_model_->proc()->tiles_remove(tile->id());
  sync_ui();
}

//----------------------------------------------------------------
// the_main_window_t::tiles_clear
// 
void
the_main_window_t::tiles_clear()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  the_mosaic_proc_t * proc =
    doc_so_->document()->active_procedure<the_mosaic_proc_t>();
  if (first_dependent<the_tile_t>(proc) == NULL) return;
  
  doc_so().save_undo_record();
  proc->tiles_clear();
  sync_ui();
}

//----------------------------------------------------------------
// the_main_window_t::mosaic_layout
// 
void
the_main_window_t::mosaic_layout()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  the_mosaic_proc_t * proc =
    document()->active_procedure<the_mosaic_proc_t>();
  
  std::list<the_tile_t *> tile_primitives;
  unsigned int num_tiles =
    direct_dependents<the_tile_t>(proc, tile_primitives);
  if (num_tiles < 2) return;
  
  // FIXME: finish this:
  mosaic_layout_controller_t controller(tile_primitives, num_tiles);
  tile_match_transaction_t transaction(&controller);
  
  ThreadPoolDialog tmp(1, this);
  tmp.pool_.push_back(&transaction);
  if (tmp.exec() != QDialog::Accepted) return;
  
  // apply the new transforms:
  doc_so().save_undo_record();

  unsigned int i = 0;
  for (std::list<the_tile_t *>::iterator iter = tile_primitives.begin();
       iter != tile_primitives.end(); ++iter, ++i)
  {
    the_tile_t * tile = *iter;
    update_transform(tile, controller.affine_[i].GetPointer());
  }
  
  proc->request_regeneration();
  sync_ui();
  view(0)->restore_orientation();
}

//----------------------------------------------------------------
// the_main_window_t::mosaic_refine
// 
void
the_main_window_t::mosaic_refine()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  the_mosaic_proc_t * proc =
    document()->active_procedure<the_mosaic_proc_t>();
  
  std::list<the_tile_t *> tile_primitives;
  unsigned int num_tiles =
    direct_dependents<the_tile_t>(proc, tile_primitives);
  if (num_tiles < 2) return;
  
  // FIXME: finish this:
  mosaic_refinement_controller_t controller(tile_primitives, num_tiles);
  mosaic_refinement_transaction_t transaction(&controller);
  
  ThreadPoolDialog tmp(1, this);
  tmp.pool_.push_back(&transaction);
#if 0
  if (tmp.exec() != QDialog::Accepted) return;
#else
  tmp.exec();
#endif
  
  // apply the new transforms:
  doc_so().save_undo_record();

  unsigned int i = 0;
  for (std::list<the_tile_t *>::iterator iter = tile_primitives.begin();
       iter != tile_primitives.end(); ++iter, ++i)
  {
    the_tile_t * tile = *iter;
    update_transform(tile, controller.transform_[i].GetPointer());
  }
  
  proc->request_regeneration();
  sync_ui();
}

//----------------------------------------------------------------
// the_main_window_t::help_about
// 
void
the_main_window_t::help_about()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  AboutDialog about(this);
  about.exec();
}

//----------------------------------------------------------------
// the_main_window_t::tile_load_transaction
// 
the_tile_transaction_t *
the_main_window_t::tile_load_transaction(const QString & filename,
					 unsigned int shrink_factor,
					 double spacing_x,
					 double spacing_y)
{
  the_mosaic_proc_t * proc =
    doc_so_->document()->active_procedure<the_mosaic_proc_t>();
  
  unsigned int max_texture = OpenGL(0).max_texture_;
  bool use_luminance_textures = prefsUseLuminanceTextures();
  bool use_fragment_shader = prefsUseFragmentShader();
  bool use_std_mask = prefsUseStdMask();
  
  QFileInfo file_info(filename);
  const QString path(file_info.absoluteFilePath());
  
  unsigned int tile_id = proc->tiles_add(the_text_t(qPrintable(path)),
					 shrink_factor,
					 spacing_x,
					 spacing_y,
					 use_std_mask);
  
  the_tile_t * tile = proc->registry()->elem<the_tile_t>(tile_id);
  the_tile_transaction_t * loader =
    new the_tile_transaction_t(tile->id(),
			       tile->pyramid_,
			       shrink_factor,
			       max_texture,
			       use_fragment_shader,
			       use_std_mask,
			       use_luminance_textures);
  return loader;
}

//----------------------------------------------------------------
// the_main_window_t::sync_ui
// 
void
the_main_window_t::sync_ui()
{
  static int recursion_depth = 0;
  the_scoped_variable_t<int> scope(recursion_depth,
				   recursion_depth + 1,
				   recursion_depth);
  if (recursion_depth > 1) return;
  
  // shortcut:
  const the_document_t * doc = document();
  bool has_doc = (doc != NULL);
  
  // shortcut:
  the_procedure_ui_t * proc_ui = active_proc_ui();
  if (proc_ui != NULL) proc_ui->sync_ui();
  
  // update the views:
  for (unsigned int i = 0; i < view_.size(); i++)
  {
    the_qt_view_t * v = view<the_qt_view_t>(i);
    if (v->isHidden()) continue;
    v->refresh();
  }
  
  // update the actions:
  actionSave->setEnabled(has_doc);

  bool undo_ok = has_doc && !doc_so().undo_stack_empty();
  actionUndo->setEnabled(undo_ok);

  bool redo_ok = has_doc && !doc_so().redo_stack_empty();
  actionRedo->setEnabled(redo_ok);
  
  the_mosaic_proc_t * p = tiles_model_->proc();
  if (p->has_selection())
  {
    the_registry_t * r = p->registry();
    the_tile_t * t = r->elem<the_tile_t>(p->selection());
    if (t != NULL)
    {
      QItemSelectionModel * sel_model = listView->selectionModel();
      const char * fn = t->pyramid().filename_.text();
      const QFileInfo info(fn);
      const QString path(info.absoluteFilePath());
      QModelIndex index = tiles_model_->index(path);
      QModelIndex curr_index = sel_model->currentIndex();
      if (index != curr_index)
      {
	sel_model->clear();
	sel_model->setCurrentIndex(index, QItemSelectionModel::Select);
      }
    }
  }
  
  actionUnwarp->setEnabled(p->has_selection());
  listView->reset();
  
  actionSave->setEnabled(p->regenerated());
  actionLayout->setEnabled(p->regenerated());
  actionRefine->setEnabled(p->regenerated());
  
  // FIXME:
#if 0
  if (!view(0)->is_hidden())
  {
    static unsigned int frame = 0;
    the_qt_view_t * v = view<the_qt_view_t>(0);
    QImage img = v->grabFrameBuffer();
    static const the_text_t fn_prefix("/scratch/crcns/frames/ir-mosaic/");
    the_text_t fn = fn_prefix + the_text_t::number(frame, 6, '0') + ".png";
    img.save(QString(fn.text()));
    frame++;
  }
#endif
}

//----------------------------------------------------------------
// the_main_window_t::kill_ui
// 
void
the_main_window_t::kill_ui()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  file_exit();
}

//----------------------------------------------------------------
// the_main_window_t::get_visible_views
// 
void
the_main_window_t::get_visible_views(std::vector<the_view_t *> & views) const
{
  QWidget * active_tab = tabWidget->currentWidget();
  if (active_tab == layoutTab)
  {
    views.resize(1);
    views[0] = view_[0];
  }
  else
  {
    views.resize(0);
  }
}

//----------------------------------------------------------------
// the_main_window_t::closeEvent
// 
void
the_main_window_t::closeEvent(QCloseEvent * e)
{
  e->ignore();
  kill_ui();
}

//----------------------------------------------------------------
// the_main_window_t::dragEnterEvent
// 
void
the_main_window_t::dragEnterEvent(QDragEnterEvent * e)
{
  const QMimeData * mime = e->mimeData();
  if (mime->hasUrls() && (e->possibleActions() & Qt::MoveAction))
  {
    const QList<QUrl> & urls = mime->urls();
    for (QList<QUrl>::const_iterator i = urls.begin(); i != urls.end(); ++i)
    {
      const QUrl & url = *i;
      QFileInfo info(url.toLocalFile());
      if (info.isFile())
      {
	e->acceptProposedAction();
      }
    }
  }
}

//----------------------------------------------------------------
// the_main_window_t::dropEvent
// 
void
the_main_window_t::dropEvent(QDropEvent * e)
{
  QString mosaic_file;
  QStringList filenames;
  
  const QMimeData * mime = e->mimeData();
  if (mime->hasUrls() && (e->possibleActions() & Qt::MoveAction))
  {
    const QList<QUrl> & urls = mime->urls();
    for (QList<QUrl>::const_iterator i = urls.begin(); i != urls.end(); ++i)
    {
      const QUrl & url = *i;
      QFileInfo info(url.toLocalFile());
      if (info.isFile())
      {
	if (info.fileName().endsWith(".mosaic", Qt::CaseInsensitive))
	{
	  mosaic_file = info.absoluteFilePath();
	}
	else
	{
	  filenames.push_back(info.absoluteFilePath());
	}
	e->acceptProposedAction();
      }
    }
  }
  
  if (!filenames.empty())
  {
    tiles_add(filenames, 1, 1, 1);
  }
  else if (mosaic_file.size() != 0)
  {
    file_open(mosaic_file);
  }
}
