// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : the_mosaic_proc.cxx
// Author       : Pavel A. Koshevoy
// Created      : Wed Feb 14 12:11:00 MST 2007
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : The document procedure containing the mosaic layout.

// local includes:
#include "the_mosaic_proc.hxx"
#include "the_tile_transaction.hxx"
#include "PreferencesDialog.h"
#include "mosaic_cache.hxx"

// the includes:
#include <ui/the_document_ui.hxx>
#include <eh/the_input_device_eh.hxx>
#include <doc/the_graph.hxx>
#include <sel/the_pick_rec.hxx>
#include <math/the_plane.hxx>
#include <math/the_view_volume.hxx>
#include <math/the_ray.hxx>
#include <math/the_coord_sys.hxx>
#include <math/the_color.hxx>
#include <opengl/the_view.hxx>
#include <opengl/image_tile_dl_elem.hxx>
#include <utils/the_indentation.hxx>
#include <Qt/the_qt_thread_pool.hxx>

// Cg includes:
#include <Cg/cgGL.h>

// forward declarations:
extern the_qt_thread_pool_t * TILE_THREAD_POOL();

#ifdef USE_CG
extern CGprogram * CG_LA_FRAGMENT_PROGRAM_PTR;
extern CGcontext cg_context;
extern CGprofile cg_profile;
#endif


//----------------------------------------------------------------
// the_mosaic_proc_t::the_mosaic_proc_t
// 
the_mosaic_proc_t::the_mosaic_proc_t():
  the_procedure_t(),
  dl_(new the_disp_list_t()),
  layout_(new mosaic_cache_pyramid_t()),
  layout_dl_elem_(new std::vector<boost::shared_ptr<mosaic_dl_elem_t> >()),
  layout_dl_ready_(~0),
  layout_id_(std::numeric_limits<unsigned int>::max()),
  selection_(std::numeric_limits<unsigned int>::max())
{}

//----------------------------------------------------------------
// the_mosaic_proc_t::~the_mosaic_proc_t
// 
the_mosaic_proc_t::~the_mosaic_proc_t()
{}

//----------------------------------------------------------------
// the_mosaic_proc_t::added_to_the_registry
// 
void
the_mosaic_proc_t::added_to_the_registry(the_registry_t * registry,
					 const unsigned int & id)
{
  the_procedure_t::added_to_the_registry(registry, id);
  
  the_tile_mosaic_t * layout = new the_tile_mosaic_t();
  registry->add(layout);
  layout_id_ = layout->id();
  establish_supporter_dependent(registry, id, layout_id_);
  
  the_tile_warp_widget_t * widget = new the_tile_warp_widget_t();
  registry->add(widget);
  establish_supporter_dependent(registry, id, widget->id());
}

//----------------------------------------------------------------
// the_mosaic_proc_t::regenerate
// 
bool
the_mosaic_proc_t::regenerate()
{
  the_document_ui_t * doc_ui = the_document_ui_t::doc_ui();
  
  if (layout_->cache_.size() == 0)
  {
    setup_pyramids();
  }
  // clear the view dependent primitive lists:
  the_view_t * view_layout = doc_ui->view(0);
  view_layout->gl_make_current();
  view_layout->clear_primitives();
  dl_->clear(); // FIXME: this is inefficient
  
  // regenerate the dependents that need to be regenerated:
  the_registry_t * r = registry();
  the_graph_t graph;
  graph.set_roots(r, direct_dependents());
  regenerated_ = graph.regenerate();
  
  // update the view dependent selectable primitive lists:
  for (std::list<unsigned int>::const_iterator
	 i = direct_dependents().begin(); i != direct_dependents().end(); ++i)
  {
    the_primitive_t * primitive = r->elem(*i);
    if (!primitive->regenerated()) continue;
    
    the_tile_transform_t * tile_transform =
      dynamic_cast<the_tile_transform_t *>(primitive);
    
    the_tile_mosaic_t * tile_mosaic =
      dynamic_cast<the_tile_mosaic_t *>(primitive);
    
    the_tile_warp_widget_t * tile_widget =
      dynamic_cast<the_tile_warp_widget_t *>(primitive);
    
    if (tile_transform != NULL)
    {
      view_layout->add_primitive(tile_transform);
    }
    else if (tile_mosaic != NULL)
    {
      if (prefsUseOpenGLForMosaicking())
      {
	dl_->push_back(new the_tile_mosaic_dl_elem_t(tile_mosaic, selection_));
      }
    }
    else if (tile_widget != NULL)
    {
      dl_->push_back(tile_widget->dl_elem());
      view_layout->add_primitive(tile_widget);
    }
  }
  
  view_layout->gl_done_current();
  return regenerated_;
}

//----------------------------------------------------------------
// the_mosaic_proc_t::draw
// 
void
the_mosaic_proc_t::draw(const the_view_t & view) const
{
  view.view_mgr().reset_opengl_viewing();
  view.view_mgr().setup_opengl_3d_viewing();
  
  glPushAttrib(GL_ENABLE_BIT);
  {
    glDisable(GL_DEPTH_TEST);
    glEnable(GL_BLEND);
    
    the_document_ui_t * doc_ui = the_document_ui_t::doc_ui();
    if (&view == doc_ui->view(0))
    {
      // mosaic layout view:
      if (layout_dl_ready_ < layout_dl_elem_->size() &&
	  !prefsUseOpenGLForMosaicking())
      {
	layout_dl_elem(layout_dl_ready_)->draw();
      }
    }
    
    dl_->draw();
  }
  glPopAttrib();
}

//----------------------------------------------------------------
// the_mosaic_proc_t::calc_bbox
// 
void
the_mosaic_proc_t::calc_bbox(const the_view_t & view,
			     the_bbox_t & bbox) const
{
  std::list<the_tile_transform_t *> transforms;
  ::direct_dependents<the_tile_transform_t>(this, transforms);
  for (std::list<the_tile_transform_t *>::iterator i = transforms.begin();
       i != transforms.end(); ++i)
  {
    the_tile_transform_t * s = *i;
    if (!s->transform_.is_ready()) continue;
    
    const pnt2d_t & min = s->transform_.grid_min_;
    pnt2d_t max = min + s->transform_.grid_ext_;
    bbox << p3x1_t(min[0], min[1], 0) << p3x1_t(max[0], max[1], 0);
  }
}

//----------------------------------------------------------------
// the_mosaic_proc_t::tiles_add
// 
unsigned int
the_mosaic_proc_t::tiles_add(const the_text_t & filename,
			     unsigned int shrink_factor,
			     double spacing_x,
			     double spacing_y,
			     bool use_std_mask)
{
  the_tile_t * t = new the_tile_t();
  
  // shortcuts:
  the_registry_t * r = registry();
  the_image_pyramid_t & pyramid = *(t->pyramid_.get());
  pyramid.shrink_factor_ = shrink_factor;
  pyramid.spacing_x_ = spacing_x;
  pyramid.spacing_y_ = spacing_y;
  pyramid.use_std_mask_ = use_std_mask;
  pyramid.filename_ = filename;
  
  // add the tile to the registry:
  r->add(t);
  establish_supporter_dependent(r, id(), t->id());
  
  the_tile_transform_t * s = new the_tile_transform_t();
  r->add(s);
  establish_supporter_dependent(r, id(), s->id());
  establish_supporter_dependent(r, t->id(), s->id());
  
  the_tile_mosaic_t * l = layout_prim();
  establish_supporter_dependent(r, s->id(), l->id());
  
  l->request_regeneration();
  return t->id();
}

//----------------------------------------------------------------
// the_mosaic_proc_t::tiles_remove
// 
void
the_mosaic_proc_t::tiles_remove(unsigned int tile_id)
{
  if (!direct_supporter_of(tile_id)) return;
  
  if (tile_id == selection_)
  {
    unselect();
  }
  
  the_registry_t * r = registry();
  the_tile_t * t = r->elem<the_tile_t>(tile_id);
  the_tile_transform_t * s = ::first_dependent<the_tile_transform_t>(t);
  the_tile_mosaic_t * l = layout_prim();
  
  breakdown_supporter_dependent(r, s->id(), l->id());
  breakdown_supporter_dependent(r, tile_id, s->id());
  
  breakdown_supporter_dependent(r, id(), s->id());
  r->del(s);
  delete s;
  
  breakdown_supporter_dependent(r, id(), tile_id);
  r->del(t);
  delete t;
  
  l->request_regeneration();
  the_document_ui_t::doc_ui()->sync_ui();
}

//----------------------------------------------------------------
// the_mosaic_proc_t::tiles_clear
// 
void
the_mosaic_proc_t::tiles_clear()
{
  std::list<the_tile_t *> tiles;
  ::direct_dependents<the_tile_t>(this, tiles);
  
  while (!tiles.empty())
  {
    the_tile_t * t = remove_head(tiles);
    tiles_remove(t->id());
  }
}

//----------------------------------------------------------------
// the_mosaic_proc_t::apply_prefs
// 
void
the_mosaic_proc_t::apply_prefs()
{
  std::list<the_tile_t *> tiles;
  unsigned int num_tiles = ::direct_dependents<the_tile_t>(this, tiles);
  if (num_tiles == 0) return;
  
  unsigned int max_texture = OpenGL(0).max_texture_;
  bool use_luminance_textures = prefsUseLuminanceTextures();
  bool use_fragment_shader = prefsUseFragmentShader();
  bool use_std_mask = prefsUseStdMask();

  std::list<the_transaction_t *> schedule;
  for (std::list<the_tile_t *>::iterator i = tiles.begin();
       i != tiles.end(); ++i)
  {
    the_tile_t * tile = *i;
    the_tile_transaction_t * loader =
      new the_tile_transaction_t(tile->id(),
				 tile->pyramid_,
				 tile->pyramid_->shrink_factor_,
				 max_texture,
				 use_fragment_shader,
				 use_std_mask,
				 use_luminance_textures);
    schedule.push_back(loader);
  }
  TILE_THREAD_POOL()->stop_and_go(schedule);
  request_regeneration();
}

//----------------------------------------------------------------
// the_mosaic_proc_t::layout_ready
// 
void
the_mosaic_proc_t::layout_ready(const unsigned int & level)
{
  // get the display element:
  mosaic_dl_elem_t * e = NULL;
  {
    the_lock_t<the_mutex_interface_t> lock(layout_->cache_[level]->mutex());
    e = (mosaic_dl_elem_t *)(layout_->cache_[level]->dl_elem_.release());
  }
  
  // cerr << "FIXME: mosaic ready: " << level << ' ' << e << endl;
  if (e != NULL)
  {
    // assert(e->shared_->active() == selection_);
    
    // make the mosaic view context current right here:
    the_document_ui_t * doc_ui = the_document_ui_t::doc_ui();
    the_view_t * view_layout = doc_ui->view(0);
    view_layout->gl_make_current();
    layout_dl_elem(level) = boost::shared_ptr<mosaic_dl_elem_t>(e);
    view_layout->gl_done_current();
  }
  
  mosaic_dl_elem_t * dl_elem = layout_dl_elem(level).get();
  if (dl_elem == NULL) return;
  
#if 1
  if (level > layout_dl_ready_)
  {
    // higher pyramid level means higher pixel granularity -- lower precision,
    // therefore the current ready level should be examined to verify whether
    // is is actually necessary to switch to the newly delivered level:
    
    the_document_ui_t * doc_ui = the_document_ui_t::doc_ui();
    the_view_t * view_layout = doc_ui->view(0);
    
    const the_view_mgr_t & view_mgr = view_layout->view_mgr();
    the_view_volume_t vol;
    view_mgr.setup_view_volume(vol);
    
    p3x1_t view_min;
    p3x1_t view_max;
    vol.scs_to_wcs(p2x1_t(0, 0), view_min);
    vol.scs_to_wcs(p2x1_t(1, 1), view_max);
    view_min[2] = 0;
    view_max[2] = 0;
    
    the_aa_bbox_t requested_region;
    requested_region << view_min << view_max;
    
    std::vector<base_transform_t::ConstPointer> transform;
    get_transforms(this, transform);
    
    // make sure this object doesn't get deleted from some other thread:
    mosaic_cache_t * cache = layout_->cache_[layout_dl_ready_];
    boost::shared_ptr<mosaic_cache_shared_data_t> hold(cache->shared_);
    
    // shortcut:
    mosaic_cache_shared_data_t & shared = *(hold.get());
    
    if (shared.is_ready(requested_region, transform, selection_))
    {
      // cerr << "FIXME: staying at level: " << layout_dl_ready_ << endl;
      return;
    }
  }
#endif
  
#ifndef NDEBUG
  if (level != layout_dl_ready_)
  {
    cerr << "FIXME: switching to level: " << level << endl;
  }
#endif
  
  layout_dl_ready_ = level;
}

//----------------------------------------------------------------
// the_mosaic_proc_t::regenerate_layout
// 
bool
the_mosaic_proc_t::regenerate_layout()
{
  if (prefsUseOpenGLForMosaicking())
  {
    return true;
  }
  
  // regenerate the visible portion of the layout:
  the_document_ui_t * doc_ui = the_document_ui_t::doc_ui();
  the_view_t * view_layout = doc_ui->view(0);
  view_layout->eh_stack().view_cb(view_layout);
  return true;
}

//----------------------------------------------------------------
// the_mosaic_proc_t::setup_pyramids
// 
void
the_mosaic_proc_t::setup_pyramids()
{
  the_tile_t * tile = first_dependent<the_tile_t>(this);
  if (tile == NULL) return;

  const the_image_pyramid_t & pyramid = tile->pyramid();
  const unsigned int num_levels = pyramid.image_.size();
  if (num_levels == 0) return;
  
  layout_->setup(num_levels,
		 pyramid.image_[0].spacing_x_,
		 pyramid.image_[0].spacing_y_);
  layout_dl_elem_->resize(num_levels);
}

//----------------------------------------------------------------
// the_mosaic_proc_t::layout_prim
// 
the_tile_mosaic_t *
the_mosaic_proc_t::layout_prim() const
{
  return registry()->elem<the_tile_mosaic_t>(layout_id_);
}

//----------------------------------------------------------------
// the_mosaic_proc_t::select
// 
void
the_mosaic_proc_t::select(const unsigned int & selection)
{
  if (selection == selection_) return;
  
  the_tile_warp_widget_t * widget =
    ::first_dependent<the_tile_warp_widget_t>(this);
  the_registry_t * r = registry();
  
  if (selection_ != std::numeric_limits<unsigned int>::max())
  {
    the_tile_t * t = r->elem<the_tile_t>(selection_);
    the_tile_transform_t * s = ::first_dependent<the_tile_transform_t>(t);
    breakdown_supporter_dependent(r, s->id(), widget->id());
  }
  
  selection_ = selection;
  
  if (selection_ != std::numeric_limits<unsigned int>::max())
  {
    the_tile_t * t = r->elem<the_tile_t>(selection_);
    the_tile_transform_t * s = ::first_dependent<the_tile_transform_t>(t);
    establish_supporter_dependent(r, s->id(), widget->id());
    
    widget->deactivate_all();
  }
  
  request_regeneration();
}


//----------------------------------------------------------------
// get_transforms
// 
void
get_transforms(const std::list<the_tile_t *> & tiles,
	       const unsigned int num_tiles,
	       std::vector<base_transform_t::ConstPointer> & xform)
{
  xform.resize(num_tiles);
  unsigned int index = 0;
  for (std::list<the_tile_t *>::const_iterator iter = tiles.begin();
       iter != tiles.end(); ++iter, ++index)
  {
    const the_tile_t * t = *iter;
    the_tile_transform_t * s = first_dependent<the_tile_transform_t>(t);
    
    itk::GridTransform::Pointer itk_gt = itk::GridTransform::New();
    itk_gt->setup(s->transform_);
    xform[index] = itk_gt;
  }
}

//----------------------------------------------------------------
// get_transforms
// 
void
get_transforms(const the_mosaic_proc_t * proc,
	       std::vector<base_transform_t::ConstPointer> & xform)
{
  std::list<the_tile_t *> tiles;
  unsigned int num_tiles = ::direct_dependents<the_tile_t>(proc, tiles);
  get_transforms(tiles, num_tiles, xform);
}


//----------------------------------------------------------------
// the_tile_t::the_tile_t
// 
the_tile_t::the_tile_t():
  the_primitive_t(),
  pyramid_(new the_image_pyramid_t())
{}

//----------------------------------------------------------------
// the_tile_t::~the_tile_t
// 
the_tile_t::~the_tile_t()
{}

//----------------------------------------------------------------
// the_tile_t::regenerate
// 
bool
the_tile_t::regenerate()
{
  regenerated_ = pyramid_->is_ready();
  if (regenerated_)
  {
    GLenum filter = prefsAllowLinearFiltering() ? GL_LINEAR : GL_NEAREST;
    dl_elem_ =
      boost::shared_ptr<image_tile_dl_elem_t>
      (new image_tile_dl_elem_t(pyramid_->image_[0],
				filter,
				filter,
				CG_LA_FRAGMENT_PROGRAM_PTR));
  }
  else
  {
    dl_elem_ = boost::shared_ptr<image_tile_dl_elem_t>();
  }
  
  return regenerated_;
}


//----------------------------------------------------------------
// the_tile_transform_t::the_tile_transform_t
// 
the_tile_transform_t::the_tile_transform_t()
{}

//----------------------------------------------------------------
// the_tile_transform_t::regenerate
// 
bool
the_tile_transform_t::regenerate()
{
  the_tile_t * t = tile();
  regenerated_ =
    transform_.rows_ != 0 &&
    transform_.cols_ != 0 &&
    t != NULL;
  
  if (regenerated_ &&
      prefsUseOpenGLForMosaicking() &&
      t->dl_elem_.get() != NULL)
  {
    dl_elem_ = 
      boost::shared_ptr<the_warped_image_dl_elem_t>
      (new the_warped_image_dl_elem_t(t->dl_elem_, transform_));
  }
  else
  {
    dl_elem_ = boost::shared_ptr<the_warped_image_dl_elem_t>();
  }
  
  return regenerated_;
}

//----------------------------------------------------------------
// the_tile_transform_t::intersect
// 
bool
the_tile_transform_t::intersect(const the_view_volume_t & volume,
				std::list<the_pick_data_t> & data) const
{
  the_tile_t * t = tile();
  if (t == NULL)
  {
    return false;
  }
  
  const the_image_pyramid_t & pyramid = t->pyramid();
  if (pyramid.image_.size() == 0)
  {
    return false;
  }
  
  // 1. intersect the volume with the plane of the tile:
  const the_ray_t ray = volume.to_ray(p2x1_t(0.5, 0.5));
  const the_coord_sys_t csys;
  const the_plane_t plane(csys);
  float x_param = FLT_MAX;
  if (!plane.intersect(ray, x_param))
  {
    return false;
  }
  
  // 2. if the intersection point falls inside the tile
  //    test the intersection point against the tile mask.
  p3x1_t wcs_pt = ray * x_param;
  
  pnt2d_t mosaic_pt;
  mosaic_pt[0] = wcs_pt[0];
  mosaic_pt[1] = wcs_pt[1];
  
  // check where this point maps to in the tile:
  pnt2d_t uv_01;
  if (!transform_.transform(mosaic_pt, uv_01))
  {
    return false;
  }
  
  // because the_grid_transform_t maps into the [0, 1]^2 range,
  // remap the uv coordinates into physical tile coordinates:
  pnt2d_t tile_pt;
  tile_pt[0] =
    pyramid.image_[0].origin_x_ +
    uv_01[0] * pyramid.image_[0].w_ * pyramid.image_[0].spacing_x_;
  
  tile_pt[1] =
    pyramid.image_[0].origin_x_ +
    uv_01[1] * pyramid.image_[0].h_ * pyramid.image_[0].spacing_y_;
  
  mask_t::IndexType tile_ix;
  if (!pyramid.itk_mask_[0]->TransformPhysicalPointToIndex(tile_pt, tile_ix))
  {
    return false;
  }
  
  mask_t::PixelType alpha = pyramid.itk_mask_[0]->GetPixel(tile_ix);
  if (alpha < 128)
  {
    return false;
  }
  
  // shortcuts:
  p3x1_t cyl_pt;
  float & radius = cyl_pt.x();
  float & angle  = cyl_pt.y();
  float & depth  = cyl_pt.z();
  radius = 0;
  angle = 0;
  depth = x_param;
  
  // this point falls within the volume boundaries:
  data.push_back(the_pick_data_t(cyl_pt,
				 new the_tile_transform_ref_t
				 (id(), p2x1_t(uv_01[0], uv_01[1]))));
  return true;
}

//----------------------------------------------------------------
// the_tile_transform_t::set_anchor
// 
void
the_tile_transform_t::set_anchor()
{
  anchor_ = transform_;
}

//----------------------------------------------------------------
// the_tile_transform_t::move
// 
bool
the_tile_transform_t::move(const v3x1_t & wcs_vec)
{
  transform_ = anchor_;
  vec2d_t shift;
  shift[0] = wcs_vec[0];
  shift[1] = wcs_vec[1];
  transform_.shift(shift);
  request_regeneration();
  return true;
}

//----------------------------------------------------------------
// the_tile_transform_t::warp
// 
bool
the_tile_transform_t::warp(const p3x1_t & wcs_a,
			   const v3x1_t & wcs_vec,
			   const float & radius)
{
  pnt2d_t origin;
  origin[0] = wcs_a[0];
  origin[1] = wcs_a[1];
  
  vec2d_t shift;
  shift[0] = wcs_vec[0];
  shift[1] = wcs_vec[1];
  
  transform_ = anchor_;
  
  unsigned int n_vx = transform_.mesh_.size();
  for (unsigned int i = 0; i < n_vx; i++)
  {
    pnt2d_t & xy = transform_.mesh_[i].xy_;
    vec2d_t v = xy - origin;
    float sx = v[0] / radius;
    float sy = v[1] / radius;
    float t = expf(-(sx * sx + sy * sy));
    xy += shift * t;
  }
  
  transform_.rebuild_grid();
  request_regeneration();
  return true;
}


//----------------------------------------------------------------
// the_tile_transform_ref_t::the_tile_transform_ref_t
// 
the_tile_transform_ref_t::the_tile_transform_ref_t(const unsigned int & id,
						   const p2x1_t & uv):
  the_reference_t(id),
  uv_(uv)
{}

//----------------------------------------------------------------
// the_tile_transform_ref_t::eval
// 
bool
the_tile_transform_ref_t::eval(the_registry_t * r, p3x1_t & pt) const
{
  the_tile_transform_t * s = r->elem<the_tile_transform_t>(id());
  if (s == NULL) return false;
  
  pnt2d_t xy;
  s->transform_.transform_inv(pnt2d(uv_[0], uv_[1]), xy);
  
  pt[0] = xy[0];
  pt[1] = xy[1];
  pt[2] = 0.0f;
  return true;
}

//----------------------------------------------------------------
// the_tile_transform_ref_t::move
// 
// virtual:
bool
the_tile_transform_ref_t::move(the_registry_t * /* registry */,
			   const the_view_mgr_t & /* view_mgr */,
			   const p3x1_t & /* wcs_pt */)
{
  return false;
}

//----------------------------------------------------------------
// the_tile_transform_ref_t::dump
// 
void
the_tile_transform_ref_t::dump(ostream & strm, unsigned int indent) const
{
  strm << INDSCP << "the_tile_transform_ref_t(" << (void *)this << ")" << endl
       << INDSCP << "{" << endl;
  the_reference_t::dump(strm, INDNXT);
  strm << INDSCP << "}" << endl << endl;
}


//----------------------------------------------------------------
// the_tile_mosaic_dl_elem_t::the_tile_mosaic_dl_elem_t
// 
the_tile_mosaic_dl_elem_t::
the_tile_mosaic_dl_elem_t(const the_tile_mosaic_t * mosaic,
			  const unsigned int & active):
  active_(std::numeric_limits<unsigned int>::max())
{
  std::list<the_tile_transform_t *> transforms;
  unsigned int num_tiles =
    ::direct_supporters<the_tile_transform_t>(mosaic, transforms);
  
  // cerr << this << " new mosaic dl elem: " << num_tiles << endl;
  
  tile_dl_elem_.resize(num_tiles);
  unsigned int index = 0;
  for (std::list<the_tile_transform_t *>::iterator iter = transforms.begin();
       iter != transforms.end(); ++iter, ++index)
  {
    the_tile_transform_t * tt = *iter;
    tile_dl_elem_[index] = tt->dl_elem_;
    
    the_tile_t * t = ::first_supporter<the_tile_t>(tt);
    if (t->id() == active)
    {
      active_ = index;
    }
  }
}

//----------------------------------------------------------------
// the_tile_mosaic_dl_elem_t::draw
// 
void
the_tile_mosaic_dl_elem_t::draw() const
{
  static const the_color_t zebra[] = {
    the_color_t(0, 0.5, 1, 1),
    the_color_t(1, 0.5, 0, 0)
  };
  
  unsigned int num_tiles = tile_dl_elem_.size();
  // cerr << this << " drawing the mosaic: " << num_tiles << endl;
  
  for (unsigned int i = 0; i < num_tiles; i++)
  {
    the_warped_image_dl_elem_t * e = tile_dl_elem_[i].get();
    e->draw(the_color_t::BLACK, false);
  }
  
  for (unsigned int i = 0; i < num_tiles; i++)
  {
    if (i == active_) continue;
    the_warped_image_dl_elem_t * e = tile_dl_elem_[i].get();
    e->draw(zebra[0], true);
  }
  
  if (active_ < num_tiles)
  {
    the_warped_image_dl_elem_t * e = tile_dl_elem_[active_].get();
    e->draw(zebra[1], true);
  }
}

//----------------------------------------------------------------
// the_tile_mosaic_dl_elem_t::update_bbox
// 
void
the_tile_mosaic_dl_elem_t::update_bbox(the_bbox_t & bbox) const
{
  unsigned int num_tiles = tile_dl_elem_.size();
  for (unsigned int i = 0; i < num_tiles; i++)
  {
    the_warped_image_dl_elem_t * e = tile_dl_elem_[i].get();
    e->update_bbox(bbox);
  }
}


//----------------------------------------------------------------
// the_tile_mosaic_t::the_tile_mosaic_t
// 
the_tile_mosaic_t::the_tile_mosaic_t()
{}

//----------------------------------------------------------------
// the_tile_mosaic_t::regenerate
// 
bool
the_tile_mosaic_t::regenerate()
{
  regenerated_ = true;
  the_mosaic_proc_t * proc = first_supporter<the_mosaic_proc_t>(this);
  proc->regenerate_layout();
  /*
  if (dl_elem_.get() == NULL || dl_elem_->active_ != proc->selection_)
  {
    dl_elem_ =
      boost::shared_ptr<the_tile_mosaic_dl_elem_t>
      (new the_tile_mosaic_dl_elem_t(this, proc->selection_));
  }
  */
  return regenerated_;
}


//----------------------------------------------------------------
// _3rd
// 
static const double _3rd[] = {
  0.0f,
  1.0f / 3.0f,
  2.0f / 3.0f,
  1.0f
};

//----------------------------------------------------------------
// the_tile_warp_widget_t::uv_
// 
const p2x1_t
the_tile_warp_widget_t::uv_[12] = {
  p2x1_t(_3rd[0], _3rd[0]),
  p2x1_t(_3rd[1], _3rd[0]),
  p2x1_t(_3rd[2], _3rd[0]),
  p2x1_t(_3rd[3], _3rd[0]),
  p2x1_t(_3rd[0], _3rd[1]),
  p2x1_t(_3rd[3], _3rd[1]),
  p2x1_t(_3rd[0], _3rd[2]),
  p2x1_t(_3rd[3], _3rd[2]),
  p2x1_t(_3rd[0], _3rd[3]),
  p2x1_t(_3rd[1], _3rd[3]),
  p2x1_t(_3rd[2], _3rd[3]),
  p2x1_t(_3rd[3], _3rd[3])
};

//----------------------------------------------------------------
// the_tile_warp_widget_t::the_tile_warp_widget_t
// 
the_tile_warp_widget_t::the_tile_warp_widget_t()
{
  deactivate_all();
}

//----------------------------------------------------------------
// the_tile_warp_widget_t::~the_tile_warp_widget_t
// 
the_tile_warp_widget_t::~the_tile_warp_widget_t()
{}

//----------------------------------------------------------------
// the_tile_warp_widget_t::regenerate
// 
bool
the_tile_warp_widget_t::regenerate()
{
  regenerated_ = false;
  
  // look up the supporting procedure:
  the_mosaic_proc_t * proc = ::first_supporter<the_mosaic_proc_t>(this);
  
  if (!proc->has_selection())
  {
    // done:
    regenerated_ = true;
  }
  else
  {
    // look up the active tile transform:
    the_tile_t * tile = registry()->elem<the_tile_t>(proc->selection());
    if (tile->regenerated())
    {
      the_tile_transform_t * transform =
	::first_dependent<the_tile_transform_t>(tile);
      if (transform->regenerated())
      {
	// update the control points:
	for (unsigned int i = 0; i < 12; i++)
	{
	  pnt2d_t itk_uv;
	  itk_uv[0] = uv_[i][0];
	  itk_uv[1] = uv_[i][1];
	  
	  pnt2d_t itk_xy;
	  transform->transform_.transform_inv(itk_uv, itk_xy);
	  xy_[i][0] = itk_xy[0];
	  xy_[i][1] = itk_xy[1];
	  xy_[i][2] = 0.0;
	}
	
	regenerated_ = true;
      }
    }
  }
  
  return regenerated_;
}

//----------------------------------------------------------------
// the_tile_warp_widget_t::intersect
// 
bool
the_tile_warp_widget_t::intersect(const the_view_volume_t & volume,
				  std::list<the_pick_data_t> & data) const
{
  // test the control points:
  if (!regenerated_) return false;
  
  the_tile_transform_t * transform =
    ::first_supporter<the_tile_transform_t>(this);
  if (transform == NULL) return false;
  
  for (unsigned int i = 0; i < 12; i++)
  {
    const p3x1_t & wcs_pt = xy_[i];
    
    // shortcuts:
    p3x1_t cyl_pt;
    float & radius = cyl_pt.x();
    float & angle  = cyl_pt.y();
    float & depth  = cyl_pt.z();
    
    // find the depth of this point within the volume:
    depth = volume.depth_of_wcs_pt(wcs_pt);
    
    // find the UV frame within the volume at the specified depth:
    the_cyl_uv_csys_t uv_frame;
    volume.uv_frame_at_depth(depth, uv_frame);
    
    // calculate the polar coordinates of this point within the UV frame:
    uv_frame.wcs_to_lcs(wcs_pt, radius, angle);
    
    // if the radius of this point within the UV frame exceeds 1.0 then
    // this point is outside the volume - does not intersect:
    if (radius < 1.0)
    {
      // this point falls within the volume boundaries:
      data.push_back(the_pick_data_t
		     (cyl_pt, new the_tile_warp_widget_ref_t(id(), i)));
      return true;
    }
  }
  
  return false;
}

//----------------------------------------------------------------
// the_tile_warp_widget_t::dl_elem
// 
the_dl_elem_t *
the_tile_warp_widget_t::dl_elem() const
{
  if (!regenerated_) return NULL;
  
  the_tile_transform_t * transform =
    ::first_supporter<the_tile_transform_t>(this);
  if (transform == NULL) return NULL;
  
  // return the display list element:
  return new the_tile_warp_widget_dl_elem_t(this);
}

//----------------------------------------------------------------
// the_tile_warp_widget_t::activate
// 
void
the_tile_warp_widget_t::activate(const unsigned int & pt_index)
{
  active_[pt_index] = true;
}

//----------------------------------------------------------------
// the_tile_warp_widget_t::deactivate
// 
void
the_tile_warp_widget_t::deactivate(const unsigned int & pt_index)
{
  active_[pt_index] = false;
}

//----------------------------------------------------------------
// the_tile_warp_widget_t::toggle
// 
bool
the_tile_warp_widget_t::toggle(const unsigned int & pt_index)
{
  active_[pt_index] = !active_[pt_index];
  return active_[pt_index];
}

//----------------------------------------------------------------
// the_tile_warp_widget_t::is_active
// 
bool
the_tile_warp_widget_t::is_active(const unsigned int & pt_index) const
{
  return active_[pt_index];
}

//----------------------------------------------------------------
// the_tile_warp_widget_t::has_active_points
// 
bool
the_tile_warp_widget_t::has_active_points() const
{
  bool active = active_[0];
  for (unsigned int i = 1; i < 12 && !active; i++)
  {
    active = active || active_[i];
  }
  
  return active;
}

//----------------------------------------------------------------
// the_tile_warp_widget_t::deactivate_all
// 
void
the_tile_warp_widget_t::deactivate_all()
{
  for (unsigned int i = 0; i < 12; i++)
  {
    active_[i] = false;
  }
}

//----------------------------------------------------------------
// the_tile_warp_widget_t::get_bsplines
// 
void
the_tile_warp_widget_t::get_bsplines(the_bspline_geom_t * geom) const
{
  // setup the curves:
  static const unsigned int pt_idx[4][4] = {
    { 0, 1, 2, 3 },	// (u, 0)
    { 8, 9, 10, 11 },	// (u, 1)
    { 0, 4, 6, 8 },	// (0, v)
    { 3, 5, 7, 11 }	// (1, v)
  };
  
  for (unsigned int i = 0; i < 4; i++)
  {
    std::vector<p3x1_t> pts(4);
    std::vector<float> wts(4);
    std::vector<float> kts(4);
    
    for (unsigned int j = 0; j < 4; j++)
    {
      pts[j] = xy_[pt_idx[i][j]];
      kts[j] = float(j) / 3.0f;
      wts[j] = 1.0f;
    }
    
    the_interpolation_bspline_t bspline;
    bspline.update_geom(pts, wts, kts);
    geom[i] = bspline.geom();
  }
}

//----------------------------------------------------------------
// evaluate_surface
// 
bool
evaluate_surface(const the_bspline_geom_t & uv0,
		 const the_bspline_geom_t & uv1,
		 const the_bspline_geom_t & vu0,
		 const the_bspline_geom_t & vu1,
		 const p2x1_t & uv,
		 p3x1_t & xy)
{
  p3x1_t pt_u[2];
  p3x1_t pt_v[2];
  p3x1_t corner[4];
  
  if (uv0.position(uv[0], pt_u[0]) &&
      uv1.position(uv[0], pt_u[1]) &&
      vu0.position(uv[1], pt_v[0]) &&
      vu1.position(uv[1], pt_v[1]) &&
      
      // FIXME: this is wasteful:
      uv0.position(uv0.t_min(), corner[0]) &&
      uv0.position(uv0.t_max(), corner[1]) &&
      uv1.position(uv1.t_min(), corner[2]) &&
      uv1.position(uv1.t_max(), corner[3]))
  {
    p3x1_t a = pt_v[0] + (pt_v[1] - pt_v[0]) * uv[0];
    p3x1_t b = pt_u[0] + (pt_u[1] - pt_u[0]) * uv[1];
    
    p3x1_t v0 = corner[0] + (corner[1] - corner[0]) * uv[0];
    p3x1_t v1 = corner[2] + (corner[3] - corner[2]) * uv[0];
    p3x1_t c = v0 + (v1 - v0) * uv[1];
    
    xy[0] = a[0] + b[0] - c[0];
    xy[1] = a[1] + b[1] - c[1];
    xy[2] = a[2] + b[2] - c[2];
    return true;
  }
  
  return false;
}

//----------------------------------------------------------------
// get_local_frame
// 
bool
get_local_frame(const the_bspline_geom_t & uv0,
		const the_bspline_geom_t & uv1,
		const the_bspline_geom_t & vu0,
		const the_bspline_geom_t & vu1,
		const p2x1_t & uv_origin,
		the_coord_sys_t & csys)
{
  float u_unit = 1e-2f;
  float v_unit = 1e-2f;
  
  float shift_u =
    (uv_origin[0] < 0) ? -uv_origin[0] :
    (uv_origin[0] + u_unit > 1) ? 1 - (uv_origin[0] + u_unit) : 0;
  
  float shift_v =
    (uv_origin[1] < 0) ? -uv_origin[1] :
    (uv_origin[1] + v_unit > 1) ? 1 - (uv_origin[1] + v_unit) : 0;
  
  p2x1_t q00(uv_origin[0] + shift_u, uv_origin[1] + shift_v);
  p2x1_t q10(q00[0] + u_unit, q00[1]);
  p2x1_t q01(q00[0], q00[1] + v_unit);
  
  p3x1_t p00;
  p3x1_t p10;
  p3x1_t p01;
  if (evaluate_surface(uv0, uv1, vu0, vu1, q00, p00) &&
      evaluate_surface(uv0, uv1, vu0, vu1, q10, p10) &&
      evaluate_surface(uv0, uv1, vu0, vu1, q01, p01))
  {
    csys.origin() = p00;
    csys.x_axis() = (p10 - p00) / u_unit;
    csys.y_axis() = (p01 - p00) / v_unit;
    csys.z_axis() = csys.x_axis() % csys.y_axis();
    return true;
  }
  
  return false;
}

//----------------------------------------------------------------
// closest_point_on_surface
// 
bool
closest_point_on_surface(const the_bspline_geom_t & uv0,
			 const the_bspline_geom_t & uv1,
			 const the_bspline_geom_t & vu0,
			 const the_bspline_geom_t & vu1,
			 const p3x1_t & xy,
			 const p2x1_t & uv_seed,
			 p2x1_t & uv)
{
#ifdef DEBUG_SURFACE_PROJECTION
  cerr << "- - - - - - - - - - - - - - - - - - - - - - - - - - - - -" << endl;
#endif
  
  static const unsigned int max_iterations = 8;
  static const unsigned int pick_up_pace_steps = 5;
  static const float max_step_scale = 1;
  static const float min_step_scale = 1e-12f;
  static const float min_error_sqrd = 1e-16f;
  
  // temporaries:
  the_coord_sys_t frame_i;
  p3x1_t xy_i;
  p2x1_t uv_i(uv_seed);
  v3x1_t error_xy;
  v3x1_t error_uv(0, 0, 0);
  float err_i = std::numeric_limits<double>::max();
  
  unsigned int successful_steps = 0;
  float step_scale = max_step_scale;
  
  for (unsigned int i = 0; i < max_iterations; i++)
  {
    bool step_ok = evaluate_surface(uv0, uv1, vu0, vu1, uv_i, xy_i);
    assert(step_ok);
    
    if (step_ok)
    {
      // take a look at what happened to the error:
      error_xy = xy_i - xy;
      float err = error_xy.norm_sqrd();
      if (err < err_i)
      {
#ifdef DEBUG_SURFACE_PROJECTION
	cerr << i << ". error: " << err;
#endif
	uv = uv_i;
	
	if (err < min_error_sqrd)
	{
#ifdef DEBUG_SURFACE_PROJECTION
	  cerr << ", done" << endl;
#endif
	  break;
	}
	
	err_i = err;
	get_local_frame(uv0, uv1, vu0, vu1, uv, frame_i);
	error_uv = frame_i.to_lcs(error_xy);
	successful_steps++;
	
	if (successful_steps % pick_up_pace_steps == 0)
	{
#ifdef DEBUG_SURFACE_PROJECTION
	  float new_step_scale = std::min(max_step_scale, step_scale * 2);
	  if (step_scale != new_step_scale)
	  {
	    step_scale = new_step_scale;
	    cerr << ", " << successful_steps
		 << " successful steps -- increasing pace, new step scale: "
		 << step_scale << endl;
	  }
	  else
	  {
	    cerr << ", ok -- can't increase pace past the max" << endl;
	  }
#else
	  step_scale = std::min(max_step_scale, step_scale * 2);
#endif
	}
#ifdef DEBUG_SURFACE_PROJECTION
	else
	{
	  cerr << ", ok" << endl;
	}
#endif
      }
      else
      {
	step_ok = false;
      }
    }

    if (!step_ok)
    {
      step_scale /= 2.0;
      if (step_scale < min_step_scale)
      {
#ifdef DEBUG_SURFACE_PROJECTION
	cerr << i << ". step scale too small, giving up..." << endl;
#endif
	break;
      }
      
      successful_steps = 0;
#ifdef DEBUG_SURFACE_PROJECTION
      cerr << i << ". relaxing and backtracking, new step scale: "
	   << step_scale << endl;
#endif
    }
    
    // setup the next trial point:
    uv_i[0] = std::min(1.0f, std::max(0.0f, uv[0] - step_scale * error_uv[0]));
    uv_i[1] = std::min(1.0f, std::max(0.0f, uv[1] - step_scale * error_uv[1]));
  }
  
  return true;
}

//----------------------------------------------------------------
// the_tile_warp_widget_t::set_anchor
// 
void
the_tile_warp_widget_t::set_anchor()
{
  // store the current warp surface:
  get_bsplines(anchor_);
  
  // store the current interpolation points:
  for (unsigned int i = 0; i < 12; i++)
  {
    anchor_xy_[i] = xy_[i];
  }
  
  // lookup the transform and project the interpolation points onto the
  // surface:
  the_tile_transform_t * transform =
    ::first_supporter<the_tile_transform_t>(this);
  assert(transform != NULL);
  if (transform == NULL) return;

  // shortcut:
  transform->set_anchor();
  
  const the_grid_transform_t & gt = transform->anchor_;
  const unsigned int n_rows = gt.rows_;
  const unsigned int n_cols = gt.cols_;
  
  anchor_uv_.resize(gt.mesh_.size());
  p2x1_t uv;
  for (unsigned int row = 0; row <= n_rows; row++)
  {
    uv[1] = double(row) / double(n_rows);
    for (unsigned int col = 0; col <= n_cols; col++)
    {
      uv[0] = double(col) / double(n_cols);
      
      const vertex_t & vx = gt.vertex(row, col);
      unsigned int index = row * (n_cols + 1) + col;
      closest_point_on_surface(anchor_[0],
			       anchor_[1],
			       anchor_[2],
			       anchor_[3],
			       p3x1_t(vx.xy_[0], vx.xy_[1], 0),
			       uv,
			       anchor_uv_[index]);
    }
  }
}

//----------------------------------------------------------------
// the_tile_warp_widget_t::move
// 
bool
the_tile_warp_widget_t::move(const v3x1_t & wcs_vec)
{
  // construct a bspline surface, interpolating the boundary curves,
  for (unsigned int i = 0; i < 12; i++)
  {
    if (active_[i])
    {
      xy_[i] = anchor_xy_[i] + wcs_vec;
    }
    else
    {
      xy_[i] = anchor_xy_[i];
    }
  }
  
  the_bspline_geom_t geom[4];
  get_bsplines(geom);
  
  // lookup the transform and project the interpolation points onto the
  // surface:
  the_tile_transform_t * transform =
    ::first_supporter<the_tile_transform_t>(this);
  assert(transform != NULL);
  if (transform == NULL) return false;
  
  const the_grid_transform_t & gt_anchor = transform->anchor_;
  the_grid_transform_t & gt = transform->transform_;
  const unsigned int n_rows = gt.rows_;
  const unsigned int n_cols = gt.cols_;
  
  // find the displacements:
  p3x1_t xy[2];
  fast_array_t<vec2d_t> shift(gt.mesh_.size());
  for (unsigned int row = 0; row <= n_rows; row++)
  {
    for (unsigned int col = 0; col <= n_cols; col++)
    {
      unsigned int index = row * (n_cols + 1) + col;
      
      // evaluate the anchor:
      evaluate_surface(anchor_[0],
		       anchor_[1],
		       anchor_[2],
		       anchor_[3],
		       anchor_uv_[index],
		       xy[0]);
      
      // evaluate the warped surface:
      evaluate_surface(geom[0],
		       geom[1],
		       geom[2],
		       geom[3],
		       anchor_uv_[index],
		       xy[1]);
      
      vec2d_t & v = shift[index];
      v[0] = xy[1][0] - xy[0][0];
      v[1] = xy[1][1] - xy[0][1];
    }
  }
  
  gt.mesh_ = gt_anchor.mesh_;
  gt.update(&shift[0]);
  transform->request_regeneration();
  return true;
}


//----------------------------------------------------------------
// the_tile_warp_widget_ref_t::the_tile_warp_widget_ref_t
// 
the_tile_warp_widget_ref_t::
the_tile_warp_widget_ref_t(const unsigned int & id,
			   const unsigned int & pt_index):
  the_reference_t(id),
  pt_index_(pt_index)
{}

//----------------------------------------------------------------
// the_tile_warp_widget_ref_t::eval
// 
bool
the_tile_warp_widget_ref_t::eval(the_registry_t * r, p3x1_t & pt) const
{
  the_tile_warp_widget_t * w = r->elem<the_tile_warp_widget_t>(id());
  if (w == NULL) return false;
  
  pt = w->xy_[pt_index_];
  return true;
}

//----------------------------------------------------------------
// the_tile_warp_widget_ref_t::move
// 
// virtual:
bool
the_tile_warp_widget_ref_t::move(the_registry_t * /* registry */,
				 const the_view_mgr_t & /* view_mgr */,
				 const p3x1_t & /* wcs_pt */)
{
  return false;
}

//----------------------------------------------------------------
// the_tile_warp_widget_ref_t::dump
// 
void
the_tile_warp_widget_ref_t::dump(ostream & strm, unsigned int indent) const
{
  strm << INDSCP << "the_tile_warp_widget_ref_t(" << (void *)this << ")"
       << endl
       << INDSCP << "{" << endl;
  the_reference_t::dump(strm, INDNXT);
  strm << INDSCP << "}" << endl << endl;
}


//----------------------------------------------------------------
// the_tile_warp_widget_dl_elem_t::the_tile_warp_widget_dl_elem_t
// 
the_tile_warp_widget_dl_elem_t::
the_tile_warp_widget_dl_elem_t(const the_tile_warp_widget_t * widget)
{
  // setup the curves:
  widget->get_bsplines(geom_);
  for (unsigned int i = 0; i < 4; i++)
  {
    curve_[i] = new the_bspline_geom_dl_elem_t(geom_[i], the_color_t::RED);
  }
  
  // setup the interpolation points:
  for (unsigned int i = 0; i < 12; i++)
  {
    const the_color_t & color =
      widget->active_[i] ?
      the_color_t::RED :
      the_color_t::WHITE;
    
    point_[i] =
      new the_masked_symbol_dl_elem_t
      (widget->xy_[i],
       color,
       the_color_t::BLACK,
       THE_POINT_SYMBOLS,
       THE_SMALL_FILLED_CIRCLE_SYMBOL_E);
  }
}

//----------------------------------------------------------------
// the_tile_warp_widget_dl_elem_t::~the_tile_warp_widget_dl_elem_t
// 
the_tile_warp_widget_dl_elem_t::~the_tile_warp_widget_dl_elem_t()
{
  for (unsigned int i = 0; i < 12; i++)
  {
    delete point_[i];
  }
  
  for (unsigned int i = 0; i < 4; i++)
  {
    delete curve_[i];
  }
}

//----------------------------------------------------------------
// the_tile_warp_widget_dl_elem_t::draw
// 
void
the_tile_warp_widget_dl_elem_t::draw() const
{
  for (unsigned int i = 0; i < 4; i++)
  {
    curve_[i]->draw();
  }
  
  for (unsigned int i = 0; i < 12; i++)
  {
    point_[i]->draw();
  }
}

//----------------------------------------------------------------
// the_tile_warp_widget_dl_elem_t::update_bbox
// 
void
the_tile_warp_widget_dl_elem_t::update_bbox(the_bbox_t & bbox) const
{
  for (unsigned int i = 0; i < 12; i++)
  {
    point_[i]->update_bbox(bbox);
  }
}
