// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : the_mosaic_proc.hxx
// Author       : Pavel A. Koshevoy
// Created      : Wed Feb 14 12:10:00 MST 2007
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : The document procedure containing the mosaic layout.

#ifndef THE_MOSAIC_PROC_HXX_
#define THE_MOSAIC_PROC_HXX_

// Boost includes:
#include <boost/shared_ptr.hpp>

// the includes:
#include <doc/the_procedure.hxx>
#include <geom/the_point.hxx>
#include <geom/the_bspline.hxx>
#include <utils/the_text.hxx>
#include <utils/the_dynamic_array.hxx>
#include <image/image_tile_generator.hxx>
#include <opengl/the_disp_list.hxx>
#include <itk/common.hxx>
#include <itk/the_grid_transform.hxx>
#include <itk/itkGridTransform.h>

// Qt includes:
#include <QImage>

// Cg includes:
#include <Cg/cgGL.h>

// local includes:
#include "mosaic_cache.hxx"
#include "mosaic_cache_pyramid.hxx"
#include "the_warped_image_dl_elem.hxx"

// forward declarations:
extern CGprogram * CG_RGBA16_FRAGMENT_PROGRAM_PTR;
class the_bbox_t;
class the_view_t;
class the_tile_t;
class the_tile_transform_t;
class the_tile_mosaic_t;
class the_warped_image_dl_elem_t;


//----------------------------------------------------------------
// mosaic_dl_elem_t
// 
class mosaic_dl_elem_t : public image_tile_dl_elem_t
{
public:
  mosaic_dl_elem_t(boost::shared_ptr<mosaic_cache_shared_data_t> & shared,
		   GLenum min_filter = GL_NEAREST,
		   GLenum mag_filter = GL_NEAREST):
#ifdef USE_CG
    image_tile_dl_elem_t(shared->data_,
			 min_filter,
			 mag_filter,
			 CG_RGBA16_FRAGMENT_PROGRAM_PTR),
#else
    image_tile_dl_elem_t(shared->data_,
			 min_filter,
			 mag_filter),
#endif
    shared_(shared)
  {}
  
  ~mosaic_dl_elem_t()
  {}
  
  // virtual:
  void draw() const
  {
    // upload the texture data:
#if 1
    {
      // NOTE: this is an optimization which avoids uploading
      // all texture tiles, and instead uploads only the updated
      // portions of affected texture tiles:
      
      the_lock_t<the_mutex_interface_t> locker(shared_->mutex_);
      upload_.splice(upload_.end(), shared_->upload_);
    }
#endif
    
#ifdef USE_CG
    if (has_fragment_program())
    {
      cgGLLoadProgram(*fragment_program_);
    }
#endif
    
    image_tile_dl_elem_t::draw();
  }
  
  boost::shared_ptr<mosaic_cache_shared_data_t> shared_;
  
private:
  mosaic_dl_elem_t();
};

//----------------------------------------------------------------
// the_mosaic_proc_t
// 
class the_mosaic_proc_t : public the_procedure_t
{
public:
  the_mosaic_proc_t();
  ~the_mosaic_proc_t();
  
  // virtual:
  the_primitive_t * clone() const
  { return new the_mosaic_proc_t(*this); }
  
  // virtual:
  const char * name() const
  { return "the_mosaic_proc_t"; }
  
  // virtual:
  void added_to_the_registry(the_registry_t * registry,
			     const unsigned int & id);
  
  // virtual:
  bool regenerate();
  
  // virtual:
  void draw(const the_view_t & view) const;
  
  // virtual: calculate the model bounding box:
  void calc_bbox(const the_view_t & view, the_bbox_t & bbox) const;
  
  // add a tile to the mosaic, return the tile primitive id:
  unsigned int tiles_add(const the_text_t & filename,
			 unsigned int shrink_factor,
			 double spacing_x,
			 double spacing_y,
			 bool use_std_mask);
  
  // remove a tile based on its primitive id:
  void tiles_remove(unsigned int tile_id);
  
  // remove all tiles:
  void tiles_clear();
  
  // this will be called to regenerate the opengl textures:
  void apply_prefs();
  
  // trigger mosaic cache regeneration:
  bool regenerate_layout();
  
  // update the mosaic display list:
  void layout_ready(const unsigned int & level);
  
  // setup the mosaic cache pyramid:
  void setup_pyramids();
  
  // helpers:
  the_tile_mosaic_t * layout_prim() const;
  
  // accessor:
  inline mosaic_cache_pyramid_t & layout()
  { return *layout_; }
  
  // selection set accessors:
  inline const unsigned int & selection() const
  { return selection_; }
  
  inline void unselect()
  { select(std::numeric_limits<unsigned int>::max()); }
  
  inline bool has_selection() const
  { return selection_ != std::numeric_limits<unsigned int>::max(); }
  
  void select(const unsigned int & selection);
  
protected:
  // helpers for accessing a given mosaic level display element:
  inline boost::shared_ptr<mosaic_dl_elem_t> &
  layout_dl_elem(const unsigned int & level)
  { return (*(layout_dl_elem_.get()))[level]; }
  
  inline const mosaic_dl_elem_t *
  layout_dl_elem(const unsigned int & level) const
  { return (*(layout_dl_elem_.get()))[level].get(); }
  
  boost::shared_ptr<the_disp_list_t> dl_;
  
  // multi resolution mosaics:
  boost::shared_ptr<mosaic_cache_pyramid_t> layout_;
  
  // multi-resolution display elements for the mosaic:
  boost::shared_ptr<std::vector<boost::shared_ptr<mosaic_dl_elem_t> > >
  layout_dl_elem_;
  unsigned int layout_dl_ready_;
  
  // id's of placeholder primitives:
  unsigned int layout_id_;
  
  // a simple selection set:
  unsigned int selection_;
};

//----------------------------------------------------------------
// get_transforms
// 
extern void
get_transforms(const std::list<the_tile_t *> & tiles,
	       const unsigned int num_tiles,
	       std::vector<base_transform_t::ConstPointer> & xform);

//----------------------------------------------------------------
// get_transforms
// 
extern void
get_transforms(const the_mosaic_proc_t * proc,
	       std::vector<base_transform_t::ConstPointer> & xform);


//----------------------------------------------------------------
// the_image_pyramid_t
// 
class the_image_pyramid_t
{
public:
  typedef enum {
    NOT_READY_E,
    LOADING_E,
    FAILED_E,
    READY_E
  } state_t;
  
  the_image_pyramid_t():
    state_(NOT_READY_E)
  {}
  
  // check whether the pyramid is ready:
  bool is_ready() const
  { return state_ == READY_E; }
  
  // the image source file name:
  the_text_t filename_;
  
  // downsample the image at load time:
  unsigned int shrink_factor_;
  
  // pixel spacing:
  double spacing_x_;
  double spacing_y_;
  
  // the Marc Lab standard mask flag:
  bool use_std_mask_;
  
  // multi resolution image pyramid, level 0 is the highest resolution,
  // every following level is at half the resolution of the level
  // preceeding it.
  the_dynamic_array_t<native_image_t::Pointer> itk_image_;
  the_dynamic_array_t<mask_t::Pointer> itk_mask_;
  
  // a copy of the image pyramid optimized for OpenGL:
  the_dynamic_array_t<image_tile_generator_t> image_;
  
  // current state of the pyramid:
  state_t state_;
  
  // the thumbnail:
  QImage thumbnail_;
};


//----------------------------------------------------------------
// the_tile_t
// 
class the_tile_t : public the_primitive_t
{
public:
  the_tile_t();
  ~the_tile_t();
  
  // virtual:
  the_primitive_t * clone() const
  { return new the_tile_t(*this); }
  
  // virtual:
  const char * name() const
  { return "the_tile_t"; }
  
  // virtual:
  bool regenerate();
  
  // pyramid accessors:
  inline const the_image_pyramid_t & pyramid() const
  { return *(pyramid_.get()); }
  
  inline the_image_pyramid_t & pyramid()
  { return *(pyramid_.get()); }
  
  // multi-resolution image pyramid:
  boost::shared_ptr<the_image_pyramid_t> pyramid_;
  
  boost::shared_ptr<image_tile_dl_elem_t> dl_elem_;
  
  // imported transform -- use it for layout/mosaic:
  base_transform_t::Pointer imported_;
};


//----------------------------------------------------------------
// the_tile_transform_t
// 
class the_tile_transform_t : public the_primitive_t
{
public:
  the_tile_transform_t();
  
  // virtual:
  the_primitive_t * clone() const
  { return new the_tile_transform_t(*this); }
  
  // virtual:
  const char * name() const
  { return "the_tile_transform_t"; }
  
  // virtual:
  bool regenerate();
  
  // virtual: this is used during intersection/proximity testing:
  bool intersect(const the_view_volume_t & volume,
		 std::list<the_pick_data_t> & data) const;
  
  // helper:
  inline the_tile_t * tile() const
  { return ::first_supporter<the_tile_t>(this); }
  
  void set_anchor();
  bool move(const v3x1_t & wcs_vec);
  bool warp(const p3x1_t & wcs_a,
	    const v3x1_t & wcs_vec,
	    const float & radius);
  
  // the tile warp transform:
  the_grid_transform_t transform_;
  the_grid_transform_t anchor_;
  
  // the warped tile display list element:
  boost::shared_ptr<the_warped_image_dl_elem_t> dl_elem_;
};

//----------------------------------------------------------------
// the_tile_transform_ref_t
// 
// Reference to a tile transform:
// 
class the_tile_transform_ref_t : public the_reference_t
{
public:
  the_tile_transform_ref_t(const unsigned int & id,
			   const p2x1_t & uv);
  
  // virtual: a method for cloning references (potential memory leak):
  the_reference_t * clone() const
  { return new the_tile_transform_ref_t(*this); }
  
  // virtual:
  const char * name() const
  { return "the_tile_transform_ref_t"; }
  
  // virtual: Calculate the 3D value of this reference:
  bool eval(the_registry_t * registry, p3x1_t & pt) const;
  
  // virtual:
  bool move(the_registry_t * registry,
	    const the_view_mgr_t & view_mgr,
	    const p3x1_t & wcs_pt);
  
  // virtual:
  the_point_symbol_id_t symbol() const
  { return THE_CORNERS_SYMBOL_E; }
  
  // virtual:
  void dump(ostream & strm, unsigned int indent = 0) const;
  
private:
  the_tile_transform_ref_t();
  
  // mesh coordinates of the reference:
  p2x1_t uv_;
};


//----------------------------------------------------------------
// the_tile_mosaic_dl_elem_t
// 
class the_tile_mosaic_dl_elem_t : public the_dl_elem_t
{
public:
  the_tile_mosaic_dl_elem_t(const the_tile_mosaic_t * mosaic,
			    const unsigned int & active);
  
  // virtual:
  void draw() const;
  
  // virtual:
  void update_bbox(the_bbox_t & bbox) const;
  
private:
  the_tile_mosaic_dl_elem_t();

  // a vector of display elements for the warped tiles:
  std::vector<boost::shared_ptr<the_warped_image_dl_elem_t> > tile_dl_elem_;
  
  // active tile:
  unsigned int active_;
};

//----------------------------------------------------------------
// the_tile_mosaic_t
// 
class the_tile_mosaic_t : public the_primitive_t
{
public:
  the_tile_mosaic_t();
  
  // virtual:
  the_primitive_t * clone() const
  { return new the_tile_mosaic_t(*this); }
  
  // virtual:
  const char * name() const
  { return "the_tile_mosaic_t"; }
  
  // virtual:
  bool regenerate();
  
  // the mosaic display list element:
  boost::shared_ptr<the_tile_mosaic_dl_elem_t> dl_elem_;
};


//----------------------------------------------------------------
// the_tile_warp_widget_t
// 
// The widget is represented by 12 points, which form the interpolation
// points of 4 cubic b-spline curves, which in turn are used to form
// a b-spline surface patch
// 
class the_tile_warp_widget_t : public the_primitive_t
{
public:
  the_tile_warp_widget_t();
  ~the_tile_warp_widget_t();
  
  // virtual:
  the_primitive_t * clone() const
  { return new the_tile_warp_widget_t(*this); }
  
  // virtual:
  const char * name() const
  { return "the_tile_warp_widget_t"; }
  
  // virtual:
  bool regenerate();
  
  // virtual:
  bool intersect(const the_view_volume_t & volume,
		 std::list<the_pick_data_t> & data) const;
  
  // virtual:
  the_dl_elem_t * dl_elem() const;
  
  // control point selection set accessors:
  void activate(const unsigned int & pt_index);
  void deactivate(const unsigned int & pt_index);
  bool toggle(const unsigned int & pt_index);
  bool is_active(const unsigned int & pt_index) const;
  bool has_active_points() const;
  void deactivate_all();
  
  // get boundary bsplines bounding the warp surface
  // (approximating the grid transform):
  void get_bsplines(the_bspline_geom_t * geom) const;
  
  // warp the tile by moving the active points:
  void set_anchor();
  bool move(const v3x1_t & wcs_vec);
  
  // control points expressed in the [0, 1] x [0, 1] uv space of the
  // grid transform:
  static const p2x1_t uv_[12];
  
  // control points expressed in the world space coordinate system:
  p3x1_t xy_[12];
  
  // flags indicating which of the control points are active:
  bool active_[12];
  
  // anchor of the surface approximating the grid transform:
  the_bspline_geom_t anchor_[4];
  
  // anchor surface boundary interpolation points:
  p3x1_t anchor_xy_[12];
  
  // projection of the grid transform control points on the anchor surface:
  fast_array_t<p2x1_t> anchor_uv_;
};


//----------------------------------------------------------------
// the_tile_warp_widget_ref_t
// 
// Reference to a tile transform control widget:
// 
class the_tile_warp_widget_ref_t : public the_reference_t
{
public:
  the_tile_warp_widget_ref_t(const unsigned int & id,
			     const unsigned int & pt_index);
  
  // virtual: a method for cloning references (potential memory leak):
  the_reference_t * clone() const
  { return new the_tile_warp_widget_ref_t(*this); }
  
  // virtual:
  const char * name() const
  { return "the_tile_warp_widget_ref_t"; }
  
  // virtual: Calculate the 3D value of this reference:
  bool eval(the_registry_t * registry, p3x1_t & pt) const;
  
  // virtual:
  bool move(the_registry_t * registry,
	    const the_view_mgr_t & view_mgr,
	    const p3x1_t & wcs_pt);
  
  // virtual:
  the_point_symbol_id_t symbol() const
  { return THE_SMALL_FILLED_CIRCLE_SYMBOL_E; }
  
  // virtual:
  void dump(ostream & strm, unsigned int indent = 0) const;
  
private:
  the_tile_warp_widget_ref_t();
  
public:
  // mesh coordinates of the reference:
  unsigned int pt_index_;
};


//----------------------------------------------------------------
// the_tile_warp_widget_dl_elem_t
// 
class the_tile_warp_widget_dl_elem_t : public the_dl_elem_t
{
public:
  the_tile_warp_widget_dl_elem_t(const the_tile_warp_widget_t * widget);
  ~the_tile_warp_widget_dl_elem_t();
  
  // virtual:
  void draw() const;
  
  // virtual:
  void update_bbox(the_bbox_t & bbox) const;
  
  // image bounding box:
  the_bbox_t bbox_;
  
private:
  the_tile_warp_widget_dl_elem_t();
  
  the_symbol_dl_elem_t * point_[12];
  the_bspline_geom_dl_elem_t * curve_[4];
  the_bspline_geom_t geom_[4];
};


#endif // THE_MOSAIC_PROC_HXX_
