// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : the_mosaic_proc_ui.cxx
// Author       : Pavel A. Koshevoy
// Created      : Wed Feb 14 12:12:00 MST 2007
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : UI for the mosaic layout procedure.

// local includes:
#include "the_mosaic_proc_ui.hxx"
#include "the_mosaic_proc.hxx"
#include "the_main_window.hxx"
#include "mosaic_cache.hxx"

// the includes:
#include <doc/the_registry.hxx>
#include <doc/the_reference.hxx>
#include <eh/the_view_mgr_eh.hxx>
#include <ui/the_trail.hxx>
#include <sel/the_pick_list.hxx>
#include <sel/the_pick_filter.hxx>
#include <sel/the_pick_rec.hxx>
#include <Qt/the_qt_trail.hxx>
#include <Qt/the_qt_thread_pool.hxx>
#include <utils/debug.hxx>

// system includes:
#include <fstream>
#include <memory>


//----------------------------------------------------------------
// MOSAIC_THREAD_POOL
// 
the_qt_thread_pool_t *
MOSAIC_THREAD_POOL()
{
  static the_qt_thread_pool_t * pool_ = NULL;
  if (pool_ == NULL)
  {
    pool_ = new the_qt_thread_pool_t(4);
  }
  
  return pool_;
}


//----------------------------------------------------------------
// mosaic_transaction_t
//
// Update the requested mosaic region.
// If the max_texture transaction parameter is non zero
// a corresponding display element will also be created.
// 
class mosaic_transaction_t : public the_transaction_t
{
public:
  // constructor:
  mosaic_transaction_t
  (mosaic_cache_t & cache,
   const unsigned int & prefetch_level,
   const the_aa_bbox_t & requested_region,
   const unsigned int active,
   const std::vector<image_tile_generator_t> & tile,
   const std::vector<the_color_t> & tint,
   const std::vector<base_transform_t::ConstPointer> & xform,
   const unsigned int max_texture,
   const GLenum & filter,
   const bool & rgba_compressed):
    cache_(cache),
    prefetch_level_(prefetch_level),
    requested_region_(requested_region),
    active_(active),
    tile_(tile),
    tint_(tint),
    transform_(xform),
    max_texture_(max_texture),
    filter_(filter),
    rgba_compressed_(rgba_compressed)
  {
    assert(transform_.size() == tile_.size());
  }
  
  // virtual:
  void execute(the_thread_interface_t * /* thread */)
  {
    // update the requested region:
    // cerr << "thread " << the_thread_storage().thread_id() << " executing "
    //      << this << " on cache level: " << cache_.id_ << endl;
    cache_.update(requested_region_,
		  tile_,
		  tint_,
		  transform_,
		  active_,
		  rgba_compressed_);
    
    make_dl_elem();
    
    if (prefetch_level_ < 2)
    {
      // increase the requested region for each prefetch iteration:
      mosaic_transaction_t * t =
	new mosaic_transaction_t(cache_,
				 prefetch_level_ + 1,
				 requested_region_ * 1.6,
				 active_,
				 tile_,
				 tint_,
				 transform_,
				 max_texture_,
				 filter_,
				 rgba_compressed_);
      MOSAIC_THREAD_POOL()->push_back(t, true);
    }
  }
  
  void make_dl_elem()
  {
    // make sure the shared data doesn't get deleted from some other thread:
    boost::shared_ptr<mosaic_cache_shared_data_t> hold_shared(cache_.shared_);
    
    // shortcut:
    mosaic_cache_shared_data_t & shared = *(hold_shared.get());
    the_lock_t<the_mutex_interface_t> locker(shared.mutex_);
    
    if (shared.data_.buffer_.get() == NULL)
    {
      // cut new tiles:
      if (rgba_compressed_)
      {
	shared.make_tiles(GL_UNSIGNED_BYTE,
			  GL_LUMINANCE,	// internal data format
			  GL_LUMINANCE,	// data format
			  max_texture_ / 2);
      }
      else
      {
	shared.make_tiles(GL_UNSIGNED_INT_8_8_8_8_REV,
			  GL_RGBA8,	// internal data format
			  GL_BGRA_EXT,	// data format
			  max_texture_);
      }
      
      // setup the display element:
      cache_.dl_elem_ =
	std::auto_ptr<the_dl_elem_t>
	(new mosaic_dl_elem_t(hold_shared, filter_, filter_));
    }
  }
  
  // trasaction parameters:
  mosaic_cache_t & cache_;
  const unsigned int prefetch_level_;
  const the_aa_bbox_t requested_region_;
  const unsigned int active_;
  std::vector<image_tile_generator_t> tile_;
  std::vector<the_color_t> tint_;
  std::vector<base_transform_t::ConstPointer> transform_;
  const unsigned int max_texture_;
  GLenum filter_;
  const bool rgba_compressed_;
};


//----------------------------------------------------------------
// the_mosaic_proc_eh_t::the_mosaic_proc_eh_t
// 
the_mosaic_proc_eh_t::the_mosaic_proc_eh_t(the_document_ui_t * doc_ui):
  the_document_eh_t<the_document_t>(doc_ui),
  mosaic_state_(HAS_NOTHING_E)
{}

//----------------------------------------------------------------
// the_mosaic_proc_eh_t::is_modifier_active_locally
// 
bool
the_mosaic_proc_eh_t::is_modifier_active_locally(the_eh_modifier_t mod) const
{
#if 0 
  cout << "--------------------------------------------------------" << endl;
  ::dump(THE_MOUSE);
  ::dump(THE_KEYBD);
#endif
  
  switch (mod)
  {
    case THE_EH_MOD_VIEW_ZOOM_E:
      return ((state() != THE_EH_NOT_READY_E) &&
	      THE_MOUSE.verify_pressed_btns(THE_MOUSE.m_btn()));
      
    case THE_EH_MOD_VIEW_PAN_E:
      return ((state() != THE_EH_NOT_READY_E) &&
	      THE_MOUSE.verify_pressed_btns(THE_MOUSE.l_btn()));
      
    default:
      break;
  }
  
  return the_input_device_eh_t::is_modifier_active_locally(mod);
}

//----------------------------------------------------------------
// transaction_payload_t
// 
class transaction_payload_t
{
public:
  transaction_payload_t():
    transaction_(NULL)
  {}
  
  ~transaction_payload_t()
  {
    delete transaction_;
  }
  
  void set_payload(the_transaction_t * t)
  {
    delete transaction_;
    transaction_ = t;
  }
  
  the_transaction_t * get_payload()
  {
    the_transaction_t * t = transaction_;
    transaction_ = NULL;
    return t;
  }
  
  inline bool has_payload() const
  { return transaction_ != NULL; }
  
  QMutex mutex_;
  the_transaction_t * transaction_;
};

//----------------------------------------------------------------
// rgba_done_event_t
// 
class rgba_done_event_t : public QEvent
{
public:
  rgba_done_event_t(transaction_payload_t & payload):
    QEvent(QEvent::Type(QEvent::User + 1)),
    payload_(payload)
  {}
  
  transaction_payload_t & payload_;
};

//----------------------------------------------------------------
// rgba_done_cb
// 
static void
rgba_done_cb(void * obj,
	     the_transaction_t * transaction,
	     the_transaction_t::state_t state)
{
  // cerr << transaction << " transaction " << state << endl;
  switch (state)
  {
    case the_transaction_t::PENDING_E:
    case the_transaction_t::STARTED_E:
      return;
      
    default:
      break;
  }
  
  if (state == the_transaction_t::DONE_E)
  {
    // take ownership of the payload:
    static transaction_payload_t payload;
    QMutexLocker locker(&payload.mutex_);
    
    bool had_payload = payload.has_payload();
    payload.set_payload(transaction);
    if (had_payload) return;
    
    // send a custom event to trigger the rgbaDone slot:
    the_mosaic_proc_ui_t * ui = (the_mosaic_proc_ui_t *)(obj);
    QCoreApplication::postEvent(ui, new rgba_done_event_t(payload));
  }
  else
  {
    // discard the transaction:
    delete transaction;
  }
}

//----------------------------------------------------------------
// get_tiles_at_resolution
// 
static void
get_tiles_at_resolution(const std::list<the_tile_t *> & tiles,
			const unsigned int & resolution_level,
			std::vector<image_tile_generator_t> & tile)
{
  unsigned int index = 0;
  for (std::list<the_tile_t *>::const_iterator i = tiles.begin();
       i != tiles.end(); ++i, ++index)
  {
    const the_tile_t * t = *i;
    const the_image_pyramid_t & pyramid = t->pyramid();
    if (pyramid.image_.size() <= resolution_level)
    {
      tile[index] = pyramid.image_[pyramid.image_.size() - 1];
    }
    else
    {
      tile[index] = pyramid.image_[resolution_level];
    }
  }
}

//----------------------------------------------------------------
// the_mosaic_proc_eh_t::view_handler
// 
bool
the_mosaic_proc_eh_t::view_handler(the_view_t * view)
{
  if (prefsUseOpenGLForMosaicking())
  {
    return false;
  }
  
  // FIXME: this is wrong:
  // prevent recursive calls:
  static unsigned int call_count = 0;
  if (call_count > 0) return false;
  call_count++;
  
  bool view_refreshed = false;
  
  the_main_window_t * main_window =
    dynamic_cast<the_main_window_t *>(&doc_ui());
  the_mosaic_proc_ui_t * mosaic_ui = main_window->mosaic_proc_ui();
  the_mosaic_proc_t * proc = the_mosaic_proc_eh_t::proc();
  the_view_t * view_layout = doc_ui().view(0);
  the_tile_mosaic_t * mosaic = proc->layout_prim();
  
#if 0
  cerr << "view handler: " << mosaic;
  if (mosaic != NULL) cerr << '\t' << mosaic->regenerated();
  cerr << endl;
#endif
  
  if (view == view_layout && mosaic != NULL && mosaic->regenerated())
  {
    // shortcuts:
    const unsigned int max_texture = OpenGL(0).max_texture_;
    const GLenum filter = prefsAllowLinearFiltering() ? GL_LINEAR : GL_NEAREST;
    const bool rgba_compressed = prefsUseFragmentShader();
    
    const the_view_mgr_t & view_mgr = view_layout->view_mgr();
    the_view_volume_t vol;
    view_mgr.setup_view_volume(vol);
    
    p3x1_t view_min;
    p3x1_t view_max;
    vol.scs_to_wcs(p2x1_t(0, 0), view_min);
    vol.scs_to_wcs(p2x1_t(1, 1), view_max);
    view_min[2] = 0;
    view_max[2] = 0;
    
    the_aa_bbox_t requested_region;
    requested_region << view_min << view_max;
    
    // get the tiles:
    std::list<the_tile_t *> tiles;
    unsigned int num_tiles = ::direct_dependents<the_tile_t>(proc, tiles);
    
    // allocate a vector to hold the tiles:
    std::vector<image_tile_generator_t> tile(num_tiles);
    
    // get the tile tint and transforms:
    std::vector<the_color_t> tint(num_tiles);
    std::vector<base_transform_t::ConstPointer> xform(num_tiles);
    unsigned int index = 0;
    for (std::list<the_tile_t *>::iterator i = tiles.begin();
	 i != tiles.end(); ++i, ++index)
    {
      the_tile_t * t = *i;
      the_tile_transform_t * s = first_dependent<the_tile_transform_t>(t);
      
      itk::GridTransform::Pointer itk_gt = itk::GridTransform::New();
      itk_gt->setup(s->transform_);
      xform[index] = itk_gt;
      
      if (proc->selection() == t->id())
      {
	/*
	tint[index] =
	  rgba_compressed ?
	  the_color_t(1.0, 0.5, 0.0) :
	  the_color_t::RED;
	// FIXME:
	*/
	tint[index] = the_color_t(1.0, 0.5, 0.0);
      }
      else
      {
	/*
	tint[index] =
	  rgba_compressed ?
	  the_color_t(0.0, 0.5, 1.0) :
	  the_color_t::CYAN;
	// FIXME:
	*/
	tint[index] = the_color_t(0.0, 0.5, 1.0);
      }
    }
    
    // find the highest resolution pyramid level that has
    // the requested region readily available:
    mosaic_cache_pyramid_t & pyramid = proc->layout();
    
    // FIXME:
    int num_levels = pyramid.cache_.size();
    // int num_levels = 1;
    int ready_level = num_levels;
    
#if 1
    for (int i = num_levels - 1; i > -1; i--)
    {
      mosaic_cache_t * cache = pyramid.cache_[i];
      
      // make sure this object doesn't get deleted from some other thread:
      boost::shared_ptr<mosaic_cache_shared_data_t> hold(cache->shared_);
      
      // shortcut:
      mosaic_cache_shared_data_t & shared = *(hold.get());
      
      if (!shared.is_ready(requested_region, xform, proc->selection()))
      {
	// cache not ready, use the previous ready cache level:
	break;
      }
      
      // save the new ready cache candidate:
      ready_level = i;
    }
    
    if (ready_level < num_levels)
    {
      mosaic_cache_t * cache = pyramid.cache_[ready_level];
      // cerr << "FIXME: ready level: " << ready_level << endl;
      
      // draw the readily available mosaic:
      get_tiles_at_resolution(tiles, ready_level, tile);
      mosaic_transaction_t * trans = new
	mosaic_transaction_t(*cache,
			     0,
			     requested_region,
			     proc->selection(),
			     tile,
			     tint,
			     xform,
			     max_texture,
			     filter,
			     rgba_compressed);
      trans->set_notify_cb(rgba_done_cb, mosaic_ui);
      
      // generate the display list element if necessary:
      trans->make_dl_elem();
      
      // mark the transaction as completed so it wouldn't be ignored:
      trans->notify(MOSAIC_THREAD_POOL(), the_transaction_t::DONE_E);
      
      view_refreshed = true;
    }
#endif
    
    // schedule all other levels of the pyramid to be updated:
    std::list<the_transaction_t *> schedule;
    for (int i = ready_level - 1; i > -1; i--)
    {
      // cerr << "FIXME: regenerate level " << i << endl;
      get_tiles_at_resolution(tiles, i, tile);
      mosaic_transaction_t * trans =
	new mosaic_transaction_t(*(pyramid.cache_[i]),
				 0,
				 requested_region,
				 proc->selection(),
				 tile,
				 tint,
				 xform,
				 max_texture,
				 filter,
				 rgba_compressed);
      trans->set_notify_cb(rgba_done_cb, mosaic_ui);
      schedule.push_back(trans);
    }
    
    if (!schedule.empty())
    {
      MOSAIC_THREAD_POOL()->stop_and_go(schedule, true);
    }
    
#if 0
    // prefetch -- anticipate further requests by scheduling expanded
    // region updates:
    for (int prefetch = 0; prefetch < 1; prefetch++)
    {
      // double the requested region for each prefetch iteration:
      requested_region *= 1.6;
      
      for (int i = num_levels - 1; i > -1; i--)
      {
	mosaic_cache_t * cache = pyramid.cache_[i];
	
	// make sure this object doesn't get deleted from some other thread:
	boost::shared_ptr<mosaic_cache_shared_data_t> hold(cache->shared_);
	
	// shortcut:
	mosaic_cache_shared_data_t & shared = *(hold.get());
	
	if (!shared.is_ready(requested_region, xform, proc->selection()))
	{
	  get_tiles_at_resolution(tiles, i, tile);
	  mosaic_transaction_t * trans =
	    new mosaic_transaction_t(*(pyramid.cache_[i]),
				     requested_region,
				     proc->selection(),
				     tile,
				     tint,
				     xform,
				     max_texture,
				     filter,
				     rgba_compressed);
	  trans->set_notify_cb(rgba_done_cb, mosaic_ui);
	  MOSAIC_THREAD_POOL()->push_back(trans, true);
	}
      }
    }
#endif
  }
  
  call_count--;
  return view_refreshed;
}

//----------------------------------------------------------------
// the_mosaic_proc_eh_t::processor
// 
bool
the_mosaic_proc_eh_t::
processor(the_input_device_eh_t::the_event_type_t event_type)
{
  bool processed = false;
  
  if (event_type == THE_MOUSE_EVENT_E)
  {
    processed = process_mouse();
#if 0
    if (processed)
    {
      static unsigned int i = 0;
      cerr << " mouse event accepted " << i << endl;
      i++;
    }
#endif
  }
  else if (event_type == THE_KEYBD_EVENT_E)
  {
    processed = false;
  }
  
  if (processed)
  {
    // call document sync_ui, which will in turn call proc()->regenerate():
    doc_ui().sync_ui();
  }
  
  return processed;
}

//----------------------------------------------------------------
// the_mosaic_proc_eh_t::registry
// 
the_registry_t *
the_mosaic_proc_eh_t::registry() const
{
  return proc()->registry();
}

//----------------------------------------------------------------
// the_mosaic_proc_eh_t::proc
// 
the_mosaic_proc_t *
the_mosaic_proc_eh_t::proc() const
{
  the_document_t * d = document();
  if (d == NULL) return NULL;
  return d->active_procedure<the_mosaic_proc_t>();
}

//----------------------------------------------------------------
// the_mosaic_proc_eh_t::process_mouse
// 
bool
the_mosaic_proc_eh_t::process_mouse()
{
  // shortcut to the view manager event handler:
  const the_view_mgr_eh_t * view_mgr_eh =
    dynamic_cast<const the_view_mgr_eh_t *>
    (most_recent_view()->eh_stack().back());
  
  // shortcuts:
  the_mosaic_proc_t * proc = the_mosaic_proc_eh_t::proc();
  the_registry_t * r = registry();
  bool ctl_key_is_down = THE_KEYBD.ctl().down();
  
  if ((THE_KEYBD.verify_no_pressed_keys() ||
       THE_KEYBD.verify_pressed_keys(THE_KEYBD.ctl())) &&
      THE_MOUSE.l_btn().down() &&
      THE_MOUSE.l_btn().changed() &&
      !THE_MOUSE.l_btn().moved_while_down())
  {
    the_trail_t::milestone_t milestone;
    
    the_pick_list_t curr_pick;
    if (curr_pick.pick(view(),
		       scs_pt(),
		       the_pick_filter<the_tile_warp_widget_t>(),
		       view().primitives()))
    {
      const the_pick_rec_t & pick_rec = curr_pick.front();
      the_tile_warp_widget_t * widget =
	pick_rec.data().is<the_tile_warp_widget_t>(r);
      const the_tile_warp_widget_ref_t * ref =
	pick_rec.data().ref<the_tile_warp_widget_ref_t>();
      
      bool dont_let_view_mgr_have_the_event = true;
      if (ctl_key_is_down)
      {
	bool on = widget->toggle(ref->pt_index_);
	dont_let_view_mgr_have_the_event =
	  on || !widget->has_active_points();
      }
      else
      {
	if (!widget->is_active(ref->pt_index_))
	{
	  widget->deactivate_all();
	  widget->activate(ref->pt_index_);
	}
      }
      
      mosaic_state_ = widget->has_active_points() ? HAS_POINTS_E : HAS_TILE_E;
      
      widget->set_anchor();
      widget->request_regeneration();
      doc_ui().sync_ui();
      return true;
    }
    else if (curr_pick.pick(view(),
			    scs_pt(),
			    the_pick_filter<the_tile_transform_t>(),
			    view().primitives()))
    {
      mosaic_state_ = HAS_TILE_E;
      
      if (proc->has_selection())
      {
	the_tile_t * tile = r->elem<the_tile_t>(proc->selection());
	the_tile_transform_t * tile_transform =
	  first_dependent<the_tile_transform_t>(tile);
	
	if (curr_pick.contains(tile_transform->id()))
	{
	  // same tile selected:
	  tile_transform->set_anchor();
	  
	  the_tile_warp_widget_t * widget =
	    first_dependent<the_tile_warp_widget_t>(tile_transform);
	  widget->deactivate_all();
	  doc_ui().sync_ui();
	  return true;
	}
      }
      
      // new tile selected:
      const the_pick_rec_t & first = curr_pick.front();
      the_tile_transform_t * tile_transform =
	first.data().is<the_tile_transform_t>(r);
      
      proc->select(tile_transform->tile()->id());
      tile_transform->set_anchor();
      tile_transform->request_regeneration();
      doc_ui().sync_ui();
      return true;
    }
    else
    {
      // nothing selected:
      if (proc->has_selection())
      {
	proc->unselect();
	doc_ui().sync_ui();
      }
      
      mosaic_state_ = HAS_NOTHING_E;
      return false;
    }
  }
  else if (has_drag() &&
	   (THE_KEYBD.verify_no_pressed_keys() ||
	    THE_KEYBD.verify_pressed_keys(THE_KEYBD.ctl())) &&
	   THE_MOUSE.verify_pressed_btns(THE_MOUSE.l_btn()) &&
	   proc->has_selection())
  {
    the_trail_t::milestone_t milestone;
    
    the_tile_t * tile =
      r->elem<the_tile_t>(proc->selection());
    
    the_tile_transform_t * tile_transform =
      first_dependent<the_tile_transform_t>(tile);
    
    the_tile_warp_widget_t * widget =
      first_dependent<the_tile_warp_widget_t>(tile_transform);
    
    if (the_input_device_eh_t::state() == THE_EH_1ST_DRAG_E &&
	view_mgr_eh->view_mgr_eh_state() == THE_VIEW_MGR_EH_HAS_NOTHING_E)
    {
      if (widget->has_active_points())
      {
	mosaic_state_ = MOVE_POINTS_E;
	doc_so().save_undo_record();
      }
      else if (ctl_key_is_down)
      {
	mosaic_state_ = WARP_MESH_LOCALLY_E;
	doc_so().save_undo_record();
      }
      else
      {
	mosaic_state_ = MOVE_TILE_E;
	doc_so().save_undo_record();
      }
    }
    
    if (mosaic_state_ == MOVE_TILE_E)
    {
      tile_transform->move(drag_wcs());
    }
    else if (mosaic_state_ == MOVE_POINTS_E)
    {
      widget->move(drag_wcs());
    }
    else if (mosaic_state_ == WARP_MESH_LOCALLY_E)
    {
      float radius = view().view_mgr().wcs_units_per_millimeter() * 50.0f;
      tile_transform->warp(a_wcs(), drag_wcs(), radius);
    }
    else
    {
      return false;
    }
    
    draw_scs_drag_vec();
    return true;
  }
  else if (proc->has_selection() &&
	   (THE_KEYBD.verify_no_pressed_keys() ||
	    THE_KEYBD.verify_pressed_keys(THE_KEYBD.ctl())) &&
	   state() == THE_EH_END_DRAG_E &&
	   (mosaic_state_ == MOVE_POINTS_E ||
	    mosaic_state_ == MOVE_TILE_E))
  {
    the_tile_t * tile =
      r->elem<the_tile_t>(proc->selection());
    
    the_tile_transform_t * tile_transform =
      first_dependent<the_tile_transform_t>(tile);
    
    the_tile_warp_widget_t * widget =
      first_dependent<the_tile_warp_widget_t>(tile_transform);
    
    mosaic_state_ = widget->has_active_points() ? HAS_POINTS_E : HAS_TILE_E;
    
    tile_transform->request_regeneration();
    
    doc_ui().sync_ui();
    return false;
  }
  
  // reject all other events:
  return false;
}


//----------------------------------------------------------------
// the_mosaic_proc_ui_t::the_mosaic_proc_ui_t
// 
the_mosaic_proc_ui_t::the_mosaic_proc_ui_t(the_main_window_t * doc_ui):
  the_procedure_ui_t(*doc_ui),
  eh_(doc_ui),
  repaint_timer_()
{
  repaint_timer_.setSingleShot(true);
  /*
  connect(MOSAIC_THREAD_POOL(),
	  SIGNAL(transaction_finished(the_transaction_t *)),
	  this, SLOT(rgbaDone(the_transaction_t *)));
  */
  // avoid the overhead of the thread start/stop sequence -- keep the
  // thread running while putting it to sleep when it has nothing to do:
  // FIXME:
  MOSAIC_THREAD_POOL()->set_idle_sleep_duration(true, 100000);
  
  // attach the event handler to the document views:
  std::vector<the_view_t *> & views = doc_ui_.views();
  unsigned int num_views = views.size();
  for (unsigned int i = 0; i < num_views; i++)
  {
    eh_.attach_view(views[i]);
  }
}

//----------------------------------------------------------------
// the_mosaic_proc_ui_t::~the_mosaic_proc_ui_t
// 
the_mosaic_proc_ui_t::~the_mosaic_proc_ui_t()
{
  MOSAIC_THREAD_POOL()->stop();
  MOSAIC_THREAD_POOL()->wait();
  /*
  disconnect(MOSAIC_THREAD_POOL(),
	     SIGNAL(transaction_finished(the_transaction_t *)),
	     this, SLOT(rgbaDone(the_transaction_t *)));
  */
  eh_.uninstall();
  
  std::vector<the_view_t *> & views = doc_ui_.views();
  unsigned int num_views = views.size();
  for (unsigned int i = 0; i < num_views; i++)
  {
    eh_.detach_view(views[i]);
  }
}

//----------------------------------------------------------------
// the_mosaic_proc_ui_t::install
// 
void
the_mosaic_proc_ui_t::install()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  if (installed()) return;
  
  // make a new procedure:
  document()->add_proc(new the_mosaic_proc_t());
  the_procedure_ui_t::install();
  eh_.install();
  
  connect(&repaint_timer_, SIGNAL(timeout()),
	  this, SLOT(repaintMosaic()));
}

//----------------------------------------------------------------
// the_mosaic_proc_ui_t::uninstall
// 
void
the_mosaic_proc_ui_t::uninstall()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  if (!installed()) return;
  
  disconnect(&repaint_timer_, SIGNAL(timeout()),
	     this, SLOT(repaintMosaic()));
  
  the_procedure_ui_t::uninstall();
  eh_.uninstall();
}

//----------------------------------------------------------------
// the_mosaic_proc_ui_t::sync_ui
// 
void
the_mosaic_proc_ui_t::sync_ui()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  if (proc() == NULL) return;
  proc()->regenerate();
}

//----------------------------------------------------------------
// the_mosaic_proc_ui_t::proc
// 
the_mosaic_proc_t *
the_mosaic_proc_ui_t::proc()
{
  the_document_t * d = document();
  if (d == NULL) return NULL;
  return d->active_procedure<the_mosaic_proc_t>();
}

//----------------------------------------------------------------
// the_mosaic_proc_ui_t::rgbaDone
// 
void
the_mosaic_proc_ui_t::rgbaDone(the_transaction_t * t)
{
  the_mosaic_proc_t * mosaic = proc();
  if (t->done() && mosaic != NULL)
  {
    mosaic_transaction_t * transaction =
      dynamic_cast<mosaic_transaction_t *>(t);
    
    if (transaction != NULL)
    {
      // cerr << "rgbaDone: " << transaction->cache_.id_ << endl;
      
      // update the procedure:
      mosaic->layout_ready(transaction->cache_.id_);
      
      if (!repaint_timer_.isActive())
      {
	// buffer the repaints in order to avoid the jerky
	// display update and flashing due to the multi-resolution
	// pixel scale difference:
	repaint_timer_.start(41);
      }
    }
  }
  
  delete t;
}

//----------------------------------------------------------------
// the_mosaic_proc_ui_t::repaintMosaic
// 
void
the_mosaic_proc_ui_t::repaintMosaic()
{
  the_view_t * view_layout = doc_ui().view(0);
  view_layout->refresh();
}

//----------------------------------------------------------------
// the_mosaic_proc_ui_t::customEvent
// 
void
the_mosaic_proc_ui_t::customEvent(QEvent * e)
{
  // cerr << "received a custom event: " << int(e->type()) << endl;
  if (e->type() == QEvent::User + 1)
  {
    e->accept();
    rgba_done_event_t * event = (rgba_done_event_t *)(e);
    the_transaction_t * t = NULL;
    {
      transaction_payload_t & payload = event->payload_;
      QMutexLocker locker(&payload.mutex_);
      t = payload.get_payload();
    }
    rgbaDone(t);
  }
}
