// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : the_tile_transaction.cxx
// Author       : Pavel A. Koshevoy
// Created      : Fri Feb 23 12:45:00 MST 2007
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Load a mosaic tile, generate a multiresolutions pyramid
//                and OpenGL textures.

// local includes:
#include "the_tile_transaction.hxx"

// the includes:
#include <image/la_pixel_converter.hxx>
#include <utils/the_exception.hxx>

// Qt includes:
#include <QFileInfo>
#include <QString>


//----------------------------------------------------------------
// the_tile_transaction_t::the_tile_transaction_t
// 
the_tile_transaction_t::
the_tile_transaction_t(unsigned int tile_id,
		       boost::shared_ptr<the_image_pyramid_t> & pyramid,
		       unsigned int shrink_factor,
		       unsigned int max_texture,
		       bool use_fragment_shader,
		       bool use_std_mask,
		       bool use_luminance_textures):
  the_transaction_t(),
  tile_id_(tile_id),
  pyramid_(pyramid),
  shrink_factor_(shrink_factor),
  max_texture_(max_texture),
  use_fragment_shader_(use_fragment_shader),
  use_std_mask_(use_std_mask),
  use_luminance_textures_(use_luminance_textures)
{}

//----------------------------------------------------------------
// the_tile_transaction_t::execute
// 
void
the_tile_transaction_t::execute(the_thread_interface_t * thread)
{
  the_terminator_t terminator("the_tile_transaction_t");
  
  // shortcut:
  the_image_pyramid_t & pyramid = *(pyramid_.get());
  
  // load the image:
  pyramid.state_ = the_image_pyramid_t::LOADING_E;
  
  GLenum data_type = GL_UNSIGNED_BYTE;
  GLenum format_internal = 0;
  GLenum format = 0;
  size_t dst_bytes_per_pixel = 0;
  
  if (use_fragment_shader_)
  {
    format = GL_LUMINANCE;
    format_internal = GL_LUMINANCE;
    dst_bytes_per_pixel = 1;
  }
  else
  {
    format = GL_LUMINANCE_ALPHA;
    format_internal =
      use_luminance_textures_ ?
      GL_LUMINANCE_ALPHA :
      GL_RGBA8;
    
    dst_bytes_per_pixel = 2;
  }
  
  unsigned int num_levels = pyramid.itk_image_.size();
  bool reload =
    (num_levels == 0) ||
    (pyramid.shrink_factor_ != shrink_factor_) ||
    (pyramid.use_std_mask_ && !use_std_mask_);
  
  if (reload)
  {
    const char * fn = pyramid.filename_.text();
    const QFileInfo info(fn);
    
    QString filepath(info.absoluteFilePath());
    QImage qt_image(filepath);
    if (qt_image.isNull())
    {
      // FIXME: here is where I could try to load it again using ITK:
      pyramid.state_ = the_image_pyramid_t::FAILED_E;
      return;
    }
    
    switch (qt_image.format())
    {
      case QImage::Format_MonoLSB:
      case QImage::Format_Mono:
	qt_image = qt_image.convertToFormat(QImage::Format_Indexed8);
	break;
	
      default:
	break;
    }
    
    if (shrink_factor_ > 1)
    {
      // downsample the image:
      qt_image = qt_image.scaled(qt_image.width() / shrink_factor_,
				 qt_image.height() / shrink_factor_,
				 Qt::IgnoreAspectRatio,
				 Qt::SmoothTransformation);
      pyramid.shrink_factor_ = shrink_factor_;
    }
    
    // calculate the number of levels of the multi-res pyramid:
#if 0
    num_levels = 1;
#else
    static const double LOG_2 = log(2.0);
    
    double w_max = double(qt_image.width());
    double h_max = double(qt_image.height());
    
    const double x_discrete_levels = floor(0.5 + w_max);
    const double y_discrete_levels = floor(0.5 + h_max);
    const double discrete_levels = std::max(x_discrete_levels,
					    y_discrete_levels);
    const double min_discrete_levels = 64.0;
    
    const double divisions =
      std::max(log(double(discrete_levels) /
		   double(min_discrete_levels)) / LOG_2,
	       0.0);
    
    num_levels = 1 + (unsigned int)(floor(divisions));
#endif
    
    // allocate the pyramid images:
    pyramid.itk_image_.resize(num_levels);
    pyramid.itk_mask_.resize(num_levels);
    pyramid.image_.resize(num_levels);
    
    // setup the first pyramid level (copy the pixel data):
    unsigned int w0 = qt_image.width();
    unsigned int h0 = qt_image.height();
    
    blab(thread, tr("generating image, mask and texture pyramids"));
    
    native_image_t::SizeType sz;
    sz[0] = w0;
    sz[1] = h0;
    
    native_image_t::SpacingType sp;
    sp[0] = pyramid.spacing_x_ * double(shrink_factor_);
    sp[1] = pyramid.spacing_y_ * double(shrink_factor_);
    
    pyramid.itk_image_[0] = make_image<native_image_t>(sp, sz, 0);
    pyramid.itk_mask_[0] = std_mask<native_image_t>(pyramid.itk_image_[0],
						    use_std_mask_);
    
    // convert QImage into an itk::Image
    typedef itk::ImageRegionIteratorWithIndex<native_image_t> itex_t;
    itex_t itex(pyramid.itk_image_[0],
		pyramid.itk_image_[0]->GetLargestPossibleRegion());
    for (itex.GoToBegin(); !itex.IsAtEnd(); ++itex)
    {
      // make sure there hasn't been an interrupt:
      terminator.terminate_on_request();
      
      native_image_t::IndexType ix = itex.GetIndex();
      QRgb argb = qt_image.pixel(ix[0], ix[1]);
      
      itex.Set(qGray(argb));
      if (use_std_mask_) continue;
      
      int alpha = qAlpha(argb);
      pyramid.itk_mask_[0]->SetPixel(ix, alpha);
      if (alpha == 0) itex.Set(0);
    }
    
    // setup the remaining pyramid levels:
    for (unsigned int i = 1; i < num_levels; i++)
    {
      // make sure there hasn't been an interrupt:
      terminator.terminate_on_request();
      native_image_t::Pointer & prev = pyramid.itk_image_[i - 1];
      native_image_t::Pointer & next = pyramid.itk_image_[i];
      native_image_t::Pointer & prev_mask = pyramid.itk_mask_[i - 1];
      native_image_t::Pointer & next_mask = pyramid.itk_mask_[i];
      native_image_t::SpacingType prev_sp = prev->GetSpacing();
      
      // setup the next resolution level image:
      sz[0] = std::max(1u, (unsigned int)(sz[0] / 2));
      sz[1] = std::max(1u, (unsigned int)(sz[1] / 2));
      sp[0] *= 2;
      sp[1] *= 2;
      next = make_image<native_image_t>(prev->GetOrigin(), sp, sz);
      
      // setup the mask:
      if (use_std_mask_)
      {
	next_mask = std_mask<native_image_t>(next, true);
      }
      else
      {
	next_mask = make_image<mask_t>(prev->GetOrigin(), sp, sz);
      }
      
      // downsample the image:
      typedef itk::LinearInterpolateImageFunction<native_image_t, double>
	tile_interpolator_t;
      tile_interpolator_t::Pointer tile_interpolator =
	tile_interpolator_t::New();
      tile_interpolator->SetInputImage(prev);
      
      typedef itk::LinearInterpolateImageFunction<mask_t, double>
	mask_interpolator_t;
      mask_interpolator_t::Pointer mask_interpolator =
	mask_interpolator_t::New();
      mask_interpolator->SetInputImage(prev_mask);
      
      itex_t itex(next, next->GetLargestPossibleRegion());
      native_image_t::IndexType index;
      native_image_t::PointType point;
      native_image_t::PointType tmp_pt;
      
      for (itex.GoToBegin(); !itex.IsAtEnd(); ++itex)
      {
	// make sure there hasn't been an interrupt:
	terminator.terminate_on_request();
	
	index = itex.GetIndex();
	next->TransformIndexToPhysicalPoint(index, point);
	
	double p = 0.0;
	double a = 0.0;
	double n = 0.0;
	
	for (unsigned int y = 0; y < 2; y++)
	{
	  tmp_pt[1] = point[1] + prev_sp[1] * double(y);
	  for (unsigned int x = 0; x < 2; x++)
	  {
	    tmp_pt[0] = point[0] + prev_sp[0] * double(x);
	    
	    if (tile_interpolator->IsInsideBuffer(tmp_pt))
	    {
	      p += tile_interpolator->Evaluate(tmp_pt);
	      a += mask_interpolator->Evaluate(tmp_pt);
	      n += 1.0;
	    }
	  }
	}
	
	if (n != 0)
	{
	  p = std::min(255.0, std::max(0.0, p / n));
	  a = std::min(255.0, std::max(0.0, a / n));
	}
	
	itex.Set((native_image_t::PixelType)(p));
	if (use_std_mask_) continue;
	
	next_mask->SetPixel(index, (mask_t::PixelType)(a));
	if (a < 1.0) itex.Set(0);
      }
    }
  }
  else if (use_std_mask_)
  {
    for (unsigned int i = 0; i < num_levels; i++)
    {
      // it's trivial to switch from an unkown mask to a standard mask:
      pyramid.itk_mask_[i] =
	std_mask<native_image_t>(pyramid.itk_image_[i],
				 use_std_mask_);
    }
  }
  
  if (reload || use_std_mask_ != pyramid.use_std_mask_)
  {
    // reset the thumbnail:
    pyramid.thumbnail_ = QImage();
  }
  
  for (unsigned int i = 0; i < num_levels; i++)
  {
    // setup the tile generator:
    convert_tile<native_image_t, mask_t>(pyramid.image_[i],
					 pyramid.itk_image_[i],
					 pyramid.itk_mask_[i],
					 use_fragment_shader_);
    pyramid.image_[i].make_tiles(data_type,
				 format_internal,
				 format,
				 max_texture_);
  }
  
  pyramid.use_std_mask_ = use_std_mask_;
  pyramid.state_ = the_image_pyramid_t::READY_E;
}
