// -*- Mode: c++; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*  
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// File         : IRConnection.cxx
// Author       : Joel Spaltenstein
// Created      : 2008/03/11
// Copyright    : (C) 2008 University of Utah
// License      : GPLv2
// Description  :

// system includes:
#include <stdio.h>

// ITK includes:
#include <itkRegionOfInterestImageFilter.h>

// the includes:
#include <itk/fft_common.hxx>
#include <itk/mosaic_layout_common.hxx>

// local includes:
#include "IRConnection.hxx"
#include "IRImageLoader.hxx"
#include "IRTransform.hxx"


//----------------------------------------------------------------
// IRConnection::IRConnection
// 
IRConnection::IRConnection(IRTransform* fristTransform,
						   IRTransform* secondTransform) : 
	_firstTransformation(fristTransform),
	_secondTransformation(secondTransform),
	_matchValue(0.0)
{	
	printf("building an IRConnection between \n");
	_firstTransformation->printImageID();
	printf("\n");
	_secondTransformation->printImageID();
	printf("\n");
	return;
}

//----------------------------------------------------------------
// IRConnection::~IRConnection
// 
IRConnection::~IRConnection()
{
	return;
}

//----------------------------------------------------------------
// IRConnection::tensionOnTransformation
// 
vec2d_t
IRConnection::tensionOnTransformation(IRTransform * transform)
{
	vec2d_t tension;
	if (transform != _firstTransformation &&
		transform != _secondTransformation)
	{
		printf("IRConnection::tensionOnTransformation: "
			   "WTF what transform are you talking about!\n");
		tension[0] = 0;
		tension[1] = 0;
		return tension;
	}
	
	tension = (_secondTransformation->position() -
			   _firstTransformation->position()) - _idealTransormationOffset; 
	
	if (transform == _secondTransformation)
	{
		tension[0] = -1*tension[0];
		tension[1] = -1*tension[1];
	}
	
	tension[0] = tension[0] * (_matchValue*_matchValue);
	tension[1] = tension[1] * (_matchValue*_matchValue);
	return tension;
}

//----------------------------------------------------------------
// IRConnection::idealTransormationOffset
// 
vec2d_t
IRConnection::idealTransormationOffset()
{
	return _idealTransormationOffset;
}

//----------------------------------------------------------------
// IRConnection::findIdealOffset
// 
int
IRConnection::findIdealOffset()
{
	translate_transform_t::Pointer translationTransform;
	mask_t::ConstPointer mask1, mask2;
	
	image_t::Pointer image1 = IRImageLoader::sharedImageLoader()->
		getImage(_firstTransformation->imageID());
	image_t::Pointer image2 = IRImageLoader::sharedImageLoader()->
		getImage(_secondTransformation->imageID());
	
	vec2d_t tension;
	int match_attempt = 0;
	bool match_success = false;
	double match_value = 0;
	
	while (match_success == false && match_attempt < 2)
	{
		switch (match_attempt)
		{
			case 0:
			{
				match_value =
					match_one_pair<image_t, mask_t>(*null_log(),
													true, // images_were_resampled
													false, // use_std_mask,
													image1, // fi,
													image2, // mi,
													mask1, // fi_mask,
													mask2, // mi_mask,
													.02, // overlap_min,
													.70, // overlap_max,
													translationTransform // ti
													);
				break;
			}
			
			case 1:
			default:
			{
				translate_transform_t::Pointer claheTranslationTransform;
				image2 = smooth<image_t>(image2, 2 /*sigma*/);
				image1 = CLAHE<image_t>(image1,
										32 /*nx*/,
										32 /*ny*/,
										3 /*max_slope*/,
										256 /*bins*/,
										0 /*new_min*/,
										255 /*new_max*/,
										mask1);
				
				image2 = CLAHE<image_t>(image2,
										32 /*nx*/,
										32 /*ny*/,
										3 /*max_slope*/,
										256 /*bins*/,
										0 /*new_min*/,
										255 /*new_max*/,
										mask1);
				
				match_value = match_one_pair<image_t, mask_t>
					(*null_log(),
					 true, // resampled
					 false, // use_std_mask,
					 image1, // fi,
					 image2, // mi,
					 mask1, // fi_mask,
					 mask2, // mi_mask,
					 .05, // overlap_min,
					 .70, // overlap_max,
					 claheTranslationTransform // ti
					 );
				
				translationTransform = claheTranslationTransform;
				break;
			}
		}
		
		if (!translationTransform.IsNull())
		{
			_idealTransormationOffset[0] = -1.0 * translationTransform->
				GetOffset()[0];
			_idealTransormationOffset[1] = -1.0 * translationTransform->
				GetOffset()[1];
			_matchValue = match_value;
			tension = tensionOnTransformation(_firstTransformation);
		}
		
		if (translationTransform.IsNull() ||
			(translationTransform->GetOffset()[0] == 0.0f &&
			 translationTransform->GetOffset()[1] == 0.0f))
		{
			printf("could not build a transform ");
			if (match_attempt == 1)
			{
				printf("using clahe ");
			}
			
			printf("between ");
			_firstTransformation->printImageID();
			printf(" and ");
			_secondTransformation->printImageID();
			printf("\n");
			match_success = false;
		}
		else if (sqrtf(tension*tension) > ((_firstTransformation->size()[0] +
											_firstTransformation->size()[1]) /
										   (2.0 * 4.0)))
		{
			// if the distance the image is moved is more than
			// a quarter of the average of the two sides
			printf("a displacement between ");
			_firstTransformation->printImageID();
			printf(" and ");
			_secondTransformation->printImageID();
			if (match_attempt == 1)
			{
				printf(" using clahe");
			}
			
			printf(" of %f, %f is too great\n",
				   _idealTransormationOffset[0],
				   _idealTransormationOffset[1]);
			match_success = false;
		}
		else
		{
			printf("the offset between ");
			_firstTransformation->printImageID();
			printf(" and ");
			_secondTransformation->printImageID();
			printf(" is %f, %f with a match value of %f\n",
				   _idealTransormationOffset[0],
				   _idealTransormationOffset[1],
				   match_value);
			match_success = true;
		}
	
		match_attempt++;
	}
	
	if (!match_success)
	{
		printf("gave up on finding a connection between ");
		_firstTransformation->printImageID();
		printf(" and ");
		_secondTransformation->printImageID();
		printf("\n");
		return -1;
	}

	return 0;
}
