// -*- Mode: c++; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*  
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// File         : IRImageLoader.cxx
// Author       : Joel Spaltenstein
// Created      : 2008/02/21
// Copyright    : (C) 2008 University of Utah
// License      : GPLv2
// Description  :

// ITK includes:
#include <itkImportImageFilter.h>

// the includes:
#include <itk/itkLegendrePolynomialTransform.h>
#include <utils/the_utils.hxx>

// local includes:
#include "IRImageLoader.hxx"


//----------------------------------------------------------------
// __sharedIRImageLoader
// 
static IRImageLoader * __sharedIRImageLoader = NULL;

//----------------------------------------------------------------
// __IRImageLoaderSharedImageLoaderLock
// 
static itk::SimpleMutexLock __IRImageLoaderSharedImageLoaderLock;

//----------------------------------------------------------------
// IRImageLoader::sharedImageLoader
// 
IRImageLoader *
IRImageLoader::sharedImageLoader()
{
    __IRImageLoaderSharedImageLoaderLock.Lock();
	if (__sharedIRImageLoader == NULL)
	{
		__sharedIRImageLoader = new IRImageLoader();
	}
	
    __IRImageLoaderSharedImageLoaderLock.Unlock();
	return __sharedIRImageLoader;
}

//----------------------------------------------------------------
// IRImageLoader::IRImageLoader
// 
IRImageLoader::IRImageLoader() :
	_shrinkFactor(1),
	_pixelSpacing(1.0)
{
	return;
}

//----------------------------------------------------------------
// IRImageLoader::~IRImageLoader
// 
IRImageLoader::~IRImageLoader()
{
	return;
}

//----------------------------------------------------------------
// __IRImageLoaderImageAcessLock
// 
static itk::SimpleMutexLock __IRImageLoaderImageAcessLock;

//----------------------------------------------------------------
// IRImageLoader::setTransformList
// 
void
IRImageLoader::setTransformList(const IRTransformVector & transforms,
								const TheTextVector & imageIDs)
{
	__IRImageLoaderImageAcessLock.Lock();
	
	IRTransformVector::const_iterator transformIter = transforms.begin();
	TheTextVector::const_iterator imageIDIter = imageIDs.begin();
	for (; transformIter != transforms.end(); transformIter++, imageIDIter++)
	{
		_transformations[*imageIDIter] = *transformIter;
	}
	_images = imageMap();
	
	__IRImageLoaderImageAcessLock.Unlock();
}

//----------------------------------------------------------------
// IRImageLoader::setShrinkFactor
// 
void
IRImageLoader::setShrinkFactor(unsigned int shrinkFactor)
{
	_shrinkFactor = shrinkFactor;
}

//----------------------------------------------------------------
// IRImageLoader::shrinkFactor
// 
unsigned int
IRImageLoader::shrinkFactor()
{
	return _shrinkFactor;
}

//----------------------------------------------------------------
// IRImageLoader::setPixelSpacing
// 
void
IRImageLoader::setPixelSpacing(double pixelSpacing)
{
	_pixelSpacing = pixelSpacing;
}

//----------------------------------------------------------------
// IRImageLoader::pixelSpacing
// 
double
IRImageLoader::pixelSpacing()
{
	return _pixelSpacing;
}

//----------------------------------------------------------------
// IRImageLoader::getImageSize
// 
vec2d_t
IRImageLoader::getImageSize(const the_text_t& imageID,
							base_transform_t::Pointer transform)
{
	vec2d_t imageSize;
	itk::LegendrePolynomialTransform<double, 1>::Pointer legendrePointer;
	legendrePointer =
		dynamic_cast<itk::LegendrePolynomialTransform<double, 1> *>
		(transform.GetPointer());
	
	
	if (legendrePointer.IsNotNull())
	{
		imageSize[0] = legendrePointer->GetFixedParameters()[2] * 2.0;
		imageSize[1] = legendrePointer->GetFixedParameters()[3] * 2.0;
	}
	else
	{
		// since this version does not rescale the image it is much faster
		image_t::Pointer image = getFullResImage(imageID);
		
		// first find the size of the tile
		pnt2d_t bbox_min;
		pnt2d_t bbox_max;
		calc_image_bbox<image_t>(image, bbox_min, bbox_max);
		imageSize = bbox_max - bbox_min;
		imageSize *= _pixelSpacing;
	}
	
	return imageSize;
}

//----------------------------------------------------------------
// IRImageLoader::getFullResImage
// 
// this does not cache the image
// 
image_t::Pointer
IRImageLoader::getFullResImage(const the_text_t& imageID)
{
	return std_tile<image_t>(imageID,
							 1 /*shrink factor*/,
							 1 /*pixel spacing*/);
}

//----------------------------------------------------------------
// IRImageLoader::getImage
// 
image_t::Pointer
IRImageLoader::getImage(const the_text_t& imageID)
{
	imageCheckLoad(imageID);

	__IRImageLoaderImageAcessLock.Lock();
	image_t::Pointer image = _images[imageID];
	__IRImageLoaderImageAcessLock.Unlock();

	return image;
}

//----------------------------------------------------------------
// IRImageLoader::imageCheckLoad
// 
void
IRImageLoader::imageCheckLoad(const the_text_t& imageID)
{
	__IRImageLoaderImageAcessLock.Lock();
	
	if (_images.find(imageID) != _images.end())
	{
		while (_images[imageID].IsNull())
		{
			// if another thread is already loading this same image,
			// spin until it is loaded
			__IRImageLoaderImageAcessLock.Unlock();
			sleep_msec(1000);
			__IRImageLoaderImageAcessLock.Lock();
		}
		
		__IRImageLoaderImageAcessLock.Unlock();
		return;
	}
	
	_images[imageID] = NULL;
	__IRImageLoaderImageAcessLock.Unlock();

	image_t::Pointer image = std_tile<image_t>(imageID,
											   _shrinkFactor /*shrink factor*/,
											   _pixelSpacing /*pixel spacing*/);
	
	__IRImageLoaderImageAcessLock.Lock();
	_images[imageID] = image;
	__IRImageLoaderImageAcessLock.Unlock();
}
