// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : extrema.hxx
// Author       : Pavel A. Koshevoy
// Created      : 2006/04/04 11:10
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : SIFT key extrema point and descriptor vector classes.

#ifndef EXTREMA_HXX_
#define EXTREMA_HXX_

// system includes:
#include <vector>

// ITK includes:
#include <itkPoint.h>


//----------------------------------------------------------------
// KEY_SIZE
// 
static const unsigned int KEY_SIZE = 128;


//----------------------------------------------------------------
// extrema_t
// 
class extrema_t
{
public:
  
  // shorthand notation for 2D points:
  typedef itk::Point<double, 2> pnt2d_t;
  
  // local pixel coordinates (non-physical) within the octave image:
  pnt2d_t pixel_coords_;
  
  // coordinates in the physical space of the local image:
  pnt2d_t local_coords_;
  
  // physical coordinates within the target space:
  mutable pnt2d_t target_coords_;
  
  // FIXME: extrema mass (cluster mass normalized by cluster area):
  double mass_;
  
  // point ids:
  unsigned int pyramid_;
  unsigned int octave_;
  unsigned int scale_;
};

//----------------------------------------------------------------
// descriptor_t
// 
class descriptor_t
{
public:
  descriptor_t(extrema_t * extrema = NULL):
    extrema_(extrema)
  {}
  
  descriptor_t(const descriptor_t & key)
  { *this = key; }
  
  descriptor_t & operator = (const descriptor_t & key)
  {
    if (this == &key) return *this;
    
    extrema_ = key.extrema_;
    local_orientation_ = key.local_orientation_;
    target_orientation_ = key.target_orientation_;
    descriptor_ = key.descriptor_;
    
    return *this;
  }
  
  // reference to the extrema point:
  extrema_t * extrema_;
  
  // coordinate system orientation angle expressed in the local image space:
  double local_orientation_;
  
  // orientation angle in the target image space:
  mutable double target_orientation_;
  
  // the descriptor vector:
  std::vector<double> descriptor_;
};

//----------------------------------------------------------------
// ext_wrapper_t
// 
class ext_wrapper_t
{
public:
  ext_wrapper_t():
    key_(NULL)
  {}
  
  // vector-like accessor to the extrema target space coordinates:
  inline const double & operator [] (const unsigned int & index) const
  { return key_->extrema_->target_coords_[index]; }
  
  // pointer to the key:
  const descriptor_t * key_;
};

//----------------------------------------------------------------
// key_wrapper_t
// 
class key_wrapper_t
{
public:
  key_wrapper_t():
    key_(NULL)
  {}
  
  // vector-like accessor to the key descriptor:
  inline const double & operator [] (const unsigned int & index) const
  { return key_->descriptor_[index]; }
  
  // pointer to the key:
  const descriptor_t * key_;
};


#endif // EXTREMA_HXX_
