// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : ir-assemble-rgb-movie.cxx
// Author       : Pavel A. Koshevoy
// Created      : 2006/11/20 22:48
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Assemble a sequence of colored mosaic images
//                from a progressive sequence of .mosaic file.
//                The sequence of mosaic files is generated by ir-refine.

// the includes:
#include <itk/common.hxx>
#include <itk/draw.hxx>
#include <itk/itkImageMosaicVarianceMetric.h>
#include <utils/the_dynamic_array.hxx>
#include <thread/the_boost_mutex.hxx>
#include <thread/the_boost_thread.hxx>

// ITK includes:
#include <itkRGBPixel.h>

// system includes:
#include <math.h>


//----------------------------------------------------------------
// load_mosaic
// 
template <typename T>
bool
load_mosaic(const char * fn_mosaic,
	    const unsigned int & shrink_factor,
	    const double & clahe_slope,
	    const bool & setup_images_and_masks,
	    std::vector<base_transform_t::Pointer> & transform,
	    std::vector<typename T::Pointer> & image,
	    std::vector<mask_t::ConstPointer> & mask)
{
  cout << "    reading " << fn_mosaic;
  
  std::fstream fin;
  fin.open(fn_mosaic, ios::in);
  if (!fin.is_open())
  {
    cout << ", failed...." << endl;
    return false;
  }
  cout << endl;
  
  std::list<the_text_t> fn_tiles;
  
  double pixel_spacing = 1.0;
  bool use_std_mask = false;
  load_mosaic<base_transform_t>(fin,
				pixel_spacing,
				use_std_mask,
				fn_tiles,
				transform);
  fin.close();
  
  unsigned int num_images = transform.size();
  assert(num_images > 1);
  assert(pixel_spacing != 0);
  
  if (!setup_images_and_masks)
  {
    assert(num_images == image.size());
    return true;
  }
  
  image.resize(num_images);
  mask.resize(num_images);
  
  unsigned int i = 0;
  for (std::list<the_text_t>::const_iterator iter = fn_tiles.begin();
       iter != fn_tiles.end(); ++iter, i++)
  {
    const the_text_t & fn_image = *iter;
    
    // read the image:
    cout << setw(3) << i << " ";
    image[i] = std_tile<T>(fn_image, shrink_factor, pixel_spacing);
    
    // generate a standard tile mask:
    if (use_std_mask)
    {
      mask[i] = std_mask<T>(image[i]);
    }
    
    if (clahe_slope > 1.0)
    {
      // enhance image contrast:
      image[i] = CLAHE<T>(image[i],
			  255,
			  255,
			  clahe_slope,
			  256,
			  0.0,
			  255.0,
			  mask[i]);
    }
  }
  
  return true;
}

//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL)
{
  cerr << "USAGE: " << argv[0] << " "
       << "[-sh shrink_factor] "
       << "[-clahe slope] "
       << "[[-variance] | [-feathering [none | blend | binary]]] "
       << "[-roi tile pixel_w pixel_h] "
       << "-save suffix "
       << "-load mosaic-0 [mosaic-N]* "
       << endl
       << "EXAMPLE: " << argv[0] << " "
       << "-sh 8 "
       << "-feathering none "
       << "-save .png "
       << "-load *.mosaic "
       << endl;
  
  if (message != NULL)
  {
    cerr << "ERROR: " << message << endl;
  }
  
  exit(1);
}

//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  // this is so that the printouts look better:
  cout.precision(2);
  cout.setf(ios::scientific);
  
  // setup thread storage for the main thread:
  set_the_thread_storage_provider(the_boost_thread_t::thread_storage);
  the_boost_thread_t MAIN_THREAD_DUMMY;
  MAIN_THREAD_DUMMY.set_stopped(false);
  
  // setup thread and mutex interface creators:
  the_mutex_interface_t::set_creator(the_boost_mutex_t::create);
  the_thread_interface_t::set_creator(the_boost_thread_t::create);
  
  // parse the command line arguments:
  unsigned int shrink_factor = 1;
  double clahe_slope = 1.0;
  std::list<the_text_t> fn_load;
  the_text_t fn_save;
  bool save_variance = false;
  feathering_t feathering = FEATHER_NONE_E;
  unsigned int consume_default_tokens = 0;
  
  unsigned int roi_tile = std::numeric_limits<unsigned int>::max();
  unsigned int roi_width = 0;
  unsigned int roi_height = 0;
  
  for (int i = 1; i < argc; i++)
  {
    bool ok = true;
    
    if (strcmp(argv[i], "-sh") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -sh option");
      
      i++;
      shrink_factor = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -sh option");
    }
    else if (strcmp(argv[i], "-clahe") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -clahe option");
      
      i++;
      clahe_slope = the_text_t(argv[i]).toDouble(&ok);
      if (!ok || clahe_slope < 1.0) usage(argv, "bad -clahe option");
    }
    else if (strcmp(argv[i], "-variance") == 0)
    {
      save_variance = true;
    }
    else if (strcmp(argv[i], "-feathering") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -feathering option");
      
      i++;
      if (strcmp(argv[i], "blend") == 0)
      {
	feathering = FEATHER_BLEND_E;
      }
      else if (strcmp(argv[i], "binary") == 0)
      {
	feathering = FEATHER_BINARY_E;
      }
      else if (strcmp(argv[i], "none") == 0)
      {
	feathering = FEATHER_NONE_E;
      }
      else
      {
	usage(argv, "could not parse -feathering option");
      }
    }
    else if (strcmp(argv[i], "-roi") == 0)
    {
      if ((argc - i) <= 3) usage(argv, "could not parse -roi option");
      
      i++;
      roi_tile = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -roi tile parameter");
      
      i++;
      roi_width = the_text_t(argv[i]).toUInt(&ok);
      if (!ok || roi_width == 0) usage(argv, "bad -roi width parameter");
      
      i++;
      roi_height = the_text_t(argv[i]).toUInt(&ok);
      if (!ok || roi_height == 0) usage(argv, "bad -roi height parameter");
    }
    else if (strcmp(argv[i], "-load") == 0)
    {
      if ((argc - i) <= 2) usage(argv, "could not parse -load option");
      consume_default_tokens = 2;
    }
    else if (strcmp(argv[i], "-save") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -save option");
      
      i++;
      fn_save = the_text_t(argv[i]);
    }
    else if (consume_default_tokens > 0)
    {
      fn_load.push_back(the_text_t(argv[i]));
      consume_default_tokens++;
    }
    else
    {
      usage(argv,
	    the_text_t("unknown option \'") +
	    the_text_t(argv[i]) +
	    the_text_t("\'"));
    }
    
    if (consume_default_tokens > 0)
    {
      consume_default_tokens--;
    }
  }
  
  // sanity checks:
  if (fn_save.size() == 0)
  {
    usage(argv, "must specify -save option");
  }
  
  if (fn_load.empty())
  {
    usage(argv, "must specify -load option");
  }
  
  cout << "shrink factor: " << shrink_factor << endl
       << "clahe slope: " << clahe_slope << endl
       << "save variance: " << save_variance << endl
       << "feathering: " << feathering << endl
       << endl;
  
  // sort the list:
  fn_load.sort();
  
  std::vector<the_text_t> fn_mosaic(fn_load.begin(), fn_load.end());
  const unsigned int num_frames = fn_mosaic.size();
  
  std::vector<std::vector<base_transform_t::Pointer> > transform(num_frames);
  std::vector<image_t::Pointer> image;
  std::vector<mask_t::ConstPointer> mask;
  
  // load transforms, calculate mosaic bounding boxes:
  std::vector<pnt2d_t> mosaic_min(num_frames);
  std::vector<pnt2d_t> mosaic_max(num_frames);
  for (unsigned int i = 0; i < num_frames; i++)
  {
    bool ok = load_mosaic<image_t>(fn_mosaic[i],
				   shrink_factor,
				   clahe_slope,
				   i == 0, // setup images and masks only once
				   transform[i],
				   image,
				   mask);
    if (!ok)
    {
      usage(argv, the_text_t("could not load ") + fn_mosaic[i]);
    }
    
    calc_mosaic_bbox<image_t::Pointer, base_transform_t::Pointer>
      (transform[i],
       image,
       mosaic_min[i],
       mosaic_max[i]);
  }
  
  // calculate the union of all bounding boxes:
  pnt2d_t global_min = mosaic_min[0];
  pnt2d_t global_max = mosaic_max[0];
  for (unsigned int i = 1; i < num_frames; i++)
  {
    update_bbox(global_min, global_max, mosaic_min[i]);
    update_bbox(global_min, global_max, mosaic_max[i]);
  }
  
  cout << "GLOBAL MIN: " << global_min << endl
       << "GLOBAL MAX: " << global_max << endl;
  
  // setup the color table:
  const unsigned int num_images = image.size();
  xyz_t background_color = xyz(255, 255, 255);
  std::vector<xyz_t> color_table(num_images);
  
  static const xyz_t BLACK	= xyz(0, 0, 0);
  static const xyz_t WHITE	= xyz(255, 255, 255);
  
  static const xyz_t RED	= xyz(255, 0, 0);
  static const xyz_t GREEN	= xyz(0, 255, 0);
  static const xyz_t BLUE	= xyz(0, 0, 255);
  static const xyz_t CYAN	= xyz(0, 255, 255);
  static const xyz_t MAGENTA	= xyz(255, 0, 255);
  static const xyz_t YELLOW	= xyz(255, 255, 0);
  static const xyz_t ORANGE	= xyz(255, 127, 0);
  
  static const xyz_t LGREEN	= xyz(128, 255, 128);
  static const xyz_t GRAPE	= xyz(128, 36, 182);
  static const xyz_t BROWN	= xyz(145, 58, 8);
  static const xyz_t AMPAD	= xyz(112, 207, 159);
  
  static const xyz_t temperature[] = {
    BLACK,
    BLUE,
    CYAN,
    GREEN,
    LGREEN,
    YELLOW,
    ORANGE,
    RED,
    WHITE,
    WHITE
  };
  
#if 0
  static const xyz_t zebra[] = {
    CYAN,
    RED,
    BLUE,
    ORANGE,
    GREEN,
    MAGENTA,
    YELLOW,
    GRAPE,
    AMPAD,
    BROWN
  };
  
  for (unsigned int i = 0; i < num_images; i++)
  {
    color_table[i] = zebra[i % (sizeof(zebra) / sizeof(xyz_t))];
  }
#else
  make_colors(num_images, color_table);
#endif
  
  std::vector<double> max_variance(num_frames);
  std::vector<double> avg_variance(num_frames);
  std::vector<the_text_t> filename(num_frames);
  
  image_t::SpacingType mosaic_sp = image[0]->GetSpacing();
  
  // generate the frames:
  for (unsigned int i = 0; i < num_frames; i++)
  {
    // calculate the region of interest:
    pnt2d_t roi_min;
    image_t::SizeType roi_sz;
    
    if (roi_tile == std::numeric_limits<unsigned int>::max())
    {
      roi_min = global_min;
      roi_sz[0] = (unsigned int)((global_max[0] - global_min[0]) /
				 mosaic_sp[0]);
      roi_sz[1] = (unsigned int)((global_max[1] - global_min[1]) /
				 mosaic_sp[1]);
    }
    else
    {
      base_transform_t::Pointer inverse = transform[i][roi_tile]->GetInverse();
      roi_min = inverse->TransformPoint(image[roi_tile]->GetOrigin());
      roi_min[0] -= 0.1 * mosaic_sp[0] * double(roi_width);
      roi_min[1] -= 0.1 * mosaic_sp[1] * double(roi_height);
      roi_sz[0] = roi_width;
      roi_sz[1] = roi_height;
    }
    
    if (!save_variance)
    {
      image_t::Pointer mosaic[3];
      make_mosaic_rgb<image_t::Pointer, base_transform_t::Pointer>
	(mosaic,
	 mosaic_sp,
	 roi_min,
	 roi_sz,
	 background_color,
	 color_table,
	 std::vector<bool>(num_images, false), // omit nothing
	 transform[i],
	 image,
	 mask,
	 feathering);
      
      filename[i] =
	fn_mosaic[i] + fn_save + "-" + the_text_t::number(i, 3, '0') + ".png";
      save_rgb<image_t::Pointer>(mosaic, filename[i]);
    }
    else
    {
      typedef itk::LinearInterpolateImageFunction<image_t, double>
	image_interpolator_t;
      
      typedef itk::ImageMosaicVarianceMetric<image_t, image_interpolator_t>
	mosaic_metric_t;
      
      mosaic_metric_t::Pointer mosaic_metric = mosaic_metric_t::New();
      mosaic_metric->image_.resize(num_images);
      mosaic_metric->mask_.resize(num_images);
      mosaic_metric->transform_.resize(num_images);
      for (unsigned int j = 0; j < num_images; j++)
      {
	mosaic_metric->image_[j] = image[j];
	mosaic_metric->mask_[j] = mask[j];
	mosaic_metric->transform_[j] = transform[i][j];
      }
      
      std::vector<bool> param_shared;
      std::vector<bool> param_active;
      param_shared.assign(transform[i][0]->GetNumberOfParameters(), false);
      param_active.assign(transform[i][0]->GetNumberOfParameters(), false);
      mosaic_metric->setup_param_map(param_shared, param_active);
      mosaic_metric->Initialize();
      
      image_t::Pointer mosaic = mosaic_metric->variance(mosaic_sp,
							roi_min,
							roi_sz,
							max_variance[i],
							avg_variance[i]);
      
      filename[i] =
	fn_mosaic[i] + fn_save + "-" + the_text_t::number(i, 3, '0') + ".mha";
      save<image_t>(mosaic, filename[i]);
    }
  }
  
  if (save_variance)
  {
    double max_var_global = max_variance[0];
    for (unsigned int i = 1; i < num_frames; i++)
    {
      max_var_global = std::max(max_var_global, max_variance[i]);
    }
    
    // apply the colormap:
    typedef itk::RGBPixel<unsigned char> rgb_t;
    typedef itk::Image<rgb_t, 2> rgb_image_t;
    for (unsigned int i = 0; i < num_frames; i++)
    {
      image_t::Pointer var_img = load<image_t>(filename[i]);
      
      rgb_image_t::Pointer rgb_img =
	make_image<rgb_image_t>(var_img->GetSpacing(),
				var_img->GetLargestPossibleRegion().GetSize());
      rgb_img->SetOrigin(var_img->GetOrigin());
      
      typedef itk::ImageRegionConstIteratorWithIndex<image_t> itex_t;
      itex_t itex(var_img, var_img->GetLargestPossibleRegion());
      for (itex.GoToBegin(); !itex.IsAtEnd(); ++itex)
      {
	double v = itex.Get();
	double t = v / max_var_global;
	double ct = t * 8.0;
	unsigned int i0 = (unsigned int)(floor(ct));
	unsigned int i1 = i0 + 1;
	double w1 = ct - double(i0);
	double w0 = 1.0 - w1;
	
	xyz_t color = (temperature[i0] * w0) + (temperature[i1] * w1);
	rgb_t pixel;
	pixel[0] = (unsigned char)(std::min(255.0, std::max(0.0, color[0])));
	pixel[1] = (unsigned char)(std::min(255.0, std::max(0.0, color[1])));
	pixel[2] = (unsigned char)(std::min(255.0, std::max(0.0, color[2])));
	rgb_img->SetPixel(itex.GetIndex(), pixel);
      }
      
      save<rgb_image_t>(rgb_img, filename[i] + ".png");
    }
  }
  
  // done:
  return 0;
}
