// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : ir-assemble.cxx
// Author       : Pavel A. Koshevoy
// Created      : 2005/03/29 16:55
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Assemble a mosaic image from a .mosaic file.

// the includes:
#include <itk/common.hxx>
#include <itk/itkImageMosaicVarianceMetric.h>
#include <utils/the_dynamic_array.hxx>
#include <thread/the_boost_mutex.hxx>
#include <thread/the_boost_thread.hxx>

// system includes:
#include <cmath>


//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL)
{
  cerr << "USAGE: " << argv[0] << " "
       << "[-sh shrink_factor] "
       << "[-sp pixel_spacing] "
       << "[-remap] "
       << "[-clahe slope] "
       << "[[-variance] | [-feathering [none | blend | binary]]] "
       << "-load transform_params_file_in "
       << "-save mosaic_file_out "
       << "[-mask mosaic_mask_file_out] "
       << endl
       << "EXAMPLE: " << argv[0] << " "
       << "-sh 8 "
       << "-feathering blend "
       << "-load params.mosaic "
       << "-save mosaic.tif "
       << "-mask mosaic_mask.tif "
       << endl;
  
  if (message != NULL)
  {
    cerr << "ERROR: " << message << endl;
  }
  
  exit(1);
}

//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  // this is so that the printouts look better:
  cout.precision(2);
  cout.setf(ios::scientific);
  
  // setup thread storage for the main thread:
  set_the_thread_storage_provider(the_boost_thread_t::thread_storage);
  the_boost_thread_t MAIN_THREAD_DUMMY;
  MAIN_THREAD_DUMMY.set_stopped(false);
  
  // setup thread and mutex interface creators:
  the_mutex_interface_t::set_creator(the_boost_mutex_t::create);
  the_thread_interface_t::set_creator(the_boost_thread_t::create);
  
  // parse the command line arguments:
  unsigned int shrink_factor = 1;
  bool pixel_spacing_user_override = false;
  bool remap_values = false;
  double pixel_spacing = 1.0;
  bool use_std_mask = false;
  double clahe_slope = 1.0;
  std::list<the_text_t> in;
  the_text_t fn_save;
  the_text_t fn_mask;
  bool save_variance = false;
  feathering_t feathering = FEATHER_NONE_E;
  
  std::vector<base_transform_t::Pointer> transform;
  std::vector<image_t::ConstPointer> image;
  std::vector<mask_t::ConstPointer> mask;
  
  for (int i = 1; i < argc; i++)
  {
    bool ok = true;
    
    if (strcmp(argv[i], "-sh") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -sh option");
      
      i++;
      shrink_factor = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -sh option");
    }
    else if (strcmp(argv[i], "-sp") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -sp option");
      
      i++;
      pixel_spacing = the_text_t(argv[i]).toDouble(&ok);
      if (!ok) usage(argv, "bad -sp option");
      
      pixel_spacing_user_override = true;
    }
    else if (strcmp(argv[i], "-remap") == 0)
    {
      remap_values = true;
    }
    else if (strcmp(argv[i], "-clahe") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -clahe option");
      
      i++;
      clahe_slope = the_text_t(argv[i]).toDouble(&ok);
      if (!ok || clahe_slope < 1.0) usage(argv, "bad -clahe option");
    }
    else if (strcmp(argv[i], "-variance") == 0)
    {
      save_variance = true;
    }
    else if (strcmp(argv[i], "-feathering") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -feathering option");
      
      i++;
      if (strcmp(argv[i], "blend") == 0)
      {
	feathering = FEATHER_BLEND_E;
      }
      else if (strcmp(argv[i], "binary") == 0)
      {
	feathering = FEATHER_BINARY_E;
      }
      else if (strcmp(argv[i], "none") == 0)
      {
	feathering = FEATHER_NONE_E;
      }
      else
      {
	usage(argv, "could not parse -feathering option");
      }
    }
    else if (strcmp(argv[i], "-load") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -load option");
      
      i++;
      std::fstream fin;
      fin.open(argv[i], ios::in);
      if (!fin.is_open())
      {
	usage(argv,
	      the_text_t("could not open \'") +
	      the_text_t(argv[i]) +
	      the_text_t("\' for reading"));
      }
      
      double mosaic_pixel_spacing = 0;
      load_mosaic<base_transform_t>(fin,
				    mosaic_pixel_spacing,
				    use_std_mask,
				    in,
				    transform);
      
      if (!pixel_spacing_user_override)
      {
	pixel_spacing = mosaic_pixel_spacing;
      }
      
      unsigned int num_images = transform.size();
      assert(num_images > 1);
      assert(pixel_spacing != 0);
      
      image.resize(num_images);
      mask.resize(num_images);
      
      unsigned int i = 0;
      for (std::list<the_text_t>::const_iterator iter = in.begin();
	   iter != in.end(); ++iter, i++)
      {
	const the_text_t & fn_image = *iter;
	
	// read the image:
	cout << setw(3) << i << " ";
	image[i] = std_tile<image_t>(fn_image, shrink_factor, pixel_spacing);
	if (use_std_mask)
	{
	  mask[i] = std_mask<image_t>(image[i]);
	}
	
	if (clahe_slope > 1.0)
	{
	  image[i] = CLAHE<image_t>(image[i],
				    255,
				    255,
				    clahe_slope,
				    256,
				    0.0,
				    255.0,
				    mask[i]);
	}
      }
      
      fin.close();
    }
    else if (strcmp(argv[i], "-save") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -save option");
      
      i++;
      fn_save = the_text_t(argv[i]);
    }
    else if (strcmp(argv[i], "-mask") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -mask option");
      
      i++;
      fn_mask = the_text_t(argv[i]);
    }
    else
    {
      usage(argv,
	    the_text_t("unknown option \'") +
	    the_text_t(argv[i]) +
	    the_text_t("\'"));
    }
  }
  
  // sanity checks:
  if (fn_save.size() == 0)  usage(argv, "must specify -save option");
  if (in.empty()) usage(argv, "must specify at least 1 input image");
  
  cout << "shrink factor: " << shrink_factor << endl
       << "pixel spacing: " << pixel_spacing << endl
       << "save variance: " << save_variance << endl
       << "use feathering: " << feathering << endl
       << "remap min max: " << remap_values << endl
       << endl;
  
  cout << "assembling mosaic..." << endl;
  image_t::Pointer mosaic;
  
  if (save_variance)
  {
    typedef itk::LinearInterpolateImageFunction<image_t, double>
      image_interpolator_t;
    
    typedef itk::ImageMosaicVarianceMetric<image_t, image_interpolator_t>
      mosaic_metric_t;
    
    mosaic_metric_t::Pointer mosaic_metric = mosaic_metric_t::New();
    unsigned int num_images = in.size();
    mosaic_metric->image_.resize(num_images);
    mosaic_metric->mask_.resize(num_images);
    mosaic_metric->transform_.resize(num_images);
    for (unsigned int i = 0; i < num_images; i++)
    {
      mosaic_metric->image_[i] = image[i];
      mosaic_metric->mask_[i] = mask[i];
      mosaic_metric->transform_[i] = transform[i];
    }
    
    std::vector<bool> param_shared;
    std::vector<bool> param_active;
    param_shared.assign(transform[0]->GetNumberOfParameters(), false);
    param_active.assign(transform[0]->GetNumberOfParameters(), false);
    mosaic_metric->setup_param_map(param_shared, param_active);
    mosaic_metric->Initialize();
    
    mosaic_metric_t::measure_t max_var;
    mosaic_metric_t::measure_t avg_var;
    mosaic = mosaic_metric->variance(max_var, avg_var);
    cout << "max variance: " << max_var << endl
	 << "avg variance: " << avg_var << endl;
  }
  else
  {
    mosaic = make_mosaic<image_t::ConstPointer, base_transform_t::Pointer>
      (image,
       transform,
       feathering,
       ~0,
       ~0,
       mask);
  }
  
  if (remap_values)
  {
    save<native_image_t>(cast<image_t, native_image_t>
			 (remap_min_max<image_t>(mosaic)),
			 fn_save);
  }
  else if (fn_save.match_tail(".mha") ||
	   fn_save.match_tail(".mhd") ||
	   fn_save.match_tail(".nrrd"))
  {
    save<image_t>(mosaic, fn_save);
  }
  else
  {
    save<native_image_t>(cast<image_t, native_image_t>(mosaic), fn_save);
  }
  
  if (fn_mask.is_empty())
  {
    // done:
    return 0;
  }
  
  // assemble the mosaic mask:
  cout << "assembling mosaic mask..." << endl;
  mask_t::Pointer mosaic_mask;
  
  if (!use_std_mask)
  {
    unsigned int num_images = in.size();
    for (unsigned int i = 0; i < num_images; i++)
    {
      // since the standard mask was not used assume the whole image is masked:
      mask[i] = std_mask<image_t>(image[i], false);
      
      // free up the image memory, we no longer need it:
      image[i] = image_t::Pointer(NULL);
    }
  }
  
  mosaic_mask = make_mask<base_transform_t::Pointer>(transform, mask);
  save<mask_t>(mosaic_mask, fn_mask);
  
  // done:
  return 0;
}
