// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : ir-autocrop.cxx
// Author       : Pavel A. Koshevoy
// Created      : 2005/09/30 12:27
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Automatically align the image frame to the canonical
//                x/y axis, then crop the frame. Frame can be dark (black)
//                or light (white).

// the includes:
#include <itk/common.hxx>
#include <utils/the_utils.hxx>
#include <thread/the_boost_mutex.hxx>
#include <thread/the_boost_thread.hxx>

// ITK includes:
#include <itkFixedCenterOfRotationAffineTransform.h>

// VXL includes:
#include <vnl/algo/vnl_svd.h>

// system includes:
#include <math.h>

// local typedefs:
typedef image_t::RegionType rn_t;
typedef image_t::IndexType  ix_t;
typedef image_t::SizeType   sz_t;


//----------------------------------------------------------------
// PI_OVER_2
// 
static const double PI_OVER_2 = M_PI / 2.0;


//----------------------------------------------------------------
// DEBUG_EDGE
// 
bool DEBUG_EDGE = false;


//----------------------------------------------------------------
// solve_for_line_parameters
// 
void
solve_for_line_parameters(const bool & vertical,
			  const std::vector<image_t::IndexType> & pts,
			  const std::list<unsigned int> & inliers,
			  double & a,
			  double & b,
			  double & c)
{
  const unsigned int num_inliers = inliers.size();
  
  // Mx = z;
  vnl_matrix<double> M(num_inliers, 2);
  vnl_vector<double> z(num_inliers);
  
  std::list<unsigned int>::const_iterator iter = inliers.begin();
  for (unsigned int i = 0; i < num_inliers; i++, ++iter)
  {
    const image_t::IndexType & xy = pts[*iter];
    M[i][0] = vertical ? xy[1] : xy[0];
    M[i][1] = 1.0;
    z[i] = vertical ? -xy[0] : -xy[1];
  }
  
  vnl_svd<double> svd(M);
  vnl_vector<double> x = svd.solve(z);
  
  c = x[1];
  if (vertical)
  {
    a = 1.0;
    b = x[0];
  }
  else
  {
    a = x[0];
    b = 1.0;
  }
}

//----------------------------------------------------------------
// refine_inliers
// 
bool
refine_inliers(const double & a,
	       const double & b,
	       const double & c,
	       const double & t,
	       const std::vector<image_t::IndexType> & pts,
	       std::vector<bool> & is_inlier,
	       std::list<unsigned int> & inliers,
	       double & model_quality)
{
  // point to line distance for a line specified by a general equation
  // ax + by + c = 0 is very well explained at the cut-the-knot.org
  // 
  // http://www.cut-the-knot.org/Curriculum/Calculus/StraightLine.shtml
  // http://www.cut-the-knot.org/Curriculum/Calculus/DistanceToLine.shtml
  // 
  // Given point P = [u,v]T and line L(x, y) = Ax + By + C = 0
  // dist(P, L) = abs(L(u, v)) / norm([A, B]T)
  // 
  const double ab_norm = sqrt(a * a + b * b);
  
  // shortcut:
  const unsigned int & num_pts = pts.size();
  
  double inlier_error = 0.0;
  double total_error = 0.0;
  model_quality = 0.0;
  
  bool removed_inliers = false;
  bool added_inliers = false;
  for (unsigned int i = 0; i < num_pts; i++)
  {
    // transform one of the points in the match pair and see how
    // closely it lands to its counterpart:
    const image_t::IndexType & xy = pts[i];
    
    double d = fabs(a * double(xy[0]) + b * double(xy[1]) + c) / ab_norm;
    total_error += d;
    
    if (d <= t)
    {
      inlier_error += d;
      
      if (!is_inlier[i])
      {
	inliers.push_back(i);
	is_inlier[i] = true;
	added_inliers = true;
      }
    }
    else if (is_inlier[i])
    {
      inliers.remove(i);
      is_inlier[i] = false;
      removed_inliers = true;
    }
  }
  
  model_quality = double(inliers.size());
  return removed_inliers || added_inliers;
}

//----------------------------------------------------------------
// RANSAC
// 
bool
RANSAC(// are we looking for a vertical line?
       const bool vertical,
       
       // candidate line points:
       const std::vector<image_t::IndexType> & pts,
       
       // RANSAC controls:
       const unsigned int k,
       const unsigned int n,
       const double t,
       
       // pass back the results via these variables:
       double & best_a,
       double & best_b,
       double & best_c,
       
       // FIXME: this is for debugging, really...
       std::list<unsigned int> & bestfit,
       double & bestquality)
{
  const double initial_bestquality = bestquality;
  double a = best_a;
  double b = best_b;
  double c = best_c;
  
  // shortcut:
  const unsigned int & num_pts = pts.size();
  
  // at least n sample points are required to estimate the model parameters:
  if (num_pts < n) return false;
  
  for (unsigned int i = 0; i < k; i++)
  {
    std::list<unsigned int> inliers;
    std::vector<bool> is_inlier;
    is_inlier.assign(num_pts, false);
    
    // pick n randomly selected points:
    for (unsigned int j = 0; j < n; j++)
    {
      // assume the points are sorted in the order of decreasing importance,
      // implement importance sampling to take advantage of this:
      unsigned int index;
      while (true)
      {
	// importance sampling of the points, better points are sampled
	// more frequently:
	double nx = double(num_pts);
	double q = drand();
	
	// PDF(x) = 1 - x;
	// CDF(x) = 2x - x^2;
	double x = nx * (1 - sqrt(q));
	
	index = std::min(num_pts - 1, (unsigned int)(x));
	if (!is_inlier[index]) break;
      }
      
      inliers.push_back(index);
      is_inlier[index] = true;
    }
    
    // solve for line parameters based on the selected line points:
    solve_for_line_parameters(vertical, pts, inliers, a, b, c);
    
    // find other potential inliers and calculate the error measure:
    double model_quality = std::numeric_limits<double>::max();
    bool inliers_were_altered = refine_inliers(a,
					       b,
					       c,
					       t,
					       pts,
					       is_inlier,
					       inliers,
					       model_quality);
    
    if (model_quality > bestquality)
    {
      // allow for the removal of outliers (in case the original
      // inliers were poor):
      double quality = model_quality;
      for (unsigned int j = 0; j < 100 && inliers_were_altered; j++)
      {
	if (inliers.size() < n) break;
	
	// update the transform to reflect the changes to the inliers:
	solve_for_line_parameters(vertical, pts, inliers, a, b, c);
	
	// re-calculate the error measure, update the inliers:
	inliers_were_altered = refine_inliers(a,
					      b,
					      c,
					      t,
					      pts,
					      is_inlier,
					      inliers,
					      quality);
      }
      
      if (quality > bestquality && inliers.size() >= n)
      {
	solve_for_line_parameters(vertical, pts, inliers, a, b, c);
       	
	// save the better result:
	bestquality = quality;
	best_a = a;
	best_b = b;
	best_c = c;
	bestfit = inliers;
	cout << "FIXME: quality: " << bestquality
	     << ", iteration: " << i << endl;
      }
    }
  }
  
  return initial_bestquality < bestquality;
}

//----------------------------------------------------------------
// detect_edge
// 
// returns the number of inlier points that define the edge
// 
double
detect_edge(const bool vertical,
	    double & a,
	    double & b,
	    double & c,
	    const image_t * image,
	    const char * filter_pattern,
	    const unsigned int & samples,
	    const unsigned int & x_origin,
	    const unsigned int & x_extent,
	    const unsigned int & y_origin,
	    const unsigned int & y_extent)
{
  sz_t image_sz = image->GetLargestPossibleRegion().GetSize();
  const unsigned int image_w = image_sz[0];
  const unsigned int image_h = image_sz[1];
  
  // find the filter origin:
  unsigned int filter_len = strlen(filter_pattern);
  unsigned int filter_origin = 0;
  {
    unsigned int min = ~0;
    unsigned int max = ~0;
    for (unsigned int i = 0; i < filter_len; i++)
    {
      if (filter_pattern[i] == '+')
      {
	if (min == ~0u)
	{
	  // min:
	  min = i;
	}
	else
	{
	  // max:
	  max = i;
	}
      }
    }
    
    filter_origin = min == 0 ? max : min;
  }
  
  std::vector<double> r(samples);
  std::vector<unsigned int> x(samples);
  std::vector<unsigned int> y(samples);
  
  unsigned int * V = vertical ? &y[0] : &x[0];
  unsigned int V_coord = vertical ? 1 : 0;
  unsigned int V_origin = vertical ? y_origin : x_origin;
  unsigned int V_extent = vertical ? y_extent : x_extent;
  
  unsigned int * U = vertical ? &x[0] : &y[0];
  unsigned int U_coord = vertical ? 0 : 1;
  unsigned int U_origin = vertical ? x_origin : y_origin;
  unsigned int U_extent = vertical ? x_extent : y_extent;
  unsigned int U_length = vertical ? image_w : image_h;
  
  for (unsigned int i = 0; i < samples; i++)
  {
    r[i] = -std::numeric_limits<double>::max();
    V[i] = V_origin + (unsigned int)(double(V_extent) * (0.5 + double(i)) /
				     double(samples));
    
    ix_t index;
    index[V_coord] = V[i];
    
    for (unsigned int j = U_origin; j < U_origin + U_extent; j++)
    {
      unsigned int weight_neg = 0;
      unsigned int weight_pos = 0;
      double response_neg = 0.0;
      double response_pos = 0.0;
      for (unsigned int k = 0; k < filter_len; k++)
      {
	int Uk = j + k - filter_origin;
	if (Uk < 0 || (unsigned int)(Uk) >= U_length) continue;
	
	index[U_coord] = Uk;
	double response = image->GetPixel(index);
	
	if (filter_pattern[k] == '+')
	{
	  weight_pos++;
	  response_pos += response;
	}
	else if (filter_pattern[k] == '-')
	{
	  weight_neg++;
	  response_neg += response;
	}
      }
      
      // skip regions where only one of the traits of the filter applies:
      if (weight_neg == 0 || weight_pos == 0) continue;
      
      double scale = double(weight_neg) / double(weight_pos);
      double response = response_pos * scale - response_neg;
      if (response > r[i])
      {
	r[i] = response;
	U[i] = j;
      }
    }
  }
  
  // sort the responses from strongest to weakest:
  std::vector<unsigned int> ix(samples);
  for (unsigned int i = 0; i < samples; i++) ix[i] = i;
  for (unsigned int i = 0; i < samples; i++)
  {
    for (unsigned int j = i + 1; j < samples; j++)
    {
      if (r[ix[i]] < r[ix[j]]) std::swap(ix[i], ix[j]);
    }
  }
  
  // assemble the points:
  std::list<image_t::IndexType> pt_list;
  for (unsigned int i = 0; i < samples; i++)
  {
    // skip poor samples:
    if (r[ix[i]] == -std::numeric_limits<double>::max()) continue;
    
    image_t::IndexType index;
    index[0] = x[ix[i]];
    index[1] = y[ix[i]];
    pt_list.push_back(index);
  }
  
  // convert the list into an array:
  std::vector<image_t::IndexType> pts(pt_list.begin(), pt_list.end());
  cout << "\nFIXME: detected points: " << pts.size() << endl;
  
  // use RANSAC to select the inliers from the set of detected edge points:
  std::list<unsigned int> bestfit;
  double bestquality = 0.0;
  
  RANSAC(vertical,
	 
	 // the data points:
	 pts,
	 
	 // RANSAC controls:
	 int(pts.size()),	// k
	 2,			// n
	 1.0,			// t
	 
	 // results: the general line equation parameters:
	 a, b, c,
	 
	 // FIXME: this is for debugging, really...
	 bestfit, bestquality);
  
  // FIXME:
#if 0
  
  static unsigned int counter = 0;
  double min_r = r[0];
  double max_r = min_r;
  for (unsigned int i = 1; i < samples; i++)
  {
    if (r[i] == -std::numeric_limits<double>::max()) continue;
    min_r = std::min(min_r, r[i]);
    max_r = std::max(max_r, r[i]);
  }
  double rng_r = max_r - min_r;
  
  image_t::Pointer edge = cast<image_t, image_t>(image);
  for (std::list<unsigned int>::const_iterator iter = bestfit.begin();
       iter != bestfit.end(); ++iter)
  {
    const image_t::IndexType & index = pts[*iter];
    
    pnt2d_t point;
    image->TransformIndexToPhysicalPoint(index, point);
    
    mark<image_t>(edge, point, 255.0 * (r[ix[*iter]] - min_r) / rng_r, 20);
  }
  save<native_image_t>(cast<image_t, native_image_t>(edge),
		       the_text_t("/tmp/points-") +
		       the_text_t::number(counter++) +
		       the_text_t(".tif"));
  
#endif
  
  return bestquality;
}

//----------------------------------------------------------------
// detect_edge_twice
// 
double
detect_edge_twice(const bool vertical,
		  double & a,
		  double & b,
		  double & c,
		  const image_t * image,
		  const char * filter_pattern,
		  const unsigned int & samples,
		  const unsigned int & x_origin,
		  const unsigned int & x_extent,
		  const unsigned int & y_origin,
		  const unsigned int & y_extent)
{
  the_text_t filter_primary(filter_pattern);
  the_text_t filter_inverse(filter_primary);
  for (unsigned int i = 0; i < filter_inverse.size(); i++)
  {
    filter_inverse[i] = (filter_inverse[i] == '+') ? '-' : '+';
  }
  
  double q0 = detect_edge(vertical, a, b, c, image, filter_primary, samples,
			  x_origin, x_extent, y_origin, y_extent);
  
  double a1;
  double b1;
  double c1;
  double q1 = detect_edge(vertical, a1, b1, c1, image, filter_inverse, samples,
			  x_origin, x_extent, y_origin, y_extent);
  
  if (q0 * 2.0 > q1)
  {
    // the primary filter was appropriate:
    return q0;
  }
  
  a = a1;
  b = b1;
  c = c1;
  return q1;
}

//----------------------------------------------------------------
// detect_vertical_edge
// 
double
detect_vertical_edge(double & a,
		     double & b,
		     double & c,
		     const image_t * image,
		     const char * filter_pattern = "----++++++++++++",
		     const unsigned int & samples = 100u,
		     const unsigned int & x_origin = 0,
		     const unsigned int & x_extent = 100,
		     const unsigned int & y_origin = 0,
		     const unsigned int & y_extent = 100)
{
  return detect_edge_twice(true, a, b, c, image, filter_pattern, samples,
			   x_origin, x_extent, y_origin, y_extent);
}

//----------------------------------------------------------------
// detect_horizontal_edge
// 
double
detect_horizontal_edge(double & a,
		       double & b,
		       double & c,
		       const image_t * image,
		       const char * filter_pattern = "----++++++++++++",
		       const unsigned int & samples = 100u,
		       const unsigned int & x_origin = 0,
		       const unsigned int & x_extent = 100,
		       const unsigned int & y_origin = 0,
		       const unsigned int & y_extent = 100)
{
  return detect_edge_twice(false, a, b, c, image, filter_pattern, samples,
			   x_origin, x_extent, y_origin, y_extent);
}


//----------------------------------------------------------------
// parametric_to_generic
// 
void
parametric_to_generic(const pnt2d_t & p0,
		      const pnt2d_t & p1,
		      double & a,
		      double & b,
		      double & c)
{
  const double x0 = p0[0];
  const double y0 = p0[1];
  const double dx = p1[0] - p0[0];
  const double dy = p1[1] - p0[1];
  
  if (fabs(dy) > fabs(dx))
  {
    const double dx_dy = dx / dy;
    a = 1.0;
    b = -dx_dy;
    c = y0 * dx_dy - x0;
  }
  else
  {
    const double dy_dx = dy / dx;
    a = -dy_dx;
    b = 1.0;
    c = x0 * dy_dx - y0;
  }
}

//----------------------------------------------------------------
// angle_between_lines
// 
// http://mathworld.wolfram.com/Line.html
// 
// NOTE:
// this will always return an angle in the range [0, Pi/2]
// 
double
angle_between_lines(const double & a0, const double & b0,
		    const double & a1, const double & b1)
{
  double a = fabs(atan2(a0 * b1 - a1 * b0, a0 * a1 + b0 * b1));
  
  if (a > PI_OVER_2)
  {
    a = M_PI - a;
  }
  
  return a;
}

//----------------------------------------------------------------
// angle_between_lines
// 
double
angle_between_lines(const pnt2d_t & p0, const pnt2d_t & p1,
		    const pnt2d_t & q0, const pnt2d_t & q1)
{
  double a0;
  double b0;
  double c0;
  parametric_to_generic(p0, p1, a0, b0, c0);
  
  double a1;
  double b1;
  double c1;
  parametric_to_generic(q0, q1, a1, b1, c1);
  
  // this angle is always from 0 to Pi/2:
  double angle = angle_between_lines(a0, b0, a1, b1);
  
  vec2d_t p = p1 - p0;
  vec2d_t q = q1 - q0;
  
  // check for obtuse angle:
  double p_dot_q = p * q;
  if (p_dot_q < 0.0)
  {
    angle = M_PI - angle;
  }
  
  // determine the sign of the rotation:
  double z = p[0] * q[1] - q[0] * p[1];
  if (z < 0.0)
  {
    // change the direction of rotation:
    angle = -angle;
  }
  
  return angle;
}

//----------------------------------------------------------------
// angle_between_lines
// 
double
angle_between_lines(const pnt2d_t & p0, const pnt2d_t & p1,
		    double dx, double dy)
{
  double angle = angle_between_lines(p0, p1, pnt2d(0, 0), pnt2d(dx, dy));
  /*
  if (fabs(angle) > PI_OVER_2)
  {
    if (angle > 0.0) angle -= M_PI;
    else angle = M_PI - angle;
  }
  */
  return angle;
}

//----------------------------------------------------------------
// crooked
// 
double
crooked(const pnt2d_t & ul,
	const pnt2d_t & ll,
	const pnt2d_t & ur,
	const pnt2d_t & lr)
{
  static const char * corner[] = { "LL", "UL", "UR", "LR" };
  static const char * edge[] = { "bottom", "left", "top", "right" };
  
  const double a[] = {
    angle_between_lines(ll, lr, ll, ul), // ll
    angle_between_lines(ul, ll, ul, ur), // ul
    angle_between_lines(ur, ul, ur, lr), // ur
    angle_between_lines(lr, ur, lr, ll)  // lr
  };
  
  cout << endl;
  for (unsigned int i = 0; i < 4; i++)
  {
    cout << corner[i] << " angle: " << a[i] * 180.0 / M_PI << endl;
  }
  
  const double rotation[] = {
    angle_between_lines(ll, lr, 1.0, 0.0), // bottom
    angle_between_lines(ll, ul, 0.0, 1.0), // left
    angle_between_lines(ul, ur, 1.0, 0.0), // top
    angle_between_lines(lr, ur, 0.0, 1.0)  // bottom
  };
  
  cout << endl;
  for (unsigned int i = 0; i < 4; i++)
  {
    cout << edge[i] << " rotate: " << rotation[i] * 180.0 / M_PI << endl;
  }

  double d[4];
  for (unsigned int i = 0; i < 4; i++)
  {
    d[i] = fabs(a[i] - PI_OVER_2);
  }
  
  unsigned int best_corner = 0;
  for (unsigned int i = 0; i < 4; i++)
  {
    if (d[i] < d[best_corner])
    {
      best_corner = i;
    }
  }
  
  cout << "best corner: " << corner[best_corner]
       << ", error = " << d[best_corner] << endl;
  
  double de[4];
  for (unsigned int i = 0; i < 4; i++)
  {
    unsigned int i1 = (i + 4 - 1) % 4;
    de[i] = (d[i] + d[i1]) / 2.0;
  }
  
  cout << endl;
  double err[] = { 0.0 };
  for (unsigned int i = 0; i < 4; i++)
  {
    for (unsigned int j = 0; j < 4; j++)
    {
      if (i == j) continue;
      double dr = fabs(rotation[i] - rotation[j]);
      err[i] += dr;
      cout << edge[i] << " vs " << edge[j] << " rotation error: " << dr
	   << endl;
    }
    cout << endl;
  }

  unsigned int best_rotation = 0;
  for (unsigned int i = 0; i < 4; i++)
  {
    cout << edge[i] << " mean rotation error: " << err[i] / 3.0 << endl;
    if (err[i] < err[best_rotation])
    {
      best_rotation = i;
    }
  }
  cout << endl;
  
  // do not consider the edges with the worst rotation error:
  double num_ok = 0.0;
  double rotate = 0.0;
  
  for (unsigned int i = 0; i < 4; i++)
  {
    double r = err[i] / (err[best_rotation] + 1e-9);
    if (r > 1.4) continue;
    
    cout << edge[i] << " will be used to correct for rotation" << endl;
    num_ok += 1.0;
    rotate += rotation[i];
  }
  cout << endl;
  
  // FIXME:
  rotate /= num_ok;
  return rotate;
}

//----------------------------------------------------------------
// index_to_physical
// 
image_t::PointType
index_to_physical(const image_t * image, const image_t::PointType & ix)
{
  const image_t::SpacingType & spacing = image->GetSpacing();
  const image_t::PointType & origin = image->GetOrigin();
  
  image_t::PointType pt;
  pt[0] = origin[0] + ix[0] * spacing[0];
  pt[1] = origin[1] + ix[1] * spacing[1];
  
  return pt;
}

//----------------------------------------------------------------
// RADIUS
// 
static const int RADIUS = 4;

//----------------------------------------------------------------
// mark_pixel
// 
template <class TImage>
void
mark_pixel(typename TImage::Pointer & image,
	   typename TImage::IndexType & center,
	   typename TImage::PixelType color,
	   short int radius = 0)
{
  typename TImage::SizeType sz = image->GetLargestPossibleRegion().GetSize();
  typename TImage::IndexType ix;
  for (short int i = -radius; i <= radius; i++)
  {
    ix[0] = center[0] + i;
    for (short int j = -radius; j <= radius; j++)
    {
      ix[1] = center[1] + j;
      
      if (image_size_value_t(ix[0]) < sz[0] &&
	  image_size_value_t(ix[1]) < sz[1])
      {
	image->SetPixel(ix, color);
      }
    }
  }
}

//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL)
{
  cerr << "USAGE: " << argv[0] << " "
       << "[-border black|white] "
       << "[-invert] "
       << "[-debug] "
       << "-load in_image "
       << "-save out_image "
       << endl
       << "EXAMPLE: " << argv[0] << " "
       << "-border black "
       << "-load in/48755.tif "
       << "-save out/48755.tif "
       << endl;
  
  if (message != NULL)
  {
    cerr << "ERROR: " << message << endl;
  }
  
  exit(1);
}


//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  // this is so that the printouts look better:
  cout.precision(2);
  cout.setf(ios::scientific);
  
  // setup thread storage for the main thread:
  set_the_thread_storage_provider(the_boost_thread_t::thread_storage);
  the_boost_thread_t MAIN_THREAD_DUMMY;
  MAIN_THREAD_DUMMY.set_stopped(false);
  
  // setup thread and mutex interface creators:
  the_mutex_interface_t::set_creator(the_boost_mutex_t::create);
  the_thread_interface_t::set_creator(the_boost_thread_t::create);
  
  // parse the command line arguments:
  if (argc < 5) usage(argv);
  
  the_text_t fn_load;
  the_text_t fn_save;
  bool black_border = true;
  bool invert_image = false;
  
  for (int i = 1; i < argc; i++)
  {
    if (strcmp(argv[i], "-border") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -border option");
      
      i++;
      the_text_t border(argv[i]);
      if (border == "black") black_border = true;
      else if (border == "white") black_border = false;
      else usage(argv, "-border");
    }
    else if (strcmp(argv[i], "-invert") == 0)
    {
      invert_image = true;
    }
    else if (strcmp(argv[i], "-debug") == 0)
    {
      DEBUG_EDGE = true;
    }
    else if (strcmp(argv[i], "-load") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -load option");
      
      i++;
      fn_load = the_text_t(argv[i]);
    }
    else if (strcmp(argv[i], "-save") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -save option");
      
      i++;
      fn_save = the_text_t(argv[i]);
    }
    else
    {
      usage(argv,
	    the_text_t("unknown option \'") +
	    the_text_t(argv[i]) +
	    the_text_t("\'"));
    }
  }
  
  if (fn_load.size() == 0)
  {
    usage(argv, "must specify a file to open with the -load option");
  }
  
  if (fn_save.size() == 0)
  {
    usage(argv, "must specify a file to open with the -save option");
  }
  
  // read the input image:
  image_t::Pointer image = load<image_t>(fn_load);
  if (invert_image) invert<image_t>(image);
  
  image_t::SizeType image_sz = image->GetLargestPossibleRegion().GetSize();
  const unsigned int image_w = image_sz[0];
  const unsigned int image_h = image_sz[1];
  
  // setup the corner filters (even-integer dimensions):
  const unsigned int filter_hw = std::min(100u, std::max(2u, image_w / 2)) / 2;
  const unsigned int filter_hh = std::min(100u, std::max(2u, image_h / 2)) / 2;
  const unsigned int filter_w = filter_hw * 2;
  const unsigned int filter_h = filter_hh * 2;
  
  // helpers:
  unsigned int x_margin = std::min(image_w, std::max(filter_w, image_w / 12));
  unsigned int y_margin = std::min(image_h, std::max(filter_h, image_h / 12));
  
  // FIXME:
  double a[4] = { 0.0 };
  double b[4] = { 0.0 };
  double c[4] = { 0.0 };
  
  const char * ul_filter =
    black_border ?
    "----------++++++++++++++++++++++++++++++" :
    "++++++++++------------------------------";
  
  const char * lr_filter =
    black_border ?
    "++++++++++++++++++++++++++++++----------" :
    "------------------------------++++++++++";
  
  // left:
  detect_vertical_edge(a[0], b[0], c[0],
		       image,
		       ul_filter,
		       100,
		       0,
		       x_margin,
		       y_margin,
		       image_h - 2 * y_margin);
  cout << a[0] << ' ' << b[0] << ' ' << c[0] << endl;
  
  // right:
  detect_vertical_edge(a[1], b[1], c[1],
		       image,
		       lr_filter,
		       100,
		       image_w - x_margin - 1,
		       x_margin,
		       y_margin,
		       image_h - 2 * y_margin);
  cout << a[1] << ' ' << b[1] << ' ' << c[1] << endl;
  
  // top:
  detect_horizontal_edge(a[2], b[2], c[2],
			 image,
			 ul_filter,
			 100,
			 x_margin,
			 image_w - 2 * x_margin,
			 0,
			 y_margin);
  cout << a[2] << ' ' << b[2] << ' ' << c[2] << endl;
  
  // bottom:
  detect_horizontal_edge(a[3], b[3], c[3],
			 image,
			 lr_filter,
			 100,
			 int(2.0 * double(image_w) / 3.0),
			 image_w - int((0.06 + 2.0 / 3.0) * double(image_w)),
			 image_h - y_margin - 1,
			 y_margin);
  cout << a[3] << ' ' << b[3] << ' ' << c[3] << endl;
  
  unsigned char marker_color = black_border ? 255 : 0;
  
  if (DEBUG_EDGE)
  {
    unsigned char zebra[] = { 0x00, 0xFF };
    
    for (unsigned int x = 0; x < image_w; x++)
    {
      double y2 = -(a[2] * double(x) + c[2]) / b[2];
      double y3 = -(a[3] * double(x) + c[3]) / b[3];
      
      ix_t ix;
      ix[0] = x;
      ix[1] = (unsigned int)(floor(0.5 + y2));
      mark_pixel<image_t>(image,
			  ix,
			  zebra[((x / (RADIUS * 2 + 1)) % 10) / 5],
			  RADIUS);
      
      ix[1] = (unsigned int)(floor(0.5 + y3));
      mark_pixel<image_t>(image,
			  ix,
			  zebra[((x / (RADIUS * 2 + 1)) % 10) / 5],
			  RADIUS);
    }
    
    for (unsigned int y = 0; y < image_h; y++)
    {
      double x0 = -(b[0] * double(y) + c[0]) / a[0];
      double x1 = -(b[1] * double(y) + c[1]) / a[1];
      
      ix_t ix;
      ix[1] = y;
      ix[0] = (unsigned int)(floor(0.5 + x0));
      mark_pixel<image_t>(image,
			  ix,
			  zebra[((y / (RADIUS * 2 + 1)) % 10) / 5],
			  RADIUS);
      
      ix[0] = (unsigned int)(floor(0.5 + x1));
      mark_pixel<image_t>(image,
			  ix,
			  zebra[((y / (RADIUS * 2 + 1)) % 10) / 5],
			  RADIUS);
    }
    
    save<native_image_t>(cast<image_t, native_image_t>(image),
			 fn_save + the_text_t("-edges.tif"));
  }
  
  // find the corners, solve Mu = v for u:
  vnl_matrix<double> M(2, 2);
  vnl_vector<double> v(2);
  vnl_vector<double> u;
  
  // bottom-left corner:
  M[0][0] = a[0];
  M[0][1] = b[0];
  v[0] = -c[0];
  
  M[1][0] = a[2];
  M[1][1] = b[2];
  v[1]= -c[2];
  u = vnl_svd<double>(M).solve(v);
  
  pnt2d_t ix_ll;
  ix_ll[0] = (unsigned int)(floor(0.5 + u[0]));
  ix_ll[1] = (unsigned int)(floor(0.5 + u[1]));
  cout << "ll: " << ix_ll << endl;
  
  // bottom-right corner:
  M[0][0] = a[1];
  M[0][1] = b[1];
  v[0] = -c[1];
  u = vnl_svd<double>(M).solve(v);
  
  pnt2d_t ix_lr;
  ix_lr[0] = (unsigned int)(floor(0.5 + u[0]));
  ix_lr[1] = (unsigned int)(floor(0.5 + u[1]));
  cout << "lr: " << ix_lr << endl;

  // top-right corner:
  M[1][0] = a[3];
  M[1][1] = b[3];
  v[1] = -c[3];
  u = vnl_svd<double>(M).solve(v);
  
  pnt2d_t ix_ur;
  ix_ur[0] = (unsigned int)(floor(0.5 + u[0]));
  ix_ur[1] = (unsigned int)(floor(0.5 + u[1]));
  cout << "ur: " << ix_ur << endl;
  
  M[0][0] = a[0];
  M[0][1] = b[0];
  v[0] = -c[0];
  u = vnl_svd<double>(M).solve(v);
  
  // top-left corner:
  pnt2d_t ix_ul;
  ix_ul[0] = (unsigned int)(floor(0.5 + u[0]));
  ix_ul[1] = (unsigned int)(floor(0.5 + u[1]));
  cout << "ul: " << ix_ul << endl;
  
  pnt2d_t ll = index_to_physical(image, ix_ll);
  pnt2d_t lr = index_to_physical(image, ix_lr);
  pnt2d_t ur = index_to_physical(image, ix_ur);
  pnt2d_t ul = index_to_physical(image, ix_ul);
  
  // calculate the angle of rotation:
#if 0
  double alpha = 0.25 * (angle_between_lines(a[0], b[0], 1.0, 0.0) +
			 angle_between_lines(a[1], b[1], 1.0, 0.0) +
			 angle_between_lines(a[2], b[2], 0.0, 1.0) +
			 angle_between_lines(a[3], b[3], 0.0, 1.0));
#else
  double alpha = crooked(ix_ul, ix_ll, ix_ur, ix_lr);
#endif
  
  cout << "rotate by " << 180.0 * (alpha / M_PI) << " degrees" << endl;
  
  // use the shorter diagonal:
  double use_diag = (lr - ul).GetSquaredNorm() < (ll - ur).GetSquaredNorm();
  pnt2d_t pa = use_diag ? ul : ur;
  pnt2d_t pb = use_diag ? lr : ll;
  pnt2d_t center = pnt2d(0.5 * (pa[0] + pb[0]), 0.5 * (pa[1] + pb[1]));
  
  // rotate:
  typedef itk::FixedCenterOfRotationAffineTransform<double, 2> rotate_t;
  rotate_t::Pointer rotate = rotate_t::New();
  rotate->Rotate2D(-alpha);
  rotate->SetCenterOfRotationComponent(center);
  
  image_t::Pointer rotated = warp<image_t>(image, rotate);
  base_transform_t::Pointer rotate_inverse = rotate->GetInverse();
  assert(rotate_inverse.GetPointer() != NULL);
  
  ul = rotate_inverse->TransformPoint(ul);
  ur = rotate_inverse->TransformPoint(ur);
  lr = rotate_inverse->TransformPoint(lr);
  ll = rotate_inverse->TransformPoint(ll);
  
  if (DEBUG_EDGE)
  {
    mark<image_t>(rotated, ul, marker_color, 20);
    mark<image_t>(rotated, ur, marker_color, 20);
    mark<image_t>(rotated, lr, marker_color, 20);
    mark<image_t>(rotated, ll, marker_color, 20);
  }
  
  /*
  pa = rotate_inverse->TransformPoint(pa);
  pb = rotate_inverse->TransformPoint(pb);
  pnt2d_t min = pa;
  pnt2d_t max = pa;
  update_bbox(min, max, pb);
  */
  
  pnt2d_t min;
  min[0] = std::max(ll[0], ul[0]);
  min[1] = std::max(ll[1], lr[1]);
  
  pnt2d_t max;
  max[0] = std::min(lr[0], ur[0]);
  max[1] = std::min(ul[1], ur[1]);
  
  ix_t ix_min;
  ix_t ix_max;
  rotated->TransformPhysicalPointToIndex(min, ix_min);
  rotated->TransformPhysicalPointToIndex(max, ix_max);
  
  // try to trim some of the black frame remnants:
  ix_min[0] += 5;
  ix_min[1] += 5;
  ix_max[0] -= 5;
  ix_max[1] -= 5;
  
  cout << "min: " << min << " -> " << ix_min << endl
       << "max: " << max << " -> " << ix_max << endl;
  
  image_t::Pointer cropped = crop<image_t>(rotated, ix_min, ix_max);
  save<native_image_t>(cast<image_t, native_image_t>(cropped), fn_save);
    
  if (DEBUG_EDGE)
  {
    unsigned char zebra[] = { 0x00, 0xFF };
    
    ix_min[0] += 2 * (2 * RADIUS + 1);
    ix_min[1] += 2 * (2 * RADIUS + 1);
    ix_max[0] -= 2 * (2 * RADIUS + 1);
    ix_max[1] -= 2 * (2 * RADIUS + 1);
    
    for (int x = ix_min[0]; x <= ix_max[0]; x++)
    {
      ix_t ix;
      ix[0] = x;
      ix[1] = ix_min[1];
      // rotated->SetPixel(ix, zebra[(x % 20) / 10]);
      mark_pixel<image_t>(rotated,
			  ix,
			  zebra[((x / (RADIUS * 2 + 1)) % 20) / 10],
			  RADIUS);
      
      ix[1] = ix_max[1];
      // rotated->SetPixel(ix, zebra[(x % 20) / 10]);
      mark_pixel<image_t>(rotated,
			  ix,
			  zebra[((x / (RADIUS * 2 + 1)) % 20) / 10],
			  RADIUS);
    }
    
    for (int y = ix_min[1]; y <= ix_max[1]; y++)
    {
      ix_t ix;
      ix[1] = y;
      ix[0] = ix_min[0];
      // rotated->SetPixel(ix, zebra[(y % 20) / 10]);
      mark_pixel<image_t>(rotated,
			  ix,
			  zebra[((y / (RADIUS * 2 + 1)) % 20) / 10],
			  RADIUS);
      
      ix[0] = ix_max[0];
      // rotated->SetPixel(ix, zebra[(y % 20) / 10]);
      mark_pixel<image_t>(rotated,
			  ix,
			  zebra[((y / (RADIUS * 2 + 1)) % 20) / 10],
			  RADIUS);
    }
    
    save<native_image_t>(cast<image_t, native_image_t>(rotated),
			 fn_save + the_text_t("-rotated.tif"));
  }

  // done:
  return 0;
}
