// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : ir-blob.cxx
// Author       : Pavel A. Koshevoy
// Created      : 2006-06-26 15:03
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Blob enhancement algorithm -- classify local texture
//                properties, calculated as the inverse ratio of mean
//                pixel variance within a small local neighborhood (17x17),
//                and the median local mean pixel variance over the
//                entire image. The median is approximate, chosen from
//                a uniform grid of non-overlapping neighborhoods.
//                The ratio is thresholded (a good threshold value is 3).
//                The ratio is low when local variance exceeds the median,
//                and high when local variance is less than the median.
//                Thus, high variance regions (noisy texture regions, edges)
//                become dark, and low variance regions (smooth spots, blobs)
//                become light.

// system includes:
#include <stdlib.h>

// the includes:
#include <itk/common.hxx>
#include <utils/the_utils.hxx>
#include <thread/the_thread_pool.hxx>
#include <thread/the_boost_mutex.hxx>
#include <thread/the_boost_thread.hxx>


//----------------------------------------------------------------
// array2d
// 
#define array2d( T ) std::vector<std::vector<T> >


//----------------------------------------------------------------
// calc_variance
// 
// NOTE:
// The caller is responsible for pixel_cache memory management.
// The pixel_cache should be allocated to hold (2 * radius + 1)^2 pixels.
// 
void
calc_variance(pixel_t * pixel_cache,
	      const image_t * image,
	      const mask_t * mask,
	      const image_t::SizeType & image_size,
	      const image_t::IndexType & requested_center,
	      const int radius,
	      pixel_t & variance,
	      pixel_t & mean,
	      size_t & mass)
{
  variance = 0;
  mean = 0;
  mass = 0;
  
  typedef image_t::RegionType rn_t;
  typedef image_t::SizeType sz_t;
  typedef image_t::IndexType ix_t;

  ix_t ix;
  
  int x0 = requested_center[0] - radius;
  if (x0 < 0) x0 = 0;
  
  int x1 = x0 + radius * 2 + 1;
  if (size_t(x1) > image_size[0])
  {
    int d = x1 - image_size[0];
    x0 -= d;
    x1 -= d;
    if (x0 < 0) x0 = 0;
  }
  
  int y0 = requested_center[1] - radius;
  if (y0 < 0) y0 = 0;
  
  int y1 = y0 + radius * 2 + 1;
  if (size_t(y1) > image_size[1])
  {
    int d = y1 - image_size[1];
    y0 -= d;
    y1 -= d;
    if (y0 < 0) y0 = 0;
  }
  
  
  ix[0] = x0;
  ix[1] = y0;
  
  sz_t sz;
  sz[0] = x1 - x0;
  sz[1] = y1 - y0;
  
  rn_t rn;
  rn.SetIndex(ix);
  rn.SetSize(sz);

  typedef itk::ImageRegionConstIterator<image_t> img_iter_t;
  img_iter_t img_iter(image, rn);
  
  typedef itk::ImageRegionConstIterator<mask_t> msk_iter_t;
  msk_iter_t msk_iter(mask, rn);
  
  // calculate the mean:
  for (img_iter.GoToBegin(), msk_iter.GoToBegin();
       !img_iter.IsAtEnd(); ++img_iter, ++msk_iter)
  {
    if (!msk_iter.Get()) continue;
    
    pixel_t p = img_iter.Get();
    mean += p;
    
    pixel_cache[mass] = p;
    mass++;
  }
  
  if (!mass) return;
  mean /= float(mass);
  
  // calculate variance:
  for (size_t i = 0; i < mass; i++)
  {
    pixel_t d = pixel_cache[i] - mean;
    variance += d * d;
  }
  
  variance /= float(mass);
}


//----------------------------------------------------------------
// calc_variance_transaction_t
// 
class calc_variance_transaction_t : public the_transaction_t
{
public:
  calc_variance_transaction_t(image_t::ConstPointer image,
			      mask_t::ConstPointer mask,
			      const int radius,
			      const image_t::SizeType & image_size,
			      const std::list<image_t::IndexType> & roi,
			      image_t::Pointer output):
    m_image(image),
    m_mask(mask),
    m_radius(radius),
    m_image_size(image_size),
    m_roi(roi),
    m_output(output)
  {
    const int cell_size = radius * 2 + 1;
    const size_t cache_size = cell_size * cell_size;
    m_pixel_cache.resize(cache_size);
  }
  
  // virtual:
  void execute(the_thread_interface_t * thread)
  {
    // shortcuts:
    pixel_t * pixel_cache = &m_pixel_cache[0];
    const image_t * image = m_image.GetPointer();
    const mask_t * mask = m_mask.GetPointer();
    
    for (std::list<image_t::IndexType>::const_iterator itex = m_roi.begin();
	 itex != m_roi.end(); ++itex)
    {
      const image_t::IndexType & center = *itex;
      
      pixel_t variance = 0;
      pixel_t mean = 0;
      size_t mass = 0;
      calc_variance(pixel_cache,
		    image,
		    mask,
		    m_image_size,
		    center,
		    m_radius,
		    variance,
		    mean,
		    mass);
      if (mass)
      {
	m_output->SetPixel(center, variance);
      }
    }
  }
  
  // image:
  image_t::ConstPointer m_image;
  
  // image mask:
  mask_t::ConstPointer m_mask;
  
  // neighborhood window radius:
  const int m_radius;
  
  // image size:
  image_t::SizeType m_image_size;
  
  // region of interest:
  std::list<image_t::IndexType> m_roi;
  
  // output variance image:
  image_t::Pointer m_output;
  
  // neighborhood pixel cache:
  std::vector<pixel_t> m_pixel_cache;
};


//----------------------------------------------------------------
// qsort_pixel_comparator
// 
static int
qsort_pixel_comparator(const void * a, const void * b)
{
  const pixel_t * p0 = (const pixel_t *)a;
  const pixel_t * p1 = (const pixel_t *)b;
  pixel_t d = (*p0 - *p1);
  return (d < 0) ? -1 : (d > 0 ? 1 : 0);
}

//----------------------------------------------------------------
// enhance_blobs
// 
// for each pixel calculate its neignborhood variance, then
// calculate the output pixels as the ratio of local
// neignborhood variance to global median variance
// 
image_t::Pointer
enhance_blobs(const image_t * image,
	      const mask_t * mask,
	      const int radius,
	      const double & threshold,
	      unsigned int num_threads)
{
  const float float_max = std::numeric_limits<float>::max();
  
  image_t::RegionType rn = image->GetLargestPossibleRegion();
  image_t::SizeType sz = rn.GetSize();
  typedef image_t::IndexType ix_t;
  
  image_t::Pointer variance = cast<image_t, image_t>(image);
  variance->FillBuffer(float_max);
  
  the_thread_pool_t thread_pool(num_threads);
  thread_pool.set_idle_sleep_duration(50); // 50 usec
  std::list<the_transaction_t *> schedule;
  
  std::list<ix_t> roi;
  size_t roi_pixel_count = 0;
  const size_t pixels_per_transaction = std::min<size_t>(64, sz[0]);
  
  typedef itk::ImageRegionConstIteratorWithIndex<image_t> citex_t;
  citex_t citex(image, rn);
  for (citex.GoToBegin(); !citex.IsAtEnd(); ++citex)
  {
    const ix_t & ix = citex.GetIndex();
    roi.push_back(ix);
    roi_pixel_count++;
    if (roi_pixel_count % pixels_per_transaction == 0)
    {
      calc_variance_transaction_t * t =
	new calc_variance_transaction_t(image,
					mask,
					radius,
					sz,
					roi,
					variance);
      schedule.push_back(t);
      
      roi.clear();
      roi_pixel_count = 0;
    }
  }
  
  if (roi_pixel_count)
  {
    calc_variance_transaction_t * t =
      new calc_variance_transaction_t(image,
				      mask,
				      radius,
				      sz,
				      roi,
				      variance);
    schedule.push_back(t);
    
    roi.clear();
    roi_pixel_count = 0;
  }
  
  thread_pool.push_back(schedule);
  thread_pool.pre_distribute_work();
  thread_pool.start();
  thread_pool.wait();
  
  // find the median:
  std::vector<pixel_t> variance_vector(sz[0] * sz[1]);
  pixel_t * variance_pixels = &variance_vector[0];
  unsigned int variance_pixel_count = 0;
  
  // first, get rid of masked out variance pixels:
  typedef itk::ImageRegionConstIterator<image_t> citer_t;
  citer_t citer(variance, rn);
  for (citer.GoToBegin(); !citer.IsAtEnd(); ++citer)
  {
    pixel_t v = citer.Get();
    if (v != float_max)
    {
      variance_pixels[variance_pixel_count] = v;
      variance_pixel_count++;
    }
  }
  
  qsort(variance_pixels,
	variance_pixel_count,
	sizeof(pixel_t),
	&qsort_pixel_comparator);
  
  const pixel_t global_variance_median =
    variance_pixels[variance_pixel_count / 2];
  
  cout << "global variance median: " << global_variance_median << endl
       << endl;
  
  // enhance the blobs:
  pixel_t blob_image_mean = 0;
  size_t blob_image_mass = 0;
  
  typedef itk::ImageRegionIterator<image_t> iter_t;
  iter_t iter(variance, rn);
  for (iter.GoToBegin(); !iter.IsAtEnd(); ++iter)
  {
    pixel_t v = iter.Get();
    if (v != float_max)
    {
      double metric =
	std::min<double>(threshold,
			 double(global_variance_median + 1) /
			 double(v + 1));
      iter.Set(metric);
      
      blob_image_mean += metric;
      blob_image_mass++;
    }
  }
  
  if (blob_image_mass)
  {
    blob_image_mean /= pixel_t(blob_image_mass);
  }
  
  for (iter.GoToBegin(); !iter.IsAtEnd(); ++iter)
  {
    pixel_t v = iter.Get();
    if (v == float_max)
    {
      iter.Set(blob_image_mean);
    }
  }
  
  return variance;
}


//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL)
{
  cerr << "USAGE: " << argv[0] << " "
       << "[-sh shrink_factor] "
       << "[-r radius_in_pixels] "
       << "[-median radius] "
       << "[-max threshold] "
       << "[-std_mask] "
       << "[-mask mask_in] "
       << "[-threads number_of_threads] "
       << "-load image_in "
       << "-save image_out"
       << endl
       << "EXAMPLE: " << argv[0] << " "
       << "-sh 8 "
       << "-r 8 "
       << "-median 2 "
       << "-max 3 "
       << "-mask mask/s10.tif "
       << "-threads 8 "
       << "-load raw/s10.tif "
       << "-save blobs/s10.tif "
       << endl;
  
  if (message != NULL)
  {
    cerr << "ERROR: " << message << endl;
  }
  
  exit(1);
}


//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  // this is so that the printouts look better:
  cout.precision(6);
  cout.setf(ios::scientific);
  
  // setup thread storage for the main thread:
  set_the_thread_storage_provider(the_boost_thread_t::thread_storage);
  the_boost_thread_t MAIN_THREAD_DUMMY;
  MAIN_THREAD_DUMMY.set_stopped(false);
  
  // setup thread and mutex interface creators:
  the_mutex_interface_t::set_creator(the_boost_mutex_t::create);
  the_thread_interface_t::set_creator(the_boost_thread_t::create);
  
  if (argc < 5) usage(argv);
  the_text_t fn_mask;
  the_text_t fn_load;
  the_text_t fn_save;
  
  unsigned int shrink_factor = 1;
  int radius = 2;
  double threshold = 3;
  unsigned int median_radius = 0;
  bool use_std_mask = false;
  mask_t::Pointer image_mask;
  
  // by default run as many threads as there are cores:
  unsigned int num_threads = boost::thread::hardware_concurrency();
  
  for (int i = 1; i < argc; i++)
  {
    bool ok = true;
    
    if (strcmp(argv[i], "-sh") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -sh option");
      
      i++;
      shrink_factor = the_text_t(argv[i]).toUInt(&ok);
      if (!ok || shrink_factor == 0) usage(argv, "bad -sh option");
    }
    else if (strcmp(argv[i], "-r") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -r option");
      
      i++;
      radius = the_text_t(argv[i]).toInt(&ok);
      if (!ok || radius <= 0) usage(argv, "bad -r option");
    }
    else if (strcmp(argv[i], "-median") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -median option");
      
      i++;
      median_radius = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -median option");
    }
    else if (strcmp(argv[i], "-max") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -max option");
      
      i++;
      threshold = the_text_t(argv[i]).toDouble(&ok);
      if (!ok || threshold <= 0) usage(argv, "bad -max option");
    }
    else if (strcmp(argv[i], "-std_mask") == 0)
    {
      use_std_mask = true;
    }
    else if (strcmp(argv[i], "-mask") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -mask option");
      
      i++;
      fn_mask = the_text_t(argv[i]);
    }
    else if (strcmp(argv[i], "-threads") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -threads option");
      
      i++;
      num_threads = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -threads option");
    }
    else if (strcmp(argv[i], "-load") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -load option");
      
      i++;
      fn_load = the_text_t(argv[i]);
    }
    else if (strcmp(argv[i], "-save") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -save option");
      
      i++;
      fn_save = the_text_t(argv[i]);
    }
    else
    {
      usage(argv,
	    the_text_t("unknown option \'") +
	    the_text_t(argv[i]) +
	    the_text_t("\'"));
    }
  }
  
  if (fn_load.size() == 0)
  {
    usage(argv, "must specify a file to open with the -load option");
  }
  
  if (fn_save.size() == 0)
  {
    usage(argv, "must specify a file to open with the -save option");
  }
  
  cout << "neighborhood:  " << radius * 2 + 1 << endl
       << "threads:       " << num_threads << endl
       << endl;
  
  // load an image:
  image_t::Pointer image = std_tile<image_t>(fn_load, shrink_factor, 1);
  
  // setup the mask:
  if (fn_mask.size() > 0)
  {
    image_mask = std_tile<mask_t>(fn_mask, shrink_factor, 1);
  }
  else
  {
    image_mask = std_mask<image_t>(image, use_std_mask);
  }
  
  bool remap_values =
    !fn_save.match_tail(".mha") &&
    !fn_save.match_tail(".mhd") &&
    !fn_save.match_tail(".nrrd");
  
  // filter the image with a median filter:
  if (median_radius != 0)
  {
    image = median<image_t>(image, median_radius);
  }
  
  image_t::Pointer blobs = enhance_blobs(image,
					 image_mask,
					 radius,
					 threshold,
					 num_threads);
  
  blobs = normalize<image_t>(blobs, 1, 1, 0, 255, image_mask);
  
  if (remap_values)
  {
    save<native_image_t>(cast<image_t, native_image_t>(blobs), fn_save);
  }
  else
  {
    save<image_t>(blobs, fn_save);
  }
  
  // done:
  return 0;
}
