// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : ir-clahe.cxx
// Author       : Pavel A. Koshevoy
// Created      : 2005/09/30 12:27
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Contrast Limited Adaptive Histogram Equalization

// the includes:
#include <itk/common.hxx>
#include <thread/the_boost_mutex.hxx>
#include <thread/the_boost_thread.hxx>

// system includes:
#include <math.h>
#include <algorithm>

// ITK includes:
#include <itkMedianImageFilter.h>


//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL)
{
  cerr << "USAGE: " << argv[0] << " "
       << "[-sh shrink_factor] "
       << "[-bins num_bins] "
       << "[-window nx ny] "
       << "[-slope max] "
       << "[-remap min max] "
       << "[-smooth sigma] "
       << "[-median radius] "
       << "[-std_mask] "
       << "[-mask mask_in] "
       << "-load image_in "
       << "-save image_out"
       << endl
       << "EXAMPLE: " << argv[0] << " "
       << "-bins 256 "
       << "-window 255 255 "
       << "-median 3 "
       << "-remap 0 255 "
       << "-slope 3.0 "
       << "-load in/48755.tif -save he/48755.tif "
       << endl;
  
  if (message != NULL)
  {
    cerr << "ERROR: " << message << endl;
  }
  
  exit(1);
}


//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  // this is so that the printouts look better:
  cout.precision(2);
  cout.setf(ios::scientific);
  
  // setup thread storage for the main thread:
  set_the_thread_storage_provider(the_boost_thread_t::thread_storage);
  the_boost_thread_t MAIN_THREAD_DUMMY;
  MAIN_THREAD_DUMMY.set_stopped(false);
  
  // setup thread and mutex interface creators:
  the_mutex_interface_t::set_creator(the_boost_mutex_t::create);
  the_thread_interface_t::set_creator(the_boost_thread_t::create);
  
  // parse the command line arguments:
  if (argc < 5) usage(argv);
  the_text_t fn_mask;
  the_text_t fn_load;
  the_text_t fn_save;
  
  unsigned int shrink_factor = 1;
  unsigned int bins = 256;
  unsigned int nx = ~0;
  unsigned int ny = ~0;
  double max_slope = 0.0;
  pixel_t new_min = std::numeric_limits<pixel_t>::max();
  pixel_t new_max = -new_min;
  double sigma = std::numeric_limits<double>::max();
  unsigned int median_radius = 0;
  bool use_std_mask = false;
  mask_t::Pointer mask;
  
  for (int i = 1; i < argc; i++)
  {
    bool ok = true;
    
    if (strcmp(argv[i], "-sh") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -sh option");
      
      i++;
      shrink_factor = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -sh option");
    }
    else if (strcmp(argv[i], "-window") == 0)
    {
      if ((argc - i) <= 2) usage(argv, "could not parse -window option");
      
      i++;
      nx = the_text_t(argv[i]).toUInt(&ok);
      if (!ok || nx < 2) usage(argv, "bad -window nx option");
      
      i++;
      ny = the_text_t(argv[i]).toUInt(&ok);
      if (!ok || ny < 2) usage(argv, "bad -window ny option");
    }
    else if (strcmp(argv[i], "-slope") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -slope option");
      
      i++;
      max_slope = the_text_t(argv[i]).toDouble(&ok);
      if (!ok || max_slope <= 1.0) usage(argv, "bad -slope option");
    }
    else if (strcmp(argv[i], "-median") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -median option");
      
      i++;
      median_radius = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -window median option");
    }
    else if (strcmp(argv[i], "-std_mask") == 0)
    {
      use_std_mask = true;
    }
    else if (strcmp(argv[i], "-bins") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -bins option");
      
      i++;
      bins = the_text_t(argv[i]).toUInt(&ok);
      if (!ok || bins < 2) usage(argv, "bad -window bins option");
    }
    else if (strcmp(argv[i], "-remap") == 0)
    {
      if ((argc - i) <= 2) usage(argv, "could not parse -remap option");
      
      i++;
      new_min = pixel_t(the_text_t(argv[i]).toDouble(&ok));
      if (!ok) usage(argv, "bad -window new_min option");
      
      i++;
      new_max = pixel_t(the_text_t(argv[i]).toDouble(&ok));
      if (!ok) usage(argv, "bad -window new_max option");
    }
    else if (strcmp(argv[i], "-smooth") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -smooth option");
      
      i++;
      sigma = the_text_t(argv[i]).toDouble(&ok);
      if (!ok || sigma <= 0.0) usage(argv, "bad -smooth option");
    }
    else if (strcmp(argv[i], "-mask") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -mask option");
      
      i++;
      fn_mask = the_text_t(argv[i]);
    }
    else if (strcmp(argv[i], "-load") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -load option");
      
      i++;
      fn_load = the_text_t(argv[i]);
    }
    else if (strcmp(argv[i], "-save") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -save option");
      
      i++;
      fn_save = the_text_t(argv[i]);
    }
    else
    {
      usage(argv,
	    the_text_t("unknown option \'") +
	    the_text_t(argv[i]) +
	    the_text_t("\'"));
    }
  }
  
  if (fn_load.size() == 0)
  {
    usage(argv, "must specify a file to open with the -load option");
  }
  
  if (fn_save.size() == 0)
  {
    usage(argv, "must specify a file to open with the -save option");
  }
  
  // read the input image:
  image_t::Pointer image = std_tile<image_t>(fn_load, shrink_factor, 1);
  if (use_std_mask)
  {
    mask = std_mask<image_t>(image);
  }
  else if (fn_mask.size() > 0)
  {
    mask = std_tile<mask_t>(fn_mask, shrink_factor, 1);
  }
  
  image_t::SizeType sz = image->GetLargestPossibleRegion().GetSize();
  if (nx == ~0u) nx = sz[0];
  if (ny == ~0u) ny = sz[1];
  
  // filter the image with a median filter:
  if (median_radius != 0)
  {
    image = median<image_t>(image, median_radius);
  }
  
  // smooth the image prior to histogram equalization:
  if (sigma != std::numeric_limits<double>::max())
  {
    image = smooth<image_t>(image, sigma);
  }
  
  image_t::Pointer out =
    CLAHE<image_t>(image, nx, ny, max_slope, bins, new_min, new_max, mask);
  
  bool remap_values = (fn_save.match_tail(".png") ||
		       fn_save.match_tail(".tif") ||
		       fn_save.match_tail(".bmp"));
  if (remap_values)
  {
    save<native_image_t>(cast<image_t, native_image_t>(out), fn_save);
  }
  else
  {
    save<image_t>(out, fn_save);
  }
  
  // done:
  return 0;
}
