// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : ir-diffusion.cxx
// Author       : Tolga Tasdizen, ported by Pavel Koshevoy
// Created      : Wed Nov 22 14:40:00 MDT 2006
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : This is an ITK port of Tolgas diffusion filter.

// the includes:
#include <itk/common.hxx>
#include <utils/the_text.hxx>

// ITK includes:
#include <itkImageFileReader.h>
#include <itkImageFileWriter.h>
#include <itkCastImageFilter.h>
#include <itkDiscreteGaussianImageFilter.h>

// system includes:
#include <iostream>
#include <iomanip>
#include <cstdlib>
#include <vector>
#include <list>
#include <limits>

// namespace access:
using std::cout;
using std::cerr;
using std::endl;
using std::setw;
using std::ios;
using std::flush;

#define SAVE_IT 10

//----------------------------------------------------------------
// vector_t
// 
typedef itk::Vector<pixel_t, 2> vector_t;

//----------------------------------------------------------------
// vector
// 
inline static const vector_t
vector(const pixel_t & x, const pixel_t & y)
{
  vector_t vc;
  vc[0] = x;
  vc[1] = y;
  return vc;
}

//----------------------------------------------------------------
// vector_image_t
// 
typedef itk::Image<vector_t, 2> vector_image_t;

//----------------------------------------------------------------
// tensor_t
// 
typedef itk::Matrix<pixel_t, 2, 2> tensor_t;

//----------------------------------------------------------------
// tensor
// 
inline static const tensor_t
tensor(const pixel_t & aNN = pixel_t(0))
{
  tensor_t t;
  t(0, 0) = aNN;
  t(0, 1) = aNN;
  t(1, 0) = aNN;
  t(1, 1) = aNN;
  return t;
}

//----------------------------------------------------------------
// transpose
// 
inline static const tensor_t
transpose(const tensor_t & in)
{
  tensor_t t(in);
  std::swap(t(0, 1), t(1, 0));
  return t;
}

//----------------------------------------------------------------
// tensor
// 
inline static const tensor_t
tensor(const pixel_t & a00,
       const pixel_t & a01,
       const pixel_t & a10,
       const pixel_t & a11)
{
  tensor_t t;
  t(0, 0) = a00;
  t(0, 1) = a01;
  t(1, 0) = a10;
  t(1, 1) = a11;
  return t;
}

//----------------------------------------------------------------
// exterior
// 
// Compute exterior product of 2 vectors
// 
inline static const tensor_t
exterior(const vector_t & a,
	 const vector_t & b)
{
  tensor_t t;
  t(0, 0) = a[0] * b[0];
  t(0, 1) = a[0] * b[1];
  t(1, 0) = a[1] * b[0];
  t(1, 1) = a[1] * b[1];
  return t;
}

//----------------------------------------------------------------
// tensor_image_t
// 
typedef itk::Image<tensor_t, 2> tensor_image_t;

//----------------------------------------------------------------
// ix
// 
inline static const image_t::IndexType
ix(int x, int y)
{
  image_t::IndexType i;
  i[0] = x;
  i[1] = y;
  return i;
}

//----------------------------------------------------------------
// gaussDiffuse
// 
template <typename TImage>
typename TImage::Pointer
gaussDiffuse(const TImage * in, double rho)
{
  return smooth<TImage>(in, rho, 0.001);
}

//----------------------------------------------------------------
// gauss_matrix_image
// 
tensor_image_t::Pointer
gauss_matrix_image(const tensor_image_t::Pointer & in, float rho)
{
  tensor_image_t::SizeType sz = in->GetLargestPossibleRegion().GetSize();

  int r = tensor_t::RowDimensions;
  int c = tensor_t::ColumnDimensions;
  
  tensor_image_t::Pointer out = make_image<tensor_image_t>(sz, tensor(0));
  image_t::Pointer component = make_image<image_t>(sz);
  tensor_t temp;
  tensor_image_t::IndexType ix;
  
  // for each component
  for (int i = 0; i < r; i++)
  {
    for (int j = 0; j < c; j++)
    {
      typedef itk::ImageRegionIteratorWithIndex<image_t> itex_t;
      itex_t itex(component, component->GetLargestPossibleRegion());
      for (itex.GoToBegin(); !itex.IsAtEnd(); ++itex)
      {
	ix = itex.GetIndex();
	tensor_t t = in->GetPixel(ix);
	itex.Set(t(i, j));
      }
      
      image_t::Pointer scomponent = gaussDiffuse<image_t>(component, rho);
      
      itex = itex_t(scomponent, scomponent->GetLargestPossibleRegion());
      for (itex.GoToBegin(); !itex.IsAtEnd(); ++itex)
      {
	ix = itex.GetIndex();
	tensor_t t = in->GetPixel(ix);
	t(i, j) = itex.Get();
	out->SetPixel(ix, t);
      }
    }
  }
  
  return out;
}

//----------------------------------------------------------------
// ComputeStructureTensorImage
// 
tensor_image_t::Pointer
ComputeStructureTensorImage(const vector_image_t::Pointer & grad_image,
			    float rho)
{
  tensor_image_t::Pointer J0 =
    make_image<tensor_image_t>
    (grad_image->GetLargestPossibleRegion().GetSize(), tensor(0.0f));
  
  typedef itk::ImageRegionIteratorWithIndex<vector_image_t> itex_t;
  itex_t itex(grad_image, grad_image->GetLargestPossibleRegion());
  for (itex.GoToBegin(); !itex.IsAtEnd(); ++itex)
  { 
    vector_t grad = itex.Get();
    tensor_t t = exterior(grad, grad);
    J0->SetPixel(itex.GetIndex(), t);
  }
  
  if (rho > 0.0f)
  {
    return gauss_matrix_image(J0, rho);
  }
  
  return J0;
}

//----------------------------------------------------------------
// ComputeHessianImage
// 
tensor_image_t::Pointer
ComputeHessianImage(image_t::Pointer & in, float sigma)
{
  image_t::SizeType sz = in->GetLargestPossibleRegion().GetSize();
  int w = sz[0];
  int h = sz[1];
  
  image_t::Pointer image = in;
  if (sigma > 0.0f)
  {
    image = gaussDiffuse<image_t>(in, sigma);
  }
  
  tensor_t H = tensor(0.0f);
  tensor_image_t::Pointer Himage = make_image<tensor_image_t>(sz, H);
  pixel_t pixel[3][3];
  image_t::IndexType tmp;
  
  typedef itk::ImageRegionIteratorWithIndex<image_t> itex_t;
  itex_t itex(image, image->GetLargestPossibleRegion());
  for (itex.GoToBegin(); !itex.IsAtEnd(); ++itex)
  {
    image_t::IndexType ix = itex.GetIndex();
    
    // sample the neighbors:
    for (int dy = -1; dy <= 1; dy++)
    {
      int y = std::min(h - 1, std::max(0, int(ix[1] + dy)));
      tmp[1] = y;
      
      for (int dx = -1; dx <= 1; dx++)
      {
	int x = std::min(w - 1, std::max(0, int(ix[0] + dx)));
	tmp[0] = x;
	
	pixel[dx + 1][dy + 1] = image->GetPixel(tmp);
      }
    }
    
    H(0, 0) = pixel[2][1] + pixel[0][1] - 2 * pixel[1][1];
    H(1, 1) = pixel[1][2] + pixel[1][0] - 2 * pixel[1][1];
    H(0, 1) = 0.25 * ((pixel[0][0] + pixel[2][2]) -
		      (pixel[0][2] + pixel[2][0]));
    H(1, 0) = H(0, 1);
    Himage->SetPixel(ix, H);
  }
  
  return Himage;
}

//----------------------------------------------------------------
// ComputeGradientImage
// 
vector_image_t::Pointer
ComputeGradientImage(image_t::Pointer & in,
		     int bctype,
		     float sigma)
{
  image_t::RegionType rn = in->GetLargestPossibleRegion();
  image_t::SizeType sz = rn.GetSize();
  int w = sz[0];
  int h = sz[1];
  
  image_t::Pointer image = in;
  if (sigma > 0.0)
  {
    image = gaussDiffuse<image_t>(in, sigma);
  }
  
  vector_image_t::Pointer grad = make_image<vector_image_t>(sz, vector(0, 0));
  {
    image_t::RegionType r = rn;
    image_t::IndexType r_ix = r.GetIndex();
    image_t::SizeType r_sz = r.GetSize();
    r_ix[0]++;
    r_ix[1]++;
    r_sz[0] -= 2;
    r_sz[1] -= 2;
    r.SetIndex(r_ix);
    r.SetSize(r_sz);
    
    int stencil[4][2] = {
      {  1,  0 },
      { -1,  0 },
      {  0,  1 },
      {  0, -1 }
    };
    
    pixel_t pixel[4];
    image_t::IndexType tmp;
    
    typedef itk::ImageRegionIteratorWithIndex<image_t> itex_t;
    itex_t itex(image, r);
    for (itex.GoToBegin(); !itex.IsAtEnd(); ++itex)
    {
      image_t::IndexType ix = itex.GetIndex();
      
      for (int i = 0; i < 4; i++)
      {
	tmp[0] = ix[0] + stencil[i][0];
	tmp[1] = ix[1] + stencil[i][1];
	pixel[i] = image->GetPixel(tmp);
      }

      grad->SetPixel(ix, vector(0.5 * (pixel[0] - pixel[1]),
				0.5 * (pixel[2] - pixel[3])));
    }
  }
  
  // if (bctype==1)
  {
    for (int x = 1; x < w - 1; x++)
    {
      grad->SetPixel(ix(x, 0),
		     vector(0.5 * (image->GetPixel(ix(x + 1, 0)) -
				   image->GetPixel(ix(x - 1, 0))),
			    image->GetPixel(ix(x, 1)) -
			    image->GetPixel(ix(x, 0))));
      
      grad->SetPixel(ix(x, h - 1),
		     vector(0.5 * (image->GetPixel(ix(x + 1, h - 1)) -
				   image->GetPixel(ix(x - 1, h - 1))),
			    image->GetPixel(ix(x, h - 1)) -
			    image->GetPixel(ix(x, h - 2))));
    }
    
    for (int y = 1; y < h - 1; y++)
    {
      grad->SetPixel(ix(0, y),
		     vector(image->GetPixel(ix(1, y)) -
			    image->GetPixel(ix(0, y)),
			    0.5 * (image->GetPixel(ix(0, y + 1)) -
				   image->GetPixel(ix(0, y - 1)))));
      
      grad->SetPixel(ix(w - 1, y),
		     vector(image->GetPixel(ix(w - 1, y)) -
			    image->GetPixel(ix(w - 2, y)),
			    0.5 * (image->GetPixel(ix(w - 1, y + 1)) -
				   image->GetPixel(ix(w - 1, y - 1)))));
    }
    
    grad->SetPixel(ix(0, 0),
		   vector(image->GetPixel(ix(1, 0)) -
			  image->GetPixel(ix(0, 0)),
			  image->GetPixel(ix(0, 1)) -
			  image->GetPixel(ix(0, 0))));
    
    grad->SetPixel(ix(w - 1, 0),
		   vector(image->GetPixel(ix(w - 1, 0)) -
			  image->GetPixel(ix(w - 2, 0)),
			  image->GetPixel(ix(w - 1, 1)) -
			  image->GetPixel(ix(w - 1, 0))));
    
    grad->SetPixel(ix(0, h - 1),
		   vector(image->GetPixel(ix(1, h - 1)) -
			  image->GetPixel(ix(0, h - 1)),
			  image->GetPixel(ix(0, h - 1)) -
			  image->GetPixel(ix(0, h - 2))));
    
    grad->SetPixel(ix(w - 1, h - 1),
		   vector(image->GetPixel(ix(w - 1, h - 1)) -
			  image->GetPixel(ix(w - 2, h - 1)),
			  image->GetPixel(ix(w - 1, h - 1)) -
			  image->GetPixel(ix(w - 1, h - 2))));
  }
  
  return grad;
}

//----------------------------------------------------------------
// ComputeSortedEigenValues_2x2
// 
vector_image_t::Pointer
ComputeSortedEigenValues_2x2(tensor_image_t::Pointer & T_image)
{
  tensor_image_t::SizeType sz = T_image->GetLargestPossibleRegion().GetSize();
  
  vector_image_t::Pointer eigenvalue =
    make_image<vector_image_t>(sz, vector(0, 0));
  
  typedef itk::ImageRegionIteratorWithIndex<tensor_image_t> itex_t;
  itex_t itex(T_image, T_image->GetLargestPossibleRegion());
  for (itex.GoToBegin(); !itex.IsAtEnd(); ++itex)
  {
    tensor_t T = itex.Get();
    tensor_image_t::IndexType ix = itex.GetIndex();
    
    float dif = T(0, 0) - T(1, 1);
    float det = 0.5 * sqrt(dif * dif + 4.0 * T(0, 1) * T(0, 1));
    float m = 0.5 * (T(0, 0) + T(1, 1));
    float l1 = m + det;
    float l2 = m - det;
    
    if (fabs(l1) > fabs(l2))
    {
      eigenvalue->SetPixel(ix, vector(l1, l2));
    }
    else
    {
      eigenvalue->SetPixel(ix, vector(l2, l1));
    }
  }
  
  return eigenvalue;
}

//----------------------------------------------------------------
// compute_2x2_eigenvector
// 
void
compute_2x2_eigenvector(const tensor_t & M,
			float l,
			float * evec,
			int flag = 0)
{
  float num, theta;
  
  if (fabs(M(0, 0)) > fabs(M(1, 1)))
  {
    num = l - M(0, 0);
    theta = atan2(num, M(0, 1));
  }
  else
  {
    num = l - M(1, 1);
    theta = atan2(M(0, 1), num);
  }
  
  evec[0] = cos(theta);
  evec[1] = sin(theta);
  
  if (flag)
  {
    cout << theta << ' ' << evec[0] << ' ' << evec[1] << endl;
  }
}

//----------------------------------------------------------------
// ComputeSortedEigenVectors_2x2
// 
vector_image_t::Pointer
ComputeSortedEigenVectors_2x2(tensor_image_t::Pointer & T_image,
			      tensor_image_t::Pointer & evec_image)
{
  tensor_image_t::SizeType sz = T_image->GetLargestPossibleRegion().GetSize();
  float temp[2];
  
  vector_image_t::Pointer eigenvalue =
    make_image<vector_image_t>(sz, vector(0, 0));
  
  typedef itk::ImageRegionIteratorWithIndex<tensor_image_t> itex_t;
  itex_t itex(T_image, T_image->GetLargestPossibleRegion());
  for (itex.GoToBegin(); !itex.IsAtEnd(); ++itex)
  {
    tensor_t T = itex.Get();
    tensor_image_t::IndexType ix = itex.GetIndex();
    
    float dif = T(0, 0) - T(1, 1);
    float det = 0.5 * sqrt(dif * dif + 4.0 * T(0, 1) * T(0, 1));
    float m = 0.5 * (T(0, 0) + T(1, 1));
    float l1 = m + det;
    float l2 = m - det;
    
    tensor_t evec;
    if (fabs(l1) > fabs(l2)) 
    {
      eigenvalue->SetPixel(ix, vector(l1, l2));
      
      // column 0 is l1
      compute_2x2_eigenvector(T, l1, temp);
      evec(0, 0) = temp[0];
      evec(1, 0) = temp[1];
      evec(0, 1) = -temp[1];
      evec(1, 1) = temp[0];
    }
    else  
    {
      eigenvalue->SetPixel(ix, vector(l2, l1));
      
      // column 0 is l2
      compute_2x2_eigenvector(T, l2, temp);
      evec(0, 0) = temp[0];
      evec(1, 0) = temp[1];
      evec(0, 1) = -temp[1];
      evec(1, 1) = temp[0];
    }
    evec_image->SetPixel(ix, evec);
  }
  
  return eigenvalue;
}

//----------------------------------------------------------------
// ComputeScalarDiffusionCoefficient
// 
image_t::Pointer
ComputeScalarDiffusionCoefficient(vector_image_t::Pointer & im, float T)
{
  vector_image_t::SizeType sz = im->GetLargestPossibleRegion().GetSize();
  image_t::Pointer r = make_image<image_t>(sz);
  
  pixel_t min = std::numeric_limits<pixel_t>::max();
  pixel_t max = -min;
  pixel_t avg = pixel_t(0);
  pixel_t num = pixel_t(0);
  
  typedef itk::ImageRegionIteratorWithIndex<vector_image_t> itex_t;
  itex_t itex(im, im->GetLargestPossibleRegion());
  for (itex.GoToBegin(); !itex.IsAtEnd(); ++itex)
  {
    vector_image_t::IndexType ix = itex.GetIndex();
    vector_t vec = itex.Get();
    
    float dif = fabs(vec[0]) - fabs(vec[1]);
    dif *= dif;

    if (dif > 0.0)
    {
      r->SetPixel(ix, exp(-1.0 * T / dif));
    }
    else
    {
      r->SetPixel(ix, 0.0f);
    }
    
    pixel_t p = r->GetPixel(ix);
    min = std::min(min, p);
    max = std::max(max, p);
    avg += p;
    num += 1.0f;
  }
  
  if (num > 0.0f)
  {
    avg /= num;
  }
  
  cout << min << ' ' << avg << ' ' << max << endl;
  return r;
}

//----------------------------------------------------------------
// angle_image
// 
image_t::Pointer
angle_image(tensor_image_t::Pointer & in)
{
  static const float TINY = 1.0e-6;
  
  tensor_image_t::SizeType sz = in->GetLargestPossibleRegion().GetSize();
  image_t::Pointer ang = make_image<image_t>(sz);
  
  typedef itk::ImageRegionIteratorWithIndex<tensor_image_t> itex_t;
  itex_t itex(in, in->GetLargestPossibleRegion());
  for (itex.GoToBegin(); !itex.IsAtEnd(); ++itex)
  {
    tensor_t t = itex.Get();
    
    double a = atan(t(1, 1) / (t(0, 1) + TINY));
    a += 0.5 * M_PI;
    
    ang->SetPixel(itex.GetIndex(), a);
  }
  
  return ang;
}

//----------------------------------------------------------------
// tolgas_diffusion
// 
image_t::Pointer
tolgas_diffusion(image_t::Pointer & im,
		 float sigma1,
		 float sigma2,
		 float sigma3,
		 float alpha,
		 float T,
		 int hflag,
		 const unsigned int & max_it,
		 const float & dt,
		 const the_text_t & prefix,
		 const the_text_t & suffix)
{
  image_t::SizeType sz = im->GetLargestPossibleRegion().GetSize();
  int w = sz[0];
  int h = sz[1];
  
  int xl, xh, yl, yh;
  
  image_t::Pointer flux_x = make_image<image_t>(w + 1, h, 1.0);
  image_t::Pointer flux_y = make_image<image_t>(w, h + 1, 1.0);
  image_t::Pointer diffcoeff;
  
  tensor_image_t::Pointer TensorImage =
    make_image<tensor_image_t>(sz, tensor(0.0f));
  
  tensor_image_t::Pointer HessianImage =
    make_image<tensor_image_t>(sz, tensor(0.0f));
  
  tensor_image_t::Pointer EigenVectorImage =
    make_image<tensor_image_t>(sz, tensor(0.0f));
  
  vector_image_t::Pointer GradImage;
  vector_image_t::Pointer EigenvalueImage;
  vector_image_t::Pointer TensorEigenvalueImage;
  
  vector_t grad;
  tensor_t D;
  tensor_t U;
  tensor_t W;
  tensor_t U2;
  
  unsigned int k = 1;
  while (k <= max_it)
  {
    // cout << "iteration " << k << endl;
    
    if (hflag != 0)
    {
      HessianImage = ComputeHessianImage(im, sigma1);
    }
    
    if (hflag < 2)
    {
      GradImage = ComputeGradientImage(im, 1, sigma2);
      TensorImage = ComputeStructureTensorImage(GradImage, sigma3);
      
      if (hflag == 1)
      {
        EigenvalueImage = ComputeSortedEigenValues_2x2(HessianImage);
        TensorEigenvalueImage =
	  ComputeSortedEigenVectors_2x2(TensorImage, EigenVectorImage);
      }
      else
      {
        EigenvalueImage =
	  ComputeSortedEigenVectors_2x2(TensorImage, EigenVectorImage);
      }
    }
    else
    {
      EigenvalueImage =
	ComputeSortedEigenVectors_2x2(HessianImage, EigenVectorImage);
    }
    
#if 0
    if (k == 1)
    {
      image_t::Pointer ang = angle_image(EigenVectorImage);
      the_text_t fn = prefix + "-angle." + suffix;
      
      save<native_image_t>(cast<image_t, native_image_t>
			   (remap_min_max<image_t>(ang)),
			   fn);
    }
#endif
    
    diffcoeff = ComputeScalarDiffusionCoefficient(EigenvalueImage, T);
    
    for (int y = 0; y < h; y++)
    {
      for (int x = 1; x < w; x++)
      {
        W = tensor(0.0f);
        xl = std::max(x - 1, 0);
        yl = std::max(y - 1, 0);
        yh = std::min(y + 1, h - 1);
        
        float df = sqrt(diffcoeff->GetPixel(ix(xl, y)) *
			diffcoeff->GetPixel(ix(x, y)));
        W(0, 0) = alpha;
        W(1, 1) = alpha + (1.0 - alpha) * df;
        
        U = EigenVectorImage->GetPixel(ix(x, y));
        U2 = EigenVectorImage->GetPixel(ix(xl, y));
	
	// dot product of main diffusion:
        float dp = U(0, 1) * U2(0, 1) + U(1, 1) * U2(1, 1);
        U(0, 1) = dp * U(0, 1);
        U(1, 1) = dp * U(1, 1);
        D = U * W * (transpose(U));
        
        grad[0] = im->GetPixel(ix(x, y)) - im->GetPixel(ix(xl, y));
        grad[1] = 0.25 * ((im->GetPixel(ix(x, yh)) +
			   im->GetPixel(ix(xl, yh))) -
			  (im->GetPixel(ix(x, yl)) +
			   im->GetPixel(ix(xl, yl))));
        flux_x->SetPixel(ix(x, y), (D * grad)[0]);
      }
      flux_x->SetPixel(ix(0, y), flux_x->GetPixel(ix(1, y)));
      flux_x->SetPixel(ix(w, y), flux_x->GetPixel(ix(w - 1, y)));
    }
    
    for (int x = 0; x < w; x++)
    {
      for (int y = 1; y < h; y++)
      {
	W = tensor(0.0f);
	xl = std::max(x - 1, 0);
	xh = std::min(x + 1, w - 1);
	yl = std::max(y - 1, 0);
	
	float df = sqrt(diffcoeff->GetPixel(ix(x, yl)) *
			diffcoeff->GetPixel(ix(x, y)));
	W(0, 0) = alpha;
	W(1, 1) = alpha + (1.0 - alpha) * df;
	
	U = EigenVectorImage->GetPixel(ix(x, y));
	U2 = EigenVectorImage->GetPixel(ix(x, yl));
	
	// dot product of main diffusion
	float dp = U(0, 1) * U2(0, 1) + U(1, 1) * U2(1, 1);
	U(0, 1) = dp * U(0, 1);
	U(1, 1) = dp * U(1, 1);
	D = U * W * (transpose(U));
	
	grad[1] = im->GetPixel(ix(x, y)) - im->GetPixel(ix(x, yl));
	grad[0]= 0.25 * ((im->GetPixel(ix(xh, y)) +
			  im->GetPixel(ix(xh, yl))) -
			 (im->GetPixel(ix(xl, y)) +
			  im->GetPixel(ix(xl, yl))));
	flux_y->SetPixel(ix(x, y), (D * grad)[1]);
      }
      flux_y->SetPixel(ix(x, 0), flux_y->GetPixel(ix(x, 1)));
      flux_y->SetPixel(ix(x, h), flux_y->GetPixel(ix(x, h - 1)));
    }
    
    // calculate divergence
    for (int x = 0; x < w; x++)
    {
      for (int y = 0; y < h; y++)
      {
	float divx = (flux_x->GetPixel(ix(x + 1, y)) -
		      flux_x->GetPixel(ix(x, y)));
	float divy = (flux_y->GetPixel(ix(x, y + 1)) -
		      flux_y->GetPixel(ix(x, y)));
	im->SetPixel(ix(x, y), im->GetPixel(ix(x, y)) + dt * (divx + divy));
      }
    }
    
    if ((unsigned int)(floor(0.5 + k / SAVE_IT) * SAVE_IT) == k)
    {
      // save results up to the current iteration:
      {
	the_text_t fn =
	  prefix + "-" + the_text_t::number(k, 3, '0') + "." + suffix;
	
	save<native_image_t>(cast<image_t, native_image_t>
			     (remap_min_max<image_t>(im)),
			     fn);
      }
      
      // save diffusion coefficients:
      {
	the_text_t fn =
	  prefix + "-df-" + the_text_t::number(k, 3, '0') + "." + suffix;
	
	save<native_image_t>(cast<image_t, native_image_t>
			     (remap_min_max<image_t>(diffcoeff)),
			     fn);
      }
    }
    k++;
  }
  
  return im;
}


//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL)
{
  cerr << "USAGE: " << argv[0] << " " << endl
       << "\t[-it max_iterations, default 500] " << endl
       << "\t[-sigma size_of_structure_of_interest, default 5.0] " << endl
       << "\t[-alpha isotropic_diffusion [0,1], default 0.01] " << endl
       << "\t[-T membrane_strength, default 0.25] " << endl
       << "\t-load image_in " << endl
       << "\t-save image_out " << endl
       << endl
       << "EXAMPLE: " << argv[0] << " "
       << "-it 500 "
       << "-sigma 5.0 "
       << "-alpha 0.01 "
       << "-T 0.25 "
       << "-load raw/image.tif "
       << "-save enhanced/image.tif "
       << endl;
  
  if (message != NULL)
  {
    cerr << "ERROR: " << message << endl;
  }
  
  exit(1);
}

//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  // this is so that the printouts look better:
  cout.precision(6);
  cout.setf(ios::scientific);
  
  if (argc < 5)
  {
    usage(argv, "not enough parameters specified");
    return 0;
  }
  
  // parse the command line arguments:
  unsigned int max_it = 500;
  float dt = 0.25f;
  float blur = 0.0f;
  float sigma[] = { 5.0f, 1.0f, 1.0f };
  float alpha = 0.01f;
  float threshold = 0.25f;
  int H_flag = 2;
  the_text_t fn_load;
  the_text_t fn_save;
  
  for (int i = 1; i < argc; i++)
  {
    bool ok = true;
    
    if (strcmp(argv[i], "-it") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -it option");
      
      i++;
      max_it = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -it option");
    }
    else if (strcmp(argv[i], "-dt") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -dt option");
      
      i++;
      dt = the_text_t(argv[i]).toFloat(&ok);
      if (!ok || dt <= 0.0f) usage(argv, "bad -dt option");
    }
    else if (strcmp(argv[i], "-blur") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -blur option");
      
      i++;
      blur = the_text_t(argv[i]).toFloat(&ok);
      if (!ok || blur < 0.0f) usage(argv, "bad -blur option");
    }
    else if (strcmp(argv[i], "-sigma") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -sigma option");
      
      i++;
      sigma[0] = the_text_t(argv[i]).toFloat(&ok);
      if (!ok) usage(argv, "bad -sigma option");
    }
    else if (strcmp(argv[i], "-alpha") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -alpha option");
      
      i++;
      alpha = the_text_t(argv[i]).toFloat(&ok);
      if (!ok || alpha < 0.0f || alpha > 1.0f)
      {
	usage(argv, "bad -alpha option");
      }
    }
    else if (strcmp(argv[i], "-T") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -T option");
      
      i++;
      threshold = the_text_t(argv[i]).toFloat(&ok);
      if (!ok)
      {
	usage(argv, "bad -T option");
      }
    }
    else if (strcmp(argv[i], "-load") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -load option");
      
      i++;
      fn_load = the_text_t(argv[i]);
    }
    else if (strcmp(argv[i], "-save") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -save option");
      
      i++;
      fn_save = the_text_t(argv[i]);
    }
    else
    {
      usage(argv,
	    the_text_t("unknown option \'") +
	    the_text_t(argv[i]) +
	    the_text_t("\'"));
    }
  }
  
  // a minimum of two images are required for the registration:
  if (fn_load.is_empty()) usage(argv, "must specify -load option");
  if (fn_save.is_empty()) usage(argv, "must specify -save option");
  
  // for debugging:
  the_text_t prefix = fn_save.reverse().cut('.', 1, ~0).reverse();
  the_text_t suffix = fn_save.reverse().cut('.', 0).reverse();
  
  cout << "iterations: " << max_it << endl
       << "sigma:      " << sigma[0] << endl
       << "alpha:      " << alpha << endl
       << "T:          " << threshold << endl
       << "load:       " << fn_load << endl
       << "save:       " << fn_save << endl
       << "prefix:     " << prefix << endl
       << "suffix:     " << suffix << endl
       << endl;
  
  image_t::Pointer image = load<image_t>(fn_load);
  
  if (blur != 0.0)
  {
    image = gaussDiffuse<image_t>(image, blur);
  }
  
  image_t::Pointer out = tolgas_diffusion(image,
					  sigma[0],
					  sigma[1],
					  sigma[2],
					  alpha,
					  threshold,
					  H_flag,
					  max_it,
					  dt,
					  prefix,
					  suffix);
  
  save<native_image_t>(cast<image_t, native_image_t>
		       (remap_min_max<image_t>(out)),
		       fn_save);
  
  return 0;
}
