// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : ir-fft.cxx
// Author       : Pavel A. Koshevoy
// Created      : 2005/05/24 14:53
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Mosaic layout using phase correlation
//                with optional iterative refinement.

// the includes:
#include <itk/common.hxx>
#include <itk/fft_common.hxx>
#include <itk/mosaic_layout_common.hxx>
#include <itk/itkRegularStepGradientDescentOptimizer2.h>
#include <itk/itkImageMosaicVarianceMetric.h>
#include <itk/itkLegendrePolynomialTransform.h>
#include <utils/the_utils.hxx>
#include <thread/the_boost_mutex.hxx>
#include <thread/the_boost_thread.hxx>

// ITK includes:
#include <itkCommand.h>
#include <itkImageMaskSpatialObject.h>
#include <itkImageRegistrationMethod.h>
#include <itkMeanSquaresImageToImageMetric.h>


//----------------------------------------------------------------
// DEBUG_REFINEMENT
// 
#define DEBUG_REFINEMENT

//----------------------------------------------------------------
// SAVE_INTERMEDIATE_MOSAIC
// 
#define SAVE_INTERMEDIATE_MOSAIC

//----------------------------------------------------------------
// interpolator_t
// 
typedef itk::LinearInterpolateImageFunction<image_t, double> interpolator_t;

//----------------------------------------------------------------
// optimizer_t
// 
typedef itk::RegularStepGradientDescentOptimizer2 optimizer_t;



//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL)
{
  cerr << "USAGE: " << argv[0] << " "
       << "[-sh shrink_factor] "
       << "[-sp pixel_spacing] "
       << "[-py multiresolution_pyramid_levels] "
       << "[-std_mask] "
       << "[-refine] "
       << "[-ol overlap_min overlap_max] "
       << "[-clahe slope] "
       << "[-debug prefix] "
       << "-save transform_params_file_out "
       << "-data image_0 image_1 [image_n]*"
       << endl
       << "EXAMPLE: " << argv[0] << " "
       << "-sh 8 "
       << "-sp 1 "
       << "-py 3 "
       << "-ol 0.05 0.5 "
       << "-clahe 6 "
       << "-debug out/LR- .png "
       << "-save results.txt "
       << "-data L.png R.png"
       << endl;
  
  if (message != NULL)
  {
    cerr << "ERROR: " << message << endl;
  }
  
  exit(1);
}

//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  // this is so that the printouts look better:
  cout.precision(6);
  cout.setf(ios::scientific);
  
  // setup thread storage for the main thread:
  set_the_thread_storage_provider(the_boost_thread_t::thread_storage);
  the_boost_thread_t MAIN_THREAD_DUMMY;
  MAIN_THREAD_DUMMY.set_stopped(false);
  
  // setup thread and mutex interface creators:
  the_mutex_interface_t::set_creator(the_boost_mutex_t::create);
  the_thread_interface_t::set_creator(the_boost_thread_t::create);
  
  // parse the command line arguments:
  unsigned int highest_resolution_level_shrink_factor = 1;
  double pixel_spacing = 1.0;
  unsigned int pyramid_levels = 1;
  unsigned int iterations_per_level = 5;
  std::list<the_text_t> in;
  unsigned int consume_default_tokens = 0;
  double overlap_min = 0.05;
  double overlap_max = 1.0;
  double clahe_slope = 1.0;
  the_text_t fn_results;
  bool use_std_mask = false;
  bool try_refining = false;
  
  // by default run as many threads as there are cores:
  unsigned int num_threads = boost::thread::hardware_concurrency();
  fftw_init_threads();
  itk_fft::set_num_fftw_threads(num_threads);
  
  // for debugging:
  the_text_t prefix;
  
  for (int i = 1; i < argc; i++)
  {
    bool ok = true;
    
    if (strcmp(argv[i], "-sh") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -sh option");
      
      i++;
      highest_resolution_level_shrink_factor = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -sh option");
    }
    else if (strcmp(argv[i], "-sp") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -sp option");
      
      i++;
      pixel_spacing = the_text_t(argv[i]).toDouble(&ok);
      if (!ok) usage(argv, "bad -sp option");
    }
    else if (strcmp(argv[i], "-std_mask") == 0)
    {
      use_std_mask = true;
    }
    else if (strcmp(argv[i], "-refine") == 0)
    {
      try_refining = true;
    }
    else if (strcmp(argv[i], "-ol") == 0)
    {
      if ((argc - i) <= 2) usage(argv, "could not parse -ol option");
      
      i++;
      overlap_min = the_text_t(argv[i]).toDouble(&ok);
      if (!ok || overlap_min < 0.0 || overlap_min > 1.0)
      {
        usage(argv, "bad -ol option (overlap_min)");
      }
      
      i++;
      overlap_max = the_text_t(argv[i]).toDouble(&ok);
      if (!ok || overlap_max < overlap_min || overlap_max > 1.0)
      {
        usage(argv, "bad -ol option (overlap_max)");
      }
    }
    else if (strcmp(argv[i], "-clahe") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -clahe option");
      
      i++;
      clahe_slope = the_text_t(argv[i]).toDouble(&ok);
      if (!ok || clahe_slope < 1.0) usage(argv, "bad -clahe option");
    }
    else if (strcmp(argv[i], "-py") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -py option");
      
      i++;
      pyramid_levels = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -py option");
    }
    else if (strcmp(argv[i], "-it") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -it option");
      
      i++;
      iterations_per_level = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -it option");
    }
    else if (strcmp(argv[i], "-debug") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -debug option");
      
      i++;
      prefix = the_text_t(argv[i]);
    }
    else if (strcmp(argv[i], "-save") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -save option");
      
      i++;
      fn_results = the_text_t(argv[i]);
    }
    else if (strcmp(argv[i], "-data") == 0)
    {
      if ((argc - i) <= 2) usage(argv, "could not parse -data option");
      consume_default_tokens = 2;
    }
    else if (consume_default_tokens > 0)
    {
      in.push_back(the_text_t(argv[i]));
      consume_default_tokens++;
    }
    else
    {
      usage(argv,
            the_text_t("unknown option \'") +
            the_text_t(argv[i]) +
            the_text_t("\'"));
    }
    
    if (consume_default_tokens > 0)
    {
      consume_default_tokens--;
    }
  }
  
  // a minimum of two images are required for the registration:
  if (in.size() < 2) usage(argv, "must specify at least 2 input images");
  
  // save the mosaic configuration:
  std::fstream fout;
  fout.open(fn_results, ios::out);
  if (!fout.is_open())
  {
    usage(argv,
          the_text_t("could not open \'") + fn_results +
          the_text_t("\' for writing"));
  }
  
  // read the input images:
  std::vector<image_t::Pointer> image(in.size());
  std::vector<mask_t::ConstPointer> mask(in.size());
  {
    std::list<the_text_t>::iterator iter = in.begin();
    for (unsigned int i = 0; i < in.size(); i++, ++iter)
    {
      const the_text_t & fn_image = *iter;
      
      // read the image:
      cout << setw(3) << i << " ";
      image[i] = std_tile<image_t>
        (fn_image, highest_resolution_level_shrink_factor, pixel_spacing);
      if (use_std_mask)
      {
        mask[i] = std_mask<image_t>(image[i]);
      }
      
      if (clahe_slope > 1.0)
      {
        image[i] = CLAHE<image_t>(image[i],
                                  255,
                                  255,
                                  clahe_slope,
                                  256,
                                  0.0,
                                  255.0,
                                  mask[i]);
      }
    }
  }
  
  // this is just a reminder:
  const bool images_were_resampled =
    highest_resolution_level_shrink_factor != 1;
  
  // establish the correspondences between the images:
  unsigned int num_images = in.size();
  
  // setup the image pyramid:
  unsigned int shrink_factor = 1;
  array2d(image_t::Pointer) tile_pyramid(pyramid_levels);
  array2d(mask_t::ConstPointer) mask_pyramid(pyramid_levels);
  tile_pyramid[pyramid_levels - 1] = image;
  mask_pyramid[pyramid_levels - 1] = mask;
  for (int i = pyramid_levels - 2; i >= 0; i--)
  {
    tile_pyramid[i].resize(num_images);
    mask_pyramid[i].resize(num_images);
    shrink_factor *= 2;
    
    for (unsigned int j = 0; j < num_images; j++)
    {
      image_t::Pointer img = smooth<image_t>(tile_pyramid[i + 1][j], 2.0);
      img = shrink<image_t>(img, 2);
      tile_pyramid[i][j] = img;
      
      if (mask_pyramid[i + 1][j].GetPointer() != NULL)
      {
        mask_t::ConstPointer msk(shrink<mask_t>(mask_pyramid[i + 1][j], 2));
        mask_pyramid[i][j] = msk;
      }
      
      // FIXME:
#if 0
      save<native_image_t>(cast<image_t, native_image_t>(tile_pyramid[i][j]),
                           the_text_t("pyramid_") +
                           the_text_t::number(i, 2, '0') +
                           the_text_t("_") +
                           the_text_t::number(j, 2, '0') +
                           the_text_t(".png"));
#endif
    }
  }
  
  // dump the status:
  cout << "shrink min: " << highest_resolution_level_shrink_factor << endl
       << "shrink max: " << (highest_resolution_level_shrink_factor *
                             shrink_factor) << endl
       << "pixel spacing: " << pixel_spacing << endl
       << "use std mask: " << use_std_mask << endl
       << "iterations per level: " << iterations_per_level << endl
       << "overlap limits: " << overlap_min << ' ' << overlap_max << endl
       << endl;
  
  // There may be alternative mappings from one image to another via
  // a mapping through any number of intermediate images. Such cascaded
  // mapping can be used to identify and discard false mappings.
  unsigned int max_cascade_len = num_images - 2;
  
  // path[num_intermediate_images][i][j] contains the transforms
  // associated with the specified number of intermediate images
  // that maps image i to image j:
  array3d(translate_transform_t::Pointer) path;
  resize(path, max_cascade_len + 1, num_images, num_images);
  
  // cost[num_intermediate_images][i][j] contains the maximum error metric
  // due to the image mapping between each image in the cascading chain:
  array3d(double) cost;
  resize(cost, max_cascade_len + 1, num_images, num_images);
  
  // initialize the path and cost:
  reset(num_images, max_cascade_len, path, cost);
  
  // setup the initial transform:
  std::vector<affine_transform_t::Pointer> affine(num_images);
  
  // try to find matching pairs of images:
  std::vector<bool> matched(num_images);
  std::list<unsigned int> tile_order;
  layout_mosaic<image_t, mask_t>(*cerr_log(),
                                 num_images,
                                 tile_pyramid,
                                 mask_pyramid,
                                 max_cascade_len,
                                 affine,
                                 matched,
                                 tile_order,
                                 cost,
                                 path,
                                 images_were_resampled,
                                 use_std_mask,
                                 overlap_min,
                                 overlap_max,
                                 16, // max peaks
                                 try_refining,
                                 prefix);
  
  // discard unmatched images:
  unsigned int num_unmatched = 0;
  for (unsigned int i = 0; i < num_images; i++)
  {
    if (matched[i]) continue;
    
    cerr << "UNMATCHED: " << *iterator_at_index(in, i)
         << ", ignoring..."
         << endl;
    image[i] = NULL;
    num_unmatched++;
  }
  cerr << endl;
  cerr << endl;
  
  if (num_unmatched == num_images)
  {
    cerr << "could not find any matching images, aborting..." << endl;
    return 0;
  }
  
  // remap the images, forget about the mismatched images:
  affine = remap(tile_order, affine);
  in = remap(tile_order, in);
  image = remap(tile_order, image);
  mask = remap(tile_order, mask);
  num_images = in.size();
  
  save_mosaic<affine_transform_t>
    (fout, pixel_spacing, use_std_mask, in, affine);
  fout.close();
  
  // FIXME: this is just for debugging:
  // verify that the mosaic will be built up in the correct order:
  if (prefix.size() != 0)
  {
    std::vector<bool> omit(num_images);
    omit.assign(num_images, true);
    
    for (unsigned int i = 0; i < num_images; i++)
    {
      cout << i << ". assembling mosaic..." << endl;
      
      for (unsigned int j = 0; j < i; j++) omit[j] = false;
      
      image_t::Pointer prev =
        make_mosaic<image_t::Pointer, affine_transform_t::Pointer>
        (image[0]->GetSpacing(),
         num_images,
         omit,
         affine,
         image,
         mask,
         FEATHER_NONE_E,
         0);
      
      omit.assign(num_images, true);
      for (unsigned int j = 0; j < i; j++) omit[j] = true;
      omit[i] = false;
      
      image_t::Pointer next =
        make_mosaic<image_t::Pointer, affine_transform_t::Pointer>
        (image[0]->GetSpacing(),
         num_images,
         omit,
         affine,
         image,
         mask,
         FEATHER_NONE_E,
         0);
      
      save_composite<image_t>(prefix +
                              the_text_t("fft-") +
                              the_text_t::number(i, 3, '0') +
                              ".png",
                              prev,
                              next,
                              identity_transform_t::New());
    }
  }
  
  // done:
  return 0;
}
