// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : ir-flip.cxx
// Author       : Pavel A. Koshevoy
// Created      : 2006/06/07 12:26
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Flip an image along either or both axis.

// the includes:
#include <itk/common.hxx>

// system includes:
#include <algorithm>
#include <math.h>


//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL)
{
  cerr << "USAGE: " << argv[0] << " "
       << "[-flipx] "
       << "[-flipy] "
       << "[-remap | -repam_minmax] "
       << "-load image_in "
       << "-save image_out "
       << endl
       << "EXAMPLE: " << argv[0] << " "
       << "-flipx "
       << "-load in/48755.tif "
       << "-save flipped/48755.tif "
       << endl;
  
  if (message != NULL)
  {
    cerr << "ERROR: " << message << endl;
  }
  
  exit(1);
}


//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  // this is so that the printouts look better:
  cout.precision(2);
  cout.setf(ios::scientific);
  
  // parse the command line arguments:
  if (argc < 5) usage(argv);
  the_text_t fn_load;
  the_text_t fn_save;
  bool flipx = false;
  bool flipy = false;
  bool remap = false;
  bool remap_minmax = false;
  
  for (int i = 1; i < argc; i++)
  {
    if (strcmp(argv[i], "-flipx") == 0)
    {
      flipx = true;
    }
    else if (strcmp(argv[i], "-flipy") == 0)
    {
      flipy = true;
    }
    else if (strcmp(argv[i], "-remap") == 0)
    {
      remap = true;
    }
    else if (strcmp(argv[i], "-remap_minmax") == 0)
    {
      remap_minmax = true;
    }
    else if (strcmp(argv[i], "-load") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -load option");
      
      i++;
      fn_load = the_text_t(argv[i]);
    }
    else if (strcmp(argv[i], "-save") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -save option");
      
      i++;
      fn_save = the_text_t(argv[i]);
    }
    else
    {
      usage(argv,
	    the_text_t("unknown option \'") +
	    the_text_t(argv[i]) +
	    the_text_t("\'"));
    }
  }
  
  if (remap && remap_minmax)
  {
    usage(argv, "only one of -remap or -remap_minmax may be specified");
  }
  
  if (fn_load.size() == 0)
  {
    usage(argv, "must specify a file to open with the -load option");
  }
  
  if (fn_save.size() == 0)
  {
    usage(argv, "must specify a file to open with the -save option");
  }
  
  bool recast = !(fn_save.match_tail(".mha") ||
		  fn_save.match_tail(".mhd") ||
		  fn_save.match_tail(".nrrd"));
  
  cout << "loading " << fn_load << endl;
  typedef itk::ImageFileReader<image_t> reader_t;
  reader_t::Pointer reader = reader_t::New();
  reader->SetFileName(fn_load);
  reader->Update();

  image_t::Pointer image = reader->GetOutput();
  image_t::Pointer out = flip<image_t>(image, flipx, flipy);
  
  itk::ImageIOBase * io = reader->GetImageIO();
  itk::ImageIOBase::IOComponentType src_type = io->GetComponentType();
  
  pixel_t min = 0;
  pixel_t max = 0;
  switch (src_type)
  {
    case itk::ImageIOBase::UCHAR:
      min = std::numeric_limits<unsigned char>::min();
      max = std::numeric_limits<unsigned char>::max();
      break;
      
    case itk::ImageIOBase::CHAR:
      min = std::numeric_limits<char>::min();
      max = std::numeric_limits<char>::max();
      break;
      
    case itk::ImageIOBase::USHORT:
      min = std::numeric_limits<unsigned short>::min();
      max = std::numeric_limits<unsigned short>::max();
      break;
      
    case itk::ImageIOBase::SHORT:
      min = std::numeric_limits<short>::min();
      max = std::numeric_limits<short>::max();
      break;
      
    case itk::ImageIOBase::UINT:
      min = std::numeric_limits<unsigned int>::min();
      max = std::numeric_limits<unsigned int>::max();
      break;
      
    case itk::ImageIOBase::INT:
      min = std::numeric_limits<int>::min();
      max = std::numeric_limits<int>::max();
      break;
      
    case itk::ImageIOBase::ULONG:
      min = std::numeric_limits<unsigned long>::min();
      max = std::numeric_limits<unsigned long>::max();
      break;
      
    case itk::ImageIOBase::LONG:
      min = std::numeric_limits<long>::min();
      max = std::numeric_limits<long>::max();
      break;
      
    case itk::ImageIOBase::FLOAT:
    case itk::ImageIOBase::DOUBLE:
    case itk::ImageIOBase::UNKNOWNCOMPONENTTYPE:
    default:
      remap_minmax = true;
      break;
  }
  
  if (remap_minmax || (recast && !remap && (min < 0 || max > 255)))
  {
    image_min_max<image_t>(out, min, max);
    remap = true;
  }
  
  if (remap_minmax || remap)
  {
    remap_min_max_inplace<image_t>(out, min, max/*, 0, 255*/);
  }
  
  if (recast)
  {
    save<native_image_t>(cast<image_t, native_image_t>(out), fn_save);
  }
  else
  {
    save<image_t>(out, fn_save);
  }
  
  // done:
  return 0;
}
