// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : ir-layout.cxx
// Author       : Pavel A. Koshevoy
// Created      : 2007/04/26 13:31
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Manual mosaic tile layout -- for each tile
//                specified on the command line, tile coordinates
//                must be entered from the stdin.

// the includes:
#include <itk/common.hxx>


//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL)
{
  cerr << "USAGE: " << argv[0] << " "
       << "[-sp pixel_spacing] "
       << "[-std_mask] "
       << "-save mosaic_out "
       << "-data tile_0 [tile_N]+"
       << endl
       << "EXAMPLE: " << argv[0] << " "
       << "-std_mask "
       << "-save manual.mosaic "
       << "-data L.png R.png"
       << endl;
  
  if (message != NULL)
  {
    cerr << "ERROR: " << message << endl;
  }
  
  exit(1);
}

//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  // this is so that the printouts look better:
  cout.precision(6);
  cout.setf(ios::scientific);
  
  // parse the command line arguments:
  double pixel_spacing = 1.0;
  std::list<the_text_t> in;
  unsigned int consume_default_tokens = 0;
  the_text_t fn_results;
  bool use_std_mask = false;
  
  for (int i = 1; i < argc; i++)
  {
    bool ok = true;
    
    if (strcmp(argv[i], "-sp") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -sp option");
      
      i++;
      pixel_spacing = the_text_t(argv[i]).toDouble(&ok);
      if (!ok) usage(argv, "bad -sp option");
    }
    else if (strcmp(argv[i], "-std_mask") == 0)
    {
      use_std_mask = true;
    }
    else if (strcmp(argv[i], "-save") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -save option");
      
      i++;
      fn_results = the_text_t(argv[i]);
    }
    else if (strcmp(argv[i], "-data") == 0)
    {
      if ((argc - i) <= 2) usage(argv, "could not parse -data option");
      consume_default_tokens = 2;
    }
    else if (consume_default_tokens > 0)
    {
      in.push_back(the_text_t(argv[i]));
      consume_default_tokens++;
    }
    else
    {
      usage(argv,
	    the_text_t("unknown option \'") +
	    the_text_t(argv[i]) +
	    the_text_t("\'"));
    }
    
    if (consume_default_tokens > 0)
    {
      consume_default_tokens--;
    }
  }
  
  // a minimum of two images are required for the registration:
  if (in.size() < 2) usage(argv, "must specify at least 2 input images");
  
  // save the mosaic configuration:
  std::fstream fout;
  fout.open(fn_results, ios::out);
  if (!fout.is_open())
  {
    usage(argv,
	  the_text_t("could not open \'") + fn_results +
	  the_text_t("\' for writing"));
  }
  
  // establish the correspondences between the images:
  unsigned int num_images = in.size();
  
  // setup the initial transform:
  std::vector<translate_transform_t::Pointer> transform(num_images);
  
  unsigned int index = 0;
  for (std::list<the_text_t>::const_iterator iter = in.begin();
       iter != in.end(); ++iter, ++index)
  {
    const the_text_t & tile = *iter;
    vec2d_t shift = vec2d(0, 0);
    
    cout << "enter x coordinates for " << tile << endl;
    std::cin >> shift[0];
    cout << "enter y coordinates for " << tile << endl;
    std::cin >> shift[1];
    
    transform[index] = translate_transform_t::New();
    transform[index]->SetOffset(neg(shift));
  }
  
  save_mosaic<translate_transform_t>
    (fout, pixel_spacing, use_std_mask, in, transform);
  fout.close();
  
  // done:
  return 0;
}
