// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : ir-median.cxx
// Author       : Pavel A. Koshevoy
// Created      : 2006/11/20 22:48
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : The median filter -- effective against salt and pepper noise.

// the includes:
#include <itk/common.hxx>
#include <itk/itkStatisticsImageFilterWithMask.h>
#include <utils/the_utils.hxx>


//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL)
{
  cerr << "USAGE: " << argv[0] << " "
       << "[-sh shrink_factor] "
       << "[-r radius] "
       << "-load image_in "
       << "-save image_out"
       << endl
       << "EXAMPLE: " << argv[0] << " "
       << "-sh 8 "
       << "-r 2 "
       << "-load in/48755.tif "
       << "-save he/48755.tif "
       << endl;
  
  if (message != NULL)
  {
    cerr << "ERROR: " << message << endl;
  }
  
  exit(1);
}


//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  // this is so that the printouts look better:
  cout.precision(2);
  cout.setf(ios::scientific);
  
  // parse the command line arguments:
  if (argc < 5) usage(argv);
  the_text_t fn_load;
  the_text_t fn_save;
  
  unsigned int shrink_factor = 1;
  unsigned int radius = 0;
  
  for (int i = 1; i < argc; i++)
  {
    bool ok = true;
    
    if (strcmp(argv[i], "-sh") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -sh option");
      
      i++;
      shrink_factor = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -sh option");
    }
    else if (strcmp(argv[i], "-r") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -r option");
      
      i++;
      radius = the_text_t(argv[i]).toUInt(&ok);
      if (!ok || radius < 1) usage(argv, "bad -r radius option");
    }
    else if (strcmp(argv[i], "-load") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -load option");
      
      i++;
      fn_load = the_text_t(argv[i]);
    }
    else if (strcmp(argv[i], "-save") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -save option");
      
      i++;
      fn_save = the_text_t(argv[i]);
    }
    else
    {
      usage(argv,
	    the_text_t("unknown option \'") +
	    the_text_t(argv[i]) +
	    the_text_t("\'"));
    }
  }
  
  if (fn_load.size() == 0)
  {
    usage(argv, "must specify the -load option");
  }
  
  if (fn_save.size() == 0)
  {
    usage(argv, "must specify the -save option");
  }
  
  if (radius == 0)
  {
    usage(argv, "must specify the -r radius option");
  }
  
  // read the input image:
  image_t::Pointer image = std_tile<image_t>(fn_load, shrink_factor, 1);
  image_t::Pointer out = median<image_t>(image, radius);
  bool recast_values =
    fn_save.match_tail(".tif") || fn_save.match_tail(".png");
  
  if (recast_values)
  {
    save<native_image_t>(cast<image_t, native_image_t>(out), fn_save);
  }
  else
  {
    save<image_t>(out, fn_save);
  }
  
  // done:
  return 0;
}
