// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : ir-normalize.cxx
// Author       : Pavel A. Koshevoy
// Created      : 2006/06/16 17:00
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Image normalization filter based on blending overlapping
//                neighborhoods, where each neighborhood is normalized
//                such that pixels conform to a normal distribution,
//                the pixel intensities are clamped to the [-3, 3] range
//                and the belnded pixels are remapped into a user specified
//                [min, max] range. This filter produces results
//                similar to CLAHE.

// the includes:
#include <itk/common.hxx>
#include <itk/itkStatisticsImageFilterWithMask.h>
#include <utils/the_utils.hxx>


//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL)
{
  cerr << "USAGE: " << argv[0] << " "
       << "[-sh shrink_factor] "
       << "[-sz tile_width | -nt tiles_per_row] "
       << "[-std_mask] "
       << "[-mask mask_in] "
       << "-load image_in "
       << "-save image_out"
       << endl
       << "EXAMPLE: " << argv[0] << " "
       << "-sh 8 "
       << "-nt 10 "
       << "-std_mask "
       << "-load in/48755.tif "
       << "-save he/48755.tif "
       << endl;
  
  if (message != NULL)
  {
    cerr << "ERROR: " << message << endl;
  }
  
  exit(1);
}


//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  // this is so that the printouts look better:
  cout.precision(2);
  cout.setf(ios::scientific);
  
  // parse the command line arguments:
  if (argc < 5) usage(argv);
  the_text_t fn_mask;
  the_text_t fn_load;
  the_text_t fn_save;
  
  unsigned int shrink_factor = 1;
  unsigned int cols = 1;
  unsigned int rows = 1;
  unsigned int size = 0;
  bool use_std_mask = false;
  mask_t::Pointer mask;
  
  for (int i = 1; i < argc; i++)
  {
    bool ok = true;
    
    if (strcmp(argv[i], "-sh") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -sh option");
      
      i++;
      shrink_factor = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -sh option");
    }
    else if (strcmp(argv[i], "-std_mask") == 0)
    {
      use_std_mask = true;
    }
    else if (strcmp(argv[i], "-sz") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -sz option");
      
      i++;
      size = the_text_t(argv[i]).toUInt(&ok);
      if (!ok || size < 2) usage(argv, "bad -sz option");
    }
    else if (strcmp(argv[i], "-nt") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -nt option");
      
      i++;
      cols = the_text_t(argv[i]).toUInt(&ok);
      if (!ok || cols == 0) usage(argv, "bad -nt option");
    }
    else if (strcmp(argv[i], "-mask") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -mask option");
      
      i++;
      fn_mask = the_text_t(argv[i]);
    }
    else if (strcmp(argv[i], "-load") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -load option");
      
      i++;
      fn_load = the_text_t(argv[i]);
    }
    else if (strcmp(argv[i], "-save") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -save option");
      
      i++;
      fn_save = the_text_t(argv[i]);
    }
    else
    {
      usage(argv,
	    the_text_t("unknown option \'") +
	    the_text_t(argv[i]) +
	    the_text_t("\'"));
    }
  }
  
  if (fn_load.size() == 0)
  {
    usage(argv, "must specify a file to open with the -load option");
  }
  
  if (fn_save.size() == 0)
  {
    usage(argv, "must specify a file to open with the -save option");
  }
  
  // read the input image:
  image_t::Pointer image = std_tile<image_t>(fn_load, shrink_factor, 1);
  if (use_std_mask)
  {
    mask = std_mask<image_t>(image);
  }
  else if (fn_mask.size() > 0)
  {
    mask = std_tile<mask_t>(fn_mask, shrink_factor, 1);
  }
  
  // split the image into a set of overlapping tiles:
  image_t::SizeType sz = image->GetLargestPossibleRegion().GetSize();
  
  if (size != 0)
  {
    cols = std::max(1, int(floor(double(2 * sz[0]) / double(size) + 0.5) - 1));
  }
  
  if (size != 0 || cols > 1)
  {
    rows =
      (unsigned int)
      (floor(double(cols) * double(sz[1]) / double(sz[0]) + 0.5));
  }
  
  cout << "rows: " << rows << endl
       << "cols: " << cols << endl;
  /*
  typedef itk::StatisticsImageFilterWithMask<image_t> statistics_t;
  statistics_t::Pointer statistics = statistics_t::New();
  statistics->SetInput(image);
  statistics->SetImageMask(mask_so(mask));
  statistics->Update();
  cout << "min:      " << statistics->GetMinimum() << endl
       << "max:      " << statistics->GetMaximum() << endl
       << "mean:     " << statistics->GetMean() << endl
       << "sigma:    " << statistics->GetSigma() << endl
       << "variance: " << statistics->GetVariance() << endl
       << "sum:      " << statistics->GetSum() << endl << endl;
  */
  
  image_t::Pointer out;
  if (rows == 1 && cols == 1)
  {
    out = normalize<image_t>(image, mask);
    out = threshold<image_t>(out, -3, 3, -3, 3);
  }
  else
  {
    out = normalize<image_t>(image, cols, rows, 0, 255, mask);
  }
  
  /*
  statistics->SetInput(out);
  statistics->Update();
  cout << "min:      " << statistics->GetMinimum() << endl
       << "max:      " << statistics->GetMaximum() << endl
       << "mean:     " << statistics->GetMean() << endl
       << "sigma:    " << statistics->GetSigma() << endl
       << "variance: " << statistics->GetVariance() << endl
       << "sum:      " << statistics->GetSum() << endl << endl;
  */
  remap_min_max_inplace<image_t>(out);
  
  bool recast_values =
    fn_save.match_tail(".tif") || fn_save.match_tail(".png");
  if (recast_values)
  {
    save<native_image_t>(cast<image_t, native_image_t>(out), fn_save);
  }
  else
  {
    save<image_t>(out, fn_save);
  }
  
  // done:
  return 0;
}
