// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : ir-pyramid.cxx
// Author       : Pavel A. Koshevoy
// Created      : 2006/06/13 17:13
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Pre-compute and save a Difference of Gaussians pyramid
//                so it can be quickly loaded into ir-stos. This is
//                useful when debugging feature based (SIFT key) image
//                registration.

// the includes:
#include <itk/common.hxx>
#include <itk/stos_common.hxx>
#include <utils/the_text.hxx>
#include <utils/the_utils.hxx>
#include <thread/the_boost_mutex.hxx>
#include <thread/the_boost_thread.hxx>

// local includes:
#include "pyramid.hxx"
#include "match.hxx"

// system includes:
#include <math.h>
#include <stdlib.h>
#include <algorithm>
#include <functional>


//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL)
{
  cerr << "USAGE: " << argv[0] << " "
       << "[-sh shrink_factor] "
       << "[-clahe slope] "
       << "[-flip] "
       << "[-s scales_per_octave] "
       << "-load some.mosaic "
       << "-save id some.pyramid "
       << endl
       << "EXAMPLE: " << argv[0] << " "
       << "-sh 8 "
       << "-clahe 2 "
       << "-s 3 "
       << "-load s7.mosaic "
       << "-save index s7.pyramid "
       << endl;
  
  if (message != NULL)
  {
    cerr << "ERROR: " << message << endl;
  }
  
  exit(1);
}

//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  // this is so that the printouts look better:
  cout.precision(6);
  cout.setf(ios::scientific);
  
  // setup thread storage for the main thread:
  set_the_thread_storage_provider(the_boost_thread_t::thread_storage);
  the_boost_thread_t MAIN_THREAD_DUMMY;
  MAIN_THREAD_DUMMY.set_stopped(false);
  
  // setup thread and mutex interface creators:
  the_mutex_interface_t::set_creator(the_boost_mutex_t::create);
  the_thread_interface_t::set_creator(the_boost_thread_t::create);
  
  // parse the command line arguments:
  if (argc < 5) usage(argv);
  
  the_text_t fn_load;
  bool flip = false;
  
  the_text_t fn_save;
  double clahe_slope = 1.0;
  unsigned int shrink_factor = 1;
  unsigned int pyramid_index = 0;
  unsigned int scales_per_octave = 1;
  
  // parse the command line arguments:
  for (int i = 1; i < argc; i++)
  {
    bool ok = true;
    
    if (strcmp(argv[i], "-sh") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -sh option");
      
      i++;
      shrink_factor = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -sh option");
    }
    else if (strcmp(argv[i], "-clahe") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -clahe option");
      
      i++;
      clahe_slope = the_text_t(argv[i]).toDouble(&ok);
      if (!ok || clahe_slope < 1.0) usage(argv, "bad -clahe option");
    }
    else if (strcmp(argv[i], "-flip") == 0)
    {
      flip = true;
    }
    else if (strcmp(argv[i], "-s") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -s option");
      
      i++;
      scales_per_octave = the_text_t(argv[i]).toUInt(&ok);
      if (!ok || scales_per_octave < 1) usage(argv, "bad -s option");
    }
    else if (strcmp(argv[i], "-save") == 0)
    {
      if ((argc - i) <= 2) usage(argv, "could not parse -save option");

      i++;
      pyramid_index = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -save option (pyramid id)");
      
      i++;
      fn_save = the_text_t(argv[i]);
    }
    else if (strcmp(argv[i], "-load") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -load option");
      
      i++;
      fn_load.assign(argv[i]);
    }
    else
    {
      usage(argv,
	    the_text_t("unknown option \'") +
	    the_text_t(argv[i]) +
	    the_text_t("\'"));
    }
  }
  
  // dump the status:
  cout << "shrink factor: " << shrink_factor << endl
       << endl;
  
  // load the images, assemble the pyramids, match them:
  image_t::Pointer mosaic;
  mask_t::Pointer mosaic_mask;
  
  if (fn_load.match_tail(".mosaic"))
  {
    load_slice(fn_load,
	       flip,
	       shrink_factor,
	       clahe_slope,
	       mosaic,
	       mosaic_mask);
  }
  else
  {
    // load an image:
    mosaic = std_tile<image_t>(fn_load, shrink_factor, 1);
    mosaic_mask = std_mask<image_t>(mosaic);
  }
  
  pyramid_t pyramid;
  setup_pyramid(pyramid,
		pyramid_index,
		fn_load,
		mosaic,
		mosaic_mask,
		~0,
		scales_per_octave,
		true);
  
  pyramid.save(fn_save);
  
  // done:
  return 0;
}
