// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : ir-refine-grid.cxx
// Author       : Pavel A. Koshevoy, Joel Spaltenstein
// Created      : 2007/04/09 08:53
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Refine a mosaic layout via concurrent mesh warping
//                of the mosaic tiles.

// the includes:
#include <itk/common.hxx>
#include <itk/mosaic_refinement_common.hxx>
#include <itk/itkImageMosaicVarianceMetric.h>
#include <itk/itkRadialDistortionTransform.h>
#include <itk/itkLegendrePolynomialTransform.h>
#include <itk/itkRegularStepGradientDescentOptimizer2.h>
#include <utils/the_dynamic_array.hxx>
#include <thread/the_boost_mutex.hxx>
#include <thread/the_boost_thread.hxx>

// system includes:
#include <math.h>

// ITK includes:
#include <itkDiscreteGaussianImageFilter.h>
#include <itkShrinkImageFilter.h>

//----------------------------------------------------------------
// SAVE_INTERMEDIATE_MOSAIC
// 
#define SAVE_INTERMEDIATE_MOSAIC


//----------------------------------------------------------------
// optimizer_t
//
typedef itk::RegularStepGradientDescentOptimizer2 optimizer_t;


//----------------------------------------------------------------
// optimization_observer_t
// 
template<class mosaic_metric_t>
class optimization_observer_t : public itk::Command
{
public:
  typedef optimization_observer_t Self;
  typedef itk::Command Superclass;
  typedef itk::SmartPointer<Self> Pointer;
  
  itkNewMacro(Self);
  
  void Execute(itk::Object *caller, const itk::EventObject & event)
  { Execute((const itk::Object *)(caller), event); }
  
  void Execute(const itk::Object * object, const itk::EventObject & event)
  {
    if (typeid(event) != typeid(itk::IterationEvent)) return;
    
    const optimizer_t * optimizer = dynamic_cast<const optimizer_t *>(object);
    cout << (unsigned int)(optimizer->GetCurrentIteration()) << '\t'
	 << optimizer->GetValue() << '\t'
	 << metric_->GetNumberOfPixelsCounted() << '\t'
	 << optimizer->GetCurrentPosition() << endl;
    
    // FIXME: 2005/11/09
#ifdef SAVE_INTERMEDIATE_MOSAIC
    {
      static unsigned int counter = 0;
      
      the_text_t fn =
	prefix_ + the_text_t::number(counter, 3, '0') + the_text_t(".mosaic");
      std::fstream fout;
      fout.open(fn, ios::out);
      save_mosaic<base_transform_t>(fout,
				    pixel_spacing_,
				    use_std_mask_,
				    in_,
				    transform_);
      fout.close();
      counter++;
    }
#endif
  }
  
  // FIXME: 2005/11/09
  typename mosaic_metric_t::Pointer metric_;
  std::list<the_text_t> in_;
  double pixel_spacing_;
  bool use_std_mask_;
  std::vector<base_transform_t::Pointer> transform_;
  the_text_t prefix_;
  
protected:
  optimization_observer_t() {}
};


//----------------------------------------------------------------
// radial_distortion_t
//
typedef itk::RadialDistortionTransform<double, 2>
radial_distortion_t;

//----------------------------------------------------------------
// affine_transform_t
// 
typedef itk::LegendrePolynomialTransform<double, 1>
affine_transform_t;

//----------------------------------------------------------------
// cubic_transform_t
// 
typedef itk::LegendrePolynomialTransform<double, 3>
cubic_transform_t;

//----------------------------------------------------------------
// image_interpolator_t
// 
typedef itk::LinearInterpolateImageFunction<image_t, double>
image_interpolator_t;


//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL)
{
  cerr << "USAGE: " << argv[0] << " "
       << "[-sh shrink_factor] "
       << "[-it iterations] "
       << "[-sp pixel_spacing] "
       << "[-cell size] "
       << "[-mesh rows cols] "
       << "[-threads number_of_threads] "
       << "-load any_transform_in "
       << "-save grid_transform_out "
       << endl
       << "EXAMPLE: " << argv[0] << " "
       << "-sh 8 "
       << "-it 3 "
       << "-cell 96 "
       << "-mesh 8 8 "
       << "-load fft/s01.mosaic "
       << "-save grid/s01.mosaic"
       << endl;
  
  if (message != NULL)
  {
    cerr << "ERROR: " << message << endl;
  }
  
  exit(1);
}

//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  // this is so that the printouts look better:
  cout.precision(3);
  cout.setf(ios::scientific);
  
  // setup thread storage for the main thread:
  set_the_thread_storage_provider(the_boost_thread_t::thread_storage);
  the_boost_thread_t MAIN_THREAD_DUMMY;
  MAIN_THREAD_DUMMY.set_stopped(false);
  
  // setup thread and mutex interface creators:
  the_mutex_interface_t::set_creator(the_boost_mutex_t::create);
  the_thread_interface_t::set_creator(the_boost_thread_t::create);
  
  // parse the command line arguments:
  unsigned int shrink_factor = 1;
  bool pixel_spacing_user_override = false;
  double pixel_spacing = 1.0;
  bool use_std_mask = false;
  unsigned int iterations = 3;
  unsigned int cell_size = 0;
  unsigned int mesh_rows = 0;
  unsigned int mesh_cols = 0;
  the_text_t fn_results;
  
  // by default run as many threads as there are cores:
  unsigned int num_threads = boost::thread::hardware_concurrency();
  fftw_init_threads();
  itk_fft::set_num_fftw_threads(1);
  
  std::list<the_text_t> in;
  std::vector<base_transform_t::Pointer> tbase;
  std::vector<image_t::Pointer> image;
  std::vector<mask_t::ConstPointer> mask;
  
  for (int i = 1; i < argc; i++)
  {
    bool ok = true;
    
    if (strcmp(argv[i], "-sh") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -sh option");
      
      i++;
      shrink_factor = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -sh option");
    }
    else if (strcmp(argv[i], "-sp") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -sp option");
      
      i++;
      pixel_spacing = the_text_t(argv[i]).toDouble(&ok);
      pixel_spacing_user_override = true;
      if (!ok) usage(argv, "bad -sp option");
    }
    else if (strcmp(argv[i], "-it") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -it option");
      
      i++;
      iterations = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -it option");
    }
    else if (strcmp(argv[i], "-cell") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -cell option");
      
      i++;
      cell_size = the_text_t(argv[i]).toUInt(&ok);
      if (!ok || cell_size < 4) usage(argv, "bad -cell option");
    }
    else if (strcmp(argv[i], "-mesh") == 0)
    {
      if ((argc - i) <= 2) usage(argv, "could not parse -mesh option");
      
      i++;
      mesh_rows = the_text_t(argv[i]).toUInt(&ok);
      if (!ok || mesh_rows < 2) usage(argv, "bad -mesh rows option");
      
      i++;
      mesh_cols = the_text_t(argv[i]).toUInt(&ok);
      if (!ok || mesh_cols < 2) usage(argv, "bad -mesh cols option");
    }
    else if (strcmp(argv[i], "-threads") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -threads option");
      
      i++;
      num_threads = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -threads option");
    }
    else if (strcmp(argv[i], "-load") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -load option");
      
      i++;
      std::fstream fin;
      fin.open(argv[i], ios::in);
      if (!fin.is_open())
      {
	usage(argv,
	      the_text_t("could not open \'") +
	      the_text_t(argv[i]) +
	      the_text_t("\' for reading"));
      }
      
      double mosaic_pixel_spacing = 0;
      load_mosaic<base_transform_t>(fin,
				    mosaic_pixel_spacing,
				    use_std_mask,
				    in,
				    tbase);
      if (!pixel_spacing_user_override)
      {
	pixel_spacing = mosaic_pixel_spacing;
      }
      
      unsigned int num_images = tbase.size();
      assert(num_images > 1);
      assert(pixel_spacing != 0);
      
      image.resize(num_images);
      mask.resize(num_images);
      
      unsigned int i = 0;
      for (std::list<the_text_t>::const_iterator iter = in.begin();
	   iter != in.end(); ++iter, ++i)
      {
	const the_text_t & fn_image = *iter;
	
	// read the image:
	cout << setw(3) << i << " ";
	image[i] = std_tile<image_t>(fn_image, shrink_factor, pixel_spacing);
	
	if (use_std_mask)
	{
	  mask[i] = std_mask<image_t>(image[i]);
	}
      }
      
      fin.close();
    }
    else if (strcmp(argv[i], "-save") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -save option");
      
      i++;
      fn_results = the_text_t(argv[i]);
    }
    else
    {
      usage(argv,
	    the_text_t("unknown option \'") +
	    the_text_t(argv[i]) +
	    the_text_t("\'"));
    }
  }
  
  // a minimum of two images are required for the registration:
  unsigned int num_images = in.size();
  if (num_images < 2)
  {
    usage(argv, "something is wrong with the -load parameter");
  }
  
  if (cell_size == 0 && (mesh_rows == 0 || mesh_cols == 0))
  {
    usage(argv, "you must specify either the -cell or -mesh parameter");
  }
  else if (cell_size == 0)
  {
    // calculate the cell size from the rows/cols:
    image_t::SizeType sz = image[0]->GetLargestPossibleRegion().GetSize();
    unsigned int nx = 3 * sz[0] / (mesh_cols - 1);
    unsigned int ny = 3 * sz[1] / (mesh_rows - 1);
    cell_size = std::max(nx, ny);
  }
  else if (mesh_rows == 0 || mesh_cols == 0)
  {
    // calculate the cell size from the rows/cols:
    image_t::SizeType sz = image[0]->GetLargestPossibleRegion().GetSize();
    mesh_cols = 1 + 3 * sz[0] / cell_size;
    mesh_rows = 1 + 3 * sz[1] / cell_size;
  }
  
  // get the image bounding boxes:
  std::vector<pnt2d_t> tile_min(num_images);
  std::vector<pnt2d_t> tile_max(num_images);
  calc_image_bboxes<image_t::Pointer>(image, tile_min, tile_max);
  
  // setup the mesh transforms:
  std::vector<itk::GridTransform::Pointer> transform(num_images);
  for (unsigned int i = 0; i < num_images; i++)
  {
    transform[i] = itk::GridTransform::New();
    bool ok = setup_grid_transform(transform[i]->transform_,
				   mesh_rows - 1,	// grid rows
				   mesh_cols - 1,	// grid cols
				   tile_min[i],
				   tile_max[i],
				   mask[i].GetPointer(),
				   tbase[i].GetPointer(),
				   100,			// max iterations
				   1e-12,		// min step scale
				   1e-16,		// min error squared
				   5);			// pick up pace steps
    transform[i]->setup(transform[i]->transform_);
    if (!ok) assert(false);
  }
  
  cout << "shrink factor: " << shrink_factor << endl
       << "pixel spacing: " << pixel_spacing << endl
       << "iterations:    " << iterations << endl
       << "neighborhood:  " << cell_size << endl
       << "mesh rows:     " << mesh_rows << endl
       << "mesh cols:     " << mesh_cols << endl
       << "threads:       " << num_threads << endl
       << endl;
  
  refine_mosaic_mt<image_t::Pointer, mask_t::ConstPointer>
    (*cout_log(),
     transform,
     image,
     mask,
     cell_size,
     true,		// prewarp tiles
     0.25,		// minimum overlap
     1,			// median filter radius
     iterations,	// number of refinement passes
     false,		// all tiles are moving
     num_threads);	// number of threads to launch
  
  // save the results:
  std::fstream fout;
  fout.open(fn_results, ios::out);
  
  if (!fout.is_open())
  {
    usage(argv,
	  the_text_t("could not open \'") + fn_results +
	  the_text_t("\' for writing"));
  }
  
  save_mosaic<itk::GridTransform>(fout,
				  pixel_spacing,
				  use_std_mask,
				  in,
				  transform);
  fout.close();
  
  // done:
  return 0;
}
