// -*- Mode: c++; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*  
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// File         : ir-refine-translate.cxx
// Author       : Joel Spaltenstein
// Created      : 2008/02/19
// Copyright    : (C) 2008 University of Utah
// License      : GPLv2
// Description  :

// the includes:
#include <itk/common.hxx>
#include <itk/mosaic_refinement_common.hxx>
#include <itk/itkImageMosaicVarianceMetric.h>
#include <itk/itkRadialDistortionTransform.h>
#include <itk/itkLegendrePolynomialTransform.h>
#include <itk/itkRegularStepGradientDescentOptimizer2.h>
#include <utils/the_dynamic_array.hxx>
#include <thread/the_boost_mutex.hxx>
#include <thread/the_boost_thread.hxx>

// system includes:
#include <math.h>

// ITK includes:
#include <itkDiscreteGaussianImageFilter.h>
#include <itkShrinkImageFilter.h>

// local includes:
#include "IRRefineTranslateCanvas.hxx"
#include "IRImageLoader.hxx"


//----------------------------------------------------------------
// affine_transform_t
// 
typedef itk::LegendrePolynomialTransform<double, 1>
affine_transform_t;

//----------------------------------------------------------------
// cubic_transform_t
// 
typedef itk::LegendrePolynomialTransform<double, 3>
cubic_transform_t;

//----------------------------------------------------------------
// image_interpolator_t
// 
typedef itk::LinearInterpolateImageFunction<image_t, double>
image_interpolator_t;


//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL)
{
	cerr << "USAGE: " << argv[0] << " "
		 << "[-sh shrink_factor] "
		 << "[-sp pixel_spacing] "
		 << "[-it iterations] "
		 << "[-threads number_of_threads] "
		 << "-load any_transform_in "
		 << "-save grid_transform_out "
		 << endl
		 << "EXAMPLE: " << argv[0] << " "
		 << "-sh 8 "
		 << "-load fft/s01.mosaic "
		 << "-save grid/s01.mosaic"
		 << endl;
	
	if (message != NULL)
	{
		cerr << "ERROR: " << message << endl;
	}
	
	exit(1);
}

//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
	// this is so that the printouts look better:
	cout.precision(3);
	cout.setf(ios::scientific);
	
	// setup thread storage for the main thread:
	set_the_thread_storage_provider(the_boost_thread_t::thread_storage);
	the_boost_thread_t MAIN_THREAD_DUMMY;
	MAIN_THREAD_DUMMY.set_stopped(false);
	
	// setup thread and mutex interface creators:
	the_mutex_interface_t::set_creator(the_boost_mutex_t::create);
	the_thread_interface_t::set_creator(the_boost_thread_t::create);
	
	// parse the command line arguments:
	unsigned int shrink_factor = 1;
	bool pixel_spacing_user_override = false;
	double pixel_spacing = 1.0;
	bool use_std_mask = false;
	unsigned int num_threads = boost::thread::hardware_concurrency();
	the_text_t fn_results;
	std::vector<the_text_t> imagePathVector;
	
	std::list<the_text_t> in;
	std::vector<base_transform_t::Pointer> tbase;
	
	IRRefineTranslateCanvas canvas;
	
	for (int i = 1; i < argc; i++)
	{
		bool ok = true;
		
		if (strcmp(argv[i], "-sh") == 0)
		{
			if ((argc - i) <= 1) usage(argv, "could not parse -sh option");
			
			i++;
			shrink_factor = the_text_t(argv[i]).toUInt(&ok);
			if (!ok) usage(argv, "bad -sh option");
		}
		else if (strcmp(argv[i], "-sp") == 0)
		{
			if ((argc - i) <= 1) usage(argv, "could not parse -sp option");
			
			i++;
			pixel_spacing = the_text_t(argv[i]).toDouble(&ok);
			pixel_spacing_user_override = true;
			if (!ok) usage(argv, "bad -sp option");
		}
		else if (strcmp(argv[i], "-threads") == 0)
		{
			if ((argc - i) <= 1) usage(argv, "could not parse -threads option");
			
			i++;
			num_threads = the_text_t(argv[i]).toUInt(&ok);
			if (!ok) usage(argv, "bad -threads option");
		}
		else if (strcmp(argv[i], "-load") == 0)
		{
			if ((argc - i) <= 1) usage(argv, "could not parse -load option");
			
			i++;
			std::fstream fin;
			fin.open(argv[i], ios::in);
			if (!fin.is_open())
			{
				usage(argv,
					  the_text_t("could not open \'") +
					  the_text_t(argv[i]) +
					  the_text_t("\' for reading"));
			}
			
			double mosaic_pixel_spacing = 0;
			load_mosaic<base_transform_t>(fin,
										  mosaic_pixel_spacing,
										  use_std_mask,
										  in,
										  tbase);
			if (!pixel_spacing_user_override)
			{
				pixel_spacing = mosaic_pixel_spacing;
			}
			
			unsigned int num_images = tbase.size();
			if (num_images < 2) usage(argv, "less than 2 images in the mosaic");
			if (!pixel_spacing) usage(argv, "pixel spacing is zero");
			
			for (std::list<the_text_t>::const_iterator iter = in.begin();
				 iter != in.end(); ++iter)
			{
				imagePathVector.push_back(*iter);
			}
			
			fin.close();
		}
		else if (strcmp(argv[i], "-save") == 0)
		{
			if ((argc - i) <= 1) usage(argv, "could not parse -save option");
			
			i++;
			fn_results = the_text_t(argv[i]);
		}
		else
		{
			usage(argv,
				  the_text_t("unknown option \'") +
				  the_text_t(argv[i]) +
				  the_text_t("\'"));
		}
	}
	
	// a minimum of two images are required for the registration:
	unsigned int num_images = in.size();
	if (num_images < 2)
	{
		usage(argv, "something is wrong with the -load parameter");
	}
		
	cout << "shrink factor: " << shrink_factor << endl
		 << "pixel spacing: " << pixel_spacing << endl
		 << "threads:       " << num_threads << endl
		 << endl;
	
	IRImageLoader::sharedImageLoader()->setShrinkFactor(shrink_factor);
	IRImageLoader::sharedImageLoader()->setPixelSpacing(pixel_spacing);
	
	canvas.setBaseTransforms(tbase, imagePathVector);
	canvas.buildConnections();
	canvas.findIdealTransformationOffsets(num_threads);
	canvas.fillGroupIDs();
	canvas.pruneSmallGroups(.05f);
	

	cout << "system energy is " << canvas.systemEnergy()
		 << " max tesion is " << canvas.maximumTension()
		 << " max pull is " << canvas.maximumPullOnTransformation() << endl;
	
	const size_t MAX_ITERATIONS = 1000000;
	const double MAX_PULL_GOAL = 1e-2;
	size_t i = 0;
	while (i < MAX_ITERATIONS &&
		   canvas.maximumPullOnTransformation() > MAX_PULL_GOAL)
	{
		canvas.releaseTensionOnSystem();
		i++;
	}
	
	if (i == MAX_ITERATIONS)
	{
		cout << "WARNING: did not reach maximum pull goal of "
			 << MAX_PULL_GOAL
			 << " after having reached the maximum iterations of "
			 << MAX_ITERATIONS << endl;
	}
	
	cout << "iterated " << i << " times" << endl;
	cout << "system energy is " << canvas.systemEnergy()
		 << " max tesion is " << canvas.maximumTension()
		 << " max pull is " << canvas.maximumPullOnTransformation() << endl;
	
	std::vector<base_transform_t::Pointer> transforms;
	std::list<the_text_t> imageIDs;
	canvas.fillTransformAndImageIDVectors(transforms, imageIDs);
	
	// save the results:
	std::fstream fout;
	fout.open(fn_results, ios::out);
	
	if (!fout.is_open())
	{
		usage(argv,
			  the_text_t("could not open \'") + fn_results +
			  the_text_t("\' for writing"));
	}
	
	save_mosaic<base_transform_t>(fout,
								  pixel_spacing,
								  use_std_mask,
								  imageIDs,
								  transforms);
	fout.close();
	
	// done:
	return 0;
}
