// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : ir-refine.cxx
// Author       : Pavel A. Koshevoy
// Created      : 2005/10/27 10:31
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Refine a mosaic layout via concurrent polynomial warping
//                of the mosaic tiles, where the tile warp transforms
//                may have shared or unique transform parameters.

// the includes:
#include <itk/common.hxx>
#include <itk/itkImageMosaicVarianceMetric.h>
#include <itk/itkRadialDistortionTransform.h>
#include <itk/itkLegendrePolynomialTransform.h>
#include <itk/itkRegularStepGradientDescentOptimizer2.h>
#include <utils/the_dynamic_array.hxx>
#include <thread/the_boost_mutex.hxx>
#include <thread/the_boost_thread.hxx>

// system includes:
#include <math.h>

// ITK includes:
#include <itkDiscreteGaussianImageFilter.h>
#include <itkShrinkImageFilter.h>

//----------------------------------------------------------------
// SAVE_INTERMEDIATE_MOSAIC
// 
#define SAVE_INTERMEDIATE_MOSAIC


//----------------------------------------------------------------
// optimizer_t
//
typedef itk::RegularStepGradientDescentOptimizer2 optimizer_t;


//----------------------------------------------------------------
// optimization_observer_t
// 
template<class mosaic_metric_t>
class optimization_observer_t : public itk::Command
{
public:
  typedef optimization_observer_t Self;
  typedef itk::Command Superclass;
  typedef itk::SmartPointer<Self> Pointer;
  
  itkNewMacro(Self);
  
  void Execute(itk::Object *caller, const itk::EventObject & event)
  { Execute((const itk::Object *)(caller), event); }
  
  void Execute(const itk::Object * object, const itk::EventObject & event)
  {
    if (typeid(event) != typeid(itk::IterationEvent)) return;
    
    const optimizer_t * optimizer = dynamic_cast<const optimizer_t *>(object);
    cout << (unsigned int)(optimizer->GetCurrentIteration()) << '\t'
	 << optimizer->GetValue() << '\t'
	 << metric_->GetNumberOfPixelsCounted() << '\t'
	 << optimizer->GetCurrentPosition() << endl;
    
    // FIXME: 2005/11/09
#ifdef SAVE_INTERMEDIATE_MOSAIC
    {
      static unsigned int counter = 0;
      
      the_text_t fn =
	prefix_ + the_text_t::number(counter, 3, '0') + the_text_t(".mosaic");
      std::fstream fout;
      fout.open(fn, ios::out);
      save_mosaic<base_transform_t>(fout,
				    pixel_spacing_,
				    use_std_mask_,
				    in_,
				    transform_);
      fout.close();
      counter++;
    }
#endif
  }
  
  // FIXME: 2005/11/09
  typename mosaic_metric_t::Pointer metric_;
  std::list<the_text_t> in_;
  double pixel_spacing_;
  bool use_std_mask_;
  std::vector<base_transform_t::Pointer> transform_;
  the_text_t prefix_;
  
protected:
  optimization_observer_t() {}
};


//----------------------------------------------------------------
// radial_distortion_t
//
typedef itk::RadialDistortionTransform<double, 2>
radial_distortion_t;

//----------------------------------------------------------------
// affine_transform_t
// 
typedef itk::LegendrePolynomialTransform<double, 1>
affine_transform_t;

//----------------------------------------------------------------
// cubic_transform_t
// 
typedef itk::LegendrePolynomialTransform<double, 3>
cubic_transform_t;

//----------------------------------------------------------------
// image_interpolator_t
// 
typedef itk::LinearInterpolateImageFunction<image_t, double>
image_interpolator_t;


//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL)
{
  cerr << "USAGE: " << argv[0] << " "
       << "[-sh shrink_factor] "
       << "[-py multiresolution_pyramid_levels] "
       << "[-it iterations_per_level] "
       << "[-clahe slope] "
       << "[-transform { tr | rd | affine | cubic } ] "
       << "[-params unique|shared] "
       << "-load transform_params_file_in "
       << "-save transform_params_file_out "
       << endl
       << "EXAMPLE: " << argv[0] << " "
       << "-sh 8 "
       << "-py 1 "
       << "-it 10 "
       << "-transform cubic "
       << "-params shared "
       << "-load fft/s01.mosaic "
       << "-save cubic/s01.mosaic"
       << endl;
  
  if (message != NULL)
  {
    cerr << "ERROR: " << message << endl;
  }
  
  exit(1);
}

//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  // this is so that the printouts look better:
  cout.precision(3);
  cout.setf(ios::scientific);
  
  // setup thread storage for the main thread:
  set_the_thread_storage_provider(the_boost_thread_t::thread_storage);
  the_boost_thread_t MAIN_THREAD_DUMMY;
  MAIN_THREAD_DUMMY.set_stopped(false);
  
  // setup thread and mutex interface creators:
  the_mutex_interface_t::set_creator(the_boost_mutex_t::create);
  the_thread_interface_t::set_creator(the_boost_thread_t::create);
  
  // parse the command line arguments:
  unsigned int highest_resolution_level_shrink_factor = 1;
  unsigned int pyramid_levels = 1;
  double pixel_spacing = 1.0;
  bool use_std_mask = false;
  unsigned int iterations_per_level = 5;
  bool use_shared_params = false;
  double Rmax = 0.0;
  double Umax = 0.0;
  double Vmax = 0.0;
  the_text_t transform_type("tr");
  double clahe_slope = 1.0;
  the_text_t fn_results;
  
  std::list<the_text_t> in;
  std::vector<base_transform_t::Pointer> tbase;
  std::vector<image_t::Pointer> image;
  std::vector<mask_t::ConstPointer> mask;
  
  for (int i = 1; i < argc; i++)
  {
    bool ok = true;
    
    if (strcmp(argv[i], "-sh") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -sh option");
      
      i++;
      highest_resolution_level_shrink_factor = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -sh option");
    }
    else if (strcmp(argv[i], "-clahe") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -clahe option");
      
      i++;
      clahe_slope = the_text_t(argv[i]).toDouble(&ok);
      if (!ok || clahe_slope < 1.0) usage(argv, "bad -clahe option");
    }
    else if (strcmp(argv[i], "-py") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -py option");
      
      i++;
      pyramid_levels = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -py option");
    }
    else if (strcmp(argv[i], "-it") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -it option");
      
      i++;
      iterations_per_level = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -it option");
    }
    else if (strcmp(argv[i], "-params") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -params option");
      
      i++;
      use_shared_params = (the_text_t(argv[i]) == the_text_t("shared"));
    }
    else if (strcmp(argv[i], "-transform") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -transform option");
      
      i++;
      transform_type = the_text_t(argv[i]);
      if (transform_type != "tr" &&
	  transform_type != "rd" &&
	  transform_type != "affine" &&
	  transform_type != "cubic")
      {
	usage(argv, "bad -transform option");
      }
    }
    else if (strcmp(argv[i], "-load") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -load option");
      
      i++;
      std::fstream fin;
      fin.open(argv[i], ios::in);
      if (!fin.is_open())
      {
	usage(argv,
	      the_text_t("could not open \'") +
	      the_text_t(argv[i]) +
	      the_text_t("\' for reading"));
      }
      
      load_mosaic<base_transform_t>(fin,
				    pixel_spacing,
				    use_std_mask,
				    in,
				    tbase);
      unsigned int num_images = tbase.size();
      assert(num_images > 1);
      assert(pixel_spacing != 0);
      
      image.resize(num_images);
      mask.resize(num_images);
      
      unsigned int i = 0;
      for (std::list<the_text_t>::const_iterator iter = in.begin();
	   iter != in.end(); ++iter, i++)
      {
	const the_text_t & fn_image = *iter;
	
	// read the image:
	cout << setw(3) << i << " ";
	image[i] = std_tile<image_t>
	  (fn_image, highest_resolution_level_shrink_factor, pixel_spacing);
	if (use_std_mask)
	{
	  mask[i] = std_mask<image_t>(image[i]);
	}
	
	if (clahe_slope > 1.0)
	{
	  image[i] = CLAHE<image_t>(image[i],
				    255,
				    255,
				    clahe_slope,
				    256,
				    0.0,
				    255.0,
				    mask[i]);
	}
      }
      
      fin.close();
    }
    else if (strcmp(argv[i], "-save") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -save option");
      
      i++;
      fn_results = the_text_t(argv[i]);
    }
    else
    {
      usage(argv,
	    the_text_t("unknown option \'") +
	    the_text_t(argv[i]) +
	    the_text_t("\'"));
    }
  }
  
  // a minimum of two images are required for the registration:
  unsigned int num_images = in.size();
  if (num_images < 2)
  {
    usage(argv, "something is wrong with the -load parameter");
  }
  
  // some transforms cannot share the parameters:
  use_shared_params =
    use_shared_params && (transform_type == "rd" ||
			  transform_type == "cubic");
  
  if (use_shared_params)
  {
    // set all transforms to the same :
    image_t::SizeType sz = image[0]->GetLargestPossibleRegion().GetSize();
    image_t::SpacingType sp = image[0]->GetSpacing();
    double w = sz[0] * sp[0];
    double h = sz[1] * sp[1];
    Rmax = sqrt(w * w + h * h) / 2.0;
    // FIXME: -1 : 1
    Umax = w / 2.0;
    Vmax = h / 2.0;
  }
  
  // setup the shared parameters mask:
  std::vector<bool> param_shared;
  std::vector<bool> param_active;
  if (transform_type == "tr")
  {
    param_shared.assign(2, false);
    param_active.assign(2, true);
  }
  else if (transform_type == "rd")
  {
    radial_distortion_t::setup_shared_params_mask(use_shared_params,
						  param_shared);
    param_active.assign(param_shared.size(), true);
  }
  else if (transform_type == "affine")
  {
    affine_transform_t::setup_shared_params_mask(use_shared_params,
						 param_shared);
    param_active.assign(param_shared.size(), true);
  }
  else if (transform_type == "cubic")
  {
    cubic_transform_t::setup_shared_params_mask(use_shared_params,
						param_shared);
    param_active.assign(param_shared.size(), true);
  }
  
  // update the fixed transform parameters:
  std::vector<base_transform_t::Pointer> transform(num_images);
  for (unsigned int i = 0; i < num_images; i++)
  {
    base_transform_t::Pointer & ti = transform[i];
    translate_transform_t::Pointer  ti_tr;
    radial_distortion_t::Pointer    ti_rd;
    affine_transform_t::Pointer     ti_affine;
    cubic_transform_t::Pointer      ti_cubic;
    
    image_t::SizeType sz = image[i]->GetLargestPossibleRegion().GetSize();
    image_t::SpacingType sp = image[i]->GetSpacing();
    pnt2d_t bbox_min = image[i]->GetOrigin();
    pnt2d_t bbox_max;
    bbox_max[0] = bbox_min[0] + sp[0] * double(sz[0]);
    bbox_max[1] = bbox_min[1] + sp[1] * double(sz[1]);
    
    if (transform_type == "tr")
    {
      ti_tr = translate_transform_t::New();
      ti = ti_tr.GetPointer();
    }
    else if (transform_type == "rd")
    {
      ti_rd = radial_distortion_t::New();
      ti_rd->setup(bbox_min[0],
		   bbox_max[0],
		   bbox_min[1],
		   bbox_max[1],
		   Rmax);
      
      ti = ti_rd.GetPointer();
    }
    else if (transform_type == "affine")
    {
      ti_affine = affine_transform_t::New();
      ti_affine->setup(bbox_min[0],
		       bbox_max[0],
		       bbox_min[1],
		       bbox_max[1],
		       Umax,
		       Vmax);
      
      ti = ti_affine.GetPointer();
    }
    else if (transform_type == "cubic")
    {
      ti_cubic = cubic_transform_t::New();
      ti_cubic->setup(bbox_min[0],
			 bbox_max[0],
			 bbox_min[1],
			 bbox_max[1],
			 Umax,
			 Vmax);
      
      ti = ti_cubic.GetPointer();
    }
    
    translate_transform_t::Pointer tr =
      dynamic_cast<translate_transform_t *>(tbase[i].GetPointer());
    
    radial_distortion_t::Pointer rd =
      dynamic_cast<radial_distortion_t *>(tbase[i].GetPointer());
    
    affine_transform_t::Pointer affine =
      dynamic_cast<affine_transform_t *>(tbase[i].GetPointer());
    
    cubic_transform_t::Pointer cubic =
      dynamic_cast<cubic_transform_t *>(tbase[i].GetPointer());
    
    // calculate the shift:
    pnt2d_t center = pnt2d(bbox_min[0] + (bbox_max[0] - bbox_min[0]) / 2.0,
			 bbox_min[1] + (bbox_max[1] - bbox_min[1]) / 2.0);
    vec2d_t shift = tbase[i]->TransformPoint(center) - center;
    
    if (ti_tr.GetPointer() != NULL)
    {
      ti_tr->SetOffset(shift);
    }
    else if (ti_rd.GetPointer() != NULL)
    {
      if (rd.GetPointer() != NULL)
      {
	ti_rd->SetParameters(rd->GetParameters());
      }
      else
      {
	ti_rd->setup_translation(shift[0], shift[1]);
      }
    }
    else if (ti_affine.GetPointer() != NULL)
    {
      if (affine.GetPointer() != NULL)
      {
	ti_affine->SetParameters(affine->GetParameters());
	ti_affine->SetFixedParameters(affine->GetFixedParameters());
      }
      else
      {
	// FIXME: we can do better -- sample the loaded transform behaviour
	// and solve for the corresponding affine transform coefficients:
	// ti_affine->setup_translation(shift[0], shift[1]);
	solve_for_transform<image_t, affine_transform_t>(image[i],
							 mask[i],
							 tbase[i],
							 ti_affine);
      }
    }
    else if (ti_cubic.GetPointer() != NULL)
    {
      if (cubic.GetPointer() != NULL)
      {
	ti_cubic->SetParameters(cubic->GetParameters());
	ti_cubic->SetFixedParameters(cubic->GetFixedParameters());
      }
      else
      {
	// FIXME: we can do better -- sample the loaded transform behaviour
	// and solve for the corresponding cubic transform coefficients:
	// ti_cubic->setup_translation(shift[0], shift[1]);
	solve_for_transform<image_t, cubic_transform_t>(image[i],
							mask[i],
							tbase[i],
							ti_cubic);
      }
    }
    else
    {
      usage(argv,
	    the_text_t("unrecognized transform \'") +
	    the_text_t(tbase[i]->GetTransformTypeAsString().c_str()) +
	    the_text_t("\'"));
    }
  }
  
  // setup the image pyramid:
  unsigned int shrink_factor = 1;
  std::vector<std::vector<image_t::Pointer> > pyramid(pyramid_levels);
  pyramid[pyramid_levels - 1] = image;
  for (int i = pyramid_levels - 2; i >= 0; i--)
  {
    pyramid[i].resize(num_images);
    shrink_factor *= 2;
    
    for (unsigned int j = 0; j < num_images; j++)
    {
      image_t::Pointer tmp = smooth<image_t>(pyramid[i + 1][j], 2.0);
      // FIXME:
      tmp = shrink<image_t>(tmp, 2);
      pyramid[i][j] = tmp;
      
      // FIXME:
#if 0
      save<native_image_t>(cast<image_t, native_image_t>(pyramid[i][j]),
			   the_text_t("pyramid_") +
			   the_text_t::number(i, 2, '0') +
			   the_text_t("_") +
			   the_text_t::number(j, 2, '0') +
			   the_text_t(".tif"));
#endif
    }
  }
  
  cout << "shrink min: " << highest_resolution_level_shrink_factor << endl
       << "shrink max: " << (highest_resolution_level_shrink_factor *
			     shrink_factor) << endl
       << "pixel spacing: " << pixel_spacing << endl
       << "iterations per level: " << iterations_per_level << endl
       << "transform type: " << transform[0]->GetTransformTypeAsString()
       << endl;
  cout << "parameters: ";
  if (use_shared_params)
  {
    cout << "shared" << endl
	 << "Rmax: " << Rmax << endl
	 << "Umax: " << Umax << endl
	 << "Vmax: " << Vmax << endl;
  }
  else
  {
    cout << "unique" << endl;
  }
  
  // try global refinement of the mosaic:
  typedef itk::ImageMosaicVarianceMetric<image_t, image_interpolator_t>
    mosaic_metric_t;
  
  mosaic_metric_t::Pointer mosaic_metric = mosaic_metric_t::New();
  mosaic_metric->image_.resize(num_images);
  mosaic_metric->mask_.resize(num_images);
  mosaic_metric->transform_.resize(num_images);
  for (unsigned int i = 0; i < num_images; i++)
  {
    mosaic_metric->image_[i] = image[i];
    mosaic_metric->mask_[i] = mask[i];
    mosaic_metric->transform_[i] = transform[i];
  }
  mosaic_metric->setup_param_map(param_shared, param_active);
  mosaic_metric->Initialize();
  
  // setup the optimizer scales:
  mosaic_metric_t::params_t parameter_scales = 
    mosaic_metric->GetTransformParameters();
  parameter_scales.Fill(1.0);
  
#if 1
  // FIXME: de-prioritize the even power polynomial transform terms:
  if (transform_type == "cubic")
  {
  #if 0 // anything goes:
    for (unsigned int k = 0; k < num_images; k++)
    {
      const unsigned int * addr = &(mosaic_metric->address_[k][0]);
      
      for (unsigned int i = 0; i <= cubic_transform_t::Degree; i++)
      {
	for (unsigned int j = 0; j <= cubic_transform_t::Degree - i; j++)
	{
	  unsigned int a_ij = cubic_transform_t::index_a(i, j);
	  unsigned int b_ij = cubic_transform_t::index_b(i, j);
	  
	  // skip the translation coefficients:
	  if (i + j == 0)
	  {
	    // FIXME: continue;
	    
	    // encourage translation:
	    if (param_active[a_ij]) parameter_scales[addr[a_ij]] = 1e-1;
	    if (param_active[b_ij]) parameter_scales[addr[b_ij]] = 1e-1;
	  }
	  
	  // skip the odd-degree coefficients:
	  if ((i + j) % 2 != 0) continue;
	  
	  /*
	  // FIXME: discourage even-degree coefficients:
	  // double scale = pow(10.0, double(i + j));
	  double scale = 10.0;
	  if (param_active[a_ij]) parameter_scales[addr[a_ij]] = scale;
	  if (param_active[b_ij]) parameter_scales[addr[b_ij]] = scale;
	  */
	}
      }
    }
  #endif
  }
  else if (transform_type == "affine")
  {
    for (unsigned int k = 0; k < num_images; k++)
    {
      const unsigned int * addr = &(mosaic_metric->address_[k][0]);
      
      // discourage skewing and rotation:
      parameter_scales[addr[affine_transform_t::index_a(0, 1)]] = 1e+6;
      parameter_scales[addr[affine_transform_t::index_b(1, 0)]] = 1e+6;
    }
  }
#endif
  
  for (unsigned int level = 0;
       level < pyramid_levels && iterations_per_level > 0;
       level++)
  {
    for (unsigned int i = 0; i < num_images; i++)
    {
      mosaic_metric->image_[i] = pyramid[level][i];
    }
    
    mosaic_metric_t::measure_t metric_before =
      mosaic_metric->GetValue(mosaic_metric->GetTransformParameters());
    
    // run several iterations of the optimizer:
    for (unsigned int k = 0; k < 20; k++)
    {
      mosaic_metric_t::params_t params_before = 
	mosaic_metric->GetTransformParameters();
      
      mosaic_metric_t::measure_t metric_after =
	std::numeric_limits<mosaic_metric_t::measure_t>::max();
      
      // use global refinement:
      optimizer_t::Pointer optimizer = optimizer_t::New();
      optimization_observer_t<mosaic_metric_t>::Pointer observer =
	optimization_observer_t<mosaic_metric_t>::New();
      observer->metric_ = mosaic_metric;
      observer->in_ = in;
      observer->pixel_spacing_ = pixel_spacing;
      observer->use_std_mask_ = use_std_mask;
      observer->transform_ = transform;
      observer->prefix_ =
	fn_results.extract(0,
			   fn_results.size() -
			   the_text_t(".mosaic").size()) +
	the_text_t("-");
      
      optimizer->AddObserver(itk::IterationEvent(), observer);
      optimizer->SetMinimize(true);
      optimizer->SetNumberOfIterations(iterations_per_level);
      
      // FIXME: these parameters vary among transforms:
      if (transform_type == "tr")
      {
	optimizer->SetMinimumStepLength(1e-12);	// min_step
	optimizer->SetMaximumStepLength(5e-1);	// max_step
      }
      else if (transform_type == "affine")
      {
	optimizer->SetMinimumStepLength(1e-12);	// min_step
	optimizer->SetMaximumStepLength(1e-6);	// max_step
      }
      else // cubic
      {
	optimizer->SetMinimumStepLength(1e-12);	// min_step
	optimizer->SetMaximumStepLength(1e-6);	// max_step
      }
      
      // FIXME:
      /*
      optimizer->SetMaximumStepLength
	(optimizer->GetMaximumStepLength() *
	 integer_power<double>(2.0, pyramid.size() - level - 1));
      cout << "maximum step: " << optimizer->GetMaximumStepLength() << endl;
      */
      
      optimizer->SetGradientMagnitudeTolerance(1e-6);
      optimizer->SetRelaxationFactor(5e-1);
      optimizer->SetCostFunction(mosaic_metric);
      optimizer->SetInitialPosition(params_before);
      optimizer->SetScales(parameter_scales);
      optimizer->SetPickUpPaceSteps(5);
      optimizer->SetBackTracking(true);
      
      // refine the mosaic:
      try
      {
	cout << "\n" << level << '.' << k << ": refining distortion transforms"
	     << endl;
	
	optimizer->StartOptimization();
      }
      catch (itk::ExceptionObject & exception)
      {
	static unsigned int counter = 0;
	
	// oops:
	cerr << "optimizer threw an exception:" << endl << exception << endl;
	
	// FIXME:
	the_text_t fn_exc =
	  the_text_t("exception-") + the_text_t::number(counter) + ".mosaic";
	cout << "saving " << fn_exc << endl;
	
	std::fstream tmp;
	tmp.open(fn_exc, ios::out);
	save_mosaic<base_transform_t>(tmp,
				      pixel_spacing,
				      use_std_mask,
				      in,
				      transform);
	tmp.close();
      }
      mosaic_metric->SetTransformParameters(optimizer->GetBestParams());
      metric_after = optimizer->GetBestValue();
      
      mosaic_metric_t::params_t params_after = 
	mosaic_metric->GetTransformParameters();
      
      cout << "before: METRIC = " << metric_before
	   << ", PARAMS = " << params_before << endl
	   << "after:  METRIC = " << metric_after
	   << ", PARAMS = " << params_after << endl;
      
      // quantify the improvement:
      double improvement = 1.0 - metric_after / metric_before;
      bool failed_to_improve = (metric_after - metric_before) >= 0.0;
      bool negligible_improvement = !failed_to_improve && (improvement < 1e-3);
      
      if (!failed_to_improve)
      {
	cout << "IMPROVEMENT: " << setw(3) << int(100.0 * improvement)
	     << "%" << endl;
      }
      
      if (failed_to_improve)
      {
	// FIXME:
	save_mosaic<base_transform_t>(cout,
				      pixel_spacing,
				      use_std_mask,
				      in,
				      transform);
	
	cout << "NOTE: minimization failed, ignoring registration results..."
	     << endl;
	
	// previous transform was better:
	mosaic_metric->SetTransformParameters(params_before);
	break;
      }
      else if (negligible_improvement)
      {
	cout << "NOTE: improvement is negligible..." << endl;
	break;
      }
      
      // avoid recalculating the same metric:
      metric_before = metric_after;
    }
  }
  
  // save the results:
  std::fstream fout;
  fout.open(fn_results, ios::out);
  
  if (!fout.is_open())
  {
    usage(argv,
	  the_text_t("could not open \'") + fn_results +
	  the_text_t("\' for writing"));
  }
  
  save_mosaic<base_transform_t>(fout,
				pixel_spacing,
				use_std_mask,
				in,
				transform);
  fout.close();
  
  // done:
  return 0;
}
