// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : ir-shrink.cxx
// Author       : Pavel A. Koshevoy
// Created      : 2005/06/22 17:19
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Scale down an image.

// the includes:
#include <itk/common.hxx>

// system includes:
#include <math.h>


//----------------------------------------------------------------
// image_interpolator_t
// 
typedef itk::LinearInterpolateImageFunction<image_t, double>
image_interpolator_t;


//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL)
{
  if (message != NULL)
  {
    cerr << "ERROR: " << message << endl;
  }

  cerr << "USAGE: " << argv[0] << " "
       << "-sh shrink_factor "
       << "[-remap]"
       << "-load image_in "
       << "-save image_out "
       << endl
       << "EXAMPLE: " << argv[0] << " "
       << "-sh 8 "
       << "-load original.tif "
       << "-save shrunk.tif"
       << endl;
  exit(1);
}

//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  // this is so that the printouts look better:
  cout.precision(2);
  cout.setf(ios::scientific);
  
  // parse the command line arguments:
  unsigned int shrink_factor = 1;
  the_text_t fn_load;
  the_text_t fn_save;

  bool remap_values = false;
  
  for (int i = 1; i < argc; i++)
  {
    bool ok = true;
    
    if (strcmp(argv[i], "-sh") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -sh option");
      
      i++;
      shrink_factor = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -sh option");
    }
    else if (strcmp(argv[i], "-remap") == 0)
    {
      remap_values = true;
    }
    else if (strcmp(argv[i], "-load") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -load option");
      
      i++;
      fn_load = the_text_t(argv[i]);
    }
    else if (strcmp(argv[i], "-save") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -save option");
      
      i++;
      fn_save = the_text_t(argv[i]);
    }
    else
    {
      usage(argv,
	    the_text_t("unknown option \'") +
	    the_text_t(argv[i]) +
	    the_text_t("\'"));
    }
  }

  // sanity checks:
  if (fn_load.size() == 0)
  {
    usage(argv, "must specify a file to open with the -load option");
  }
  
  if (fn_save.size() == 0)
  {
    usage(argv, "must specify a file to open with the -save option");
  }
  
  if (!remap_values && shrink_factor == 1)
  {
    cerr << "WARNING: shrink factor is 1 and you aren't remapping [min, max]"
	 << " -- perhaps you simply want to copy the file?" << endl;
  }
  
  try
  {
    // read the input image, then shrink it:
    image_t::Pointer image = shrink<image_t>(load<image_t>(fn_load),
					     shrink_factor);
    
    if (remap_values)
    {
      // remap image [min, max] to [0, 255] and save as 8-bit image:
      save<native_image_t>(cast<image_t, native_image_t>
			   (remap_min_max<image_t>(image)),
			   fn_save);
    }
    else if (fn_save.match_tail(".mha") ||
	     fn_save.match_tail(".mhd") ||
	     fn_save.match_tail(".nrrd"))
    {
      // save as floating point image:
      save<image_t>(image, fn_save);
    }
    else
    {
      // save as 8-bit image:
      save<native_image_t>(cast<image_t, native_image_t>(image), fn_save);
    }
  }
  catch (...)
  {
    cerr << "WARNING: caught an unexpected exception, quitting..." << endl;
  }
  
  return 0;
}
