// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : ir-slice.cxx
// Author       : Pavel A. Koshevoy
// Created      : 2006/07/06 15:58
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Given a .stom file, generate a warped mosaic image
//                for assembling a mosaic stack volume.

// local includes:
#include "pyramid.hxx"
#include "match.hxx"

// the includes:
#include <itk/common.hxx>
#include <itk/stos.hxx>
#include <itk/stos_common.hxx>
#include <itk/itkCascadedTransform.h>
#include <itk/itkLegendrePolynomialTransform.h>
#include <utils/the_text.hxx>
#include <utils/the_utils.hxx>
#include <thread/the_boost_mutex.hxx>
#include <thread/the_boost_thread.hxx>

// ITK includes:
#include <itkCommand.h>
#include <itkImageMaskSpatialObject.h>
#include <itkMeanSquaresImageToImageMetric.h>

// system includes:
#include <math.h>
#include <stdlib.h>
#include <algorithm>
#include <functional>


//----------------------------------------------------------------
// interpolator_t
// 
typedef itk::LinearInterpolateImageFunction<native_image_t, double>
interpolator_t;

//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL)
{
  cerr << "USAGE: " << argv[0] << " "
       << "[-sh shrink_factor] "
       << "-load slice "
       << "-save image "
       << endl
       << "EXAMPLE: " << argv[0] << " "
       << "-sh 8 "
       << "-load slice.mtos "
       << "-save slice.tif "
       << endl;
  
  if (message != NULL)
  {
    cerr << "ERROR: " << message << endl;
  }
  
  exit(1);
}

//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  // this is so that the printouts look better:
  cout.precision(6);
  cout.setf(ios::scientific);
  
  // setup thread storage for the main thread:
  set_the_thread_storage_provider(the_boost_thread_t::thread_storage);
  the_boost_thread_t MAIN_THREAD_DUMMY;
  MAIN_THREAD_DUMMY.set_stopped(false);
  
  // setup thread and mutex interface creators:
  the_mutex_interface_t::set_creator(the_boost_mutex_t::create);
  the_thread_interface_t::set_creator(the_boost_thread_t::create);
  
  // parameters:
  if (argc < 5) usage(argv);
  the_text_t fn_load;
  the_text_t fn_save;
  unsigned int shrink_factor = 1;
  
  // parse the command line arguments:
  for (int i = 1; i < argc; i++)
  {
    bool ok = true;
    
    if (strcmp(argv[i], "-sh") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -sh option");
      
      i++;
      shrink_factor = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -sh option");
    }
    else if (strcmp(argv[i], "-load") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -load option");
      
      i++;
      fn_load = the_text_t(argv[i]);
    }
    else if (strcmp(argv[i], "-save") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -save option");
      
      i++;
      fn_save = the_text_t(argv[i]);
    }
    else
    {
      usage(argv,
	    the_text_t("unknown option \'") +
	    the_text_t(argv[i]) +
	    the_text_t("\'"));
    }
  }
  
  // a minimum of two images are required for the registration:
  if (fn_load.size() == 0)
  {
    usage(argv, "must specify the -load option");
  }
  
  if (fn_save.size() == 0)
  {
    usage(argv, "must specify the -save option");
  }
  
  // dump the status:
  cout << "shrink factor: " << shrink_factor << endl
       << endl;
  
  unsigned int index;
  the_text_t fn_mosaic;
  double spacing_x = 1;
  double spacing_y = 1;
  base_transform_t::Pointer transform_mtos;
  pnt2d_t overall_min;
  pnt2d_t overall_max;
  pnt2d_t slice_min;
  pnt2d_t slice_max;
  bool flip;
  
  std::fstream si;
  si.open(fn_load, ios::in);
  if (!si.is_open())
  {
    cerr << "could not open " << fn_load << "for reading, aborting..." << endl;
    ::exit(1);
  }
  else
  {
    cout << "loading " << fn_load << endl;
    load_volume_slice<base_transform_t>(si,
					index,
					fn_mosaic,
					spacing_x,
					spacing_y,
					transform_mtos,
					overall_min,
					overall_max,
					slice_min,
					slice_max,
					flip);
    si.close();
  }
  
  // allocate the slice image:
  native_image_t::SizeType slice_sz =
    calc_size(overall_min, overall_max, double(shrink_factor));
  cout << "mosaic size: " << slice_sz[0] << " x " << slice_sz[1] << endl;
  
  native_image_t::SpacingType slice_sp;
  slice_sp[0] = double(shrink_factor);
  slice_sp[1] = double(shrink_factor);
  
  native_image_t::Pointer slice = native_image_t::New();
  slice->SetRegions(slice_sz);
  slice->SetSpacing(slice_sp);
  slice->SetOrigin(pnt2d(floor(overall_min[0]), floor(overall_min[1])));
  
  try { slice->Allocate(); }
  catch (itk::ExceptionObject &)
  {
    cout << "caught an exception trying to allocate memory "
	 << "for a volume slice, size "
	 << slice_sz[0] << " x " << slice_sz[1]
	 << ", aborting..." << endl;
    ::exit(2);
  }
  
  // assemble the slices at full resolution:
  native_image_t::Pointer image;
  mask_t::Pointer mask;
  if (fn_mosaic.match_tail(".mosaic"))
  {
    if (!load_mosaic<native_image_t>(fn_mosaic,
				     flip,
				     shrink_factor,
				     1.0, // clahe slope
				     image,
				     mask))
    {
      ::exit(3);
    }
  }
  else
  {
    // load an image:
    image = std_tile<native_image_t>(fn_mosaic,
				     shrink_factor,
				     std::max(spacing_x, spacing_y));
    mask = std_mask<native_image_t>(image);
  }
  
  interpolator_t::Pointer interpolator = interpolator_t::New();
  interpolator->SetInputImage(image);
  
  // reset the slice:
  slice->FillBuffer(itk::NumericTraits<native_image_t::PixelType>::Zero);
  
  // fill in the slice:
  itk::ImageRegionIteratorWithIndex<native_image_t>
    itex(slice, slice->GetLargestPossibleRegion());
  for (itex.GoToBegin(); !itex.IsAtEnd(); ++itex)
  {
    // point expressed in the coordinate system of the volume slice:
    native_image_t::PointType uv;
    slice->TransformIndexToPhysicalPoint(itex.GetIndex(), uv);
    
    // avoid undesirable distortion artifacts:
    if (!inside_bbox(slice_min, slice_max, uv)) continue;
    
    // point expressed in the coordinate system of the slice:
    native_image_t::PointType xy = transform_mtos->TransformPoint(uv);
    
    // make sure the pixel maps into the image mask:
    if (!pixel_in_mask<mask_t>(mask, xy)) continue;
    
    // make sure the pixel maps into the image:
    if (!interpolator->IsInsideBuffer(xy)) continue;
    
    double pixel = interpolator->Evaluate(xy);
    itex.Set(native_image_t::PixelType(pixel));
  }
  
  // save the image:
  save<native_image_t>(slice, fn_save);
  
  return 0;
}
