// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : ir-smooth.cxx
// Author       : Pavel A. Koshevoy
// Created      : 2006/11/28 10:34
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : The Gaussian smoothing filter.

// the includes:
#include <itk/common.hxx>
#include <utils/the_text.hxx>


//----------------------------------------------------------------
// gaussDiffuse
// 
template <typename TImage>
typename TImage::Pointer
gaussDiffuse(const TImage * in, double rho)
{
  return smooth<TImage>(in, rho, 0.001);
}


//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL)
{
  cerr << "USAGE: " << argv[0] << " "
       << "-sigma s "
       << "-load image_in "
       << "-save image_out "
       << endl
       << "EXAMPLE: " << argv[0] << " "
       << "-sigma 5.0 "
       << "-load 001.mtos.crop2.tif "
       << "-save 001.mtos.crop2-smooth.tif "
       << endl;
  
  if (message != NULL)
  {
    cerr << "ERROR: " << message << endl;
  }
  
  exit(1);
}

//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  // this is so that the printouts look better:
  cout.precision(6);
  cout.setf(ios::scientific);
  
  // parse the command line arguments:
  double sigma = 0.0;
  the_text_t fn_load;
  the_text_t fn_save;
  
  for (int i = 1; i < argc; i++)
  {
    bool ok = true;
    
    if (strcmp(argv[i], "-sigma") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -sigma option");
      
      i++;
      sigma = the_text_t(argv[i]).toDouble(&ok);
      if (!ok || sigma <= 0.0) usage(argv, "bad -sigma option");
    }
    else if (strcmp(argv[i], "-load") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -load option");
      
      i++;
      fn_load = the_text_t(argv[i]);
    }
    else if (strcmp(argv[i], "-save") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -save option");
      
      i++;
      fn_save = the_text_t(argv[i]);
    }
    else
    {
      usage(argv,
	    the_text_t("unknown option \'") +
	    the_text_t(argv[i]) +
	    the_text_t("\'"));
    }
  }
  
  // sanity check:
  if (sigma == 0.0) usage(argv, "must specify -sigma option");
  if (fn_load.is_empty()) usage(argv, "must specify -load option");
  if (fn_save.is_empty()) usage(argv, "must specify -save option");
  
  cout << "sigma: " << sigma << endl
       << "load:  " << fn_load << endl
       << "save:  " << fn_save << endl
       << endl;
  
  image_t::Pointer image = load<image_t>(fn_load);
  image = gaussDiffuse<image_t>(image, sigma);
  
  bool remap_values =
    !fn_save.match_tail(".mha") &&
    !fn_save.match_tail(".mhd") &&
    !fn_save.match_tail(".nrrd");
  
  if (remap_values)
  {
    save<native_image_t>(cast<image_t, native_image_t>
			 (remap_min_max<image_t>(image)),
			 fn_save);
  }
  else
  {
    save<image_t>(image, fn_save);
  }
  
  // done:
  return 0;
}
