// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : ir-stom-approx.cxx
// Author       : Pavel A. Koshevoy
// Created      : 2006/05/31 14:29
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Given a set of slice to slice registration transforms,
//                for each slice N generate an approximate cascaded transform
//                for mapping from the space of slice 1 of the slice stack
//                to the space of slice N. The first N-1 transforms in the
//                cascade are approximated using a 2nd order Legendre
//                polynomial transform, and the Nth transform in the cascade
//                is the exact mapping from slice N-1 to slice N --
//                a 4th order Legendre polynomial transform.

// local includes:
#include "pyramid.hxx"
#include "match.hxx"

// the includes:
#include <itk/common.hxx>
#include <itk/stos.hxx>
#include <itk/stos_common.hxx>
#include <itk/itkCascadedTransform.h>
#include <itk/itkLegendrePolynomialTransform.h>
#include <utils/the_text.hxx>
#include <utils/the_utils.hxx>
#include <thread/the_boost_mutex.hxx>
#include <thread/the_boost_thread.hxx>

// ITK includes:
#include <itkCommand.h>
#include <itkImageMaskSpatialObject.h>
#include <itkMeanSquaresImageToImageMetric.h>

// system includes:
#include <math.h>
#include <stdlib.h>
#include <algorithm>
#include <functional>

//template class __declspec(dllimport) CStringT<TCHAR, StrTraitMFC<TCHAR, ChTraitsCRT<TCHAR> > >;
//template class __declspec(dllimport) CSimpleStringT<TCHAR>;

//----------------------------------------------------------------
// cascaded_transform_t
// 
typedef itk::CascadedTransform<double, 2> cascaded_transform_t;

//----------------------------------------------------------------
// order2_transform_t
// 
typedef itk::LegendrePolynomialTransform<double, 1> order2_transform_t;

//----------------------------------------------------------------
// order4_transform_t
// 
typedef itk::LegendrePolynomialTransform<double, 3> order4_transform_t;


//----------------------------------------------------------------
// sort
// 
// Sort the list so that the slices are sequential.
// 
// Return true if the list was sorted successfully.
// 
// Return false otherwise (may happen if there are
// disconnects or duplicates present).
// 
static bool
sort(std::list<stos_t<image_t> > & stos)
{
  if (stos.empty()) return true;
  
  std::list<stos_t<image_t> > sorted;
  sorted.push_back(remove_head(stos));
  
  while (!stos.empty())
  {
    // find a match for the head:
    bool found_head_match = false;
    for (std::list<stos_t<image_t> >::iterator i = stos.begin();
	 i != stos.end();)
    {
      const the_text_t & head = sorted.front().fn_[0];
      const the_text_t & tail = (*i).fn_[1];
      if (head == tail)
      {
	found_head_match = true;
	sorted.push_front(*i);
	i = stos.erase(i);
      }
      else
      {
	++i;
      }
    }
    
    // find a match for the tail:
    bool found_tail_match = false;
    for (std::list<stos_t<image_t> >::iterator i = stos.begin();
	 i != stos.end();)
    {
      const the_text_t & tail = sorted.back().fn_[1];
      const the_text_t & head = (*i).fn_[0];
      if (head == tail)
      {
	found_tail_match = true;
	sorted.push_back(*i);
	i = stos.erase(i);
      }
      else
      {
	++i;
      }
    }
    
    if (!found_head_match && !found_tail_match)
    {
      // could not find a match for either head or tail:
      break;
    }
  }
  
  if (!stos.empty()) return false;
  
  stos.splice(stos.end(), sorted);
  return true;
}


//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL)
{
  cerr << "USAGE: " << argv[0] << " "
       << "-save prefix "
       << "-load stos_0 [stos_i]*"
       << endl
       << "EXAMPLE: " << argv[0] << " "
       << "-save slice- "
       << "-load 000.stos 001.stos"
       << endl;
  
  if (message != NULL)
  {
    cerr << "ERROR: " << message << endl;
  }
  
  exit(1);
}

//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  // this is so that the printouts look better:
  cout.precision(6);
  cout.setf(ios::scientific);
  
  // setup thread storage for the main thread:
  set_the_thread_storage_provider(the_boost_thread_t::thread_storage);
  the_boost_thread_t MAIN_THREAD_DUMMY;
  MAIN_THREAD_DUMMY.set_stopped(false);
  
  // setup thread and mutex interface creators:
  the_mutex_interface_t::set_creator(the_boost_mutex_t::create);
  the_thread_interface_t::set_creator(the_boost_thread_t::create);
  
  // parameters:
  std::list<stos_t<image_t> > stos_list;
  the_text_t fn_prefix;
  
  // parse the command line arguments:
  unsigned int consume_default_tokens = 0;
  for (int i = 1; i < argc; i++)
  {
    if (strcmp(argv[i], "-save") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -save option");
      
      i++;
      fn_prefix = the_text_t(argv[i]);
    }
    else if (strcmp(argv[i], "-load") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -load option");
      consume_default_tokens = 2;
    }
    else if (consume_default_tokens > 0)
    {
      stos_list.push_back(stos_t<image_t>(argv[i]));
      consume_default_tokens++;
    }
    else
    {
      usage(argv,
	    the_text_t("unknown option \'") +
	    the_text_t(argv[i]) +
	    the_text_t("\'"));
    }
    
    if (consume_default_tokens > 0)
    {
      consume_default_tokens--;
    }
  }
  
  // a minimum of two images are required for the registration:
  const unsigned int num_pairs = stos_list.size();
  if (num_pairs == 0)
  {
    usage(argv, "must specify the -load option");
  }
  
  // shortcut:
  const unsigned int num_slices = num_pairs + 1;
  
  if (!sort(stos_list))
  {
    cerr << "ERROR: Could not sort the stos transforms, " << endl
	 << "       missing or duplicate transforms present. " << endl
	 << "       Could not use the following transforms: " << endl;
    
    std::list<stos_t<image_t> >::const_iterator iter;
    for (iter = stos_list.begin(); iter != stos_list.end(); ++iter)
    {
      cerr << (*iter).fn_load_ << endl;
    }
    ::exit(1);
  }
  
  // setup random access vectors:
  std::vector<stos_t<image_t> > stos(stos_list.begin(), stos_list.end());
  std::vector<the_text_t> fn_mosaic(num_slices);
  std::vector<bool> flip_mosaic(num_slices);
  std::vector<base_transform_t::Pointer> transform_stos(num_slices);
  std::vector<base_transform_t::Pointer> transform_mtos(num_slices);
  {
    fn_mosaic[0] = stos[0].fn_[0];
    flip_mosaic[0] = stos[0].flipped_[0];
    
    // first slice is fixed:
    transform_stos[0] = identity_transform_t::New();
    transform_mtos[0] = order4_transform_t::New();
    
    for (unsigned int i = 0; i < num_pairs; i++)
    {
      fn_mosaic[i + 1] = stos[i].fn_[1];
      flip_mosaic[i + 1] = stos[i].flipped_[1];
      transform_stos[i + 1] =
	const_cast<base_transform_t *>(stos[i].t01_.GetPointer());
    }
  }
  
  // initialize the mosaic bounding boxes (per slice and overall):
  std::vector<image_t::PointType> slice_min(num_slices);
  std::vector<image_t::PointType> slice_max(num_slices);
  calc_tile_mosaic_bbox(transform_mtos[0],
			stos[0].tile_min(0),
			stos[0].tile_max(0),
			slice_min[0],
			slice_max[0]);
  
  image_t::PointType overall_min = slice_min[0];
  image_t::PointType overall_max = slice_max[0];
  
  // load the images, estimate the bounding box:
  base_transform_t::Pointer prev_cascade =
    identity_transform_t::New().GetPointer();
  
  for (unsigned int i = 0; i < num_pairs; i++)
  {
    // setup the cascaded transform:
    cascaded_transform_t::Pointer cascade = cascaded_transform_t::New();
    cascade->append(prev_cascade);
    cascade->append(transform_stos[i + 1]);
    
    // initialize the mosaic-to-slice transform:
#if 0
    transform_mtos[i + 1] = cascade;
#else
    order4_transform_t::Pointer t_order4 =
      setup_transform<order4_transform_t>(stos[i].tile_min(1),
					  stos[i].tile_max(1));
    transform_mtos[i + 1] = t_order4;
    
    // calculate the corresponding slice-to-mosaic transform:
    solve_for_transform<order4_transform_t>(stos[i].tile_min(1),
					    stos[i].tile_max(1),
					    NULL,
					    cascade,
					    t_order4);

    // FIXME:
    {
      cerr << "t_order4: " << t_order4 << endl;
      solve_for_transform<order4_transform_t>(stos[i].tile_min(1),
					      stos[i].tile_max(1),
					      NULL,
					      cascade,
					      t_order4,
					      16,
					      1,
					      true);
    }
#endif
    
    // fold a rough approximation of the transform into a short cascade:
    order2_transform_t::Pointer t_order2 =
      setup_transform<order2_transform_t>(overall_min, overall_max);
    solve_for_transform<order2_transform_t>(stos[i].tile_min(1),
					    stos[i].tile_max(1),
					    NULL,
					    cascade,
					    t_order2);
    prev_cascade = t_order2;
    
    // update the bounding box:
    calc_tile_mosaic_bbox(// t_order2, // FIXME
			  t_order4,
			  // cascade,
			  stos[i].tile_min(1),
			  stos[i].tile_max(1),
			  slice_min[i + 1],
			  slice_max[i + 1]);
    update_bbox(overall_min, overall_max, slice_min[i + 1]);
    update_bbox(overall_min, overall_max, slice_max[i + 1]);
    
    image_t::SizeType sz = calc_size(overall_min, overall_max, 1.0);
    cout << "added " << stos[i].fn_[1]
	 << ",\tmosaic size:\t" << sz[0] << " x " << sz[1] << endl;
  }
  
  for (unsigned int i = 0; i < num_slices; i++)
  {
    unsigned int j = i == 0 ? 0 : i - 1;
    unsigned int k = i == 0 ? 0 : 1;
    
    the_text_t fn = fn_prefix + the_text_t::number(i, 3, '0') + ".mtos";
    std::fstream so;
    so.open(fn, ios::out);
    if (!so.is_open())
    {
      cerr << "could not open " << fn << " for writing, skipping..." << endl;
    }
    else
    {
      cout << "saving " << fn << endl;
      save_volume_slice<base_transform_t>(so,
					  i,
					  fn_mosaic[i],
					  stos[j].sp_[k][0],
					  stos[j].sp_[k][1],
					  transform_mtos[i],
					  overall_min,
					  overall_max,
					  slice_min[i],
					  slice_max[i],
					  flip_mosaic[i]);
      so.close();
    }
  }
  
  return 0;
}
