// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : ir-stom.cxx
// Author       : Pavel A. Koshevoy
// Created      : 2006/12/19 15:13
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Given a set of slice to slice registration transforms,
//                for each slice N create a cascaded transform
//                for mapping from the space of slice 1 of the slice stack
//                to the space of slice N. Generate a complete overlap
//                mask for all the slices, then generate a warped image
//                of each slice mapping into the space of the 1st slice.
//                The warped images are cropped to the bounding box of
//                the complete overlap mask. The mask is necessary, because
//                the grid transform is discontinuous, and a mapping
//                does not exist for pixels in one slice where it does
//                not overlap with one of its neighbors. In short, if
//                the complete overlap mask is not used then one of the
//                slices will throw an exception when trying to map
//                a pixel from a region where a transform isn't defined.

// local includes:
#include "pyramid.hxx"
#include "match.hxx"

// the includes:
#include <itk/common.hxx>
#include <itk/stos.hxx>
#include <itk/stos_common.hxx>
#include <itk/itkGridTransform.h>
#include <itk/itkLegendrePolynomialTransform.h>
#include <itk/itkCascadedTransform.h>
#include <utils/the_text.hxx>
#include <utils/the_utils.hxx>
#include <thread/the_boost_mutex.hxx>
#include <thread/the_boost_thread.hxx>

// ITK includes:
#include <itkCommand.h>
#include <itkImageMaskSpatialObject.h>
#include <itkMeanSquaresImageToImageMetric.h>

// system includes:
#include <math.h>
#include <stdlib.h>
#include <algorithm>
#include <functional>


//----------------------------------------------------------------
// cascaded_transform_t
// 
typedef itk::CascadedTransform<double, 2> cascaded_transform_t;

//----------------------------------------------------------------
// grid_transform_t
// 
typedef itk::GridTransform grid_transform_t;


//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL)
{
  cerr << "USAGE: " << argv[0] << " "
       << "[-sh shrink_factor] "
       << "[-clahe window slope] "
       << "[-draw-grid] "
       << "-save prefix "
       << "-load stos_0 [stos_i]*"
       << endl
       << "EXAMPLE: " << argv[0] << " "
       << "-sh 1 "
       << "-clahe 64 3.0 "
       << "-save slice- "
       << "-load 000.stos 001.stos"
       << endl;
  
  if (message != NULL)
  {
    cerr << "ERROR: " << message << endl;
  }
  
  exit(1);
}

//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  // this is so that the printouts look better:
  cout.precision(6);
  cout.setf(ios::scientific);
  
  // setup thread storage for the main thread:
  set_the_thread_storage_provider(the_boost_thread_t::thread_storage);
  the_boost_thread_t MAIN_THREAD_DUMMY;
  MAIN_THREAD_DUMMY.set_stopped(false);
  
  // setup thread and mutex interface creators:
  the_mutex_interface_t::set_creator(the_boost_mutex_t::create);
  the_thread_interface_t::set_creator(the_boost_thread_t::create);
  
  // parameters:
  std::list<stos_t<image_t> > stos_list;
  the_text_t fn_prefix;
  bool visualize_warp = false;
  unsigned int shrink_factor = 1;
  unsigned int clahe_window = 64;
  double clahe_slope = 1.0;
  
  // parse the command line arguments:
  unsigned int consume_default_tokens = 0;
  for (int i = 1; i < argc; i++)
  {
    bool ok = true;
    
    if (strcmp(argv[i], "-sh") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -sh option");
      
      i++;
      shrink_factor = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -sh option");
    }
    else if (strcmp(argv[i], "-clahe") == 0)
    {
      if ((argc - i) <= 2) usage(argv, "could not parse -clahe option");
      
      i++;
      clahe_window = the_text_t(argv[i]).toUInt(&ok);
      if (!ok || clahe_window < 2) usage(argv, "bad -clahe window option");
      
      i++;
      clahe_slope = the_text_t(argv[i]).toDouble(&ok);
      if (!ok || clahe_slope < 1.0) usage(argv, "bad -clahe slope option");
    }
    else if (strcmp(argv[i], "-draw-grid") == 0)
    {
      visualize_warp = true;
    }
    else if (strcmp(argv[i], "-save") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -save option");
      
      i++;
      fn_prefix = the_text_t(argv[i]);
    }
    else if (strcmp(argv[i], "-load") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -load option");
      consume_default_tokens = 2;
    }
    else if (consume_default_tokens > 0)
    {
      stos_list.push_back(stos_t<image_t>(argv[i]));
      consume_default_tokens++;
    }
    else
    {
      usage(argv,
	    the_text_t("unknown option \'") +
	    the_text_t(argv[i]) +
	    the_text_t("\'"));
    }
    
    if (consume_default_tokens > 0)
    {
      consume_default_tokens--;
    }
  }
  
  // a minimum of two images are required for the registration:
  const unsigned int num_pairs = stos_list.size();
  if (num_pairs == 0)
  {
    usage(argv, "must specify the -load option");
  }
  
  if (visualize_warp)
  {
    // no point in wasting time on CLAHE -- the results
    // will be destroyed anyway:
    clahe_slope = 1.0;
  }
  
  // dump the status:
  cout << "shrink factor: " << shrink_factor << endl;
  if (clahe_slope > 1.0)
  {
    cout << "clahe window: " << clahe_window << endl
	 << "clahe slope: " << clahe_slope << endl;
  }
  
  // shortcut:
  const unsigned int num_slices = num_pairs + 1;
  
  if (!sort<image_t>(stos_list))
  {
    cerr << "ERROR: Could not sort the stos transforms, " << endl
	 << "       missing or duplicate transforms present. " << endl
	 << "       Could not use the following transforms: " << endl;
    
    std::list<stos_t<image_t> >::const_iterator iter;
    for (iter = stos_list.begin(); iter != stos_list.end(); ++iter)
    {
      cerr << (*iter).fn_load_ << endl;
    }
    ::exit(1);
  }
  
  // setup random access vectors:
  std::vector<stos_t<image_t> > stos(stos_list.begin(), stos_list.end());
  std::vector<the_text_t> fn_mosaic(num_slices);
  std::vector<bool> flip_mosaic(num_slices);
  std::vector<base_transform_t::Pointer> transform_stos(num_slices);
  std::vector<base_transform_t::Pointer> transform_mtos(num_slices);
  {
    fn_mosaic[0] = stos[0].fn_[0];
    flip_mosaic[0] = stos[0].flipped_[0];
    
    // first slice is fixed:
    transform_stos[0] = identity_transform_t::New();
    transform_mtos[0] = identity_transform_t::New();
    
    for (unsigned int i = 0; i < num_pairs; i++)
    {
      fn_mosaic[i + 1] = stos[i].fn_[1];
      flip_mosaic[i + 1] = stos[i].flipped_[1];
      transform_stos[i + 1] =
	const_cast<base_transform_t *>(stos[i].t01_.GetPointer());
    }
  }
  
  // find the mosaic section mask such that for each pixel there exists
  // a mapping to all of the sections:
  native_image_t::Pointer img[2];
  mask_t::Pointer msk[2];
  
  // load the initial mask:
  if (stos[num_pairs - 1].fn_[1].match_tail(".mosaic"))
  {
    load_mosaic<native_image_t>(stos[num_pairs - 1].fn_[1],
				stos[num_pairs - 1].flipped_[1],
				shrink_factor,
				false,	// assemble mosaic
				img[1],
				true,	// assemble mosaic mask
				msk[1],
				clahe_slope,
				clahe_window);
  }
  else
  {
    img[1] = std_tile<native_image_t>(stos[num_pairs - 1].fn_[1],
				      shrink_factor,
				      std::max(stos[num_pairs - 1].sp_[1][0],
					       stos[num_pairs - 1].sp_[1][1]));
    
    if (stos[num_pairs - 1].fn_mask_[1].is_empty())
    {
      msk[1] = cast<native_image_t, mask_t>(img[1]);
      msk[1]->FillBuffer(1);
    }
    else
    {
      msk[1] = std_tile<mask_t>(stos[num_pairs - 1].fn_mask_[1],
				shrink_factor,
				std::max(stos[num_pairs - 1].sp_[1][0],
					 stos[num_pairs - 1].sp_[1][1]));
    }
  }
  
  for (int i = num_pairs - 1; i >= 0; i--)
  {
    // load the next image:
    if (stos[i].fn_[0].match_tail(".mosaic"))
    {
      load_mosaic<native_image_t>(stos[i].fn_[0],
				  stos[i].flipped_[0],
				  shrink_factor,
				  false,// assemble mosaic
				  img[0],
				  true,	// assemble mosaic mask
				  msk[0],
				  clahe_slope,
				  clahe_window);
    }
    else
    {
      img[0] = std_tile<native_image_t>(stos[i].fn_[0],
					shrink_factor,
					std::max(stos[i].sp_[0][0],
						 stos[i].sp_[0][1]));
      if (stos[i].fn_mask_[0].is_empty())
      {
	msk[0] = cast<native_image_t, mask_t>(img[0]);
	msk[0]->FillBuffer(1);
      }
      else
      {
	msk[0] = std_tile<mask_t>(stos[i].fn_mask_[0],
				  shrink_factor,
				  std::max(stos[i].sp_[0][0],
					   stos[i].sp_[0][1]));
      }
    }
    
    // allocate the clipped mask:
    mask_t::Pointer next_mask =
      make_image<mask_t>(msk[0]->GetOrigin(),
			 msk[0]->GetSpacing(),
			 msk[0]->GetLargestPossibleRegion().GetSize(),
			 0);
    
    // iterate over the previous mask, generate the next mask:
    typedef itk::ImageRegionConstIteratorWithIndex<mask_t> iter_t;
    iter_t iter(msk[0], msk[0]->GetLargestPossibleRegion());

    // temporaries:
    mask_t::IndexType index;
    pnt2d_t pt[2];
    const base_transform_t * t = stos[i].t01_.GetPointer();
    typedef itk::LinearInterpolateImageFunction<mask_t, double> interpolator_t;
    interpolator_t::Pointer interpolator = interpolator_t::New();
    interpolator->SetInputImage(msk[1]);
    
    for (iter.GoToBegin(); !iter.IsAtEnd(); ++iter)
    {
      if (iter.Get() == 0) continue;
      
      index = iter.GetIndex();
      msk[0]->TransformIndexToPhysicalPoint(index, pt[0]);
      pt[1] = t->TransformPoint(pt[0]);
      
      if (!interpolator->IsInsideBuffer(pt[1])) continue;
      double p = interpolator->Evaluate(pt[1]);
      if (p < 0.5) continue;
      
      next_mask->SetPixel(index, 1);
    }

    msk[0] = next_mask;
    
    // save results for the next pair:
    msk[1] = msk[0];
    img[1] = img[0];
  }
  
  // find the bounding box of the final mask:
  mask_t::Pointer cropped_mask;
  {
    mask_t::IndexType min;
    mask_t::IndexType max;
    
    min[0] = std::numeric_limits<int>::max();
    min[1] = std::numeric_limits<int>::max();
    
    max[0] = std::numeric_limits<int>::min();
    max[1] = std::numeric_limits<int>::min();
    
    mask_t::IndexType index;
    
    typedef itk::ImageRegionConstIteratorWithIndex<mask_t> iter_t;
    iter_t iter(msk[0], msk[0]->GetLargestPossibleRegion());
    
    for (iter.GoToBegin(); !iter.IsAtEnd(); ++iter)
    {
      if (iter.Get() == 0) continue;
      
      index = iter.GetIndex();
      
      min[0] = std::min(min[0], index[0]);
      min[1] = std::min(min[1], index[1]);
      
      max[0] = std::max(max[0], index[0]);
      max[1] = std::max(max[1], index[1]);
    }
    
    cropped_mask = crop<mask_t>(msk[0], min, max);
  }
  
  // FIXME: this is for debugging:
#if 0
  save<mask_t>(remap_min_max<mask_t>(cropped_mask),
	       fn_prefix + "mask.nrrd");
#endif
  
  // generate the slices:
  for (unsigned int i = 0; i < fn_mosaic.size(); i++)
  {
    // load the image:
    native_image_t::Pointer slice;
    mask_t::Pointer slice_mask;
    
    if (fn_mosaic[i].match_tail(".mosaic"))
    {
      load_mosaic<native_image_t>(fn_mosaic[i],
				  flip_mosaic[i],
				  shrink_factor,
				  true,	// assemble mosaic
				  slice,
				  // assemble mosaic mask only if CLAHE
				  // is requested, otherwise it's useless:
				  clahe_slope > 1.0,
				  slice_mask,
				  clahe_slope,
				  clahe_window);
    }
    else
    {
      unsigned int j = i == 0 ? 0 : i - 1;
      unsigned int k = i == 0 ? 0 : 1;
      
      slice = std_tile<native_image_t>(fn_mosaic[i],
				       shrink_factor,
				       std::max(stos[j].sp_[k][0],
						stos[j].sp_[k][1]),
				       clahe_slope,
				       clahe_window,
				       true);
    }
    
    if (visualize_warp)
    {
      // draw a grid every N pixels:
      const unsigned int N = 16;
      slice->FillBuffer(0);
      
      typedef itk::ImageRegionIteratorWithIndex<native_image_t> iter_t;
      iter_t iter(slice, slice->GetLargestPossibleRegion());
      
      for (iter.GoToBegin(); !iter.IsAtEnd(); ++iter)
      {
	native_image_t::IndexType index = iter.GetIndex();
	if (index[0] % N == 0 || index[1] % N == 0)
	{
	  iter.Set(255);
	}
      }
    }
    
    native_image_t::Pointer mosaic =
      make_image<native_image_t>
      (cropped_mask->GetOrigin(),
       cropped_mask->GetSpacing(),
       cropped_mask->GetLargestPossibleRegion().GetSize());
    
    // temporaries:
    native_image_t::IndexType index;
    pnt2d_t pt[2];
    typedef itk::LinearInterpolateImageFunction<native_image_t, double>
      interpolator_t;
    interpolator_t::Pointer interpolator = interpolator_t::New();
    interpolator->SetInputImage(slice);
    
    // iterate over the mosaic image, check against the mask,
    // and fill in the slice pixels:
    typedef itk::ImageRegionIteratorWithIndex<native_image_t> iter_t;
    iter_t iter(mosaic, mosaic->GetLargestPossibleRegion());
    
    for (iter.GoToBegin(); !iter.IsAtEnd(); ++iter)
    {
      index = iter.GetIndex();
      
      // 1. check the mask:
      if (cropped_mask->GetPixel(index) == 0) continue;
      
      // 2. find where this point maps to in the current slice:
      mosaic->TransformIndexToPhysicalPoint(index, pt[0]);
      for (unsigned int j = 0; j <= i; j++)
      {
	const base_transform_t * t = transform_stos[j].GetPointer();
	pt[1] = t->TransformPoint(pt[0]);
	pt[0] = pt[1];
      }
      
      if (!interpolator->IsInsideBuffer(pt[0])) continue;
      double p = interpolator->Evaluate(pt[0]);
      iter.Set((unsigned char)(p));
    }
    
    // save the slice:
    the_text_t fn_save = fn_prefix + the_text_t::number(i, 3, '0') + ".tif";
    save<native_image_t>(mosaic, fn_save);
  }
  
  return 0;
}
