// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : LoadMosaicTransaction.h
// Author       : Pavel A. Koshevoy
// Created      : Tue Sep 19 16:13:00 MDT 2006
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Transaction for loading a mosaic and
//                generating OpenGL textures.

#ifndef LOAD_MOSAIC_TRANSACTION_H_
#define LOAD_MOSAIC_TRANSACTION_H_

// local includes:
#include "StdTileTransaction.h"

// thelib includes:
#include <thread/the_thread_interface.hxx>
#include <thread/the_transaction.hxx>
#include <itk/common.hxx>
#include <utils/the_text.hxx>


//----------------------------------------------------------------
// LoadMosaicTransaction
// 
template <typename TImage>
class LoadMosaicTransaction : public the_transaction_t
{
public:
  LoadMosaicTransaction(const the_text_t & filename = "",
			const unsigned int & shrink_factor = 1,
			const bool flip = false,
			const bool blab = true):
    filename_(filename),
    shrink_factor_(shrink_factor),
    flip_(flip),
    blab_(blab)
  {}
  
  // virtual:
  void execute(the_thread_interface_t * thread)
  {
    itk_terminator_t terminator("LoadMosaicTransaction::execute");
    
    std::fstream fin;
    fin.open(filename_, ios::in);
    if (!fin.is_open())
    {
      itk::ExceptionObject e(__FILE__, __LINE__);
      std::ostringstream os;
      os << "could not open " << filename_
	 << " for reading, aborting...";
      e.SetDescription(os.str().c_str());
      throw e;
    }
    
    if (blab_)
    {
      blab(thread, tr(the_text_t("    reading ") + filename_));
    }
    
    double pixel_spacing = 1.0;
    bool use_std_mask = false;
    std::list<the_text_t> fn_tiles;
    std::vector<base_transform_t::Pointer> transform;
    
    ::load_mosaic<base_transform_t>(fin,
				    pixel_spacing,
				    use_std_mask,
				    fn_tiles,
				    transform);
    fin.close();
    
    unsigned int num_tiles = transform.size();
    if (num_tiles == 0)
    {
      itk::ExceptionObject e(__FILE__, __LINE__);
      e.SetDescription("empty mosaic, aborting...");
      throw e;
    }
    
    if (pixel_spacing == 0.0)
    {
      itk::ExceptionObject e(__FILE__, __LINE__);
      e.SetDescription("0 units/pixel, aborting...");
      throw e;
    }
    
    // load the individual tiles:
    std::vector<image_tile_generator_t> tile(num_tiles);
    
    unsigned int i = 0;
    for (std::list<the_text_t>::const_iterator iter = fn_tiles.begin();
	 iter != fn_tiles.end(); ++iter, i++)
    {
      const the_text_t & fn_image = *iter;
      
      // read the image:
      if (blab_)
      {
	std::ostringstream os;
	os << setw(3) << i << " loading " << fn_image;
	blab(thread, tr(os.str().c_str()));
      }
      
      try
      {
	StdTileTransaction<TImage> t(fn_image,
				     use_std_mask,
				     shrink_factor_,
				     pixel_spacing,
				     false);
	t.execute(thread);
	tile[i] = t.tile_;
      }
      catch (itk::ProcessAborted & e)
      {
	throw e;
      }
      catch (itk::ExceptionObject &)
      {
	itk::ExceptionObject e(__FILE__, __LINE__);
	std::ostringstream os;
	os << filename_
	   << " references a file that could not be loaded: "
	   << fn_image;
	e.SetDescription(os.str().c_str());
	throw e;
      }
    }
    
    // assemble the mosaic:
    if (blab_)
    {
      blab(thread, tr("    assembling tiles into a mosaic"));
    }
    
    // collect tile bounding boxes expressed in the tile coordinate system:
    std::vector<pnt2d_t> tile_min(num_tiles);
    std::vector<pnt2d_t> tile_max(num_tiles);
    for (unsigned int i = 0; i < num_tiles; i++)
    {
      tile_min[i][0] = tile[i].bbox_.min_[0];
      tile_min[i][1] = tile[i].bbox_.min_[1];
      tile_max[i][0] = tile[i].bbox_.max_[0];
      tile_max[i][1] = tile[i].bbox_.max_[1];
    }
    
    // collect tile bounding boxes expressed in the mosaic coordinate system:
    std::vector<pnt2d_t> mosaic_tile_min(num_tiles);
    std::vector<pnt2d_t> mosaic_tile_max(num_tiles);
    calc_mosaic_bboxes(transform,
		       tile_min,
		       tile_max,
		       mosaic_tile_min,
		       mosaic_tile_max);
    
    // calculate the mosaic bounding box:
    pnt2d_t mosaic_min;
    pnt2d_t mosaic_max;
    calc_mosaic_bbox(mosaic_tile_min, mosaic_tile_max, mosaic_min, mosaic_max);
    
    the_aa_bbox_t mosaic_bbox;
    mosaic_bbox << p3x1_t(mosaic_min[0], mosaic_min[1], 0)
		<< p3x1_t(mosaic_max[0], mosaic_max[1], 0);
    
    // mosaic origin:
    double origin_x = mosaic_bbox.min_.x();
    double origin_y = mosaic_bbox.min_.y();
    
    // mosaic width and height in world coordinate system:
    double mw = mosaic_bbox.max_.x() - origin_x;
    double mh = mosaic_bbox.max_.y() - origin_y;
    
    // FIXME: the mosaic currently supports uniform pixel spacing only:
    const double spacing_x = pixel_spacing;
    const double spacing_y = pixel_spacing;
    
    // corresponding discrete image dimensions
    // (equal or greater than the mosaic dimensions):
    unsigned int iw = int(ceil(mw / spacing_x));
    unsigned int ih = int(ceil(mh / spacing_y));
    
    // this will throw an exception if the mosaic image is too large:
    mosaic_.layout(iw, ih, origin_x, origin_y, spacing_x, spacing_y);
    
    // sanity check -- padding due to discretization should not be negative:
    assert((double(iw) * spacing_x - mw) >= 0.0 &&
	   (double(ih) * spacing_y - mh) >= 0.0);
    
    // allocate the mosaic:
    mosaic_.allocate(tile[0].bytes_per_pixel_);
    
    // local temporaries:
    pnt2d_t mosaic_pt;
    pnt2d_t tile_pt;
    
    const unsigned char * pixel[4];
    double weight[4];
    
    // setup the luminance/alpha interpolator:
    bool la_compressed = tile[0].bytes_per_pixel_ == 1;
    la_interpolator_t interpolator = la_compressed
      ? interpolate_compressed_luminance_alpha
      : interpolate_luminance_alpha;
    
    // iterate over the requested mosaic region:
    for (unsigned int y = 0; y < ih; y++)
    {
      mosaic_pt[1] = origin_y + spacing_y * double(y);
      
      for (unsigned int x = 0; x < iw; x++)
      {
	// make sure there hasn't been an interrupt:
	terminator.terminate_on_request();
	
	mosaic_pt[0] = origin_x + spacing_x * double(x);
	
	// pixel intensity accumulator, pixel alpha accumulator,
	// and pixel feathered weight accumulator:
	double p = 0.0;
	double a = 0.0;
	double w = 0.0;
	
	for (unsigned int k = 0; k < num_tiles; k++)
	{
	  if (!inside_bbox(mosaic_tile_min[k], mosaic_tile_max[k], mosaic_pt))
	  {
	    // the pixel falls outside the mosaic tile bounding box,
	    // don't even bother transforming it into the tile coordinate
	    // system -- it will still be outside the tile:
	    continue;
	  }
	  
	  // transform the mosaic point into the tile coordinate system:
	  tile_pt = transform[k]->TransformPoint(mosaic_pt);
	  
	  // see whether the point falls inside the tile:
	  if (tile[k].evaluate(tile_pt[0], tile_pt[1], pixel, weight))
	  {
	    // get the feathered weight:
	    double feathered_weight =
	      ::calc_pixel_weight(tile_min[k], tile_max[k], tile_pt);
	    
	    // get luminance/alpha from the tile:
	    double luminance = 0.0;
	    double alpha = 0.0;
	    interpolator(luminance, alpha, pixel, weight);
	    
	    if (alpha > 0.0)
	    {
	      // feather out the edges by giving them a tiny weight:
	      double wa = alpha == 255.0 ? 1e-0 : 1e-6;
	      
	      p += feathered_weight * luminance * wa;
	      a = std::max(a, alpha);
	      w += feathered_weight * wa;
	    }
	  }
	}
	
	if (w > 0)
	{
	  p = std::min(255.0, std::max(0.0, floor(p / w + 0.5)));
	  a = std::min(255.0, std::max(0.0, floor(a + 0.5)));
	}
	
	unsigned char * la_bytes =
	  mosaic_.scanline(y) + x * mosaic_.bytes_per_pixel_;
	assert(la_bytes + mosaic_.bytes_per_pixel_ <=
	       mosaic_.buffer_->data() + mosaic_.buffer_->size() &&
	       la_bytes >= mosaic_.buffer_->data());

	if (la_compressed)
	{
	  encode_la((unsigned char)(p), (unsigned char)(a), la_bytes[0]);
	}
	else
	{
	  la_bytes[0] = (unsigned char)(p);
	  la_bytes[1] = (unsigned char)(a);
	}
      }
    }
    
    // reset the mosaic origin:
    mosaic_.layout(iw, ih, 0.0, 0.0, spacing_x, spacing_y);
    
    // flip the mosaic as necesary:
    if (flip_)
    {
      mosaic_.flip();
    }
  }
  
  // trasaction parameters:
  the_text_t filename_;
  unsigned int shrink_factor_;
  bool flip_;
  bool blab_;
  
  // transaction output:
  image_tile_generator_t mosaic_;
};


#endif // LOAD_MOSAIC_TRANSACTION_H_
