// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : LoadMosaicTransactionITK.h
// Author       : Pavel A. Koshevoy
// Created      : Tue Sep 19 16:13:00 MDT 2006
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Transaction wrapper for loading a mosaic into an ITK image.

#ifndef LOAD_MOSAIC_TRANSACTION_ITK_H_
#define LOAD_MOSAIC_TRANSACTION_ITK_H_

// local includes:
#include <thread/the_thread_interface.hxx>
#include <thread/the_transaction.hxx>
#include <itk/common.hxx>
#include <utils/the_text.hxx>


//----------------------------------------------------------------
// LoadMosaicTransactionITK
// 
template <typename TImage>
class LoadMosaicTransactionITK : public the_transaction_t
{
public:
  LoadMosaicTransactionITK(const the_text_t & filename = the_text_t(""),
			   const unsigned int & shrink_factor = 1,
			   const bool flip = false,
			   const bool make_mask = true):
    filename_(filename),
    shrink_factor_(shrink_factor),
    flip_(flip),
    make_mask_(make_mask)
  {}
  
  // virtual:
  void execute(the_thread_interface_t * thread)
  {
    std::fstream fin;
    fin.open(filename_, ios::in);
    if (!fin.is_open())
    {
      itk::ExceptionObject e(__FILE__, __LINE__);
      std::ostringstream os;
      os << "could not open " << filename_
	 << " for reading, aborting...";
      e.SetDescription(os.str().c_str());
      throw e;
    }
    else
    {
      blab(thread, tr(the_text_t("    reading ") + filename_));
    }
    
    double pixel_spacing = 1.0;
    bool use_std_mask = false;
    std::list<the_text_t> fn_tiles;
    std::vector<base_transform_t::Pointer> transform;
    
    ::load_mosaic<base_transform_t>(fin,
				    pixel_spacing,
				    use_std_mask,
				    fn_tiles,
				    transform);
    fin.close();
    
    unsigned int num_images = transform.size();
    if (num_images == 0)
    {
      itk::ExceptionObject e(__FILE__, __LINE__);
      e.SetDescription("empty mosaic, aborting...");
      throw e;
    }
    
    if (pixel_spacing == 0.0)
    {
      itk::ExceptionObject e(__FILE__, __LINE__);
      e.SetDescription("0 units/pixel, aborting...");
      throw e;
    }
    
    // load the individual tiles:
    std::vector<typename TImage::Pointer> image(num_images);
    std::vector<mask_t::ConstPointer> mask(num_images);
    
    unsigned int i = 0;
    for (std::list<the_text_t>::const_iterator iter = fn_tiles.begin();
	 iter != fn_tiles.end(); ++iter, i++)
    {
      const the_text_t & fn_image = *iter;
      
      // read the image:
      std::ostringstream os;
      os << setw(3) << i << " loading " << fn_image;
      blab(thread, tr(os.str().c_str()));

      try
      {
	image[i] =
	  std_tile<TImage>(fn_image, shrink_factor_, pixel_spacing, false);
      }
      catch (itk::ProcessAborted & e)
      {
	throw e;
      }
      catch (itk::ExceptionObject &)
      {
	itk::ExceptionObject e(__FILE__, __LINE__);
	std::ostringstream os;
	os << filename_
	   << " references a file that could not be loaded: "
	   << fn_image;
	e.SetDescription(os.str().c_str());
	throw e;
      }
      
      // blab(thread, tr("    generating tile mask"));
      if (use_std_mask)
      {
	mask[i] = ::std_mask<TImage>(image[i]);
      }
    }
    
    // assempble the mosaic:
    blab(thread, tr("    assembling tiles into a mosaic"));
    mosaic_ =
      make_mosaic<typename TImage::Pointer, base_transform_t::Pointer>
      (FEATHER_BLEND_E, transform, image, mask);
    
    // reset the tile image origin and spacing:
    pnt2d_t origin = mosaic_->GetOrigin();
    origin[0] = 0;
    origin[1] = 0;
    mosaic_->SetOrigin(origin);

    if (make_mask_)
    {
      blab(thread, tr("    assembling mosaic mask"));
      mosaic_mask_ = make_mask<base_transform_t::Pointer>(transform, mask);
      mosaic_mask_->SetOrigin(origin);
    }
    
    if (flip_)
    {
      ::flip<TImage>(mosaic_);
      
      if (make_mask_)
      {
	::flip<mask_t>(mosaic_mask_);
      }
    }
  }
  
  // trasaction parameters:
  the_text_t filename_;
  unsigned int shrink_factor_;
  bool flip_;
  bool make_mask_;
  
  // transaction output:
  typename TImage::Pointer mosaic_;
  mask_t::Pointer mosaic_mask_;
};


#endif // LOAD_MOSAIC_TRANSACTION_ITK_H_
