// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : image_data.hxx
// Author       : Pavel A. Koshevoy
// Created      : Thu Jul 20 11:41:00 MDT 2006
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : An image and its attributes.

#ifndef IMAGE_DATA_HXX_
#define IMAGE_DATA_HXX_

// the includes:
#include <image/image_tile_generator.hxx>
#include <opengl/image_tile_dl_elem.hxx>
#include <math/the_bbox.hxx>
#include <utils/the_text.hxx>

// system includes:
#include <memory>

// Cg includes:
#include <Cg/cgGL.h>

// forward declarations:
extern CGprogram * CG_LA_FRAGMENT_PROGRAM_PTR;


//----------------------------------------------------------------
// image_data_t
//
class image_data_t
{
public:
  image_data_t():
    shrink_factor_(0),
    flipped_(false),
    max_texture_(0)
  {}
  
  // check whether the image data is ready:
  inline bool isNull() const
  { return tile_generator_.bbox_.is_empty(); }
  
  // return the bounding box of this image:
  inline the_bbox_t bbox() const
  {
    the_bbox_t b;
    const std::vector<image_tile_t> & tiles = tile_generator_.tiles_;
    for (unsigned int i = 0; i < tiles.size(); i++)
    {
      tiles[i].update_bbox(b);
    }
    
    return b;
  }
  
  // return image size:
  inline unsigned int size(unsigned int dimension) const
  {
    if (isNull()) return 0;
    
    return
      (dimension == 0) ?
      tile_generator_.w_ :
      tile_generator_.h_;
  }
  
  // return pixel spacing:
  inline double spacing(unsigned int dimension = 0) const
  {
    if (isNull()) return 0.0;
    
    return
      (dimension == 0) ?
      tile_generator_.spacing_x_ :
      tile_generator_.spacing_y_;
  }
  
  // return image size information string:
  inline the_text_t imageSize() const
  {
    if (isNull()) return "";
    
    the_text_t txt =
      the_text_t::number(tile_generator_.w_) + " x " +
      the_text_t::number(tile_generator_.h_);
    return txt;
  }
  
  // return image pixel size information string:
  inline the_text_t pixelSize() const
  {
    if (isNull()) return "";
    
    the_text_t txt =
      the_text_t::number(tile_generator_.spacing_x_) + "mm x " +
      the_text_t::number(tile_generator_.spacing_y_) + "mm";
    return txt;
  }
  
  // generate a display list element:
  inline std::auto_ptr<image_tile_dl_elem_t>
  dl_elem(const GLenum & min_filter,
	  const GLenum & mag_filter) const
  {
#ifdef USE_CG
    return std::auto_ptr<image_tile_dl_elem_t>(new image_tile_dl_elem_t
					       (tile_generator_,
						min_filter,
						mag_filter,
						CG_LA_FRAGMENT_PROGRAM_PTR));
#else
    return std::auto_ptr<image_tile_dl_elem_t>(new image_tile_dl_elem_t
					       (tile_generator_,
						min_filter,
						mag_filter));
#endif
  }
  
  // image attributes:
  the_text_t name_;
  unsigned int shrink_factor_;
  bool flipped_;
  
  // OpenGL texture attributes:
  unsigned int max_texture_;
  
  // texture tiles:
  image_tile_generator_t tile_generator_;
};


#endif // IMAGE_DATA_HXX_
