// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : the_main_window.cxx
// Author       : Pavel A. Koshevoy
// Created      : 2007/01/03 17:20
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : main application window widget

// system includes:
#include <vector>

// Qt includes:
#include <QWidget>
#include <QTabWidget>
#include <QApplication>
#include <QMessageBox>
#include <QInputDialog>
#include <QFileDialog>
#include <QFileInfo>
#include <QAction>
#include <QListView>
#include <QPixmap>
#include <QGridLayout>
#include <QList>
#include <QDateTime>

// local includes:
#include "the_main_window.hxx"
#include "the_track_proc_ui.hxx"
#include "the_track_proc.hxx"

// thelib includes:
#include <math/the_color.hxx>
#include <opengl/the_appearance.hxx>
#include <eh/the_view_mgr_eh.hxx>
#include <io/the_file_io.hxx>
#include <utils/the_indentation.hxx>
#include <Qt/the_qt_trail.hxx>
#include <Qt/the_qt_view.hxx>


//----------------------------------------------------------------
// orientation
// 
static the_view_mgr_orientation_t orientation[] =
{
  THE_BOTTOM_VIEW_E
};


//----------------------------------------------------------------
// the_main_window_t::the_main_window_t
// 
the_main_window_t::the_main_window_t():
  QMainWindow(NULL, 0),
  the_document_ui_t("ir-track/20070102"),
  track_proc_ui_(NULL),
  preferences_ui_(this)
{
  the_qt_view_t * shared =
    new the_qt_view_t(this, "shared", NULL, THE_BOTTOM_VIEW_E);
  
  the_view_t * qt_view[] = {
    new the_qt_view_t(this, "feature tracking view", shared, orientation[0])
  };
  setup(shared, qt_view, 1);
  
  // disable 3D rotation:
#ifndef DEBUG_TEXTURES
  view_mgr_eh_[0]->allow_spin(false);
#endif
  
  setupUi(this);
  
  // signals and slots connections:
  actionNew->setShortcutContext(Qt::ApplicationShortcut);
  actionLoad->setShortcutContext(Qt::ApplicationShortcut);
  actionQuit->setShortcutContext(Qt::ApplicationShortcut);
  actionRedo->setShortcutContext(Qt::ApplicationShortcut);
  actionUndo->setShortcutContext(Qt::ApplicationShortcut);
  actionDelete->setShortcutContext(Qt::ApplicationShortcut);
  actionRestore->setShortcutContext(Qt::ApplicationShortcut);
  actionAbout->setShortcutContext(Qt::ApplicationShortcut);
  
  connect(actionNew, SIGNAL(triggered()),
	  this, SLOT(file_new()));
  connect(actionQuit, SIGNAL(triggered()),
	  this, SLOT(file_exit()));
  
  connect(actionRedo, SIGNAL(triggered()),
	  this, SLOT(edit_redo()));
  connect(actionUndo, SIGNAL(triggered()),
	  this, SLOT(edit_undo()));
  connect(actionPreferences, SIGNAL(triggered()),
	  this, SLOT(edit_preferences()));
  
  connect(actionRestore, SIGNAL(triggered()),
	  this, SLOT(orientation_restore()));
  
  connect(actionAbout, SIGNAL(triggered()),
	  this, SLOT(helpAbout()));
  
  // UI objects of various procedures:
  track_proc_ui_ = new the_track_proc_ui_t(this);
  
  QGridLayout * layout = new QGridLayout(centralwidget);
  layout->setSpacing(0);
  layout->setMargin(0);
  view<the_qt_view_t>(0)->setParent(centralwidget);
  layout->addWidget(view<the_qt_view_t>(0), 0, 0);
  view<the_qt_view_t>(0)->show();
  
  // initialize the UI:
  sync_ui();
}

//----------------------------------------------------------------
// the_main_window_t::~the_main_window_t
// 
the_main_window_t::~the_main_window_t()
{
  assert(track_proc_ui_ == NULL);
}

//----------------------------------------------------------------
// ok_to_discard_current_document
// 
static bool
ok_to_discard_current_document(the_main_window_t * main_window,
			       the_document_so_t & doc_so)
{
  if (doc_so.changes_saved()) return true;
  
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  QMessageBox mb(main_window);
  mb.setIcon(QMessageBox::Question);	
  mb.setWindowTitle(QString("Save changes?"));
  
  std::ostringstream os;
  os << "The document \"" << doc_so.document()->name()
     <<"\" has unsaved changes." << endl;
  mb.setText(QString(os.str().c_str()));
  
  os.str("");
  os << "Click [Save] to save the changes and proceed." << endl
     << "Click [Discard] to discard the changes and proceed." << endl
     << "Click [Cancel] to keep the changes and continue working." << endl;
  mb.setDetailedText(QString(os.str().c_str()));
  
  QAbstractButton * btn_save =
    mb.addButton(QObject::tr("Save"), QMessageBox::ActionRole);
  
  QAbstractButton * btn_discard =
    mb.addButton(QObject::tr("Discard"), QMessageBox::ActionRole);
  
  // QAbstractButton * btn_cancel =
  mb.addButton(QObject::tr("Cancel"), QMessageBox::ActionRole);
  
  mb.exec();
  if (mb.clickedButton() == btn_save)
  {
    cerr << "implement file->save" << endl;
    assert(0);
    if (doc_so.changes_saved()) return true;
    
    QMessageBox::critical(main_window,
			  QString("Could not save changes"),
			  QString("Could not save the document, aborting..."),
			  QMessageBox::Ok,
			  QMessageBox::NoButton,
			  QMessageBox::NoButton);
    return false;
  }
  
  return (mb.clickedButton() == btn_discard);
}

//----------------------------------------------------------------
// the_main_window_t::file_new
// 
void
the_main_window_t::file_new()
{
  // check whether the existing document needs to be saved:
  if (ok_to_discard_current_document(this, doc_so()) == false) return;
  
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  // start a new document:
  doc_so().new_document(new the_document_t
			(QDateTime::currentDateTime().
			 toString("yyyyMMdd-hhmm").toUtf8().data()));
  if (!track_proc_ui_->installed())
  {
    tool_track();
  }
  else
  {
    document()->add_proc(new the_track_proc_t());
  }
  
  sync_ui();
}

//----------------------------------------------------------------
// the_main_window_t::file_exit
// 
void
the_main_window_t::file_exit()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  // check whether the existing document needs to be saved:
  if (ok_to_discard_current_document(this, doc_so()) == false) return;
  
  // stop trail playback:
  THE_TRAIL.stop();
  
  // shut down the procedure UI:
  track_proc_ui_->uninstall();
  delete track_proc_ui_;
  track_proc_ui_ = NULL;
  
  // close the document:
  doc_so().close_document();
  
  // shut dowm the document UI:
  shutdown();
  
  // done:
  qApp->quit();
}

//----------------------------------------------------------------
// the_main_window_t::edit_undo
// 
void
the_main_window_t::edit_undo()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  doc_so().undo();
  document()->active_procedure<the_track_proc_t>()->request_regeneration();
  sync_ui();
}

//----------------------------------------------------------------
// the_main_window_t::edit_redo
// 
void
the_main_window_t::edit_redo()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  doc_so().redo();
  document()->active_procedure<the_track_proc_t>()->request_regeneration();
  sync_ui();
}

extern CGprogram CG_LA_FRAGMENT_PROGRAM;
extern CGprogram * CG_LA_FRAGMENT_PROGRAM_PTR;
extern bool CG_LA_FRAGMENT_PROGRAM_OK;

//----------------------------------------------------------------
// the_main_window_t::edit_preferences
// 
void
the_main_window_t::edit_preferences()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  int r = preferences_ui_.exec();
  if (r == QDialog::Accepted)
  {
    document()->active_procedure()->request_regeneration();
    sync_ui();
  }
}

//----------------------------------------------------------------
// the_main_window_t::toggle_track_ui
// 
void
the_main_window_t::toggle_track_ui(bool on)
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  if (on && track_proc_ui_->isHidden())
  {
    track_proc_ui_->show();
  }
  else if (!on && track_proc_ui_->isVisible())
  {
    track_proc_ui_->hide();
  }
}

//----------------------------------------------------------------
// the_main_window_t::orientation_restore
// 
void
the_main_window_t::orientation_restore()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  std::vector<the_view_t *> views;
  get_visible_views(views);
  unsigned int num_views = views.size();
  for (unsigned int i = 0; i < num_views; i++)
  {
    views[i]->restore_orientation();
  }
}

//----------------------------------------------------------------
// the_main_window_t::helpAbout
// 
void
the_main_window_t::helpAbout()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  AboutDialog about(this);
  about.exec();
}

//----------------------------------------------------------------
// the_main_window_t::tool_track
// 
void
the_main_window_t::tool_track()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  toggle_track_ui(true);
  track_proc_ui_->install();
}

//----------------------------------------------------------------
// the_main_window_t::sync_ui
// 
void
the_main_window_t::sync_ui()
{
  // setup 3D viewing (necessary in order for the compiled display
  // lists to tesselate the splines correctly):
  shared()->gl_make_current();
  
  // shortcut:
  const the_document_t * doc = document();
  bool has_doc = (doc != NULL);
  
  // shortcut:
  the_procedure_ui_t * proc_ui = active_proc_ui();
  if (proc_ui != NULL) proc_ui->sync_ui();
  
  // update the views:
  for (unsigned int i = 0; i < view_.size(); i++)
  {
    the_qt_view_t * v = view<the_qt_view_t>(i);
    if (v->isHidden()) continue;
    v->refresh();
  }
  
  // update the actions:
  bool undo_ok = has_doc && !doc_so().undo_stack_empty();
  actionUndo->setEnabled(undo_ok);
  actionUndoIcon->setEnabled(undo_ok);

  bool redo_ok = has_doc && !doc_so().redo_stack_empty();
  actionRedo->setEnabled(redo_ok);
  actionRedoIcon->setEnabled(redo_ok);
}

//----------------------------------------------------------------
// the_main_window_t::kill_ui
// 
void
the_main_window_t::kill_ui()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  file_exit();
}

//----------------------------------------------------------------
// the_main_window_t::get_visible_views
// 
void
the_main_window_t::get_visible_views(std::vector<the_view_t *> & views) const
{
  views = view_;
}

//----------------------------------------------------------------
// the_main_window_t::closeEvent
// 
void
the_main_window_t::closeEvent(QCloseEvent * e)
{
  e->ignore();
  kill_ui();
}
