// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : the_track_proc.cxx
// Author       : Pavel A. Koshevoy
// Created      : Tue Jan 2 16:01:00 MDT 2007
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : The document procedure containing the registered slices
//                and tracking data.

// local includes:
#include "the_track_proc.hxx"
#include "the_main_window.hxx"
#include "ImageDataTransaction.h"
#include "ImageProcessingDialog.h"
#include "PreferencesDialog.h"

// thelib includes:
#include <doc/the_graph.hxx>
#include <ui/the_document_ui.hxx>
#include <ui/the_trail.hxx>
#include <eh/the_input_device_eh.hxx>
#include <itk/common.hxx>
#include <math/the_color.hxx>
#include <opengl/the_view.hxx>
#include <opengl/the_disp_list.hxx>
#include <opengl/the_point_symbols.hxx>


//----------------------------------------------------------------
// the_track_proc_t::the_track_proc_t
// 
the_track_proc_t::the_track_proc_t():
  the_procedure_t(),
  image_(new the_dynamic_array_t<image_data_t>()),
  color_(8),
  active_slice_(0),
  active_label_(0),
  dl_image_(new the_disp_list_t()),
  dl_(new the_disp_list_t())
{
  the_color_t colors[8];
  make_rainbow(8, colors, false);
  for (unsigned int i = 0; i < 8; i++)
  {
    color_[i] = colors[i];
  }
}

//----------------------------------------------------------------
// the_track_proc_t::~the_track_proc_t
// 
the_track_proc_t::~the_track_proc_t()
{}

//----------------------------------------------------------------
// the_track_proc_t::added_to_the_registry
// 
void
the_track_proc_t::added_to_the_registry(the_registry_t * registry,
					const unsigned int & id)
{
  the_procedure_t::added_to_the_registry(registry, id);
}

//----------------------------------------------------------------
// the_points_dl_elem_t
// 
class the_points_dl_elem_t : public the_dl_elem_t
{
public:
  the_points_dl_elem_t(const std::list<mask_pt_t> & points,
		       const the_color_t & color):
    points_(points),
    color_(color)
  {}
  
  // virtual:
  void draw() const
  {
    glDisable(GL_LIGHTING);
    glEnable(GL_BLEND);
    
    for (std::list<mask_pt_t>::const_iterator iter = points_.begin();
	 iter != points_.end(); ++iter)
    {
      const mask_pt_t & mask_pt = *iter;
      
#if 1
      pnt2d_t mosaic_pt;
      mosaic_pt[0] = (mask_pt.mosaic_quad_[0][0] +
		      mask_pt.mosaic_quad_[1][0] +
		      mask_pt.mosaic_quad_[2][0] +
		      mask_pt.mosaic_quad_[3][0]) / 4.0;
      
      mosaic_pt[1] = (mask_pt.mosaic_quad_[0][1] +
		      mask_pt.mosaic_quad_[1][1] +
		      mask_pt.mosaic_quad_[2][1] +
		      mask_pt.mosaic_quad_[3][1]) / 4.0;

      p3x1_t pt(mosaic_pt[0], mosaic_pt[1], 0);
      the_symbol_dl_elem_t e(pt, color_,
			     THE_POINT_SYMBOLS,
			     THE_SMALL_FILLED_CIRCLE_SYMBOL_E);
      e.draw();
#endif
      
      glBegin(GL_QUADS);
      {
	glColor4f(color_[0], color_[1], color_[2], 0.5f);
	glVertex2f(mask_pt.mosaic_quad_[0][0],
		   mask_pt.mosaic_quad_[0][1]);
	glVertex2f(mask_pt.mosaic_quad_[1][0],
		   mask_pt.mosaic_quad_[1][1]);
	glVertex2f(mask_pt.mosaic_quad_[2][0],
		   mask_pt.mosaic_quad_[2][1]);
	glVertex2f(mask_pt.mosaic_quad_[3][0],
		   mask_pt.mosaic_quad_[3][1]);
      }
      glEnd();
    }
  }
  
  // virtual:
  void update_bbox(the_bbox_t & bbox) const
  {
    for (std::list<mask_pt_t>::const_iterator iter = points_.begin();
	 iter != points_.end(); ++iter)
    {
      const mask_pt_t & mask_pt = *iter;
      for (unsigned int i = 0; i < 4; i++)
      {
	bbox << p3x1_t(mask_pt.mosaic_quad_[i][0],
		       mask_pt.mosaic_quad_[i][1],
		       0);
      }
    }
  }
  
protected:
  const std::list<mask_pt_t> & points_;
  the_color_t color_;
};

//----------------------------------------------------------------
// the_track_proc_t::regenerate
// 
bool
the_track_proc_t::regenerate()
{
  // rebuild the display lists:
  dl_->clear();
  dl_image_->clear();
  
  // regenerate the dependents that need to be regenerated:
  the_registry_t * r = registry();
  the_graph_t graph;
  graph.set_roots(r, direct_dependents());
  regenerated_ = graph.regenerate();
  
  GLenum filter = prefsAllowLinearFiltering() ? GL_LINEAR : GL_NEAREST;
  if (image_ready())
  {
    dl_image_->push_back(image().dl_elem(filter, filter).release());
    
    // regenerate the mask display list:
    for (unsigned int i = 0; i < color_.size(); i++)
    {
      const std::list<mask_pt_t> & mask = mask_[i][active_slice_];
      
      // make the mask semi-transparent:
      the_color_t color = color_[i];
      dl_->push_back(new the_points_dl_elem_t(mask, color));
    }
  }
  
  return regenerated_;
}

//----------------------------------------------------------------
// the_track_proc_t::draw
// 
void
the_track_proc_t::draw(const the_view_t & view) const
{
  view.view_mgr().reset_opengl_viewing();
  view.view_mgr().setup_opengl_3d_viewing();
  
  glPushAttrib(GL_ENABLE_BIT);
  {
    glDisable(GL_DEPTH_TEST);
    glEnable(GL_BLEND);
    dl_image_->draw();
    dl_->execute();
  }
  glPopAttrib();
}

//----------------------------------------------------------------
// the_track_proc_t::calc_bbox
// 
void
the_track_proc_t::calc_bbox(const the_view_t & /* view */,
			    the_bbox_t & bbox) const
{
  bbox += dl_->bbox();
    
  if (image_ready())
  {
    bbox += image().bbox();
  }
}

//----------------------------------------------------------------
// the_track_proc_t::reset
// 
void
the_track_proc_t::
reset(const the_dynamic_array_t<stos_t<native_image_t> > & stos,
      const the_dynamic_array_t<image_data_t> & image_data,
      const the_dynamic_array_t<base_transform_t::ConstPointer> &
      exact_local,
      const the_dynamic_array_t<base_transform_t::ConstPointer> &
      approx_global)
{
  stos_ = stos;
  *(image_.get()) = image_data;
  exact_local_ = exact_local;
  approx_global_ = approx_global;
  
  // reset the masks:
  mask_ = labeled_masks_t(color_.size(), 8, slice_masks_t(image_data.size()));
  active_slice_ = 0;
}

//----------------------------------------------------------------
// the_track_proc_t::image_ready
// 
bool
the_track_proc_t::image_ready() const
{
  return image_.get() != NULL && !image().isNull();
}

//----------------------------------------------------------------
// the_track_proc_t::set_active_slice
// 
void
the_track_proc_t::set_active_slice(const unsigned int & index)
{
  active_slice_ = index;
  request_regeneration();
}

//----------------------------------------------------------------
// the_track_proc_t::set_active_label
// 
void
the_track_proc_t::set_active_label(const unsigned int & index)
{
  active_label_ = std::min(index, color_.size() - 1);
}

//----------------------------------------------------------------
// the_track_proc_t::save
// 
bool
the_track_proc_t::save(std::ostream & stream) const
{
  return the_procedure_t::save(stream);
}

//----------------------------------------------------------------
// the_track_proc_t::load
// 
bool
the_track_proc_t::load(std::istream & stream)
{
  return the_procedure_t::load(stream);
}

//----------------------------------------------------------------
// the_track_proc_t::dump
// 
void
the_track_proc_t::dump(ostream & strm, unsigned int indent) const
{
  strm << INDSCP << "the_track_proc_t(" << (void *)this << ")" << endl
       << INDSCP << "{" << endl
       << INDSCP << "}" << endl << endl;
}
