// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : the_track_proc.hxx
// Author       : Pavel A. Koshevoy
// Created      : Tue Jan 2 16:00:00 MDT 2007
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : The document procedure containing the registered slices
//                and tracking data.

#ifndef THE_TRACK_PROC_HXX_
#define THE_TRACK_PROC_HXX_

// Boost includes:
#include <boost/shared_ptr.hpp>

// local includes:
#include "image_data.hxx"

// thelib includes:
#include <doc/the_procedure.hxx>
#include <itk/stos.hxx>
#include <math/the_color.hxx>
#include <utils/the_dynamic_array.hxx>
#include <utils/the_unique_list.hxx>
#include <opengl/the_disp_list.hxx>

// forward declarations:
class the_bbox_t;
class the_view_t;
class the_view_volume_t;


//----------------------------------------------------------------
// mask_pt_t
// 
class mask_pt_t
{
public:
  inline bool operator == (const mask_pt_t & p) const
  { return tile_pt_ == p.tile_pt_; }
  
  pnt2d_t tile_pt_;
  pnt2d_t mosaic_quad_[4];
};


//----------------------------------------------------------------
// the_track_proc_t
// 
class the_track_proc_t : public the_procedure_t
{
public:
  the_track_proc_t();
  ~the_track_proc_t();
  
  // virtual:
  the_primitive_t * clone() const
  { return new the_track_proc_t(*this); }
  
  // virtual:
  const char * name() const
  { return "the_track_proc_t"; }
  
  // virtual:
  void added_to_the_registry(the_registry_t * registry,
			     const unsigned int & id);
  
  // virtual:
  bool regenerate();
  
  // virtual:
  void draw(const the_view_t & view) const;
  
  // virtual: calculate the model bounding box:
  void calc_bbox(const the_view_t & view, the_bbox_t & bbox) const;
  
  // re-initialize the procedure:
  void reset(const the_dynamic_array_t<stos_t<native_image_t> > & stos,
	     const the_dynamic_array_t<image_data_t> & image_data,
	     
	     const the_dynamic_array_t<base_transform_t::ConstPointer> &
	     exact_local,
	     const the_dynamic_array_t<base_transform_t::ConstPointer> &
	     approx_global);
  
  // accessors:
  inline const image_data_t & image() const
  { return image_->elem(active_slice_); }
  
  inline image_data_t & image()
  { return image_->elem(active_slice_); }
  
  // check whether the fixed/moving images are available:
  bool image_ready() const;
  
  // activate a slice (load the image data):
  void set_active_slice(const unsigned int & slice_index);
  
  // set active label index:
  void set_active_label(const unsigned int & label_index);
  
  // virtual: file io:
  bool save(std::ostream & stream) const;
  bool load(std::istream & stream);
  
  // virtual: For debugging, dumps the value
  void dump(ostream & strm, unsigned int indent = 0) const;
  
public:
  // the slice to slice registration records:
  the_dynamic_array_t<stos_t<native_image_t> > stos_;
  
  // fixed/moving/mosaic image data:
  boost::shared_ptr<the_dynamic_array_t<image_data_t> > image_;
  
  // transforms mapping from one slice into the next:
  the_dynamic_array_t<base_transform_t::ConstPointer> exact_local_;
  
  // transforms mapping from the first slice into each slice:
  the_dynamic_array_t<base_transform_t::ConstPointer> approx_global_;
  
  // label masks (collection of points) for each slice:
  typedef the_dynamic_array_t<the::unique_list<mask_pt_t> > slice_masks_t;
  typedef the_dynamic_array_t<slice_masks_t> labeled_masks_t;
  labeled_masks_t mask_;
  
  // label colors:
  the_dynamic_array_t<the_color_t> color_;
  
  // currently active (displayed) slice:
  int active_slice_;
  
  // currently active (used to paint) label color:
  int active_label_;
  
protected:
  // texture display lists (not compiled):
  boost::shared_ptr<the_disp_list_t> dl_image_;
  
  // misc display lists (compiled):
  boost::shared_ptr<the_disp_list_t> dl_;
};


#endif // THE_TRACK_PROC_HXX_
