// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : ImageDataTransaction.h
// Author       : Pavel A. Koshevoy
// Created      : Tue Sep 19 16:23:00 MDT 2006
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Transaction for setting up the image_data_t structure.

#ifndef IMAGE_DATA_TRANSACTION_H_
#define IMAGE_DATA_TRANSACTION_H_

// local includes:
#include "StdTileTransaction.h"
#include "LoadMosaicTransaction.h"
#include "image_data.hxx"

// thelib includes:
#include <thread/the_thread_interface.hxx>
#include <thread/the_transaction.hxx>
#include <image/image_tile_generator.hxx>
#include <itk/common.hxx>


//----------------------------------------------------------------
// ImageDataTransaction
// 
template <typename TImage>
class ImageDataTransaction : public the_transaction_t
{
public:
  ImageDataTransaction(const the_text_t & filename = "",
		       unsigned int shrink_factor = 1,
		       double pixel_spacing = 1.0,
		       bool flip = false,
		       unsigned int max_texture = 64,
		       GLenum internal_format = GL_LUMINANCE_ALPHA):
    internal_format_(internal_format)
  {
    data_.name_ = filename;
    data_.shrink_factor_ = shrink_factor;
    data_.pixel_spacing_ = pixel_spacing;
    data_.flipped_ = flip;
    data_.max_texture_ = max_texture;
  }
  
  // virtual:
  void execute(the_thread_interface_t * thread)
  {
    image_tile_generator_t tile_generator;
    
    if (data_.name_.match_tail(".mosaic"))
    {
      // load the mosaic:
      LoadMosaicTransaction<TImage> transaction(data_.name_,
						data_.shrink_factor_,
						data_.flipped_);
      transaction.execute(thread);
      tile_generator = transaction.mosaic_;
    }
    else
    {
      // load a standard tile:
      StdTileTransaction<TImage> transaction(data_.name_,
					     false,
					     data_.shrink_factor_,
					     data_.pixel_spacing_);
      transaction.execute(thread);
      tile_generator = transaction.tile_;
      
      if (data_.flipped_)
      {
	tile_generator.flip();
      }
    }
    
    blab(thread, tr("generating OpenGL texture tile layout\n"));
    
    if (tile_generator.bytes_per_pixel_ == 2)
    {
      tile_generator.make_tiles(GL_UNSIGNED_BYTE,	// data type
				internal_format_,	// internal format
				GL_LUMINANCE_ALPHA,	// external format
				data_.max_texture_);
    }
    else if (tile_generator.bytes_per_pixel_ == 1)
    {
      tile_generator.make_tiles(GL_UNSIGNED_BYTE,	// data type
				GL_LUMINANCE,		// internal format
				GL_LUMINANCE,		// external format
				data_.max_texture_);
    }
    else
    {
      assert(false);
    }
    
    data_.tile_generator_ = tile_generator;
  }
  
  // transaction parameters:
  GLenum internal_format_;
  
  // transaction output:
  image_data_t data_;
};


#endif // IMAGE_DATA_TRANSACTION_H_
