// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : ImageLoaderDialog.cpp
// Author       : Pavel A. Koshevoy
// Created      : Thu Jul 13 12:01:00 MDT 2006
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : UI for the image loader.

// local includes:
#include "ImageLoaderDialog.h"
#include "ImageProcessingDialog.h"
#include "PreferencesDialog.h"
#include "ImageDataTransaction.h"

// thelib includes:
#include <ui/the_trail.hxx>
#include <opengl/OpenGLCapabilities.h>

// Qt includes:
#include <QLineEdit>
#include <QPushButton>
#include <QStackedWidget>
#include <QTextEdit>
#include <QFileDialog>
#include <QCloseEvent>
#include <QFile>
#include <QTime>


//----------------------------------------------------------------
// ImageLoaderDialog::ImageLoaderDialog
// 
ImageLoaderDialog::ImageLoaderDialog(QWidget * parent, Qt::WFlags f):
  QDialog(parent, f),
  Ui::ImageLoaderDialog(),
  file_dialog_(new QFileDialog(this))
{
  Ui::ImageLoaderDialog::setupUi(this);
  
  // signals and slots connections
  connect(changeSourceFileBtn, SIGNAL(clicked()),
	  this, SLOT(changeSourceFile()));
  connect(okButton, SIGNAL(clicked()), this, SLOT(load()));
  connect(cancelButton, SIGNAL(clicked()), this, SLOT(cancel()));
  
  file_dialog_->setFileMode(QFileDialog::ExistingFile);
  QStringList filters;
  filters << tr("Mosaic or image (*.mosaic *.tif *.png *.jpg "
		"*.bmp *.gif *.xpm *.mha *.mhd *.nrrd)")
	  << tr("Anything (*.*)");
  file_dialog_->setFilters(filters);
  file_dialog_->setViewMode(QFileDialog::Detail);
  file_dialog_->setDirectory(QDir::current());
}

//----------------------------------------------------------------
// ImageLoaderDialog::~ImageLoaderDialog
// 
ImageLoaderDialog::~ImageLoaderDialog()
{}

//----------------------------------------------------------------
// ImageLoaderDialog::exec
// 
int
ImageLoaderDialog::exec(image_data_t & data)
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  sourceFileTxt->setText(QString::fromUtf8(data.name_));
  sync_ui();
  
  int r = QDialog::exec();
  if (r == QDialog::Accepted)
  {
    // accepted:
    QString filename = sourceFileTxt->text();
    unsigned int shrink_factor = changeShrinkFactorSpinBox->value();
    double pixel_spacing = changePixelSpacingSpinBox->value();
    
    // lookup the prefered internal format for the grayscale images:
    GLenum internal_format =
      prefsUseLuminanceTextures() ?
      GL_LUMINANCE6_ALPHA2 :
      GL_RGBA8;
    
    ImageDataTransaction<native_image_t>
      t(filename.toUtf8().constData(),
	shrink_factor,
	pixel_spacing,
	
	// by default, do not flip:
	false,
	
	// FIXME: this is wrong, it should be a parameter to the loader:
	OpenGL(0).max_texture_,
	internal_format);
    
    ImageProcessingDialog tmp(this);
    tmp.thread_.add_transaction(&t);
    
    r = tmp.exec();
    if (r == QDialog::Accepted && t.state() == the_transaction_t::DONE_E)
    {
      data = t.data_;
    }
  }
  
  return r;
}

//----------------------------------------------------------------
// ImageLoaderDialog::changeSourceFile
// 
void
ImageLoaderDialog::changeSourceFile()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  QString fn;
  
#if 0
  if (!file_dialog_->exec()) return;
  
  QStringList fn_list = file_dialog_->selectedFiles();
  if (!fn_list.empty())
  {
    fn = fn_list.back();
  }
#else
  fn =
    QFileDialog::getOpenFileName
    (this,	// parent
     QString(),	// caption
     QString(),	// dir
     tr("Mosaic or image (*.mosaic *.tif *.png *.jpg *.bmp "
	"*.gif *.xpm *.mha *.mhd *.nrrd);;"
	"Anything (*.*)"));
#endif
  
  if (!fn.isEmpty())
  {
    sourceFileTxt->setText(fn);
    
    // change current working directory:
    QDir::setCurrent(QFileInfo(fn).dir().absolutePath());
  }
  
  sync_ui();
}

//----------------------------------------------------------------
// ImageLoaderDialog::load
// 
void
ImageLoaderDialog::load()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  done(QDialog::Accepted);
}

//----------------------------------------------------------------
// ImageLoaderDialog::cancel
// 
void
ImageLoaderDialog::cancel()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  done(QDialog::Rejected);
}

//----------------------------------------------------------------
// ImageLoaderDialog::sync_ui
// 
void
ImageLoaderDialog::sync_ui()
{
  okButton->setEnabled(QFile::exists(sourceFileTxt->text()));
  bool enableSpacing =
    !sourceFileTxt->text().isEmpty() &&
    !sourceFileTxt->text().endsWith(QString::fromUtf8(".mosaic"));
  changePixelSpacingSpinBox->setEnabled(enableSpacing);
}

//----------------------------------------------------------------
// ImageLoaderDialog::closeEvent
// 
void
ImageLoaderDialog::closeEvent(QCloseEvent * e)
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  e->ignore();
  ImageLoaderDialog::cancel();
}
