// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : WarpImageTransaction.h
// Author       : Pavel A. Koshevoy
// Created      : Sun Nov 19 19:26:00 MDT 2006
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : transaction for warping an image

#ifndef WARP_IMAGE_TRANSACTION_H_
#define WARP_IMAGE_TRANSACTION_H_

// local includes:
#include "LoadMosaicTransactionITK.h"

// thelib includes:
#include <thread/the_thread_interface.hxx>
#include <itk/common.hxx>
#include <thread/the_terminator.hxx>


//----------------------------------------------------------------
// WarpImageTransaction
// 
template <typename TImage>
class WarpImageTransaction : public the_transaction_t
{
public:
  WarpImageTransaction(const image_data_t & image,
		       base_transform_t::ConstPointer transform,
		       const bool blab = true):
    filename_(image.name_),
    flipped_(image.flipped_),
    transform_(transform),
    blab_(blab)
  {
    // shortcut:
    const image_tile_generator_t & tile = image.tile_generator_;
    
    // collect tile bounding boxes expressed in the tile coordinate system:
    pnt2d_t tile_min = pnt2d(tile.bbox_.min_[0], tile.bbox_.min_[1]);
    pnt2d_t tile_max = pnt2d(tile.bbox_.max_[0], tile.bbox_.max_[1]);
    
    // calculate the warped image bounding box:
    ::calc_tile_mosaic_bbox(transform_,
			    tile_min,
			    tile_max,
			    mosaic_min_,
			    mosaic_max_);
    
    // warped image pixel spacing:
    spacing_[0] = tile.spacing_x_ / image.shrink_factor_;
    spacing_[1] = tile.spacing_y_ / image.shrink_factor_;
  }
  
  WarpImageTransaction(const the_text_t & filename,
		       const bool & flipped,
		       base_transform_t::ConstPointer transform,
		       const pnt2d_t & min,
		       const pnt2d_t & max,
		       const vec2d_t & spacing,
		       const bool blab = true):
    filename_(filename),
    flipped_(flipped),
    transform_(transform),
    mosaic_min_(min),
    mosaic_max_(max),
    spacing_(spacing),
    blab_(blab)
  {}
  
  // virtual:
  void execute(the_thread_interface_t * thread)
  {
    itk_terminator_t terminator("WarpImageTransaction::execute");
    
    // re-read in the data:
    if (blab_)
    {
      blab(thread, tr("re-reading the data at full resolution"));
    }
    
    typename TImage::Pointer image;
    if (filename_.match_tail(".mosaic"))
    {
      LoadMosaicTransactionITK<TImage> t(filename_, 1, flipped_, false);
      t.execute(thread);
      image = t.mosaic_;
    }
    else
    {
      image = ::std_tile<TImage>(filename_, 1, spacing_[0], false);
    }
    
    // assemble the mosaic:
    if (blab_)
    {
      blab(thread, tr("warping the image"));
    }
    
    // calculate the warped image bounding box:
    pnt2d_t tile_mosaic_min;
    pnt2d_t tile_mosaic_max;
    ::calc_tile_mosaic_bbox<TImage>(transform_,
				    image,
				    tile_mosaic_min,
				    tile_mosaic_max);
    
    // mosaic width and height in world coordinate system:
    double mw = mosaic_max_[0] - mosaic_min_[0];
    double mh = mosaic_max_[1] - mosaic_min_[1];
    
    // corresponding discrete image dimensions:
    unsigned int iw = int(ceil(mw / spacing_[0]));
    unsigned int ih = int(ceil(mh / spacing_[1]));
    typename TImage::RegionType::SizeType sz;
    sz[0] = iw;
    sz[1] = ih;
    
    // allocate the image:
    warped_ = ::make_image<TImage>(spacing_, sz);
    warped_->SetOrigin(mosaic_min_);
    
    // local temporaries:
    pnt2d_t warped_pt;
    pnt2d_t tile_pt;
    
    // connect an interpolator to the image:
    typedef typename itk::LinearInterpolateImageFunction<TImage, double>
      interpolator_t;
    typename interpolator_t::Pointer interpolator = interpolator_t::New();
    interpolator->SetInputImage(image);
    
    // these are used for clamping the luminance value:
    typedef typename TImage::PixelType pixel_t;
    bool integer_pixel = std::numeric_limits<pixel_t>::is_integer;
    double pixel_max = double(std::numeric_limits<pixel_t>::max());
    double pixel_min = integer_pixel ?
      double(std::numeric_limits<pixel_t>::min()) : -pixel_max;
    
    // assemble the warped image:
    typedef typename itk::ImageRegionIteratorWithIndex<TImage> itex_t;
    itex_t itex(warped_, warped_->GetLargestPossibleRegion());
    typename TImage::IndexType index;
    for (itex.GoToBegin(); !itex.IsAtEnd(); ++itex)
    {
      // make sure there hasn't been an interrupt:
      terminator.terminate_on_request();
      
      index = itex.GetIndex();
      warped_->TransformIndexToPhysicalPoint(index, warped_pt);
      
      double luminance = 0.0;
      if (::inside_bbox(tile_mosaic_min, tile_mosaic_max, warped_pt))
      {
	// transform the warped point into the tile coordinate system:
	tile_pt = transform_->TransformPoint(warped_pt);
	
	// see whether the point falls inside the tile:
	if (interpolator->IsInsideBuffer(tile_pt))
	{
	  // get luminance from the tile:
	  luminance = interpolator->Evaluate(tile_pt);
	}
      }
      
      luminance = std::min(pixel_max, std::max(pixel_min, luminance));
      itex.Set((unsigned char)(luminance));
    }
  }
  
  // trasaction parameters:
  the_text_t filename_;
  bool flipped_;
  base_transform_t::ConstPointer transform_;
  pnt2d_t mosaic_min_;
  pnt2d_t mosaic_max_;
  vec2d_t spacing_;
  bool blab_;
  
  // transaction output:
  typename TImage::Pointer warped_;
};


#endif // WARP_IMAGE_TRANSACTION_H_
