// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : mosaic_cache.hxx
// Author       : Pavel A. Koshevoy
// Created      : Wed Oct 18 10:34:00 MDT 2006
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : mosaic cache -- supports on-demand mosaic image regeneration.

#ifndef MOSAIC_CACHE_HXX_
#define MOSAIC_CACHE_HXX_

// the includes:
#include <thread/the_mutex_interface.hxx>
#include <opengl/image_tile_dl_elem.hxx>
#include <image/image_tile_generator.hxx>
#include <utils/the_utils.hxx>
#include <math/the_aa_bbox.hxx>
#include <itk/common.hxx>

// Boost includes:
#include <boost/shared_ptr.hpp>

// system includes:
#include <list>
#include <memory>


//----------------------------------------------------------------
// mosaic_cache_shared_data_t
//
class mosaic_cache_shared_data_t
{
  friend class mosaic_cache_t;
  
public:
  mosaic_cache_shared_data_t();
  mosaic_cache_shared_data_t(const base_transform_t::ConstPointer & transform,
			     const image_tile_generator_t & mosaic,
			     const the_aa_bbox_t & mosaic_bbox,
			     const unsigned int & bytes_per_pixel);
  ~mosaic_cache_shared_data_t();
  
  // update the shared image tiles data:
  void make_tiles(const GLenum & data_type,
		  const GLenum & format_internal,
		  const GLenum & format,
		  const unsigned int max_texture);
  
  // mosaic bounding box accessor:
  inline const the_aa_bbox_t & mosaic_bbox() const
  { return mosaic_bbox_; }
  
  // check whether the requested region is available and valid:
  bool is_ready(const the_aa_bbox_t & requested_region,
		const base_transform_t::ConstPointer & transform) const;
  
  // not thread safe:
  inline bool under_construction() const
  { return work_in_progress_ || !work_.empty(); }
  
protected:
  // check whether the mosaic is the same:
  bool same_mosaic(const base_transform_t::ConstPointer & transform) const;
  
  // check whether the requested region is readily available:
  bool region_ready(const the_aa_bbox_t & requested_region) const;
  
  // Find all ready regions that overlap the requested region
  // by at least as much as the given overlap ratio. The default
  // overlap ratio of 0 will return all regions that overlap
  // the requested region, so long as the overlap area is non-zero.
  // The requested region must be clamped to the mosaic bounding box.
  void find_overlapping_regions(std::list<the_aa_bbox_t> & found,
				const the_aa_bbox_t & requested,
				const double & overlap_ratio = 0.0) const;
  
  // partition the requested region into several sub-regions:
  void get_work_regions(const the_aa_bbox_t & requested_region,
			std::list<the_aa_bbox_t> & work,
			const double & target_region_size = 0) const;
  
  // generate a mosaic for the requested region, return true
  // if the tile have to be regenerated as well:
  void update(const the_aa_bbox_t & requested_region,
	      const image_tile_generator_t & fixed,
	      const image_tile_generator_t & moving);
  
  // NOTE: The function will return true if it was interrupted by a terminator.
  // Update the requested region of the mosaic (assumed already
  // clamped to the mosaic bounding box), and calculate the bounding box
  // of the processed region and the corresponding texture image region;
  // both regions will be updated incrementally, scanline by scanline,
  // so the results will be valid even when the function is interrupted
  // by a terminator:
  bool process(const the_aa_bbox_t & clamped_region,
	       const image_tile_generator_t & fixed,
	       const image_tile_generator_t & moving,
	       the_aa_bbox_t & processed,
	       image_tile_t::quad_t & upload);
  
  // mosaic transform:
  base_transform_t::ConstPointer transform_;
  
  // the cached mosaic:
  image_tile_generator_t mosaic_;
  
  // mosaic bounding box:
  the_aa_bbox_t mosaic_bbox_;
  
  // most recently regenerated regions:
  std::list<the_aa_bbox_t> ready_;
  
  // the portion of the mosaic that every thread should be working on:
  the_aa_bbox_t work_region_;
  std::list<the_aa_bbox_t> work_;
  unsigned int work_in_progress_;
  
public:
  // texture regions that have yet to be upload to the graphics card:
  std::list<image_tile_t::quad_t> upload_;
  
  // image tiles and buffer pointer:
  image_tile_generator_t data_;
  
  // mutex controlling access to this cache:
  mutable the_mutex_interface_t * mutex_;
};


//----------------------------------------------------------------
// mosaic_cache_t
// 
class mosaic_cache_t
{
public:
  mosaic_cache_t(const unsigned int & id,
		 const double & spacing_x,
		 const double & spacing_y);
  
  // generate a mosaic for the requested region, return true
  // if the tile have to be regenerated as well:
  void update(const the_aa_bbox_t & requested_region,
	      const image_tile_generator_t & fixed,
	      const image_tile_generator_t & moving,
	      base_transform_t::ConstPointer transform,
	      const bool & mosaic_compressed);
  
  // helper:
  inline the_mutex_interface_t * mutex() const
  { return shared_->mutex_; }
  
  // cache id:
  unsigned int id_;
  
  // pixel spacing for this mosaic:
  double spacing_x_;
  double spacing_y_;
  
  // the shared component of the display element:
  boost::shared_ptr<mosaic_cache_shared_data_t> shared_;
  
  // display element:
  std::auto_ptr<the_dl_elem_t> dl_elem_;
};


#endif // MOSAIC_CACHE_HXX_
