// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : the_main.cxx
// Author       : Pavel A. Koshevoy
// Created      : 2006/07/11 13:20
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : A landmark correspondence based interactive
//                slice to slice image registration tool.

// configure default trail recording behaviour based on build type:
#ifndef NDEBUG
static const bool RECORD_BY_DEFAULT = true;
#else
static const bool RECORD_BY_DEFAULT = false;
#endif

// system includes:
#include <iostream>

// the includes:
#include <Qt/the_qt_mutex.hxx>
#include <Qt/the_qt_thread.hxx>
#include <Qt/the_qt_trail.hxx>
#include <Qt/the_qt_view.hxx>
#include <Qt/the_qt_desktop_metrics.hxx>
#include <opengl/OpenGLCapabilities.h>
#include <opengl/glsl.hxx>
#include <image/image_tile_generator.hxx>
#include <utils/the_utils.hxx>

// Qt includes:
#include <QTimer>

// Cg includes:
#include <Cg/cgGL.h>

// local includes:
#include "the_main_window.hxx"

// include fragment shaders:
#include "shader_la.hxx"
#include "shader_la2.hxx"
#include "shader_mosaic.hxx"
#include "shader_mosaic2.hxx"

// namespace accessors:
using std::cerr;
using std::cout;
using std::endl;


// global fragment shader handles:
CGcontext cg_context;
CGprofile cg_profile;
CGprogram CG_LA_FRAGMENT_PROGRAM;
CGprogram CG_MOSAIC_FRAGMENT_PROGRAM;

CGprogram * CG_LA_FRAGMENT_PROGRAM_PTR = NULL;
CGprogram * CG_MOSAIC_FRAGMENT_PROGRAM_PTR = NULL;

bool CG_LA_FRAGMENT_PROGRAM_OK = false;
bool CG_MOSAIC_FRAGMENT_PROGRAM_OK = false;

//----------------------------------------------------------------
// the_application_t
// 
class the_application_t : public the_qt_trail_t
{
public:
  the_application_t(int & argc, char ** argv, bool record_by_default = false):
    the_qt_trail_t(argc, argv, record_by_default)
  {}
  
protected:
  bool event(QEvent * e)
  {
    if (e->type() != QEvent::FileOpen)
    {
      return the_qt_trail_t::event(e);
    }
    
    // handle the apple event to open a document:
    QString fn_load = static_cast<QFileOpenEvent *>(e)->file();
    the_main_window_t * main_window =
      dynamic_cast<the_main_window_t *>(the_document_ui_t::doc_ui());
    main_window->file_load(fn_load);
    return true;
  }
};

//----------------------------------------------------------------
// cg_error
// 
CGerror cg_error = CG_NO_ERROR;

//----------------------------------------------------------------
// cg_error_callback
//
static void cg_error_callback()
{
  CGerror last_error = cgGetError();
  if (!last_error) return;
  cg_error = last_error;
  
  cerr << cgGetErrorString(last_error) << endl
       << cgGetLastListing(cg_context) << endl
       << cg_error << endl
       << endl;
}

//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL, int exit_code = 1)
{
  cerr << "USAGE: " << argv[0]
       << " [-ask] [-replay in.txt] [-record out.txt]"
       << " [-wait seconds] [file_to_open]" << endl
       << endl
       << "\t-ask:    confirm every major event prior to playback" << endl
       << "\t-replay: replay a given event trail file" << endl
       << "\t-record: record events to a specific file" << endl
       << "\t-wait:   number of seconds to wait for a milestone" << endl
       << endl;
  
  if (RECORD_BY_DEFAULT)
  {
    cerr << "NOTE:\tIf -record is not specified, the event trail file\n"
	 << "\tname is picked automatically as .dont_record.txt"<<endl;
  }
  
  if (message != NULL)
  {
    cerr << "NOTE:\t" << message << endl;
  }
  
  exit(exit_code);
}

//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  // this is so that the printouts look better:
  cout.precision(8);
  cerr.precision(8);
  cout.setf(ios::scientific);
  cerr.setf(ios::scientific);
  
  // necessary on some platforms for resources stored in a static library:
  Q_INIT_RESOURCE(resources);
  
  // parse command line for input/output trail file options:
  the_application_t A(argc, argv, RECORD_BY_DEFAULT);
  
  // setup the desktop metrics provider:
  the_desktop_metrics(new the_qt_desktop_metrics_t());
  
  // parse the rest of the command line:
  QString fn_load;
  
  for (int i = 1; i < argc; i++)
  {
    if (strcmp(argv[i], "--help") == 0 ||
	strcmp(argv[i], "-help") == 0 ||
	strcmp(argv[i], "-h") == 0)
    {
      usage(argv, argv[i], 0);
    }
    else
    {
      QFileInfo fi(argv[i]);
      if (!fi.isFile())
      {
	usage(argv, the_text_t("unrecogrinzed parameter: ") + argv[i], 1);
      }
      else
      {
	fn_load = QString::fromUtf8(argv[i]);
      }
    }
  }
  
  // setup thread storage for the main thread:
  set_the_thread_storage_provider(the_qt_thread_t::thread_storage);
  the_qt_thread_t MAIN_THREAD_DUMMY;
  MAIN_THREAD_DUMMY.set_stopped(false);
  
  // setup thread and mutex interface creators:
  the_mutex_interface_t::set_creator(the_qt_mutex_t::create);
  the_thread_interface_t::set_creator(the_qt_thread_t::create);
  
  // setup the UI:
  the_main_window_t main_window;
  main_window.show();
  main_window.setAcceptDrops(true);
  main_window.file_new();
  
  // load user preferences:
  main_window.preferences().loadSettings();
  prefsSetFragmentShaderAvailable(false);
  
  // initialize the opengl:
  OpenGLCapabilities & gl0 = OpenGL(0, main_window.view(0));
  gl0.max_texture_ = OpenGLCapabilities::maxTextureSize(GL_LUMINANCE,
							GL_LUMINANCE,
							GL_UNSIGNED_BYTE);
  // FIXME:
  // gl0.max_texture_ = std::min(GLuint(512), gl0.max_texture_);
  
  OpenGLCapabilities & gl1 = OpenGL(1, main_window.view(1));
  gl1.max_texture_ = OpenGLCapabilities::maxTextureSize(GL_LUMINANCE,
							GL_LUMINANCE,
							GL_UNSIGNED_BYTE);
  // FIXME:
  // gl1.max_texture_ = std::min(GLuint(512), gl1.max_texture_);
  
  OpenGLCapabilities & gl2 = OpenGL(2, main_window.view(2));
  gl2.max_texture_ =
    OpenGLCapabilities::maxTextureSize(GL_RGBA8,
				       GL_BGRA_EXT,
				       GL_UNSIGNED_INT_8_8_8_8_REV);
  // FIXME:
  // gl2.max_texture_ = std::min(GLuint(512), gl2.max_texture_);
  
  // FIXME:
#if 0
  gl0.max_texture_ = 64;
  gl1.max_texture_ = 64;
  gl2.max_texture_ = 64;
#endif
  
  // setup the shaders:
  main_window.view(0)->gl_make_current();
  if (glsl_init())
  {
#ifdef USE_CG
    GLint num_texture_units = 0;
    glGetIntegerv(GL_MAX_TEXTURE_IMAGE_UNITS_ARB, &num_texture_units);
    
    if (num_texture_units > 0)
    {
      cgSetErrorCallback(cg_error_callback);
      cg_context = cgCreateContext();
      
      cg_profile = cgGLGetLatestProfile(CG_GL_FRAGMENT);
      if (cg_profile != CG_PROFILE_UNKNOWN)
      {
	cgGLSetOptimalOptions(cg_profile);
	FIXME_OPENGL("cgGLSetOptimalOptions");
	
	static const char * cgc_args[] = { "-DDISABLE_AA", NULL };
	const char ** args[] = { NULL, cgc_args };
	const char * shaders[][2] = {
	  // FIXME:
	  { shader_la_src, shader_mosaic_src },		// advanced
	  { shader_la2_src, shader_mosaic2_src }	// simple
	};
	
	encode_la = NULL;
	decode_la = NULL;
	for (unsigned int i = 0; i < 2; i++)
	{
	  for (unsigned int j = 0; j < 2; j++)
	  {
	    cg_error = CG_NO_ERROR;
	    
	    // setup the fixed/moving image shader program:
	    CG_LA_FRAGMENT_PROGRAM =
	      cgCreateProgram(cg_context,
			      CG_SOURCE,
			      shaders[i][0],
			      cg_profile,
			      "main",
			      args[j]);
	    if (cg_error != CG_NO_ERROR) continue;
	    
	    // setup the mosaic image shader program:
	    CG_MOSAIC_FRAGMENT_PROGRAM =
	      cgCreateProgram(cg_context,
			      CG_SOURCE,
			      shaders[i][1],
			      cg_profile,
			      "main",
			      args[j]);
	    if (cg_error != CG_NO_ERROR) continue;
	    break;
	  }
	  
	  if (cg_error != CG_NO_ERROR) continue;
	  
	  // both fragment programs compiled successfully, setup
	  // the matching pixel encoders:
	  if (i == 0)
	  {
	    encode_la = encode_la_v1;
	    decode_la = decode_la_v1;
	  }
	  else
	  {
	    encode_la = encode_la_v2;
	    decode_la = decode_la_v2;
	  }
	  break;
	}
	
	if (cg_error == CG_NO_ERROR)
	{
	  CG_LA_FRAGMENT_PROGRAM_OK = true;
	  CG_MOSAIC_FRAGMENT_PROGRAM_OK = true;
	  
	  prefsSetFragmentShaderAvailable(true);
	  if (prefsUseFragmentShader())
	  {
	    CG_LA_FRAGMENT_PROGRAM_PTR = &CG_LA_FRAGMENT_PROGRAM;
	    CG_MOSAIC_FRAGMENT_PROGRAM_PTR = &CG_MOSAIC_FRAGMENT_PROGRAM;
	  }
	}
      }
    }
#endif
  }
  
  cerr << "la shader ....... " << CG_LA_FRAGMENT_PROGRAM_OK << endl
       << "mosaic shader ... " << CG_MOSAIC_FRAGMENT_PROGRAM_OK << endl;
  
  // load the parameter:
  if (fn_load.size() != 0)
  {
    // cerr << "FIXME: load " << qPrintable(fn_load) << endl;
    main_window.file_load(fn_load);
  }
  
  // give the window manager a little time to set up the window decorations,
  // and then start the trail replay:
  QTimer::singleShot(178, &A, SLOT(replay()));
  
  // start the event loop:
  int result = A.exec();
  return result;
}
