// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : the_stos_proc.cxx
// Author       : Pavel A. Koshevoy
// Created      : 2006/07/11 13:20
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Landmark based slice to slice registration procedure.

// local includes:
#include "the_stos_proc.hxx"
#include "ImageProcessingDialog.h"
#include "the_main_window.hxx"
#include "PreferencesDialog.h"
#include "ImageDataTransaction.h"
#include "mosaic_cache.hxx"

// the includes:
#include <doc/the_graph.hxx>
#include <opengl/the_view.hxx>
#include <ui/the_document_ui.hxx>
#include <ui/the_trail.hxx>
#include <eh/the_input_device_eh.hxx>
#include <itk/common.hxx>
#include <itk/grid_common.hxx>
#include <utils/the_indentation.hxx>


//----------------------------------------------------------------
// the_stos_proc_t::the_stos_proc_t
// 
the_stos_proc_t::the_stos_proc_t():
  the_procedure_t(),
  fi_(new image_data_t()),
  mi_(new image_data_t()),
  mosaic_(new mosaic_cache_pyramid_t()),
  mosaic_dl_elem_(new std::vector<boost::shared_ptr<mosaic_dl_elem_t> >()),
  mosaic_dl_ready_(~0),
  dl_landmarks_fi_(new the_disp_list_t()),
  dl_landmarks_mi_(new the_disp_list_t()),
  dl_landmarks_mosaic_(new the_disp_list_t()),
  dl_(new the_disp_list_t()),
  fi_id_(std::numeric_limits<unsigned int>::max()),
  mi_id_(std::numeric_limits<unsigned int>::max()),
  xf_id_(std::numeric_limits<unsigned int>::max()),
  mz_id_(std::numeric_limits<unsigned int>::max()),
  selection_(std::numeric_limits<unsigned int>::max())
{}

//----------------------------------------------------------------
// the_stos_proc_t::the_stos_proc_t
// 
the_stos_proc_t::the_stos_proc_t(const the_stos_proc_t & proc):
  the_procedure_t(proc),
  fi_(proc.fi_),
  mi_(proc.mi_),
  mosaic_(proc.mosaic_),
  mosaic_dl_elem_(proc.mosaic_dl_elem_),
  mosaic_dl_ready_(proc.mosaic_dl_ready_),
  dl_landmarks_fi_(proc.dl_landmarks_fi_),
  dl_landmarks_mi_(proc.dl_landmarks_mi_),
  dl_landmarks_mosaic_(proc.dl_landmarks_mosaic_),
  dl_(proc.dl_),
  fi_id_(proc.fi_id_),
  mi_id_(proc.mi_id_),
  xf_id_(proc.xf_id_),
  mz_id_(proc.mz_id_),
  selection_(proc.selection_)
{}

//----------------------------------------------------------------
// the_stos_proc_t::~the_stos_proc_t
// 
the_stos_proc_t::~the_stos_proc_t()
{}

//----------------------------------------------------------------
// the_stos_proc_t::added_to_the_registry
// 
void
the_stos_proc_t::added_to_the_registry(the_registry_t * registry,
				       const unsigned int & id)
{
  the_procedure_t::added_to_the_registry(registry, id);
  
  the_image_t * fi_prim = new the_image_t();
  registry->add(fi_prim);
  fi_id_ = fi_prim->id();
  establish_supporter_dependent(registry, id, fi_id_);
  
  the_image_t * mi_prim = new the_image_t();
  registry->add(mi_prim);
  mi_id_ = mi_prim->id();
  establish_supporter_dependent(registry, id, mi_id_);
  
  the_xform_t * xform = new the_xform_t();
  registry->add(xform);
  xf_id_ = xform->id();
  establish_supporter_dependent(registry, id, xf_id_);
  establish_supporter_dependent(registry, fi_id_, xf_id_);
  establish_supporter_dependent(registry, mi_id_, xf_id_);
  
  the_mosaic_t * mosaic = new the_mosaic_t();
  registry->add(mosaic);
  mz_id_ = mosaic->id();
  establish_supporter_dependent(registry, id, mz_id_);
  establish_supporter_dependent(registry, xf_id_, mz_id_);
}


//----------------------------------------------------------------
// the_stos_proc_t::regenerate
// 
bool
the_stos_proc_t::regenerate()
{
  the_document_ui_t * doc_ui = the_document_ui_t::doc_ui();
  
  // clear the view dependent primitive lists:
  the_view_t * view_fixed = doc_ui->view(0);
  the_view_t * view_moving = doc_ui->view(1);
  the_view_t * view_mosaic = doc_ui->view(2);
  
  view_fixed->clear_primitives();
  view_moving->clear_primitives();
  view_mosaic->clear_primitives();
  
  // rebuild the display lists:
  dl_landmarks_fi_->clear();
  dl_landmarks_mi_->clear();
  dl_landmarks_mosaic_->clear();
  dl_->clear();
  
  // regenerate the dependents that need to be regenerated:
  the_registry_t * r = registry();
  the_graph_t graph;
  graph.set_roots(r, direct_dependents());
  regenerated_ = graph.regenerate();
  
  // get the inverse transform:
  the_xform_t * xform = xform_prim();
  stos_transform_t::Pointer forward = xform->forward_;
  stos_transform_t::Pointer inverse = xform->inverse_;
  bool display_landmarks = prefsDisplayLandmarks();
  
  // add the points to the display lists, update the view dependent
  // selectable primitive lists:
  for (std::list<unsigned int>::const_iterator
	 i = direct_dependents().begin(); i != direct_dependents().end(); ++i)
  {
    the_primitive_t * primitive = r->elem(*i);
    if (!primitive->regenerated()) continue;
    
    point_pair_t * pp = dynamic_cast<point_pair_t *>(primitive);
    if (pp != NULL)
    {
      the_point_t * a = pp->ref_[0]->references<the_point_t>(r);
      the_point_t * b = pp->ref_[1]->references<the_point_t>(r);
      
      view_fixed->add_primitive(a);
      view_moving->add_primitive(b);
      
      the_color_t color = a->current_state() == THE_SELECTED_STATE_E ?
	the_color_t::RED :
	the_color_t(0.2, 0.4, 0.6, 1);
      
      dl_landmarks_fi_->push_back(new the_masked_symbol_dl_elem_t
				  (a->value(),
				   color,
				   the_color_t::WHITE,
				   THE_POINT_SYMBOLS,
				   THE_SMALL_FILLED_CIRCLE_SYMBOL_E));
      
      dl_landmarks_mi_->push_back(new the_masked_symbol_dl_elem_t
				  (b->value(),
				   color,
				   the_color_t::WHITE,
				   THE_POINT_SYMBOLS,
				   THE_SMALL_FILLED_CIRCLE_SYMBOL_E));
      
      if (display_landmarks)
      {
	view_mosaic->add_primitive(a);
	
	p3x1_t pt_a = a->value();
	p3x1_t pt_b = b->value();
	
	pnt2d_t tmp;
	if (find_inverse(forward.GetPointer(),
			 inverse.GetPointer(),
			 pnt2d(pt_b.x(), pt_b.y()),
			 tmp))
	{
	  pt_b[0] = tmp[0];
	  pt_b[1] = tmp[1];
	  dl_landmarks_mosaic_->push_back(new the_masked_symbol_dl_elem_t
					  (pt_b,
					   color,
					   the_color_t::WHITE,
					   THE_POINT_SYMBOLS,
					   THE_CIRCLE_SYMBOL_E));
	}
	
	dl_landmarks_mosaic_->push_back(new the_masked_symbol_dl_elem_t
					(pt_a,
					 color,
					 the_color_t::WHITE,
					 THE_POINT_SYMBOLS,
					 THE_SMALL_FILLED_CIRCLE_SYMBOL_E));
      }
    }
  }
  
  if (prefsUseOpenGLForMosaicking())
  {
    dl_->push_back(new the_stos_mosaic_dl_elem_t(this));
  }
  
  return regenerated_;
}

//----------------------------------------------------------------
// the_stos_proc_t::draw
// 
void
the_stos_proc_t::draw(const the_view_t & view) const
{
  view.view_mgr().reset_opengl_viewing();
  view.view_mgr().setup_opengl_3d_viewing();
  
  glPushAttrib(GL_ENABLE_BIT);
  {
    glDisable(GL_DEPTH_TEST);
    glEnable(GL_BLEND);
    
    the_document_ui_t * doc_ui = the_document_ui_t::doc_ui();
    if (&view == doc_ui->view(0))
    {
      // fixed image view:
      if (fi_prim()->regenerated())
      {
	fi_prim()->dl_elem_->draw();
      }
      
      dl_landmarks_fi_->execute();
    }
    else if (&view == doc_ui->view(1))
    {
      // moving image view:
      if (mi_prim()->regenerated())
      {
	mi_prim()->dl_elem_->draw();
      }
      
      dl_landmarks_mi_->execute();
    }
    else if (&view == doc_ui->view(2))
    {
      // mosaic image view:
      if (!prefsUseOpenGLForMosaicking())
      {
	if (mosaic_dl_ready_ < mosaic_dl_elem_->size())
	{
	  mosaic_dl_elem(mosaic_dl_ready_)->draw();
	}
      }
      else
      {
	dl_->draw();
      }
      
      dl_landmarks_mosaic_->execute();
    }
  }
  glPopAttrib();
}

//----------------------------------------------------------------
// get_bbox
// 
void
get_bbox(const the_bbox_t & bbox,
	 pnt2d_t & min,
	 pnt2d_t & max)
{
  p3x1_t wcs_min = bbox.wcs_min();
  p3x1_t wcs_max = bbox.wcs_max();
  min[0] = wcs_min[0];
  min[1] = wcs_min[1];
  max[0] = wcs_max[0];
  max[1] = wcs_max[1];
}

//----------------------------------------------------------------
// the_stos_proc_t::calc_bbox
// 
void
the_stos_proc_t::calc_bbox(const the_view_t & view,
			   the_bbox_t & bbox) const
{
  the_document_ui_t * doc_ui = the_document_ui_t::doc_ui();
  if (&view == doc_ui->view(0))
  {
    // fixed image view:
    bbox += dl_landmarks_fi_->bbox();
    
    if (fi_ready())
    {
      bbox += fi_->bbox();
    }
  }
  else if (&view == doc_ui->view(1))
  {
    // moving image view:
    bbox += dl_landmarks_mi_->bbox();
    
    if (mi_ready())
    {
      bbox += mi_->bbox();
    }
  }
  else if (&view == doc_ui->view(2))
  {
    // mosaic image view:
    bbox += dl_landmarks_mosaic_->bbox();
    
    if (fi_ready() && mi_ready() && xform_prim()->regenerated())
    {
      // calculate actual mosaic bounding box:
      
      std::vector<pnt2d_t> bbox_min(2);
      std::vector<pnt2d_t> bbox_max(2);
      get_bbox(fi().bbox(), bbox_min[0], bbox_max[0]);
      get_bbox(mi().bbox(), bbox_min[1], bbox_max[1]);
      
      std::vector<base_transform_t::ConstPointer> xform(2);
      xform[0] = identity_transform_t::New();
      xform[1] = xform_prim()->forward_.GetPointer();
      
      std::vector<pnt2d_t> mosaic_min(2);
      std::vector<pnt2d_t> mosaic_max(2);
      calc_mosaic_bboxes(xform, bbox_min, bbox_max, mosaic_min, mosaic_max);
      
      pnt2d_t min;
      pnt2d_t max;
      calc_mosaic_bbox(mosaic_min, mosaic_max, min, max);
      
      bbox << p3x1_t(min[0], min[1], 0)
	   << p3x1_t(max[0], max[1], 0);
    }
  }
}

//----------------------------------------------------------------
// the_stos_proc_t::swap_images
// 
void
the_stos_proc_t::swap_images()
{
  std::swap(fi_, mi_);
  
  // iterate through the point pairs and swap the points:
  the_registry_t * r = registry();
  
  for (std::list<unsigned int>::const_iterator
	 i = direct_dependents().begin(); i != direct_dependents().end(); ++i)
  {
    point_pair_t * pp = r->elem<point_pair_t>(*i);
    if (pp == NULL) continue;
    
    std::swap(pp->ref_[0], pp->ref_[1]);
  }
  
  fi_prim()->request_regeneration();
  mi_prim()->request_regeneration();
}

//----------------------------------------------------------------
// the_stos_proc_t::calc_transforms
// 
void
the_stos_proc_t::calc_transforms(stos_transform_t::Pointer & forward,
				 stos_transform_t::Pointer & inverse) const
{
  the_dynamic_array_t<p3x1_t> a;
  the_dynamic_array_t<p3x1_t> b;
  
  stos_transform_t::Pointer ft = stos_transform_t::New();
  stos_transform_t::Pointer it = stos_transform_t::New();
  
  // collect all point pairs:
  the_registry_t * r = registry();
  for (std::list<unsigned int>::const_iterator
	 i = direct_dependents().begin(); i != direct_dependents().end(); ++i)
  {
    point_pair_t * pp = r->elem<point_pair_t>(*i);
    if (pp == NULL) continue;
    if (!pp->regenerated()) continue;
    
    the_point_t * pt_a = pp->ref_[0]->references<the_point_t>(r);
    the_point_t * pt_b = pp->ref_[1]->references<the_point_t>(r);
    a.append(pt_a->value());
    b.append(pt_b->value());
  }
  
  // solve for transform parameters:
  if (a.size() == 0 || !images_are_ready())
  {
    // no point pairs defined -- assume identity transform:
    forward = ft;
    inverse = it;
    return;
  }
  
  std::vector<pnt2d_t> tile_min(2);
  std::vector<pnt2d_t> tile_max(2);
  get_bbox(fi().bbox(), tile_min[0], tile_max[0]);
  get_bbox(mi().bbox(), tile_min[1], tile_max[1]);

  if (a.size() == 1)
  {
    // add 2 more point pairs with the same translation vector:
    const p3x1_t & a0 = a[0];
    const p3x1_t & b0 = b[0];
    
    a.append(a0 + v3x1_t(1, 0, 0));
    b.append(b0 + v3x1_t(1, 0, 0));
    
    a.append(a0 + v3x1_t(0, 1, 0));
    b.append(b0 + v3x1_t(0, 1, 0));
  }
  else if (a.size() == 2)
  {
    // add another point pair with the same rotation and scale:
    const v3x1_t z(0, 0, 1);
    
    const p3x1_t & a0 = a[0];
    const p3x1_t & a1 = a[1];
    const v3x1_t va = a1 - a0;
    const v3x1_t ua = ((!va) % z) * ~va;
    a.append(a0 + ua);
    
    const p3x1_t & b0 = b[0];
    const p3x1_t & b1 = b[1];
    const v3x1_t vb = b1 - b0;
    const v3x1_t ub = ((!vb) % z) * ~vb;
    b.append(b0 + ub);
  }
  
  std::vector<pnt2d_t> uv(a.size());
  std::vector<pnt2d_t> xy(b.size());
  for (unsigned int i = 0; i < a.size(); i++)
  {
    // mosaic space:
    uv[i][0] = a[i][0];
    uv[i][1] = a[i][1];
    
    // tile space:
    xy[i][0] = b[i][0];
    xy[i][1] = b[i][1];
  }
  
  // first, solve for the low-order coefficients:
#ifdef USE_LEGENDRE
  ft = setup_transform<stos_transform_t>(tile_min[1], tile_max[1]);
  it = setup_transform<stos_transform_t>(tile_min[0], tile_max[0]);
  ft->solve_for_parameters(0, 2, uv, xy);
  it->solve_for_parameters(0, 2, xy, uv);
  
  // unsigned int o2 = stos_transform_t::count_coefficients(0, 2);
  unsigned int o3 = stos_transform_t::count_coefficients(0, 3);
  unsigned int o4 = stos_transform_t::count_coefficients(0, 4);
  /*
  cerr << "point pairs: " << a.size() << endl
       << "2nd order coefficients: " << o2 << endl
       << "3rd order coefficients: " << o3 << endl
       << "4th order coefficients: " << o4 << endl;
  */
  if (a.size() >= o4)
  {
    // solve for the 4th order coefficients:
    ft->solve_for_parameters(2, 2, uv, xy);
    it->solve_for_parameters(2, 2, xy, uv);
  }
  else if (a.size() >= o3)
  {
    // solve for the 3rd order coefficients:
    ft->solve_for_parameters(2, 1, uv, xy);
    it->solve_for_parameters(2, 1, xy, uv);
  }
  
#else
  ft->setup(tile_min[1], tile_max[1], uv.size(), &(uv[0]), &(xy[0]));
  it->setup(tile_min[0], tile_max[0], xy.size(), &(xy[0]), &(uv[0]));
#endif
  
  forward = ft;
  inverse = it;
}

//----------------------------------------------------------------
// regenerate_mosaic
// 
bool
the_stos_proc_t::regenerate_mosaic()
{
  // generate a pyramid of mosaics:
  if (!images_are_ready()) return false;
  
  if (prefsUseOpenGLForMosaicking())
  {
    return true;
  }
  
  // FIXME: we may avoid regenerating the mosaic during a drag:
  // if (!THE_MOUSE.verify_no_pressed_btns()) return false;
  
  // regenerate the visible portion of the mosaic:
  the_document_ui_t * doc_ui = the_document_ui_t::doc_ui();
  the_view_t * view_mosaic = doc_ui->view(2);
  view_mosaic->eh_stack().view_cb(view_mosaic);
  
  return true;
}

//----------------------------------------------------------------
// the_stos_proc_t::mosaic_ready
// 
void
the_stos_proc_t::mosaic_ready(const unsigned int & level)
{
  // get the display element:
  mosaic_dl_elem_t * e = NULL;
  {
    the_lock_t<the_mutex_interface_t> lock(mosaic_->cache_[level]->mutex());
    e = (mosaic_dl_elem_t *)(mosaic_->cache_[level]->dl_elem_.release());
  }
  
  // cerr << "FIXME: mosaic ready: " << level << ' ' << e << endl;
  if (e != NULL)
  {
    // make the mosaic view context current right here:
    the_document_ui_t * doc_ui = the_document_ui_t::doc_ui();
    the_view_t * view_mosaic = doc_ui->view(2);
    view_mosaic->gl_make_current();
    mosaic_dl_elem(level) = boost::shared_ptr<mosaic_dl_elem_t>(e);
    view_mosaic->gl_done_current();
  }
  
  mosaic_dl_elem_t * dl_elem = mosaic_dl_elem(level).get();
  if (dl_elem == NULL) return;
  
#if 1
  if (level > mosaic_dl_ready_)
  {
    // higher pyramid level means higher pixel granularity -- lower precision,
    // therefore the current ready level should be examined to verify whether
    // is is actually necessary to switch to the newly delivered level:
    
    the_document_ui_t * doc_ui = the_document_ui_t::doc_ui();
    the_view_t * view_mosaic = doc_ui->view(2);
    
    const the_view_mgr_t & view_mgr = view_mosaic->view_mgr();
    the_view_volume_t vol;
    view_mgr.setup_view_volume(vol);
    
    p3x1_t view_min;
    p3x1_t view_max;
    vol.scs_to_wcs(p2x1_t(0, 0), view_min);
    vol.scs_to_wcs(p2x1_t(1, 1), view_max);
    view_min[2] = 0;
    view_max[2] = 0;
    
    the_aa_bbox_t requested_region;
    requested_region << view_min << view_max;
    
    stos_transform_t::Pointer & forward = xform_prim()->forward_;
    mosaic_cache_t * cache = mosaic_->cache_[mosaic_dl_ready_];
    
    // make sure this object doesn't get deleted from some other thread:
    boost::shared_ptr<mosaic_cache_shared_data_t> hold(cache->shared_);
    
    // shortcut:
    mosaic_cache_shared_data_t & shared = *(hold.get());
    
    if (shared.is_ready(requested_region, forward.GetPointer()))
    {
      // cerr << "FIXME: staying at level: " << mosaic_dl_ready_ << endl;
      return;
    }
  }
#endif
  
#ifndef NDEBUG
  if (level != mosaic_dl_ready_)
  {
    cerr << "FIXME: switching to level: " << level << endl;
  }
#endif
  
  mosaic_dl_ready_ = level;
}

//----------------------------------------------------------------
// the_stos_proc_t::setup_mosaic_pyramid
// 
void
the_stos_proc_t::setup_mosaic_pyramid()
{
  if (!fi_ready() || !mi_ready()) return;
  
  double w_max;
  double h_max;
  double spacing_x;
  double spacing_y;
  
  if (fi().spacing(0) < mi().spacing(0))
  {
    spacing_x = fi().spacing(0);
    w_max = spacing_x * double(fi().size(0));
  }
  else
  {
    spacing_x = mi().spacing(0);
    w_max = spacing_x * double(mi().size(0));
  }
  
  if (fi().spacing(1) < mi().spacing(1))
  {
    spacing_y = fi().spacing(1);
    h_max = spacing_y * double(fi().size(1));
  }
  else
  {
    spacing_y = mi().spacing(1);
    h_max = spacing_y * double(mi().size(1));
  }
  
  mosaic_->setup(w_max, h_max, spacing_x, spacing_y);
  mosaic_dl_elem_->resize(mosaic_->cache_.size());
}

//----------------------------------------------------------------
// the_stos_proc_t::fi_prim
// 
the_image_t *
the_stos_proc_t::fi_prim() const
{
  return registry()->elem<the_image_t>(fi_id_);
}

//----------------------------------------------------------------
// the_stos_proc_t::mi_prim
// 
the_image_t *
the_stos_proc_t::mi_prim() const
{
  return registry()->elem<the_image_t>(mi_id_);
}

//----------------------------------------------------------------
// the_stos_proc_t::xform_prim
// 
the_xform_t *
the_stos_proc_t::xform_prim() const
{
  return registry()->elem<the_xform_t>(xf_id_);
}

//----------------------------------------------------------------
// the_stos_proc_t::mosaic_prim
// 
the_mosaic_t *
the_stos_proc_t::mosaic_prim() const
{
  return registry()->elem<the_mosaic_t>(mz_id_);
}

//----------------------------------------------------------------
// the_stos_proc_t::fi_ready
// 
bool
the_stos_proc_t::fi_ready() const
{
  return fi_.get() != NULL && !fi().isNull();
}

//----------------------------------------------------------------
// the_stos_proc_t::mi_ready
// 
bool
the_stos_proc_t::mi_ready() const
{
  return mi_.get() != NULL && !mi().isNull();
}

//----------------------------------------------------------------
// the_stos_proc_t::save
// 
bool
the_stos_proc_t::save(std::ostream & stream) const
{
  ::save(stream, fi_->name_);
  ::save(stream, fi_->shrink_factor_);
  ::save(stream, fi_->pixel_spacing_);
  ::save(stream, fi_->flipped_);
  
  ::save(stream, mi_->name_);
  ::save(stream, mi_->shrink_factor_);
  ::save(stream, mi_->pixel_spacing_);
  ::save(stream, mi_->flipped_);
  
  ::save(stream, fi_id_);
  ::save(stream, mi_id_);
  ::save(stream, xf_id_);
  ::save(stream, mz_id_);
  return the_procedure_t::save(stream);
}

//----------------------------------------------------------------
// the_stos_proc_t::load
// 
bool
the_stos_proc_t::load(std::istream & stream)
{
  ::load(stream, fi_->name_);
  ::load(stream, fi_->shrink_factor_);
  ::load(stream, fi_->pixel_spacing_);
  ::load(stream, fi_->flipped_);
  
  ::load(stream, mi_->name_);
  ::load(stream, mi_->shrink_factor_);
  ::load(stream, mi_->pixel_spacing_);
  ::load(stream, mi_->flipped_);
  
  ::load(stream, fi_id_);
  ::load(stream, mi_id_);
  ::load(stream, xf_id_);
  ::load(stream, mz_id_);
  
  if (!the_procedure_t::load(stream))
  {
    return false;
  }
  
  return reload();
}

//----------------------------------------------------------------
// the_stos_proc_t::reload
// 
bool
the_stos_proc_t::reload()
{
  // lookup the prefered internal format for the grayscale images:
  GLenum internal_format =
    prefsUseLuminanceTextures() ?
    GL_LUMINANCE6_ALPHA2 :
    GL_RGBA8;
  
  // now load the actual images themselves:
  ImageDataTransaction<native_image_t>
    fi_trans(fi_->name_,
	     fi_->shrink_factor_,
	     fi_->pixel_spacing_,
	     fi_->flipped_,
	     OpenGL(0).max_texture_,
	     internal_format);
  
  ImageDataTransaction<native_image_t>
    mi_trans(mi_->name_,
	     mi_->shrink_factor_,
	     mi_->pixel_spacing_,
	     mi_->flipped_,
	     OpenGL(1).max_texture_,
	     internal_format);
  
  the_main_window_t * main_window =
    dynamic_cast<the_main_window_t *>(the_document_ui_t::doc_ui());
  ImageProcessingDialog tmp(main_window);
  
  if (fi_->name_.size() != 0)
  {
    tmp.thread_.add_transaction(&fi_trans);
  }
  
  if (mi_->name_.size() != 0)
  {
    tmp.thread_.add_transaction(&mi_trans);
  }
  
  if (tmp.thread_.has_work())
  {
    int r = tmp.exec();
    if (r != QDialog::Accepted)
    {
      return false;
    }
  }
  
  *(fi_.get()) = fi_trans.data_;
  *(mi_.get()) = mi_trans.data_;
  
  setup_mosaic_pyramid();
  return true;
}

//----------------------------------------------------------------
// the_stos_proc_t::dump
// 
void
the_stos_proc_t::dump(ostream & strm, unsigned int indent) const
{
  strm << INDSCP << "the_stos_proc_t(" << (void *)this << ")" << endl
       << INDSCP << "{" << endl;
  the_procedure_t::dump(strm, INDNXT);
  strm << INDSTR << "fi_id_ = " << fi_id_ << endl
       << INDSTR << "mi_id_ = " << mi_id_ << endl
       << INDSTR << "xf_id_ = " << xf_id_ << endl
       << INDSTR << "mz_id_ = " << mz_id_ << endl
       << INDSCP << "}" << endl << endl;
}


//----------------------------------------------------------------
// point_pair_t::point_pair_t
// 
point_pair_t::point_pair_t()
{
  ref_[0] = NULL;
  ref_[1] = NULL;
}

//----------------------------------------------------------------
// point_pair_t::point_pair_t
// 
point_pair_t::point_pair_t(const the_point_ref_t & a,
			   const the_point_ref_t & b)
{
  ref_[0] = (the_point_ref_t *)(a.clone());
  ref_[1] = (the_point_ref_t *)(b.clone());
}

//----------------------------------------------------------------
// point_pair_t::point_pair_t
// 
point_pair_t::point_pair_t(const point_pair_t & f):
  the_primitive_t(f)
{
  ref_[0] = (the_point_ref_t *)(f.ref_[0]->clone());
  ref_[1] = (the_point_ref_t *)(f.ref_[1]->clone());
}

//----------------------------------------------------------------
// point_pair_t::~point_pair_t
// 
point_pair_t::~point_pair_t()
{
  delete ref_[0];
  delete ref_[1];
}

//----------------------------------------------------------------
// point_pair_t::added_to_the_registry
// 
void
point_pair_t::added_to_the_registry(the_registry_t * registry,
				    const unsigned int & id)
{
  the_primitive_t::added_to_the_registry(registry, id);
  
  // update the graph:
  establish_supporter_dependent(registry, ref_[0]->id(), id);
  establish_supporter_dependent(registry, ref_[1]->id(), id);
}

//----------------------------------------------------------------
// point_pair_t::regenerate
// 
bool
point_pair_t::regenerate()
{
  the_registry_t * r = registry();
  regenerated_ =
    ref_[0]->references<the_primitive_t>(r)->regenerate() &&
    ref_[1]->references<the_primitive_t>(r)->regenerate();
  
  return regenerated_;
}

//----------------------------------------------------------------
// point_pair_t::save
// 
bool
point_pair_t::save(std::ostream & stream) const
{
  ::save(stream, ref_[0]);
  ::save(stream, ref_[1]);
  return the_primitive_t::save(stream);
}

//----------------------------------------------------------------
// point_pair_t::load
// 
bool
point_pair_t::load(std::istream & stream)
{
  the_reference_t * r0 = NULL;
  ::load(stream, r0);
  ref_[0] = dynamic_cast<the_point_ref_t *>(r0);
  
  the_reference_t * r1 = NULL;
  ::load(stream, r1);
  ref_[1] = dynamic_cast<the_point_ref_t *>(r1);
  
  return the_primitive_t::load(stream);
}

//----------------------------------------------------------------
// point_pair_t::dump
// 
void
point_pair_t::dump(ostream & strm, unsigned int indent) const
{
  strm << INDSCP << "point_pair_t(" << (void *)this << ")" << endl
       << INDSCP << "{" << endl;
  the_primitive_t::dump(strm, INDNXT);
  strm << INDSTR << "ref_[0] = ";
  ref_[0]->dump(strm, INDNXT);
  strm << INDSTR << "ref_[1] = ";
  ref_[1]->dump(strm, INDNXT);
  strm << INDSCP << "}" << endl << endl;
}

//----------------------------------------------------------------
// the_image_t::regenerate
// 
bool
the_image_t::regenerate()
{
  the_stos_proc_t * proc = first_supporter<the_stos_proc_t>(this);
  const image_data_t & data =
    (id() == proc->fi_prim()->id()) ?
    proc->fi() :
    proc->mi();
  
  regenerated_ = !data.isNull();
  
  if (regenerated_)
  {
    GLenum filter = prefsAllowLinearFiltering() ? GL_LINEAR : GL_NEAREST;
    
    if (dl_elem_.get() == NULL ||
	dl_elem_->data_.buffer_.get() !=
	data.tile_generator_.buffer_.get())
    {
      dl_elem_ =
	boost::shared_ptr<image_tile_dl_elem_t>
	(data.dl_elem(filter, filter).release());
    }
    else
    {
      dl_elem_->min_filter_ = filter;
      dl_elem_->mag_filter_ = filter;
    }
  }
  else
  {
    dl_elem_ = boost::shared_ptr<image_tile_dl_elem_t>();
  }
  
  return regenerated_;
}


//----------------------------------------------------------------
// the_xform_t::the_xform_t
// 
the_xform_t::the_xform_t():
  the_primitive_t()
{}

//----------------------------------------------------------------
// the_xform_t::the_xform_t
// 
the_xform_t::the_xform_t(const the_xform_t & t):
  the_primitive_t(t),
  dl_elem_(t.dl_elem_)
{
  forward_ = stos_transform_t::New();
  inverse_ = stos_transform_t::New();
  
  if (t.forward_)
  {
    forward_->SetFixedParameters(t.forward_->GetFixedParameters());
    forward_->SetParameters(t.forward_->GetParameters());
  }
  
  if (t.inverse_)
  {
    inverse_->SetFixedParameters(t.inverse_->GetFixedParameters());
    inverse_->SetParameters(t.inverse_->GetParameters());
  }
}

//----------------------------------------------------------------
// the_xform_t::regenerate
// 
bool
the_xform_t::regenerate()
{
  the_stos_proc_t * proc = first_supporter<the_stos_proc_t>(this);
  proc->calc_transforms(forward_, inverse_);
  regenerated_ = proc->mi_prim()->regenerated();
  
  if (prefsUseOpenGLForMosaicking() && regenerated_)
  {
    const image_tile_generator_t & image = proc->mi().tile_generator_;
    
    // resample the transform onto a grid transform:
    the_grid_transform_t transform;
    
    pnt2d_t tile_min = pnt2d(image.bbox_.min_[0],
			     image.bbox_.min_[1]);
    pnt2d_t tile_max = pnt2d(image.bbox_.max_[0],
			     image.bbox_.max_[1]);
    
    // FIXME: perhaps this should be in the preferences?
#if 0
    static const unsigned int rows = 10;
    static const unsigned int cols = 10;
#else
    static const unsigned int rows = 17;
    static const unsigned int cols = 17;
#endif
    
#ifndef NDEBUG
    bool ok =
#endif
      setup_grid_transform(transform,
			   rows,
			   cols,
			   tile_min,
			   tile_max,
			   NULL,
			   forward_.GetPointer(),
			   100,		// iterations
			   1e-12,	// min step scale
			   1e-16,	// min error squared
			   5);		// successfull steps to pick up pace
    assert(ok);
    
    dl_elem_ =
      boost::shared_ptr<the_warped_image_dl_elem_t>
      (new the_warped_image_dl_elem_t(proc->mi_prim()->dl_elem_, transform));
  }
  else
  {
    dl_elem_ = boost::shared_ptr<the_warped_image_dl_elem_t>();
  }
  
  return regenerated_;
}


//----------------------------------------------------------------
// the_stos_mosaic_dl_elem_t::the_stos_mosaic_dl_elem_t
// 
the_stos_mosaic_dl_elem_t::
the_stos_mosaic_dl_elem_t(const the_stos_proc_t * proc)
{
  fi_ = proc->fi_prim()->dl_elem_;
  mi_ = proc->xform_prim()->dl_elem_;
}

//----------------------------------------------------------------
// the_stos_mosaic_dl_elem_t::draw
// 
void
the_stos_mosaic_dl_elem_t::draw() const
{
  static const the_color_t zebra[] = {
    the_color_t(0, 0.5, 1, 1),
    the_color_t(1, 0.5, 0, 0)
  };

  bool fi_ready = fi_.get() != NULL;
  bool mi_ready = mi_.get() != NULL;

  if (fi_ready)
  {
    fi_->draw(the_color_t::BLACK, false);
  }
  
  if (mi_ready)
  {
    mi_->draw(the_color_t::BLACK, false);
  }
  
  if (fi_ready)
  {
    fi_->draw(zebra[0], true);
  }
  
  if (mi_ready)
  {
    mi_->draw(zebra[1], true);
  }
}

//----------------------------------------------------------------
// the_stos_mosaic_dl_elem_t::update_bbox
// 
void
the_stos_mosaic_dl_elem_t::update_bbox(the_bbox_t & bbox) const
{
  bool fi_ready = fi_.get() != NULL;
  bool mi_ready = mi_.get() != NULL;
  
  if (fi_ready)
  {
    fi_->update_bbox(bbox);
  }
  
  if (mi_ready)
  {
    mi_->update_bbox(bbox);
  }
}
