// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : the_stos_proc.hxx
// Author       : Pavel A. Koshevoy
// Created      : 2006/07/11 13:20
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Landmark based slice to slice registration procedure.

#ifndef THE_STOS_PROC_HXX_
#define THE_STOS_PROC_HXX_

// Boost includes:
#include <boost/shared_ptr.hpp>

// local includes:
#include "image_data.hxx"
#include "mosaic_cache.hxx"
#include "mosaic_cache_pyramid.hxx"
#include "the_warped_image_dl_elem.hxx"

// the includes:
#include <utils/the_text.hxx>
#include <opengl/the_disp_list.hxx>
#include <doc/the_procedure.hxx>
#include <geom/the_point.hxx>
#include <itk/itkLegendrePolynomialTransform.h>
#include <itk/itkRBFTransform.h>

// Cg includes:
#include <Cg/cgGL.h>

// forward declarations:
extern CGprogram * CG_MOSAIC_FRAGMENT_PROGRAM_PTR;
class the_bbox_t;
class the_view_t;
class the_view_volume_t;
class the_image_t;
class the_xform_t;
class the_mosaic_t;


//----------------------------------------------------------------
// stos_transform_t
//
// #define USE_LEGENDRE
#ifdef USE_LEGENDRE
typedef itk::LegendrePolynomialTransform<double, 3> stos_transform_t;
#else
typedef itk::RBFTransform stos_transform_t;
#endif



//----------------------------------------------------------------
// mosaic_dl_elem_t
// 
class mosaic_dl_elem_t : public image_tile_dl_elem_t
{
public:
  mosaic_dl_elem_t(boost::shared_ptr<mosaic_cache_shared_data_t> & shared,
		   GLenum min_filter = GL_NEAREST,
		   GLenum mag_filter = GL_NEAREST):
#ifdef USE_CG
    image_tile_dl_elem_t(shared->data_,
			 min_filter,
			 mag_filter,
			 CG_MOSAIC_FRAGMENT_PROGRAM_PTR),
#else
    image_tile_dl_elem_t(shared->data_,
			 min_filter,
			 mag_filter),
#endif
    shared_(shared)
  {}
  
  ~mosaic_dl_elem_t()
  {}
  
  // virtual:
  void draw() const
  {
    // upload the texture data:
#if 1
    {
      // NOTE: this is an optimization which avoids uploading
      // all texture tiles, and instead uploads only the updated
      // portions of affected texture tiles:
      
      the_lock_t<the_mutex_interface_t> locker(shared_->mutex_);
      upload_.splice(upload_.end(), shared_->upload_);
    }
#endif
    
#ifdef USE_CG
    if (has_fragment_program())
    {
      cgGLLoadProgram(*fragment_program_);
      
      CGparameter fi_param = cgGetNamedParameter(*fragment_program_,
						 "fi_tint");
      CGparameter mi_param = cgGetNamedParameter(*fragment_program_,
						 "mi_tint");
      
      // cgGLSetParameter3f(fi_param, 1.0f, 0.0f, 0.0f);
      // cgGLSetParameter3f(mi_param, 0.0f, 1.0f, 1.0f);
      v3x1_t fi_tint(1.0f, 0.5f, 0.0f);
      v3x1_t mi_tint = v3x1_t(1, 1, 1) - fi_tint;
      cgGLSetParameter3f(fi_param, fi_tint[0], fi_tint[1], fi_tint[2]);
      cgGLSetParameter3f(mi_param, mi_tint[0], mi_tint[1], mi_tint[2]);
    }
#endif
    
    image_tile_dl_elem_t::draw();
  }
  
  boost::shared_ptr<mosaic_cache_shared_data_t> shared_;
  
private:
  mosaic_dl_elem_t();
};


//----------------------------------------------------------------
// the_stos_proc_t
// 
class the_stos_proc_t : public the_procedure_t
{
public:
  the_stos_proc_t();
  the_stos_proc_t(const the_stos_proc_t & proc);
  ~the_stos_proc_t();
  
  // virtual:
  the_primitive_t * clone() const
  { return new the_stos_proc_t(*this); }
  
  // virtual:
  const char * name() const
  { return "the_stos_proc_t"; }
  
  // virtual:
  void added_to_the_registry(the_registry_t * registry,
			     const unsigned int & id);
  
  // virtual:
  bool regenerate();
  
  // virtual:
  void draw(const the_view_t & view) const;
  
  // virtual: calculate the model bounding box:
  void calc_bbox(const the_view_t & view, the_bbox_t & bbox) const;
  
  // helper functions:
  void swap_images();
  
  // generate forward/inverse transforms based on the current set
  // of point pairs:
  void calc_transforms(stos_transform_t::Pointer & forward,
		       stos_transform_t::Pointer & inverse) const;
  
  // regenerate the image display list:
  bool regenerate_image(const unsigned int & image_id);
  
  // regenerate the mosaic image display list:
  bool regenerate_mosaic();
  
  // update the mosaic display list:
  void mosaic_ready(const unsigned int & level);
  
  // setup the mosaic cache pyramid:
  void setup_mosaic_pyramid();
  
  // helpers:
  the_image_t * fi_prim() const;
  the_image_t * mi_prim() const;
  the_xform_t * xform_prim() const;
  the_mosaic_t * mosaic_prim() const;
  
  // accessors:
  inline const image_data_t & fi() const { return *fi_; }
  inline const image_data_t & mi() const { return *mi_; }
  inline const mosaic_cache_pyramid_t & mosaic() const { return *mosaic_; }
  
  inline image_data_t & fi() { return *fi_; }
  inline image_data_t & mi() { return *mi_; }
  inline mosaic_cache_pyramid_t & mosaic() { return *mosaic_; }
  
  // check whether the fixed/moving images are available:
  bool fi_ready() const;
  bool mi_ready() const;
  
  inline bool images_are_ready() const
  { return fi_ready() && mi_ready(); }
  
  // common: file io:
  static bool loader(std::istream & stream, the_primitive_t *& primitive)
  {
    the_stos_proc_t * stos = new the_stos_proc_t();
    bool ok = stos->load(stream);
    primitive = stos;
    return ok;
  }
  
  // virtual: file io:
  bool save(std::ostream & stream) const;
  bool load(std::istream & stream);
  
  bool reload();
  
  // virtual: For debugging, dumps the value
  void dump(ostream & strm, unsigned int indent = 0) const;
  
  // FIXME: protected:
  
  // helpers for accessing a given mosaic level display element:
  inline boost::shared_ptr<mosaic_dl_elem_t> &
  mosaic_dl_elem(const unsigned int & level)
  { return (*(mosaic_dl_elem_.get()))[level]; }
  
  inline const mosaic_dl_elem_t *
  mosaic_dl_elem(const unsigned int & level) const
  { return (*(mosaic_dl_elem_.get()))[level].get(); }
  
  // fixed/moving/mosaic image data:
  boost::shared_ptr<image_data_t> fi_;
  boost::shared_ptr<image_data_t> mi_;
  boost::shared_ptr<mosaic_cache_pyramid_t> mosaic_;
  
  // multi-resolution display elements for the mosaic:
  boost::shared_ptr<std::vector<boost::shared_ptr<mosaic_dl_elem_t> > >
  mosaic_dl_elem_;
  unsigned int mosaic_dl_ready_;
  
  // landmark display lists (compiled):
  boost::shared_ptr<the_disp_list_t> dl_landmarks_fi_;
  boost::shared_ptr<the_disp_list_t> dl_landmarks_mi_;
  boost::shared_ptr<the_disp_list_t> dl_landmarks_mosaic_;
  boost::shared_ptr<the_disp_list_t> dl_;
  
  // id's of various primitives:
  unsigned int fi_id_; // fixed image id
  unsigned int mi_id_; // moving image id
  unsigned int xf_id_; // transform id
  unsigned int mz_id_; // mosaic id
  
public:
  // a simple selection set:
  unsigned int selection_;
};


//----------------------------------------------------------------
// point_pair_t
// 
class point_pair_t : public the_primitive_t
{
public:
  point_pair_t();
  point_pair_t(const the_point_ref_t & a,
	       const the_point_ref_t & b);
  point_pair_t(const point_pair_t & f);
  ~point_pair_t();
  
  // virtual:
  the_primitive_t * clone() const
  { return new point_pair_t(*this); }
  
  // virtual:
  const char * name() const
  { return "point_pair_t"; }
  
  // virtual:
  void added_to_the_registry(the_registry_t * registry,
			     const unsigned int & id);
  
  // virtual:
  bool regenerate();
  
  // virtual: file io:
  bool save(std::ostream & stream) const;
  bool load(std::istream & stream);
  
  // virtual: For debugging, dumps the value
  void dump(ostream & strm, unsigned int indent = 0) const;
  
  // references to points that define this point pair:
  the_point_ref_t * ref_[2];
};


//----------------------------------------------------------------
// the_image_t
//
// Even though the image primitive is a shallow class and does
// not store any image data internally due to performance concerns,
// it is necessary in order to simplify the regeneration process.
// 
class the_image_t : public the_primitive_t
{
public:
  // virtual:
  the_primitive_t * clone() const
  { return new the_image_t(*this); }
  
  // virtual:
  const char * name() const
  { return "the_image_t"; }
  
  // virtual:
  bool regenerate();
  
  boost::shared_ptr<image_tile_dl_elem_t> dl_elem_;
};


//----------------------------------------------------------------
// the_xform_t
//
// The transform is defined by the point pairs.
// 
class the_xform_t : public the_primitive_t
{
public:
  the_xform_t();
  the_xform_t(const the_xform_t & t);
  
  // virtual:
  the_primitive_t * clone() const
  { return new the_xform_t(*this); }
  
  // virtual:
  const char * name() const
  { return "the_xform_t"; }
  
  // virtual:
  bool regenerate();
  
  // the forward/inverse transforms:
  stos_transform_t::Pointer forward_;
  stos_transform_t::Pointer inverse_;
  
  // the warped tile display list element:
  boost::shared_ptr<the_warped_image_dl_elem_t> dl_elem_;
};


//----------------------------------------------------------------
// the_mosaic_t
//
// The mosaic is defined by the transform.
// Even though the mosaic primitive is a shallow class and does
// not store any data internally due to performance concerns,
// it is necessary in order to simplify the regeneration process.
// 
class the_mosaic_t : public the_primitive_t
{
public:
  the_mosaic_t():
    first_time_regenerating_(true)
  {}
  
  // virtual:
  the_primitive_t * clone() const
  { return new the_mosaic_t(*this); }
  
  // virtual:
  const char * name() const
  { return "the_mosaic_t"; }
  
  // virtual:
  bool regenerate()
  {
    the_stos_proc_t * proc = first_supporter<the_stos_proc_t>(this);
    regenerated_ = proc->regenerate_mosaic();
    return regenerated_;
  }
  
  // This flag indicates that the mosaic is being regenerated for
  // the first time. This flag controls whether the view manager
  // will be reset when the actual mosaic regeneration completes:
  bool first_time_regenerating_;
};


//----------------------------------------------------------------
// the_stos_mosaic_dl_elem_t
// 
class the_stos_mosaic_dl_elem_t : public the_dl_elem_t
{
public:
  the_stos_mosaic_dl_elem_t(const the_stos_proc_t * proc);
  
  // virtual:
  void draw() const;
  
  // virtual:
  void update_bbox(the_bbox_t & bbox) const;
  
private:
  the_stos_mosaic_dl_elem_t();
  
  boost::shared_ptr<image_tile_dl_elem_t> fi_;
  boost::shared_ptr<the_warped_image_dl_elem_t> mi_;
};


#endif // THE_STOS_PROC_HXX_
