// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : the_stos_proc_ui.hxx
// Author       : Pavel A. Koshevoy
// Created      : 2006/07/11 13:20
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : UI for the landmark registration tool
//                used for slice to slice registration.

// local includes:
#include "the_stos_proc_ui.hxx"
#include "the_stos_proc.hxx"
#include "the_main_window.hxx"
#include "mosaic_cache.hxx"
#include "ImageProcessingDialog.h"
#include "FlipImageTransaction.h"
#include "WarpImageTransaction.h"

// the includes:
#include <geom/the_point.hxx>
#include <doc/the_graph.hxx>
#include <sel/the_pick_list.hxx>
#include <math/the_bbox.hxx>
#include <math/the_aa_bbox.hxx>
#include <image/image_tile_generator.hxx>
#include <itk/common.hxx>
#include <itk/stos.hxx>
#include <itk/stos_common.hxx>
#include <Qt/the_qt_trail.hxx>
#include <Qt/the_qt_thread_pool.hxx>

// system includes:
#include <fstream>
#include <memory>

// Qt includes:
#include <QAction>
#include <QLineEdit>
#include <QButtonGroup>
#include <QRadioButton>
#include <QPushButton>
#include <QStackedWidget>


//----------------------------------------------------------------
// MOSAIC_THREAD_POOL
// 
the_qt_thread_pool_t *
MOSAIC_THREAD_POOL()
{
  static the_qt_thread_pool_t * pool_ = NULL;
  if (pool_ == NULL)
  {
    pool_ = new the_qt_thread_pool_t(4);
  }
  
  return pool_;
}


//----------------------------------------------------------------
// mosaic_transaction_t
//
// Update the requested mosaic region.
// If the max_texture transaction parameter is non zero
// a corresponding display element will also be created.
// 
class mosaic_transaction_t : public the_transaction_t
{
public:
  // constructor:
  mosaic_transaction_t(mosaic_cache_t & cache,
		       const unsigned int & prefetch_level,
		       const the_aa_bbox_t & requested_region,
		       const image_tile_generator_t & fixed,
		       const image_tile_generator_t & moving,
		       const base_transform_t::ConstPointer & t,
		       const unsigned int max_texture,
		       const GLenum & filter,
   const bool & rgba_compressed):
    cache_(cache),
    prefetch_level_(prefetch_level),
    requested_region_(requested_region),
    fixed_(fixed),
    moving_(moving),
    transform_(t),
    max_texture_(max_texture),
    filter_(filter),
    rgba_compressed_(rgba_compressed)
  {}
  
  // virtual:
  void execute(the_thread_interface_t * /* thread */)
  {
    // update the requested region:
    cache_.update(requested_region_,
		  fixed_,
		  moving_,
		  transform_,
		  rgba_compressed_);
    
    make_dl_elem();
    
    if (prefetch_level_ < 2)
    {
      // increase the requested region for each prefetch iteration:
      mosaic_transaction_t * t =
	new mosaic_transaction_t(cache_,
				 prefetch_level_ + 1,
				 requested_region_ * 1.6,
				 fixed_,
				 moving_,
				 transform_,
				 max_texture_,
				 filter_,
				 rgba_compressed_);
      MOSAIC_THREAD_POOL()->push_back(t, true);
    }
  }
  
  void make_dl_elem()
  {
    // make sure the shared data doesn't get deleted from some other thread:
    boost::shared_ptr<mosaic_cache_shared_data_t> hold_shared(cache_.shared_);
    
    // shortcut:
    mosaic_cache_shared_data_t & shared = *(hold_shared.get());
    the_lock_t<the_mutex_interface_t> locker(shared.mutex_);
    
    if (shared.data_.buffer_.get() == NULL)
    {
      // cut new tiles:
      if (rgba_compressed_)
      {
	shared.make_tiles(GL_UNSIGNED_BYTE,
			  GL_LUMINANCE,	// internal data format
			  GL_LUMINANCE,	// data format
			  max_texture_ / 2);
      }
      else
      {
	shared.make_tiles(GL_UNSIGNED_INT_8_8_8_8_REV,
			  GL_RGBA8,	// internal data format
			  GL_BGRA_EXT,	// data format
			  max_texture_);
      }
      
      // setup the display element:
      cache_.dl_elem_ =
	std::auto_ptr<the_dl_elem_t>
	(new mosaic_dl_elem_t(hold_shared, filter_, filter_));
    }
  }
  
  // trasaction parameters:
  mosaic_cache_t & cache_;
  const unsigned int prefetch_level_;
  const the_aa_bbox_t requested_region_;
  const image_tile_generator_t & fixed_;
  const image_tile_generator_t & moving_;
  base_transform_t::ConstPointer transform_;
  const unsigned int max_texture_;
  GLenum filter_;
  const bool rgba_compressed_;
};


//----------------------------------------------------------------
// activate_point_pair
// 
static void
activate_point_pair(the_point_t * pt)
{
  if (pt == NULL) return;
  
  point_pair_t * pp = first_dependent<point_pair_t>(pt);
  
  the_registry_t * r = pt->registry();
  the_point_t * pt_a = pp->ref_[0]->references<the_point_t>(r);
  the_point_t * pt_b = pp->ref_[1]->references<the_point_t>(r);
  
  pt_a->set_current_state(THE_SELECTED_STATE_E);
  pt_b->set_current_state(THE_SELECTED_STATE_E);
}

//----------------------------------------------------------------
// deactivate_point_pair
// 
static void
deactivate_point_pair(the_point_t * pt)
{
  if (pt == NULL) return;
  
  point_pair_t * pp = first_dependent<point_pair_t>(pt);
  
  the_registry_t * r = pt->registry();
  the_point_t * pt_a = pp->ref_[0]->references<the_point_t>(r);
  the_point_t * pt_b = pp->ref_[1]->references<the_point_t>(r);
  
  pt_a->clear_state(THE_SELECTED_STATE_E);
  pt_b->clear_state(THE_SELECTED_STATE_E);
}


//----------------------------------------------------------------
// the_stos_proc_eh_t::the_stos_proc_eh_t
// 
the_stos_proc_eh_t::the_stos_proc_eh_t(the_document_ui_t * doc_ui):
  the_document_eh_t<the_document_t>(doc_ui)
{}

//----------------------------------------------------------------
// the_stos_proc_eh_t::is_modifier_active_locally
// 
bool
the_stos_proc_eh_t::is_modifier_active_locally(the_eh_modifier_t mod) const
{
#if 0 
  cout << "--------------------------------------------------------" << endl;
  ::dump(THE_MOUSE);
  ::dump(THE_KEYBD);
#endif
  
#if 0
  switch (mod)
  {
    case THE_EH_MOD_VIEW_ZOOM_E:
      return ((state() != THE_EH_NOT_READY_E) &&
	      THE_MOUSE.verify_pressed_btns(THE_MOUSE.m_btn()) &&
	      THE_KEYBD.verify_no_pressed_keys());
      
    case THE_EH_MOD_VIEW_PAN_E:
      return ((state() != THE_EH_NOT_READY_E) &&
	      THE_MOUSE.verify_pressed_btns(THE_MOUSE.l_btn()) &&
	      THE_KEYBD.verify_no_pressed_keys());
      
    default:
      break;
  }
#endif
  
  return the_input_device_eh_t::is_modifier_active_locally(mod);
}

//----------------------------------------------------------------
// the_stos_proc_eh_t::delete_something
// 
void
the_stos_proc_eh_t::delete_something()
{
  // cerr << "delete_something" << endl;
  
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  if (active_pt() != NULL)
  {
    doc_so().save_undo_record();
    
    point_pair_t * pp = first_dependent<point_pair_t>(active_pt());
    the_registry_t * r = registry();
    
    the_point_t * pt_a = pp->ref_[0]->references<the_point_t>(r);
    the_point_t * pt_b = pp->ref_[1]->references<the_point_t>(r);
    
    unsigned int pp_id = pp->id();
    breakdown_supporter_dependent(r, proc()->id(), pp_id);
    breakdown_supporter_dependent(r, pp_id, proc()->xform_prim()->id());
    r->del(pp);
    delete pp;
    proc()->xform_prim()->request_regeneration();
    
    breakdown_supporter_dependent(r, proc()->id(), pt_a->id());
    r->del(pt_a);
    delete pt_a;
    
    breakdown_supporter_dependent(r, proc()->id(), pt_b->id());
    r->del(pt_b);
    delete pt_b;
  }
  proc()->selection_ = std::numeric_limits<unsigned int>::max();
  
  // regenerate and update the UI:
  doc_ui().sync_ui();
}

//----------------------------------------------------------------
// the_stos_proc_eh_t::clear_all
// 
void
the_stos_proc_eh_t::clear_all()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  the_stos_proc_t * p = proc();
  the_registry_t * r = p->registry();
  std::list<point_pair_t *> landmarks;
  
  for (std::list<unsigned int>::const_iterator
	 i = p->direct_dependents().begin();
       i != p->direct_dependents().end(); ++i)
  {
    point_pair_t * pp = r->elem<point_pair_t>(*i);
    if (pp != NULL) landmarks.push_back(pp);
  }
  
  if (landmarks.empty()) return;
  
  doc_so().save_undo_record();
  while (!landmarks.empty())
  {
    point_pair_t * pp = remove_head(landmarks);
    
    the_point_t * pt_a = pp->ref_[0]->references<the_point_t>(r);
    the_point_t * pt_b = pp->ref_[1]->references<the_point_t>(r);
    
    unsigned int pp_id = pp->id();
    breakdown_supporter_dependent(r, p->id(), pp_id);
    breakdown_supporter_dependent(r, pp_id, p->xform_prim()->id());
    r->del(pp);
    delete pp;
    p->xform_prim()->request_regeneration();
    
    breakdown_supporter_dependent(r, p->id(), pt_a->id());
    r->del(pt_a);
    delete pt_a;
    
    breakdown_supporter_dependent(r, p->id(), pt_b->id());
    r->del(pt_b);
    delete pt_b;
  }
  p->selection_ = std::numeric_limits<unsigned int>::max();
  
  // regenerate and update the UI:
  doc_ui().sync_ui();
}

//----------------------------------------------------------------
// transaction_payload_t
// 
class transaction_payload_t
{
public:
  transaction_payload_t():
    transaction_(NULL)
  {}
  
  ~transaction_payload_t()
  {
    delete transaction_;
  }
  
  void set_payload(the_transaction_t * t)
  {
    delete transaction_;
    transaction_ = t;
  }
  
  the_transaction_t * get_payload()
  {
    the_transaction_t * t = transaction_;
    transaction_ = NULL;
    return t;
  }
  
  inline bool has_payload() const
  { return transaction_ != NULL; }
  
  QMutex mutex_;
  the_transaction_t * transaction_;
};

//----------------------------------------------------------------
// rgba_done_event_t
// 
class rgba_done_event_t : public QEvent
{
public:
  rgba_done_event_t(transaction_payload_t & payload):
    QEvent(QEvent::Type(QEvent::User + 1)),
    payload_(payload)
  {}
  
  transaction_payload_t & payload_;
};

//----------------------------------------------------------------
// rgba_done_cb
// 
static void
rgba_done_cb(void * obj,
	     the_transaction_t * transaction,
	     the_transaction_t::state_t state)
{
  // cerr << "transaction " << state << endl;
  switch (state)
  {
    case the_transaction_t::PENDING_E:
    case the_transaction_t::STARTED_E:
      return;
      
    default:
      break;
  }

  if (state == the_transaction_t::DONE_E)
  {
    // take ownership of the payload:
    static transaction_payload_t payload;
    QMutexLocker locker(&payload.mutex_);
    
    bool had_payload = payload.has_payload();
    payload.set_payload(transaction);
    if (had_payload) return;
    
    // send a custom event to trigger the rgbaDone slot:
    the_stos_proc_ui_t * ui = (the_stos_proc_ui_t *)(obj);
    QCoreApplication::postEvent(ui, new rgba_done_event_t(payload));
  }
  else
  {
    // discard the transaction:
    delete transaction;
  }
}

//----------------------------------------------------------------
// the_stos_proc_eh_t::view_handler
// 
bool
the_stos_proc_eh_t::view_handler(the_view_t * view)
{
  if (prefsUseOpenGLForMosaicking())
  {
    return false;
  }
  
  // prevent recursive calls:
  static unsigned int call_count = 0;
  if (call_count > 0) return false;
  call_count++;
  
  bool view_refreshed = false;
  
  the_main_window_t * main_window =
    dynamic_cast<the_main_window_t *>(&doc_ui());
  the_stos_proc_ui_t * stos_ui = main_window->stos_proc_ui();
  the_stos_proc_t * stos = proc();
  
  the_view_t * view_mosaic = doc_ui().view(2);
  if (view == view_mosaic &&
      stos->fi_ready() &&
      stos->mi_ready() &&
      stos->xform_prim()->regenerated() &&
      prefsAutoRegeneration())
  {
    // shortcuts:
    const unsigned int max_texture = OpenGL(2).max_texture_;
    const GLenum filter = prefsAllowLinearFiltering() ? GL_LINEAR : GL_NEAREST;
    const bool rgba_compressed = prefsUseFragmentShader();
    
    const the_view_mgr_t & view_mgr = view_mosaic->view_mgr();
    the_view_volume_t vol;
    view_mgr.setup_view_volume(vol);
    
    p3x1_t view_min;
    p3x1_t view_max;
    vol.scs_to_wcs(p2x1_t(0, 0), view_min);
    vol.scs_to_wcs(p2x1_t(1, 1), view_max);
    view_min[2] = 0;
    view_max[2] = 0;
    
    the_aa_bbox_t requested_region;
    requested_region << view_min << view_max;
    
    the_xform_t * xform = stos->xform_prim();
    stos_transform_t::Pointer forward = stos_transform_t::New();
    forward->SetFixedParameters(xform->forward_->GetFixedParameters());
    forward->SetParameters(xform->forward_->GetParameters());
    
    const image_tile_generator_t & fixed = stos->fi().tile_generator_;
    const image_tile_generator_t & moving = stos->mi().tile_generator_;
    
    // find the highest resolution pyramid level that has
    // the requested region readily available:
    mosaic_cache_pyramid_t & pyramid = stos->mosaic();
    
    // FIXME:
    int num_levels = pyramid.cache_.size();
    // int num_levels = 1;
    int ready_level = num_levels;
    
#if 1
    for (int i = num_levels - 1; i > -1; i--)
    {
      mosaic_cache_t * cache = pyramid.cache_[i];
      
      // make sure this object doesn't get deleted from some other thread:
      boost::shared_ptr<mosaic_cache_shared_data_t> hold(cache->shared_);
      
      // shortcut:
      mosaic_cache_shared_data_t & shared = *(hold.get());
      
      if (!shared.is_ready(requested_region, forward.GetPointer()))
      {
	// cache not ready, use the previous ready cache level:
	break;
      }
      
      // save the new ready cache candidate:
      ready_level = i;
    }
    
    if (ready_level < num_levels)
    {
      mosaic_cache_t * cache = pyramid.cache_[ready_level];
      // cerr << "FIXME: ready level: " << ready_level << endl;
      
      // draw the readily available mosaic:
      mosaic_transaction_t * trans =
	new mosaic_transaction_t(*cache,
				 0,
				 requested_region,
				 fixed,
				 moving,
				 forward.GetPointer(),
				 max_texture,
				 filter,
				 rgba_compressed);
      trans->set_notify_cb(rgba_done_cb, stos_ui);
      
      // generate the display list element if necessary:
      trans->make_dl_elem();
      
      // mark the transaction as completed so it wouldn't be ignored:
      trans->notify(MOSAIC_THREAD_POOL(), the_transaction_t::DONE_E);
      
      view_refreshed = true;
    }
#endif
    
    // schedule all other levels of the pyramid to be updated:
    std::list<the_transaction_t *> schedule;
    for (int i = ready_level - 1; i > -1; i--)
    {
      // cerr << "FIXME: regenerate level " << i << endl;
      mosaic_transaction_t * trans =
	new mosaic_transaction_t(*(pyramid.cache_[i]),
				 0,
				 requested_region,
				 fixed,
				 moving,
				 forward.GetPointer(),
				 max_texture,
				 filter,
				 rgba_compressed);
      trans->set_notify_cb(rgba_done_cb, stos_ui);
      schedule.push_back(trans);
    }
    
    if (!schedule.empty())
    {
      MOSAIC_THREAD_POOL()->stop_and_go(schedule, true);
    }
    
#if 0
    // prefetch -- anticipate further requests by scheduling expanded
    // region updates:
    for (int prefetch = 0; prefetch < 1; prefetch++)
    {
      // double the requested region for each prefetch iteration:
      requested_region *= 1.6;
      
      for (int i = num_levels - 1; i > -1; i--)
      {
	mosaic_cache_t * cache = pyramid.cache_[i];
	
	// make sure this object doesn't get deleted from some other thread:
	boost::shared_ptr<mosaic_cache_shared_data_t> hold(cache->shared_);
	
	// shortcut:
	mosaic_cache_shared_data_t & shared = *(hold.get());
	
	if (!shared.is_ready(requested_region, forward.GetPointer()))
	{
	  // cerr << "FIXME: prefetch level " << i << endl;
	  mosaic_transaction_t * trans =
	    new mosaic_transaction_t(*(pyramid.cache_[i]),
				     prefetch + 1,
				     requested_region,
				     fixed,
				     moving,
				     forward.GetPointer(),
				     max_texture,
				     filter,
				     rgba_compressed);
	  trans->set_notify_cb(rgba_done_cb, stos_ui);
	  MOSAIC_THREAD_POOL()->push_back(trans, true);
	}
      }
    }
#endif
  }
  
  call_count--;
  return view_refreshed;
}

//----------------------------------------------------------------
// the_stos_proc_eh_t::processor
// 
bool
the_stos_proc_eh_t::
processor(the_input_device_eh_t::the_event_type_t event_type)
{
  bool processed = false;
  
  if (event_type == THE_MOUSE_EVENT_E)
  {
    processed = process_mouse();
  }
  else if (event_type == THE_KEYBD_EVENT_E)
  {
    processed = false;
  }
  
  if (processed)
  {
    // call document sync_ui, which will in turn call proc()->regenerate():
    doc_ui().sync_ui();
  }
  
  return processed;
}

//----------------------------------------------------------------
// the_stos_proc_eh_t::registry
// 
the_registry_t *
the_stos_proc_eh_t::registry() const
{
  return proc()->registry();
}

//----------------------------------------------------------------
// the_stos_proc_eh_t::proc
// 
the_stos_proc_t *
the_stos_proc_eh_t::proc() const
{
  the_document_t * d = document();
  if (d == NULL) return NULL;
  return d->active_procedure<the_stos_proc_t>();
}

//----------------------------------------------------------------
// the_stos_proc_eh_t::process_mouse
// 
bool
the_stos_proc_eh_t::process_mouse()
{
  // both images must be defined prior to creating or
  // manipulating the point pairs:
  if (proc()->images_are_ready() == false) return false;
  
  the_view_t * view_fixed = doc_ui().view(0);
  the_view_t * view_moving = doc_ui().view(1);
  the_view_t * view_mosaic = doc_ui().view(2);
  
  if (THE_KEYBD.verify_pressed_keys(THE_KEYBD.ctl()) &&
      THE_MOUSE.l_btn().down() &&
      THE_MOUSE.l_btn().changed() &&
      !THE_MOUSE.l_btn().moved_while_down())
  {
    the_trail_t::milestone_t milestone;
    
    // create a new point pair:
    the_pick_list_t curr_pick;
    if (!curr_pick.pick(view(),
			scs_pt(),
			the_pick_list_t::default_filter_,
			view().primitives()))
    {
      // cout << "picked nothing -- create a new point pair" << endl;
      doc_so().save_undo_record();
      deactivate_point_pair(active_pt());
      
      // Depending on which view the point is being created in,
      // either pt_a or pt_b has to be transformed by the current
      // estimate of the registration transform (or its inverse).
      the_xform_t * xform = proc()->xform_prim();
      stos_transform_t::Pointer forward = xform->forward_;
      stos_transform_t::Pointer inverse = xform->inverse_;
      
      p3x1_t wcs_a = b_wcs();
      p3x1_t wcs_b = b_wcs();
      
      if (&view() == view_fixed || &view() == view_mosaic)
      {
	pnt2d_t a = pnt2d(wcs_a.x(), wcs_a.y());
	pnt2d_t b = forward->TransformPoint(a);
	wcs_b[0] = b[0];
	wcs_b[1] = b[1];
      }
      else if (&view() == view_moving)
      {
	pnt2d_t b = pnt2d(wcs_b.x(), wcs_b.y());
	// pnt2d_t a = inverse->TransformPoint(b);
	pnt2d_t a;
	if (!find_inverse(forward.GetPointer(), inverse.GetPointer(), b, a))
	{
	  return false;
	}
	
	wcs_a[0] = a[0];
	wcs_a[1] = a[1];
      }
      
      // create a couple of points:
      the_registry_t * r = registry();
      
      the_hard_point_t * pt_a = new the_hard_point_t(wcs_a);
      r->add(pt_a);
      pt_a->set_anchor();
      establish_supporter_dependent(r, proc()->id(), pt_a->id());
      
      the_hard_point_t * pt_b = new the_hard_point_t(wcs_b);
      r->add(pt_b);
      pt_b->set_anchor();
      establish_supporter_dependent(r, proc()->id(), pt_b->id());
      
      point_pair_t * pp =
	new point_pair_t(the_point_ref_t(pt_a->id()),
			 the_point_ref_t(pt_b->id()));
      r->add(pp);
      establish_supporter_dependent(r, proc()->id(), pp->id());
      establish_supporter_dependent(r, pp->id(), proc()->xform_prim()->id());
      pp->request_regeneration();
      
      // update the selection set:
      if (&view() == view_fixed || &view() == view_mosaic)
      {
	proc()->selection_ = pt_a->id();
      }
      else if (&view() == view_moving)
      {
	proc()->selection_ = pt_b->id();
      }
      
      activate_point_pair(active_pt());
    }
    else
    {
      // cout << "picked something -- CANNOT create a new point pair" << endl;
      deactivate_point_pair(active_pt());
      proc()->selection_ = std::numeric_limits<unsigned int>::max();
    }
    return true;
  }
  else if (THE_MOUSE.l_btn().down() &&
	   THE_MOUSE.l_btn().changed() &&
	   !THE_MOUSE.l_btn().moved_while_down())
  {
    the_trail_t::milestone_t milestone;
    
    deactivate_point_pair(active_pt());
    proc()->selection_ = std::numeric_limits<unsigned int>::max();
    
    // select a point:
    the_pick_list_t curr_pick;
    if (curr_pick.pick(view(),
		       scs_pt(),
		       the_point_pick_filter_t(),
		       view().primitives()))
    {
      // cout << "picked a point -- select the point pair" << endl;
      
      // lookup the point pair that references the selected point:
      proc()->selection_ = curr_pick.front().data().id();
      active_pt()->set_anchor();
      
      activate_point_pair(active_pt());
      return true;
    }
    else
    {
      // cout << "picked noting -- selection cleared" << endl;
      doc_ui().sync_ui();
    }
  }
  else if (active_pt() != NULL &&
	   has_drag() &&
	   THE_MOUSE.verify_pressed_btns(THE_MOUSE.l_btn()) &&
	   THE_KEYBD.verify_no_pressed_keys())
  {
    the_trail_t::milestone_t milestone;
    
    if (the_input_device_eh_t::state() == THE_EH_1ST_DRAG_E)
    {
      doc_so().save_undo_record();
    }
    
    active_pt()->move(view().view_mgr(), drag_wcs());
    draw_scs_drag_vec();
    return true;
  }
  else if (active_pt() != NULL &&
	   state() == THE_EH_END_DRAG_E &&
	   THE_KEYBD.verify_no_pressed_keys())
  {
    // accept the end-drag event so that the mosaic would regenerate:
    return true;
  }
  
  // reject all other events:
  return false;
}


//----------------------------------------------------------------
// the_stos_proc_ui_t::the_stos_proc_ui_t
// 
the_stos_proc_ui_t::the_stos_proc_ui_t(the_main_window_t * doc_ui):
  QDialog(doc_ui,
	  Qt::Window |
	  Qt::WindowTitleHint |
	  Qt::WindowSystemMenuHint |
	  Qt::WindowMinMaxButtonsHint ),
  the_procedure_ui_t(*doc_ui),
  eh_(doc_ui),
  image_loader_(doc_ui),
  repaint_timer_(this)
{
  repaint_timer_.setSingleShot(true);
  
  setupUi(this);
  /*
  connect(MOSAIC_THREAD_POOL(),
	  SIGNAL(transaction_finished(the_transaction_t *)),
	  this, SLOT(rgbaDone(the_transaction_t *)));
  */
  // avoid the overhead of the thread start/stop sequence -- keep the
  // thread running while putting it to sleep when it has nothing to do:
  MOSAIC_THREAD_POOL()->set_idle_sleep_duration(true, 100000);
  
  // attach the event handler to the document views:
  std::vector<the_view_t *> & views = doc_ui_.views();
  unsigned int num_views = views.size();
  for (unsigned int i = 0; i < num_views; i++)
  {
    eh_.attach_view(views[i]);
  }
}

//----------------------------------------------------------------
// the_stos_proc_ui_t::~the_stos_proc_ui_t
// 
the_stos_proc_ui_t::~the_stos_proc_ui_t()
{
  /*
  disconnect(MOSAIC_THREAD_POOL(),
	     SIGNAL(transaction_finished(the_transaction_t *)),
	     this, SLOT(rgbaDone(the_transaction_t *)));
  */
  eh_.uninstall();
  
  std::vector<the_view_t *> & views = doc_ui_.views();
  unsigned int num_views = views.size();
  for (unsigned int i = 0; i < num_views; i++)
  {
    eh_.detach_view(views[i]);
  }
}

//----------------------------------------------------------------
// the_stos_proc_ui_t::install
// 
void
the_stos_proc_ui_t::install()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  if (installed()) return;
  
  show();
  
  // make a new procedure:
  document()->add_proc(new the_stos_proc_t());
  the_procedure_ui_t::install();
  eh_.install();
  
  // hookup the slots:
  the_main_window_t * main_window =
    dynamic_cast<the_main_window_t *>(&doc_ui_);
  
  connect(main_window->actionExportStoS, SIGNAL(triggered()),
	  this, SLOT(exportStoS()));
  connect(main_window->actionExportMosaic, SIGNAL(triggered()),
	  this, SLOT(exportMosaic()));
  connect(main_window->actionExportWarped, SIGNAL(triggered()),
	  this, SLOT(exportWarped()));
  connect(main_window->actionExportRegistered, SIGNAL(triggered()),
	  this, SLOT(exportRegistered()));
  connect(main_window->actionDelete, SIGNAL(triggered()),
	  this, SLOT(slot_edit_delete()));
  connect(main_window->actionClearAll, SIGNAL(triggered()),
	  this, SLOT(slot_clear_all()));
  connect(main_window->actionChangeFixed, SIGNAL(triggered()),
	  this, SLOT(changeFixed()));
  connect(main_window->actionChangeMoving, SIGNAL(triggered()),
	  this, SLOT(changeMoving()));
  connect(main_window->actionFlipFixed, SIGNAL(triggered()),
	  this, SLOT(flipFixed()));
  connect(main_window->actionFlipMoving, SIGNAL(triggered()),
	  this, SLOT(flipMoving()));
  connect(main_window->actionSwapImages, SIGNAL(triggered()),
	  this, SLOT(swapImages()));
  connect(&repaint_timer_, SIGNAL(timeout()),
	  this, SLOT(repaintMosaic()));
  
  doc_ui().sync_ui();
}

//----------------------------------------------------------------
// the_stos_proc_ui_t::uninstall
// 
void
the_stos_proc_ui_t::uninstall()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  if (!installed()) return;
  
  MOSAIC_THREAD_POOL()->stop();
  MOSAIC_THREAD_POOL()->wait();
  
  sync_ui();
  hide();
  
  // unhook the slots:
  the_main_window_t * main_window =
    dynamic_cast<the_main_window_t *>(&doc_ui_);
  
  disconnect(main_window->actionExportStoS, SIGNAL(triggered()),
	     this, SLOT(exportStoS()));
  disconnect(main_window->actionExportMosaic, SIGNAL(triggered()),
	     this, SLOT(exportMosaic()));
  disconnect(main_window->actionExportWarped, SIGNAL(triggered()),
	     this, SLOT(exportWarped()));
  disconnect(main_window->actionExportRegistered, SIGNAL(triggered()),
	     this, SLOT(exportRegistered()));
  disconnect(main_window->actionDelete, SIGNAL(triggered()),
	     this, SLOT(slot_edit_delete()));
  disconnect(main_window->actionClearAll, SIGNAL(triggered()),
	     this, SLOT(slot_clear_all()));
  disconnect(main_window->actionChangeFixed, SIGNAL(triggered()),
	     this, SLOT(changeFixed()));
  disconnect(main_window->actionChangeMoving, SIGNAL(triggered()),
	     this, SLOT(changeMoving()));
  disconnect(main_window->actionFlipFixed, SIGNAL(triggered()),
	     this, SLOT(flipFixed()));
  disconnect(main_window->actionFlipMoving, SIGNAL(triggered()),
	     this, SLOT(flipMoving()));
  disconnect(main_window->actionSwapImages, SIGNAL(triggered()),
	     this, SLOT(swapImages()));
  disconnect(&repaint_timer_, SIGNAL(timeout()),
	     this, SLOT(repaintMosaic()));
  
  the_procedure_ui_t::uninstall();
  eh_.uninstall();
  
  doc_ui().sync_ui();
}

//----------------------------------------------------------------
// the_stos_proc_ui_t::sync_ui
// 
void
the_stos_proc_ui_t::sync_ui()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  if (proc() == NULL) return;
  proc()->regenerate();
  
  // FIXME: write me: enable/disable context sensitive actions here:
  // File->Export, Image->Flip, etc...
}

//----------------------------------------------------------------
// the_stos_proc_ui_t::exportStoS
// 
void
the_stos_proc_ui_t::exportStoS()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  the_main_window_t * main_window =
    dynamic_cast<the_main_window_t *>(&doc_ui_);
  
  QFileInfo original(QString::fromUtf8(document()->name()));
  QString suggest_fn = original.baseName() + QString::fromUtf8(".stos");
  
  QString filename(QFileDialog::getSaveFileName
		   (this,
		    tr("Export as slice to slice registration"),
		    suggest_fn,
		    tr("slice to slice registration files (*.stos)")));
  if (filename.isEmpty()) return;
  
  std::fstream file;
  file.open(filename.toUtf8().constData(), ios::out);
  if (!file.is_open())
  {
    QString text =
      tr("Could not save to file: \"") + filename +
      QString("\".\n") +
      tr("Operation will be aborted...\n");
    
    QMessageBox::critical(main_window,
			  tr("Could not export slice to slice registration"),
			  text,
			  QMessageBox::Ok,
			  QMessageBox::NoButton,
			  QMessageBox::NoButton);
    return;
  }
  
  the_stos_proc_t * p = proc();
  stos_t<native_image_t> stos(p->fi().name_,
			      p->mi().name_,
			      the_text_t(), // fixed image mask
			      the_text_t(), // moving image mask
			      p->fi().flipped_,
			      p->mi().flipped_,
			      p->fi().spacing(0),
			      p->fi().spacing(1),
			      p->mi().spacing(0),
			      p->mi().spacing(1),
			      p->fi().size(0),
			      p->fi().size(1),
			      p->mi().size(0),
			      p->mi().size(1),
			      p->xform_prim()->forward_.GetPointer());
  stos.save(file);
  file.close();
}

//----------------------------------------------------------------
// the_stos_proc_ui_t::exportMosaic
// 
void
the_stos_proc_ui_t::exportMosaic()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  the_main_window_t * main_window =
    dynamic_cast<the_main_window_t *>(&doc_ui_);
  
  QFileInfo original(QString::fromUtf8(document()->name()));
  QString suggest_fn = original.baseName() + QString::fromUtf8(".mosaic");
  
  QString filename(QFileDialog::getSaveFileName
		   (this,
		    tr("Export as in-plane mosaic"),
		    suggest_fn,
		    tr("mosaic files (*.mosaic)")));
  if (filename.isEmpty()) return;
  
  std::fstream file;
  file.open(filename.toUtf8().data(), ios::out);
  if (!file.is_open())
  {
    QString text =
      tr("Could not save to file: \"") + filename +
      QString("\".\n") +
      tr("Operation will be aborted...\n");
    
    QMessageBox::critical(main_window,
			  tr("Could not export mosaic"),
			  text,
			  QMessageBox::Ok,
			  QMessageBox::NoButton,
			  QMessageBox::NoButton);
    return;
  }
  
  the_stos_proc_t * p = proc();
  std::list<the_text_t> images;
  images.push_back(the_text_t(p->fi().name_));
  images.push_back(the_text_t(p->mi().name_));
  std::vector<base_transform_t::Pointer> transforms(2);
  transforms[0] = stos_transform_t::New();
  transforms[1] = p->xform_prim()->forward_.GetPointer();
  
  // FIXME: why are pixel spacing and std_mask flag hard coded?
  save_mosaic<base_transform_t>(file, 1, false, images, transforms);
  file.close();
}

//----------------------------------------------------------------
// ExportImageTransaction
// 
template <typename TImage>
class ExportImageTransaction : public the_transaction_t
{
public:
  ExportImageTransaction(WarpImageTransaction<TImage> & warp,
			 const the_text_t & filename):
    warp_(warp),
    filename_(filename)
  {}
  
  // virtual:
  void execute(the_thread_interface_t * thread)
  {
    warp_.execute(thread);
    
    if (warp_.blab_)
    {
      blab(thread, tr("saving ") + filename_);
    }
    
    save<TImage>(warp_.warped_,
		 filename_,
		 false);
  }
  
  WarpImageTransaction<TImage> & warp_;
  the_text_t filename_;
};

//----------------------------------------------------------------
// the_stos_proc_ui_t::exportWarped
// 
void
the_stos_proc_ui_t::exportWarped()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  the_stos_proc_t * stos = proc();
  if (!stos->mi_ready()) return;
  
  QFileInfo original(QString::fromUtf8(stos->mi().name_));
  QString suggest_fn =
    original.baseName() + QString::fromUtf8("-warped.") + original.suffix();
  QString filename(QFileDialog::getSaveFileName
		   (this,
		    tr("Export as warped image"),
		    suggest_fn,
		    tr("Anything (*.*)")));
  if (filename.isEmpty()) return;
  
  // setup the warp:
  WarpImageTransaction<native_image_t>
    warp(stos->mi(),
	 stos->xform_prim()->forward_.GetPointer(),
	 true);
  ExportImageTransaction<native_image_t>
    transaction(warp, the_text_t(filename.toUtf8().constData()));
  
  // warp the image:
  ImageProcessingDialog tmp(this);
  tmp.thread_.add_transaction(&transaction);
  tmp.exec();
}

//----------------------------------------------------------------
// the_stos_proc_ui_t::exportRegistered
// 
void
the_stos_proc_ui_t::exportRegistered()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  the_stos_proc_t * stos = proc();
  if (!stos->mi_ready()) return;
  
  QFileInfo original(QString(stos->mi().name_));
  QString suggest_fn =
    original.baseName() + QString::fromUtf8("-registered.") + original.suffix();
  QString filename(QFileDialog::getSaveFileName
		   (this,
		    tr("Export as registered image"),
		    suggest_fn,
		    tr("Anything (*.*)")));
  if (filename.isEmpty()) return;
  
  // setup the warp:
  image_data_t & fi = stos->fi();
  image_data_t & mi = stos->mi();
  
  pnt2d_t fi_min;
  pnt2d_t fi_max;
  {
    const image_tile_generator_t & tile = fi.tile_generator_;
    fi_min = pnt2d(tile.bbox_.min_[0], tile.bbox_.min_[1]);
    fi_max = pnt2d(tile.bbox_.max_[0], tile.bbox_.max_[1]);
  }
  
  vec2d_t spacing;
  {
    const image_tile_generator_t & tile = mi.tile_generator_;
    spacing[0] = tile.spacing_x_ / double(mi.shrink_factor_);
    spacing[1] = tile.spacing_y_ / double(mi.shrink_factor_);
  }
  
  WarpImageTransaction<native_image_t>
    warp(mi.name_,
	 mi.flipped_,
	 stos->xform_prim()->forward_.GetPointer(),
	 fi_min,
	 fi_max,
	 spacing,
	 true);
  ExportImageTransaction<native_image_t>
    transaction(warp, the_text_t(filename.toUtf8().constData()));
  
  // warp the image:
  ImageProcessingDialog tmp(this);
  tmp.thread_.add_transaction(&transaction);
  tmp.exec();
}

//----------------------------------------------------------------
// the_stos_proc_ui_t::changeFixed
// 
void
the_stos_proc_ui_t::changeFixed()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  the_main_window_t * main_window =
    dynamic_cast<the_main_window_t *>(&doc_ui());
  
  // disable auto save:
  QTimer & autosave = main_window->autosave();
  bool autosave_on = autosave.isActive();
  autosave.stop();
  
  MOSAIC_THREAD_POOL()->stop();
  MOSAIC_THREAD_POOL()->wait();
  
  int r = image_loader_.exec(proc()->fi());
  if (r == QDialog::Accepted)
  {
    doc_so().big_changes();
    doc_so().set_filename("");
    proc()->fi_prim()->request_regeneration();
    proc()->setup_mosaic_pyramid();
    proc()->regenerate();
    main_window->view(0)->restore_orientation();
    main_window->sync_ui();
  }
  
  // restore autosave:
  if (autosave_on)
  {
    main_window->file_auto_save(true);
  }
}

//----------------------------------------------------------------
// the_stos_proc_ui_t::changeMoving
// 
void
the_stos_proc_ui_t::changeMoving()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  the_main_window_t * main_window =
    dynamic_cast<the_main_window_t *>(&doc_ui());
  
  // disable auto save:
  QTimer & autosave = main_window->autosave();
  bool autosave_on = autosave.isActive();
  autosave.stop();
  
  MOSAIC_THREAD_POOL()->stop();
  MOSAIC_THREAD_POOL()->wait();
  
  int r = image_loader_.exec(proc()->mi());
  if (r == QDialog::Accepted)
  {
    doc_so().big_changes();
    doc_so().set_filename("");
    proc()->mi_prim()->request_regeneration();
    proc()->setup_mosaic_pyramid();
    proc()->regenerate();
    main_window->view(1)->restore_orientation();
    main_window->sync_ui();
  }
  
  // restore autosave:
  if (autosave_on)
  {
    main_window->file_auto_save(true);
  }
}

//----------------------------------------------------------------
// the_stos_proc_ui_t::flipFixed
// 
void
the_stos_proc_ui_t::flipFixed()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  if (!proc()->fi_ready()) return;
  
  MOSAIC_THREAD_POOL()->stop();
  MOSAIC_THREAD_POOL()->wait();
  
  ImageProcessingDialog tmp(this);
  FlipImageTransaction<native_image_t> transaction(proc()->fi());
  tmp.thread_.add_transaction(&transaction);
  int r = tmp.exec();
  if (r == QDialog::Accepted && transaction.done())
  {
    doc_so().big_changes();
    proc()->fi_prim()->request_regeneration();
    proc()->regenerate();
    sync_ui();
    doc_ui().sync_ui();
  }
}

//----------------------------------------------------------------
// the_stos_proc_ui_t::flipMoving
// 
void
the_stos_proc_ui_t::flipMoving()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  if (!proc()->mi_ready()) return;
  
  MOSAIC_THREAD_POOL()->stop();
  MOSAIC_THREAD_POOL()->wait();
  
  ImageProcessingDialog tmp(this);
  FlipImageTransaction<native_image_t> transaction(proc()->mi());
  tmp.thread_.add_transaction(&transaction);
  int r = tmp.exec();
  if (r == QDialog::Accepted && transaction.done())
  {
    doc_so().big_changes();
    proc()->mi_prim()->request_regeneration();
    proc()->regenerate();
    sync_ui();
    doc_ui().sync_ui();
  }
}

//----------------------------------------------------------------
// the_stos_proc_ui_t::swapImages
// 
void
the_stos_proc_ui_t::swapImages()
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  doc_so().save_undo_record();
  proc()->swap_images();
  proc()->regenerate();
  
  the_main_window_t * main_window =
    dynamic_cast<the_main_window_t *>(&doc_ui());
  main_window->view(0)->restore_orientation();
  main_window->view(1)->restore_orientation();
  main_window->sync_ui();
}

//----------------------------------------------------------------
// the_stos_proc_ui_t::rgbaDone
// 
void
the_stos_proc_ui_t::rgbaDone(the_transaction_t * t)
{
  the_stos_proc_t * stos = proc();
  if (t->done() && stos != NULL)
  {
    mosaic_transaction_t * transaction =
      dynamic_cast<mosaic_transaction_t *>(t);
    
    if (transaction != NULL)
    {
      // cerr << "rgbaDone: " << transaction->cache_.id_ << endl;
      
      // update the procedure:
      stos->mosaic_ready(transaction->cache_.id_);
      
      // update the mosaic view:
      the_view_t * mosaic_view = doc_ui().view(2);
      
      if (stos->mosaic_prim()->first_time_regenerating_)
      {
	stos->mosaic_prim()->first_time_regenerating_ = false;
	mosaic_view->restore_orientation();
      }
      
      if (!repaint_timer_.isActive())
      {
	// buffer the repaints in order to avoid the jerky
	// display update and flashing due to the multi-resolution
	// pixel scale difference:
	repaint_timer_.start(41);
      }
    }
  }
  
  delete t;
}

//----------------------------------------------------------------
// the_stos_proc_ui_t::closeEvent
// 
void
the_stos_proc_ui_t::closeEvent(QCloseEvent * e)
{
  // setup the milestone markers:
  the_trail_t::milestone_t milestone;
  
  e->ignore();
  the_main_window_t * main_window =
    dynamic_cast<the_main_window_t *>(&doc_ui());
  main_window->toggle_stos_ui(false);
}

//----------------------------------------------------------------
// the_stos_proc_ui_t::customEvent
// 
void
the_stos_proc_ui_t::customEvent(QEvent * e)
{
  if (e->type() == QEvent::User + 1)
  {
    e->accept();
    rgba_done_event_t * event = (rgba_done_event_t *)(e);
    the_transaction_t * t = NULL;
    {
      transaction_payload_t & payload = event->payload_;
      QMutexLocker locker(&payload.mutex_);
      t = payload.get_payload();
    }
    rgbaDone(t);
  }
}

//----------------------------------------------------------------
// the_stos_proc_ui_t::repaintMosaic
// 
void
the_stos_proc_ui_t::repaintMosaic()
{
  the_view_t * mosaic_view = doc_ui().view(2);
  // mosaic_view->gl_make_current();
  mosaic_view->refresh();
}

//----------------------------------------------------------------
// the_stos_proc_ui_t::proc
// 
the_stos_proc_t *
the_stos_proc_ui_t::proc()
{
  the_document_t * d = document();
  if (d == NULL) return NULL;
  return d->active_procedure<the_stos_proc_t>();
}
