/*  
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : the_iv_proc_ui.cxx
// Author       : Pavel Aleksandrovich Koshevoy
// Created      : Sun Sep 24 16:54:00 MDT 2006
// Copyright    : (C) 2006-2008 University of Utah
// License      : GPLv2
// Description  : UI for the image viewer procedure.

// local includes:
#include "the_iv_proc_ui.hxx"
#include "the_iv_proc.hxx"
#include "the_main_window.hxx"

// the includes:
#include <math/the_bbox.hxx>
#include <image/image_tile.hxx>
#include <image/image_tile_generator.hxx>
#include <image/texture_data.hxx>
#include <image/texture.hxx>
#include <utils/the_exception.hxx>
#include <thread/the_terminator.hxx>
#include <Qt/the_qt_thread.hxx>
#include <Qt/the_qt_trail.hxx>
#include <Qt/qimage_pixel_converter.hxx>
#include <opengl/image_tile_dl_elem.hxx>
#include <opengl/the_disp_list.hxx>

// Qt includes:
#include <QAction>
#include <QFileDialog>
#include <QImageReader>

// Boost includes:
#include <boost/shared_ptr.hpp>


//----------------------------------------------------------------
// IMAGE_READER_THREAD
// 
// the pixmap generating thread:
// 
the_qt_thread_t *
IMAGE_READER_THREAD()
{
  static the_qt_thread_t * thread_ = NULL;
  if (thread_ == NULL)
  {
    thread_ = new the_qt_thread_t();
  }
  
  return thread_;
}


//----------------------------------------------------------------
// ImageReaderTransaction
// 
class ImageReaderTransaction : public the_transaction_t
{
public:
  ImageReaderTransaction(const the_text_t & fn,
			 const GLenum & filter,
			 unsigned int max_texture):
    fn_(fn),
    filter_(filter),
    max_texture_(max_texture)
  {}
  
  // virtual:
  void execute(the_thread_interface_t * thread)
  {
    if (fn_.is_empty()) throw the_exception_t("empty filename");
    
    QFileInfo fi(fn_.text());
    the_text_t fn_short(qPrintable(fi.fileName()));
    
    blab(thread, tr("\nloading ") + fn_short);
    QImage image(fn_.text());
    bool ok = !image.isNull();
    if (!ok)
    {
      throw the_exception_t("could not load the image");
    }
    
    blab(thread, tr("calculating texture tile layout"));
    image_tile_generator_t tile_generator;
    tile_generator.layout(image.width(), image.height());
    
    w_ = tile_generator.w_;
    h_ = tile_generator.h_;
    
    // gather some stats about the image:
    size_t src_bytes_per_pixel = 0;
    size_t dst_bytes_per_pixel = 0;
    bool has_alpha = false;
    bool is_grayscale = true;
    bool must_convert = false;
    
    switch (image.format())
    {
      case QImage::Format_MonoLSB:
      case QImage::Format_Mono:
	// FIXME: this is wasteful:
	image = image.convertToFormat(QImage::Format_Indexed8);
	must_convert = true;
	is_grayscale = true;
	has_alpha = false;
	src_bytes_per_pixel = 1;
	break;
	
      case QImage::Format_Indexed8:
      {
	// check the color table -- if each element is
	// equal to its index than we can do the optimized
	// thing and copy the data directly without conversion:
	QVector<QRgb> palette = image.colorTable();
	for (int i = 0; i < palette.size(); i++)
	{
	  QRgb rgba = palette[i];
	  
	  int r = qRed(rgba);
	  int g = qGreen(rgba);
	  int b = qBlue(rgba);
	  int a = qAlpha(rgba);
	  
#if 0
	  is_grayscale = is_grayscale && (r == g && g == b);
	  must_convert = must_convert || i != r || i != g || i != b;
#else
	  is_grayscale = is_grayscale && (r == i && g == i && b == i);
#endif
	  
	  has_alpha = has_alpha || (a != 0xFF);
	}
	
	must_convert = must_convert || !is_grayscale || has_alpha;
	src_bytes_per_pixel = 1;
      }
      break;

      case QImage::Format_RGB32:
      case QImage::Format_ARGB32:
      case QImage::Format_ARGB32_Premultiplied:
      {
	is_grayscale = image.isGrayscale();
	has_alpha = image.hasAlphaChannel();
	must_convert = is_grayscale;
	src_bytes_per_pixel = 4;
      }
      break;
      
      default:
	throw the_exception_t("image format not supported");
    }
    
    GLenum data_type = 0;
    GLenum format_internal = 0;
    GLenum format = 0;
    const unsigned char * src = image.bits();
    
#if 0
    // FIXME: force a particular format:
    must_convert = true;
    is_grayscale = true;
    has_alpha = true;
#endif
    
    if (!must_convert)
    {
      if (is_grayscale)
      {
	data_type = GL_UNSIGNED_BYTE;
	format_internal = GL_LUMINANCE;
	format = GL_LUMINANCE;
	blab(thread, tr("copying image luminance data into a texture"));
      }
      else
      {
	data_type = GL_UNSIGNED_INT_8_8_8_8_REV;
	format_internal = GL_RGBA8;
	format = GL_BGRA_EXT;
	blab(thread, tr("copying image RGBA data into a texture"));
      }
      
      dst_bytes_per_pixel = src_bytes_per_pixel;
      tile_generator.convert_and_pad(src,
				     4,
				     src_bytes_per_pixel,
				     dst_bytes_per_pixel,
				     copy_pixels_t());
    }
    else
    {
      if (is_grayscale)
      {
	data_type = GL_UNSIGNED_BYTE;
	if (has_alpha)
	{
	  format = GL_LUMINANCE_ALPHA;
	  format_internal = GL_LUMINANCE_ALPHA;
	  dst_bytes_per_pixel = 2;
	  blab(thread, tr("converting image data into a "
			  "luminance/alpha texture"));
	}
	else
	{
	  format = GL_LUMINANCE;
	  format_internal = GL_LUMINANCE;
	  dst_bytes_per_pixel = 1;
	  blab(thread, tr("converting image data into a luminance texture"));
	}
      }
      else
      {
	if (has_alpha)
	{
	  data_type = GL_UNSIGNED_BYTE;
	  format_internal = GL_RGBA8;
	  format = GL_RGBA;
	  dst_bytes_per_pixel = 4;
	  blab(thread, tr("converting image data into an RGBA texture"));
	}
	else
	{
	  data_type = GL_UNSIGNED_BYTE;
	  format_internal = GL_RGB8;
	  format = GL_RGB;
	  dst_bytes_per_pixel = 3;
	  blab(thread, tr("converting image data into an RGB texture"));
	}
      }
      
      tile_generator.
	convert_and_pad(src,
			4,
			src_bytes_per_pixel,
			dst_bytes_per_pixel,
			qimage_pixel_converter_t(image,
						 src_bytes_per_pixel,
						 dst_bytes_per_pixel,
						 format));
    }
    
    tile_generator.make_tiles(data_type,
			      format_internal,
			      format,
			      max_texture_);
    
    std::auto_ptr<image_tile_dl_elem_t> new_img
      (new image_tile_dl_elem_t(tile_generator, filter_, filter_));
    img_ = new_img;
    
    the_text_t short_info = (the_text_t("width: ") +
			     the_text_t::number(image.width()) +
			     the_text_t(", height: ") +
			     the_text_t::number(image.height()));
    blab(thread, short_info);
  }
  
  // transaction parameters:
  the_text_t fn_;
  GLenum filter_;
  unsigned int max_texture_;
  
  // transaction result:
  unsigned int w_;
  unsigned int h_;
  std::auto_ptr<image_tile_dl_elem_t> img_;
};


//----------------------------------------------------------------
// get_image_filters
// 
QString
get_image_filters()
{
  QList<QByteArray> supported = QImageReader::supportedImageFormats();
  the_dynamic_array_t<QString> format;
  the_dynamic_array_t<QString> extension;
  for(QList<QByteArray>::const_iterator i = supported.begin();
      i != supported.end(); i++)
  {
    QString f(*i);
    format.append(f);
    extension.append(f.toLower());
  }
  
  QString filter = "Images (";
  for (unsigned int i = 0; i < format.size(); i++)
  {
    filter += QString("*.") + extension[i] + QString(" ");
  }
  filter += ")";
  return filter;
}


//----------------------------------------------------------------
// get_image_filters
// 
void
get_image_filters(QStringList & filters, bool decorated)
{
  QList<QByteArray> supported = QImageReader::supportedImageFormats();
  the_dynamic_array_t<QString> format;
  the_dynamic_array_t<QString> extension;
  for(QList<QByteArray>::const_iterator i = supported.begin();
      i != supported.end(); i++)
  {
    QString f(*i);
    format.append(f);
    extension.append(f.toLower());
  }
  
  for (unsigned int i = 0; i < format.size(); i++)
  {
    QString filter;
    if (decorated)
    {
      filter = format[i] + " images (*." + extension[i] + ")";
    }
    else
    {
      filter = QString("*.") + extension[i];
    }
    
    filters << filter;
  }
}


//----------------------------------------------------------------
// the_iv_proc_eh_t::is_modifier_active_locally
// 
bool
the_iv_proc_eh_t::is_modifier_active_locally(the_eh_modifier_t mod) const
{
#if 0 
  cout << "--------------------------------------------------------" << endl;
  ::dump(THE_MOUSE);
  ::dump(THE_KEYBD);
#endif
  
  switch (mod)
  {
    case THE_EH_MOD_VIEW_ZOOM_E:
      return ((state() != THE_EH_NOT_READY_E) &&
	      THE_MOUSE.verify_pressed_btns(THE_MOUSE.m_btn()));
      
    case THE_EH_MOD_VIEW_PAN_E:
      return ((state() != THE_EH_NOT_READY_E) &&
	      THE_MOUSE.verify_pressed_btns(THE_MOUSE.l_btn()));
      
    default:
      break;
  }
  
  return the_input_device_eh_t::is_modifier_active_locally(mod);
}

//----------------------------------------------------------------
// the_iv_proc_eh_t::processor
// 
bool
the_iv_proc_eh_t::processor(the_input_device_eh_t::the_event_type_t et)
{
  the_main_window_t * main_window = 
    dynamic_cast<the_main_window_t *>(&doc_ui());
  
  if (et == THE_MOUSE_EVENT_E)
  {
    if (THE_MOUSE.l_btn().double_click())
    {
      main_window->actionFullScreen->trigger();
      return true;
    }
  }
  else if (et == THE_KEYBD_EVENT_E)
  {
    const the_keybd_key_t & key_escape = THE_KEYBD.key(Qt::Key_Escape);
    if (key_escape.changed() && key_escape.up())
    {
      if (main_window->isFullScreen())
      {
	main_window->actionFullScreen->trigger();
	return true;
      }
    }
    
    const the_keybd_key_t & key_space = THE_KEYBD.key(Qt::Key_Space);
    if (key_space.changed() && key_space.up())
    {
      main_window->toggleSlideShow();
      return true;
    }
  }
  
  return false;
}


//----------------------------------------------------------------
// the_iv_proc_ui_t::the_iv_proc_ui_t
// 
the_iv_proc_ui_t::the_iv_proc_ui_t(the_main_window_t * doc_ui):
  QObject(doc_ui),
  the_procedure_ui_t(*doc_ui),
  eh_(doc_ui),
  file_dialog_(new QFileDialog(doc_ui))
{
  // attach the event handler to the document views:
  std::vector<the_view_t *> & views = doc_ui_.views();
  unsigned int num_views = views.size();
  for (unsigned int i = 0; i < num_views; i++)
  {
    eh_.attach_view(views[i]);
  }
  
  // setup the file dialog filter:
  QStringList filters;
  filters << get_image_filters();
  get_image_filters(filters, true);
  filters << QString("Anything (*.*)");
  
  file_dialog_->setFileMode(QFileDialog::ExistingFile);
  file_dialog_->setFilters(filters);
  file_dialog_->setViewMode(QFileDialog::Detail);
  file_dialog_->setDirectory(QDir::current());
  
  connect(IMAGE_READER_THREAD(),
	  SIGNAL(transaction_finished(the_transaction_t *)),
	  this, SLOT(transactionFinished(the_transaction_t *)));
  
  connect(IMAGE_READER_THREAD(),
	  SIGNAL(status_update(const QString &)),
	  doc_ui->statusBar(),
	  SLOT(showMessage(const QString &)));
}

//----------------------------------------------------------------
// the_iv_proc_ui_t::~the_iv_proc_ui_t
// 
the_iv_proc_ui_t::~the_iv_proc_ui_t()
{
  IMAGE_READER_THREAD()->stop();
  
  eh_.uninstall();
  
  std::vector<the_view_t *> & views = doc_ui_.views();
  unsigned int num_views = views.size();
  for (unsigned int i = 0; i < num_views; i++)
  {
    eh_.detach_view(views[i]);
  }
  
  delete file_dialog_;
  file_dialog_ = NULL;
}

//----------------------------------------------------------------
// the_iv_proc_ui_t::proc
// 
the_iv_proc_t *
the_iv_proc_ui_t::proc()
{
  the_document_t * d = document();
  if (d == NULL) return NULL;
  return d->active_procedure<the_iv_proc_t>();
}

//----------------------------------------------------------------
// the_iv_proc_ui_t::install
// 
void
the_iv_proc_ui_t::install()
{
  if (installed()) return;
  
  // make a new procedure:
  document()->add_proc(new the_iv_proc_t());
  the_procedure_ui_t::install();
  eh_.install();
  
  doc_ui().sync_ui();
}

//----------------------------------------------------------------
// the_iv_proc_ui_t::uninstall
// 
void
the_iv_proc_ui_t::uninstall()
{
  if (!installed()) return;
  
  sync_ui();
  
  the_procedure_ui_t::uninstall();
  eh_.uninstall();
  
  doc_ui().sync_ui();
}

//----------------------------------------------------------------
// the_iv_proc_ui_t::file_load
// 
void
the_iv_proc_ui_t::file_load(const QString & fn)
{
  // shortcuts:
  GLenum filter = prefsAllowLinearFiltering() ? GL_LINEAR : GL_NEAREST;
  unsigned int max_texture = OpenGL(0).max_texture_;
  ImageReaderTransaction * t =
    new ImageReaderTransaction(qPrintable(fn), filter, max_texture);
  
  IMAGE_READER_THREAD()->flush();
  IMAGE_READER_THREAD()->the_thread_interface_t::start(t);
}

//----------------------------------------------------------------
// the_iv_proc_ui_t::transactionFinished
// 
void
the_iv_proc_ui_t::transactionFinished(the_transaction_t * trans)
{
  // shortcut:
  the_main_window_t * main_window =
    dynamic_cast<the_main_window_t *>(&doc_ui());
  
  if (trans->done())
  {
    ImageReaderTransaction * t = static_cast<ImageReaderTransaction *>(trans);
    
    // update the procedure:
    the_iv_proc_t * p = proc();
    p->dl_->clear();
    p->dl_->push_back(t->img_.release());
    
    bool same_size = p->w_ == t->w_ && p->h_ == t->h_;
    p->w_ = t->w_;
    p->h_ = t->h_;
    
    if (!main_window->isFullScreen() && prefsShouldFitToImage())
    {
      // resize the window:
      main_window->fitToImage();
    }
    
    // change current working directory:
    QFileInfo fileinfo(t->fn_.text());
    QDir::setCurrent(fileinfo.dir().absolutePath());
    
    // change the window title:
    main_window->setWindowTitle(fileinfo.fileName());
    
    // update the ui:
    if (!same_size)
    {
      the_view_t * view = main_window->view(0);
      view->restore_orientation();
    }
    main_window->sync_ui();
  }
  
  delete trans;
}

//----------------------------------------------------------------
// the_iv_proc_ui_t::file_load
// 
void
the_iv_proc_ui_t::file_load()
{
  QString fn;
#if 1
  if (!file_dialog_->exec()) return;
  
  QStringList fn_list = file_dialog_->selectedFiles();
  if (!fn_list.empty())
  {
    fn = fn_list.back();
  }
#else
  fn =
    QFileDialog::getOpenFileName
    (main_window,	// parent
     QString(),		// caption
     QString(),		// dir
     "Image (*.tif *.png *.jpg *.bmp *.xpm *.gif);;"
     "Anything (*.*)");
#endif
  
  file_load(fn);
}
