/*  
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// File         : the_main_window.cxx
// Author       : Pavel Aleksandrovich Koshevoy
// Created      : Sun Sep 24 16:41:00 MDT 2006
// Copyright    : (C) 2006-2008 University of Utah
// License      : GPLv2
// Description  : main application window widget

// system includes:
#include <vector>

// Qt includes:
#include <QWidget>
#include <QTabWidget>
#include <QApplication>
#include <QMessageBox>
#include <QInputDialog>
#include <QFileDialog>
#include <QAction>
#include <QPixmap>
#include <QGridLayout>
#include <QList>
#include <QDateTime>
#include <QListView>
#include <QStringList>
#include <QFileIconProvider>
#include <QIconEngine>
#include <QIcon>
#include <QDesktopWidget>

// local includes:
#include "the_main_window.hxx"
#include "the_iv_proc_ui.hxx"
#include "the_iv_proc.hxx"
#include "the_thumbnails.hxx"

// the includes:
#include <eh/the_view_mgr_eh.hxx>
#include <math/the_color.hxx>
#include <opengl/the_appearance.hxx>
#include <Qt/the_qt_view.hxx>
#include <Qt/the_qt_trail.hxx>

// forward declarations:
extern void get_image_filters(QStringList & filters, bool decorated);


//----------------------------------------------------------------
// FileSysModel::trigger_update
// 
void
FileSysModel::trigger_update(const QModelIndex & index)
{
  // cerr << "trigger update: " << qPrintable(fileName(index)) << endl;
  emit dataChanged(index, index);
}


//----------------------------------------------------------------
// KeyPressObserver::KeyPressObserver
// 
KeyPressObserver::KeyPressObserver(QObject * parent):
  QObject(parent)
{}

//----------------------------------------------------------------
// KeyPressObserver::eventFilter
// 
bool
KeyPressObserver::eventFilter(QObject * obj, QEvent * event)
{
  if (event->type() == QEvent::KeyPress)
  {
    QKeyEvent * ke = static_cast<QKeyEvent *>(event);
    if (ke->key() == Qt::Key_Enter ||
	ke->key() == Qt::Key_Return)
    {
      the_main_window_t * main_window =
	dynamic_cast<the_main_window_t *>(the_document_ui_t::doc_ui());
      main_window->go_to_selected();
      return true;
    }
  }
  
  // standard event processing
  return QObject::eventFilter(obj, event);
}


//----------------------------------------------------------------
// get_index
// 
static QModelIndex
get_index(const QDirModel * model,
	  const QModelIndex & root,
	  const QModelIndex & curr,
	  const int & delta_row,
	  const int & offset = 0)
{
  int rows = model->rowCount(root);
  for (int i = curr.row() + offset; i > -1 && i < rows; i = i + delta_row)
  {
    QModelIndex next = model->index(i, curr.column(), root);
    if (next.isValid() && !model->isDir(next))
    {
      return next;
      break;
    }
  }
  
  return QModelIndex();
}

//----------------------------------------------------------------
// get_sibling
// 
static inline QModelIndex
get_sibling(const QDirModel * model,
	    const QModelIndex & root,
	    const QModelIndex & curr,
	    const int & delta_row)
{
  return get_index(model, root, curr, delta_row, delta_row);
}

//----------------------------------------------------------------
// enter_dir
// 
// return the previous directory
// 
static QString
enter_dir(QListView * iconList,
	  QDirModel * filesys_,
	  const QModelIndex & root)
{
  // if we are changing directories, than there is no need to
  // keep working on those thumbnails:
  THUMBNAIL_THREAD()->stop();
  
  QString prev = filesys_->filePath(iconList->rootIndex());
  iconList->setRootIndex(root);
  filesys_->refresh(root);
  
  // reset the selection:
  QModelIndex curr = filesys_->index(0, 0, root);
  curr = get_index(filesys_, root, curr, 1, 0);
  if (curr.isValid())
  {
    QItemSelectionModel * sel = iconList->selectionModel();
    sel->clear();
    sel->setCurrentIndex(curr, QItemSelectionModel::Select);
  }
  
  return prev;
}


//----------------------------------------------------------------
// orientation
// 
static the_view_mgr_orientation_t orientation[] =
{
  THE_BOTTOM_VIEW_E
};


//----------------------------------------------------------------
// the_main_window_t::the_main_window_t
// 
the_main_window_t::the_main_window_t():
  QMainWindow(NULL, 0),
  the_document_ui_t("iv/20070111"),
  iv_proc_ui_(NULL),
  preferences_ui_(this)
{
  the_qt_view_t * shared =
    new the_qt_view_t(this, "shared", NULL, THE_BOTTOM_VIEW_E);
  
  the_view_t * qt_view[] = {
    new the_qt_view_t(this, "", shared, orientation[0])
  };
  setup(shared, qt_view, 1);
  
  // disable 3D rotation:
#ifndef DEBUG_TEXTURES
  view_mgr_eh_[0]->allow_spin(false);
#endif
  
  setupUi(this);
  
  // signals and slots connections:
  actionUp->setShortcut(Qt::ALT + Qt::Key_Up);
  actionBack->setShortcut(Qt::ALT + Qt::Key_Left);
  actionForward->setShortcut(Qt::ALT + Qt::Key_Right);
  
  actionPrev->setShortcut(Qt::Key_PageUp);
  actionNext->setShortcut(Qt::Key_PageDown);
  
  actionUp->setShortcutContext(Qt::ApplicationShortcut);
  actionBack->setShortcutContext(Qt::ApplicationShortcut);
  actionForward->setShortcutContext(Qt::ApplicationShortcut);
  
  actionQuit->setShortcutContext(Qt::ApplicationShortcut);
  actionPrev->setShortcutContext(Qt::ApplicationShortcut);
  actionNext->setShortcutContext(Qt::ApplicationShortcut);
  
  actionFullScreen->setShortcutContext(Qt::ApplicationShortcut);
  actionShrinkWrap->setShortcutContext(Qt::ApplicationShortcut);
  actionRestore->setShortcutContext(Qt::ApplicationShortcut);
  
  connect(actionQuit, SIGNAL(triggered()),
	  this, SLOT(file_exit()));
  connect(actionPreferences, SIGNAL(triggered()),
	  this, SLOT(edit_preferences()));
  connect(actionUp, SIGNAL(triggered()),
	  this, SLOT(go_up()));
  connect(actionBack, SIGNAL(triggered()),
	  this, SLOT(go_back()));
  connect(actionForward, SIGNAL(triggered()),
	  this, SLOT(go_forward()));
  connect(addressText, SIGNAL(returnPressed()),
	  this, SLOT(go_to_address()));
  connect(iconList, SIGNAL(activated(QModelIndex)),
	  this, SLOT(enterDirectory(QModelIndex)));
  connect(&timer_, SIGNAL(timeout()),
	  this, SLOT(slot_next()));
  
  connect(actionPrev, SIGNAL(triggered()),
	  this, SLOT(slot_prev()));
  connect(actionNext, SIGNAL(triggered()),
	  this, SLOT(slot_next()));
  connect(actionPlay, SIGNAL(triggered()),
	  this, SLOT(slot_play()));
  connect(actionStop, SIGNAL(triggered()),
	  this, SLOT(slot_stop()));
  
  connect(actionFullScreen, SIGNAL(triggered(bool)),
	  this, SLOT(full_screen(bool)));
  connect(actionShrinkWrap, SIGNAL(triggered()),
	  this, SLOT(fitToImage()));
  connect(actionRestore, SIGNAL(triggered()),
	  this, SLOT(orientation_restore()));
  
  connect(actionAbout, SIGNAL(triggered()),
	  this, SLOT(helpAbout()));
  
  // this is a workaround for the OSX (where activated is not emitted
  // when the user presses the enter key):
  iconList->installEventFilter(new KeyPressObserver(this));
  
  // add the OpenGL view to the UI:
  QGridLayout * layout = new QGridLayout(frameFixed);
  layout->setSpacing(0);
  layout->setMargin(0);
  view<the_qt_view_t>(0)->setParent(frameFixed);
  layout->addWidget(view<the_qt_view_t>(0), 0, 0);
  view<the_qt_view_t>(0)->show();
  
  // setup the list view:
  QStringList filters;
  get_image_filters(filters, false);
  
  filesys_ = new FileSysModel();
  filesys_->setFilter(QDir::Drives |
		      QDir::AllDirs |
		      QDir::Files |
		      QDir::Hidden |
		      QDir::CaseSensitive |
		      QDir::NoDotAndDotDot);
  filesys_->setNameFilters(filters);
  QDir::SortFlags sort = QDir::SortFlags(QDir::Name |
					 QDir::LocaleAware |
					 QDir::IgnoreCase |
					 QDir::DirsFirst);
  filesys_->setSorting(sort);
  filesys_->setReadOnly(true);
  filesys_->setLazyChildCount(true);
  
  ThumbnailProvider * provider = new ThumbnailProvider();
  connect(THUMBNAIL_THREAD(),
	  SIGNAL(transaction_finished(the_transaction_t *)),
	  this, SLOT(thumbnailReady(the_transaction_t *)));
  
  filesys_->setIconProvider(provider);
  // iconList->setResizeMode(QListView::Adjust);
  iconList->setModel(filesys_);
  iconList->setDragEnabled(true);
  
  // initialize the document:
  doc_so().new_document(new the_document_t("iv"));
  
  // setup the iv procedure UI:
  iv_proc_ui_ = new the_iv_proc_ui_t(this);
  iv_proc_ui_->install();
  
  // update the icon list:
  enter_dir(filesys_->index(QDir::currentPath()));
  
  // initialize the UI:
  sync_ui();
}

//----------------------------------------------------------------
// the_main_window_t::~the_main_window_t
// 
the_main_window_t::~the_main_window_t()
{
  assert(iv_proc_ui_ == NULL);
}

//----------------------------------------------------------------
// the_main_window_t::thumbnailReady
// 
void
the_main_window_t::thumbnailReady(the_transaction_t * t)
{
  if (t->done())
  {
    ThumbnailTransaction * trans = static_cast<ThumbnailTransaction *>(t);
    QModelIndex i = filesys_->index(trans->thumb_.info().absoluteFilePath());
    QModelIndex parent = i.parent();
    
    // avoid unnecessary updates:
    if (parent == iconList->rootIndex())
    {
      filesys_->trigger_update(i);
      // iconList->repaint();
      // QCoreApplication::processEvents();
    }
  }
  
  delete t;
}

//----------------------------------------------------------------
// the_main_window_t::file_exit
// 
void
the_main_window_t::file_exit()
{
  // stop trail playback:
  THE_TRAIL.stop();
  
  // shut down the UI:
  iv_proc_ui_->uninstall();
  delete iv_proc_ui_;
  iv_proc_ui_ = NULL;
  
  // close the document:
  doc_so().close_document();
  
  // done:
  qApp->quit();
}

//----------------------------------------------------------------
// the_main_window_t::edit_preferences
// 
void
the_main_window_t::edit_preferences()
{
  int r = preferences_ui_.exec();
  if (r == QDialog::Accepted)
  {
    document()->active_procedure()->request_regeneration();
    sync_ui();
  }
}

//----------------------------------------------------------------
// get_supported_ext
// 
the_dynamic_array_t<QString>
get_supported_ext()
{
  the_dynamic_array_t<QString> ext;
  
  QList<QByteArray> supported = QImageReader::supportedImageFormats();
  for (QList<QByteArray>::const_iterator i = supported.begin();
       i != supported.end(); ++i)
  {
    QString f(*i);
    ext.append(f.toLower());
    // cerr << qPrintable(f) << endl;
  }
  
  return ext;
}

//----------------------------------------------------------------
// is_valid_path
// 
static bool
is_valid_path(const QString & path)
{
  QFileInfo fileinfo(path);
  if (fileinfo.isDir()) return true;
  if (!fileinfo.isFile()) return false;
  
  // make sure it's an image file:
  static const the_dynamic_array_t<QString> supported = get_supported_ext();
  const QString suffix = fileinfo.suffix().toLower();
  
  for (unsigned int i = 0; i < supported.size(); i++)
  {
    if (suffix == supported[i])
    {
      return true;
    }
  }
  
  return false;
}

//----------------------------------------------------------------
// the_main_window_t::go_here
// 
void
the_main_window_t::go_here(const QString & path)
{
  if (!is_valid_path(path)) return;
  
  QFileInfo fileinfo(path);
  if (fileinfo.isDir())
  {
    QModelIndex root = filesys_->index(path);
    enterDirectory(root);
    return;
  }
  
  // enter that directory, select that file:
  QString dir = fileinfo.dir().absolutePath();
  QModelIndex root = filesys_->index(dir);
  enterDirectory(root);
  
  QItemSelectionModel * sel = iconList->selectionModel();
  sel->clear();
  QModelIndex idx = filesys_->index(path);
  sel->setCurrentIndex(idx, QItemSelectionModel::Select);
  
  enterDirectory(idx);
}

//----------------------------------------------------------------
// the_main_window_t::go_up
// 
void
the_main_window_t::go_up()
{
  QModelIndex root = iconList->rootIndex().parent();
  if (!root.isValid() || filesys_->isDir(root))
  {
    enterDirectory(root);
  }
}

//----------------------------------------------------------------
// the_main_window_t::go_back
// 
void
the_main_window_t::go_back()
{
  if (history_[0].empty()) return;
  
  QFileInfo info(history_[0].front());
  if (info.isDir())
  {
    QString curr = enter_dir(filesys_->index(remove_head(history_[0])));
    history_[1].push_front(curr);
    sync_ui();
  }
}

//----------------------------------------------------------------
// the_main_window_t::go_forward
// 
void
the_main_window_t::go_forward()
{
  if (history_[1].empty()) return;
  
  QFileInfo info(history_[1].front());
  if (info.isDir())
  {
    QString curr = enter_dir(filesys_->index(remove_head(history_[1])));
    history_[0].push_front(curr);
    sync_ui();
  }
}

//----------------------------------------------------------------
// the_main_window_t::go_to_address
// 
void
the_main_window_t::go_to_address()
{
  go_here(addressText->text());
}

//----------------------------------------------------------------
// the_main_window_t::go_to_selected
// 
void
the_main_window_t::go_to_selected()
{
  // cerr << "go to selected" << endl;
  QItemSelectionModel * sel = iconList->selectionModel();
  QModelIndex curr = sel->currentIndex();
  enterDirectory(curr);
}

//----------------------------------------------------------------
// the_main_window_t::slot_prev
// 
void
the_main_window_t::slot_prev()
{
  QItemSelectionModel * sel = iconList->selectionModel();
  QModelIndex curr = sel->currentIndex();
  QModelIndex next = get_sibling(filesys_, iconList->rootIndex(), curr, -1);
  if (next.isValid())
  {
    sel->clear();
    sel->setCurrentIndex(next, QItemSelectionModel::Select);
    enterDirectory(next);
  }
}
  
//----------------------------------------------------------------
// the_main_window_t::slot_next
// 
void
the_main_window_t::slot_next()
{
  QItemSelectionModel * sel = iconList->selectionModel();
  QModelIndex root = iconList->rootIndex();
  QModelIndex curr = sel->currentIndex();
  if (!curr.isValid()) curr = filesys_->index(0, 0, root);
  
  QModelIndex next = get_sibling(filesys_, root, curr, 1);
  if (!next.isValid() && timer_.isActive())
  {
    // try to wrap around -- if you cant then stop the slideshow:
    next = get_index(filesys_, root, filesys_->index(0, 0, root), 1, 0);
    if (!next.isValid()) slot_stop();
  }
  
  if (next.isValid())
  {
    sel->clear();
    sel->setCurrentIndex(next, QItemSelectionModel::Select);
    enterDirectory(next);
  }
}

//----------------------------------------------------------------
// the_main_window_t::slot_play
// 
void
the_main_window_t::slot_play()
{
  QItemSelectionModel * sel = iconList->selectionModel();
  QModelIndex root = iconList->rootIndex();
  QModelIndex curr = sel->currentIndex();
  if (!curr.isValid()) curr = filesys_->index(0, 0, root);
  
  curr = get_index(filesys_, root, curr, 1, 0);
  if (curr.isValid())
  {
    sel->clear();
    sel->setCurrentIndex(curr, QItemSelectionModel::Select);
    enterDirectory(curr);
    timer_.start(prefsSlideShowDelay());
    sync_ui();
  }
}

//----------------------------------------------------------------
// the_main_window_t::slot_stop
// 
void
the_main_window_t::slot_stop()
{
  timer_.stop();
  sync_ui();
}

//----------------------------------------------------------------
// the_main_window_t::orientation_restore
// 
void
the_main_window_t::orientation_restore()
{
  view(0)->restore_orientation();
}

//----------------------------------------------------------------
// the_main_window_t::full_screen
// 
void
the_main_window_t::full_screen(bool on)
{
  static QRect prev_geom;
  
  if (on)
  {
    prev_geom = geometry();
    
    frameFixed->setParent(NULL,
			  Qt::Window |
			  Qt::FramelessWindowHint |
			  Qt::WindowStaysOnTopHint);
    
    QDesktopWidget * desktop = QApplication::desktop();
    QRect rect = desktop->screenGeometry(frameFixed);
    frameFixed->setGeometry(rect);
    frameFixed->show();
    frameFixed->showFullScreen();
  }
  else
  {
    splitter->insertWidget(0, frameFixed);
    setGeometry(prev_geom);
    
    if (prefsShouldFitToImage())
    {
      fitToImage();
    }
  }
}

//----------------------------------------------------------------
// the_main_window_t::helpAbout
// 
void
the_main_window_t::helpAbout()
{
  AboutDialog about(this);
  about.exec();
}

//----------------------------------------------------------------
// the_main_window_t::enterDirectory
// 
void
the_main_window_t::enterDirectory(QModelIndex index)
{
  // cout << "enterDirectory" << endl;
  QPersistentModelIndex persistent(index);
  
  if (!persistent.isValid() || filesys_->isDir(persistent))
  {
    QString prev = enter_dir(persistent);
    history_[0].push_front(prev);
    history_[1].clear();
    sync_ui();
  }
  else
  {
    // shortcut:
    if (iv_proc_ui_ != NULL)
    {
      iv_proc_ui_->file_load(filesys_->filePath(index));
    }
  }
}

//----------------------------------------------------------------
// the_main_window_t::fitToImage
// 
void
the_main_window_t::fitToImage()
{
  the_iv_proc_t * p = iv_proc_ui()->proc();
  if (p->w_ == 0 || p->h_ == 0) return;
  
  // make sure there are no pending geometry-related events:
  QCoreApplication::processEvents();
  
  QRect rect_max = QApplication::desktop()->availableGeometry(this);
  QRect rect_old = frameGeometry();
  QRect rect_old_frameless = geometry();
  
  int fx0 = rect_old_frameless.x() - rect_old.x();
  int fy0 = rect_old_frameless.y() - rect_old.y();
  int fx1 = ((rect_old.x() + rect_old.width()) -
	     (rect_old_frameless.x() + rect_old_frameless.width()));
  int fy1 = ((rect_old.y() + rect_old.height()) -
	     (rect_old_frameless.y() + rect_old_frameless.height()));
  
  // FIXME: cerr << "image dimensions: " << p->w_ << " x " << p->h_ << endl;
  
  the_view_t * v = view(0);
  int vw = v->width();
  int vh = v->height();
  int dx = p->w_ - vw;
  int dy = p->h_ - vh;
  
  int ideal_w = rect_old.width() + dx;
  int ideal_h = rect_old.height() + dy;
  
  int new_w = std::min(ideal_w, rect_max.width());
  int new_h = std::min(ideal_h, rect_max.height());
  
  int max_x0 = rect_max.x();
  int max_y0 = rect_max.y();
  int max_x1 = max_x0 + rect_max.width() - 1;
  int max_y1 = max_y0 + rect_max.height() - 1;
  
  int new_x0 = rect_old.x();
  int new_y0 = rect_old.y();
  int new_x1 = new_x0 + new_w - 1;
  int new_y1 = new_y0 + new_h - 1;
  
  int shift_x = std::min(0, max_x1 - new_x1);
  int shift_y = std::min(0, max_y1 - new_y1);
  
  int new_x = new_x0 + shift_x;
  int new_y = new_y0 + shift_y;
  
  // apply the new window geometry:
  resize(new_w - (fx0 + fx1), new_h - (fy0 + fy1)); 
  move(new_x, new_y);
  v->restore_orientation();
  QCoreApplication::processEvents();
  
  // make sure the current selection is visible in the icon list:
  QItemSelectionModel * sel = iconList->selectionModel();
  QModelIndex curr = sel->currentIndex();
  iconList->scrollTo(curr);
}

//----------------------------------------------------------------
// the_main_window_t::sync_ui
// 
void
the_main_window_t::sync_ui()
{
  // setup 3D viewing (necessary in order for the compiled display
  // lists to tesselate the splines correctly):
  shared()->gl_make_current();
  
  // shortcut:
  the_procedure_ui_t * proc_ui = active_proc_ui();
  if (proc_ui != NULL) proc_ui->sync_ui();
  
  addressText->setText(filesys_->filePath(iconList->rootIndex()));
  
  actionBack->setEnabled(!history_[0].empty());
  backButton->setEnabled(!history_[0].empty());
  
  actionForward->setEnabled(!history_[1].empty());
  forwardButton->setEnabled(!history_[1].empty());
  
  actionUp->setEnabled(iconList->rootIndex().parent().isValid());
  upButton->setEnabled(iconList->rootIndex().parent().isValid());
  
  actionPlay->setEnabled(!timer_.isActive());
  buttonPlay->setEnabled(!timer_.isActive());
  
  actionStop->setEnabled(timer_.isActive());
  buttonStop->setEnabled(timer_.isActive());
  
  QModelIndex root = iconList->rootIndex();
  QModelIndex curr = iconList->selectionModel()->currentIndex();
  bool prev_ok = get_sibling(filesys_, root, curr, -1).isValid();
  actionPrev->setEnabled(prev_ok);
  buttonPrev->setEnabled(prev_ok);
  
  bool next_ok = get_sibling(filesys_, root, curr, 1).isValid();
  actionNext->setEnabled(next_ok);
  buttonNext->setEnabled(next_ok);
  
  // update the views:
  for (unsigned int i = 0; i < view_.size(); i++)
  {
    the_qt_view_t * v = view<the_qt_view_t>(i);
    if (v->isHidden()) continue;
    v->refresh();
  }
}

//----------------------------------------------------------------
// the_main_window_t::kill_ui
// 
void
the_main_window_t::kill_ui()
{
  file_exit();
}

//----------------------------------------------------------------
// the_main_window_t::get_visible_views
// 
void
the_main_window_t::get_visible_views(std::vector<the_view_t *> & views) const
{
  views.resize(1);
  views[0] = view_[0];
}

//----------------------------------------------------------------
// the_main_window_t::isFullScreen
// 
bool
the_main_window_t::isFullScreen() const
{
  return actionFullScreen->isChecked();
}

//----------------------------------------------------------------
// the_main_window_t::toggleSlideShow
// 
void
the_main_window_t::toggleSlideShow()
{
  if (timer_.isActive())
  {
    actionStop->trigger();
  }
  else
  {
    actionPlay->trigger();
  }
}

//----------------------------------------------------------------
// the_main_window_t::closeEvent
// 
void
the_main_window_t::closeEvent(QCloseEvent * e)
{
  e->ignore();
  kill_ui();
}

//----------------------------------------------------------------
// the_main_window_t::dragEnterEvent
// 
void
the_main_window_t::dragEnterEvent(QDragEnterEvent * e)
{
  if (e->mimeData()->hasUrls() && (e->possibleActions() & Qt::MoveAction))
  {
    QString path = e->mimeData()->urls().front().path();
    if (is_valid_path(path))
    {
      e->acceptProposedAction();
    }
  }
}

//----------------------------------------------------------------
// the_main_window_t::dropEvent
// 
void
the_main_window_t::dropEvent(QDropEvent * e)
{
  if (e->mimeData()->hasUrls() && (e->possibleActions() & Qt::MoveAction))
  {
    QString path = e->mimeData()->urls().front().path();
    if (is_valid_path(path))
    {
      go_here(path);
      e->acceptProposedAction();
    }
  }
}

//----------------------------------------------------------------
// the_main_window_t::enter_dir
// 
QString
the_main_window_t::enter_dir(const QModelIndex & root)
{
  QString prev = ::enter_dir(iconList, filesys_, root);
  if (timer_.isActive())
  {
    go_to_selected();
  }
  return prev;
}
