/*  
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// File         : the_thumbnails.cxx
// Author       : Pavel Aleksandrovich Koshevoy
// Created      : Wed Feb 14 13:43:00 MST 2007
// Copyright    : (C) 2007-2008 University of Utah
// License      : GPLv2
// Description  : multi-threaded image thumbnail generator


// local includes:
#include "the_thumbnails.hxx"

// Qt includes:
#include <QPainter>
#include <QRect>

// system includes:
#include <iostream>
#include <math.h>

// namespace access:
using std::cerr;
using std::endl;


//----------------------------------------------------------------
// THUMBNAIL_THREAD
// 
// the pixmap generating thread:
// 
the_qt_thread_pool_t *
THUMBNAIL_THREAD()
{
  static the_qt_thread_pool_t * pool_ = NULL;
  if (pool_ == NULL)
  {
    pool_ = new the_qt_thread_pool_t(2);
  }
  
  return pool_;
}

//----------------------------------------------------------------
// Thumbnail::clock_
// 
unsigned int
Thumbnail::clock_ = 0;

//----------------------------------------------------------------
// Thumbnail::Thumbnail
// 
Thumbnail::Thumbnail(const QFileInfo & info):
  failed_to_load_(false),
  info_(info),
  size_(0, 0),
  ts_(0)
{}

//----------------------------------------------------------------
// Thumbnail::pixmap
// 
QPixmap
Thumbnail::pixmap()
{
  clock_++;
  ts_ = clock_;
  
  // avoid working with pixmaps in secondary threads:
  if (!image_.isNull() && pixmap_.isNull())
  {
    pixmap_ = QPixmap::fromImage(image_);
  }
  
  return pixmap_;
}

//----------------------------------------------------------------
// Thumbnail::update
// 
bool
Thumbnail::update(const QSize & size)
{
  if (isReady(size)) return true;
  
  // load the image:
  QString filepath(info_.absoluteFilePath());
  QImage tmp(filepath);
  bool ok = !tmp.isNull();
  
  if (!ok)
  {
    // cerr << "could not load " << qPrintable(filepath) << endl;
    failed_to_load_ = true;
    return false;
  }
  
  int w = tmp.width();
  int h = tmp.height();
  double wr = double(w) / double(size.width());
  double hr = double(h) / double(size.height());
  double ar = double(w) / double(h);
  
  // scale the image to fit the thumbnail:
  if (w > size.width() || h > size.height())
  {
    QSize sz;
    if (wr > hr)
    {
      sz.setWidth(size.width());
      sz.setHeight(std::max(1, int(double(size.width()) / ar)));
    }
    else
    {
      sz.setWidth(std::max(1, int(double(size.height()) * ar)));
      sz.setHeight(size.height());
    }
    
    // FIXME: Qt image scaling is buggy for large images, try a workaround:
    // calculate the number of levels of the multi-res pyramid:
    static const double LOG_2 = log(2.0);
    
    const double x_discrete_levels = floor(w / sz.width());
    const double y_discrete_levels = floor(h / sz.height());
    const double discrete_levels = std::min(x_discrete_levels,
					    y_discrete_levels);
    
    const double divisions = std::max(log(double(discrete_levels)) / LOG_2,
				      0.0);
    const unsigned int num_levels = (unsigned int)(floor(divisions));
    
    // cerr << "---------------------------------------" << endl;
    for (unsigned int i = 0; i < num_levels; i++)
    {
#if 0
      cerr << tmp.width() << " -> " << tmp.width() / 2 << endl
	   << tmp.height() << " -> " << tmp.height() / 2 << endl
	   << endl;
#endif
      
      QImage half_img = tmp.scaled(tmp.width() / 2,
				   tmp.height() / 2,
				   Qt::IgnoreAspectRatio,
				   Qt::SmoothTransformation);
      tmp = half_img;
    }
    
#if 0
    cerr << tmp.width() << " -> " << sz.width() << endl
	 << tmp.height() << " -> " << sz.height() << endl
	 << endl;
#endif
    
    QImage scaled = tmp.scaled(sz,
			       Qt::IgnoreAspectRatio,
			       Qt::SmoothTransformation);
    tmp = scaled;
  }
  
  image_ = tmp;
  size_ = size;
  
  // cerr << qPrintable(filepath) << " finished" << endl;
  return true;
}


//----------------------------------------------------------------
// ThumbnailTransaction::ThumbnailTransaction
// 
ThumbnailTransaction::ThumbnailTransaction(Thumbnail & thumb,
					   const QSize & size):
  thumb_(thumb),
  size_(size)
{}

//----------------------------------------------------------------
// ThumbnailTransaction::execute
// 
void
ThumbnailTransaction::execute(the_thread_interface_t * /* thread */)
{
  thumb_.update(size_);
}


//----------------------------------------------------------------
// ThumbnailProvider::ThumbnailProvider
// 
ThumbnailProvider::ThumbnailProvider():
  num_thumbs_(0)
{}

//----------------------------------------------------------------
// ThumbnailProvider::icon
// 
QIcon
ThumbnailProvider::icon(const QFileInfo & info) const
{
  if (!info.isFile())
  {
    QIcon ic = QFileIconProvider::icon(info);
    return QIcon(new PaddedThumbnailEngine(ic));
  }
  
  ThumbnailProvider * fake = const_cast<ThumbnailProvider *>(this);
  return QIcon(new ThumbnailEngine(info, fake));
}

//----------------------------------------------------------------
// ThumbnailProvider::thumbnail
// 
Thumbnail &
ThumbnailProvider::thumbnail(const QFileInfo & info)
{
  // first, see if we already have a matching pixmap:
  std::list<Thumbnail>::iterator i_thumb = thumbs_.begin();
  for (; i_thumb != thumbs_.end(); ++i_thumb)
  {
    const Thumbnail & t = *i_thumb;
    if (t.info() == info) break;
  }
  
  if (i_thumb == thumbs_.end())
  {
    // thumbnail not found, make a new one:
    thumbs_.push_front(Thumbnail(info));
    num_thumbs_++;
    
    /*
    // FIXME: write a garbage collector to remove unused thumbnails:
    if (num_thumbs_ > 10240)
    {
    // remove the last icon:
    thumbs_.erase(thumbs_.rbegin());
    num_thumbs_--;
    }
    */
    
    i_thumb = thumbs_.begin();
  }
  
  Thumbnail & t = *i_thumb;
  return t;
}

//----------------------------------------------------------------
// ThumbnailProvider::update
// 
void
ThumbnailProvider::update(Thumbnail & thumb, const QSize & size)
{
  // cerr << "updating: " << qPrintable(thumb.info().fileName()) << endl;
  ThumbnailTransaction * t = new ThumbnailTransaction(thumb, size);
  THUMBNAIL_THREAD()->start(t);
}


//----------------------------------------------------------------
// ThumbnailEngine::ThumbnailEngine
// 
ThumbnailEngine::ThumbnailEngine(const QFileInfo & info,
				 ThumbnailProvider * provider):
  info_(info),
  provider_(provider)
{}

//----------------------------------------------------------------
// ThumbnailEngine::paint
// 
void
ThumbnailEngine::paint(QPainter * painter,
		       const QRect & rect,
		       QIcon::Mode /* mode */,
		       QIcon::State /* state */)
{
  /*
    cerr << "ThumbnailEngine::paint, rect: " << rect.width() << " "
    << rect.height() << endl;
  */
  Thumbnail & t = provider_->thumbnail(info_);
  if (t.isReady(rect.size()))
  {
    const QImage & icon = t.image();
    int x = rect.x() + (rect.width() - icon.width()) / 2;
    int y = rect.y() + (rect.height() - icon.height()) / 2;
    // painter->fillRect(rect, QBrush(QColor(64, 128, 255, 255)));
    painter->drawImage(QPoint(x, y), t.image());
  }
  else
  {
    if (!t.failedToLoad())
    {
      provider_->update(t, rect.size());
    }
    
    painter->fillRect(rect, QBrush(QColor(64, 128, 255, 255)));
  }
}

//----------------------------------------------------------------
// ThumbnailEngine::pixmap
// 
QPixmap
ThumbnailEngine::pixmap(const QSize & size,
			QIcon::Mode mode,
			QIcon::State state)
{
  // cerr << "ThumbnailEngine::pixmap" << endl;
  Thumbnail & t = provider_->thumbnail(info_);
  if (t.isReady(size)) return t.pixmap();
  
  // tell the provider to make a new pixmap:
  provider_->update(t, size);
  
  return QIconEngine::pixmap(size, mode, state);
}


//----------------------------------------------------------------
// PaddedThumbnailEngine::PaddedThumbnailEngine
// 
PaddedThumbnailEngine::PaddedThumbnailEngine(const QIcon & icon):
  icon_(icon)
{}

//----------------------------------------------------------------
// PaddedThumbnailEngine::paint
// 
void
PaddedThumbnailEngine::paint(QPainter * painter,
			     const QRect & rect,
			     QIcon::Mode mode,
			     QIcon::State state)
{
  /*
    cerr << "PaddedThumbnailEngine::paint, rect: "
    << rect.width() << " " << rect.height() << endl;
  */
  QSize isz = icon_.actualSize(rect.size(), mode, state);
  int w = isz.width();
  int h = isz.height();
  int x = rect.x() + (rect.width() - w) / 2;
  int y = rect.y() + (rect.height() - h) / 2;
  // painter->fillRect(rect, QBrush(QColor(64, 128, 255, 255)));
  icon_.paint(painter, QRect(x, y, w, h), Qt::AlignCenter, mode, state);
}

//----------------------------------------------------------------
// PaddedThumbnailEngine::pixmap
// 
QPixmap
PaddedThumbnailEngine::pixmap(const QSize & size,
			      QIcon::Mode mode,
			      QIcon::State state)
{
  // cerr << "PaddedThumbnailEngine::pixmap" << endl;
  QPixmap pm(size);
  QPainter painter(&pm);
  paint(&painter, QRect(0, 0, size.width(), size.height()), mode, state);
  return pm;
}
