/*  
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// File         : the_thumbnails.hxx
// Author       : Pavel Aleksandrovich Koshevoy
// Created      : Wed Feb 14 13:37:00 MST 2007
// Copyright    : (C) 2007, 2008 University of Utah
// License      : GPLv2
// Description  : multi-threaded image thumbnail generator

#ifndef THE_THUMBNAILS_HXX_
#define THE_THUMBNAILS_HXX_

// system includes:
#include <list>

// Qt includes:
#include <QFileInfo>
#include <QPixmap>
#include <QImage>
#include <QSize>
#include <QFileIconProvider>
#include <QIcon>
#include <QIconEngine>

// the includes:
#include <Qt/the_qt_thread_pool.hxx>
#include <thread/the_transaction.hxx>

// forward declarations:
class QPainter;
class QRect;

//----------------------------------------------------------------
// THUMBNAIL_THREAD
// 
extern the_qt_thread_pool_t * THUMBNAIL_THREAD();


//----------------------------------------------------------------
// Thumbnail
// 
class Thumbnail
{
public:
  Thumbnail(const QFileInfo & info);
  
  QPixmap pixmap();
  
  bool update(const QSize & size);
  
  inline bool isReady(const QSize & size) const
  { return !image_.isNull() && (size == size_); }
  
  inline bool failedToLoad() const
  { return failed_to_load_; }
  
  // accessors:
  inline const QFileInfo & info() const
  { return info_; }
  
  inline const QSize & size() const
  { return size_; }
  
  inline const QImage & image() const
  { return image_; }
  
  inline const unsigned int & ts() const
  { return ts_; }
  
protected:
  bool failed_to_load_;
  QFileInfo info_;
  QSize size_;
  QImage image_;
  QPixmap pixmap_;
  unsigned int ts_;
  static unsigned int clock_;
};


//----------------------------------------------------------------
// ThumbnailTransaction
// 
class ThumbnailTransaction : public the_transaction_t
{
public:
  ThumbnailTransaction(Thumbnail & thumb, const QSize & size);
  
  // virtual:
  void execute(the_thread_interface_t * /* thread */);
  
  // transaction parameters:
  Thumbnail & thumb_;
  QSize size_;
};


//----------------------------------------------------------------
// ThumbnailProvider
// 
class ThumbnailProvider : public QFileIconProvider
{
public:
  ThumbnailProvider();
  
  // virtual:
  QIcon icon(QFileIconProvider::IconType type) const
  { return QFileIconProvider::icon(type); }
  
  // virtual:
  QIcon icon(const QFileInfo & info) const;
  
  // helper:
  Thumbnail & thumbnail(const QFileInfo & info);
  
  // generate a thumbnail with a helper thread:
  void update(Thumbnail & thumb, const QSize & size);
  
  // the thumbnails:
  unsigned int num_thumbs_;
  std::list<Thumbnail> thumbs_;
};


//----------------------------------------------------------------
// ThumbnailEngine
// 
class ThumbnailEngine : public QIconEngine
{
public:
  ThumbnailEngine(const QFileInfo & info, ThumbnailProvider * provider);
  
  // virtual:
  QSize actualSize(const QSize & size, QIcon::Mode, QIcon::State)
  { return size; }
  
  // virtual:
  void paint(QPainter * painter,
	     const QRect & rect,
	     QIcon::Mode /* mode */,
	     QIcon::State /* state */);
  
  // virtual:
  QPixmap pixmap(const QSize & size, QIcon::Mode mode, QIcon::State state);
  
private:
  QFileInfo info_;
  ThumbnailProvider * provider_;
};


//----------------------------------------------------------------
// PaddedThumbnailEngine
// 
class PaddedThumbnailEngine : public QIconEngine
{
public:
  PaddedThumbnailEngine(const QIcon & icon);
  
  // virtual:
  QSize actualSize(const QSize & size, QIcon::Mode, QIcon::State)
  { return size; }
  
  // virtual:
  void paint(QPainter * painter,
	     const QRect & rect,
	     QIcon::Mode mode,
	     QIcon::State state);
  
  // virtual:
  QPixmap pixmap(const QSize & size, QIcon::Mode mode, QIcon::State state);
  
private:
  QIcon icon_;
};


#endif // THE_THUMBNAILS_HXX_
