#include "itkBinaryThresholdImageFilter.h"
#include "itkImage.h"
#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"
#include "../track/util.h"
#include "itkConnectedComponentImageFilter.h"

typedef  float  InputPixelType;
typedef itk::Image< InputPixelType,  2 >   ImageType;
typedef itk::BinaryThresholdImageFilter
               < ImageType, ImageType >  ThresholdFilterType;

typedef unsigned char char_type;
typedef itk::RGBPixel< char_type >   RGBPixelType;
typedef itk::Image< RGBPixelType, 2 > RGBImageType;
typedef itk::ImageRegionIterator< ImageType > ImageIterator;
typedef itk::ImageRegionIterator< RGBImageType > RGBImageIterator;
typedef itk::ConnectedComponentImageFilter
                < lImageType, ImageType > ConnectedComponentFilter;


void convertToRGB( RGBImageType::Pointer& rgb_img, 
              ImageType::Pointer img );
void combine( RGBImageType::Pointer& img, 
              ImageType::Pointer thresh, float color[]);

int main( int argc, char * argv[] )
{
  if( argc < 6 )
    {
    std::cerr << "Usage: " << argv[0];
    std::cerr << " image_to_threshold background_image ";  
    std::cerr << " lowerThreshold higherThreshold";  
    std::cerr << " output_filename_prefix "  << std::endl;  
    return EXIT_FAILURE;
    }
 
  char* thresh_file = argv[ 1 ];
  char* background_file = argv[ 2 ];
  int low_thresh = atoi ( argv[ 3 ] );
  int high_thresh = atoi ( argv[ 4 ] );
  char* prefix = argv[ 5 ]; 
  char filename[100];

  ImageType::Pointer thresh_image = ImageType::New();
  thresh_image = load< ImageType >( thresh_file );

  ImageType::Pointer back_image = ImageType::New();
  back_image = load< ImageType >( background_file );

  //ImageType::PixelType min;
  //ImageType::PixelType max;
  //image_min_max<ImageType>(thresh_image, min, max);
  //remap_min_max_inplace<ImageType>(thresh_image, max, min);

  //
  //  apply the first threshold filter
  //
  ThresholdFilterType::Pointer low_thresh_filter = ThresholdFilterType::New();
  low_thresh_filter->SetInput( thresh_image );
  const InputPixelType outsideValue = 0;
  const InputPixelType insideValue  = 255;
  low_thresh_filter->SetOutsideValue( outsideValue );
  low_thresh_filter->SetInsideValue( insideValue  );
  low_thresh_filter->SetLowerThreshold( low_thresh );
  low_thresh_filter->SetUpperThreshold( 255 );
  low_thresh_filter->Update();

  ConnectedComponentImageFilter::Pointer cc1_filter =
                              ConnectedComponentImageFilter::New()
  cc1_filter->SetInput( low_threshold_filter->GetOutput() );
  cc1_filter->Update();

  //
  //  apply the second threshold filter
  //
  ThresholdFilterType::Pointer high_thresh_filter = ThresholdFilterType::New();
  high_thresh_filter->SetInput( thresh_image );
  high_thresh_filter->SetOutsideValue( outsideValue );
  high_thresh_filter->SetInsideValue( insideValue  );
  high_thresh_filter->SetLowerThreshold( high_thresh );
  high_thresh_filter->SetUpperThreshold( 255 );
  high_thresh_filter->Update();

  ConnectedComponentImageFilter::Pointer cc2_filter =
                              ConnectedComponentImageFilter::New()
  cc2_filter->SetInput( high_threshold_filter->GetOutput() );
  cc2_filter->Update();

  //sprintf( filename, "%s_thresh%03i.mha", prefix, thresh );
  //save< ImageType >( thresh_filter->GetOutput(), filename );

  //
  //  combine threshold with background image
  //
  RGBImageType::Pointer thresh_rgb_image = new2d< RGBImageType >
     ( thresh_image->GetLargestPossibleRegion().GetSize()[0],
       thresh_image->GetLargestPossibleRegion().GetSize()[1] );
  convertToRGB( thresh_rgb_image, back_image );
  float color[3]; color[0] = 1; color[1] = 0; color[2] = 0;
  combine( thresh_rgb_image, low_thresh_filter->GetOutput(), color );
  color[0] = 0; color[1] = 1; color[2] = 0;
  combine( thresh_rgb_image, high_thresh_filter->GetOutput(), color ); 
  sprintf( filename, "%s_thresh%03i-%03i_comb.png", prefix, 
                       low_thresh, high_thresh );
  save< RGBImageType >( thresh_rgb_image, filename );
     
  return EXIT_SUCCESS;
}

void convertToRGB( RGBImageType::Pointer& rgb_img, 
              ImageType::Pointer img )
{
   img = rescale<ImageType>(img, 0, 255);
   ImageIterator image_iter(img, img->GetLargestPossibleRegion());
   RGBImageIterator rgb_image_iter(rgb_img, 
                            rgb_img->GetLargestPossibleRegion());

   for(rgb_image_iter.GoToBegin(); 
         !rgb_image_iter.IsAtEnd();
         ++rgb_image_iter)
   {
      float val = image_iter.Get();
      RGBPixelType rgb_pixel;
      rgb_pixel[0] = static_cast<unsigned char>(val); 
      rgb_pixel[1] = static_cast<unsigned char>(val);
      rgb_pixel[2] = static_cast<unsigned char>(val);
      rgb_image_iter.Set(rgb_pixel);

      ++image_iter;
   }
   rgb_img->Update();
}

void combine( RGBImageType::Pointer& rgb_img,
              ImageType::Pointer thresh_img, float color[] )
{
   ImageIterator image_iter(thresh_img, thresh_img->GetLargestPossibleRegion());
   RGBImageIterator rgb_image_iter(rgb_img, 
                            rgb_img->GetLargestPossibleRegion());

   image_iter.GoToBegin();
   for(rgb_image_iter.GoToBegin(); 
         !rgb_image_iter.IsAtEnd();
         ++rgb_image_iter)
   {
      float val = image_iter.Get();
      if( val > 0 )
      {
         RGBPixelType rgb_pixel;
         rgb_pixel[0] = static_cast<unsigned char>(255 * color[0]); 
         rgb_pixel[1] = static_cast<unsigned char>(255 * color[1] );
         rgb_pixel[2] = static_cast<unsigned char>(255 * color[2]);
         rgb_image_iter.Set(rgb_pixel);
      }

      ++image_iter;
   }
   rgb_img->Update();
}

