
#include "itkRescaleIntensityImageFilter.h"
#include "itkIntensityWindowingImageFilter.h"
#include <itkStatisticsImageFilter.h>
#include "itkSubtractConstantFromImageFilter.h"
#include "itkAddConstantToImageFilter.h"
#include "itkMultiplyByConstantImageFilter.h"
#include "itkDivideByConstantImageFilter.h"
#include "itkIntensityWindowingImageFilter.h"

#include "itkImage.h"
#include "../track/util.h"

//
//  ITK types we're using
//
typedef unsigned char out_type;
const     unsigned int    Dimension = 2;
typedef itk::Image< double, Dimension >  ImageType;
typedef itk::Image< out_type, Dimension >  OutputImageType;

//
//  ITK Filters we're using
//
typedef itk::ImageRegionIterator< ImageType > ImageRegionIterator;
typedef itk::StatisticsImageFilter< ImageType > StatFilterType;
typedef itk::SubtractConstantFromImageFilter
               < ImageType, double, ImageType > SubtractFilter;
typedef itk::AddConstantToImageFilter
               < ImageType, double, ImageType > AddFilter;
typedef itk::MultiplyByConstantImageFilter
               < ImageType, double, ImageType > MultiplyFilter;
typedef itk::DivideByConstantImageFilter
               < ImageType, double, ImageType > DivideFilter;
typedef itk::IntensityWindowingImageFilter
               < ImageType, OutputImageType > IntensityWindowingImageFilter;


using namespace std;

int main( int argc, char ** argv )
{
   // Verify the number of parameters in the command line
   if( argc < 5 )
   {
      std::cerr << "Usage: " << std::endl;
      std::cerr << argv[0] << " avg_mean/null agv_std/null "
                << "outputDirectory inputFile" << std::endl;
      return EXIT_FAILURE;
   }

   char* avg_mean_char = argv[1];
   char* avg_std_char = argv[2];
   bool calc_averages = true;
   double avg_mean = 0, avg_std = 0;
   if( ( strcmp( avg_mean_char, "null" ) != 0 ) &&
       ( strcmp( avg_std_char, "null" ) != 0 ) )
   {
      cout << "found the avg_mean and avg_std" << endl ;
      calc_averages = false;
      avg_mean = atof( avg_mean_char );
      avg_std = atof( avg_std_char );
   }


   char* output_directory = argv[3];
   cout << "output dir: " << output_directory << endl ;

   vector< string > input_filenames;
   for( int i = 4; i < argc; i++ )
   {
      string afile( argv[i] );
      input_filenames.push_back( afile );
   }
   unsigned int s = input_filenames.size();

   //
   //  if the avg std and mean were not given in the command line,
   //  calculate them here.
   //
   if( calc_averages )
   {

      //
      //  read each file and calculate the mean and std
      //
      double n = 0;
      for( unsigned int i = 0; i < s; i++ )
      {
         ImageType::Pointer image = load< ImageType >
            ( input_filenames[ i ].c_str() );

         StatFilterType::Pointer statFilter = StatFilterType::New();
         statFilter->SetInput( image );
         statFilter->Update();
         float var = statFilter->GetVariance();
         float mean = statFilter->GetMean();
         float min = statFilter->GetMinimum();
         float max = statFilter->GetMaximum();
         float stdev = sqrt( var );
         cout << "var: " << var << ", stdev: " << sqrt(var) << ", mean: " << mean 
            << ", min: " << min << ", max: " << max << endl;

         // dont avg in the flat regions
         if( stdev > 20 )
         {
            avg_mean += mean;
            avg_std += stdev;
            n++;
         }
      }

      avg_mean = avg_mean / n;//s;
      avg_std = avg_std / n;//s;
   }

   cout << endl << "avg stdev: " << avg_std << ", avg mean: " << avg_mean
      << endl << endl ;

   //
   //  read each file and adjust the intensity 
   //
   for( unsigned int i = 0; i < s; i++ )
   {
      ImageType::Pointer image = load< ImageType >( input_filenames[ i ].c_str());

      //
      //  get the original stats again
      //
      StatFilterType::Pointer statFilter = StatFilterType::New();
      statFilter->SetInput( image );
      statFilter->Update();
      float var = statFilter->GetVariance();
      float stdev = sqrt( var );
      float mean = statFilter->GetMean();
      float min = statFilter->GetMinimum();
      float max = statFilter->GetMaximum();
      cout << "old var: " << var << ", stdev: " << sqrt(var) << ", mean: " << mean 
         << ", min: " << min << ", max: " << max << endl;

      //
      //  adjust the intensities
      //
      SubtractFilter::Pointer sub = SubtractFilter::New();
      sub->SetInput( image );
      sub->SetConstant( mean );

      DivideFilter::Pointer div = DivideFilter::New();
      div->SetInput( sub->GetOutput() );
      div->SetConstant( stdev );

      MultiplyFilter::Pointer mult = MultiplyFilter::New();
      mult->SetInput( div->GetOutput() );
      mult->SetConstant( avg_std );

      AddFilter::Pointer add = AddFilter::New();
      add->SetInput( mult->GetOutput() );
      add->SetConstant( avg_mean );
      add->Update();

      //
      //  get the new image stats
      //
      StatFilterType::Pointer statFilter2 = StatFilterType::New();
      statFilter2->SetInput( add->GetOutput() );
      statFilter2->Update();
      var = statFilter2->GetVariance();
      mean = statFilter2->GetMean();
      min = statFilter2->GetMinimum();
      max = statFilter2->GetMaximum();
      cout << "new var: " << var << ", stdev: " << sqrt(var) << ", mean: " << mean 
         << ", min: " << min << ", max: " << max << endl;

      //
      //  "rescale the image for char"
      //
      IntensityWindowingImageFilter::Pointer intensityWindowing =
         IntensityWindowingImageFilter::New();
      intensityWindowing->SetWindowMinimum( mean - (3*sqrt(var)) ); 
      intensityWindowing->SetWindowMaximum( mean + (3*sqrt(var)) );
      intensityWindowing->SetOutputMinimum( (int)numeric_limits< out_type >::min( )  );
      intensityWindowing->SetOutputMaximum( (int)numeric_limits< out_type >::max( ) );
      intensityWindowing->SetInput( add->GetOutput() );
      cout << "min_max_type: " << (int)numeric_limits< out_type >::min( )
         << " " << (int)numeric_limits< out_type >::max() << endl ;
         

      string simp_file = input_filenames[i];
      int ind = simp_file.find("/",0);  // get the basename
      simp_file.erase( 0, ind );
      //ind = simp_file.find(".png",0);  // remove the file extension
      simp_file.erase( simp_file.size() - 4 ); // remove the file ext.
      char filename[100];
      sprintf( filename, "%s%s_adj.png", output_directory, simp_file.c_str() );
      //save< OutputImageType >( add->GetOutput(), filename ); 
      save< OutputImageType >( intensityWindowing->GetOutput(), filename ); 
      sprintf( filename, "%s%s_adj.mha", output_directory, simp_file.c_str() );
      save< ImageType >( add->GetOutput(), filename ); 

   }




   return EXIT_SUCCESS;
}




