#include "itkBinaryThresholdImageFilter.h"
#include "itkImage.h"
#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"
#include "../track/util.h"

typedef  float  InputPixelType;
typedef itk::Image< InputPixelType,  2 >   ImageType;
typedef itk::BinaryThresholdImageFilter
               < ImageType, ImageType >  ThresholdFilterType;

typedef unsigned char char_type;
typedef itk::RGBPixel< char_type >   RGBPixelType;
typedef itk::Image< RGBPixelType, 2 > RGBImageType;
typedef itk::ImageRegionIterator< ImageType > ImageIterator;
typedef itk::ImageRegionIterator< RGBImageType > RGBImageIterator;

void convertToRGB( RGBImageType::Pointer& rgb_img, 
              ImageType::Pointer img );
void combine( RGBImageType::Pointer& img, 
              ImageType::Pointer thresh );

int main( int argc, char * argv[] )
{
  if( argc < 7 )
    {
    std::cerr << "Usage: " << argv[0];
    std::cerr << " image_to_threshold background_image ";  
    std::cerr << " lowerThreshold highThreshold increment";  
    std::cerr << " output_filename_prefix [calc_inv(0/1)]"  << std::endl;  
    return EXIT_FAILURE;
    }
 
  char* thresh_file = argv[ 1 ];
  char* background_file = argv[ 2 ];
  int low_thresh = atoi ( argv[ 3 ] );
  int high_thresh = atoi ( argv[ 4 ] );
  double inc = atof ( argv[ 5 ] );
  char* prefix = argv[ 6 ]; 
  char filename[100];

  ImageType::Pointer thresh_image = ImageType::New();
  thresh_image = load< ImageType >( thresh_file );

  ImageType::Pointer back_image = ImageType::New();
  back_image = load< ImageType >( background_file );

  if( argc > 7 )
  {
     if( atoi( argv[7] ) )
     {
        ImageType::PixelType min;
        ImageType::PixelType max;
        image_min_max<ImageType>(thresh_image, min, max);
        remap_min_max_inplace<ImageType>(thresh_image, max, min);
     }
  }

  for( double thresh = low_thresh; thresh < high_thresh; thresh+=inc )
  {
     cout << "thresh: " << thresh << endl;
     //
     //  apply the threshold filter
     //
     ThresholdFilterType::Pointer thresh_filter = ThresholdFilterType::New();
     thresh_filter->SetInput( thresh_image );
     const InputPixelType outsideValue = 0;
     const InputPixelType insideValue  = 255;
     thresh_filter->SetOutsideValue( outsideValue );
     thresh_filter->SetInsideValue( insideValue  );

     //const InputPixelType lowerThreshold = low_thresh; 
     thresh_filter->SetLowerThreshold( thresh );
     thresh_filter->SetUpperThreshold( high_thresh );
     thresh_filter->Update();

     sprintf( filename, "%s_thresh%03i.mha", prefix, (int)thresh );
     save< ImageType >( thresh_filter->GetOutput(), filename );

     //
     //  combine threshold with background image
     //
     RGBImageType::Pointer thresh_rgb_image = new2d< RGBImageType >
               ( thresh_image->GetLargestPossibleRegion().GetSize()[0],
                 thresh_image->GetLargestPossibleRegion().GetSize()[1] );
     convertToRGB( thresh_rgb_image, back_image );
     combine( thresh_rgb_image, thresh_filter->GetOutput() ); 
     sprintf( filename, "%s_thresh%03i_comb.png", prefix, (int)thresh );
     save< RGBImageType >( thresh_rgb_image, filename );
     
  }

  return EXIT_SUCCESS;
}

void convertToRGB( RGBImageType::Pointer& rgb_img, 
              ImageType::Pointer img )
{
   img = rescale<ImageType>(img, 0, 255);
   ImageIterator image_iter(img, img->GetLargestPossibleRegion());
   RGBImageIterator rgb_image_iter(rgb_img, 
                            rgb_img->GetLargestPossibleRegion());

   for(rgb_image_iter.GoToBegin(); 
         !rgb_image_iter.IsAtEnd();
         ++rgb_image_iter)
   {
      float val = image_iter.Get();
      RGBPixelType rgb_pixel;
      rgb_pixel[0] = static_cast<unsigned char>(val); 
      rgb_pixel[1] = static_cast<unsigned char>(val);
      rgb_pixel[2] = static_cast<unsigned char>(val);
      rgb_image_iter.Set(rgb_pixel);

      ++image_iter;
   }
   rgb_img->Update();
}

void combine( RGBImageType::Pointer& rgb_img,
              ImageType::Pointer thresh_img )
{
   ImageIterator image_iter(thresh_img, thresh_img->GetLargestPossibleRegion());
   RGBImageIterator rgb_image_iter(rgb_img, 
                            rgb_img->GetLargestPossibleRegion());

   image_iter.GoToBegin();
   for(rgb_image_iter.GoToBegin(); 
         !rgb_image_iter.IsAtEnd();
         ++rgb_image_iter)
   {
      float val = image_iter.Get();
      if( val > 0 )
      {
         RGBPixelType rgb_pixel;
         rgb_pixel[0] = static_cast<unsigned char>(255); 
         rgb_pixel[1] = static_cast<unsigned char>(0);
         rgb_pixel[2] = static_cast<unsigned char>(0);
         rgb_image_iter.Set(rgb_pixel);
      }

      ++image_iter;
   }
   rgb_img->Update();
}
