/*
Copyright 2004-2007 University of Utah

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/


// File         : crop.cxx
// Author       : Pavel Aleksandrovich Koshevoy
// Created      : 2005/09/19 14:54
// Copyright    : (C) 2005
// License      : MIT
// Description  : Crop a region of interest from a series of images.

// ITK includes:
#include <itkImageFileReader.h>
#include <itkImageFileWriter.h>
#include <itkExtractImageFilter.h>
#include <itkImageRegionConstIteratorWithIndex.h>

// system includes:
#include <algorithm>
#include <iostream>
#include <string>
#include <list>
#include <errno.h>
#include <stdlib.h>

// namespace access:
using std::cout;
using std::cerr;
using std::endl;

typedef std::string text_t;
typedef std::list<text_t> text_list_t;

// typedef unsigned char pix_t;
// typedef short pix_t;
typedef float pix_t;
typedef itk::Image<pix_t, 2> image_t;
typedef itk::Image<pix_t, 3> volume_t;

#define WRAP(x)

//----------------------------------------------------------------
// load
// 
template <class image_t>
typename image_t::Pointer
load(const char * filename, bool blab = true)
{
  typedef typename itk::ImageFileReader<image_t> reader_t;
  typename reader_t::Pointer reader = reader_t::New();
  reader->SetFileName(filename);
  
  if (blab) cout << "loading " << filename << endl;
  reader->Update();
  
  return reader->GetOutput();
}

//----------------------------------------------------------------
// save
// 
template <class image_t>
void
save(const image_t * image, const char * filename, bool blab = true)
{
  typedef typename itk::ImageFileWriter<image_t> writer_t;
  typename writer_t::Pointer writer = writer_t::New();
  writer->SetFileName(filename);
  writer->SetInput(image);
  
  if (blab) cout << "saving " << filename << endl;
  writer->Update();
}

//----------------------------------------------------------------
// crop
// 
template <class image_t>
typename image_t::Pointer
crop(const image_t * image,
     const typename image_t::IndexType & min,
     const typename image_t::IndexType & max)
{
  typedef typename image_t::RegionType::SizeType sz_t;
  
  sz_t img_sz = image->GetLargestPossibleRegion().GetSize();
  sz_t reg_sz;
  const unsigned int d = image_t::GetImageDimension();
  for (unsigned int i = 0; i < d; i++)
  {
    reg_sz[i] = std::min((unsigned int)(max[i] - min[i] + 1),
			 (unsigned int)(img_sz[i] - min[i]));
  }
  
  typename image_t::RegionType region;
  region.SetIndex(min);
  region.SetSize(reg_sz);
  
  typedef typename itk::ExtractImageFilter<image_t, image_t> crop_t;
  typename crop_t::Pointer filter = crop_t::New();
  filter->SetInput(image);
  filter->SetExtractionRegion(region);
  filter->Update();
  
  return filter->GetOutput();
}

//----------------------------------------------------------------
// image_min_max
// 
// Find min/max pixel values, return mean pixel value (average):
// 
template <typename T>
double
image_min_max(const T * a,
	      typename T::PixelType & min,
	      typename T::PixelType & max)
{
  WRAP(itk_terminator_t terminator("image_min_max"));
  
  typedef typename T::PixelType pixel_t;
  typedef typename itk::ImageRegionConstIteratorWithIndex<T> iter_t;
  
  min = std::numeric_limits<pixel_t>::max();
  max = -min;
  
  double mean = 0.0;
  double counter = 0.0;
  iter_t iter(a, a->GetLargestPossibleRegion());
  for (iter.GoToBegin(); !iter.IsAtEnd(); ++iter)
  {
    // make sure there hasn't been an interrupt:
    WRAP(terminator.terminate_on_request());
    
    pixel_t v = iter.Get();
    min = std::min(min, v);
    max = std::max(max, v);
    
    mean += double(v);
    counter += 1.0;
  }
  
  mean /= counter;
  return mean;
}

//----------------------------------------------------------------
// remap_min_max_inplace
// 
template <class image_t>
bool
remap_min_max_inplace(image_t * a,
		      typename image_t::PixelType new_min = 0,
		      typename image_t::PixelType new_max = 255)
{
  WRAP(itk_terminator_t terminator("remap_min_max_inplace"));
  
  typedef typename image_t::PixelType pixel_t;
  typedef typename itk::ImageRegionIterator<image_t> iter_t;
  
  double min = std::numeric_limits<pixel_t>::max();
  double max = -min;
  
  iter_t iter(a, a->GetLargestPossibleRegion());
  for (iter.GoToBegin(); !iter.IsAtEnd(); ++iter)
  {
    // make sure there hasn't been an interrupt:
    WRAP(terminator.terminate_on_request());
    
    double v = iter.Get();
    min = std::min(min, v);
    max = std::max(max, v);
  }
  
  double rng = max - min;
  if (rng == 0.0) return false;
  
  // rescale the intensities:
  double new_rng = new_max - new_min;
  
  for (iter.GoToBegin(); !iter.IsAtEnd(); ++iter)
  {
    // make sure there hasn't been an interrupt:
    WRAP(terminator.terminate_on_request());

    pixel_t p_old = iter.Get();
    
    double t = (double(p_old) - min) / rng;
    pixel_t p_new =
      pixel_t(std::min(double(itk::NumericTraits<pixel_t>::max()),
		       std::max(double(itk::NumericTraits<pixel_t>::min()),
				new_min + t * new_rng)));
    iter.Set(p_new);
  }
  
  return true;
}

//----------------------------------------------------------------
// toUInt
// 
unsigned int
toUInt(const char * text, bool * ok = NULL, int base = 10)
{
  char * endptr = NULL;
  unsigned long int num = strtoul(text, &endptr, base);
  if (ok != NULL) *ok = !(text == endptr || errno == ERANGE);
  return num;
}


//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL)
{
  if (message != NULL)
  {
    cerr << "ERROR: " << message << endl;
  }

  cerr << "USAGE: " << argv[0] << " "
       << "(-xx x0 x1 | -xw x0 w) "
       << "(-yy y0 y1 | -yh y0 h) "
       << "(-zz z0 z1 | -zd z0 d) "
       << "[-invert]  "
       << "-load image1 [imageN]* "
       << "-save prefix suffix "
       << endl
       << "EXAMPLE: " << argv[0] << " "
       << "-xw 400 300 "
       << "-yh 800 100 "
       << "-zh 0 10 "
       << "-invert "
       << "-load *.tif "
       << "-save cropped- .mhd"
       << endl;
  exit(1);
}

//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  if (argc < 10) usage(argv, "not enough parameters");
  
  // parse the command line arguments:
  unsigned int roi_x = ~0;
  unsigned int roi_y = ~0;
  unsigned int roi_z = ~0;
  unsigned int roi_w = ~0;
  unsigned int roi_h = ~0;
  unsigned int roi_d = ~0;
  
  bool invert = false;
  text_list_t in;
  text_t prefix("cropped-");
  text_t suffix(".mhd");
  
  unsigned int consume_default_tokens = 0;
  
  for (int i = 1; i < argc; i++)
  {
    bool ok = true;
    
    if (strcmp(argv[i], "-xw") == 0)
    {
      if ((argc - i) <= 2) usage(argv, "could not parse -xw option");
      
      i++;
      roi_x = toUInt(argv[i], &ok);
      if (!ok) usage(argv, "bad -xw x option");
      
      i++;
      roi_w = toUInt(argv[i], &ok);
      if (!ok) usage(argv, "bad -xw w option");
    }
    else if (strcmp(argv[i], "-yh") == 0)
    {
      if ((argc - i) <= 2) usage(argv, "could not parse -yh option");
      
      i++;
      roi_y = toUInt(argv[i], &ok);
      if (!ok) usage(argv, "bad -yh y option");
      
      i++;
      roi_h = toUInt(argv[i], &ok);
      if (!ok) usage(argv, "bad -yh h option");
    }
    else if (strcmp(argv[i], "-zd") == 0)
    {
      if ((argc - i) <= 2) usage(argv, "could not parse -yh option");
      
      i++;
      roi_z = toUInt(argv[i], &ok);
      if (!ok) usage(argv, "bad -zd z option");
      
      i++;
      roi_d = toUInt(argv[i], &ok);
      if (!ok) usage(argv, "bad -zd d option");
    }
    else if (strcmp(argv[i], "-xx") == 0)
    {
      if ((argc - i) <= 2) usage(argv, "could not parse -xx option");
      
      i++;
      roi_x = toUInt(argv[i], &ok);
      if (!ok) usage(argv, "bad -xx x0 option");
      
      i++;
      roi_w = toUInt(argv[i], &ok) - roi_x + 1;
      if (!ok) usage(argv, "bad -xx x1 option");
    }
    else if (strcmp(argv[i], "-yy") == 0)
    {
      if ((argc - i) <= 2) usage(argv, "could not parse -yy option");
      
      i++;
      roi_y = toUInt(argv[i], &ok);
      if (!ok) usage(argv, "bad -yy y0 option");
      
      i++;
      roi_h = toUInt(argv[i], &ok) - roi_y + 1;
      if (!ok) usage(argv, "bad -yy y1 option");
    }
    else if (strcmp(argv[i], "-zz") == 0)
    {
      if ((argc - i) <= 2) usage(argv, "could not parse -zz option");
      
      i++;
      roi_z = toUInt(argv[i], &ok);
      if (!ok) usage(argv, "bad -zz z0 option");
      
      i++;
      roi_d = toUInt(argv[i], &ok) - roi_z + 1;
      if (!ok) usage(argv, "bad -zz z1 option");
    }
    else if (strcmp(argv[i], "-invert") == 0)
    {
      invert = true;
    }
    else if (strcmp(argv[i], "-load") == 0)
    {
      if ((argc - i) <= 2) usage(argv, "could not parse -load option");
      consume_default_tokens = 2;
    }
    else if (strcmp(argv[i], "-save") == 0)
    {
      if ((argc - i) <= 2) usage(argv, "could not parse -save option");
      
      i++;
      prefix = text_t(argv[i]);
      
      i++;
      suffix = text_t(argv[i]);
    }
    else if (consume_default_tokens > 0)
    {
      in.push_back(text_t(argv[i]));
      consume_default_tokens++;
    }
    else
    {
      usage(argv, argv[i]);
    }
    
    if (consume_default_tokens > 0)
    {
      consume_default_tokens--;
    }
  }
  
  // iterate over the input files:
  for (text_list_t::iterator i = in.begin(); i != in.end(); ++i)
  {
    text_t filein = *i;
    
    if (roi_z == ~0)
    {
      // read the 2D image:
      image_t::Pointer image = load<image_t>(filein.c_str());
      
      // setup the cropping parameters:
      image_t::RegionType::SizeType sz =
	image->GetLargestPossibleRegion().GetSize();
      
      image_t::IndexType index[2];
      index[0][0] = roi_x;
      index[0][1] = roi_y;
      index[1][0] = roi_x + roi_w - 1;
      index[1][1] = roi_y + roi_h - 1;
      
      // crop the image:
      image_t::Pointer cropped = crop<image_t>(image, index[0], index[1]);
      
      // rescale the intensities:
      if (invert)
      {
	image_t::PixelType min;
	image_t::PixelType max;
	image_min_max<image_t>(cropped, min, max);
	remap_min_max_inplace<image_t>(cropped, max, min);
      }
      
      std::ostringstream os;
      os << prefix << filein << suffix;
      text_t fileout = os.str();
      save<image_t>(cropped, fileout.c_str());
    }
    else
    {
      // read the 3D image:
      volume_t::Pointer image = load<volume_t>(filein.c_str());
      
      // setup the cropping parameters:
      volume_t::RegionType::SizeType sz =
	image->GetLargestPossibleRegion().GetSize();
      
      volume_t::IndexType index[2];
      index[0][0] = roi_x;
      index[0][1] = roi_y;
      index[0][2] = roi_z;
      index[1][0] = roi_x + roi_w - 1;
      index[1][1] = roi_y + roi_h - 1;
      index[1][2] = roi_z + roi_d - 1;
      
      // crop the image:
      volume_t::Pointer cropped = crop<volume_t>(image, index[0], index[1]);
      
      // rescale the intensities:
      if (invert)
      {
	volume_t::PixelType min;
	volume_t::PixelType max;
	image_min_max<volume_t>(cropped, min, max);
	remap_min_max_inplace<volume_t>(cropped, max, min);
      }
      
      std::ostringstream os;
      os << prefix << filein << suffix;
      text_t fileout = os.str();
      save<volume_t>(cropped, fileout.c_str());
    }
  }
  
  return 0;
}
