/*
Copyright 2004-2007 University of Utah

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/


// File         : img_to_vol.cxx
// Author       : Pavel Aleksandrovich Koshevoy
// Created      : 2005/09/19 14:54
// Copyright    : (C) 2005
// License      : MIT
// Description  : Assemble a series of images into a volume.

// ITK includes:
#include <itkImageSeriesReader.h>
#include <itkImageFileWriter.h>
#include <itkExtractImageFilter.h>

// system includes:
#include <algorithm>
#include <iostream>
#include <iomanip>
#include <string>
#include <vector>
#include <errno.h>
#include <stdlib.h>

// namespace access:
using std::cout;
using std::cerr;
using std::endl;


//----------------------------------------------------------------
// toUInt
// 
// This code was extracted from the_text.cxx
// 
unsigned int
toUInt(const std::string & text, bool * ok = NULL, int base = 10)
{
  char * endptr = NULL;
  unsigned long int num = strtoul(text.c_str(), &endptr, base);
  if (ok != NULL) *ok = !(text == endptr || errno == ERANGE);
  return num;
}

//----------------------------------------------------------------
// split
//
// This code was extracted from the_text.cxx
// 
unsigned int
split(std::vector<std::string> & tokens,
      const std::string & token_string,
      const char & separator,
      const bool & empty_ok = false)
{
  unsigned int string_length = token_string.size();
  if (string_length == 0)
  {
    tokens.resize(0);
    return 0;
  }
  
  // find the separators:
  typedef std::list<unsigned int> list_t;
  list_t separators;
  for (unsigned int i = 0; i < string_length; i++)
  {
    if (token_string[i] != separator) continue;
    separators.push_back(i);
  }
  separators.push_back(string_length);
  
  std::list<std::string> token_list;
  
  typedef std::list<unsigned int>::iterator iter_t;
  unsigned int a = 0;
  for (iter_t i = separators.begin(); i != separators.end(); ++i)
  {
    unsigned int b = *i;
    
    if (b - a == 0 && empty_ok)
    {
      token_list.push_back(std::string(""));
    }
    else if (b - a > 0)
    {
      token_list.push_back(std::string(&token_string[a], b - a));
    }
    
    a = b + 1;
  }
  
  tokens.assign(token_list.begin(), token_list.end());
  return tokens.size();
}


//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const std::string & message = std::string(""))
{
  if (message.size() != 0)
  {
    cerr << "ERROR: " << message << endl;
  }

  cerr << "USAGE: " << argv[0] << " "
       << "[-skip skip_n_outof_n+1_images] "
       << "[-replace a[,b]* ] "
       << "-load image1 [imageN]* "
       << "-save volume "
       << endl
       << "EXAMPLE: " << argv[0] << " "
       << "-skip 9 "
       << "-replace 165,232,250 "
       << "-load *.tif "
       << "-save volume.mha "
       << endl
       << endl
       << "NOTE: " << endl
       << "1. the -replace switch specifies the slices that have to "
       << "be replaced by their nearest neighbor, either preceding "
       << "or following (depending which is available)"
       << endl
       << "2. the -skip switch specifies how many intermediate slices have "
       << "to be discarded from the volume"
       << endl;
  
  exit(1);
}

//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  if (argc < 5) usage(argv, "not enough parameters");
  
  // parse the command line arguments:
  std::vector<std::string> load;
  std::string save("volume.mha");
  
  unsigned int skip = 0;
  unsigned int consume_default_tokens = 0;
  std::list<unsigned int> replace;
  
  for (int i = 1; i < argc; i++)
  {
    bool ok = true;
    
    if (strcmp(argv[i], "-skip") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -skip option");
      
      i++;
      skip = toUInt(argv[i], &ok);
      if (!ok) usage(argv, "bad -skip option");
    }
    else if (strcmp(argv[i], "-replace") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -replace option");
      
      i++;
      
      std::string clump(argv[i]);
      std::vector<std::string> tokens;
      split(tokens, clump, ',');
      
      for (unsigned int j = 0; j < tokens.size(); j++)
      {
	bool ok = true;
	unsigned int index = toUInt(tokens[j], &ok);
	if (!ok)
	{
	  usage(argv,
		std::string("bad index specified for the -replace option: ") +
		tokens[j]);
	}
	
	replace.push_back(index);
      }
    }
    else if (strcmp(argv[i], "-load") == 0)
    {
      if ((argc - i) <= 2) usage(argv, "could not parse -load option");
      consume_default_tokens = 2;
    }
    else if (strcmp(argv[i], "-save") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -save option");
      
      i++;
      save = std::string(argv[i]);
    }
    else if (consume_default_tokens > 0)
    {
      load.push_back(std::string(argv[i]));
      consume_default_tokens++;
    }
    else
    {
      usage(argv, argv[i]);
    }
    
    if (consume_default_tokens > 0)
    {
      consume_default_tokens--;
    }
  }
  
  // sort the file names:
  std::sort(load.begin(), load.end());
  
  // replace some slices by their duplicates:
  for (std::list<unsigned int>::const_iterator
	 iter = replace.begin(); iter != replace.end(); ++iter)
  {
    unsigned int index = *iter;
    
    // find a replacement:
    unsigned int replacement = index - 1;
    if (index == 0)
    {
      for (replacement = 1; replacement < load.size(); replacement++)
      {
	std::list<unsigned int>::const_iterator where =
	  find(replace.begin(), replace.end(), replacement);
	
	if (where == replace.end()) break;
      }
    }
    
    // cout << "replacing " << load[index] << " with " << load[replacement]
    //      << endl;
    load[index] = load[replacement];
  }
  
  //::exit(0);
  
  // remove n out of n+1 file names:
  std::vector<std::string> load_shortened(load.size() / (1 + skip));
  for (unsigned int i = 0, j = 0;
       j < load_shortened.size();
       i = i + 1 + skip, j++)
  {
    load_shortened[j] = load[i];
    cout << std::setw(3) << j << ". loading " << load_shortened[j] << endl;
  }
  
  typedef itk::Image<unsigned char, 3> volume_t;
  //typedef itk::Image<short int, 3> volume_t;
  //typedef itk::Image<float, 3> volume_t;
  typedef itk::ImageSeriesReader<volume_t> reader_t;
  reader_t::Pointer reader = reader_t::New();
  reader->SetFileNames(load_shortened);
  
  try
  {
    reader->Update();
  }
  catch(itk::ExceptionObject & err)
  {
    cerr << "Error loading the series:" << endl << err << endl;
  }
  
  typedef itk::ImageFileWriter<volume_t> writer_t;
  writer_t::Pointer writer = writer_t::New();
  writer->SetFileName(save.c_str());
  writer->SetInput(reader->GetOutput());
  
  try
  { 
    cout << "saving " << save << endl;
    writer->Update();
  }
  catch(itk::ExceptionObject & err)
  {
    cerr << "Error saving the volume:" << endl << err << endl;
  }
  
  return 0;
}
