
#include "itkImageFileWriter.h"
#include "itkImageFileReader.h"
#include "itkImageRegionConstIteratorWithIndex.h"
#include "itkImageRegionIterator.h"
#include "itkNumericTraits.h"
#include <vector>

using namespace std;

//----------------------------------------------------------------
// image_min_max
//
// Find min/max pixel values, return mean pixel value (average):
//
template <typename image_t>
double
image_min_max(const image_t * a,
      typename image_t::PixelType & min,
      typename image_t::PixelType & max)
{
   typedef typename image_t::PixelType pixel_t;
   typedef typename itk::ImageRegionConstIteratorWithIndex<image_t> iter_t;

   min = std::numeric_limits<pixel_t>::max();
   max = -min;

   double mean = 0.0;
   double counter = 0.0;
   iter_t iter(a, a->GetLargestPossibleRegion());
   for (iter.GoToBegin(); !iter.IsAtEnd(); ++iter)
   {
      pixel_t v = iter.Get();
      min = std::min(min, v);
      max = std::max(max, v);

      mean += double(v);
      counter += 1.0;
   }

   mean /= counter;
   return mean;
}

//----------------------------------------------------------------
// remap_min_max_inplace
// 
template <class image_t>
bool
remap_min_max_inplace(image_t * a,
		      typename image_t::PixelType new_min = 0,
		      typename image_t::PixelType new_max = 255)
{
  
  typedef typename image_t::PixelType pixel_t;
  typedef typename itk::ImageRegionIterator<image_t> iter_t;
  
  double min = std::numeric_limits<pixel_t>::max();
  double max = -min;
  
  iter_t iter(a, a->GetLargestPossibleRegion());
  for (iter.GoToBegin(); !iter.IsAtEnd(); ++iter)
  {
    double v = iter.Get();
    min = std::min(min, v);
    max = std::max(max, v);
  }
  
  double rng = max - min;
  if (rng == 0.0) return false;
  
  // rescale the intensities:
  double new_rng = new_max - new_min;
  for (iter.GoToBegin(); !iter.IsAtEnd(); ++iter)
  {
    double t = (iter.Get() - min) / rng;
    //cout<< "pretest: " << iter.Get() << endl;
    iter.
      Set(pixel_t(std::min(double(itk::NumericTraits<pixel_t>::max()),
			   std::max(double(itk::NumericTraits<pixel_t>::min()),
				    new_min + t * new_rng))));
    //cout << "test: " << pixel_t(std::min(double(itk::NumericTraits<pixel_t>::max()), std::max(double(itk::NumericTraits<pixel_t>::min()), new_min + t * new_rng))) << endl ;
  }
  
  return true;
}

//
// load
//
template <class image_t>
typename image_t::Pointer
load(const char * filename, bool blab = true)
{
   typedef typename itk::ImageFileReader<image_t> reader_t;
   typename reader_t::Pointer reader = reader_t::New();
   reader->SetFileName(filename);

   if (blab) cout << "loading " << filename << endl;
   try
   {
      reader->Update();
   }
   catch (itk::ExceptionObject &ex)
   {
      std::cout << "------ Caught expected exception!" << std::endl;
      std::cout << ex;
      exit(1);
   }

   return reader->GetOutput();
}

//
// save
//
template <class image_t>
void
save(const image_t * image, const char * filename, bool blab = true)
{
   typedef typename itk::ImageFileWriter<image_t> writer_t;
   typename writer_t::Pointer writer = writer_t::New();
   writer->SetInput(image);

   if (blab) cout << "saving " << filename << endl;
   writer->SetFileName(filename);
   writer->Update();
   if ( !writer->GetImageIO()->CanReadFile(filename) )
   {
      cout << " ------------ CANNOT READ " << filename << endl ;
      exit(1);
      //return EXIT_FAILURE;
   }
}


