// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : the_document.hxx
// Author       : Pavel A. Koshevoy
// Created      : Tue Apr 04 15:38:30 MDT 2004
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Document framework class.

#ifndef THE_DOCUMENT_HXX_
#define THE_DOCUMENT_HXX_

// local includes:
#include "doc/the_registry.hxx"
#include "doc/the_procedure.hxx"
#include "utils/the_text.hxx"

// forward declarations:
class the_view_t;
class the_bbox_t;


//----------------------------------------------------------------
// the_document_t
// 
class the_document_t
{
public: 
  the_document_t(const the_text_t & name);
  virtual ~the_document_t() {}
 
  virtual the_document_t * clone() const
  { return new the_document_t(*this); }
  
  // regenerate the unrolled procedures geometry:
  virtual bool regenerate();
  
  // draw the unrolled procedures geometry:
  virtual void draw(const the_view_t & view) const;
  
  // calculate the bounding box of unrolled procedures geometry:
  virtual void calc_bbox(const the_view_t & view, the_bbox_t & bbox) const;
  
  // this function splits procedures into active and rolled back:
  void unroll(the_procedure_t * proc);
  
  // append a new procedure to the unrolled procedure list:
  inline void add_proc(the_procedure_t * proc)
  {
    registry().add(proc);
    procs_.push_back(proc->id());
  }
  
  // remove a procedure from the unrolled procedure list:
  inline void del_proc(the_procedure_t * proc)
  {
    procs_.remove(proc->id());
    registry().del(proc);
  }
  
  // accessor to the currently active procedure:
  inline the_procedure_t * active_procedure() const
  {
    if (procs_.empty()) return NULL;
    return registry().elem<the_procedure_t>(procs_.back());
  }
  
  // check whether a procedure of a given type is currently active:
  template <typename proc_t>
  inline proc_t * active_procedure() const
  {
    if (procs_.empty()) return NULL;
    return dynamic_cast<proc_t *>(registry().elem(procs_.back()));
  }
  
  // accessors:
  inline const the_text_t & name() const { return name_; }
  inline       the_text_t & name()       { return name_; }
  
  inline const the_registry_t & registry() const { return registry_; }
  inline       the_registry_t & registry()       { return registry_; }
  
  // accessor to the currently unrolled procedures:
  inline const std::list<unsigned int> & procs() const
  { return procs_; }
  
  inline const std::list<unsigned int> & rolled_back_procs() const
  { return rolled_back_procs_; }
  
  // file io:
  bool save(std::ostream & stream) const;
  bool load(std::istream & stream);
  
private:
  // the name of this document:
  the_text_t name_;
  
  // the document registry:
  the_registry_t registry_;
  
  // unrolled and rolled-back procedures of the document:
  std::list<unsigned int> procs_;
  std::list<unsigned int> rolled_back_procs_;
};


#endif // THE_DOCUMENT_HXX_
