// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : the_document_so.hxx
// Author       : Pavel A. Koshevoy
// Created      : Fri Apr 07 13:37:00 MDT 2004
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Document state object class.

// local includes:
#include "doc/the_document_so.hxx"
#include "doc/the_document.hxx"
#include "utils/the_indentation.hxx"
#include "io/the_file_io.hxx"


//----------------------------------------------------------------
// the_document_so_t::the_document_so_t
// 
the_document_so_t::the_document_so_t(const char * magic):
  changes_saved_(true),
  magic_(magic)
{}

//----------------------------------------------------------------
// the_document_so_t::~the_document_so_t
// 
the_document_so_t::~the_document_so_t()
{
  close_document();
}

//----------------------------------------------------------------
// the_document_so_t::cancel_stack_save
// 
void
the_document_so_t::cancel_stack_save()
{
  assert(document_);
  if (!document_) return;
  
  the_document_so_t::restore_t saved;
  saved.document_ = boost::shared_ptr<the_document_t>(document_->clone());
  saved.undo_ = undo_;
  saved.undo_.push(boost::shared_ptr<the_document_t>(document_->clone()));
  saved.redo_ = redo_;
  ::clear_stack(undo_);
  ::clear_stack(redo_);
  cancel_.push(saved);
}

//----------------------------------------------------------------
// the_document_so_t::cancel_stack_restore
// 
bool
the_document_so_t::cancel_stack_restore()
{
  if (cancel_.empty()) return false;
  
  the_document_so_t::restore_t saved = cancel_.top();
  cancel_.pop();
  saved.undo_.pop();
  undo_ = saved.undo_;
  redo_ = saved.redo_;
  document_ = saved.document_;
  document_->regenerate();
  return true;
}

//----------------------------------------------------------------
// the_document_so_t::cancel_stack_pop
// 
bool
the_document_so_t::cancel_stack_dismiss()
{
  if (cancel_.empty()) return false;
  
  the_document_so_t::restore_t saved = cancel_.top();
  cancel_.pop();
  undo_ = saved.undo_;
  ::clear_stack(redo_);
  return true;
}

//----------------------------------------------------------------
// the_document_so_t::save_undo_record
// 
void
the_document_so_t::save_undo_record()
{
  assert(document_ != NULL);
  undo_.push(boost::shared_ptr<the_document_t>(document_->clone()));
  ::clear_stack(redo_);
  
  // the state of the document is about to change,
  // the file will probably be out of sync:
  changes_saved_ = false;
}

//----------------------------------------------------------------
// the_document_so_t::undo
// 
void
the_document_so_t::undo()
{
  if (undo_.empty()) return;
  
  boost::shared_ptr<the_document_t> undo_document = undo_.top();
  undo_.pop();
  
  boost::shared_ptr<the_document_t> curr_document = document_;
  redo_.push(curr_document);
  
  document_ = undo_document;
  document_->regenerate();
  
  // changes were undone, the file may be out of sync:
  changes_saved_ = false;
}

//----------------------------------------------------------------
// the_document_so_t::redo
// 
void
the_document_so_t::redo()
{
  if (redo_.empty()) return;
  
  boost::shared_ptr<the_document_t> redo_document = redo_.top();
  redo_.pop();
  
  boost::shared_ptr<the_document_t> curr_document = document_;
  undo_.push(curr_document);
  
  document_ = redo_document;
  document_->regenerate();
  
  // changes were redone, the file may be out of sync:
  changes_saved_ = false;
}

//----------------------------------------------------------------
// the_document_so_t::new_document
// 
void
the_document_so_t::new_document(the_document_t * document)
{
  close_document();
  document_ = boost::shared_ptr<the_document_t>(document);
  changes_saved_ = true;
}

//----------------------------------------------------------------
// the_document_so_t::close_document
// 
void
the_document_so_t::close_document()
{
  document_ = boost::shared_ptr<the_document_t>();
  
  ::clear_stack(cancel_);
  clear_undo_redo();
  filename_.clear();
}

//----------------------------------------------------------------
// the_document_so_t::load_document
// 
bool
the_document_so_t::load_document(const the_text_t & filename)
{
  the_document_t * new_doc = NULL;
  bool load_ok = ::load(magic_, filename, new_doc);
  if (!load_ok)
  {
    delete new_doc;
    return false;
  }
  
  new_document(new_doc);
  document()->regenerate();
  // FIXME: document()->registry().assert_sanity();
  
  set_filename(filename);
  return true;
}

//----------------------------------------------------------------
// the_document_so_t::save_document
// 
bool
the_document_so_t::save_document(const the_text_t & filename)
{
  assert(document_ != NULL);
  changes_saved_ = ::save(magic_, filename, document_.get());
  
  if (changes_saved_)
  {
    set_filename(filename);
  }
  
  return changes_saved_;
}

//----------------------------------------------------------------
// the_document_so_t::clear_undo_redo
// 
void
the_document_so_t::clear_undo_redo()
{
  ::clear_stack(undo_);
  ::clear_stack(redo_);
}
