// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : cluster.cxx
// Author       : Pavel A. Koshevoy
// Created      : 2006/04/05 12:36
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Helper functions for identifying pixel clusters in
//                thresholded 1D/2D discrete function/image.

// local includes:
#include "itk/cluster.hxx"


//----------------------------------------------------------------
// identify_clusters_cm
// 
// Identify clusters in the 1D data (anything that's not 0),
// calculate a center of mass for each cluster.
// 
bool
identify_clusters_cm(const double * data,
		     const unsigned int & size,
		     std::list<centerofmass_t<1> > & centerofmass,
		     const bool periodic)
{
  // classify the clusters:
  typedef std::list<int> cluster_1d_t;
  the_dynamic_array_t<cluster_1d_t> clusters;
  std::vector<unsigned int> cluster_map;
  cluster_map.assign(size, ~0);
  
  for (unsigned int i = 0; i < size; i++)
  {
    const double & v = data[i];
    if (v == 0.0) continue;
    
    unsigned int prev = (i + size - 1) % size;
    unsigned int id = cluster_map[prev];
    if (id == ~0u)
    {
      // make a new cluster:
      clusters.append(cluster_1d_t());
      id = clusters.end_index(true);
      
      cluster_map[i] = id;
      clusters[id].push_back(i);
    }
    else
    {
      // add to the existing cluster:
      cluster_map[i] = id;
      clusters[id].push_back(i);
    }
  }
  
  // merge the clusters that are broken up across the periodic boundary:
  if (periodic &&
      cluster_map[0] != ~0u &&
      cluster_map[size - 1] != ~0u)
  {
    unsigned int id = cluster_map[0];
    cluster_1d_t & cluster_a = clusters[id];
    cluster_1d_t & cluster_b = clusters[cluster_map[size - 1]];
    
    if (&cluster_a != &cluster_b)
    {
      while (!cluster_b.empty() )
      {
	int i = remove_head(cluster_b);
	
	cluster_map[i] = id;
	cluster_a.push_back(i - size);
      }
    }
  }
  
  // calculate the center of mass for each cluster:
  for (unsigned int i = 0; i < clusters.size(); i++)
  {
    const cluster_1d_t & cluster = clusters[i];
    if (cluster.empty()) continue;
    
    double mx = 0.0;
    double mt = 0.0;
    
    for (std::list<int>::const_iterator j = cluster.begin();
	 j != cluster.end(); ++j)
    {
      double x = double(*j);
      
      // adjust index for periodicity:
      unsigned int index = (*j + size) % size;
      
      const double & m = data[index];
      mx += m * x;
      mt += m;
    }
    
    centerofmass_t<1> cm;
    cm[0] = fmod(double(size) + mx / mt, double(size));
    cm.mass_ = mt / double(cluster.size());
    centerofmass.push_back(cm);
  }
  
  // sort the max points so that the best candidate is firts:
  centerofmass.sort(std::greater<centerofmass_t<1> >());
  
  return !centerofmass.empty();
}


//----------------------------------------------------------------
// identify_clusters_cm
// 
// Calculate the centers of mass for a given set of
// 2D pixel clusters.
// 
bool
identify_clusters_cm(const image_t * image,
		     const the_dynamic_array_t<cluster_t> & clusters,
		     std::list<centerofmass_t<2> > & centerofmass)
{
  typedef image_t::IndexType index_t;
  
  const unsigned int num_clusters = clusters.size();
  
  for (unsigned int j = 0; j < num_clusters; j++)
  {
    const cluster_t & cluster = clusters[j];
    if (cluster.empty()) continue;
    
    double cluster_min = std::numeric_limits<double>::max();
    double cluster_max = -cluster_min;
    
    for (std::list<index_t>::const_iterator k = cluster.begin();
	 k != cluster.end(); ++k)
    {
      const index_t & index = *k;
      double pixel = double(image->GetPixel(index));
      cluster_min = std::min(pixel, cluster_min);
      cluster_max = std::max(pixel, cluster_max);
    }
    
    double cluster_x = 0.0;
    double cluster_y = 0.0;
    double cluster_m = 0.0;
    
    for (std::list<index_t>::const_iterator k = cluster.begin();
	 k != cluster.end(); ++k)
    {
      const index_t & index = *k;
      double pixel = double(image->GetPixel(index));
      double mass = pixel;
      
      cluster_x += double(index[0]) * mass;
      cluster_y += double(index[1]) * mass;
      cluster_m += mass;
    }
    
    assert(cluster_m != 0.0);
    
    centerofmass_t<2> cm;
    cm[0] = cluster_x / double(cluster_m);
    cm[1] = cluster_y / double(cluster_m);
    cm.mass_ = cluster_m / double(cluster.size());
    centerofmass.push_back(cm);
  }
  
  return !centerofmass.empty();
}
