// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : fft.cxx
// Author       : Pavel A. Koshevoy
// Created      : 2005/06/03 10:16
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Wrapper class and helper functions for working with
//                FFTW3 and ITK images.

// system includes:
#ifndef _USE_MATH_DEFINES
#define _USE_MATH_DEFINES
#endif
#include <string.h>
#include <math.h>

// ITK includes:
#include <itkImage.h>
#include <itkImageRegionIteratorWithIndex.h>
#include <itkImageRegionConstIteratorWithIndex.h>

// local includes:
#include "itk/fft.hxx"
#include "thread/the_mutex_interface.hxx"
#include "utils/the_utils.hxx"


namespace itk_fft
{
  //----------------------------------------------------------------
  // fftw_mutex
  // 
  static the_mutex_interface_t *
  fftw_mutex()
  {
    static the_mutex_interface_t * mutex = the_mutex_interface_t::create();
    return mutex;
  }

  //----------------------------------------------------------------
  // NUM_FFTW_THREADS
  // 
  static std::size_t NUM_FFTW_THREADS = 1;
  
  //----------------------------------------------------------------
  // set_num_fftw_threads
  // 
  // set's number of threads used by fftw, returns previous value:
  std::size_t set_num_fftw_threads(std::size_t num_threads)
  {
    the_lock_t<the_mutex_interface_t> lock(fftw_mutex());
    std::size_t prev = NUM_FFTW_THREADS;
    if (num_threads > 0)
    {
      
      NUM_FFTW_THREADS = num_threads;
    }
    
    return prev;
  }
  
  
  //----------------------------------------------------------------
  // fft_data_t::fft_data_t
  // 
  fft_data_t::fft_data_t(const itk_image_t::Pointer & real):
    data_(NULL),
    nx_(0),
    ny_(0)
  {
    setup(real);
  }
  
  //----------------------------------------------------------------
  // fft_data_t::fft_data_t
  // 
  fft_data_t::fft_data_t(const unsigned int w, const unsigned int h):
    data_(NULL),
    nx_(0),
    ny_(0)
  {
    resize(w, h);
  }
  
  //----------------------------------------------------------------
  // fft_data_t::fft_data_t
  // 
  fft_data_t::fft_data_t(const itk_image_t::Pointer & real,
			 const itk_image_t::Pointer & imag):
    data_(NULL),
    nx_(0),
    ny_(0)
  {
    setup(real, imag);
  }
  
  //----------------------------------------------------------------
  // fft_data_t::fft_data_t
  // 
  fft_data_t::fft_data_t(const fft_data_t & data):
    data_(NULL),
    nx_(0),
    ny_(0)
  {
    *this = data;
  }
  
  //----------------------------------------------------------------
  // fft_data_t::operator =
  // 
  fft_data_t &
  fft_data_t::operator = (const fft_data_t & data)
  {
    assert(this != &data);
    
    if (data.data_ != NULL)
    {
      resize(data.nx_, data.ny_);
      data_ = (fft_complex_t *)(memcpy(data_,
				       data.data_,
				       nx_ * ny_ * sizeof(fft_complex_t)));
    }
    else
    {
      cleanup();
    }
    
    return *this;
  }
  
  //----------------------------------------------------------------
  // fft_data_t::cleanup
  // 
  void
  fft_data_t::cleanup()
  {
    if (data_ != NULL) fftw_free((fft_complex_t *)(data_));
    data_ = NULL;
    nx_ = 0;
    ny_ = 0;
  }
  
  //----------------------------------------------------------------
  // fft_data_t::resize
  // 
  void
  fft_data_t::resize(const unsigned int w, const unsigned int h)
  {
    const unsigned int new_sz = w  * h;
    const unsigned int old_sz = nx_ * ny_;
    if (old_sz == new_sz) return;
    
    if (data_ != NULL) fftw_free((fft_complex_t *)(data_));
    data_ = (fft_complex_t *)(fftw_malloc(new_sz * sizeof(fft_complex_t)));
    assert(data_ != NULL);
    
    nx_ = w;
    ny_ = h;
  }
  
  //----------------------------------------------------------------
  // fft_data_t::fill
  // 
  void
  fft_data_t::fill(const double real, const double imag)
  {
    fftw_complex * hack = reinterpret_cast<fftw_complex *>(data_);
    
    for (unsigned int x = 0; x < nx_; x++)
    {
      for (unsigned int y = 0; y < ny_; y++)
      {
	unsigned int i = y + ny_ * x;
	
	hack[i][0] = real;
	hack[i][1] = imag;
      }
    }
  }
  
  //----------------------------------------------------------------
  // fft_data_t::setup
  // 
  void
  fft_data_t::setup(const itk_image_t::Pointer & real,
		    const itk_image_t::Pointer & imag)
  {
    typedef itk::ImageRegionConstIteratorWithIndex<itk_image_t> itex_t;
    typedef itk_image_t::IndexType index_t;
    
    itk::Size<2> size = real->GetLargestPossibleRegion().GetSize();
    resize(size[0], size[1]);
    
    // shortcut:
    fftw_complex * hack = reinterpret_cast<fftw_complex *>(data_);
    
    // iterate over the real component image:
    itex_t itex(real, real->GetLargestPossibleRegion());
    for (itex.GoToBegin(); !itex.IsAtEnd(); ++itex)
    {
      const index_t index = itex.GetIndex();
      const unsigned int x = index[0];
      const unsigned int y = index[1];
      const unsigned int i = y + ny_ * x;
      
      hack[i][0] = itex.Get();
      hack[i][1] = 0.0;
    }
    
    // iterate over the imaginary component image:
    if (imag.GetPointer() == NULL) return;
    itex = itex_t(imag, real->GetLargestPossibleRegion());
    for (itex.GoToBegin(); !itex.IsAtEnd(); ++itex)
    {
      const index_t index = itex.GetIndex();
      const unsigned int x = index[0];
      const unsigned int y = index[1];
      const unsigned int i = y + ny_ * x;
      
      hack[i][1] = itex.Get();
    }
  }
  
  //----------------------------------------------------------------
  // fft_data_t::component
  // 
  itk_image_t::Pointer
  fft_data_t::component(const bool imag) const
  {
    typedef itk::ImageRegionIteratorWithIndex<itk_image_t> itex_t;
    typedef itk_image_t::IndexType index_t;
    
    itk_image_t::Pointer out = itk_image_t::New();
    itk::Size<2> size;
    size[0] = nx_;
    size[1] = ny_;
    out->SetRegions(size);
    out->Allocate();
    
    // shortcut:
    const fftw_complex * hack = reinterpret_cast<const fftw_complex *>(data_);
    
    // iterate over the image:
    itex_t itex(out, out->GetLargestPossibleRegion());
    for (itex.GoToBegin(); !itex.IsAtEnd(); ++itex)
    {
      const index_t index = itex.GetIndex();
      const unsigned int x = index[0];
      const unsigned int y = index[1];
      const unsigned int i = y + ny_ * x;
      
      itex.Set(hack[i][imag]);
    }
    
    return out;
  }
  
  //----------------------------------------------------------------
  // fft_data_t::apply_lp_filter
  // 
  void
  fft_data_t::apply_lp_filter(const double r, const double s)
  {
    if (r > ::sqrt(2.0)) return;
    
    const unsigned int hx = nx_ / 2;
    const unsigned int hy = ny_ / 2;
    
    const double r0 = (r - s);
    const double r1 = (r + s);
    const double dr = r1 - r0;
    
    for (unsigned int x = 0; x < nx_; x++)
    {
      double sx = 2.0 * (double((x + hx) % nx_) - double(hx)) / double(nx_);
      double x2 = sx * sx;
      
      for (unsigned int y = 0; y < ny_; y++)
      {
	double sy = 2.0 * (double((y + hy) % ny_) - double(hy)) / double(ny_);
	double y2 = sy * sy;
	double d = ::sqrt(x2 + y2);
	
	double v =
	  (d < r0) ? 1.0 :
	  (d > r1) ? 0.0 :
	  (1.0 + cos(M_PI * (d - r0) / dr)) / 2.0;
	
	unsigned int i = y + ny_ * x;
	data_[i] *= v;
      }
    }
  }
  
  
  //----------------------------------------------------------------
  // fft
  // 
  bool
  fft(const itk_image_t::Pointer & in, fft_data_t & out)
  {
    typedef itk::ImageRegionConstIteratorWithIndex<itk_image_t> itex_t;
    typedef itk_image_t::IndexType index_t;
    
    itk::Size<2> size = in->GetLargestPossibleRegion().GetSize();
    const unsigned int & w = size[0];
    const unsigned int & h = size[1];
    
    const unsigned int h_complex = h / 2 + 1;
    const unsigned int h_padded = 2 * h_complex;
    std::vector<double> din(w * h_padded);
    
    // iterate over the image:
    itex_t itex(in, in->GetLargestPossibleRegion());
    for (itex.GoToBegin(); !itex.IsAtEnd(); ++itex)
    {
      const index_t index = itex.GetIndex();
      const unsigned int x = index[0];
      const unsigned int y = index[1];
      const unsigned int i = y + h_padded * x;
      
      din[i] = itex.Get();
    }
    
    fftw_plan plan = NULL;
    
    // FFTW is not re-entrant, except for fftw_execute:
    {
      the_lock_t<the_mutex_interface_t> lock(fftw_mutex());
      fftw_plan_with_nthreads(NUM_FFTW_THREADS);
      plan = fftw_plan_dft_r2c_2d(w,
				  h,
				  &din[0],
				  (fftw_complex *)(&din[0]),
				  FFTW_DESTROY_INPUT | FFTW_ESTIMATE);
    }
    
    if (plan == NULL) return false;
    fftw_execute(plan);
    
    // FFTW is not re-entrant, except for fftw_execute:
    {
      the_lock_t<the_mutex_interface_t> lock(fftw_mutex());
      fftw_destroy_plan(plan);
    }
    
    // fill in the rest of the output data:
    out.resize(w, h);
    for (unsigned int x = 0; x < w; x++)
    {
      for (unsigned int y = 0; y < h_complex; y++)
      {
	unsigned int i = y + h_complex * x;
	out(x, y) = ((fft_complex_t *)(&din[0]))[i];
      }
      
      for (unsigned int y = h_complex; y < h; y++)
      {
	unsigned int i = (h - y) + h_complex * ((w - x) % w);
	out(x, y) = std::conj(((fft_complex_t *)(&din[0]))[i]);
      }
    }
    
    return true;
  }
  
  
  //----------------------------------------------------------------
  // fft
  // 
  bool
  fft(const fft_data_t & in, fft_data_t & out, int sign)
  {
    out.resize(in.nx(), in.ny());
    
    fftw_plan plan = NULL;
    
    // FFTW is not re-entrant, except for fftw_execute:
    {
      the_lock_t<the_mutex_interface_t> lock(fftw_mutex());
      fftw_plan_with_nthreads(NUM_FFTW_THREADS);
      plan = fftw_plan_dft_2d(in.nx(),
			      in.ny(),
			      (fftw_complex *)(in.data()),
			      (fftw_complex *)(out.data()),
			      sign,
			      FFTW_ESTIMATE);
    }
    
    if (plan == NULL) return false;
    fftw_execute(plan);

    // FFTW is not re-entrant, except for fftw_execute:
    {
      the_lock_t<the_mutex_interface_t> lock(fftw_mutex());
      fftw_destroy_plan(plan);
    }
    
    return true;
  }

  //----------------------------------------------------------------
  // ifft
  // 
  bool
  ifft(const fft_data_t & in, fft_data_t & out)
  {
    return fft(in, out, FFTW_BACKWARD);
  }
  
  //----------------------------------------------------------------
  // elem_by_elem
  // 
  void
  elem_by_elem(fn_fft_c_t f,
	       const fft_data_t & in,
	       fft_data_t & out)
  {
    const unsigned nx = in.nx();
    const unsigned ny = in.ny();
    
    out.resize(nx, ny);
    fft_complex_t * dout = out.data();
    const fft_complex_t * din = in.data();
    
    const unsigned int size = nx * ny;
    for (unsigned int i = 0; i < size; i++)
    {
      dout[i] = f(din[i]);
    }
  }
  
  
  //----------------------------------------------------------------
  // elem_by_elem
  //
  void
  elem_by_elem(fft_complex_t(*f)(const double & a,
				 const fft_complex_t & b),
	       const double & a,
	       const fft_data_t & b,
	       fft_data_t & c)
  {
    assert(&b != &c);
    
    const unsigned nx = b.nx();
    const unsigned ny = b.ny();
    c.resize(nx, ny);
    
    fft_complex_t * dc = c.data();
    const fft_complex_t * db = b.data();
    
    const unsigned int size = nx * ny;
    for (unsigned int i = 0; i < size; i++)
    {
      dc[i] = f(a, db[i]);
    }
  }
  
  
  //----------------------------------------------------------------
  // elem_by_elem
  // 
  void
  elem_by_elem(fft_complex_t(*f)(const fft_complex_t & a,
				 const double & b),
	       const fft_data_t & a,
	       const double & b,
	       fft_data_t & c)
  {
    assert(&a != &c);
    
    const unsigned nx = a.nx();
    const unsigned ny = a.ny();
    c.resize(nx, ny);
    
    fft_complex_t * dc = c.data();
    const fft_complex_t * da = a.data();
    
    const unsigned int size = nx * ny;
    for (unsigned int i = 0; i < size; i++)
    {
      dc[i] = f(da[i], b);
    }
  }
  
  
  //----------------------------------------------------------------
  // elem_by_elem
  //
  void
  elem_by_elem(fft_complex_t(*f)(const fft_complex_t & a,
				 const fft_complex_t & b),
	       const fft_complex_t & a,
	       const fft_data_t & b,
	       fft_data_t & c)
  {
    assert(&b != &c);
    
    const unsigned nx = b.nx();
    const unsigned ny = b.ny();
    c.resize(nx, ny);
    
    fft_complex_t * dc = c.data();
    const fft_complex_t * db = b.data();
    
    const unsigned int size = nx * ny;
    for (unsigned int i = 0; i < size; i++)
    {
      dc[i] = f(a, db[i]);
    }
  }
  
  
  //----------------------------------------------------------------
  // elem_by_elem
  // 
  void
  elem_by_elem(fft_complex_t(*f)(const fft_complex_t & a,
				 const fft_complex_t & b),
	       const fft_data_t & a,
	       const fft_complex_t & b,
	       fft_data_t & c)
  {
    assert(&a != &c);
    
    const unsigned nx = a.nx();
    const unsigned ny = a.ny();
    c.resize(nx, ny);
    
    fft_complex_t * dc = c.data();
    const fft_complex_t * da = a.data();
    
    const unsigned int size = nx * ny;
    for (unsigned int i = 0; i < size; i++)
    {
      dc[i] = f(da[i], b);
    }
  }
  
  
  //----------------------------------------------------------------
  // elem_by_elem
  //
  void
  elem_by_elem(fn_fft_cc_t f,
	       const fft_data_t & a,
	       const fft_data_t & b,
	       fft_data_t & c)
  {
    assert(&a != &c);
    assert(&b != &c);
    
    const unsigned nx = a.nx();
    const unsigned ny = a.ny();
    assert(nx == b.nx() && ny == b.ny());
    
    c.resize(nx, ny);
    fft_complex_t * dc = c.data();
    
    const fft_complex_t * da = a.data();
    const fft_complex_t * db = b.data();
    
    const unsigned int size = nx * ny;
    for (unsigned int i = 0; i < size; i++)
    {
      dc[i] = f(da[i], db[i]);
    }
  }
  
} // namespace itk_fft
