// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : fft_common.hxx
// Author       : Pavel A. Koshevoy
// Created      : 2005/11/10 14:05
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Helper functions for image alignment (registration)
//                using phase correlation to find the translation vector.

#ifndef FFT_COMMON_HXX_
#define FFT_COMMON_HXX_

// local includes:
#include "itk/common.hxx"
#include "itk/fft.hxx"
#include "utils/the_dynamic_array.hxx"
#include "utils/the_log.hxx"

// system includes:
#include <math.h>
#include <list>
#include <limits.h>

#ifndef WIN32
#include <unistd.h>
#endif

// namespace access:
using namespace itk_fft;

//----------------------------------------------------------------
// DEBUG_PDF
// 
// #define DEBUG_PDF

#ifdef DEBUG_PDF
// FIXME: this is not thread safe:
#define DEBUG_PDF_PFX the_text_t("PDF-")
#define DEBUG_CLUSTERS
#define DEBUG_MARKERS
extern unsigned int DEBUG_COUNTER1;
extern unsigned int DEBUG_COUNTER2;
#else
#define DEBUG_PDF_PFX the_text_t("")
#endif


//----------------------------------------------------------------
// local_max_t
// 
class local_max_t
{
public:
  local_max_t(const double & value,
	      const double & x,
	      const double & y,
	      const unsigned int & area):
    value_(value),
    x_(x),
    y_(y),
    area_(area)
  {}
  
  inline bool operator == (const local_max_t & lm) const
  { return (value_ == lm.value_) && (x_ == lm.x_) && (y_ == lm.y_); }
  
  inline bool operator < (const local_max_t & lm) const
  { return value_ < lm.value_; }
  
  inline bool operator > (const local_max_t & lm) const
  { return value_ > lm.value_; }
  
  // cluster value:
  double value_;
  
  // center of mass of the cluster:
  double x_;
  double y_;
  
  // area of the cluster:
  unsigned int area_;
};

//----------------------------------------------------------------
// operator << 
// 
inline std::ostream &
operator << (std::ostream & sout, const local_max_t & lm)
{
  return sout << lm.value_ << '\t'
	      << lm.x_ << '\t'
	      << lm.y_ << '\t'
	      << lm.area_ << endl;
}

//----------------------------------------------------------------
// cluster_bbox_t
// 
class cluster_bbox_t
{
public:
  cluster_bbox_t()
  { reset(); }
  
  void reset()
  {
    min_[0] = std::numeric_limits<int>::max();
    min_[1] = std::numeric_limits<int>::max();
    max_[0] = std::numeric_limits<int>::min();
    max_[1] = std::numeric_limits<int>::min();
  }
  
  void update(int x, int y)
  {
    min_[0] = std::min(min_[0], x);
    min_[1] = std::min(min_[1], y);
    max_[0] = std::max(max_[0], x);
    max_[1] = std::max(max_[1], y);
  }
  
  int min_[2];
  int max_[2];
};

//----------------------------------------------------------------
// find_maxima_cm
//
// The percentage below refers to the number of pixels that fall
// below the maxima. Thus, the number of pixels above the maxima
// is 1 - percentage. This way it is possible to specify a
// thresholding value without knowing anything about the image.
// 
// The given image is thresholded, the resulting clusters/blobs
// are identified/classified, the center of mass of each cluster
// is treated as a maxima in the image.
//
// Returns the number of maxima found.
// 
extern unsigned int
find_maxima_cm(std::list<local_max_t> & max_list,
	       const itk_image_t::Pointer & image,
	       const double percentage = 0.9995,
	       const the_text_t & prefix = DEBUG_PDF_PFX,
	       const the_text_t & suffix = the_text_t(".png"));


//----------------------------------------------------------------
// find_correlation
// 
template <class TImage>
unsigned int
find_correlation(std::list<local_max_t> & max_list,
		 const TImage * fi,
		 const TImage * mi,
		 double lp_filter_r,
		 double lp_filter_s)
{
  itk_image_t::Pointer z0 = cast<TImage, itk_image_t>(fi);
  itk_image_t::Pointer z1 = cast<TImage, itk_image_t>(mi);
  
  itk_image_t::SizeType max_sz = calc_padding<itk_image_t>(z0, z1);
  z0 = pad<itk_image_t>(z0, max_sz);
  z1 = pad<itk_image_t>(z1, max_sz);
  
  fft_data_t f0;
  fft(z0, f0);
  
  // resampled data requires less smoothing:
  f0.apply_lp_filter(lp_filter_r, lp_filter_s);
  
  fft_data_t f1;
  fft(z1, f1);
  
  // resampled data requires less smoothing:
  f1.apply_lp_filter(lp_filter_r, lp_filter_s);
  
  const unsigned int & nx = f0.nx();
  const unsigned int & ny = f0.ny();
  fft_data_t P(nx, ny);
  
  for (unsigned int x = 0; x < nx; x++)
  {
    for (unsigned int y = 0; y < ny; y++)
    {
#if 1
      // Girod-Kuo, normalized cross power spectrum,
      // corresponds to phase correlation in spatial domain:
      fft_complex_t p10 = f1(x, y) * std::conj(f0(x, y));
      P(x, y) =	p10 / (std::sqrt(p10 * std::conj(p10)) + 1e-8);
#else
      // cross power spectrum,
      // corresponds to cross correlation in spatial domain:
      P(x, y) = f1(x, y) * std::conj(f0(x, y));
#endif
    }
  }
  
  // resampled data produces less noisy PDF and requires less smoothing:
  P.apply_lp_filter(lp_filter_r * 0.8, lp_filter_s);
  
  // calculate the displacement probability density function:
  fft_data_t ifft_P;
  
#ifndef NDEBUG // get around an annoying compiler warning:
  bool ok =
#endif
    ifft(P, ifft_P);
  assert(ok);
  
  itk_image_t::Pointer PDF = ifft_P.real();
  
  // look for the maxima in the PDF:
  double area = double(max_sz[0] * max_sz[1]);
  
  // a minimum of 5 pixels and a maximum of 64 pixels may be attributed
  // to local maxima in the image:
  // double fraction = std::min(64.0 / area, std::max(5.0 / area, 1e-3));
  double fraction = std::min(64.0 / area, std::max(5.0 / area, 1e-2));
  
  // the entire image should never be treated as a maxima cluster:
  assert(fraction < 1.0);
  
  // find the maxima clusters:
#ifdef DEBUG_PDF
  the_text_t pfx = (the_text_t::number(DEBUG_COUNTER1, 3, '0') +
		    the_text_t("-PDF-"));
#else
  const the_text_t & pfx = DEBUG_PDF_PFX;
#endif
  the_text_t sfx(".png");
  
  unsigned int found = find_maxima_cm(max_list, PDF, 1.0 - fraction, pfx);
  return found;
}

//----------------------------------------------------------------
// find_correlation
//
// This is a backwards compatibility API
// 
template <class TImage>
unsigned int
find_correlation(const TImage * fi,
		 const TImage * mi,
		 std::list<local_max_t> & max_list,
		 bool resampled_data)
{
  double lp_filter_r = resampled_data ? 0.9 : 0.5;
  return find_correlation<TImage>(max_list, fi, mi, lp_filter_r, 0.1);
}

//----------------------------------------------------------------
// threshold_maxima
// 
// Discard maxima whose mass is below a given threshold ratio
// of the total mass of all maxima:
// 
void
threshold_maxima(std::list<local_max_t> & max_list,
		 const double & threshold);

//----------------------------------------------------------------
// reject_negligible_maxima
// 
// Discard maxima that are worse than the best maxima by a factor
// greater than the given threshold ratio:
// 
// Returns the size of the new list.
// 
unsigned int
reject_negligible_maxima(std::list<local_max_t> & max_list,
			 const double & threshold);

//----------------------------------------------------------------
// overlap_t
// 
class overlap_t
{
public:
  overlap_t():
    id_(~0),
    overlap_(0.0)
  {}
  
  overlap_t(const unsigned int & id, const double & overlap):
    id_(id),
    overlap_(overlap)
  {}
  
  inline bool operator == (const overlap_t & d) const
  { return id_ == d.id_; }
  
  inline bool operator < (const overlap_t & d) const
  { return overlap_ < d.overlap_; }
  
  inline bool operator > (const overlap_t & d) const
  { return overlap_ > d.overlap_; }
  
  unsigned int id_;
  double overlap_;
};

//----------------------------------------------------------------
// reject_negligible_overlap
// 
void
reject_negligible_overlap(std::list<overlap_t> & ol,
			  const double & threshold);


//----------------------------------------------------------------
// estimate_displacement
//
template <class TImage>
double
estimate_displacement(the_log_t & log,
		      const TImage * a,
		      const TImage * b,
		      const local_max_t & lm,
		      translate_transform_t::Pointer & transform,
		      const double overlap_min = 0.0,
		      const double overlap_max = 1.0,
		      const mask_t * mask_a = NULL,
		      const mask_t * mask_b = NULL)
{
  // FIXME:
#ifdef DEBUG_PDF
  the_text_t fn_debug =
    the_text_t::number(DEBUG_COUNTER1, 3, '0') + the_text_t("-") +
    the_text_t::number(DEBUG_COUNTER2, 1, '0') + the_text_t("-");
#endif
  
  itk_image_t::SizeType max_sz = calc_padding<TImage>(a, b);
  const unsigned int & w = max_sz[0];
  const unsigned int & h = max_sz[1];
  
  // evaluate 4 permutations of the maxima:
  // typedef itk::NormalizedCorrelationImageToImageMetric<TImage, TImage>
  typedef itk::MeanSquaresImageToImageMetric<TImage, TImage> metric_t;
  typedef itk::LinearInterpolateImageFunction<TImage, double> interpolator_t;
  
  typedef typename TImage::SpacingType spacing_t;
  spacing_t spacing = a->GetSpacing();
  double sx = spacing[0];
  double sy = spacing[1];
  
  // evaluate 4 permutation of the maxima:
  const double x = lm.x_;
  const double y = lm.y_;
  
  const vec2d_t t[] = {
    vec2d(sx * x, sy * y),
    vec2d(sx * (x - w), sy * y),
    vec2d(sx * x, sy * (y - h)),
    vec2d(sx * (x - w), sy * (y - h))
  };
  
  double best_metric = std::numeric_limits<double>::max();
  for (unsigned int i = 0; i < 4; i++)
  {
    translate_transform_t::Pointer ti = translate_transform_t::New();
    ti->SetOffset(t[i]);
    
    // FIXME:
#ifdef DEBUG_PDF
    the_text_t fn = fn_debug + the_text_t::number(i) + ".png";
    save_rgb<TImage>(fn, a, b, ti, mask_a, mask_b);
#endif
    
    const double area_ratio = overlap_ratio<TImage>(a, b, ti);
    int old_precision = log.precision();
    log.precision(2);
    log << i << ": " << setw(3) << std::fixed << area_ratio * 100.0
	<< "% of overlap, ";
    log.precision(old_precision);
    
    if (area_ratio < overlap_min || area_ratio > overlap_max)
    {
      log << "skipping..." << endl;
      continue;
    }
    
    // double metric = eval_metric<metric_t, interpolator_t>(ti, a, b);
    double metric = my_metric<TImage>(a,
				      b,
				      ti,
				      mask_a,
				      mask_b,
				      overlap_min,
				      overlap_max);
    log << std::scientific << metric;
    if (metric < best_metric)
    {
      transform = ti;
      best_metric = metric;
      log << " - better..." << endl;
      
#ifdef DEBUG_PDF
      save_rgb<TImage>(fn_debug + the_text_t::number(i) + "-better.png",
			a, b, ti, mask_a, mask_b);
#endif
    }
    else
    {
      log << " - worse..." << endl;
    }
  }
  
  return best_metric;
}


//----------------------------------------------------------------
// match_one_pair
// 
// match two images, find the best matching transform and
// return the corresponding match metric value
// 
template <typename TImage, typename TMask>
double
match_one_pair(the_log_t & log,
	       const bool & images_were_resampled,
	       const bool & use_std_mask,
	       const TImage * fi,
	       const TImage * mi,
	       const TMask * fi_mask,
	       const TMask * mi_mask,
	       
	       const double & overlap_min,
	       const double & overlap_max,
	       
	       translate_transform_t::Pointer & ti,
	       std::list<local_max_t> & peaks,
	       unsigned int & num_peaks,
	       
	       // ideally this should be one, but radial distortion may
	       // generate several valid peaks (up to 4), so it may be
	       // necessary to consider more peaks for the unmatched images:
	       const unsigned int max_peaks)
{
#ifdef DEBUG_PDF
  DEBUG_COUNTER1++;
#endif
  
  ti = NULL;
  
  unsigned int total_peaks = 0;
  if (use_std_mask)
  {
    // ugh, blank out 5 percent of the image at the bottom
    // to cover up the image id:
    typename TImage::SizeType fi_sz = fi->GetLargestPossibleRegion().GetSize();
    typename TImage::SizeType mi_sz = mi->GetLargestPossibleRegion().GetSize();
    
    unsigned int fi_y = (unsigned int)(0.95 * fi_sz[1]);
    unsigned int mi_y = (unsigned int)(0.95 * mi_sz[1]);

    typename TImage::Pointer fi_filled = cast<TImage, TImage>(fi);
    fill<TImage>(fi_filled, 0, fi_y, fi_sz[0], fi_sz[1] - fi_y, 0);
    
    typename TImage::Pointer mi_filled = cast<TImage, TImage>(mi);
    fill<TImage>(mi_filled, 0, mi_y, mi_sz[0], mi_sz[1] - mi_y, 0);
    
    total_peaks = find_correlation<TImage>(fi_filled,
					   mi_filled,
					   peaks,
					   images_were_resampled);
  }
  else
  {
    total_peaks = find_correlation<TImage>(fi,
					   mi,
					   peaks,
					   images_were_resampled);
  }
  
  num_peaks = reject_negligible_maxima(peaks, 3.0);
  log << num_peaks << '/' << total_peaks << " eligible peaks, ";
  
  if (num_peaks > max_peaks)
  {
    log << "skipping..." << endl;
    return std::numeric_limits<double>::max();
  }
  
  // choose the best peak:
  double best_metric = std::numeric_limits<double>::max();

#ifdef DEBUG_PDF
  DEBUG_COUNTER2 = 0;
#endif
  for (std::list<local_max_t>::iterator j = peaks.begin();
       j != peaks.end(); ++j)
  {
    log << "evaluating permutations..." << endl;
    const local_max_t & lm = *j;
    
    translate_transform_t::Pointer tmp = translate_transform_t::New();
    double metric = estimate_displacement<TImage>(log,
						  fi,
						  mi,
						  lm,
						  tmp,
						  overlap_min,
						  overlap_max,
						  fi_mask,
						  mi_mask);
    if (metric < best_metric)
    {
      best_metric = metric;
      ti = tmp;
    }
    
#ifdef DEBUG_PDF
    DEBUG_COUNTER2++;
#endif
  }
  
  return best_metric;
}


//----------------------------------------------------------------
// match_one_pair
// 
// match two images, find the best matching transform and
// return the corresponding match metric value
// 
template <typename TImage, typename TMask>
double
match_one_pair(the_log_t & log,
	       const bool & images_were_resampled,
	       const bool & use_std_mask,
	       const TImage * fi,
	       const TImage * mi,
	       const TMask * fi_mask,
	       const TMask * mi_mask,
	       
	       const double & overlap_min,
	       const double & overlap_max,
	       
	       const unsigned int & node_id,
	       translate_transform_t::Pointer & ti,
	       std::pair<unsigned int, std::list<local_max_t> > & peaks,
	       
	       // ideally this should be one, but radial distortion may
	       // generate several valid peaks (up to 4), so it may be
	       // necessary to consider more peaks for the unmatched images:
	       const unsigned int max_peaks)
{
  unsigned int peak_list_size = 0;
  std::list<local_max_t> peak_list;
  double m = match_one_pair<TImage, TMask>(log,
					   images_were_resampled,
					   use_std_mask,
					   fi,
					   mi,
					   fi_mask,
					   mi_mask,
					   overlap_min,
					   overlap_max,
					   ti,
					   peak_list,
					   peak_list_size,
					   max_peaks);
  
  // this info will be used when trying to match the unmatched images:
  if (peak_list_size != 0 &&
      peak_list_size <= max_peaks &&
      (peaks.second.empty() || peaks.second.size() > peak_list_size))
  {
    peaks.first = node_id;
    peaks.second.clear();
    peaks.second.splice(peaks.second.end(), peak_list);
  }
  
  return m;
}

//----------------------------------------------------------------
// match_one_pair
// 
template <typename TImage, typename TMask>
double
match_one_pair(the_log_t & log,
	       const bool & images_were_resampled,
	       const bool & use_std_mask,
	       const TImage * fi,
	       const TImage * mi,
	       const TMask * fi_mask,
	       const TMask * mi_mask,
	       const double & overlap_min,
	       const double & overlap_max,
	       translate_transform_t::Pointer & ti)
{
  std::list<local_max_t> peaks;
  unsigned int num_peaks = 0;
  return match_one_pair<TImage, TMask>(log,
				       images_were_resampled,
				       use_std_mask,
				       fi,
				       mi,
				       fi_mask,
				       mi_mask,
				       overlap_min,
				       overlap_max,
				       ti,
				       peaks,
				       num_peaks,
				       UINT_MAX);
}


#endif // FFT_COMMON_HXX_
