#ifndef __itkBlobImageFilterWithMask_h
#define __itkBlobImageFilterWithMask_h

#include "itkImageToImageFilter.h"
#include "itkNumericTraits.h"
#include "itkArray.h"

namespace itk
{
  
/** \class BlobImageFilterWithMask
 * \brief Implements comparison between two images.
 *
 * This filter is used by the testing system to compute the difference between
 * a valid image and an image produced by the test. The comparison value is
 * computed by visiting all the pixels in the baseline images and comparing
 * their values with the pixel values in the neighborhood of the homologous
 * pixel in the other image.
 * 
 * \ingroup IntensityImageFilters   Multithreaded
 */
template <class TInputImage>
class ITK_EXPORT BlobImageFilterWithMask :
    public ImageToImageFilter<TInputImage, TInputImage> 
{
public:
  /** Standard class typedefs. */
  typedef BlobImageFilterWithMask  Self;
  typedef ImageToImageFilter<TInputImage, TInputImage>  Superclass;
  typedef SmartPointer<Self>   Pointer;
  typedef SmartPointer<const Self>  ConstPointer;

  /** Method for creation through the object factory. */
  itkNewMacro(Self);
  
  /** Run-time type information (and related methods). */
  itkTypeMacro(BlobImageFilterWithMask, ImageToImageFilter);
  
  /** Some convenient typedefs. */
  typedef TInputImage InputImageType;
  typedef TInputImage OutputImageType;
  typedef typename OutputImageType::PixelType OutputPixelType;
  typedef typename OutputImageType::RegionType OutputImageRegionType;
  typedef typename NumericTraits<OutputPixelType>::RealType RealType;
  
  /**  Type for the mask of the fixed image. Only pixels that are "inside"
       this mask will be considered for the computation of the metric */
  typedef SpatialObject<TInputImage::ImageDimension> ImageMaskType;
  typedef typename ImageMaskType::Pointer ImageMaskPointer;
  
  /** Set/Get the image mask. */
  itkSetObjectMacro(ImageMask, ImageMaskType);
  itkGetConstObjectMacro(ImageMask, ImageMaskType);
  
  /** Set/Get the maximum distance away to look for a matching pixel.
      Default is 0. */
  itkSetMacro(NeighborhoodRadius, int);
  itkGetMacro(NeighborhoodRadius, int);
  
  /** Set/Get the minimum threshold for pixels to be different.
      Default is 0. */
  itkSetMacro(RatioThreshold, double);
  itkGetMacro(RatioThreshold, double);
  
protected:
  BlobImageFilterWithMask();
  virtual ~BlobImageFilterWithMask() {}
  
  void PrintSelf(std::ostream& os, Indent indent) const;

  /** BlobImageFilterWithMask can be implemented as a multithreaded
   * filter.  Therefore, this implementation provides a
   * ThreadedGenerateData() routine which is called for each
   * processing thread. The output image data is allocated
   * automatically by the superclass prior to calling
   * ThreadedGenerateData().  ThreadedGenerateData can only write to
   * the portion of the output image specified by the parameter
   * "outputRegionForThread"
   *
   * \sa ImageToImageFilter::ThreadedGenerateData(),
   *     ImageToImageFilter::GenerateData()  */
  void ThreadedGenerateData(const OutputImageRegionType& threadRegion,
                            int threadId);
  
  void BeforeThreadedGenerateData();
  void AfterThreadedGenerateData();
  
  double m_RatioThreshold;
  int m_NeighborhoodRadius;
  
  Array<unsigned int>  m_ThreadVarianceMass;
  Array<double>        m_ThreadMeanGlobalIntensity;
  
private:
  BlobImageFilterWithMask(const Self&); //purposely not implemented
  void operator=(const Self&); //purposely not implemented

  bool m_IgnoreBoundaryPixels;
};

} // end namespace itk

// Define instantiation macro for this template.
#define ITK_TEMPLATE_BlobImageFilterWithMask(_, EXPORT, x, y)		\
  namespace itk								\
  {									\
    _(2(class EXPORT BlobImageFilterWithMask< ITK_TEMPLATE_2 x >))	\
      namespace Templates						\
    {									\
      typedef BlobImageFilterWithMask< ITK_TEMPLATE_2 x >		\
	BlobImageFilterWithMask##y;					\
    }									\
  }

#if ITK_TEMPLATE_EXPLICIT
# include "Templates/itkBlobImageFilterWithMask+-.h"
#endif

#if ITK_TEMPLATE_TXX
# include "itkBlobImageFilterWithMask.txx"
#endif

#endif
