// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : stos.hxx
// Author       : Pavel A. Koshevoy
// Created      : 2006/06/23 13:40
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Slice to slice registration container class.

#ifndef STOS_HXX_
#define STOS_HXX_

// system includes:
#include <iostream>
#include <iomanip>
#include <fstream>

// local includes:
#include "itk/common.hxx"
#include "utils/the_text.hxx"


//----------------------------------------------------------------
// stos_t
//
template <typename TImage = image_t>
class stos_t
{
public:
  stos_t() {}
  
  stos_t(const the_text_t & fn_s0,
	 const the_text_t & fn_s1,
	 const the_text_t & fn_mask_s0,
	 const the_text_t & fn_mask_s1,
	 const bool & flipped_0,
	 const bool & flipped_1,
	 const TImage * s0,
	 const TImage * s1,
	 const unsigned int & shrink_factor,
	 const base_transform_t * t01):
    t01_(t01)
  {
    fn_[0] = fn_s0;
    fn_[1] = fn_s1;
    
    fn_mask_[0] = fn_mask_s0;
    fn_mask_[1] = fn_mask_s1;
    
    flipped_[0] = flipped_0;
    flipped_[1] = flipped_1;
    
    sp_[0] = s0->GetSpacing();
    sp_[1] = s1->GetSpacing();
    
    sz_[0] = s0->GetLargestPossibleRegion().GetSize();
    sz_[1] = s1->GetLargestPossibleRegion().GetSize();
    
    sp_[0][0] /= shrink_factor;
    sp_[0][1] /= shrink_factor;
    sp_[1][0] /= shrink_factor;
    sp_[1][1] /= shrink_factor;
    
    sz_[0][0] *= shrink_factor;
    sz_[0][1] *= shrink_factor;
    sz_[1][0] *= shrink_factor;
    sz_[1][1] *= shrink_factor;
  }
  
  stos_t(const the_text_t & fn_s0,
	 const the_text_t & fn_s1,
	 const the_text_t & fn_mask_s0,
	 const the_text_t & fn_mask_s1,
	 const bool & flipped_0,
	 const bool & flipped_1,
	 const double spacing_x0,
	 const double spacing_y0,
	 const double spacing_x1,
	 const double spacing_y1,
	 const unsigned int size_x0,
	 const unsigned int size_y0,
	 const unsigned int size_x1,
	 const unsigned int size_y1,
	 const base_transform_t * t01):
    t01_(t01)
  {
    fn_[0] = fn_s0;
    fn_[1] = fn_s1;
    
    fn_mask_[0] = fn_mask_s0;
    fn_mask_[1] = fn_mask_s1;
    
    flipped_[0] = flipped_0;
    flipped_[1] = flipped_1;
    
    sp_[0][0] = spacing_x0;
    sp_[0][1] = spacing_y0;
    sp_[1][0] = spacing_x1;
    sp_[1][1] = spacing_y1;
    
    sz_[0][0] = size_x0;
    sz_[0][1] = size_y0;
    sz_[1][0] = size_x1;
    sz_[1][1] = size_y1;
  }
  
  stos_t(const char * fn_load)
  {
    load(fn_load);
  }
  
  // load the filenames of the slices and the corresponding transform:
  bool load(const char * fn_load, const bool blab = true)
  {
    std::fstream si;
    si.open(fn_load, ios::in);
    if (!si.is_open())
    {
      cerr << "WARNING: could not open " << fn_load
	   << " for reading, skipping...." << endl;
      return false;
    }
    
    fn_load_.assign(fn_load);
    if (blab) cout << "loading " << fn_load_ << endl;
    
    bool ok = load(si);
    si.close();
    
    return ok;
  }
  
  bool load(std::istream & si)
  {
    getline(si, fn_[0]);
    getline(si, fn_[1]);

    int sz[2][2];
    si >> flipped_[0] >> flipped_[1]
       >> sp_[0][0] >> sp_[0][1] >> sz[0][0] >> sz[0][1]
       >> sp_[1][0] >> sp_[1][1] >> sz[1][0] >> sz[1][1];
    
    sz_[0][0] = sz[0][0];
    sz_[0][1] = sz[0][1];
    sz_[1][0] = sz[1][0];
    sz_[1][1] = sz[1][1];
    
    itk::TransformBase::Pointer tmp = load_transform(si);
    
    t01_ = dynamic_cast<base_transform_t *>(tmp.GetPointer());
    bool ok = t01_.GetPointer() != NULL;
    
    fn_mask_[0].clear();
    fn_mask_[1].clear();
    
    // skip to the next line (by skipping white space):
    si >> std::ws;
    
    the_text_t optional;
    getline(si, optional);
    if (optional == "two_user_supplied_masks:")
    {
      getline(si, fn_mask_[0]);
      getline(si, fn_mask_[1]);
    }
    
    return ok;
  }
  
  // save this slice-to-slice registration:
  bool save(const char * fn_save, const bool blab = true) const
  {
    std::fstream so;
    so.open(fn_save, ios::out);
    if (!so.is_open())
    {
      cerr << "WARNING: could not open " << fn_save
	   << " for writing, skipping...." << endl;
      return false;
    }
    if (blab) cout << "saving " << fn_save << endl;
    
    save(so);
    so.close();
    
    // FIXME: not sure if this is useful/necessary:
    fn_load_.assign(fn_save);
    
    return true;
  }
  
  void save(std::ostream & so) const
  {
    ios::fmtflags old_flags = so.setf(ios::scientific);
    int old_precision = so.precision();
    so.precision(12);
    
    so << fn_[0].text() << endl
       << fn_[1].text() << endl
       << flipped_[0] << endl
       << flipped_[1] << endl
       << sp_[0][0] << '\t'
       << sp_[0][1] << '\t'
       << int(sz_[0][0]) << '\t'
       << int(sz_[0][1]) << endl
       << sp_[1][0] << '\t'
       << sp_[1][1] << '\t'
       << int(sz_[1][0]) << '\t'
       << int(sz_[1][1]) << endl;
    
    save_transform(so, t01_);
    so << endl;
    
    if (!fn_mask_[0].is_empty() && !fn_mask_[1].is_empty())
    {
      so << "two_user_supplied_masks:" << endl
	 << fn_mask_[0].text() << endl
	 << fn_mask_[1].text() << endl;
    }
    
    so.setf(old_flags);
    so.precision(old_precision);
  }
  
  // convenience functions:
  inline pnt2d_t tile_min(unsigned int /* i */) const
  { return pnt2d(0, 0); }
  
  inline pnt2d_t tile_max(unsigned int i) const
  {
    return pnt2d(sp_[i][0] * double(sz_[i][0]),
		 sp_[i][1] * double(sz_[i][1]));
  }
  
  inline pnt2d_t tile_center(unsigned int i) const
  {
    pnt2d_t min = tile_min(i);
    pnt2d_t max = tile_max(i);
    return min + (max - min) * 0.5;
  }
  
  // datamembers:
  mutable the_text_t fn_load_;
  the_text_t fn_[2]; // filenames of the fixed/moving slices
  the_text_t fn_mask_[2]; // filenames of slice masks, optional
  bool flipped_[2];
  typename TImage::SpacingType sp_[2];
  typename TImage::RegionType::SizeType sz_[2];
  base_transform_t::ConstPointer t01_;
};

//----------------------------------------------------------------
// operator <<
// 
template <typename TImage>
inline static std::ostream &
operator << (std::ostream & so, const stos_t<TImage> & stos)
{
  stos.save(so);
  return so;
}

//----------------------------------------------------------------
// save_stos
//
template <typename TImage>
inline static bool
save_stos(const the_text_t & fn_save,
	  const the_text_t & fn_s0,
	  const the_text_t & fn_s1,
	  const the_text_t & fn_mask_s0,
	  const the_text_t & fn_mask_s1,
	  const bool & flipped_0,
	  const bool & flipped_1,
	  const TImage * s0,
	  const TImage * s1,
	  const unsigned int & shrink_factor,
	  const base_transform_t * transform,
	  const bool blab = true)
{
  stos_t<TImage> stos(fn_s0,
		      fn_s1,
		      fn_mask_s0,
		      fn_mask_s1,
		      flipped_0,
		      flipped_1,
		      s0,
		      s1,
		      shrink_factor,
		      transform);
  return stos.save(fn_save, blab);
}


#endif // STOS_HXX_
