// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : stos_common.cxx
// Author       : Pavel A. Koshevoy
// Created      : 2006/05/31 14:29
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Helper functions for slice to slice registration.

// local includes:
#include "itk/common.hxx"

// system includes:
#include <fstream>


//----------------------------------------------------------------
// load_slice
// 
bool
load_slice(const char * fn_mosaic,
	   const bool & flip_mosaic,
	   const unsigned int & shrink_factor,
	   const double & clahe_slope,
	   image_t::Pointer & mosaic,
	   mask_t::Pointer & mosaic_mask,
	   bool mosaic_mask_enabled,
	   bool dont_allocate)
{
  cout << "    reading " << fn_mosaic;
  
  std::fstream fin;
  fin.open(fn_mosaic, ios::in);
  if (!fin.is_open())
  {
    cout << ", failed...." << endl;
    return false;
  }
  cout << endl;
  
  std::list<the_text_t> fn_tiles;
  std::vector<base_transform_t::Pointer> transform;
  std::vector<image_t::ConstPointer> image;
  std::vector<mask_t::ConstPointer> mask;
  
  double pixel_spacing = 1.0;
  bool use_std_mask = false;
  load_mosaic<base_transform_t>(fin,
				pixel_spacing,
				use_std_mask,
				fn_tiles,
				transform);
  fin.close();
  
  unsigned int num_images = transform.size();
  assert(num_images > 1);
  assert(pixel_spacing != 0);
  
  image.resize(num_images);
  mask.resize(num_images);
  
  unsigned int i = 0;
  for (std::list<the_text_t>::const_iterator iter = fn_tiles.begin();
       iter != fn_tiles.end(); ++iter, i++)
  {
    const the_text_t & fn_image = *iter;
    
    // read the image:
    cout << setw(3) << i << " ";
    image[i] = std_tile<image_t>(fn_image, shrink_factor, pixel_spacing);
    
    if (use_std_mask)
    {
      mask[i] = std_mask<image_t>(image[i]);
    }
  }
  
  // assempble the mosaic:
  cout << "    assembling a mosaic " << endl;
  mosaic = make_mosaic<image_t::ConstPointer, base_transform_t::Pointer>
    (FEATHER_BLEND_E, transform, image, mask, dont_allocate);
  
  // reset the tile image origin and spacing:
  image_t::PointType origin = mosaic->GetOrigin();
  origin[0] = 0;
  origin[1] = 0;
  mosaic->SetOrigin(origin);
  
  if (mosaic_mask_enabled)
  {
    cout << "    assembling a mosaic mask " << endl;
    mosaic_mask = make_mask<base_transform_t::Pointer>(transform, mask);
    mosaic_mask->SetOrigin(origin);
  }
  
  // turn over the mosaic:
  if (flip_mosaic)
  {
    mosaic = flip<image_t>(mosaic);
    
    if (mosaic_mask_enabled)
    {
      mosaic_mask = flip<mask_t>(mosaic_mask);
    }
  }
  
  // preprocess the image with Contrast Limited Adaptive Histogram
  // Equalization algorithm, remap the intensities into the [0, 1] range:
  if (clahe_slope > 1.0)
  {
    cout << "    enhancing contrast with CLAHE " << endl;
    mosaic = CLAHE<image_t>(mosaic, 64, 64, clahe_slope, 256, 0.0, 1.0);
  }
  
  return true;
}
