// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : stos_common.hxx
// Author       : Pavel A. Koshevoy
// Created      : 2006/05/31 14:29
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Helper functions for slice to slice registration.

#ifndef STOS_COMMON_HXX_
#define STOS_COMMON_HXX_

// local includes:
#include "itk/stos.hxx"
#include "itk/common.hxx"

// system includes:
#include <fstream>


//----------------------------------------------------------------
// sort
// 
// Sort the list so that the slices are sequential.
// 
// Return true if the list was sorted successfully.
// 
// Return false otherwise (may happen if there are
// disconnects or duplicates present).
//
template <typename TImage>
bool
sort(std::list<stos_t<TImage> > & stos)
{
  if (stos.empty()) return true;
  
  std::list<stos_t<TImage> > sorted;
  sorted.push_back(remove_head(stos));
  
  while (!stos.empty())
  {
    // find a match for the head:
    bool found_head_match = false;
    for (typename std::list<stos_t<TImage> >::iterator i = stos.begin();
	 i != stos.end();)
    {
      const the_text_t & head = sorted.front().fn_[0];
      const the_text_t & tail = (*i).fn_[1];
      if (head == tail)
      {
	found_head_match = true;
	sorted.push_front(*i);
	i = stos.erase(i);
      }
      else
      {
	++i;
      }
    }
    
    // find a match for the tail:
    bool found_tail_match = false;
    for (typename std::list<stos_t<TImage> >::iterator i = stos.begin();
	 i != stos.end();)
    {
      const the_text_t & tail = sorted.back().fn_[1];
      const the_text_t & head = (*i).fn_[0];
      if (head == tail)
      {
	found_tail_match = true;
	sorted.push_back(*i);
	i = stos.erase(i);
      }
      else
      {
	++i;
      }
    }
    
    if (!found_head_match && !found_tail_match)
    {
      // could not find a match for either head or tail:
      break;
    }
  }
  
  if (!stos.empty()) return false;
  
  stos.splice(stos.end(), sorted);
  return true;
}


//----------------------------------------------------------------
// load_mosaic
// 
template <typename T>
bool
load_mosaic(const char * fn_mosaic,
	    const bool & flip_mosaic,
	    const unsigned int & shrink_factor,
	    const bool & assemble_mosaic,
	    typename T::Pointer & mosaic,
	    const bool & assemble_mosaic_mask,
	    mask_t::Pointer & mosaic_mask,
	    const double & clahe_slope = 1.0,
	    const unsigned int & clahe_window = 64,
	    typename T::PixelType clahe_new_min = 0,
	    typename T::PixelType clahe_new_max = 255,
	    const bool & dont_allocate = false)
{
  cout << "    reading " << fn_mosaic;
  
  std::fstream fin;
  fin.open(fn_mosaic, ios::in);
  if (!fin.is_open())
  {
    cout << ", failed...." << endl;
    return false;
  }
  cout << endl;
  
  std::list<the_text_t> fn_tiles;
  std::vector<base_transform_t::Pointer> transform;
  std::vector<typename T::ConstPointer> image;
  std::vector<mask_t::ConstPointer> mask;
  
  double pixel_spacing = 1.0;
  bool use_std_mask = false;
  load_mosaic<base_transform_t>(fin,
				pixel_spacing,
				use_std_mask,
				fn_tiles,
				transform);
  fin.close();
  
  unsigned int num_images = transform.size();
  assert(num_images > 1);
  assert(pixel_spacing != 0);
  
  image.resize(num_images);
  mask.resize(num_images);
  
  unsigned int i = 0;
  for (std::list<the_text_t>::const_iterator iter = fn_tiles.begin();
       iter != fn_tiles.end(); ++iter, i++)
  {
    const the_text_t & fn_image = *iter;
    
    // read the image:
    cout << setw(3) << i << " ";
    image[i] = std_tile<T>(fn_image, shrink_factor, pixel_spacing);
    
    if (use_std_mask)
    {
      mask[i] = std_mask<T>(image[i]);
    }
  }
  
  // assempble the mosaic:
  if (assemble_mosaic)
  {
    cout << "    assembling a mosaic " << endl;
    mosaic = make_mosaic<typename T::ConstPointer, base_transform_t::Pointer>
      (FEATHER_BLEND_E, transform, image, mask, dont_allocate);
    
    // reset the mosaic image origin:
    typename T::PointType origin = mosaic->GetOrigin();
    origin[0] = 0;
    origin[1] = 0;
    mosaic->SetOrigin(origin);
    
    if (flip_mosaic && !dont_allocate)
    {
      mosaic = flip<T>(mosaic);
    }
  }
  
  if (assemble_mosaic_mask)
  {
    cout << "    assembling a mosaic mask " << endl;
    mosaic_mask = make_mask<base_transform_t::Pointer>(transform, mask);
    
    // reset the mosaic image origin:
    mask_t::PointType origin = mosaic_mask->GetOrigin();
    origin[0] = 0;
    origin[1] = 0;
    mosaic_mask->SetOrigin(origin);
    
    if (flip_mosaic)
    {
      mosaic_mask = flip<mask_t>(mosaic_mask);
    }
  }
  
  // preprocess the image with Contrast Limited Adaptive Histogram
  // Equalization algorithm, remap the intensities into the [0, 1] range:
  if (assemble_mosaic && !dont_allocate && clahe_slope > 1.0)
  {
    cout << "    enhancing contrast with CLAHE " << endl;
    mosaic = CLAHE<T>(mosaic,
		      clahe_window,
		      clahe_window,
		      clahe_slope,
		      256,	// number of bins
		      clahe_new_min,
		      clahe_new_max,
		      mosaic_mask);
  }
  
  return true;
}

//----------------------------------------------------------------
// load_mosaic
// 
// Backwards compatibility API
// 
template <typename T>
bool
load_mosaic(const char * fn_mosaic,
	    const bool & flip_mosaic,
	    const unsigned int & shrink_factor,
	    const double & clahe_slope,
	    typename T::Pointer & mosaic,
	    mask_t::Pointer & mosaic_mask,
	    bool mosaic_mask_enabled = true,
	    bool dont_allocate = false,
	    typename T::PixelType clahe_new_min = 0,
	    typename T::PixelType clahe_new_max = 255)
{
  return load_mosaic<T>(fn_mosaic,
			flip_mosaic,
			shrink_factor,
			true,			// assemble mosaic
			mosaic,
			mosaic_mask_enabled,	// assemble mosaic mask
			mosaic_mask,
			clahe_slope,
			64,			// clahe window
			clahe_new_min,
			clahe_new_max,
			dont_allocate);
}

//----------------------------------------------------------------
// load_slice
// 
inline bool
load_slice(const char * fn_mosaic,
	   const bool & flip,
	   const unsigned int & shrink_factor,
	   const double & clahe_slope,
	   image_t::Pointer & mosaic,
	   mask_t::Pointer & mosaic_mask,
	   bool mosaic_mask_enabled = true,
	   bool dont_allocate = false)
{
  return load_mosaic<image_t>(fn_mosaic,
			      flip,
			      shrink_factor,
			      clahe_slope,
			      mosaic,
			      mosaic_mask,
			      mosaic_mask_enabled,
			      dont_allocate,
			      0.0,
			      1.0);
}

//----------------------------------------------------------------
// load_slice
// 
inline static bool
load_slice(const char * fn_mosaic,
	   const bool & flip,
	   const unsigned int & shrink_factor,
	   const double & clahe_slope,
	   image_t::Pointer & mosaic,
	   bool dont_allocate = false)
{
  mask_t::Pointer dummy;
  return load_slice(fn_mosaic,
		    flip,
		    shrink_factor,
		    clahe_slope,
		    mosaic,
		    dummy,
		    false,
		    dont_allocate);
}


#endif // STOS_COMMON_HXX_
