// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : the_grid_transform.cxx
// Author       : Pavel A. Koshevoy
// Created      : Thu Nov 30 13:37:18 MST 2006
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : A discontinuous transform -- a uniform grid of vertices is
//                mapped to an image. At each vertex, in addition to image
//                space coordinates, a second set of coordinates is stored.
//                This is similar to texture mapped OpenGL triangle meshes,
//                where the texture coordinates correspond to the image space
//                vertex coordinates.

// local includes:
#include "itk/the_grid_transform.hxx"


//----------------------------------------------------------------
// GRID_DENSITY
//
static const int GRID_DENSITY = 1;


//----------------------------------------------------------------
// WARPING
//
#ifdef DEBUG_WARPING
bool WARPING = false;
#endif

//----------------------------------------------------------------
// triangle_t::triangle_t
//
triangle_t::triangle_t()
{
  vertex_[0] = ~0;
  vertex_[1] = ~0;
  vertex_[2] = ~0;
}

//----------------------------------------------------------------
// triangle_t::intersect
//
bool
triangle_t::intersect(const vertex_t * v_arr,
		      const pnt2d_t & P,
		      pnt2d_t & uv) const
{
  static const double epsilon = 1e-6;

  double wb = pwb[0] * P[0] + pwb[1] * P[1] + pwb[2];
  double wc = pwc[0] * P[0] + pwc[1] * P[1] + pwc[2];

  // corner cases -- clamp when within tolerance:
  if (wb < 0 && wb > -epsilon) wb = 0;
  if (wb > 1 && (wb - 1) < epsilon) wb = 1;
  if (wc < 0 && wc > -epsilon) wc = 0;
  if (wc > 1 && (wc - 1) < epsilon) wc = 1;

  double wa = 1.0 - (wb + wc);

  const pnt2d_t & A = v_arr[vertex_[0]].uv_;
  const pnt2d_t & B = v_arr[vertex_[1]].uv_;
  const pnt2d_t & C = v_arr[vertex_[2]].uv_;

  uv[0] = A[0] * wa + B[0] * wb + C[0] * wc;
  uv[1] = A[1] * wa + B[1] * wb + C[1] * wc;

#ifdef DEBUG_WARPING
  static const double THRESHOLD = 5e-2;
  if (WARPING && (fabs(wa) < THRESHOLD ||
		  fabs(wb) < THRESHOLD ||
		  fabs(wc) < THRESHOLD))
  {
    uv[0] = 0.0;
    uv[1] = 0.0;
  }
#endif // DEBUG_WARPING

  return (wa >= 0.0 && wb >= 0.0 && wc >= 0.0);
}


//----------------------------------------------------------------
// the_grid_transform_t::the_grid_transform_t
//
the_grid_transform_t::the_grid_transform_t():
  rows_(0),
  cols_(0),
  grid_rows_(0),
  grid_cols_(0)
{
  // reset the grid bounding box:
  grid_min_[0] = std::numeric_limits<double>::max();
  grid_min_[1] = grid_min_[0];
  grid_ext_[0] = 0.0;
  grid_ext_[1] = 0.0;
}

//----------------------------------------------------------------
// the_grid_transform_t::is_ready
// 
bool
the_grid_transform_t::is_ready() const
{
  unsigned int n = rows_ * cols_;
  return (n != 0) && (tri_.size() == 2 * n);
}

//----------------------------------------------------------------
// the_grid_transform_t::xy_cell
//
unsigned int
the_grid_transform_t::xy_cell(const pnt2d_t & xy) const
{
  // find where in the grid the point lands:
  double a = (xy[0] - grid_min_[0]) / grid_ext_[0];
  double b = (xy[1] - grid_min_[1]) / grid_ext_[1];
  if (a >= 0.0 && a <= 1.0 && b >= 0.0 && b <= 1.0)
  {
    double c = std::min(double(grid_cols_ - 1), a * double(grid_cols_));
    double r = std::min(double(grid_rows_ - 1), b * double(grid_rows_));

    unsigned int col = (unsigned int)(floor(c));
    unsigned int row = (unsigned int)(floor(r));

    return row * grid_cols_ + col;
  }

  return ~0;
}

//----------------------------------------------------------------
// the_grid_transform_t::xy_triangle
//
unsigned int
the_grid_transform_t::xy_triangle(const pnt2d_t & xy, pnt2d_t & uv) const
{
  unsigned int cell_id = xy_cell(xy);

  if (cell_id != (unsigned int)(~0))
  {
    // shortcuts:
    const std::list<unsigned int> & cell = grid_[cell_id];
    const vertex_t * v_arr = &(mesh_[0]);

    // check each candidate triangle for intersection:
    for (std::list<unsigned int>::const_iterator iter = cell.begin();
	 iter != cell.end(); ++iter)
    {
      const triangle_t & tri = tri_[*iter];
      if (tri.intersect(v_arr, xy, uv))
      {
	return *iter;
      }
    }
  }

  return ~0;
}

//----------------------------------------------------------------
// the_grid_transform_t::transform
//
bool
the_grid_transform_t::transform(const pnt2d_t & xy, pnt2d_t & uv) const
{
  unsigned int t_id = xy_triangle(xy, uv);

  if (t_id == (unsigned int)(~0))
  {
    uv[0] = std::numeric_limits<double>::quiet_NaN();
    uv[1] = uv[0];
  }

  return t_id != (unsigned int)(~0);
}

//----------------------------------------------------------------
// the_grid_transform_t::transform_inv
//
bool
the_grid_transform_t::transform_inv(const pnt2d_t & uv, pnt2d_t & xy) const
{
  const double & c = uv[0];
  int c0 = int(floor(c * double(cols_)));
  int c1 = c0 + 1;
  if (c0 < 0 || c0 > int(cols_)) return false;
  if (c0 == int(cols_))
  {
    c0--;
    c1--;
  }
  
  const double & r = uv[1];
  int r0 = int(floor(r * double(rows_)));
  int r1 = r0 + 1;
  if (r0 < 0 || r0 > int(rows_)) return false;
  if (r0 == int(rows_))
  {
    r0--;
    r1--;
  }
  
  double w[4];
  {
    double wx[] = {
      double(c1) - c * double(cols_),
      c * double(cols_) - double(c0)
    };

    double wy[] = {
      double(r1) - r * double(rows_),
      r * double(rows_) - double(r0)
    };

    w[0] = wx[0] * wy[0];
    w[1] = wx[1] * wy[0];
    w[2] = wx[1] * wy[1];
    w[3] = wx[0] * wy[1];
  }

  xy[0] =
    vertex(r0, c0).xy_[0] * w[0] +
    vertex(r0, c1).xy_[0] * w[1] +
    vertex(r1, c1).xy_[0] * w[2] +
    vertex(r1, c0).xy_[0] * w[3];

  xy[1] =
    vertex(r0, c0).xy_[1] * w[0] +
    vertex(r0, c1).xy_[1] * w[1] +
    vertex(r1, c1).xy_[1] * w[2] +
    vertex(r1, c0).xy_[1] * w[3];

  return true;
}

//----------------------------------------------------------------
// the_grid_transform_t::jacobian
//
bool
the_grid_transform_t::jacobian(const pnt2d_t & P,
			       unsigned int * idx,
			       double * jac) const
{
  pnt2d_t uv;
  unsigned int t_id = xy_triangle(P, uv);
  if (t_id == (unsigned int)(~0)) return false;

  const triangle_t & tri = tri_[t_id];
  idx[0] = tri.vertex_[0];
  idx[1] = tri.vertex_[1];
  idx[2] = tri.vertex_[2];

  const vertex_t & A = mesh_[idx[0]];
  const vertex_t & B = mesh_[idx[1]];
  const vertex_t & C = mesh_[idx[2]];

  // calculate partial derivatrives of wB and wC with respect to A, B, C:
  double dw[2][6];
  {
    const double & Ax = A.xy_[0];
    const double & Ay = A.xy_[1];
    const double & Bx = B.xy_[0];
    const double & By = B.xy_[1];
    const double & Cx = C.xy_[0];
    const double & Cy = C.xy_[1];
    const double & Px = P[0];
    const double & Py = P[1];

    double bx = Bx - Ax;
    double cx = Cx - Ax;
    double px = Px - Ax;

    double by = By - Ay;
    double cy = Cy - Ay;
    double py = Py - Ay;

    double pycx_pxcy = py * cx - px * cy;
    double pxby_pybx = px * by - py * bx;
    double bycx_bxcy = by * cx - bx * cy;
    double bycx_bxcy_2 = bycx_bxcy * bycx_bxcy;

    // dwB/dAx, dwB/dAy:
    dw[0][0] = (Cy - Py) / bycx_bxcy - (Cy - By) / bycx_bxcy_2;
    dw[0][1] = (Px - Cx) / bycx_bxcy - (Bx - Cx) / bycx_bxcy_2;

    // dwB/dBx, dwB/dBy:
    dw[0][2] = cy * pycx_pxcy / bycx_bxcy_2;
    dw[0][3] = -cx * pycx_pxcy / bycx_bxcy_2;

    // dwB/dCx, dwB/dCy:
    dw[0][4] = py / bycx_bxcy - by * pycx_pxcy / bycx_bxcy_2;
    dw[0][5] = bx * pycx_pxcy / bycx_bxcy_2 - px / bycx_bxcy;

    // dwC/dAx, dwC/dAy:
    dw[1][0] = (Py - By) / bycx_bxcy - (Cy - By) * pxby_pybx / bycx_bxcy_2;
    dw[1][1] = (Bx - Px) / bycx_bxcy - (Bx - Cx) * pxby_pybx / bycx_bxcy_2;

    // dwC/dBx, dwC/dBy:
    dw[1][2] = cy * pxby_pybx / bycx_bxcy_2 - py / bycx_bxcy;
    dw[1][3] = px / bycx_bxcy - cx * pxby_pybx / bycx_bxcy_2;

    // dwC/dCx, dwC/dCy:
    dw[1][4] = -by * pxby_pybx / bycx_bxcy_2;
    dw[1][5] = bx * pxby_pybx / bycx_bxcy_2;
  }

  double bu = B.uv_[0] - A.uv_[0];
  double bv = B.uv_[1] - A.uv_[1];
  double cu = C.uv_[0] - A.uv_[0];
  double cv = C.uv_[1] - A.uv_[1];

  // shortcut:
  double * du = &(jac[0]);
  double * dv = &(jac[6]);

  for (unsigned int i = 0; i < 6; i++)
  {
    du[i] = bu * dw[0][i] + cu * dw[1][i];
    dv[i] = bv * dw[0][i] + cv * dw[1][i];
  }

  return true;
}

//----------------------------------------------------------------
// the_grid_transform_t::setup
//
void
the_grid_transform_t::setup(unsigned int rows,
			    unsigned int cols,
			    const pnt2d_t & tile_min,
			    const pnt2d_t & tile_max,
			    const fast_array_t<pnt2d_t> & xy)
{
  rows_ = rows;
  cols_ = cols;

  tile_min_ = tile_min;
  tile_ext_ = tile_max - tile_min;

  mesh_.resize((rows_ + 1) * (cols_ + 1));

  // temporaries:
  pnt2d_t uv;

  for (unsigned int row = 0; row <= rows_; row++)
  {
    uv[1] = double(row) / double(rows_);
    for (unsigned int col = 0; col <= cols_; col++)
    {
      uv[0] = double(col) / double(cols_);

      vertex_t & vx = vertex(row, col);
      vx.uv_ = uv;
      vx.xy_ = xy[row * (cols_ + 1) + col];
    }
  }

  setup_mesh();
}

//----------------------------------------------------------------
// the_grid_transform_t::update
//
void
the_grid_transform_t::update(const vec2d_t * xy_shift)
{
  unsigned int num_verts = mesh_.size();
  vertex_t * v_arr = &(mesh_[0]);
  for (unsigned int i = 0; i < num_verts; i++)
  {
    // cerr << v_arr[i].xy_ << " -> ";
    v_arr[i].xy_ += xy_shift[i];
    // cerr << v_arr[i].xy_ << endl;
  }

  rebuild_grid();
}

//----------------------------------------------------------------
// the_grid_transform_t::shift
// 
void
the_grid_transform_t::shift(const vec2d_t & xy_shift)
{
  unsigned int num_verts = mesh_.size();
  vertex_t * v_arr = &(mesh_[0]);
  for (unsigned int i = 0; i < num_verts; i++)
  {
    // cerr << v_arr[i].xy_ << " -> ";
    v_arr[i].xy_ += xy_shift;
    // cerr << v_arr[i].xy_ << endl;
  }

  rebuild_grid();
}

//----------------------------------------------------------------
// the_grid_transform_t::rebuild_grid
//
void
the_grid_transform_t::rebuild_grid()
{
  // reset the grid bounding box and destroy the the old grid:
  grid_min_[0] = std::numeric_limits<double>::max();
  grid_min_[1] = grid_min_[0];

  pnt2d_t grid_max;
  grid_max[0] = -grid_min_[0];
  grid_max[1] = -grid_min_[1];

  // find the new grid bounding box:
  unsigned int num_verts = mesh_.size();
  vertex_t * v_arr = &(mesh_[0]);
  for (unsigned int i = 0; i < num_verts; i++)
  {
    update_bbox(grid_min_, grid_max, v_arr[i].xy_);
  }
  grid_ext_ = grid_max - grid_min_;

  /*
  cerr << "grid_min: " << grid_min_ << endl
       << "grid_max: " << grid_max << endl
       << "grid_ext: " << grid_ext_ << endl;
  */

  // add triangles to the grid:
  grid_rows_ = rows_ * GRID_DENSITY;
  grid_cols_ = cols_ * GRID_DENSITY;
  grid_.resize(grid_rows_ * grid_cols_);
  unsigned int num_triangles = tri_.size();
  for (unsigned int i = 0; i < num_triangles; i++)
  {
    update_grid(i);
  }

#if 0
  for (unsigned int row = 0; row < grid_rows_; row++)
  {
    for (unsigned int col = 0; col < grid_cols_; col++)
    {
      cout << setw(3) << grid_[row * grid_cols_ + col].size();
    }
    cout << endl;
  }
#endif
}

//----------------------------------------------------------------
// intersect_lines_2d
//
// Find the intersection parameters (u, v)  of 2 parametric lines
//   A = a + b * u
//   B = c + d * v
//
static bool
intersect_lines_2d(// start point A:
		   double ax,
		   double ay,

		   // direction vector A:
		   double bx,
		   double by,

		   // start point B:
		   double cx,
		   double cy,

		   // direction vector B:
		   double dx,
		   double dy,

		   // results:
		   double & u,
		   double & v)
{
  // solve the linear system Ax = b:
  //
  // [bx  -dx] [u] = [cx - ax]
  // [by  -dy] [v]   [cy - ay]
  //

  double det_A = dx * by - bx * dy;
  if (det_A == 0.0) return false;

  double det_A_inv = 1.0 / det_A;
  double A_inv[][2] = {
    { -dy * det_A_inv, dx * det_A_inv },
    { -by * det_A_inv, bx * det_A_inv }
  };

  double b[] = { cx - ax, cy - ay };
  u = A_inv[0][0] * b[0] + A_inv[0][1] * b[1];
  v = A_inv[1][0] * b[0] + A_inv[1][1] * b[1];

  return true;
}

//----------------------------------------------------------------
// intersect_bbox_triangle
//
static bool
intersect_bbox_triangle(const pnt2d_t & min,
			const pnt2d_t & max,
			const pnt2d_t & A,
			const pnt2d_t & B,
			const pnt2d_t & C)
{
  double lines_a[][4] = {
    // horizontal:
    { min[0], min[1], max[0] - min[0], 0.0 },
    { min[0], max[1], max[0] - min[0], 0.0 },
    // vertical:
    { min[0], min[1], 0.0, max[1] - min[1] },
    { max[0], min[1], 0.0, max[1] - min[1] }
  };

  double lines_b[][4] = {
    { A[0], A[1], B[0] - A[0], B[1] - A[1] },
    { A[0], A[1], C[0] - A[0], C[1] - A[1] },
    { B[0], B[1], C[0] - B[0], C[1] - B[1] }
  };

  double u;
  double v;
  for (unsigned int i = 0; i < 4; i++)
  {
    for (unsigned int j = 0; j < 3; j++)
    {
      if (intersect_lines_2d(lines_a[i][0],
			     lines_a[i][1],
			     lines_a[i][2],
			     lines_a[i][3],

			     lines_b[j][0],
			     lines_b[j][1],
			     lines_b[j][2],
			     lines_b[j][3],

			     u,
			     v))
      {
	if (u >= 0.0 && u <= 1.0 && v >= 0.0 && v <= 1.0)
	{
	  return true;
	}
      }
    }
  }

  return false;
}

//----------------------------------------------------------------
// the_grid_transform_t::update_grid
//
void
the_grid_transform_t::update_grid(unsigned int t_idx)
{
  // shortcuts:
  triangle_t & tri = tri_[t_idx];
  const vertex_t * v_arr = &(mesh_[0]);

  const vertex_t & v0 = mesh_[tri.vertex_[0]];
  const vertex_t & v1 = mesh_[tri.vertex_[1]];
  const vertex_t & v2 = mesh_[tri.vertex_[2]];

  const pnt2d_t & A = v0.xy_;
  const pnt2d_t & B = v1.xy_;
  const pnt2d_t & C = v2.xy_;

  // precompute the barycentric calculation coefficients:
  {
    vec2d_t b = B - A;
    vec2d_t c = C - A;
    double bycx_bxcy = b[1] * c[0] - b[0] * c[1];

    tri.pwb[0] = -c[1] / bycx_bxcy;
    tri.pwb[1] = c[0] / bycx_bxcy;
    tri.pwb[2] = (c[1] * A[0] - c[0] * A[1]) / bycx_bxcy;

    tri.pwc[0] = b[1] / bycx_bxcy;
    tri.pwc[1] = -b[0] / bycx_bxcy;
    tri.pwc[2] = (b[0] * A[1] - b[1] * A[0]) / bycx_bxcy;
  }

  pnt2d_t min = A;
  pnt2d_t max = min;
  update_bbox(min, max, B);
  update_bbox(min, max, C);

  const double & gw = grid_ext_[0];
  double a[] = {
    (min[0] - grid_min_[0]) / gw,
    (max[0] - grid_min_[0]) / gw
  };

  unsigned int c[] = {
    std::min(grid_cols_ - 1, (unsigned int)(floor(a[0] * double(grid_cols_)))),
    std::min(grid_cols_ - 1, (unsigned int)(floor(a[1] * double(grid_cols_))))
  };

  const double & gh = grid_ext_[1];
  double b[] = {
    (min[1] - grid_min_[1]) / gh,
    (max[1] - grid_min_[1]) / gh
  };

  unsigned int r[] = {
    std::min(grid_rows_ - 1, (unsigned int)(floor(b[0] * double(grid_rows_)))),
    std::min(grid_rows_ - 1, (unsigned int)(floor(b[1] * double(grid_rows_))))
  };

  // FIXME:
  for (unsigned int row = r[0]; row <= r[1]; row++)
  {
    for (unsigned int col = c[0]; col <= c[1]; col++)
    {
      unsigned int i = row * grid_cols_ + col;
      std::list<unsigned int> & cell = grid_[i];

#if 1
      // calculate the bounding box of this grid cell:
      pnt2d_t min(grid_min_);
      min[0] += gw * double(col) / double(grid_cols_);
      min[1] += gh * double(row) / double(grid_rows_);

      pnt2d_t max(grid_min_);
      max[0] += gw * double(col + 1) / double(grid_cols_);
      max[1] += gh * double(row + 1) / double(grid_rows_);

      pnt2d_t uv;
      if (inside_bbox(min, max, A) ||
	  inside_bbox(min, max, B) ||
	  inside_bbox(min, max, C) ||
	  tri.intersect(v_arr, pnt2d(min[0], min[1]), uv) ||
	  tri.intersect(v_arr, pnt2d(max[0], min[1]), uv) ||
	  tri.intersect(v_arr, pnt2d(max[0], max[1]), uv) ||
	  tri.intersect(v_arr, pnt2d(min[0], max[1]), uv) ||
	  intersect_bbox_triangle(min, max, A, B, C))
      {
	cell.push_back(t_idx);
      }
#else
      cell.push_back(t_idx);
#endif
    }
  }
}

//----------------------------------------------------------------
// the_grid_transform_t::setup_mesh
//
void
the_grid_transform_t::setup_mesh()
{
  tri_.resize(rows_ * cols_ * 2);

  // temporaries:
  unsigned int v_idx[4];
  unsigned int t_idx[2];

  for (unsigned int row = 0; row < rows_; row++)
  {
    for (unsigned int col = 0; col < cols_; col++)
    {
      // vertex indices:
      v_idx[0] = (cols_ + 1) * row + col;
      v_idx[1] = v_idx[0] + (cols_ + 1);
      v_idx[2] = v_idx[1] + 1;
      v_idx[3] = v_idx[0] + 1;

      // triangle indices:
      t_idx[0] = (cols_ * row + col) * 2;
      t_idx[1] = t_idx[0] + 1;

      // setup triangle A:
      triangle_t & tri_a = tri_[t_idx[0]];
      tri_a.vertex_[0] = v_idx[0];
      tri_a.vertex_[1] = v_idx[1];
      tri_a.vertex_[2] = v_idx[2];

      // setup triangle B:
      triangle_t & tri_b = tri_[t_idx[1]];
      tri_b.vertex_[0] = v_idx[0];
      tri_b.vertex_[1] = v_idx[2];
      tri_b.vertex_[2] = v_idx[3];
    }
  }

  // initialize the acceleration structure:
  rebuild_grid();
}
