// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : the_grid_transform.hxx
// Author       : Pavel A. Koshevoy
// Created      : Thu Nov 30 10:45:47 MST 2006
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : A discontinuous transform -- a uniform grid of vertices is
//                mapped to an image. At each vertex, in addition to image
//                space coordinates, a second set of coordinates is stored.
//                This is similar to texture mapped OpenGL triangle meshes,
//                where the texture coordinates correspond to the image space
//                vertex coordinates.

#ifndef THE_GRID_TRANSFORM_HXX_
#define THE_GRID_TRANSFORM_HXX_

// local includes:
#include "itk/common.hxx"


//----------------------------------------------------------------
// fast_array_t
//
template <typename T>
class fast_array_t
{
public:
  fast_array_t(size_t size = 0, const T & fill = T()):
    data_(NULL),
    size_(size)
  {
    if (size_ > 0)
    {
      data_ = new T [size_];
      for (unsigned int i = 0; i < size_; i++)
      {
	data_[i] = fill;
      }
    }
  }
  
  fast_array_t(const fast_array_t<T> & a):
    data_(NULL),
    size_(0)
  { *this = a; }
  
  ~fast_array_t()
  { clear(); }
  
  // clear the array and resize it to be zero-length:
  void clear()
  {
    delete [] data_;
    data_ = NULL;
    size_ = 0;
  }
  
  // deep copy via assignment operator:
  fast_array_t<T> & operator = (const fast_array_t<T> & a)
  {
    if (this != &a)
    {
      delete [] data_;
      data_ = NULL;
      size_ = 0;
      
      if (a.size_ > 0)
      {
	data_ = new T [a.size_];
	size_ = a.size_;
	
	for (size_t i = 0; i < size_; i++)
	{
	  data_[i] = a.data_[i];
	}
      }
    }
    
    return *this;
  }
  
  inline void resize(size_t new_size)
  {
    delete [] data_;
    data_ = NULL;
    size_ = new_size;
    
    if (new_size > 0)
    {
      data_ = new T [size_];
    }
  }
  
  inline void assign(size_t new_size, const T & data)
  {
    resize(new_size);
    for (unsigned int i = 0; i < size_; i++)
    {
      data_[i] = data;
    }
  }
  
  // accessors:
  inline const T & operator [] (size_t i) const
  { return data_[i]; }
  
  inline T & operator [] (size_t i)
  { return data_[i]; }
  
  inline size_t size() const
  { return size_; }
  
  // data:
  T * data_;
  size_t size_;
};

//----------------------------------------------------------------
// vertex_t
// 
class vertex_t
{
public:
  pnt2d_t xy_;
  pnt2d_t uv_;
};

//----------------------------------------------------------------
// triangle_t
// 
class triangle_t
{
public:
  triangle_t();
  
  bool intersect(const vertex_t * v_arr,
		 const pnt2d_t & xy,
		 pnt2d_t & uv) const;
  
  // triangle vertex indices, counterclockwise winding:
  unsigned int vertex_[3];
  
  // precomputed fast barycentric coordinate calculation coefficientse:
  double pwb[3];
  double pwc[3];
};

//----------------------------------------------------------------
// the_grid_transform_t
// 
class the_grid_transform_t
{
public:
  the_grid_transform_t();
  
  // check to see whether the transform has already been setup:
  bool is_ready() const;
  
  // vertex accessors:
  inline const vertex_t & vertex(size_t row, size_t col) const
  { return mesh_[row * (cols_ + 1) + col]; }
  
  inline vertex_t & vertex(size_t row, size_t col)
  { return mesh_[row * (cols_ + 1) + col]; }
  
  // find the grid cell containing a given xy point:
  unsigned int xy_cell(const pnt2d_t & xy) const;
  
  // find the triangle containing a given xy point and the barycentric
  // coordinates of the point within that triangle:
  unsigned int xy_triangle(const pnt2d_t & xy, pnt2d_t & uv) const;
  
  // transform the point:
  bool transform(const pnt2d_t & xy, pnt2d_t & uv) const;
  
  // inverse transform the point:
  bool transform_inv(const pnt2d_t & uv, pnt2d_t & xy) const;
  
  // calculate the derivatives of the transforms with respect to
  // transform parameters:
  bool jacobian(const pnt2d_t & xy, unsigned int * idx, double * jac) const;
  
  // setup the transform:
  void setup(unsigned int rows,
	     unsigned int cols,
	     const pnt2d_t & tile_min,
	     const pnt2d_t & tile_max,
	     const fast_array_t<pnt2d_t> & xy);
  
  // update the vertex xy coordinates and rebuild the grid
  void update(const vec2d_t * xy_shift);
  void shift(const vec2d_t & xy_shift);
  
  // rebuild the acceleration grid:
  void rebuild_grid();
  
private:
  // helper used to rebuild the grid:
  void update_grid(unsigned int t_idx);
  
  // helper used to setup the triangle mesh:
  void setup_mesh();
  
public:
  // the deformable mesh vertices:
  fast_array_t<vertex_t> mesh_;
  size_t rows_;
  size_t cols_;
  
  // the tile bounding box:
  pnt2d_t tile_min_;
  vec2d_t tile_ext_;
  
  // the acceleration structure:
  fast_array_t<std::list<unsigned int> > grid_;
  unsigned int grid_rows_;
  unsigned int grid_cols_;
  
  // the grid bounding box:
  pnt2d_t grid_min_;
  vec2d_t grid_ext_;
  
  // the triangle mesh:
  fast_array_t<triangle_t> tri_;
};


#endif // THE_GRID_TRANSFORM_HXX_
