// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : threshold.hxx
// Author       : Pavel A. Koshevoy
// Created      : 2006/04/05 12:36
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Helper functions for thresholding 2D image to isolate
//                peaks/clusters in the data.

#ifndef THRESHOLD_HXX_
#define THRESHOLD_HXX_

// local includes:
#include "itk/common.hxx"


//----------------------------------------------------------------
// threshold_by_intensity_inplace
// 
// The image is thresholded, intensities below/above threshold are set
// to a background value, specified by bg_offset.
// 
// Depending on whether the image is being thresholded above or below
// the threshold, the background value is calculated as:
// 
//   a. keep pixels above threshold,
//      bg = clip - bg_offset * (max - clip)
// 
//   b. keep pixels below threshold,
//      bg = clip + bg_offset * (clip_max - min)
// 
// The function returns the background value of the thresholded image.
// 
template <class image_t>
double
threshold_by_intensity_inplace(typename image_t::Pointer & image,
			       const double clip_param,
			       const bool keep_pixels_above_threshold = true,
			       const unsigned int bins = 256,
			       const double bg_offset = 1e-3)
{
  typedef itk::ImageRegionConstIterator<image_t> iter_t;
  typedef typename image_t::IndexType index_t;
  
  // first find minima/maxima of the image:
  double v_min = std::numeric_limits<double>::max();
  double v_max = -v_min;
  
  iter_t iter(image, image->GetLargestPossibleRegion());
  for (iter.GoToBegin(); !iter.IsAtEnd(); ++iter)
  {
    const double & v = iter.Get();
    v_min = std::min(v_min, v);
    v_max = std::max(v_max, v);
  }
  
  // calculate the min/max range:
  const double v_rng = v_max - v_min;
  
  if (v_rng == 0.0)
  {
    // there are no peaks in this image:
    return v_min;
  }
  
  double clip = v_min + clip_param * v_rng;
  
  // threshold the peaks:
  double background =
    keep_pixels_above_threshold
    ? clip - bg_offset * v_rng
    : clip + bg_offset * v_rng;
  
  if (keep_pixels_above_threshold)
  {
    image = threshold<image_t>(image, clip, v_max, background, v_max);
  }
  else
  {
    image = threshold<image_t>(image, v_min, clip, v_min, background);
  }
  
  return background;
}

//----------------------------------------------------------------
// threshold_by_intensity
// 
template <class image_t>
typename image_t::Pointer
threshold_by_intensity(const image_t * image,
		       const double clip_param,
		       const bool keep_pixels_above_threshold = true,
		       const unsigned int bins = 256,
		       const double bg_offset = 1e-3)
{
  typename image_t::Pointer peaks = cast<image_t, image_t>(image);
  threshold_by_intensity_inplace(peaks,
				 clip_param,
				 keep_pixels_above_threshold,
				 bins,
				 bg_offset);
  return peaks;
}

//----------------------------------------------------------------
// threshold_by_area_inplace
// 
// The clip_param below refers to the number of pixels that fall
// below/above the threshold. Thus, the number of pixels above
// the maxima is 1 - clip_param.
// 
// Thus, it is possible to specify a threshold value without
// knowing anything about the image.
// 
// The image is thresholded, intensities below/above threshold are set
// to a background value, specified by bg_offset.
// 
// Depending on whether the image is being thresholded above or below
// the threshold, the background value is calculated as:
// 
//   a. keep pixels above threshold,
//      bg = clip - bg_offset * (max - clip)
// 
//   b. keep pixels below threshold,
//      bg = clip + bg_offset * (clip_max - min)
// 
// The function returns the background value of the thresholded image
// 
template <class image_t>
double
threshold_by_area_inplace(typename image_t::Pointer & image,
			  const double clip_param,
			  const bool keep_pixels_above_threshold = true,
			  const unsigned int bins = 4096,
			  const double bg_offset = 1e-3)
{
  typedef itk::ImageRegionConstIterator<image_t> iter_t;
  typedef typename image_t::IndexType index_t;
  
  // first find minima/maxima of the image:
  double v_min = std::numeric_limits<double>::max();
  double v_max = -v_min;
  
  iter_t iter(image, image->GetLargestPossibleRegion());
  for (iter.GoToBegin(); !iter.IsAtEnd(); ++iter)
  {
    const double v = iter.Get();
    v_min = std::min(v_min, v);
    v_max = std::max(v_max, v);
  }
  
  // calculate the min/max range:
  const double v_rng = v_max - v_min;
  
  if (v_rng == 0.0)
  {
    // there are no peaks in this image:
    return v_min;
  }
  
  // build a histogram:
  std::vector<unsigned int> pdf;
  pdf.assign(bins, 0);
  
  for (iter.GoToBegin(); !iter.IsAtEnd(); ++iter)
  {
    const double v = iter.Get();
    const unsigned int bin =
      (unsigned int)(double((v - v_min) / v_rng) * double(bins - 1));
    pdf[bin]++;
  }
  
  // build the cumulative histogram:
  std::vector<unsigned int> cdf(bins);
  cdf[0] = pdf[0];
  for (unsigned int i = 1; i < bins; i++)
  {
    cdf[i] = cdf[i - 1] + pdf[i];
  }
  
  // calculate the total number of pixels in the image:
  typename image_t::SizeType size =
    image->GetLargestPossibleRegion().GetSize();
  
  double total_number_of_pixels = 1.0;
  for (unsigned int i = 0; i < size.GetSizeDimension(); i++)
  {
    total_number_of_pixels *= size[i];
  }
  
  // find the CDF bin that contains a given percentage of the total image:
  double clip = 0.0;
  for (unsigned int i = 1; i < bins; i++)
  {
    clip = v_min + (double(i) / double(bins - 1)) * v_rng;
    if (double(cdf[i]) >= clip_param * total_number_of_pixels)
    {
      break;
    }
  }
  
  // threshold the peaks:
  double background =
    keep_pixels_above_threshold
    ? clip - bg_offset * v_rng
    : clip + bg_offset * v_rng;
  
  if (keep_pixels_above_threshold)
  {
    image = threshold<image_t>(image, clip, v_max, background, v_max);
  }
  else
  {
    image = threshold<image_t>(image, v_min, clip, v_min, background);
  }
  
  return background;
}

//----------------------------------------------------------------
// threshold_by_area
// 
template <class image_t>
typename image_t::Pointer
threshold_by_area(const image_t * image,
		  const double clip_param,
		  const bool keep_pixels_above_threshold = true,
		  const unsigned int bins = 256,
		  const double bg_offset = 1e-3)
{
  typename image_t::Pointer peaks = cast<image_t, image_t>(image);
  threshold_by_area_inplace(peaks,
			    clip_param,
			    keep_pixels_above_threshold,
			    bins,
			    bg_offset);
  return peaks;
}


#endif // THRESHOLD_HXX_
