// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : the_linear_algebra.hxx
// Author       : Pavel A. Koshevoy
// Created      : Sat Oct  2 12:12:30 MDT 2004
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Common linear algebra functions.

#ifndef THE_LINEAR_ALGEBRA_HXX_
#define THE_LINEAR_ALGEBRA_HXX_


//----------------------------------------------------------------
// the_determinant_2x2
// 
// Calculate the determinant of a 2x2 matrix.
//
#define the_determinant_2x2( a11, a12, \
                             a21, a22 ) \
  ( a11 * a22 - a12 * a21 )


//----------------------------------------------------------------
// the_cramers_rule_2x2
// 
// Solve a 2x2 linear system using the Cramer's Rule.
// 
template <class T>
bool
the_cramers_rule_2x2(const T & a11, const T & a12,
		     const T & a21, const T & a22,
		     const T & b1,
		     const T & b2,
		     T & x1,
		     T & x2,
		     T determinant_tolerance = 1e-6)
{
  T d = the_determinant_2x2(a11, a12,
			    a21, a22);
  if (fabs(d) <= determinant_tolerance || d != d) return false;
  
  T d1 = the_determinant_2x2(b1, a12,
			     b2, a22);
  
  T d2 = the_determinant_2x2(a11, b1,
			     a21, b2);
  
  x1 = d1 / d;
  x2 = d2 / d;
  
  return true;
}


//----------------------------------------------------------------
// the_determinant_3x3
// 
// Calculate the determinant of a 3x3 matrix.
//
#define the_determinant_3x3( a11, a12, a13, \
                             a21, a22, a23, \
			     a31, a32, a33 )\
  ( a11 * a22 * a33 - a11 * a32 * a23 + \
    a21 * a32 * a13 - a21 * a12 * a33 + \
    a31 * a12 * a23 - a31 * a22 * a13 )


// shorthand, undefined at the end of the file:
#define a11 A[0]
#define a12 A[1]
#define a13 A[2]
#define a21 A[3]
#define a22 A[4]
#define a23 A[5]
#define a31 A[6]
#define a32 A[7]
#define a33 A[8]
#define x1  x[0]
#define x2  x[1]
#define x3  x[2]
#define b1  b[0]
#define b2  b[1]
#define b3  b[2]

//----------------------------------------------------------------
// the_cramers_rule_3x3
// 
// Solve a 3x3 linear system using the Cramer's Rule.
// 
template <class T>
bool
the_cramers_rule_3x3(const T * A,
		     const T * b,
		     T * x,
		     T determinant_tolerance = 1e-6)
{
  T d = the_determinant_3x3(a11, a12, a13,
			    a21, a22, a23,
			    a31, a32, a33);
  if (fabs(d) <= determinant_tolerance || d != d) return false;
  
  T d1 = the_determinant_3x3(b1, a12, a13,
			     b2, a22, a23,
			     b3, a32, a33);
  
  T d2 = the_determinant_3x3(a11, b1, a13,
			     a21, b2, a23,
			     a31, b3, a33);
  
  T d3 = the_determinant_3x3(a11, a12, b1,
			     a21, a22, b2,
			     a31, a32, b3);
  
  x1 = d1 / d;
  x2 = d2 / d;
  x3 = d3 / d;
  
  return true;
}

// undefine shorthand:
#undef a11
#undef a12
#undef a13
#undef a21
#undef a22
#undef a23
#undef a31
#undef a32
#undef a33
#undef x1
#undef x2
#undef x3
#undef b1
#undef b2
#undef b3


#endif // THE_LINEAR_ALGEBRA_HXX_
