// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : the_trail.hxx
// Author       : Pavel A. Koshevoy
// Created      : 2001/06/24 16:47:24
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : event trail recoring/playback abstract interface,
//                used for regression testing and debugging.

#ifndef THE_TRAIL_HXX_
#define THE_TRAIL_HXX_

// local includes:
#include "eh/the_mouse_device.hxx"
#include "eh/the_keybd_device.hxx"
#include "eh/the_wacom_device.hxx"
#include "utils/the_bit_tree.hxx"

// system includes:
#include <string>
#include <fstream>
#include <iostream>


//----------------------------------------------------------------
// THE_TRAIL
// 
#define THE_TRAIL (*the_trail_t::trail_)


//----------------------------------------------------------------
// the_trail_t
//
class the_trail_t
{
public:
  the_trail_t(int & argc, char ** argv, bool record_by_default = false);
  virtual ~the_trail_t();
  
  // accessors to the input devices: 
  inline const the_mouse_t & mouse() const { return mouse_; }
  inline const the_keybd_t & keybd() const { return keybd_; }
  inline const the_wacom_t & wacom() const { return wacom_; }
  
  inline the_mouse_t & mouse() { return mouse_; }
  inline the_keybd_t & keybd() { return keybd_; }
  inline the_wacom_t & wacom() { return wacom_; }
  
  // a timer will call this periodically to let the trail know to load
  // the next event from the trail:
  virtual void timeout() {}
  
protected:
  // default implementations are no-op:
  virtual void replay() {}
  virtual void replay_one() {}
  
  // close the replay stream:
  virtual void replay_done();
  
  // milestone accessor:
  void next_milestone_achieved();
  
  // dialog bypass handling:
  virtual bool bypass_prolog(const char * /* name */) { return false; }
  virtual void bypass_epilog() {}
  
public:
  // stop the event trail replay:
  virtual void stop()
  {
    replay_done();
  }
  
  //----------------------------------------------------------------
  // milestone_t
  // 
  class milestone_t
  {
  public:
    milestone_t()
    {
      THE_TRAIL.next_milestone_achieved();
    }
    
    ~milestone_t()
    {
      THE_TRAIL.next_milestone_achieved();
    }
  };
  
  //----------------------------------------------------------------
  // bypass_t
  // 
  class bypass_t
  {
  public:
    bypass_t(const char * name):
      ok_(false)
    {
      ok_ = THE_TRAIL.bypass_prolog(name);
    }
    
    virtual ~bypass_t()
    {
      THE_TRAIL.bypass_epilog();
    }
    
    bool ok_;
  };
  
  // check whether trail records is enabled:
  inline bool is_recording() const
  { return record_stream.rdbuf()->is_open(); }
  
  // check whether trail playback is enabled:
  inline bool is_replaying() const
  { return replay_stream.rdbuf()->is_open(); }
  
  // This flag controls whether the event recording will be done
  // even when the user didn't ask for it with the -record switch.
  // It may be usefull for debugging purposes
  bool record_by_default_;
  
  // The in/out stream used for trail replay/record:
  std::ifstream replay_stream;
  std::ofstream record_stream;
  
  // the input devices (reflect the current state of the device):
  the_mouse_t mouse_;
  the_keybd_t keybd_;
  the_wacom_t wacom_;
  
  // the trail line number that is currently being read/executed:
  unsigned int line_num_;
  
  // the current application milestone marker:
  unsigned int milestone_;
  
  // this flag controls the trail playback interactivity:
  bool single_step_replay_;
  bool ask_the_user_;
  
  // these flags will be set to true periodically:
  bool dont_load_events_;
  bool dont_save_events_;
  bool dont_post_events_;
  
  // maximum number of seconds the trail replay engine should wait for
  // a matching milestone marker before declaring a trail out of sequence;
  // default wait time indefinite (max unsigned int value):
  unsigned int seconds_to_wait_;
  
  // bypass names used to synchronize trail reading and
  // application execution:
  std::string replay_bypass_name_;
  std::string record_bypass_name_;
  
  // A single instance of the trail object:
  static the_trail_t * trail_;
};


//----------------------------------------------------------------
// load_address
// 
extern bool load_address(std::istream & si, uint64_t & address);

//----------------------------------------------------------------
// save_address
// 
extern void save_address(std::ostream & so, uint64_t address);

//----------------------------------------------------------------
// save_address
// 
extern void save_address(std::ostream & so, const void * address);

//----------------------------------------------------------------
// operator >>
// 
extern istream & operator >> (std::istream & si, uint64_t & address);

//----------------------------------------------------------------
// operator <<
// 
extern ostream & operator << (std::ostream & so, const uint64_t & address);


//----------------------------------------------------------------
// encode_special_chars
// 
extern const std::string
encode_special_chars(const std::string & text_plain,
		     const char * special_chars = "");

//----------------------------------------------------------------
// decode_special_chars
// 
extern const std::string
decode_special_chars(const std::string & text_encoded);


#endif // THE_TRAIL_HXX_
